#define _CFFI_

/* We try to define Py_LIMITED_API before including Python.h.

   Mess: we can only define it if Py_DEBUG, Py_TRACE_REFS and
   Py_REF_DEBUG are not defined.  This is a best-effort approximation:
   we can learn about Py_DEBUG from pyconfig.h, but it is unclear if
   the same works for the other two macros.  Py_DEBUG implies them,
   but not the other way around.
*/
#ifndef _CFFI_USE_EMBEDDING
#  include <pyconfig.h>
#  if !defined(Py_DEBUG) && !defined(Py_TRACE_REFS) && !defined(Py_REF_DEBUG)
#    define Py_LIMITED_API
#  endif
#endif

#include <Python.h>
#ifdef __cplusplus
extern "C" {
#endif
#include <stddef.h>

/* This part is from file 'cffi/parse_c_type.h'.  It is copied at the
   beginning of C sources generated by CFFI's ffi.set_source(). */

typedef void *_cffi_opcode_t;

#define _CFFI_OP(opcode, arg)   (_cffi_opcode_t)(opcode | (((uintptr_t)(arg)) << 8))
#define _CFFI_GETOP(cffi_opcode)    ((unsigned char)(uintptr_t)cffi_opcode)
#define _CFFI_GETARG(cffi_opcode)   (((intptr_t)cffi_opcode) >> 8)

#define _CFFI_OP_PRIMITIVE       1
#define _CFFI_OP_POINTER         3
#define _CFFI_OP_ARRAY           5
#define _CFFI_OP_OPEN_ARRAY      7
#define _CFFI_OP_STRUCT_UNION    9
#define _CFFI_OP_ENUM           11
#define _CFFI_OP_FUNCTION       13
#define _CFFI_OP_FUNCTION_END   15
#define _CFFI_OP_NOOP           17
#define _CFFI_OP_BITFIELD       19
#define _CFFI_OP_TYPENAME       21
#define _CFFI_OP_CPYTHON_BLTN_V 23   // varargs
#define _CFFI_OP_CPYTHON_BLTN_N 25   // noargs
#define _CFFI_OP_CPYTHON_BLTN_O 27   // O  (i.e. a single arg)
#define _CFFI_OP_CONSTANT       29
#define _CFFI_OP_CONSTANT_INT   31
#define _CFFI_OP_GLOBAL_VAR     33
#define _CFFI_OP_DLOPEN_FUNC    35
#define _CFFI_OP_DLOPEN_CONST   37
#define _CFFI_OP_GLOBAL_VAR_F   39
#define _CFFI_OP_EXTERN_PYTHON  41

#define _CFFI_PRIM_VOID          0
#define _CFFI_PRIM_BOOL          1
#define _CFFI_PRIM_CHAR          2
#define _CFFI_PRIM_SCHAR         3
#define _CFFI_PRIM_UCHAR         4
#define _CFFI_PRIM_SHORT         5
#define _CFFI_PRIM_USHORT        6
#define _CFFI_PRIM_INT           7
#define _CFFI_PRIM_UINT          8
#define _CFFI_PRIM_LONG          9
#define _CFFI_PRIM_ULONG        10
#define _CFFI_PRIM_LONGLONG     11
#define _CFFI_PRIM_ULONGLONG    12
#define _CFFI_PRIM_FLOAT        13
#define _CFFI_PRIM_DOUBLE       14
#define _CFFI_PRIM_LONGDOUBLE   15

#define _CFFI_PRIM_WCHAR        16
#define _CFFI_PRIM_INT8         17
#define _CFFI_PRIM_UINT8        18
#define _CFFI_PRIM_INT16        19
#define _CFFI_PRIM_UINT16       20
#define _CFFI_PRIM_INT32        21
#define _CFFI_PRIM_UINT32       22
#define _CFFI_PRIM_INT64        23
#define _CFFI_PRIM_UINT64       24
#define _CFFI_PRIM_INTPTR       25
#define _CFFI_PRIM_UINTPTR      26
#define _CFFI_PRIM_PTRDIFF      27
#define _CFFI_PRIM_SIZE         28
#define _CFFI_PRIM_SSIZE        29
#define _CFFI_PRIM_INT_LEAST8   30
#define _CFFI_PRIM_UINT_LEAST8  31
#define _CFFI_PRIM_INT_LEAST16  32
#define _CFFI_PRIM_UINT_LEAST16 33
#define _CFFI_PRIM_INT_LEAST32  34
#define _CFFI_PRIM_UINT_LEAST32 35
#define _CFFI_PRIM_INT_LEAST64  36
#define _CFFI_PRIM_UINT_LEAST64 37
#define _CFFI_PRIM_INT_FAST8    38
#define _CFFI_PRIM_UINT_FAST8   39
#define _CFFI_PRIM_INT_FAST16   40
#define _CFFI_PRIM_UINT_FAST16  41
#define _CFFI_PRIM_INT_FAST32   42
#define _CFFI_PRIM_UINT_FAST32  43
#define _CFFI_PRIM_INT_FAST64   44
#define _CFFI_PRIM_UINT_FAST64  45
#define _CFFI_PRIM_INTMAX       46
#define _CFFI_PRIM_UINTMAX      47

#define _CFFI__NUM_PRIM         48
#define _CFFI__UNKNOWN_PRIM           (-1)
#define _CFFI__UNKNOWN_FLOAT_PRIM     (-2)
#define _CFFI__UNKNOWN_LONG_DOUBLE    (-3)

#define _CFFI__IO_FILE_STRUCT         (-1)


struct _cffi_global_s {
    const char *name;
    void *address;
    _cffi_opcode_t type_op;
    void *size_or_direct_fn;  // OP_GLOBAL_VAR: size, or 0 if unknown
                              // OP_CPYTHON_BLTN_*: addr of direct function
};

struct _cffi_getconst_s {
    unsigned long long value;
    const struct _cffi_type_context_s *ctx;
    int gindex;
};

struct _cffi_struct_union_s {
    const char *name;
    int type_index;          // -> _cffi_types, on a OP_STRUCT_UNION
    int flags;               // _CFFI_F_* flags below
    size_t size;
    int alignment;
    int first_field_index;   // -> _cffi_fields array
    int num_fields;
};
#define _CFFI_F_UNION         0x01   // is a union, not a struct
#define _CFFI_F_CHECK_FIELDS  0x02   // complain if fields are not in the
                                     // "standard layout" or if some are missing
#define _CFFI_F_PACKED        0x04   // for CHECK_FIELDS, assume a packed struct
#define _CFFI_F_EXTERNAL      0x08   // in some other ffi.include()
#define _CFFI_F_OPAQUE        0x10   // opaque

struct _cffi_field_s {
    const char *name;
    size_t field_offset;
    size_t field_size;
    _cffi_opcode_t field_type_op;
};

struct _cffi_enum_s {
    const char *name;
    int type_index;          // -> _cffi_types, on a OP_ENUM
    int type_prim;           // _CFFI_PRIM_xxx
    const char *enumerators; // comma-delimited string
};

struct _cffi_typename_s {
    const char *name;
    int type_index;   /* if opaque, points to a possibly artificial
                         OP_STRUCT which is itself opaque */
};

struct _cffi_type_context_s {
    _cffi_opcode_t *types;
    const struct _cffi_global_s *globals;
    const struct _cffi_field_s *fields;
    const struct _cffi_struct_union_s *struct_unions;
    const struct _cffi_enum_s *enums;
    const struct _cffi_typename_s *typenames;
    int num_globals;
    int num_struct_unions;
    int num_enums;
    int num_typenames;
    const char *const *includes;
    int num_types;
    int flags;      /* future extension */
};

struct _cffi_parse_info_s {
    const struct _cffi_type_context_s *ctx;
    _cffi_opcode_t *output;
    unsigned int output_size;
    size_t error_location;
    const char *error_message;
};

struct _cffi_externpy_s {
    const char *name;
    size_t size_of_result;
    void *reserved1, *reserved2;
};

#ifdef _CFFI_INTERNAL
static int parse_c_type(struct _cffi_parse_info_s *info, const char *input);
static int search_in_globals(const struct _cffi_type_context_s *ctx,
                             const char *search, size_t search_len);
static int search_in_struct_unions(const struct _cffi_type_context_s *ctx,
                                   const char *search, size_t search_len);
#endif

/* this block of #ifs should be kept exactly identical between
   c/_cffi_backend.c, cffi/vengine_cpy.py, cffi/vengine_gen.py
   and cffi/_cffi_include.h */
#if defined(_MSC_VER)
# include <malloc.h>   /* for alloca() */
# if _MSC_VER < 1600   /* MSVC < 2010 */
   typedef __int8 int8_t;
   typedef __int16 int16_t;
   typedef __int32 int32_t;
   typedef __int64 int64_t;
   typedef unsigned __int8 uint8_t;
   typedef unsigned __int16 uint16_t;
   typedef unsigned __int32 uint32_t;
   typedef unsigned __int64 uint64_t;
   typedef __int8 int_least8_t;
   typedef __int16 int_least16_t;
   typedef __int32 int_least32_t;
   typedef __int64 int_least64_t;
   typedef unsigned __int8 uint_least8_t;
   typedef unsigned __int16 uint_least16_t;
   typedef unsigned __int32 uint_least32_t;
   typedef unsigned __int64 uint_least64_t;
   typedef __int8 int_fast8_t;
   typedef __int16 int_fast16_t;
   typedef __int32 int_fast32_t;
   typedef __int64 int_fast64_t;
   typedef unsigned __int8 uint_fast8_t;
   typedef unsigned __int16 uint_fast16_t;
   typedef unsigned __int32 uint_fast32_t;
   typedef unsigned __int64 uint_fast64_t;
   typedef __int64 intmax_t;
   typedef unsigned __int64 uintmax_t;
# else
#  include <stdint.h>
# endif
# if _MSC_VER < 1800   /* MSVC < 2013 */
#  ifndef __cplusplus
    typedef unsigned char _Bool;
#  endif
# endif
#else
# include <stdint.h>
# if (defined (__SVR4) && defined (__sun)) || defined(_AIX) || defined(__hpux)
#  include <alloca.h>
# endif
#endif

#ifdef __GNUC__
# define _CFFI_UNUSED_FN  __attribute__((unused))
#else
# define _CFFI_UNUSED_FN  /* nothing */
#endif

#ifdef __cplusplus
# ifndef _Bool
   typedef bool _Bool;   /* semi-hackish: C++ has no _Bool; bool is builtin */
# endif
#endif

/**********  CPython-specific section  **********/
#ifndef PYPY_VERSION


#if PY_MAJOR_VERSION >= 3
# define PyInt_FromLong PyLong_FromLong
#endif

#define _cffi_from_c_double PyFloat_FromDouble
#define _cffi_from_c_float PyFloat_FromDouble
#define _cffi_from_c_long PyInt_FromLong
#define _cffi_from_c_ulong PyLong_FromUnsignedLong
#define _cffi_from_c_longlong PyLong_FromLongLong
#define _cffi_from_c_ulonglong PyLong_FromUnsignedLongLong

#define _cffi_to_c_double PyFloat_AsDouble
#define _cffi_to_c_float PyFloat_AsDouble

#define _cffi_from_c_int(x, type)                                        \
    (((type)-1) > 0 ? /* unsigned */                                     \
        (sizeof(type) < sizeof(long) ?                                   \
            PyInt_FromLong((long)x) :                                    \
         sizeof(type) == sizeof(long) ?                                  \
            PyLong_FromUnsignedLong((unsigned long)x) :                  \
            PyLong_FromUnsignedLongLong((unsigned long long)x)) :        \
        (sizeof(type) <= sizeof(long) ?                                  \
            PyInt_FromLong((long)x) :                                    \
            PyLong_FromLongLong((long long)x)))

#define _cffi_to_c_int(o, type)                                          \
    ((type)(                                                             \
     sizeof(type) == 1 ? (((type)-1) > 0 ? (type)_cffi_to_c_u8(o)        \
                                         : (type)_cffi_to_c_i8(o)) :     \
     sizeof(type) == 2 ? (((type)-1) > 0 ? (type)_cffi_to_c_u16(o)       \
                                         : (type)_cffi_to_c_i16(o)) :    \
     sizeof(type) == 4 ? (((type)-1) > 0 ? (type)_cffi_to_c_u32(o)       \
                                         : (type)_cffi_to_c_i32(o)) :    \
     sizeof(type) == 8 ? (((type)-1) > 0 ? (type)_cffi_to_c_u64(o)       \
                                         : (type)_cffi_to_c_i64(o)) :    \
     (Py_FatalError("unsupported size for type " #type), (type)0)))

#define _cffi_to_c_i8                                                    \
                 ((int(*)(PyObject *))_cffi_exports[1])
#define _cffi_to_c_u8                                                    \
                 ((int(*)(PyObject *))_cffi_exports[2])
#define _cffi_to_c_i16                                                   \
                 ((int(*)(PyObject *))_cffi_exports[3])
#define _cffi_to_c_u16                                                   \
                 ((int(*)(PyObject *))_cffi_exports[4])
#define _cffi_to_c_i32                                                   \
                 ((int(*)(PyObject *))_cffi_exports[5])
#define _cffi_to_c_u32                                                   \
                 ((unsigned int(*)(PyObject *))_cffi_exports[6])
#define _cffi_to_c_i64                                                   \
                 ((long long(*)(PyObject *))_cffi_exports[7])
#define _cffi_to_c_u64                                                   \
                 ((unsigned long long(*)(PyObject *))_cffi_exports[8])
#define _cffi_to_c_char                                                  \
                 ((int(*)(PyObject *))_cffi_exports[9])
#define _cffi_from_c_pointer                                             \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[10])
#define _cffi_to_c_pointer                                               \
    ((char *(*)(PyObject *, struct _cffi_ctypedescr *))_cffi_exports[11])
#define _cffi_get_struct_layout                                          \
    not used any more
#define _cffi_restore_errno                                              \
    ((void(*)(void))_cffi_exports[13])
#define _cffi_save_errno                                                 \
    ((void(*)(void))_cffi_exports[14])
#define _cffi_from_c_char                                                \
    ((PyObject *(*)(char))_cffi_exports[15])
#define _cffi_from_c_deref                                               \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[16])
#define _cffi_to_c                                                       \
    ((int(*)(char *, struct _cffi_ctypedescr *, PyObject *))_cffi_exports[17])
#define _cffi_from_c_struct                                              \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[18])
#define _cffi_to_c_wchar_t                                               \
    ((wchar_t(*)(PyObject *))_cffi_exports[19])
#define _cffi_from_c_wchar_t                                             \
    ((PyObject *(*)(wchar_t))_cffi_exports[20])
#define _cffi_to_c_long_double                                           \
    ((long double(*)(PyObject *))_cffi_exports[21])
#define _cffi_to_c__Bool                                                 \
    ((_Bool(*)(PyObject *))_cffi_exports[22])
#define _cffi_prepare_pointer_call_argument                              \
    ((Py_ssize_t(*)(struct _cffi_ctypedescr *,                           \
                    PyObject *, char **))_cffi_exports[23])
#define _cffi_convert_array_from_object                                  \
    ((int(*)(char *, struct _cffi_ctypedescr *, PyObject *))_cffi_exports[24])
#define _CFFI_CPIDX  25
#define _cffi_call_python                                                \
    ((void(*)(struct _cffi_externpy_s *, char *))_cffi_exports[_CFFI_CPIDX])
#define _CFFI_NUM_EXPORTS 26

struct _cffi_ctypedescr;

static void *_cffi_exports[_CFFI_NUM_EXPORTS];

#define _cffi_type(index)   (                           \
    assert((((uintptr_t)_cffi_types[index]) & 1) == 0), \
    (struct _cffi_ctypedescr *)_cffi_types[index])

static PyObject *_cffi_init(const char *module_name, Py_ssize_t version,
                            const struct _cffi_type_context_s *ctx)
{
    PyObject *module, *o_arg, *new_module;
    void *raw[] = {
        (void *)module_name,
        (void *)version,
        (void *)_cffi_exports,
        (void *)ctx,
    };

    module = PyImport_ImportModule("_cffi_backend");
    if (module == NULL)
        goto failure;

    o_arg = PyLong_FromVoidPtr((void *)raw);
    if (o_arg == NULL)
        goto failure;

    new_module = PyObject_CallMethod(
        module, (char *)"_init_cffi_1_0_external_module", (char *)"O", o_arg);

    Py_DECREF(o_arg);
    Py_DECREF(module);
    return new_module;

  failure:
    Py_XDECREF(module);
    return NULL;
}

/**********  end CPython-specific section  **********/
#else
_CFFI_UNUSED_FN
static void (*_cffi_call_python_org)(struct _cffi_externpy_s *, char *);
# define _cffi_call_python  _cffi_call_python_org
#endif


#define _cffi_array_len(array)   (sizeof(array) / sizeof((array)[0]))

#define _cffi_prim_int(size, sign)                                      \
    ((size) == 1 ? ((sign) ? _CFFI_PRIM_INT8  : _CFFI_PRIM_UINT8)  :    \
     (size) == 2 ? ((sign) ? _CFFI_PRIM_INT16 : _CFFI_PRIM_UINT16) :    \
     (size) == 4 ? ((sign) ? _CFFI_PRIM_INT32 : _CFFI_PRIM_UINT32) :    \
     (size) == 8 ? ((sign) ? _CFFI_PRIM_INT64 : _CFFI_PRIM_UINT64) :    \
     _CFFI__UNKNOWN_PRIM)

#define _cffi_prim_float(size)                                          \
    ((size) == sizeof(float) ? _CFFI_PRIM_FLOAT :                       \
     (size) == sizeof(double) ? _CFFI_PRIM_DOUBLE :                     \
     (size) == sizeof(long double) ? _CFFI__UNKNOWN_LONG_DOUBLE :       \
     _CFFI__UNKNOWN_FLOAT_PRIM)

#define _cffi_check_int(got, got_nonpos, expected)      \
    ((got_nonpos) == (expected <= 0) &&                 \
     (got) == (unsigned long long)expected)

#ifdef MS_WIN32
# define _cffi_stdcall  __stdcall
#else
# define _cffi_stdcall  /* nothing */
#endif

#ifdef __cplusplus
}
#endif

/************************************************************/

#include <wlc/wlc.h>
#include <sys/types.h>
#include <xkbcommon/xkbcommon-keysyms.h>
#include <linux/input-event-codes.h>


/************************************************************/

static void *_cffi_types[] = {
/*  0 */ _CFFI_OP(_CFFI_OP_FUNCTION, 198), // _Bool()(struct libinput_device *)
/*  1 */ _CFFI_OP(_CFFI_OP_POINTER, 334), // struct libinput_device *
/*  2 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/*  3 */ _CFFI_OP(_CFFI_OP_FUNCTION, 198), // _Bool()(struct wlc_event_source *, int32_t)
/*  4 */ _CFFI_OP(_CFFI_OP_POINTER, 335), // struct wlc_event_source *
/*  5 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21), // int32_t
/*  6 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/*  7 */ _CFFI_OP(_CFFI_OP_FUNCTION, 198), // _Bool()(uintptr_t)
/*  8 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26), // uintptr_t
/*  9 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 10 */ _CFFI_OP(_CFFI_OP_FUNCTION, 198), // _Bool()(uintptr_t, uint32_t, struct wlc_modifiers *, uint32_t, enum wlc_button_state, struct wlc_point *)
/* 11 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 12 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22), // uint32_t
/* 13 */ _CFFI_OP(_CFFI_OP_POINTER, 337), // struct wlc_modifiers *
/* 14 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 15 */ _CFFI_OP(_CFFI_OP_ENUM, 1), // enum wlc_button_state
/* 16 */ _CFFI_OP(_CFFI_OP_POINTER, 338), // struct wlc_point *
/* 17 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 18 */ _CFFI_OP(_CFFI_OP_FUNCTION, 198), // _Bool()(uintptr_t, uint32_t, struct wlc_modifiers *, uint32_t, enum wlc_key_state)
/* 19 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 20 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 21 */ _CFFI_OP(_CFFI_OP_NOOP, 13),
/* 22 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 23 */ _CFFI_OP(_CFFI_OP_ENUM, 3), // enum wlc_key_state
/* 24 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 25 */ _CFFI_OP(_CFFI_OP_FUNCTION, 198), // _Bool()(uintptr_t, uint32_t, struct wlc_modifiers const *, enum wlc_touch_type, int32_t, struct wlc_point const *)
/* 26 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 27 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 28 */ _CFFI_OP(_CFFI_OP_POINTER, 337), // struct wlc_modifiers const *
/* 29 */ _CFFI_OP(_CFFI_OP_ENUM, 12), // enum wlc_touch_type
/* 30 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 31 */ _CFFI_OP(_CFFI_OP_POINTER, 338), // struct wlc_point const *
/* 32 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 33 */ _CFFI_OP(_CFFI_OP_FUNCTION, 198), // _Bool()(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_button_state, struct wlc_point const *)
/* 34 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 35 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 36 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 37 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 38 */ _CFFI_OP(_CFFI_OP_NOOP, 15),
/* 39 */ _CFFI_OP(_CFFI_OP_NOOP, 31),
/* 40 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 41 */ _CFFI_OP(_CFFI_OP_FUNCTION, 198), // _Bool()(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_key_state)
/* 42 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 43 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 44 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 45 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 46 */ _CFFI_OP(_CFFI_OP_NOOP, 23),
/* 47 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 48 */ _CFFI_OP(_CFFI_OP_FUNCTION, 198), // _Bool()(uintptr_t, uint32_t, struct wlc_modifiers const *, uint8_t, double *)
/* 49 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 50 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 51 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 52 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18), // uint8_t
/* 53 */ _CFFI_OP(_CFFI_OP_POINTER, 323), // double *
/* 54 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 55 */ _CFFI_OP(_CFFI_OP_FUNCTION, 198), // _Bool()(uintptr_t, uint32_t, struct wlc_point *)
/* 56 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 57 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 58 */ _CFFI_OP(_CFFI_OP_NOOP, 16),
/* 59 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 60 */ _CFFI_OP(_CFFI_OP_FUNCTION, 198), // _Bool()(uintptr_t, uint32_t, struct wlc_point const *)
/* 61 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 62 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 63 */ _CFFI_OP(_CFFI_OP_NOOP, 31),
/* 64 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 65 */ _CFFI_OP(_CFFI_OP_FUNCTION, 198), // _Bool()(uintptr_t, uintptr_t const *, size_t)
/* 66 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 67 */ _CFFI_OP(_CFFI_OP_POINTER, 8), // uintptr_t const *
/* 68 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 28), // size_t
/* 69 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 70 */ _CFFI_OP(_CFFI_OP_FUNCTION, 198), // _Bool()(void)
/* 71 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 72 */ _CFFI_OP(_CFFI_OP_FUNCTION, 174), // char const *()(uintptr_t)
/* 73 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 74 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 75 */ _CFFI_OP(_CFFI_OP_FUNCTION, 324), // enum wlc_backend_type()(void)
/* 76 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 77 */ _CFFI_OP(_CFFI_OP_FUNCTION, 328), // enum wlc_positioner_anchor_bit()(uintptr_t)
/* 78 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 79 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 80 */ _CFFI_OP(_CFFI_OP_FUNCTION, 329), // enum wlc_positioner_constraint_adjustment_bit()(uintptr_t)
/* 81 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 82 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 83 */ _CFFI_OP(_CFFI_OP_FUNCTION, 330), // enum wlc_positioner_gravity_bit()(uintptr_t)
/* 84 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 85 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 86 */ _CFFI_OP(_CFFI_OP_FUNCTION, 87), // int()(int, uint32_t, void *)
/* 87 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 7), // int
/* 88 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 89 */ _CFFI_OP(_CFFI_OP_POINTER, 350), // void *
/* 90 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 91 */ _CFFI_OP(_CFFI_OP_FUNCTION, 87), // int()(uintptr_t)
/* 92 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 93 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 94 */ _CFFI_OP(_CFFI_OP_FUNCTION, 87), // int()(void *)
/* 95 */ _CFFI_OP(_CFFI_OP_NOOP, 89),
/* 96 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 97 */ _CFFI_OP(_CFFI_OP_FUNCTION, 4), // struct wlc_event_source *()(int(*)(void *), void *)
/* 98 */ _CFFI_OP(_CFFI_OP_POINTER, 94), // int(*)(void *)
/* 99 */ _CFFI_OP(_CFFI_OP_NOOP, 89),
/* 100 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 101 */ _CFFI_OP(_CFFI_OP_FUNCTION, 4), // struct wlc_event_source *()(int, uint32_t, int(*)(int, uint32_t, void *), void *)
/* 102 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 7),
/* 103 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 104 */ _CFFI_OP(_CFFI_OP_POINTER, 86), // int(*)(int, uint32_t, void *)
/* 105 */ _CFFI_OP(_CFFI_OP_NOOP, 89),
/* 106 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 107 */ _CFFI_OP(_CFFI_OP_FUNCTION, 216), // struct wlc_geometry const *()(uintptr_t)
/* 108 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 109 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 110 */ _CFFI_OP(_CFFI_OP_FUNCTION, 31), // struct wlc_point const *()(uintptr_t)
/* 111 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 112 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 113 */ _CFFI_OP(_CFFI_OP_FUNCTION, 233), // struct wlc_size const *()(uintptr_t)
/* 114 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 115 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 116 */ _CFFI_OP(_CFFI_OP_FUNCTION, 340), // struct xkb_keymap *()(void)
/* 117 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 118 */ _CFFI_OP(_CFFI_OP_FUNCTION, 342), // struct xkb_state *()(void)
/* 119 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 120 */ _CFFI_OP(_CFFI_OP_FUNCTION, 243), // uint16_t()(uintptr_t)
/* 121 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 122 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 123 */ _CFFI_OP(_CFFI_OP_FUNCTION, 344), // uint32_t const *()(size_t *)
/* 124 */ _CFFI_OP(_CFFI_OP_POINTER, 68), // size_t *
/* 125 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 126 */ _CFFI_OP(_CFFI_OP_FUNCTION, 12), // uint32_t()(uint32_t, struct wlc_modifiers const *)
/* 127 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 128 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 129 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 130 */ _CFFI_OP(_CFFI_OP_FUNCTION, 12), // uint32_t()(uintptr_t)
/* 131 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 132 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 133 */ _CFFI_OP(_CFFI_OP_FUNCTION, 345), // uintptr_t *()(uintptr_t, size_t *)
/* 134 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 135 */ _CFFI_OP(_CFFI_OP_NOOP, 124),
/* 136 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 137 */ _CFFI_OP(_CFFI_OP_FUNCTION, 67), // uintptr_t const *()(size_t *)
/* 138 */ _CFFI_OP(_CFFI_OP_NOOP, 124),
/* 139 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 140 */ _CFFI_OP(_CFFI_OP_FUNCTION, 67), // uintptr_t const *()(uintptr_t, size_t *)
/* 141 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 142 */ _CFFI_OP(_CFFI_OP_NOOP, 124),
/* 143 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 144 */ _CFFI_OP(_CFFI_OP_FUNCTION, 8), // uintptr_t()(uintptr_t)
/* 145 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 146 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 147 */ _CFFI_OP(_CFFI_OP_FUNCTION, 8), // uintptr_t()(void)
/* 148 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 149 */ _CFFI_OP(_CFFI_OP_FUNCTION, 89), // void *()(uintptr_t)
/* 150 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 151 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 152 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(_Bool(*)(struct libinput_device *))
/* 153 */ _CFFI_OP(_CFFI_OP_POINTER, 0), // _Bool(*)(struct libinput_device *)
/* 154 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 155 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(_Bool(*)(uintptr_t))
/* 156 */ _CFFI_OP(_CFFI_OP_POINTER, 7), // _Bool(*)(uintptr_t)
/* 157 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 158 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(_Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, enum wlc_touch_type, int32_t, struct wlc_point const *))
/* 159 */ _CFFI_OP(_CFFI_OP_POINTER, 25), // _Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, enum wlc_touch_type, int32_t, struct wlc_point const *)
/* 160 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 161 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(_Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_button_state, struct wlc_point const *))
/* 162 */ _CFFI_OP(_CFFI_OP_POINTER, 33), // _Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_button_state, struct wlc_point const *)
/* 163 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 164 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(_Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_key_state))
/* 165 */ _CFFI_OP(_CFFI_OP_POINTER, 41), // _Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_key_state)
/* 166 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 167 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(_Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint8_t, double *))
/* 168 */ _CFFI_OP(_CFFI_OP_POINTER, 48), // _Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint8_t, double *)
/* 169 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 170 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(_Bool(*)(uintptr_t, uint32_t, struct wlc_point const *))
/* 171 */ _CFFI_OP(_CFFI_OP_POINTER, 60), // _Bool(*)(uintptr_t, uint32_t, struct wlc_point const *)
/* 172 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 173 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(char const *, char * const *)
/* 174 */ _CFFI_OP(_CFFI_OP_POINTER, 322), // char const *
/* 175 */ _CFFI_OP(_CFFI_OP_POINTER, 321), // char * const *
/* 176 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 177 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(enum wlc_log_type, char const *)
/* 178 */ _CFFI_OP(_CFFI_OP_ENUM, 5), // enum wlc_log_type
/* 179 */ _CFFI_OP(_CFFI_OP_NOOP, 174),
/* 180 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 181 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(struct libinput_device *)
/* 182 */ _CFFI_OP(_CFFI_OP_NOOP, 1),
/* 183 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 184 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(struct wlc_event_source *)
/* 185 */ _CFFI_OP(_CFFI_OP_NOOP, 4),
/* 186 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 187 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(struct wlc_point *)
/* 188 */ _CFFI_OP(_CFFI_OP_NOOP, 16),
/* 189 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 190 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(struct wlc_point const *)
/* 191 */ _CFFI_OP(_CFFI_OP_NOOP, 31),
/* 192 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 193 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t)
/* 194 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 195 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 196 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, _Bool)
/* 197 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 198 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1), // _Bool
/* 199 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 200 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, enum wlc_view_state_bit, _Bool)
/* 201 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 202 */ _CFFI_OP(_CFFI_OP_ENUM, 14), // enum wlc_view_state_bit
/* 203 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 204 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 205 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, enum wlc_view_type_bit, _Bool)
/* 206 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 207 */ _CFFI_OP(_CFFI_OP_ENUM, 15), // enum wlc_view_type_bit
/* 208 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 209 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 210 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, struct wlc_geometry *)
/* 211 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 212 */ _CFFI_OP(_CFFI_OP_POINTER, 336), // struct wlc_geometry *
/* 213 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 214 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, struct wlc_geometry const *)
/* 215 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 216 */ _CFFI_OP(_CFFI_OP_POINTER, 336), // struct wlc_geometry const *
/* 217 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 218 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, struct wlc_point *)
/* 219 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 220 */ _CFFI_OP(_CFFI_OP_NOOP, 16),
/* 221 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 222 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, struct wlc_point const *)
/* 223 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 224 */ _CFFI_OP(_CFFI_OP_NOOP, 31),
/* 225 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 226 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, struct wlc_size *, struct wlc_size *)
/* 227 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 228 */ _CFFI_OP(_CFFI_OP_POINTER, 339), // struct wlc_size *
/* 229 */ _CFFI_OP(_CFFI_OP_NOOP, 228),
/* 230 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 231 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, struct wlc_size const *, struct wlc_size const *)
/* 232 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 233 */ _CFFI_OP(_CFFI_OP_POINTER, 339), // struct wlc_size const *
/* 234 */ _CFFI_OP(_CFFI_OP_NOOP, 233),
/* 235 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 236 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, struct wlc_size const *, uint32_t)
/* 237 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 238 */ _CFFI_OP(_CFFI_OP_NOOP, 233),
/* 239 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 240 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 241 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, uint16_t, uint16_t *, uint16_t *, uint16_t *)
/* 242 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 243 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20), // uint16_t
/* 244 */ _CFFI_OP(_CFFI_OP_POINTER, 243), // uint16_t *
/* 245 */ _CFFI_OP(_CFFI_OP_NOOP, 244),
/* 246 */ _CFFI_OP(_CFFI_OP_NOOP, 244),
/* 247 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 248 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, uint32_t)
/* 249 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 250 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 251 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 252 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, uint32_t, struct wlc_geometry const *)
/* 253 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 254 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 255 */ _CFFI_OP(_CFFI_OP_NOOP, 216),
/* 256 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 257 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, uint32_t, struct wlc_point *)
/* 258 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 259 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 260 */ _CFFI_OP(_CFFI_OP_NOOP, 16),
/* 261 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 262 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, uint32_t, struct wlc_point const *)
/* 263 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 264 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 265 */ _CFFI_OP(_CFFI_OP_NOOP, 31),
/* 266 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 267 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, uintptr_t)
/* 268 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 269 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 270 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 271 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, uintptr_t, uintptr_t)
/* 272 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 273 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 274 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 275 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 276 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(uintptr_t, void const *)
/* 277 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 26),
/* 278 */ _CFFI_OP(_CFFI_OP_POINTER, 350), // void const *
/* 279 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 280 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(void(*)(enum wlc_log_type, char const *))
/* 281 */ _CFFI_OP(_CFFI_OP_POINTER, 177), // void(*)(enum wlc_log_type, char const *)
/* 282 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 283 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(void(*)(struct libinput_device *))
/* 284 */ _CFFI_OP(_CFFI_OP_POINTER, 181), // void(*)(struct libinput_device *)
/* 285 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 286 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(void(*)(uintptr_t))
/* 287 */ _CFFI_OP(_CFFI_OP_POINTER, 193), // void(*)(uintptr_t)
/* 288 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 289 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(void(*)(uintptr_t, _Bool))
/* 290 */ _CFFI_OP(_CFFI_OP_POINTER, 196), // void(*)(uintptr_t, _Bool)
/* 291 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 292 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(void(*)(uintptr_t, enum wlc_view_state_bit, _Bool))
/* 293 */ _CFFI_OP(_CFFI_OP_POINTER, 200), // void(*)(uintptr_t, enum wlc_view_state_bit, _Bool)
/* 294 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 295 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(void(*)(uintptr_t, struct wlc_geometry const *))
/* 296 */ _CFFI_OP(_CFFI_OP_POINTER, 214), // void(*)(uintptr_t, struct wlc_geometry const *)
/* 297 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 298 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(void(*)(uintptr_t, struct wlc_point const *))
/* 299 */ _CFFI_OP(_CFFI_OP_POINTER, 222), // void(*)(uintptr_t, struct wlc_point const *)
/* 300 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 301 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(void(*)(uintptr_t, struct wlc_size const *, struct wlc_size const *))
/* 302 */ _CFFI_OP(_CFFI_OP_POINTER, 231), // void(*)(uintptr_t, struct wlc_size const *, struct wlc_size const *)
/* 303 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 304 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(void(*)(uintptr_t, uint32_t))
/* 305 */ _CFFI_OP(_CFFI_OP_POINTER, 248), // void(*)(uintptr_t, uint32_t)
/* 306 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 307 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(void(*)(uintptr_t, uint32_t, struct wlc_point const *))
/* 308 */ _CFFI_OP(_CFFI_OP_POINTER, 262), // void(*)(uintptr_t, uint32_t, struct wlc_point const *)
/* 309 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 310 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(void(*)(uintptr_t, uintptr_t, uintptr_t))
/* 311 */ _CFFI_OP(_CFFI_OP_POINTER, 271), // void(*)(uintptr_t, uintptr_t, uintptr_t)
/* 312 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 313 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(void(*)(void))
/* 314 */ _CFFI_OP(_CFFI_OP_POINTER, 316), // void(*)(void)
/* 315 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 316 */ _CFFI_OP(_CFFI_OP_FUNCTION, 350), // void()(void)
/* 317 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 318 */ _CFFI_OP(_CFFI_OP_POINTER, 10), // _Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers *, uint32_t, enum wlc_button_state, struct wlc_point *)
/* 319 */ _CFFI_OP(_CFFI_OP_POINTER, 18), // _Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers *, uint32_t, enum wlc_key_state)
/* 320 */ _CFFI_OP(_CFFI_OP_POINTER, 55), // _Bool(*)(uintptr_t, uint32_t, struct wlc_point *)
/* 321 */ _CFFI_OP(_CFFI_OP_POINTER, 322), // char *
/* 322 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 2), // char
/* 323 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14), // double
/* 324 */ _CFFI_OP(_CFFI_OP_ENUM, 0), // enum wlc_backend_type
/* 325 */ _CFFI_OP(_CFFI_OP_ENUM, 2), // enum wlc_event_bit
/* 326 */ _CFFI_OP(_CFFI_OP_ENUM, 4), // enum wlc_led_bit
/* 327 */ _CFFI_OP(_CFFI_OP_ENUM, 6), // enum wlc_modifier_bit
/* 328 */ _CFFI_OP(_CFFI_OP_ENUM, 7), // enum wlc_positioner_anchor_bit
/* 329 */ _CFFI_OP(_CFFI_OP_ENUM, 8), // enum wlc_positioner_constraint_adjustment_bit
/* 330 */ _CFFI_OP(_CFFI_OP_ENUM, 9), // enum wlc_positioner_gravity_bit
/* 331 */ _CFFI_OP(_CFFI_OP_ENUM, 10), // enum wlc_resize_edge
/* 332 */ _CFFI_OP(_CFFI_OP_ENUM, 11), // enum wlc_scroll_axis_bit
/* 333 */ _CFFI_OP(_CFFI_OP_ENUM, 13), // enum wlc_view_property_update_bit
/* 334 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 0), // struct libinput_device
/* 335 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 1), // struct wlc_event_source
/* 336 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 2), // struct wlc_geometry
/* 337 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 3), // struct wlc_modifiers
/* 338 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 4), // struct wlc_point
/* 339 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 5), // struct wlc_size
/* 340 */ _CFFI_OP(_CFFI_OP_POINTER, 341), // struct xkb_keymap *
/* 341 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 6), // struct xkb_keymap
/* 342 */ _CFFI_OP(_CFFI_OP_POINTER, 343), // struct xkb_state *
/* 343 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 7), // struct xkb_state
/* 344 */ _CFFI_OP(_CFFI_OP_POINTER, 12), // uint32_t const *
/* 345 */ _CFFI_OP(_CFFI_OP_POINTER, 8), // uintptr_t *
/* 346 */ _CFFI_OP(_CFFI_OP_POINTER, 210), // void(*)(uintptr_t, struct wlc_geometry *)
/* 347 */ _CFFI_OP(_CFFI_OP_POINTER, 218), // void(*)(uintptr_t, struct wlc_point *)
/* 348 */ _CFFI_OP(_CFFI_OP_POINTER, 226), // void(*)(uintptr_t, struct wlc_size *, struct wlc_size *)
/* 349 */ _CFFI_OP(_CFFI_OP_POINTER, 257), // void(*)(uintptr_t, uint32_t, struct wlc_point *)
/* 350 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 0), // void
};

static int _cffi_const_WLC_BACKEND_NONE(unsigned long long *o)
{
  int n = (WLC_BACKEND_NONE) <= 0;
  *o = (unsigned long long)((WLC_BACKEND_NONE) | 0);  /* check that WLC_BACKEND_NONE is an integer */
  return n;
}

static int _cffi_const_WLC_BACKEND_DRM(unsigned long long *o)
{
  int n = (WLC_BACKEND_DRM) <= 0;
  *o = (unsigned long long)((WLC_BACKEND_DRM) | 0);  /* check that WLC_BACKEND_DRM is an integer */
  return n;
}

static int _cffi_const_WLC_BACKEND_X11(unsigned long long *o)
{
  int n = (WLC_BACKEND_X11) <= 0;
  *o = (unsigned long long)((WLC_BACKEND_X11) | 0);  /* check that WLC_BACKEND_X11 is an integer */
  return n;
}

static int _cffi_const_WLC_BUTTON_STATE_RELEASED(unsigned long long *o)
{
  int n = (WLC_BUTTON_STATE_RELEASED) <= 0;
  *o = (unsigned long long)((WLC_BUTTON_STATE_RELEASED) | 0);  /* check that WLC_BUTTON_STATE_RELEASED is an integer */
  return n;
}

static int _cffi_const_WLC_BUTTON_STATE_PRESSED(unsigned long long *o)
{
  int n = (WLC_BUTTON_STATE_PRESSED) <= 0;
  *o = (unsigned long long)((WLC_BUTTON_STATE_PRESSED) | 0);  /* check that WLC_BUTTON_STATE_PRESSED is an integer */
  return n;
}

static int _cffi_const_WLC_EVENT_READABLE(unsigned long long *o)
{
  int n = (WLC_EVENT_READABLE) <= 0;
  *o = (unsigned long long)((WLC_EVENT_READABLE) | 0);  /* check that WLC_EVENT_READABLE is an integer */
  return n;
}

static int _cffi_const_WLC_EVENT_WRITABLE(unsigned long long *o)
{
  int n = (WLC_EVENT_WRITABLE) <= 0;
  *o = (unsigned long long)((WLC_EVENT_WRITABLE) | 0);  /* check that WLC_EVENT_WRITABLE is an integer */
  return n;
}

static int _cffi_const_WLC_EVENT_HANGUP(unsigned long long *o)
{
  int n = (WLC_EVENT_HANGUP) <= 0;
  *o = (unsigned long long)((WLC_EVENT_HANGUP) | 0);  /* check that WLC_EVENT_HANGUP is an integer */
  return n;
}

static int _cffi_const_WLC_EVENT_ERROR(unsigned long long *o)
{
  int n = (WLC_EVENT_ERROR) <= 0;
  *o = (unsigned long long)((WLC_EVENT_ERROR) | 0);  /* check that WLC_EVENT_ERROR is an integer */
  return n;
}

static int _cffi_const_WLC_KEY_STATE_RELEASED(unsigned long long *o)
{
  int n = (WLC_KEY_STATE_RELEASED) <= 0;
  *o = (unsigned long long)((WLC_KEY_STATE_RELEASED) | 0);  /* check that WLC_KEY_STATE_RELEASED is an integer */
  return n;
}

static int _cffi_const_WLC_KEY_STATE_PRESSED(unsigned long long *o)
{
  int n = (WLC_KEY_STATE_PRESSED) <= 0;
  *o = (unsigned long long)((WLC_KEY_STATE_PRESSED) | 0);  /* check that WLC_KEY_STATE_PRESSED is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_LED_NUM(unsigned long long *o)
{
  int n = (WLC_BIT_LED_NUM) <= 0;
  *o = (unsigned long long)((WLC_BIT_LED_NUM) | 0);  /* check that WLC_BIT_LED_NUM is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_LED_CAPS(unsigned long long *o)
{
  int n = (WLC_BIT_LED_CAPS) <= 0;
  *o = (unsigned long long)((WLC_BIT_LED_CAPS) | 0);  /* check that WLC_BIT_LED_CAPS is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_LED_SCROLL(unsigned long long *o)
{
  int n = (WLC_BIT_LED_SCROLL) <= 0;
  *o = (unsigned long long)((WLC_BIT_LED_SCROLL) | 0);  /* check that WLC_BIT_LED_SCROLL is an integer */
  return n;
}

static int _cffi_const_WLC_LOG_INFO(unsigned long long *o)
{
  int n = (WLC_LOG_INFO) <= 0;
  *o = (unsigned long long)((WLC_LOG_INFO) | 0);  /* check that WLC_LOG_INFO is an integer */
  return n;
}

static int _cffi_const_WLC_LOG_WARN(unsigned long long *o)
{
  int n = (WLC_LOG_WARN) <= 0;
  *o = (unsigned long long)((WLC_LOG_WARN) | 0);  /* check that WLC_LOG_WARN is an integer */
  return n;
}

static int _cffi_const_WLC_LOG_ERROR(unsigned long long *o)
{
  int n = (WLC_LOG_ERROR) <= 0;
  *o = (unsigned long long)((WLC_LOG_ERROR) | 0);  /* check that WLC_LOG_ERROR is an integer */
  return n;
}

static int _cffi_const_WLC_LOG_WAYLAND(unsigned long long *o)
{
  int n = (WLC_LOG_WAYLAND) <= 0;
  *o = (unsigned long long)((WLC_LOG_WAYLAND) | 0);  /* check that WLC_LOG_WAYLAND is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_MOD_SHIFT(unsigned long long *o)
{
  int n = (WLC_BIT_MOD_SHIFT) <= 0;
  *o = (unsigned long long)((WLC_BIT_MOD_SHIFT) | 0);  /* check that WLC_BIT_MOD_SHIFT is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_MOD_CAPS(unsigned long long *o)
{
  int n = (WLC_BIT_MOD_CAPS) <= 0;
  *o = (unsigned long long)((WLC_BIT_MOD_CAPS) | 0);  /* check that WLC_BIT_MOD_CAPS is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_MOD_CTRL(unsigned long long *o)
{
  int n = (WLC_BIT_MOD_CTRL) <= 0;
  *o = (unsigned long long)((WLC_BIT_MOD_CTRL) | 0);  /* check that WLC_BIT_MOD_CTRL is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_MOD_ALT(unsigned long long *o)
{
  int n = (WLC_BIT_MOD_ALT) <= 0;
  *o = (unsigned long long)((WLC_BIT_MOD_ALT) | 0);  /* check that WLC_BIT_MOD_ALT is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_MOD_MOD2(unsigned long long *o)
{
  int n = (WLC_BIT_MOD_MOD2) <= 0;
  *o = (unsigned long long)((WLC_BIT_MOD_MOD2) | 0);  /* check that WLC_BIT_MOD_MOD2 is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_MOD_MOD3(unsigned long long *o)
{
  int n = (WLC_BIT_MOD_MOD3) <= 0;
  *o = (unsigned long long)((WLC_BIT_MOD_MOD3) | 0);  /* check that WLC_BIT_MOD_MOD3 is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_MOD_LOGO(unsigned long long *o)
{
  int n = (WLC_BIT_MOD_LOGO) <= 0;
  *o = (unsigned long long)((WLC_BIT_MOD_LOGO) | 0);  /* check that WLC_BIT_MOD_LOGO is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_MOD_MOD5(unsigned long long *o)
{
  int n = (WLC_BIT_MOD_MOD5) <= 0;
  *o = (unsigned long long)((WLC_BIT_MOD_MOD5) | 0);  /* check that WLC_BIT_MOD_MOD5 is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_ANCHOR_NONE(unsigned long long *o)
{
  int n = (WLC_BIT_ANCHOR_NONE) <= 0;
  *o = (unsigned long long)((WLC_BIT_ANCHOR_NONE) | 0);  /* check that WLC_BIT_ANCHOR_NONE is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_ANCHOR_TOP(unsigned long long *o)
{
  int n = (WLC_BIT_ANCHOR_TOP) <= 0;
  *o = (unsigned long long)((WLC_BIT_ANCHOR_TOP) | 0);  /* check that WLC_BIT_ANCHOR_TOP is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_ANCHOR_BOTTOM(unsigned long long *o)
{
  int n = (WLC_BIT_ANCHOR_BOTTOM) <= 0;
  *o = (unsigned long long)((WLC_BIT_ANCHOR_BOTTOM) | 0);  /* check that WLC_BIT_ANCHOR_BOTTOM is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_ANCHOR_LEFT(unsigned long long *o)
{
  int n = (WLC_BIT_ANCHOR_LEFT) <= 0;
  *o = (unsigned long long)((WLC_BIT_ANCHOR_LEFT) | 0);  /* check that WLC_BIT_ANCHOR_LEFT is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_ANCHOR_RIGHT(unsigned long long *o)
{
  int n = (WLC_BIT_ANCHOR_RIGHT) <= 0;
  *o = (unsigned long long)((WLC_BIT_ANCHOR_RIGHT) | 0);  /* check that WLC_BIT_ANCHOR_RIGHT is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_NONE(unsigned long long *o)
{
  int n = (WLC_BIT_CONSTRAINT_ADJUSTMENT_NONE) <= 0;
  *o = (unsigned long long)((WLC_BIT_CONSTRAINT_ADJUSTMENT_NONE) | 0);  /* check that WLC_BIT_CONSTRAINT_ADJUSTMENT_NONE is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_X(unsigned long long *o)
{
  int n = (WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_X) <= 0;
  *o = (unsigned long long)((WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_X) | 0);  /* check that WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_X is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_Y(unsigned long long *o)
{
  int n = (WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_Y) <= 0;
  *o = (unsigned long long)((WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_Y) | 0);  /* check that WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_Y is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_X(unsigned long long *o)
{
  int n = (WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_X) <= 0;
  *o = (unsigned long long)((WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_X) | 0);  /* check that WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_X is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_Y(unsigned long long *o)
{
  int n = (WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_Y) <= 0;
  *o = (unsigned long long)((WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_Y) | 0);  /* check that WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_Y is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_X(unsigned long long *o)
{
  int n = (WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_X) <= 0;
  *o = (unsigned long long)((WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_X) | 0);  /* check that WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_X is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_Y(unsigned long long *o)
{
  int n = (WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_Y) <= 0;
  *o = (unsigned long long)((WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_Y) | 0);  /* check that WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_Y is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_GRAVITY_NONE(unsigned long long *o)
{
  int n = (WLC_BIT_GRAVITY_NONE) <= 0;
  *o = (unsigned long long)((WLC_BIT_GRAVITY_NONE) | 0);  /* check that WLC_BIT_GRAVITY_NONE is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_GRAVITY_TOP(unsigned long long *o)
{
  int n = (WLC_BIT_GRAVITY_TOP) <= 0;
  *o = (unsigned long long)((WLC_BIT_GRAVITY_TOP) | 0);  /* check that WLC_BIT_GRAVITY_TOP is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_GRAVITY_BOTTOM(unsigned long long *o)
{
  int n = (WLC_BIT_GRAVITY_BOTTOM) <= 0;
  *o = (unsigned long long)((WLC_BIT_GRAVITY_BOTTOM) | 0);  /* check that WLC_BIT_GRAVITY_BOTTOM is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_GRAVITY_LEFT(unsigned long long *o)
{
  int n = (WLC_BIT_GRAVITY_LEFT) <= 0;
  *o = (unsigned long long)((WLC_BIT_GRAVITY_LEFT) | 0);  /* check that WLC_BIT_GRAVITY_LEFT is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_GRAVITY_RIGHT(unsigned long long *o)
{
  int n = (WLC_BIT_GRAVITY_RIGHT) <= 0;
  *o = (unsigned long long)((WLC_BIT_GRAVITY_RIGHT) | 0);  /* check that WLC_BIT_GRAVITY_RIGHT is an integer */
  return n;
}

static int _cffi_const_WLC_RESIZE_EDGE_NONE(unsigned long long *o)
{
  int n = (WLC_RESIZE_EDGE_NONE) <= 0;
  *o = (unsigned long long)((WLC_RESIZE_EDGE_NONE) | 0);  /* check that WLC_RESIZE_EDGE_NONE is an integer */
  return n;
}

static int _cffi_const_WLC_RESIZE_EDGE_TOP(unsigned long long *o)
{
  int n = (WLC_RESIZE_EDGE_TOP) <= 0;
  *o = (unsigned long long)((WLC_RESIZE_EDGE_TOP) | 0);  /* check that WLC_RESIZE_EDGE_TOP is an integer */
  return n;
}

static int _cffi_const_WLC_RESIZE_EDGE_BOTTOM(unsigned long long *o)
{
  int n = (WLC_RESIZE_EDGE_BOTTOM) <= 0;
  *o = (unsigned long long)((WLC_RESIZE_EDGE_BOTTOM) | 0);  /* check that WLC_RESIZE_EDGE_BOTTOM is an integer */
  return n;
}

static int _cffi_const_WLC_RESIZE_EDGE_LEFT(unsigned long long *o)
{
  int n = (WLC_RESIZE_EDGE_LEFT) <= 0;
  *o = (unsigned long long)((WLC_RESIZE_EDGE_LEFT) | 0);  /* check that WLC_RESIZE_EDGE_LEFT is an integer */
  return n;
}

static int _cffi_const_WLC_RESIZE_EDGE_TOP_LEFT(unsigned long long *o)
{
  int n = (WLC_RESIZE_EDGE_TOP_LEFT) <= 0;
  *o = (unsigned long long)((WLC_RESIZE_EDGE_TOP_LEFT) | 0);  /* check that WLC_RESIZE_EDGE_TOP_LEFT is an integer */
  return n;
}

static int _cffi_const_WLC_RESIZE_EDGE_BOTTOM_LEFT(unsigned long long *o)
{
  int n = (WLC_RESIZE_EDGE_BOTTOM_LEFT) <= 0;
  *o = (unsigned long long)((WLC_RESIZE_EDGE_BOTTOM_LEFT) | 0);  /* check that WLC_RESIZE_EDGE_BOTTOM_LEFT is an integer */
  return n;
}

static int _cffi_const_WLC_RESIZE_EDGE_RIGHT(unsigned long long *o)
{
  int n = (WLC_RESIZE_EDGE_RIGHT) <= 0;
  *o = (unsigned long long)((WLC_RESIZE_EDGE_RIGHT) | 0);  /* check that WLC_RESIZE_EDGE_RIGHT is an integer */
  return n;
}

static int _cffi_const_WLC_RESIZE_EDGE_TOP_RIGHT(unsigned long long *o)
{
  int n = (WLC_RESIZE_EDGE_TOP_RIGHT) <= 0;
  *o = (unsigned long long)((WLC_RESIZE_EDGE_TOP_RIGHT) | 0);  /* check that WLC_RESIZE_EDGE_TOP_RIGHT is an integer */
  return n;
}

static int _cffi_const_WLC_RESIZE_EDGE_BOTTOM_RIGHT(unsigned long long *o)
{
  int n = (WLC_RESIZE_EDGE_BOTTOM_RIGHT) <= 0;
  *o = (unsigned long long)((WLC_RESIZE_EDGE_BOTTOM_RIGHT) | 0);  /* check that WLC_RESIZE_EDGE_BOTTOM_RIGHT is an integer */
  return n;
}

static int _cffi_const_WLC_SCROLL_AXIS_VERTICAL(unsigned long long *o)
{
  int n = (WLC_SCROLL_AXIS_VERTICAL) <= 0;
  *o = (unsigned long long)((WLC_SCROLL_AXIS_VERTICAL) | 0);  /* check that WLC_SCROLL_AXIS_VERTICAL is an integer */
  return n;
}

static int _cffi_const_WLC_SCROLL_AXIS_HORIZONTAL(unsigned long long *o)
{
  int n = (WLC_SCROLL_AXIS_HORIZONTAL) <= 0;
  *o = (unsigned long long)((WLC_SCROLL_AXIS_HORIZONTAL) | 0);  /* check that WLC_SCROLL_AXIS_HORIZONTAL is an integer */
  return n;
}

static int _cffi_const_WLC_TOUCH_DOWN(unsigned long long *o)
{
  int n = (WLC_TOUCH_DOWN) <= 0;
  *o = (unsigned long long)((WLC_TOUCH_DOWN) | 0);  /* check that WLC_TOUCH_DOWN is an integer */
  return n;
}

static int _cffi_const_WLC_TOUCH_UP(unsigned long long *o)
{
  int n = (WLC_TOUCH_UP) <= 0;
  *o = (unsigned long long)((WLC_TOUCH_UP) | 0);  /* check that WLC_TOUCH_UP is an integer */
  return n;
}

static int _cffi_const_WLC_TOUCH_MOTION(unsigned long long *o)
{
  int n = (WLC_TOUCH_MOTION) <= 0;
  *o = (unsigned long long)((WLC_TOUCH_MOTION) | 0);  /* check that WLC_TOUCH_MOTION is an integer */
  return n;
}

static int _cffi_const_WLC_TOUCH_FRAME(unsigned long long *o)
{
  int n = (WLC_TOUCH_FRAME) <= 0;
  *o = (unsigned long long)((WLC_TOUCH_FRAME) | 0);  /* check that WLC_TOUCH_FRAME is an integer */
  return n;
}

static int _cffi_const_WLC_TOUCH_CANCEL(unsigned long long *o)
{
  int n = (WLC_TOUCH_CANCEL) <= 0;
  *o = (unsigned long long)((WLC_TOUCH_CANCEL) | 0);  /* check that WLC_TOUCH_CANCEL is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_PROPERTY_TITLE(unsigned long long *o)
{
  int n = (WLC_BIT_PROPERTY_TITLE) <= 0;
  *o = (unsigned long long)((WLC_BIT_PROPERTY_TITLE) | 0);  /* check that WLC_BIT_PROPERTY_TITLE is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_PROPERTY_CLASS(unsigned long long *o)
{
  int n = (WLC_BIT_PROPERTY_CLASS) <= 0;
  *o = (unsigned long long)((WLC_BIT_PROPERTY_CLASS) | 0);  /* check that WLC_BIT_PROPERTY_CLASS is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_PROPERTY_APP_ID(unsigned long long *o)
{
  int n = (WLC_BIT_PROPERTY_APP_ID) <= 0;
  *o = (unsigned long long)((WLC_BIT_PROPERTY_APP_ID) | 0);  /* check that WLC_BIT_PROPERTY_APP_ID is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_PROPERTY_PID(unsigned long long *o)
{
  int n = (WLC_BIT_PROPERTY_PID) <= 0;
  *o = (unsigned long long)((WLC_BIT_PROPERTY_PID) | 0);  /* check that WLC_BIT_PROPERTY_PID is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_MAXIMIZED(unsigned long long *o)
{
  int n = (WLC_BIT_MAXIMIZED) <= 0;
  *o = (unsigned long long)((WLC_BIT_MAXIMIZED) | 0);  /* check that WLC_BIT_MAXIMIZED is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_FULLSCREEN(unsigned long long *o)
{
  int n = (WLC_BIT_FULLSCREEN) <= 0;
  *o = (unsigned long long)((WLC_BIT_FULLSCREEN) | 0);  /* check that WLC_BIT_FULLSCREEN is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_RESIZING(unsigned long long *o)
{
  int n = (WLC_BIT_RESIZING) <= 0;
  *o = (unsigned long long)((WLC_BIT_RESIZING) | 0);  /* check that WLC_BIT_RESIZING is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_MOVING(unsigned long long *o)
{
  int n = (WLC_BIT_MOVING) <= 0;
  *o = (unsigned long long)((WLC_BIT_MOVING) | 0);  /* check that WLC_BIT_MOVING is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_ACTIVATED(unsigned long long *o)
{
  int n = (WLC_BIT_ACTIVATED) <= 0;
  *o = (unsigned long long)((WLC_BIT_ACTIVATED) | 0);  /* check that WLC_BIT_ACTIVATED is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_OVERRIDE_REDIRECT(unsigned long long *o)
{
  int n = (WLC_BIT_OVERRIDE_REDIRECT) <= 0;
  *o = (unsigned long long)((WLC_BIT_OVERRIDE_REDIRECT) | 0);  /* check that WLC_BIT_OVERRIDE_REDIRECT is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_UNMANAGED(unsigned long long *o)
{
  int n = (WLC_BIT_UNMANAGED) <= 0;
  *o = (unsigned long long)((WLC_BIT_UNMANAGED) | 0);  /* check that WLC_BIT_UNMANAGED is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_SPLASH(unsigned long long *o)
{
  int n = (WLC_BIT_SPLASH) <= 0;
  *o = (unsigned long long)((WLC_BIT_SPLASH) | 0);  /* check that WLC_BIT_SPLASH is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_MODAL(unsigned long long *o)
{
  int n = (WLC_BIT_MODAL) <= 0;
  *o = (unsigned long long)((WLC_BIT_MODAL) | 0);  /* check that WLC_BIT_MODAL is an integer */
  return n;
}

static int _cffi_const_WLC_BIT_POPUP(unsigned long long *o)
{
  int n = (WLC_BIT_POPUP) <= 0;
  *o = (unsigned long long)((WLC_BIT_POPUP) | 0);  /* check that WLC_BIT_POPUP is an integer */
  return n;
}

static struct _cffi_externpy_s _cffi_externpy__keyboard_key =
  { "wlc.keyboard_key", (int)sizeof(_Bool) };

static _Bool keyboard_key(uintptr_t a0, uint32_t a1, struct wlc_modifiers * a2, uint32_t a3, enum wlc_key_state a4)
{
  char a[40];
  char *p = a;
  *(uintptr_t *)(p + 0) = a0;
  *(uint32_t *)(p + 8) = a1;
  *(struct wlc_modifiers * *)(p + 16) = a2;
  *(uint32_t *)(p + 24) = a3;
  *(enum wlc_key_state *)(p + 32) = a4;
  _cffi_call_python(&_cffi_externpy__keyboard_key, p);
  return *(_Bool *)p;
}

static struct _cffi_externpy_s _cffi_externpy__output_resolution =
  { "wlc.output_resolution", 0 };

static void output_resolution(uintptr_t a0, struct wlc_size * a1, struct wlc_size * a2)
{
  char a[24];
  char *p = a;
  *(uintptr_t *)(p + 0) = a0;
  *(struct wlc_size * *)(p + 8) = a1;
  *(struct wlc_size * *)(p + 16) = a2;
  _cffi_call_python(&_cffi_externpy__output_resolution, p);
}

static struct _cffi_externpy_s _cffi_externpy__pointer_button =
  { "wlc.pointer_button", (int)sizeof(_Bool) };

static _Bool pointer_button(uintptr_t a0, uint32_t a1, struct wlc_modifiers * a2, uint32_t a3, enum wlc_button_state a4, struct wlc_point * a5)
{
  char a[48];
  char *p = a;
  *(uintptr_t *)(p + 0) = a0;
  *(uint32_t *)(p + 8) = a1;
  *(struct wlc_modifiers * *)(p + 16) = a2;
  *(uint32_t *)(p + 24) = a3;
  *(enum wlc_button_state *)(p + 32) = a4;
  *(struct wlc_point * *)(p + 40) = a5;
  _cffi_call_python(&_cffi_externpy__pointer_button, p);
  return *(_Bool *)p;
}

static struct _cffi_externpy_s _cffi_externpy__pointer_motion =
  { "wlc.pointer_motion", (int)sizeof(_Bool) };

static _Bool pointer_motion(uintptr_t a0, uint32_t a1, struct wlc_point * a2)
{
  char a[24];
  char *p = a;
  *(uintptr_t *)(p + 0) = a0;
  *(uint32_t *)(p + 8) = a1;
  *(struct wlc_point * *)(p + 16) = a2;
  _cffi_call_python(&_cffi_externpy__pointer_motion, p);
  return *(_Bool *)p;
}

static struct _cffi_externpy_s _cffi_externpy__view_created =
  { "wlc.view_created", (int)sizeof(_Bool) };

static _Bool view_created(uintptr_t a0)
{
  char a[8];
  char *p = a;
  *(uintptr_t *)(p + 0) = a0;
  _cffi_call_python(&_cffi_externpy__view_created, p);
  return *(_Bool *)p;
}

static struct _cffi_externpy_s _cffi_externpy__view_destroyed =
  { "wlc.view_destroyed", 0 };

static void view_destroyed(uintptr_t a0)
{
  char a[8];
  char *p = a;
  *(uintptr_t *)(p + 0) = a0;
  _cffi_call_python(&_cffi_externpy__view_destroyed, p);
}

static struct _cffi_externpy_s _cffi_externpy__view_focus =
  { "wlc.view_focus", 0 };

static void view_focus(uintptr_t a0, _Bool a1)
{
  char a[16];
  char *p = a;
  *(uintptr_t *)(p + 0) = a0;
  *(_Bool *)(p + 8) = a1;
  _cffi_call_python(&_cffi_externpy__view_focus, p);
}

static struct _cffi_externpy_s _cffi_externpy__view_request_geometry =
  { "wlc.view_request_geometry", 0 };

static void view_request_geometry(uintptr_t a0, struct wlc_geometry * a1)
{
  char a[16];
  char *p = a;
  *(uintptr_t *)(p + 0) = a0;
  *(struct wlc_geometry * *)(p + 8) = a1;
  _cffi_call_python(&_cffi_externpy__view_request_geometry, p);
}

static struct _cffi_externpy_s _cffi_externpy__view_request_move =
  { "wlc.view_request_move", 0 };

static void view_request_move(uintptr_t a0, struct wlc_point * a1)
{
  char a[16];
  char *p = a;
  *(uintptr_t *)(p + 0) = a0;
  *(struct wlc_point * *)(p + 8) = a1;
  _cffi_call_python(&_cffi_externpy__view_request_move, p);
}

static struct _cffi_externpy_s _cffi_externpy__view_request_resize =
  { "wlc.view_request_resize", 0 };

static void view_request_resize(uintptr_t a0, uint32_t a1, struct wlc_point * a2)
{
  char a[24];
  char *p = a;
  *(uintptr_t *)(p + 0) = a0;
  *(uint32_t *)(p + 8) = a1;
  *(struct wlc_point * *)(p + 16) = a2;
  _cffi_call_python(&_cffi_externpy__view_request_resize, p);
}

static struct wlc_event_source * _cffi_d_wlc_event_loop_add_fd(int x0, uint32_t x1, int(* x2)(int, uint32_t, void *), void * x3)
{
  return wlc_event_loop_add_fd(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_event_loop_add_fd(PyObject *self, PyObject *args)
{
  int x0;
  uint32_t x1;
  int(* x2)(int, uint32_t, void *);
  void * x3;
  Py_ssize_t datasize;
  struct wlc_event_source * result;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "wlc_event_loop_add_fd", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  x0 = _cffi_to_c_int(arg0, int);
  if (x0 == (int)-1 && PyErr_Occurred())
    return NULL;

  x1 = _cffi_to_c_int(arg1, uint32_t);
  if (x1 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = (int(*)(int, uint32_t, void *))_cffi_to_c_pointer(arg2, _cffi_type(104));
  if (x2 == (int(*)(int, uint32_t, void *))NULL && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(89), arg3, (char **)&x3);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x3 = (void *)alloca((size_t)datasize);
    memset((void *)x3, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x3, _cffi_type(89), arg3) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_event_loop_add_fd(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(4));
}
#else
#  define _cffi_f_wlc_event_loop_add_fd _cffi_d_wlc_event_loop_add_fd
#endif

static struct wlc_event_source * _cffi_d_wlc_event_loop_add_timer(int(* x0)(void *), void * x1)
{
  return wlc_event_loop_add_timer(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_event_loop_add_timer(PyObject *self, PyObject *args)
{
  int(* x0)(void *);
  void * x1;
  Py_ssize_t datasize;
  struct wlc_event_source * result;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_event_loop_add_timer", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = (int(*)(void *))_cffi_to_c_pointer(arg0, _cffi_type(98));
  if (x0 == (int(*)(void *))NULL && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(89), arg1, (char **)&x1);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x1 = (void *)alloca((size_t)datasize);
    memset((void *)x1, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x1, _cffi_type(89), arg1) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_event_loop_add_timer(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(4));
}
#else
#  define _cffi_f_wlc_event_loop_add_timer _cffi_d_wlc_event_loop_add_timer
#endif

static void _cffi_d_wlc_event_source_remove(struct wlc_event_source * x0)
{
  wlc_event_source_remove(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_event_source_remove(PyObject *self, PyObject *arg0)
{
  struct wlc_event_source * x0;
  Py_ssize_t datasize;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(4), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (struct wlc_event_source *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(4), arg0) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_event_source_remove(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_event_source_remove _cffi_d_wlc_event_source_remove
#endif

static _Bool _cffi_d_wlc_event_source_timer_update(struct wlc_event_source * x0, int32_t x1)
{
  return wlc_event_source_timer_update(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_event_source_timer_update(PyObject *self, PyObject *args)
{
  struct wlc_event_source * x0;
  int32_t x1;
  Py_ssize_t datasize;
  _Bool result;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_event_source_timer_update", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(4), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (struct wlc_event_source *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(4), arg0) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int32_t);
  if (x1 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_event_source_timer_update(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, _Bool);
}
#else
#  define _cffi_f_wlc_event_source_timer_update _cffi_d_wlc_event_source_timer_update
#endif

static void _cffi_d_wlc_exec(char const * x0, char * const * x1)
{
  wlc_exec(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_exec(PyObject *self, PyObject *args)
{
  char const * x0;
  char * const * x1;
  Py_ssize_t datasize;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_exec", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(174), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (char const *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(174), arg0) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(175), arg1, (char **)&x1);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x1 = (char * const *)alloca((size_t)datasize);
    memset((void *)x1, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x1, _cffi_type(175), arg1) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_exec(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_exec _cffi_d_wlc_exec
#endif

static enum wlc_backend_type _cffi_d_wlc_get_backend_type(void)
{
  return wlc_get_backend_type();
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_get_backend_type(PyObject *self, PyObject *noarg)
{
  enum wlc_backend_type result;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_get_backend_type(); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  (void)noarg; /* unused */
  return _cffi_from_c_deref((char *)&result, _cffi_type(324));
}
#else
#  define _cffi_f_wlc_get_backend_type _cffi_d_wlc_get_backend_type
#endif

static uintptr_t _cffi_d_wlc_get_focused_output(void)
{
  return wlc_get_focused_output();
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_get_focused_output(PyObject *self, PyObject *noarg)
{
  uintptr_t result;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_get_focused_output(); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  (void)noarg; /* unused */
  return _cffi_from_c_int(result, uintptr_t);
}
#else
#  define _cffi_f_wlc_get_focused_output _cffi_d_wlc_get_focused_output
#endif

static uintptr_t const * _cffi_d_wlc_get_outputs(size_t * x0)
{
  return wlc_get_outputs(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_get_outputs(PyObject *self, PyObject *arg0)
{
  size_t * x0;
  Py_ssize_t datasize;
  uintptr_t const * result;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(124), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (size_t *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(124), arg0) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_get_outputs(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(67));
}
#else
#  define _cffi_f_wlc_get_outputs _cffi_d_wlc_get_outputs
#endif

static void * _cffi_d_wlc_handle_get_user_data(uintptr_t x0)
{
  return wlc_handle_get_user_data(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_handle_get_user_data(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  void * result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_handle_get_user_data(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(89));
}
#else
#  define _cffi_f_wlc_handle_get_user_data _cffi_d_wlc_handle_get_user_data
#endif

static void _cffi_d_wlc_handle_set_user_data(uintptr_t x0, void const * x1)
{
  wlc_handle_set_user_data(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_handle_set_user_data(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  void const * x1;
  Py_ssize_t datasize;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_handle_set_user_data", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(278), arg1, (char **)&x1);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x1 = (void const *)alloca((size_t)datasize);
    memset((void *)x1, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x1, _cffi_type(278), arg1) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_handle_set_user_data(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_handle_set_user_data _cffi_d_wlc_handle_set_user_data
#endif

static _Bool _cffi_d_wlc_init(void)
{
  return wlc_init();
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_init(PyObject *self, PyObject *noarg)
{
  _Bool result;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_init(); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  (void)noarg; /* unused */
  return _cffi_from_c_int(result, _Bool);
}
#else
#  define _cffi_f_wlc_init _cffi_d_wlc_init
#endif

static uint32_t const * _cffi_d_wlc_keyboard_get_current_keys(size_t * x0)
{
  return wlc_keyboard_get_current_keys(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_keyboard_get_current_keys(PyObject *self, PyObject *arg0)
{
  size_t * x0;
  Py_ssize_t datasize;
  uint32_t const * result;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(124), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (size_t *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(124), arg0) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_keyboard_get_current_keys(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(344));
}
#else
#  define _cffi_f_wlc_keyboard_get_current_keys _cffi_d_wlc_keyboard_get_current_keys
#endif

static uint32_t _cffi_d_wlc_keyboard_get_keysym_for_key(uint32_t x0, struct wlc_modifiers const * x1)
{
  return wlc_keyboard_get_keysym_for_key(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_keyboard_get_keysym_for_key(PyObject *self, PyObject *args)
{
  uint32_t x0;
  struct wlc_modifiers const * x1;
  Py_ssize_t datasize;
  uint32_t result;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_keyboard_get_keysym_for_key", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uint32_t);
  if (x0 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg1, (char **)&x1);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x1 = (struct wlc_modifiers const *)alloca((size_t)datasize);
    memset((void *)x1, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x1, _cffi_type(28), arg1) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_keyboard_get_keysym_for_key(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, uint32_t);
}
#else
#  define _cffi_f_wlc_keyboard_get_keysym_for_key _cffi_d_wlc_keyboard_get_keysym_for_key
#endif

static uint32_t _cffi_d_wlc_keyboard_get_utf32_for_key(uint32_t x0, struct wlc_modifiers const * x1)
{
  return wlc_keyboard_get_utf32_for_key(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_keyboard_get_utf32_for_key(PyObject *self, PyObject *args)
{
  uint32_t x0;
  struct wlc_modifiers const * x1;
  Py_ssize_t datasize;
  uint32_t result;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_keyboard_get_utf32_for_key", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uint32_t);
  if (x0 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg1, (char **)&x1);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x1 = (struct wlc_modifiers const *)alloca((size_t)datasize);
    memset((void *)x1, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x1, _cffi_type(28), arg1) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_keyboard_get_utf32_for_key(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, uint32_t);
}
#else
#  define _cffi_f_wlc_keyboard_get_utf32_for_key _cffi_d_wlc_keyboard_get_utf32_for_key
#endif

static struct xkb_keymap * _cffi_d_wlc_keyboard_get_xkb_keymap(void)
{
  return wlc_keyboard_get_xkb_keymap();
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_keyboard_get_xkb_keymap(PyObject *self, PyObject *noarg)
{
  struct xkb_keymap * result;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_keyboard_get_xkb_keymap(); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  (void)noarg; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(340));
}
#else
#  define _cffi_f_wlc_keyboard_get_xkb_keymap _cffi_d_wlc_keyboard_get_xkb_keymap
#endif

static struct xkb_state * _cffi_d_wlc_keyboard_get_xkb_state(void)
{
  return wlc_keyboard_get_xkb_state();
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_keyboard_get_xkb_state(PyObject *self, PyObject *noarg)
{
  struct xkb_state * result;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_keyboard_get_xkb_state(); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  (void)noarg; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(342));
}
#else
#  define _cffi_f_wlc_keyboard_get_xkb_state _cffi_d_wlc_keyboard_get_xkb_state
#endif

static void _cffi_d_wlc_log_set_handler(void(* x0)(enum wlc_log_type, char const *))
{
  wlc_log_set_handler(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_log_set_handler(PyObject *self, PyObject *arg0)
{
  void(* x0)(enum wlc_log_type, char const *);

  x0 = (void(*)(enum wlc_log_type, char const *))_cffi_to_c_pointer(arg0, _cffi_type(281));
  if (x0 == (void(*)(enum wlc_log_type, char const *))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_log_set_handler(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_log_set_handler _cffi_d_wlc_log_set_handler
#endif

static void _cffi_d_wlc_output_focus(uintptr_t x0)
{
  wlc_output_focus(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_focus(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_output_focus(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_output_focus _cffi_d_wlc_output_focus
#endif

static uint16_t _cffi_d_wlc_output_get_gamma_size(uintptr_t x0)
{
  return wlc_output_get_gamma_size(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_get_gamma_size(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  uint16_t result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_output_get_gamma_size(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, uint16_t);
}
#else
#  define _cffi_f_wlc_output_get_gamma_size _cffi_d_wlc_output_get_gamma_size
#endif

static uint32_t _cffi_d_wlc_output_get_mask(uintptr_t x0)
{
  return wlc_output_get_mask(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_get_mask(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  uint32_t result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_output_get_mask(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, uint32_t);
}
#else
#  define _cffi_f_wlc_output_get_mask _cffi_d_wlc_output_get_mask
#endif

static uintptr_t * _cffi_d_wlc_output_get_mutable_views(uintptr_t x0, size_t * x1)
{
  return wlc_output_get_mutable_views(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_get_mutable_views(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  size_t * x1;
  Py_ssize_t datasize;
  uintptr_t * result;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_output_get_mutable_views", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(124), arg1, (char **)&x1);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x1 = (size_t *)alloca((size_t)datasize);
    memset((void *)x1, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x1, _cffi_type(124), arg1) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_output_get_mutable_views(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(345));
}
#else
#  define _cffi_f_wlc_output_get_mutable_views _cffi_d_wlc_output_get_mutable_views
#endif

static char const * _cffi_d_wlc_output_get_name(uintptr_t x0)
{
  return wlc_output_get_name(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_get_name(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  char const * result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_output_get_name(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(174));
}
#else
#  define _cffi_f_wlc_output_get_name _cffi_d_wlc_output_get_name
#endif

static struct wlc_size const * _cffi_d_wlc_output_get_resolution(uintptr_t x0)
{
  return wlc_output_get_resolution(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_get_resolution(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  struct wlc_size const * result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_output_get_resolution(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(233));
}
#else
#  define _cffi_f_wlc_output_get_resolution _cffi_d_wlc_output_get_resolution
#endif

static uint32_t _cffi_d_wlc_output_get_scale(uintptr_t x0)
{
  return wlc_output_get_scale(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_get_scale(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  uint32_t result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_output_get_scale(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, uint32_t);
}
#else
#  define _cffi_f_wlc_output_get_scale _cffi_d_wlc_output_get_scale
#endif

static _Bool _cffi_d_wlc_output_get_sleep(uintptr_t x0)
{
  return wlc_output_get_sleep(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_get_sleep(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  _Bool result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_output_get_sleep(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, _Bool);
}
#else
#  define _cffi_f_wlc_output_get_sleep _cffi_d_wlc_output_get_sleep
#endif

static uintptr_t const * _cffi_d_wlc_output_get_views(uintptr_t x0, size_t * x1)
{
  return wlc_output_get_views(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_get_views(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  size_t * x1;
  Py_ssize_t datasize;
  uintptr_t const * result;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_output_get_views", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(124), arg1, (char **)&x1);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x1 = (size_t *)alloca((size_t)datasize);
    memset((void *)x1, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x1, _cffi_type(124), arg1) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_output_get_views(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(67));
}
#else
#  define _cffi_f_wlc_output_get_views _cffi_d_wlc_output_get_views
#endif

static struct wlc_size const * _cffi_d_wlc_output_get_virtual_resolution(uintptr_t x0)
{
  return wlc_output_get_virtual_resolution(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_get_virtual_resolution(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  struct wlc_size const * result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_output_get_virtual_resolution(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(233));
}
#else
#  define _cffi_f_wlc_output_get_virtual_resolution _cffi_d_wlc_output_get_virtual_resolution
#endif

static void _cffi_d_wlc_output_set_gamma(uintptr_t x0, uint16_t x1, uint16_t * x2, uint16_t * x3, uint16_t * x4)
{
  wlc_output_set_gamma(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_set_gamma(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  uint16_t x1;
  uint16_t * x2;
  uint16_t * x3;
  uint16_t * x4;
  Py_ssize_t datasize;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "wlc_output_set_gamma", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  x1 = _cffi_to_c_int(arg1, uint16_t);
  if (x1 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(244), arg2, (char **)&x2);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x2 = (uint16_t *)alloca((size_t)datasize);
    memset((void *)x2, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x2, _cffi_type(244), arg2) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(244), arg3, (char **)&x3);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x3 = (uint16_t *)alloca((size_t)datasize);
    memset((void *)x3, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x3, _cffi_type(244), arg3) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(244), arg4, (char **)&x4);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x4 = (uint16_t *)alloca((size_t)datasize);
    memset((void *)x4, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x4, _cffi_type(244), arg4) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_output_set_gamma(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_output_set_gamma _cffi_d_wlc_output_set_gamma
#endif

static void _cffi_d_wlc_output_set_mask(uintptr_t x0, uint32_t x1)
{
  wlc_output_set_mask(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_set_mask(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  uint32_t x1;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_output_set_mask", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  x1 = _cffi_to_c_int(arg1, uint32_t);
  if (x1 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_output_set_mask(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_output_set_mask _cffi_d_wlc_output_set_mask
#endif

static void _cffi_d_wlc_output_set_resolution(uintptr_t x0, struct wlc_size const * x1, uint32_t x2)
{
  wlc_output_set_resolution(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_set_resolution(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  struct wlc_size const * x1;
  uint32_t x2;
  Py_ssize_t datasize;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "wlc_output_set_resolution", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(233), arg1, (char **)&x1);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x1 = (struct wlc_size const *)alloca((size_t)datasize);
    memset((void *)x1, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x1, _cffi_type(233), arg1) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint32_t);
  if (x2 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_output_set_resolution(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_output_set_resolution _cffi_d_wlc_output_set_resolution
#endif

static void _cffi_d_wlc_output_set_sleep(uintptr_t x0, _Bool x1)
{
  wlc_output_set_sleep(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_set_sleep(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  _Bool x1;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_output_set_sleep", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  x1 = (_Bool)_cffi_to_c__Bool(arg1);
  if (x1 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_output_set_sleep(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_output_set_sleep _cffi_d_wlc_output_set_sleep
#endif

static _Bool _cffi_d_wlc_output_set_views(uintptr_t x0, uintptr_t const * x1, size_t x2)
{
  return wlc_output_set_views(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_output_set_views(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  uintptr_t const * x1;
  size_t x2;
  Py_ssize_t datasize;
  _Bool result;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "wlc_output_set_views", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(67), arg1, (char **)&x1);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x1 = (uintptr_t const *)alloca((size_t)datasize);
    memset((void *)x1, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x1, _cffi_type(67), arg1) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, size_t);
  if (x2 == (size_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_output_set_views(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, _Bool);
}
#else
#  define _cffi_f_wlc_output_set_views _cffi_d_wlc_output_set_views
#endif

static void _cffi_d_wlc_pointer_get_position(struct wlc_point * x0)
{
  wlc_pointer_get_position(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_pointer_get_position(PyObject *self, PyObject *arg0)
{
  struct wlc_point * x0;
  Py_ssize_t datasize;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(16), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (struct wlc_point *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(16), arg0) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_pointer_get_position(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_pointer_get_position _cffi_d_wlc_pointer_get_position
#endif

static void _cffi_d_wlc_pointer_set_position(struct wlc_point const * x0)
{
  wlc_pointer_set_position(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_pointer_set_position(PyObject *self, PyObject *arg0)
{
  struct wlc_point const * x0;
  Py_ssize_t datasize;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(31), arg0, (char **)&x0);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x0 = (struct wlc_point const *)alloca((size_t)datasize);
    memset((void *)x0, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x0, _cffi_type(31), arg0) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_pointer_set_position(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_pointer_set_position _cffi_d_wlc_pointer_set_position
#endif

static void _cffi_d_wlc_run(void)
{
  wlc_run();
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_run(PyObject *self, PyObject *noarg)
{

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_run(); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  (void)noarg; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_run _cffi_d_wlc_run
#endif

static void _cffi_d_wlc_set_compositor_ready_cb(void(* x0)(void))
{
  wlc_set_compositor_ready_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_compositor_ready_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(void);

  x0 = (void(*)(void))_cffi_to_c_pointer(arg0, _cffi_type(314));
  if (x0 == (void(*)(void))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_compositor_ready_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_compositor_ready_cb _cffi_d_wlc_set_compositor_ready_cb
#endif

static void _cffi_d_wlc_set_compositor_terminate_cb(void(* x0)(void))
{
  wlc_set_compositor_terminate_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_compositor_terminate_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(void);

  x0 = (void(*)(void))_cffi_to_c_pointer(arg0, _cffi_type(314));
  if (x0 == (void(*)(void))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_compositor_terminate_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_compositor_terminate_cb _cffi_d_wlc_set_compositor_terminate_cb
#endif

static void _cffi_d_wlc_set_input_created_cb(_Bool(* x0)(struct libinput_device *))
{
  wlc_set_input_created_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_input_created_cb(PyObject *self, PyObject *arg0)
{
  _Bool(* x0)(struct libinput_device *);

  x0 = (_Bool(*)(struct libinput_device *))_cffi_to_c_pointer(arg0, _cffi_type(153));
  if (x0 == (_Bool(*)(struct libinput_device *))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_input_created_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_input_created_cb _cffi_d_wlc_set_input_created_cb
#endif

static void _cffi_d_wlc_set_input_destroyed_cb(void(* x0)(struct libinput_device *))
{
  wlc_set_input_destroyed_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_input_destroyed_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(struct libinput_device *);

  x0 = (void(*)(struct libinput_device *))_cffi_to_c_pointer(arg0, _cffi_type(284));
  if (x0 == (void(*)(struct libinput_device *))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_input_destroyed_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_input_destroyed_cb _cffi_d_wlc_set_input_destroyed_cb
#endif

static void _cffi_d_wlc_set_keyboard_key_cb(_Bool(* x0)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_key_state))
{
  wlc_set_keyboard_key_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_keyboard_key_cb(PyObject *self, PyObject *arg0)
{
  _Bool(* x0)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_key_state);

  x0 = (_Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_key_state))_cffi_to_c_pointer(arg0, _cffi_type(165));
  if (x0 == (_Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_key_state))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_keyboard_key_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_keyboard_key_cb _cffi_d_wlc_set_keyboard_key_cb
#endif

static void _cffi_d_wlc_set_output_context_created_cb(void(* x0)(uintptr_t))
{
  wlc_set_output_context_created_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_output_context_created_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t);

  x0 = (void(*)(uintptr_t))_cffi_to_c_pointer(arg0, _cffi_type(287));
  if (x0 == (void(*)(uintptr_t))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_output_context_created_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_output_context_created_cb _cffi_d_wlc_set_output_context_created_cb
#endif

static void _cffi_d_wlc_set_output_context_destroyed_cb(void(* x0)(uintptr_t))
{
  wlc_set_output_context_destroyed_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_output_context_destroyed_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t);

  x0 = (void(*)(uintptr_t))_cffi_to_c_pointer(arg0, _cffi_type(287));
  if (x0 == (void(*)(uintptr_t))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_output_context_destroyed_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_output_context_destroyed_cb _cffi_d_wlc_set_output_context_destroyed_cb
#endif

static void _cffi_d_wlc_set_output_created_cb(_Bool(* x0)(uintptr_t))
{
  wlc_set_output_created_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_output_created_cb(PyObject *self, PyObject *arg0)
{
  _Bool(* x0)(uintptr_t);

  x0 = (_Bool(*)(uintptr_t))_cffi_to_c_pointer(arg0, _cffi_type(156));
  if (x0 == (_Bool(*)(uintptr_t))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_output_created_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_output_created_cb _cffi_d_wlc_set_output_created_cb
#endif

static void _cffi_d_wlc_set_output_destroyed_cb(void(* x0)(uintptr_t))
{
  wlc_set_output_destroyed_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_output_destroyed_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t);

  x0 = (void(*)(uintptr_t))_cffi_to_c_pointer(arg0, _cffi_type(287));
  if (x0 == (void(*)(uintptr_t))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_output_destroyed_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_output_destroyed_cb _cffi_d_wlc_set_output_destroyed_cb
#endif

static void _cffi_d_wlc_set_output_focus_cb(void(* x0)(uintptr_t, _Bool))
{
  wlc_set_output_focus_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_output_focus_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t, _Bool);

  x0 = (void(*)(uintptr_t, _Bool))_cffi_to_c_pointer(arg0, _cffi_type(290));
  if (x0 == (void(*)(uintptr_t, _Bool))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_output_focus_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_output_focus_cb _cffi_d_wlc_set_output_focus_cb
#endif

static void _cffi_d_wlc_set_output_render_post_cb(void(* x0)(uintptr_t))
{
  wlc_set_output_render_post_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_output_render_post_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t);

  x0 = (void(*)(uintptr_t))_cffi_to_c_pointer(arg0, _cffi_type(287));
  if (x0 == (void(*)(uintptr_t))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_output_render_post_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_output_render_post_cb _cffi_d_wlc_set_output_render_post_cb
#endif

static void _cffi_d_wlc_set_output_render_pre_cb(void(* x0)(uintptr_t))
{
  wlc_set_output_render_pre_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_output_render_pre_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t);

  x0 = (void(*)(uintptr_t))_cffi_to_c_pointer(arg0, _cffi_type(287));
  if (x0 == (void(*)(uintptr_t))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_output_render_pre_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_output_render_pre_cb _cffi_d_wlc_set_output_render_pre_cb
#endif

static void _cffi_d_wlc_set_output_resolution_cb(void(* x0)(uintptr_t, struct wlc_size const *, struct wlc_size const *))
{
  wlc_set_output_resolution_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_output_resolution_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t, struct wlc_size const *, struct wlc_size const *);

  x0 = (void(*)(uintptr_t, struct wlc_size const *, struct wlc_size const *))_cffi_to_c_pointer(arg0, _cffi_type(302));
  if (x0 == (void(*)(uintptr_t, struct wlc_size const *, struct wlc_size const *))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_output_resolution_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_output_resolution_cb _cffi_d_wlc_set_output_resolution_cb
#endif

static void _cffi_d_wlc_set_pointer_button_cb(_Bool(* x0)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_button_state, struct wlc_point const *))
{
  wlc_set_pointer_button_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_pointer_button_cb(PyObject *self, PyObject *arg0)
{
  _Bool(* x0)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_button_state, struct wlc_point const *);

  x0 = (_Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_button_state, struct wlc_point const *))_cffi_to_c_pointer(arg0, _cffi_type(162));
  if (x0 == (_Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint32_t, enum wlc_button_state, struct wlc_point const *))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_pointer_button_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_pointer_button_cb _cffi_d_wlc_set_pointer_button_cb
#endif

static void _cffi_d_wlc_set_pointer_motion_cb(_Bool(* x0)(uintptr_t, uint32_t, struct wlc_point const *))
{
  wlc_set_pointer_motion_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_pointer_motion_cb(PyObject *self, PyObject *arg0)
{
  _Bool(* x0)(uintptr_t, uint32_t, struct wlc_point const *);

  x0 = (_Bool(*)(uintptr_t, uint32_t, struct wlc_point const *))_cffi_to_c_pointer(arg0, _cffi_type(171));
  if (x0 == (_Bool(*)(uintptr_t, uint32_t, struct wlc_point const *))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_pointer_motion_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_pointer_motion_cb _cffi_d_wlc_set_pointer_motion_cb
#endif

static void _cffi_d_wlc_set_pointer_scroll_cb(_Bool(* x0)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint8_t, double *))
{
  wlc_set_pointer_scroll_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_pointer_scroll_cb(PyObject *self, PyObject *arg0)
{
  _Bool(* x0)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint8_t, double *);

  x0 = (_Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint8_t, double *))_cffi_to_c_pointer(arg0, _cffi_type(168));
  if (x0 == (_Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, uint8_t, double *))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_pointer_scroll_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_pointer_scroll_cb _cffi_d_wlc_set_pointer_scroll_cb
#endif

static void _cffi_d_wlc_set_touch_cb(_Bool(* x0)(uintptr_t, uint32_t, struct wlc_modifiers const *, enum wlc_touch_type, int32_t, struct wlc_point const *))
{
  wlc_set_touch_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_touch_cb(PyObject *self, PyObject *arg0)
{
  _Bool(* x0)(uintptr_t, uint32_t, struct wlc_modifiers const *, enum wlc_touch_type, int32_t, struct wlc_point const *);

  x0 = (_Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, enum wlc_touch_type, int32_t, struct wlc_point const *))_cffi_to_c_pointer(arg0, _cffi_type(159));
  if (x0 == (_Bool(*)(uintptr_t, uint32_t, struct wlc_modifiers const *, enum wlc_touch_type, int32_t, struct wlc_point const *))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_touch_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_touch_cb _cffi_d_wlc_set_touch_cb
#endif

static void _cffi_d_wlc_set_view_created_cb(_Bool(* x0)(uintptr_t))
{
  wlc_set_view_created_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_view_created_cb(PyObject *self, PyObject *arg0)
{
  _Bool(* x0)(uintptr_t);

  x0 = (_Bool(*)(uintptr_t))_cffi_to_c_pointer(arg0, _cffi_type(156));
  if (x0 == (_Bool(*)(uintptr_t))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_view_created_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_view_created_cb _cffi_d_wlc_set_view_created_cb
#endif

static void _cffi_d_wlc_set_view_destroyed_cb(void(* x0)(uintptr_t))
{
  wlc_set_view_destroyed_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_view_destroyed_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t);

  x0 = (void(*)(uintptr_t))_cffi_to_c_pointer(arg0, _cffi_type(287));
  if (x0 == (void(*)(uintptr_t))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_view_destroyed_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_view_destroyed_cb _cffi_d_wlc_set_view_destroyed_cb
#endif

static void _cffi_d_wlc_set_view_focus_cb(void(* x0)(uintptr_t, _Bool))
{
  wlc_set_view_focus_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_view_focus_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t, _Bool);

  x0 = (void(*)(uintptr_t, _Bool))_cffi_to_c_pointer(arg0, _cffi_type(290));
  if (x0 == (void(*)(uintptr_t, _Bool))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_view_focus_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_view_focus_cb _cffi_d_wlc_set_view_focus_cb
#endif

static void _cffi_d_wlc_set_view_move_to_output_cb(void(* x0)(uintptr_t, uintptr_t, uintptr_t))
{
  wlc_set_view_move_to_output_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_view_move_to_output_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t, uintptr_t, uintptr_t);

  x0 = (void(*)(uintptr_t, uintptr_t, uintptr_t))_cffi_to_c_pointer(arg0, _cffi_type(311));
  if (x0 == (void(*)(uintptr_t, uintptr_t, uintptr_t))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_view_move_to_output_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_view_move_to_output_cb _cffi_d_wlc_set_view_move_to_output_cb
#endif

static void _cffi_d_wlc_set_view_properties_updated_cb(void(* x0)(uintptr_t, uint32_t))
{
  wlc_set_view_properties_updated_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_view_properties_updated_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t, uint32_t);

  x0 = (void(*)(uintptr_t, uint32_t))_cffi_to_c_pointer(arg0, _cffi_type(305));
  if (x0 == (void(*)(uintptr_t, uint32_t))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_view_properties_updated_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_view_properties_updated_cb _cffi_d_wlc_set_view_properties_updated_cb
#endif

static void _cffi_d_wlc_set_view_render_post_cb(void(* x0)(uintptr_t))
{
  wlc_set_view_render_post_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_view_render_post_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t);

  x0 = (void(*)(uintptr_t))_cffi_to_c_pointer(arg0, _cffi_type(287));
  if (x0 == (void(*)(uintptr_t))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_view_render_post_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_view_render_post_cb _cffi_d_wlc_set_view_render_post_cb
#endif

static void _cffi_d_wlc_set_view_render_pre_cb(void(* x0)(uintptr_t))
{
  wlc_set_view_render_pre_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_view_render_pre_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t);

  x0 = (void(*)(uintptr_t))_cffi_to_c_pointer(arg0, _cffi_type(287));
  if (x0 == (void(*)(uintptr_t))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_view_render_pre_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_view_render_pre_cb _cffi_d_wlc_set_view_render_pre_cb
#endif

static void _cffi_d_wlc_set_view_request_geometry_cb(void(* x0)(uintptr_t, struct wlc_geometry const *))
{
  wlc_set_view_request_geometry_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_view_request_geometry_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t, struct wlc_geometry const *);

  x0 = (void(*)(uintptr_t, struct wlc_geometry const *))_cffi_to_c_pointer(arg0, _cffi_type(296));
  if (x0 == (void(*)(uintptr_t, struct wlc_geometry const *))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_view_request_geometry_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_view_request_geometry_cb _cffi_d_wlc_set_view_request_geometry_cb
#endif

static void _cffi_d_wlc_set_view_request_move_cb(void(* x0)(uintptr_t, struct wlc_point const *))
{
  wlc_set_view_request_move_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_view_request_move_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t, struct wlc_point const *);

  x0 = (void(*)(uintptr_t, struct wlc_point const *))_cffi_to_c_pointer(arg0, _cffi_type(299));
  if (x0 == (void(*)(uintptr_t, struct wlc_point const *))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_view_request_move_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_view_request_move_cb _cffi_d_wlc_set_view_request_move_cb
#endif

static void _cffi_d_wlc_set_view_request_resize_cb(void(* x0)(uintptr_t, uint32_t, struct wlc_point const *))
{
  wlc_set_view_request_resize_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_view_request_resize_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t, uint32_t, struct wlc_point const *);

  x0 = (void(*)(uintptr_t, uint32_t, struct wlc_point const *))_cffi_to_c_pointer(arg0, _cffi_type(308));
  if (x0 == (void(*)(uintptr_t, uint32_t, struct wlc_point const *))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_view_request_resize_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_view_request_resize_cb _cffi_d_wlc_set_view_request_resize_cb
#endif

static void _cffi_d_wlc_set_view_request_state_cb(void(* x0)(uintptr_t, enum wlc_view_state_bit, _Bool))
{
  wlc_set_view_request_state_cb(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_set_view_request_state_cb(PyObject *self, PyObject *arg0)
{
  void(* x0)(uintptr_t, enum wlc_view_state_bit, _Bool);

  x0 = (void(*)(uintptr_t, enum wlc_view_state_bit, _Bool))_cffi_to_c_pointer(arg0, _cffi_type(293));
  if (x0 == (void(*)(uintptr_t, enum wlc_view_state_bit, _Bool))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_set_view_request_state_cb(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_set_view_request_state_cb _cffi_d_wlc_set_view_request_state_cb
#endif

static void _cffi_d_wlc_terminate(void)
{
  wlc_terminate();
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_terminate(PyObject *self, PyObject *noarg)
{

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_terminate(); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  (void)noarg; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_terminate _cffi_d_wlc_terminate
#endif

static void _cffi_d_wlc_view_bring_above(uintptr_t x0, uintptr_t x1)
{
  wlc_view_bring_above(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_bring_above(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  uintptr_t x1;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_view_bring_above", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  x1 = _cffi_to_c_int(arg1, uintptr_t);
  if (x1 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_view_bring_above(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_view_bring_above _cffi_d_wlc_view_bring_above
#endif

static void _cffi_d_wlc_view_bring_to_front(uintptr_t x0)
{
  wlc_view_bring_to_front(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_bring_to_front(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_view_bring_to_front(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_view_bring_to_front _cffi_d_wlc_view_bring_to_front
#endif

static void _cffi_d_wlc_view_close(uintptr_t x0)
{
  wlc_view_close(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_close(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_view_close(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_view_close _cffi_d_wlc_view_close
#endif

static void _cffi_d_wlc_view_focus(uintptr_t x0)
{
  wlc_view_focus(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_focus(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_view_focus(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_view_focus _cffi_d_wlc_view_focus
#endif

static char const * _cffi_d_wlc_view_get_app_id(uintptr_t x0)
{
  return wlc_view_get_app_id(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_get_app_id(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  char const * result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_get_app_id(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(174));
}
#else
#  define _cffi_f_wlc_view_get_app_id _cffi_d_wlc_view_get_app_id
#endif

static char const * _cffi_d_wlc_view_get_class(uintptr_t x0)
{
  return wlc_view_get_class(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_get_class(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  char const * result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_get_class(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(174));
}
#else
#  define _cffi_f_wlc_view_get_class _cffi_d_wlc_view_get_class
#endif

static struct wlc_geometry const * _cffi_d_wlc_view_get_geometry(uintptr_t x0)
{
  return wlc_view_get_geometry(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_get_geometry(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  struct wlc_geometry const * result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_get_geometry(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(216));
}
#else
#  define _cffi_f_wlc_view_get_geometry _cffi_d_wlc_view_get_geometry
#endif

static char const * _cffi_d_wlc_view_get_instance(uintptr_t x0)
{
  return wlc_view_get_instance(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_get_instance(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  char const * result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_get_instance(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(174));
}
#else
#  define _cffi_f_wlc_view_get_instance _cffi_d_wlc_view_get_instance
#endif

static uint32_t _cffi_d_wlc_view_get_mask(uintptr_t x0)
{
  return wlc_view_get_mask(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_get_mask(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  uint32_t result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_get_mask(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, uint32_t);
}
#else
#  define _cffi_f_wlc_view_get_mask _cffi_d_wlc_view_get_mask
#endif

static uintptr_t _cffi_d_wlc_view_get_output(uintptr_t x0)
{
  return wlc_view_get_output(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_get_output(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  uintptr_t result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_get_output(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, uintptr_t);
}
#else
#  define _cffi_f_wlc_view_get_output _cffi_d_wlc_view_get_output
#endif

static uintptr_t _cffi_d_wlc_view_get_parent(uintptr_t x0)
{
  return wlc_view_get_parent(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_get_parent(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  uintptr_t result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_get_parent(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, uintptr_t);
}
#else
#  define _cffi_f_wlc_view_get_parent _cffi_d_wlc_view_get_parent
#endif

static int _cffi_d_wlc_view_get_pid(uintptr_t x0)
{
  return wlc_view_get_pid(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_get_pid(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  int result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_get_pid(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, int);
}
#else
#  define _cffi_f_wlc_view_get_pid _cffi_d_wlc_view_get_pid
#endif

static uint32_t _cffi_d_wlc_view_get_state(uintptr_t x0)
{
  return wlc_view_get_state(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_get_state(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  uint32_t result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_get_state(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, uint32_t);
}
#else
#  define _cffi_f_wlc_view_get_state _cffi_d_wlc_view_get_state
#endif

static char const * _cffi_d_wlc_view_get_title(uintptr_t x0)
{
  return wlc_view_get_title(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_get_title(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  char const * result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_get_title(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(174));
}
#else
#  define _cffi_f_wlc_view_get_title _cffi_d_wlc_view_get_title
#endif

static uint32_t _cffi_d_wlc_view_get_type(uintptr_t x0)
{
  return wlc_view_get_type(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_get_type(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  uint32_t result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_get_type(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_int(result, uint32_t);
}
#else
#  define _cffi_f_wlc_view_get_type _cffi_d_wlc_view_get_type
#endif

static void _cffi_d_wlc_view_get_visible_geometry(uintptr_t x0, struct wlc_geometry * x1)
{
  wlc_view_get_visible_geometry(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_get_visible_geometry(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  struct wlc_geometry * x1;
  Py_ssize_t datasize;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_view_get_visible_geometry", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(212), arg1, (char **)&x1);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x1 = (struct wlc_geometry *)alloca((size_t)datasize);
    memset((void *)x1, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x1, _cffi_type(212), arg1) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_view_get_visible_geometry(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_view_get_visible_geometry _cffi_d_wlc_view_get_visible_geometry
#endif

static enum wlc_positioner_anchor_bit _cffi_d_wlc_view_positioner_get_anchor(uintptr_t x0)
{
  return wlc_view_positioner_get_anchor(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_positioner_get_anchor(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  enum wlc_positioner_anchor_bit result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_positioner_get_anchor(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_deref((char *)&result, _cffi_type(328));
}
#else
#  define _cffi_f_wlc_view_positioner_get_anchor _cffi_d_wlc_view_positioner_get_anchor
#endif

static struct wlc_geometry const * _cffi_d_wlc_view_positioner_get_anchor_rect(uintptr_t x0)
{
  return wlc_view_positioner_get_anchor_rect(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_positioner_get_anchor_rect(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  struct wlc_geometry const * result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_positioner_get_anchor_rect(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(216));
}
#else
#  define _cffi_f_wlc_view_positioner_get_anchor_rect _cffi_d_wlc_view_positioner_get_anchor_rect
#endif

static enum wlc_positioner_constraint_adjustment_bit _cffi_d_wlc_view_positioner_get_constraint_adjustment(uintptr_t x0)
{
  return wlc_view_positioner_get_constraint_adjustment(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_positioner_get_constraint_adjustment(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  enum wlc_positioner_constraint_adjustment_bit result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_positioner_get_constraint_adjustment(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_deref((char *)&result, _cffi_type(329));
}
#else
#  define _cffi_f_wlc_view_positioner_get_constraint_adjustment _cffi_d_wlc_view_positioner_get_constraint_adjustment
#endif

static enum wlc_positioner_gravity_bit _cffi_d_wlc_view_positioner_get_gravity(uintptr_t x0)
{
  return wlc_view_positioner_get_gravity(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_positioner_get_gravity(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  enum wlc_positioner_gravity_bit result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_positioner_get_gravity(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_deref((char *)&result, _cffi_type(330));
}
#else
#  define _cffi_f_wlc_view_positioner_get_gravity _cffi_d_wlc_view_positioner_get_gravity
#endif

static struct wlc_point const * _cffi_d_wlc_view_positioner_get_offset(uintptr_t x0)
{
  return wlc_view_positioner_get_offset(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_positioner_get_offset(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  struct wlc_point const * result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_positioner_get_offset(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(31));
}
#else
#  define _cffi_f_wlc_view_positioner_get_offset _cffi_d_wlc_view_positioner_get_offset
#endif

static struct wlc_size const * _cffi_d_wlc_view_positioner_get_size(uintptr_t x0)
{
  return wlc_view_positioner_get_size(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_positioner_get_size(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;
  struct wlc_size const * result;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = wlc_view_positioner_get_size(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  return _cffi_from_c_pointer((char *)result, _cffi_type(233));
}
#else
#  define _cffi_f_wlc_view_positioner_get_size _cffi_d_wlc_view_positioner_get_size
#endif

static void _cffi_d_wlc_view_send_below(uintptr_t x0, uintptr_t x1)
{
  wlc_view_send_below(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_send_below(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  uintptr_t x1;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_view_send_below", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  x1 = _cffi_to_c_int(arg1, uintptr_t);
  if (x1 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_view_send_below(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_view_send_below _cffi_d_wlc_view_send_below
#endif

static void _cffi_d_wlc_view_send_to_back(uintptr_t x0)
{
  wlc_view_send_to_back(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_send_to_back(PyObject *self, PyObject *arg0)
{
  uintptr_t x0;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_view_send_to_back(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_view_send_to_back _cffi_d_wlc_view_send_to_back
#endif

static void _cffi_d_wlc_view_set_geometry(uintptr_t x0, uint32_t x1, struct wlc_geometry const * x2)
{
  wlc_view_set_geometry(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_set_geometry(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  uint32_t x1;
  struct wlc_geometry const * x2;
  Py_ssize_t datasize;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "wlc_view_set_geometry", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  x1 = _cffi_to_c_int(arg1, uint32_t);
  if (x1 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(216), arg2, (char **)&x2);
  if (datasize != 0) {
    if (datasize < 0)
      return NULL;
    x2 = (struct wlc_geometry const *)alloca((size_t)datasize);
    memset((void *)x2, 0, (size_t)datasize);
    if (_cffi_convert_array_from_object((char *)x2, _cffi_type(216), arg2) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_view_set_geometry(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_view_set_geometry _cffi_d_wlc_view_set_geometry
#endif

static void _cffi_d_wlc_view_set_mask(uintptr_t x0, uint32_t x1)
{
  wlc_view_set_mask(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_set_mask(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  uint32_t x1;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_view_set_mask", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  x1 = _cffi_to_c_int(arg1, uint32_t);
  if (x1 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_view_set_mask(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_view_set_mask _cffi_d_wlc_view_set_mask
#endif

static void _cffi_d_wlc_view_set_output(uintptr_t x0, uintptr_t x1)
{
  wlc_view_set_output(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_set_output(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  uintptr_t x1;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_view_set_output", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  x1 = _cffi_to_c_int(arg1, uintptr_t);
  if (x1 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_view_set_output(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_view_set_output _cffi_d_wlc_view_set_output
#endif

static void _cffi_d_wlc_view_set_parent(uintptr_t x0, uintptr_t x1)
{
  wlc_view_set_parent(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_set_parent(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  uintptr_t x1;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "wlc_view_set_parent", 2, 2, &arg0, &arg1))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  x1 = _cffi_to_c_int(arg1, uintptr_t);
  if (x1 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_view_set_parent(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_view_set_parent _cffi_d_wlc_view_set_parent
#endif

static void _cffi_d_wlc_view_set_state(uintptr_t x0, enum wlc_view_state_bit x1, _Bool x2)
{
  wlc_view_set_state(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_set_state(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  enum wlc_view_state_bit x1;
  _Bool x2;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "wlc_view_set_state", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x1, _cffi_type(202), arg1) < 0)
    return NULL;

  x2 = (_Bool)_cffi_to_c__Bool(arg2);
  if (x2 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_view_set_state(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_view_set_state _cffi_d_wlc_view_set_state
#endif

static void _cffi_d_wlc_view_set_type(uintptr_t x0, enum wlc_view_type_bit x1, _Bool x2)
{
  wlc_view_set_type(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_wlc_view_set_type(PyObject *self, PyObject *args)
{
  uintptr_t x0;
  enum wlc_view_type_bit x1;
  _Bool x2;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "wlc_view_set_type", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  x0 = _cffi_to_c_int(arg0, uintptr_t);
  if (x0 == (uintptr_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x1, _cffi_type(207), arg1) < 0)
    return NULL;

  x2 = (_Bool)_cffi_to_c__Bool(arg2);
  if (x2 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { wlc_view_set_type(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_wlc_view_set_type _cffi_d_wlc_view_set_type
#endif

static int _cffi_const_ABS_BRAKE(unsigned long long *o)
{
  int n = (ABS_BRAKE) <= 0;
  *o = (unsigned long long)((ABS_BRAKE) | 0);  /* check that ABS_BRAKE is an integer */
  if (!_cffi_check_int(*o, n, 10U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_DISTANCE(unsigned long long *o)
{
  int n = (ABS_DISTANCE) <= 0;
  *o = (unsigned long long)((ABS_DISTANCE) | 0);  /* check that ABS_DISTANCE is an integer */
  if (!_cffi_check_int(*o, n, 25U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_GAS(unsigned long long *o)
{
  int n = (ABS_GAS) <= 0;
  *o = (unsigned long long)((ABS_GAS) | 0);  /* check that ABS_GAS is an integer */
  if (!_cffi_check_int(*o, n, 9U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_HAT0X(unsigned long long *o)
{
  int n = (ABS_HAT0X) <= 0;
  *o = (unsigned long long)((ABS_HAT0X) | 0);  /* check that ABS_HAT0X is an integer */
  if (!_cffi_check_int(*o, n, 16U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_HAT0Y(unsigned long long *o)
{
  int n = (ABS_HAT0Y) <= 0;
  *o = (unsigned long long)((ABS_HAT0Y) | 0);  /* check that ABS_HAT0Y is an integer */
  if (!_cffi_check_int(*o, n, 17U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_HAT1X(unsigned long long *o)
{
  int n = (ABS_HAT1X) <= 0;
  *o = (unsigned long long)((ABS_HAT1X) | 0);  /* check that ABS_HAT1X is an integer */
  if (!_cffi_check_int(*o, n, 18U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_HAT1Y(unsigned long long *o)
{
  int n = (ABS_HAT1Y) <= 0;
  *o = (unsigned long long)((ABS_HAT1Y) | 0);  /* check that ABS_HAT1Y is an integer */
  if (!_cffi_check_int(*o, n, 19U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_HAT2X(unsigned long long *o)
{
  int n = (ABS_HAT2X) <= 0;
  *o = (unsigned long long)((ABS_HAT2X) | 0);  /* check that ABS_HAT2X is an integer */
  if (!_cffi_check_int(*o, n, 20U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_HAT2Y(unsigned long long *o)
{
  int n = (ABS_HAT2Y) <= 0;
  *o = (unsigned long long)((ABS_HAT2Y) | 0);  /* check that ABS_HAT2Y is an integer */
  if (!_cffi_check_int(*o, n, 21U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_HAT3X(unsigned long long *o)
{
  int n = (ABS_HAT3X) <= 0;
  *o = (unsigned long long)((ABS_HAT3X) | 0);  /* check that ABS_HAT3X is an integer */
  if (!_cffi_check_int(*o, n, 22U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_HAT3Y(unsigned long long *o)
{
  int n = (ABS_HAT3Y) <= 0;
  *o = (unsigned long long)((ABS_HAT3Y) | 0);  /* check that ABS_HAT3Y is an integer */
  if (!_cffi_check_int(*o, n, 23U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MAX(unsigned long long *o)
{
  int n = (ABS_MAX) <= 0;
  *o = (unsigned long long)((ABS_MAX) | 0);  /* check that ABS_MAX is an integer */
  if (!_cffi_check_int(*o, n, 63U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MISC(unsigned long long *o)
{
  int n = (ABS_MISC) <= 0;
  *o = (unsigned long long)((ABS_MISC) | 0);  /* check that ABS_MISC is an integer */
  if (!_cffi_check_int(*o, n, 40U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_BLOB_ID(unsigned long long *o)
{
  int n = (ABS_MT_BLOB_ID) <= 0;
  *o = (unsigned long long)((ABS_MT_BLOB_ID) | 0);  /* check that ABS_MT_BLOB_ID is an integer */
  if (!_cffi_check_int(*o, n, 56U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_DISTANCE(unsigned long long *o)
{
  int n = (ABS_MT_DISTANCE) <= 0;
  *o = (unsigned long long)((ABS_MT_DISTANCE) | 0);  /* check that ABS_MT_DISTANCE is an integer */
  if (!_cffi_check_int(*o, n, 59U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_ORIENTATION(unsigned long long *o)
{
  int n = (ABS_MT_ORIENTATION) <= 0;
  *o = (unsigned long long)((ABS_MT_ORIENTATION) | 0);  /* check that ABS_MT_ORIENTATION is an integer */
  if (!_cffi_check_int(*o, n, 52U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_POSITION_X(unsigned long long *o)
{
  int n = (ABS_MT_POSITION_X) <= 0;
  *o = (unsigned long long)((ABS_MT_POSITION_X) | 0);  /* check that ABS_MT_POSITION_X is an integer */
  if (!_cffi_check_int(*o, n, 53U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_POSITION_Y(unsigned long long *o)
{
  int n = (ABS_MT_POSITION_Y) <= 0;
  *o = (unsigned long long)((ABS_MT_POSITION_Y) | 0);  /* check that ABS_MT_POSITION_Y is an integer */
  if (!_cffi_check_int(*o, n, 54U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_PRESSURE(unsigned long long *o)
{
  int n = (ABS_MT_PRESSURE) <= 0;
  *o = (unsigned long long)((ABS_MT_PRESSURE) | 0);  /* check that ABS_MT_PRESSURE is an integer */
  if (!_cffi_check_int(*o, n, 58U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_SLOT(unsigned long long *o)
{
  int n = (ABS_MT_SLOT) <= 0;
  *o = (unsigned long long)((ABS_MT_SLOT) | 0);  /* check that ABS_MT_SLOT is an integer */
  if (!_cffi_check_int(*o, n, 47U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_TOOL_TYPE(unsigned long long *o)
{
  int n = (ABS_MT_TOOL_TYPE) <= 0;
  *o = (unsigned long long)((ABS_MT_TOOL_TYPE) | 0);  /* check that ABS_MT_TOOL_TYPE is an integer */
  if (!_cffi_check_int(*o, n, 55U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_TOOL_X(unsigned long long *o)
{
  int n = (ABS_MT_TOOL_X) <= 0;
  *o = (unsigned long long)((ABS_MT_TOOL_X) | 0);  /* check that ABS_MT_TOOL_X is an integer */
  if (!_cffi_check_int(*o, n, 60U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_TOOL_Y(unsigned long long *o)
{
  int n = (ABS_MT_TOOL_Y) <= 0;
  *o = (unsigned long long)((ABS_MT_TOOL_Y) | 0);  /* check that ABS_MT_TOOL_Y is an integer */
  if (!_cffi_check_int(*o, n, 61U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_TOUCH_MAJOR(unsigned long long *o)
{
  int n = (ABS_MT_TOUCH_MAJOR) <= 0;
  *o = (unsigned long long)((ABS_MT_TOUCH_MAJOR) | 0);  /* check that ABS_MT_TOUCH_MAJOR is an integer */
  if (!_cffi_check_int(*o, n, 48U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_TOUCH_MINOR(unsigned long long *o)
{
  int n = (ABS_MT_TOUCH_MINOR) <= 0;
  *o = (unsigned long long)((ABS_MT_TOUCH_MINOR) | 0);  /* check that ABS_MT_TOUCH_MINOR is an integer */
  if (!_cffi_check_int(*o, n, 49U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_TRACKING_ID(unsigned long long *o)
{
  int n = (ABS_MT_TRACKING_ID) <= 0;
  *o = (unsigned long long)((ABS_MT_TRACKING_ID) | 0);  /* check that ABS_MT_TRACKING_ID is an integer */
  if (!_cffi_check_int(*o, n, 57U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_WIDTH_MAJOR(unsigned long long *o)
{
  int n = (ABS_MT_WIDTH_MAJOR) <= 0;
  *o = (unsigned long long)((ABS_MT_WIDTH_MAJOR) | 0);  /* check that ABS_MT_WIDTH_MAJOR is an integer */
  if (!_cffi_check_int(*o, n, 50U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_MT_WIDTH_MINOR(unsigned long long *o)
{
  int n = (ABS_MT_WIDTH_MINOR) <= 0;
  *o = (unsigned long long)((ABS_MT_WIDTH_MINOR) | 0);  /* check that ABS_MT_WIDTH_MINOR is an integer */
  if (!_cffi_check_int(*o, n, 51U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_PRESSURE(unsigned long long *o)
{
  int n = (ABS_PRESSURE) <= 0;
  *o = (unsigned long long)((ABS_PRESSURE) | 0);  /* check that ABS_PRESSURE is an integer */
  if (!_cffi_check_int(*o, n, 24U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_RUDDER(unsigned long long *o)
{
  int n = (ABS_RUDDER) <= 0;
  *o = (unsigned long long)((ABS_RUDDER) | 0);  /* check that ABS_RUDDER is an integer */
  if (!_cffi_check_int(*o, n, 7U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_RX(unsigned long long *o)
{
  int n = (ABS_RX) <= 0;
  *o = (unsigned long long)((ABS_RX) | 0);  /* check that ABS_RX is an integer */
  if (!_cffi_check_int(*o, n, 3U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_RY(unsigned long long *o)
{
  int n = (ABS_RY) <= 0;
  *o = (unsigned long long)((ABS_RY) | 0);  /* check that ABS_RY is an integer */
  if (!_cffi_check_int(*o, n, 4U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_RZ(unsigned long long *o)
{
  int n = (ABS_RZ) <= 0;
  *o = (unsigned long long)((ABS_RZ) | 0);  /* check that ABS_RZ is an integer */
  if (!_cffi_check_int(*o, n, 5U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_THROTTLE(unsigned long long *o)
{
  int n = (ABS_THROTTLE) <= 0;
  *o = (unsigned long long)((ABS_THROTTLE) | 0);  /* check that ABS_THROTTLE is an integer */
  if (!_cffi_check_int(*o, n, 6U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_TILT_X(unsigned long long *o)
{
  int n = (ABS_TILT_X) <= 0;
  *o = (unsigned long long)((ABS_TILT_X) | 0);  /* check that ABS_TILT_X is an integer */
  if (!_cffi_check_int(*o, n, 26U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_TILT_Y(unsigned long long *o)
{
  int n = (ABS_TILT_Y) <= 0;
  *o = (unsigned long long)((ABS_TILT_Y) | 0);  /* check that ABS_TILT_Y is an integer */
  if (!_cffi_check_int(*o, n, 27U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_TOOL_WIDTH(unsigned long long *o)
{
  int n = (ABS_TOOL_WIDTH) <= 0;
  *o = (unsigned long long)((ABS_TOOL_WIDTH) | 0);  /* check that ABS_TOOL_WIDTH is an integer */
  if (!_cffi_check_int(*o, n, 28U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_VOLUME(unsigned long long *o)
{
  int n = (ABS_VOLUME) <= 0;
  *o = (unsigned long long)((ABS_VOLUME) | 0);  /* check that ABS_VOLUME is an integer */
  if (!_cffi_check_int(*o, n, 32U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_WHEEL(unsigned long long *o)
{
  int n = (ABS_WHEEL) <= 0;
  *o = (unsigned long long)((ABS_WHEEL) | 0);  /* check that ABS_WHEEL is an integer */
  if (!_cffi_check_int(*o, n, 8U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_X(unsigned long long *o)
{
  int n = (ABS_X) <= 0;
  *o = (unsigned long long)((ABS_X) | 0);  /* check that ABS_X is an integer */
  if (!_cffi_check_int(*o, n, 0))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_Y(unsigned long long *o)
{
  int n = (ABS_Y) <= 0;
  *o = (unsigned long long)((ABS_Y) | 0);  /* check that ABS_Y is an integer */
  if (!_cffi_check_int(*o, n, 1U))
    n |= 2;
  return n;
}

static int _cffi_const_ABS_Z(unsigned long long *o)
{
  int n = (ABS_Z) <= 0;
  *o = (unsigned long long)((ABS_Z) | 0);  /* check that ABS_Z is an integer */
  if (!_cffi_check_int(*o, n, 2U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_0(unsigned long long *o)
{
  int n = (BTN_0) <= 0;
  *o = (unsigned long long)((BTN_0) | 0);  /* check that BTN_0 is an integer */
  if (!_cffi_check_int(*o, n, 256U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_1(unsigned long long *o)
{
  int n = (BTN_1) <= 0;
  *o = (unsigned long long)((BTN_1) | 0);  /* check that BTN_1 is an integer */
  if (!_cffi_check_int(*o, n, 257U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_2(unsigned long long *o)
{
  int n = (BTN_2) <= 0;
  *o = (unsigned long long)((BTN_2) | 0);  /* check that BTN_2 is an integer */
  if (!_cffi_check_int(*o, n, 258U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_3(unsigned long long *o)
{
  int n = (BTN_3) <= 0;
  *o = (unsigned long long)((BTN_3) | 0);  /* check that BTN_3 is an integer */
  if (!_cffi_check_int(*o, n, 259U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_4(unsigned long long *o)
{
  int n = (BTN_4) <= 0;
  *o = (unsigned long long)((BTN_4) | 0);  /* check that BTN_4 is an integer */
  if (!_cffi_check_int(*o, n, 260U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_5(unsigned long long *o)
{
  int n = (BTN_5) <= 0;
  *o = (unsigned long long)((BTN_5) | 0);  /* check that BTN_5 is an integer */
  if (!_cffi_check_int(*o, n, 261U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_6(unsigned long long *o)
{
  int n = (BTN_6) <= 0;
  *o = (unsigned long long)((BTN_6) | 0);  /* check that BTN_6 is an integer */
  if (!_cffi_check_int(*o, n, 262U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_7(unsigned long long *o)
{
  int n = (BTN_7) <= 0;
  *o = (unsigned long long)((BTN_7) | 0);  /* check that BTN_7 is an integer */
  if (!_cffi_check_int(*o, n, 263U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_8(unsigned long long *o)
{
  int n = (BTN_8) <= 0;
  *o = (unsigned long long)((BTN_8) | 0);  /* check that BTN_8 is an integer */
  if (!_cffi_check_int(*o, n, 264U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_9(unsigned long long *o)
{
  int n = (BTN_9) <= 0;
  *o = (unsigned long long)((BTN_9) | 0);  /* check that BTN_9 is an integer */
  if (!_cffi_check_int(*o, n, 265U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_BACK(unsigned long long *o)
{
  int n = (BTN_BACK) <= 0;
  *o = (unsigned long long)((BTN_BACK) | 0);  /* check that BTN_BACK is an integer */
  if (!_cffi_check_int(*o, n, 278U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_BASE(unsigned long long *o)
{
  int n = (BTN_BASE) <= 0;
  *o = (unsigned long long)((BTN_BASE) | 0);  /* check that BTN_BASE is an integer */
  if (!_cffi_check_int(*o, n, 294U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_BASE2(unsigned long long *o)
{
  int n = (BTN_BASE2) <= 0;
  *o = (unsigned long long)((BTN_BASE2) | 0);  /* check that BTN_BASE2 is an integer */
  if (!_cffi_check_int(*o, n, 295U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_BASE3(unsigned long long *o)
{
  int n = (BTN_BASE3) <= 0;
  *o = (unsigned long long)((BTN_BASE3) | 0);  /* check that BTN_BASE3 is an integer */
  if (!_cffi_check_int(*o, n, 296U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_BASE4(unsigned long long *o)
{
  int n = (BTN_BASE4) <= 0;
  *o = (unsigned long long)((BTN_BASE4) | 0);  /* check that BTN_BASE4 is an integer */
  if (!_cffi_check_int(*o, n, 297U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_BASE5(unsigned long long *o)
{
  int n = (BTN_BASE5) <= 0;
  *o = (unsigned long long)((BTN_BASE5) | 0);  /* check that BTN_BASE5 is an integer */
  if (!_cffi_check_int(*o, n, 298U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_BASE6(unsigned long long *o)
{
  int n = (BTN_BASE6) <= 0;
  *o = (unsigned long long)((BTN_BASE6) | 0);  /* check that BTN_BASE6 is an integer */
  if (!_cffi_check_int(*o, n, 299U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_C(unsigned long long *o)
{
  int n = (BTN_C) <= 0;
  *o = (unsigned long long)((BTN_C) | 0);  /* check that BTN_C is an integer */
  if (!_cffi_check_int(*o, n, 306U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_DEAD(unsigned long long *o)
{
  int n = (BTN_DEAD) <= 0;
  *o = (unsigned long long)((BTN_DEAD) | 0);  /* check that BTN_DEAD is an integer */
  if (!_cffi_check_int(*o, n, 303U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_DIGI(unsigned long long *o)
{
  int n = (BTN_DIGI) <= 0;
  *o = (unsigned long long)((BTN_DIGI) | 0);  /* check that BTN_DIGI is an integer */
  if (!_cffi_check_int(*o, n, 320U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_DPAD_DOWN(unsigned long long *o)
{
  int n = (BTN_DPAD_DOWN) <= 0;
  *o = (unsigned long long)((BTN_DPAD_DOWN) | 0);  /* check that BTN_DPAD_DOWN is an integer */
  if (!_cffi_check_int(*o, n, 545U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_DPAD_LEFT(unsigned long long *o)
{
  int n = (BTN_DPAD_LEFT) <= 0;
  *o = (unsigned long long)((BTN_DPAD_LEFT) | 0);  /* check that BTN_DPAD_LEFT is an integer */
  if (!_cffi_check_int(*o, n, 546U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_DPAD_RIGHT(unsigned long long *o)
{
  int n = (BTN_DPAD_RIGHT) <= 0;
  *o = (unsigned long long)((BTN_DPAD_RIGHT) | 0);  /* check that BTN_DPAD_RIGHT is an integer */
  if (!_cffi_check_int(*o, n, 547U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_DPAD_UP(unsigned long long *o)
{
  int n = (BTN_DPAD_UP) <= 0;
  *o = (unsigned long long)((BTN_DPAD_UP) | 0);  /* check that BTN_DPAD_UP is an integer */
  if (!_cffi_check_int(*o, n, 544U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_EAST(unsigned long long *o)
{
  int n = (BTN_EAST) <= 0;
  *o = (unsigned long long)((BTN_EAST) | 0);  /* check that BTN_EAST is an integer */
  if (!_cffi_check_int(*o, n, 305U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_EXTRA(unsigned long long *o)
{
  int n = (BTN_EXTRA) <= 0;
  *o = (unsigned long long)((BTN_EXTRA) | 0);  /* check that BTN_EXTRA is an integer */
  if (!_cffi_check_int(*o, n, 276U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_FORWARD(unsigned long long *o)
{
  int n = (BTN_FORWARD) <= 0;
  *o = (unsigned long long)((BTN_FORWARD) | 0);  /* check that BTN_FORWARD is an integer */
  if (!_cffi_check_int(*o, n, 277U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_GAMEPAD(unsigned long long *o)
{
  int n = (BTN_GAMEPAD) <= 0;
  *o = (unsigned long long)((BTN_GAMEPAD) | 0);  /* check that BTN_GAMEPAD is an integer */
  if (!_cffi_check_int(*o, n, 304U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_GEAR_DOWN(unsigned long long *o)
{
  int n = (BTN_GEAR_DOWN) <= 0;
  *o = (unsigned long long)((BTN_GEAR_DOWN) | 0);  /* check that BTN_GEAR_DOWN is an integer */
  if (!_cffi_check_int(*o, n, 336U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_GEAR_UP(unsigned long long *o)
{
  int n = (BTN_GEAR_UP) <= 0;
  *o = (unsigned long long)((BTN_GEAR_UP) | 0);  /* check that BTN_GEAR_UP is an integer */
  if (!_cffi_check_int(*o, n, 337U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_JOYSTICK(unsigned long long *o)
{
  int n = (BTN_JOYSTICK) <= 0;
  *o = (unsigned long long)((BTN_JOYSTICK) | 0);  /* check that BTN_JOYSTICK is an integer */
  if (!_cffi_check_int(*o, n, 288U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_LEFT(unsigned long long *o)
{
  int n = (BTN_LEFT) <= 0;
  *o = (unsigned long long)((BTN_LEFT) | 0);  /* check that BTN_LEFT is an integer */
  if (!_cffi_check_int(*o, n, 272U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_MIDDLE(unsigned long long *o)
{
  int n = (BTN_MIDDLE) <= 0;
  *o = (unsigned long long)((BTN_MIDDLE) | 0);  /* check that BTN_MIDDLE is an integer */
  if (!_cffi_check_int(*o, n, 274U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_MISC(unsigned long long *o)
{
  int n = (BTN_MISC) <= 0;
  *o = (unsigned long long)((BTN_MISC) | 0);  /* check that BTN_MISC is an integer */
  if (!_cffi_check_int(*o, n, 256U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_MODE(unsigned long long *o)
{
  int n = (BTN_MODE) <= 0;
  *o = (unsigned long long)((BTN_MODE) | 0);  /* check that BTN_MODE is an integer */
  if (!_cffi_check_int(*o, n, 316U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_MOUSE(unsigned long long *o)
{
  int n = (BTN_MOUSE) <= 0;
  *o = (unsigned long long)((BTN_MOUSE) | 0);  /* check that BTN_MOUSE is an integer */
  if (!_cffi_check_int(*o, n, 272U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_NORTH(unsigned long long *o)
{
  int n = (BTN_NORTH) <= 0;
  *o = (unsigned long long)((BTN_NORTH) | 0);  /* check that BTN_NORTH is an integer */
  if (!_cffi_check_int(*o, n, 307U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_PINKIE(unsigned long long *o)
{
  int n = (BTN_PINKIE) <= 0;
  *o = (unsigned long long)((BTN_PINKIE) | 0);  /* check that BTN_PINKIE is an integer */
  if (!_cffi_check_int(*o, n, 293U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_RIGHT(unsigned long long *o)
{
  int n = (BTN_RIGHT) <= 0;
  *o = (unsigned long long)((BTN_RIGHT) | 0);  /* check that BTN_RIGHT is an integer */
  if (!_cffi_check_int(*o, n, 273U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_SELECT(unsigned long long *o)
{
  int n = (BTN_SELECT) <= 0;
  *o = (unsigned long long)((BTN_SELECT) | 0);  /* check that BTN_SELECT is an integer */
  if (!_cffi_check_int(*o, n, 314U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_SIDE(unsigned long long *o)
{
  int n = (BTN_SIDE) <= 0;
  *o = (unsigned long long)((BTN_SIDE) | 0);  /* check that BTN_SIDE is an integer */
  if (!_cffi_check_int(*o, n, 275U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_SOUTH(unsigned long long *o)
{
  int n = (BTN_SOUTH) <= 0;
  *o = (unsigned long long)((BTN_SOUTH) | 0);  /* check that BTN_SOUTH is an integer */
  if (!_cffi_check_int(*o, n, 304U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_START(unsigned long long *o)
{
  int n = (BTN_START) <= 0;
  *o = (unsigned long long)((BTN_START) | 0);  /* check that BTN_START is an integer */
  if (!_cffi_check_int(*o, n, 315U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_STYLUS(unsigned long long *o)
{
  int n = (BTN_STYLUS) <= 0;
  *o = (unsigned long long)((BTN_STYLUS) | 0);  /* check that BTN_STYLUS is an integer */
  if (!_cffi_check_int(*o, n, 331U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_STYLUS2(unsigned long long *o)
{
  int n = (BTN_STYLUS2) <= 0;
  *o = (unsigned long long)((BTN_STYLUS2) | 0);  /* check that BTN_STYLUS2 is an integer */
  if (!_cffi_check_int(*o, n, 332U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TASK(unsigned long long *o)
{
  int n = (BTN_TASK) <= 0;
  *o = (unsigned long long)((BTN_TASK) | 0);  /* check that BTN_TASK is an integer */
  if (!_cffi_check_int(*o, n, 279U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_THUMB(unsigned long long *o)
{
  int n = (BTN_THUMB) <= 0;
  *o = (unsigned long long)((BTN_THUMB) | 0);  /* check that BTN_THUMB is an integer */
  if (!_cffi_check_int(*o, n, 289U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_THUMB2(unsigned long long *o)
{
  int n = (BTN_THUMB2) <= 0;
  *o = (unsigned long long)((BTN_THUMB2) | 0);  /* check that BTN_THUMB2 is an integer */
  if (!_cffi_check_int(*o, n, 290U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_THUMBL(unsigned long long *o)
{
  int n = (BTN_THUMBL) <= 0;
  *o = (unsigned long long)((BTN_THUMBL) | 0);  /* check that BTN_THUMBL is an integer */
  if (!_cffi_check_int(*o, n, 317U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_THUMBR(unsigned long long *o)
{
  int n = (BTN_THUMBR) <= 0;
  *o = (unsigned long long)((BTN_THUMBR) | 0);  /* check that BTN_THUMBR is an integer */
  if (!_cffi_check_int(*o, n, 318U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TL(unsigned long long *o)
{
  int n = (BTN_TL) <= 0;
  *o = (unsigned long long)((BTN_TL) | 0);  /* check that BTN_TL is an integer */
  if (!_cffi_check_int(*o, n, 310U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TL2(unsigned long long *o)
{
  int n = (BTN_TL2) <= 0;
  *o = (unsigned long long)((BTN_TL2) | 0);  /* check that BTN_TL2 is an integer */
  if (!_cffi_check_int(*o, n, 312U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOOL_AIRBRUSH(unsigned long long *o)
{
  int n = (BTN_TOOL_AIRBRUSH) <= 0;
  *o = (unsigned long long)((BTN_TOOL_AIRBRUSH) | 0);  /* check that BTN_TOOL_AIRBRUSH is an integer */
  if (!_cffi_check_int(*o, n, 324U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOOL_BRUSH(unsigned long long *o)
{
  int n = (BTN_TOOL_BRUSH) <= 0;
  *o = (unsigned long long)((BTN_TOOL_BRUSH) | 0);  /* check that BTN_TOOL_BRUSH is an integer */
  if (!_cffi_check_int(*o, n, 322U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOOL_DOUBLETAP(unsigned long long *o)
{
  int n = (BTN_TOOL_DOUBLETAP) <= 0;
  *o = (unsigned long long)((BTN_TOOL_DOUBLETAP) | 0);  /* check that BTN_TOOL_DOUBLETAP is an integer */
  if (!_cffi_check_int(*o, n, 333U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOOL_FINGER(unsigned long long *o)
{
  int n = (BTN_TOOL_FINGER) <= 0;
  *o = (unsigned long long)((BTN_TOOL_FINGER) | 0);  /* check that BTN_TOOL_FINGER is an integer */
  if (!_cffi_check_int(*o, n, 325U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOOL_LENS(unsigned long long *o)
{
  int n = (BTN_TOOL_LENS) <= 0;
  *o = (unsigned long long)((BTN_TOOL_LENS) | 0);  /* check that BTN_TOOL_LENS is an integer */
  if (!_cffi_check_int(*o, n, 327U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOOL_MOUSE(unsigned long long *o)
{
  int n = (BTN_TOOL_MOUSE) <= 0;
  *o = (unsigned long long)((BTN_TOOL_MOUSE) | 0);  /* check that BTN_TOOL_MOUSE is an integer */
  if (!_cffi_check_int(*o, n, 326U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOOL_PEN(unsigned long long *o)
{
  int n = (BTN_TOOL_PEN) <= 0;
  *o = (unsigned long long)((BTN_TOOL_PEN) | 0);  /* check that BTN_TOOL_PEN is an integer */
  if (!_cffi_check_int(*o, n, 320U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOOL_PENCIL(unsigned long long *o)
{
  int n = (BTN_TOOL_PENCIL) <= 0;
  *o = (unsigned long long)((BTN_TOOL_PENCIL) | 0);  /* check that BTN_TOOL_PENCIL is an integer */
  if (!_cffi_check_int(*o, n, 323U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOOL_QUADTAP(unsigned long long *o)
{
  int n = (BTN_TOOL_QUADTAP) <= 0;
  *o = (unsigned long long)((BTN_TOOL_QUADTAP) | 0);  /* check that BTN_TOOL_QUADTAP is an integer */
  if (!_cffi_check_int(*o, n, 335U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOOL_QUINTTAP(unsigned long long *o)
{
  int n = (BTN_TOOL_QUINTTAP) <= 0;
  *o = (unsigned long long)((BTN_TOOL_QUINTTAP) | 0);  /* check that BTN_TOOL_QUINTTAP is an integer */
  if (!_cffi_check_int(*o, n, 328U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOOL_RUBBER(unsigned long long *o)
{
  int n = (BTN_TOOL_RUBBER) <= 0;
  *o = (unsigned long long)((BTN_TOOL_RUBBER) | 0);  /* check that BTN_TOOL_RUBBER is an integer */
  if (!_cffi_check_int(*o, n, 321U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOOL_TRIPLETAP(unsigned long long *o)
{
  int n = (BTN_TOOL_TRIPLETAP) <= 0;
  *o = (unsigned long long)((BTN_TOOL_TRIPLETAP) | 0);  /* check that BTN_TOOL_TRIPLETAP is an integer */
  if (!_cffi_check_int(*o, n, 334U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOP(unsigned long long *o)
{
  int n = (BTN_TOP) <= 0;
  *o = (unsigned long long)((BTN_TOP) | 0);  /* check that BTN_TOP is an integer */
  if (!_cffi_check_int(*o, n, 291U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOP2(unsigned long long *o)
{
  int n = (BTN_TOP2) <= 0;
  *o = (unsigned long long)((BTN_TOP2) | 0);  /* check that BTN_TOP2 is an integer */
  if (!_cffi_check_int(*o, n, 292U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TOUCH(unsigned long long *o)
{
  int n = (BTN_TOUCH) <= 0;
  *o = (unsigned long long)((BTN_TOUCH) | 0);  /* check that BTN_TOUCH is an integer */
  if (!_cffi_check_int(*o, n, 330U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TR(unsigned long long *o)
{
  int n = (BTN_TR) <= 0;
  *o = (unsigned long long)((BTN_TR) | 0);  /* check that BTN_TR is an integer */
  if (!_cffi_check_int(*o, n, 311U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TR2(unsigned long long *o)
{
  int n = (BTN_TR2) <= 0;
  *o = (unsigned long long)((BTN_TR2) | 0);  /* check that BTN_TR2 is an integer */
  if (!_cffi_check_int(*o, n, 313U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER(unsigned long long *o)
{
  int n = (BTN_TRIGGER) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER) | 0);  /* check that BTN_TRIGGER is an integer */
  if (!_cffi_check_int(*o, n, 288U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY) | 0);  /* check that BTN_TRIGGER_HAPPY is an integer */
  if (!_cffi_check_int(*o, n, 704U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY1(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY1) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY1) | 0);  /* check that BTN_TRIGGER_HAPPY1 is an integer */
  if (!_cffi_check_int(*o, n, 704U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY10(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY10) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY10) | 0);  /* check that BTN_TRIGGER_HAPPY10 is an integer */
  if (!_cffi_check_int(*o, n, 713U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY11(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY11) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY11) | 0);  /* check that BTN_TRIGGER_HAPPY11 is an integer */
  if (!_cffi_check_int(*o, n, 714U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY12(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY12) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY12) | 0);  /* check that BTN_TRIGGER_HAPPY12 is an integer */
  if (!_cffi_check_int(*o, n, 715U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY13(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY13) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY13) | 0);  /* check that BTN_TRIGGER_HAPPY13 is an integer */
  if (!_cffi_check_int(*o, n, 716U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY14(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY14) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY14) | 0);  /* check that BTN_TRIGGER_HAPPY14 is an integer */
  if (!_cffi_check_int(*o, n, 717U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY15(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY15) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY15) | 0);  /* check that BTN_TRIGGER_HAPPY15 is an integer */
  if (!_cffi_check_int(*o, n, 718U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY16(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY16) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY16) | 0);  /* check that BTN_TRIGGER_HAPPY16 is an integer */
  if (!_cffi_check_int(*o, n, 719U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY17(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY17) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY17) | 0);  /* check that BTN_TRIGGER_HAPPY17 is an integer */
  if (!_cffi_check_int(*o, n, 720U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY18(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY18) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY18) | 0);  /* check that BTN_TRIGGER_HAPPY18 is an integer */
  if (!_cffi_check_int(*o, n, 721U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY19(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY19) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY19) | 0);  /* check that BTN_TRIGGER_HAPPY19 is an integer */
  if (!_cffi_check_int(*o, n, 722U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY2(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY2) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY2) | 0);  /* check that BTN_TRIGGER_HAPPY2 is an integer */
  if (!_cffi_check_int(*o, n, 705U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY20(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY20) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY20) | 0);  /* check that BTN_TRIGGER_HAPPY20 is an integer */
  if (!_cffi_check_int(*o, n, 723U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY21(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY21) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY21) | 0);  /* check that BTN_TRIGGER_HAPPY21 is an integer */
  if (!_cffi_check_int(*o, n, 724U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY22(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY22) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY22) | 0);  /* check that BTN_TRIGGER_HAPPY22 is an integer */
  if (!_cffi_check_int(*o, n, 725U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY23(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY23) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY23) | 0);  /* check that BTN_TRIGGER_HAPPY23 is an integer */
  if (!_cffi_check_int(*o, n, 726U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY24(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY24) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY24) | 0);  /* check that BTN_TRIGGER_HAPPY24 is an integer */
  if (!_cffi_check_int(*o, n, 727U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY25(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY25) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY25) | 0);  /* check that BTN_TRIGGER_HAPPY25 is an integer */
  if (!_cffi_check_int(*o, n, 728U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY26(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY26) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY26) | 0);  /* check that BTN_TRIGGER_HAPPY26 is an integer */
  if (!_cffi_check_int(*o, n, 729U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY27(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY27) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY27) | 0);  /* check that BTN_TRIGGER_HAPPY27 is an integer */
  if (!_cffi_check_int(*o, n, 730U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY28(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY28) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY28) | 0);  /* check that BTN_TRIGGER_HAPPY28 is an integer */
  if (!_cffi_check_int(*o, n, 731U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY29(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY29) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY29) | 0);  /* check that BTN_TRIGGER_HAPPY29 is an integer */
  if (!_cffi_check_int(*o, n, 732U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY3(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY3) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY3) | 0);  /* check that BTN_TRIGGER_HAPPY3 is an integer */
  if (!_cffi_check_int(*o, n, 706U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY30(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY30) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY30) | 0);  /* check that BTN_TRIGGER_HAPPY30 is an integer */
  if (!_cffi_check_int(*o, n, 733U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY31(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY31) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY31) | 0);  /* check that BTN_TRIGGER_HAPPY31 is an integer */
  if (!_cffi_check_int(*o, n, 734U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY32(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY32) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY32) | 0);  /* check that BTN_TRIGGER_HAPPY32 is an integer */
  if (!_cffi_check_int(*o, n, 735U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY33(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY33) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY33) | 0);  /* check that BTN_TRIGGER_HAPPY33 is an integer */
  if (!_cffi_check_int(*o, n, 736U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY34(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY34) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY34) | 0);  /* check that BTN_TRIGGER_HAPPY34 is an integer */
  if (!_cffi_check_int(*o, n, 737U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY35(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY35) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY35) | 0);  /* check that BTN_TRIGGER_HAPPY35 is an integer */
  if (!_cffi_check_int(*o, n, 738U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY36(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY36) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY36) | 0);  /* check that BTN_TRIGGER_HAPPY36 is an integer */
  if (!_cffi_check_int(*o, n, 739U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY37(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY37) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY37) | 0);  /* check that BTN_TRIGGER_HAPPY37 is an integer */
  if (!_cffi_check_int(*o, n, 740U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY38(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY38) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY38) | 0);  /* check that BTN_TRIGGER_HAPPY38 is an integer */
  if (!_cffi_check_int(*o, n, 741U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY39(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY39) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY39) | 0);  /* check that BTN_TRIGGER_HAPPY39 is an integer */
  if (!_cffi_check_int(*o, n, 742U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY4(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY4) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY4) | 0);  /* check that BTN_TRIGGER_HAPPY4 is an integer */
  if (!_cffi_check_int(*o, n, 707U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY40(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY40) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY40) | 0);  /* check that BTN_TRIGGER_HAPPY40 is an integer */
  if (!_cffi_check_int(*o, n, 743U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY5(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY5) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY5) | 0);  /* check that BTN_TRIGGER_HAPPY5 is an integer */
  if (!_cffi_check_int(*o, n, 708U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY6(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY6) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY6) | 0);  /* check that BTN_TRIGGER_HAPPY6 is an integer */
  if (!_cffi_check_int(*o, n, 709U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY7(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY7) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY7) | 0);  /* check that BTN_TRIGGER_HAPPY7 is an integer */
  if (!_cffi_check_int(*o, n, 710U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY8(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY8) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY8) | 0);  /* check that BTN_TRIGGER_HAPPY8 is an integer */
  if (!_cffi_check_int(*o, n, 711U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_TRIGGER_HAPPY9(unsigned long long *o)
{
  int n = (BTN_TRIGGER_HAPPY9) <= 0;
  *o = (unsigned long long)((BTN_TRIGGER_HAPPY9) | 0);  /* check that BTN_TRIGGER_HAPPY9 is an integer */
  if (!_cffi_check_int(*o, n, 712U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_WEST(unsigned long long *o)
{
  int n = (BTN_WEST) <= 0;
  *o = (unsigned long long)((BTN_WEST) | 0);  /* check that BTN_WEST is an integer */
  if (!_cffi_check_int(*o, n, 308U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_WHEEL(unsigned long long *o)
{
  int n = (BTN_WHEEL) <= 0;
  *o = (unsigned long long)((BTN_WHEEL) | 0);  /* check that BTN_WHEEL is an integer */
  if (!_cffi_check_int(*o, n, 336U))
    n |= 2;
  return n;
}

static int _cffi_const_BTN_Z(unsigned long long *o)
{
  int n = (BTN_Z) <= 0;
  *o = (unsigned long long)((BTN_Z) | 0);  /* check that BTN_Z is an integer */
  if (!_cffi_check_int(*o, n, 309U))
    n |= 2;
  return n;
}

static int _cffi_const_EV_ABS(unsigned long long *o)
{
  int n = (EV_ABS) <= 0;
  *o = (unsigned long long)((EV_ABS) | 0);  /* check that EV_ABS is an integer */
  if (!_cffi_check_int(*o, n, 3U))
    n |= 2;
  return n;
}

static int _cffi_const_EV_FF(unsigned long long *o)
{
  int n = (EV_FF) <= 0;
  *o = (unsigned long long)((EV_FF) | 0);  /* check that EV_FF is an integer */
  if (!_cffi_check_int(*o, n, 21U))
    n |= 2;
  return n;
}

static int _cffi_const_EV_FF_STATUS(unsigned long long *o)
{
  int n = (EV_FF_STATUS) <= 0;
  *o = (unsigned long long)((EV_FF_STATUS) | 0);  /* check that EV_FF_STATUS is an integer */
  if (!_cffi_check_int(*o, n, 23U))
    n |= 2;
  return n;
}

static int _cffi_const_EV_KEY(unsigned long long *o)
{
  int n = (EV_KEY) <= 0;
  *o = (unsigned long long)((EV_KEY) | 0);  /* check that EV_KEY is an integer */
  if (!_cffi_check_int(*o, n, 1U))
    n |= 2;
  return n;
}

static int _cffi_const_EV_LED(unsigned long long *o)
{
  int n = (EV_LED) <= 0;
  *o = (unsigned long long)((EV_LED) | 0);  /* check that EV_LED is an integer */
  if (!_cffi_check_int(*o, n, 17U))
    n |= 2;
  return n;
}

static int _cffi_const_EV_MAX(unsigned long long *o)
{
  int n = (EV_MAX) <= 0;
  *o = (unsigned long long)((EV_MAX) | 0);  /* check that EV_MAX is an integer */
  if (!_cffi_check_int(*o, n, 31U))
    n |= 2;
  return n;
}

static int _cffi_const_EV_MSC(unsigned long long *o)
{
  int n = (EV_MSC) <= 0;
  *o = (unsigned long long)((EV_MSC) | 0);  /* check that EV_MSC is an integer */
  if (!_cffi_check_int(*o, n, 4U))
    n |= 2;
  return n;
}

static int _cffi_const_EV_PWR(unsigned long long *o)
{
  int n = (EV_PWR) <= 0;
  *o = (unsigned long long)((EV_PWR) | 0);  /* check that EV_PWR is an integer */
  if (!_cffi_check_int(*o, n, 22U))
    n |= 2;
  return n;
}

static int _cffi_const_EV_REL(unsigned long long *o)
{
  int n = (EV_REL) <= 0;
  *o = (unsigned long long)((EV_REL) | 0);  /* check that EV_REL is an integer */
  if (!_cffi_check_int(*o, n, 2U))
    n |= 2;
  return n;
}

static int _cffi_const_EV_REP(unsigned long long *o)
{
  int n = (EV_REP) <= 0;
  *o = (unsigned long long)((EV_REP) | 0);  /* check that EV_REP is an integer */
  if (!_cffi_check_int(*o, n, 20U))
    n |= 2;
  return n;
}

static int _cffi_const_EV_SND(unsigned long long *o)
{
  int n = (EV_SND) <= 0;
  *o = (unsigned long long)((EV_SND) | 0);  /* check that EV_SND is an integer */
  if (!_cffi_check_int(*o, n, 18U))
    n |= 2;
  return n;
}

static int _cffi_const_EV_SW(unsigned long long *o)
{
  int n = (EV_SW) <= 0;
  *o = (unsigned long long)((EV_SW) | 0);  /* check that EV_SW is an integer */
  if (!_cffi_check_int(*o, n, 5U))
    n |= 2;
  return n;
}

static int _cffi_const_EV_SYN(unsigned long long *o)
{
  int n = (EV_SYN) <= 0;
  *o = (unsigned long long)((EV_SYN) | 0);  /* check that EV_SYN is an integer */
  if (!_cffi_check_int(*o, n, 0))
    n |= 2;
  return n;
}

static int _cffi_const_INPUT_PROP_ACCELEROMETER(unsigned long long *o)
{
  int n = (INPUT_PROP_ACCELEROMETER) <= 0;
  *o = (unsigned long long)((INPUT_PROP_ACCELEROMETER) | 0);  /* check that INPUT_PROP_ACCELEROMETER is an integer */
  if (!_cffi_check_int(*o, n, 6U))
    n |= 2;
  return n;
}

static int _cffi_const_INPUT_PROP_BUTTONPAD(unsigned long long *o)
{
  int n = (INPUT_PROP_BUTTONPAD) <= 0;
  *o = (unsigned long long)((INPUT_PROP_BUTTONPAD) | 0);  /* check that INPUT_PROP_BUTTONPAD is an integer */
  if (!_cffi_check_int(*o, n, 2U))
    n |= 2;
  return n;
}

static int _cffi_const_INPUT_PROP_DIRECT(unsigned long long *o)
{
  int n = (INPUT_PROP_DIRECT) <= 0;
  *o = (unsigned long long)((INPUT_PROP_DIRECT) | 0);  /* check that INPUT_PROP_DIRECT is an integer */
  if (!_cffi_check_int(*o, n, 1U))
    n |= 2;
  return n;
}

static int _cffi_const_INPUT_PROP_MAX(unsigned long long *o)
{
  int n = (INPUT_PROP_MAX) <= 0;
  *o = (unsigned long long)((INPUT_PROP_MAX) | 0);  /* check that INPUT_PROP_MAX is an integer */
  if (!_cffi_check_int(*o, n, 31U))
    n |= 2;
  return n;
}

static int _cffi_const_INPUT_PROP_POINTER(unsigned long long *o)
{
  int n = (INPUT_PROP_POINTER) <= 0;
  *o = (unsigned long long)((INPUT_PROP_POINTER) | 0);  /* check that INPUT_PROP_POINTER is an integer */
  if (!_cffi_check_int(*o, n, 0))
    n |= 2;
  return n;
}

static int _cffi_const_INPUT_PROP_POINTING_STICK(unsigned long long *o)
{
  int n = (INPUT_PROP_POINTING_STICK) <= 0;
  *o = (unsigned long long)((INPUT_PROP_POINTING_STICK) | 0);  /* check that INPUT_PROP_POINTING_STICK is an integer */
  if (!_cffi_check_int(*o, n, 5U))
    n |= 2;
  return n;
}

static int _cffi_const_INPUT_PROP_SEMI_MT(unsigned long long *o)
{
  int n = (INPUT_PROP_SEMI_MT) <= 0;
  *o = (unsigned long long)((INPUT_PROP_SEMI_MT) | 0);  /* check that INPUT_PROP_SEMI_MT is an integer */
  if (!_cffi_check_int(*o, n, 3U))
    n |= 2;
  return n;
}

static int _cffi_const_INPUT_PROP_TOPBUTTONPAD(unsigned long long *o)
{
  int n = (INPUT_PROP_TOPBUTTONPAD) <= 0;
  *o = (unsigned long long)((INPUT_PROP_TOPBUTTONPAD) | 0);  /* check that INPUT_PROP_TOPBUTTONPAD is an integer */
  if (!_cffi_check_int(*o, n, 4U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_0(unsigned long long *o)
{
  int n = (KEY_0) <= 0;
  *o = (unsigned long long)((KEY_0) | 0);  /* check that KEY_0 is an integer */
  if (!_cffi_check_int(*o, n, 11U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_1(unsigned long long *o)
{
  int n = (KEY_1) <= 0;
  *o = (unsigned long long)((KEY_1) | 0);  /* check that KEY_1 is an integer */
  if (!_cffi_check_int(*o, n, 2U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_102ND(unsigned long long *o)
{
  int n = (KEY_102ND) <= 0;
  *o = (unsigned long long)((KEY_102ND) | 0);  /* check that KEY_102ND is an integer */
  if (!_cffi_check_int(*o, n, 86U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_10CHANNELSDOWN(unsigned long long *o)
{
  int n = (KEY_10CHANNELSDOWN) <= 0;
  *o = (unsigned long long)((KEY_10CHANNELSDOWN) | 0);  /* check that KEY_10CHANNELSDOWN is an integer */
  if (!_cffi_check_int(*o, n, 441U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_10CHANNELSUP(unsigned long long *o)
{
  int n = (KEY_10CHANNELSUP) <= 0;
  *o = (unsigned long long)((KEY_10CHANNELSUP) | 0);  /* check that KEY_10CHANNELSUP is an integer */
  if (!_cffi_check_int(*o, n, 440U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_2(unsigned long long *o)
{
  int n = (KEY_2) <= 0;
  *o = (unsigned long long)((KEY_2) | 0);  /* check that KEY_2 is an integer */
  if (!_cffi_check_int(*o, n, 3U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_3(unsigned long long *o)
{
  int n = (KEY_3) <= 0;
  *o = (unsigned long long)((KEY_3) | 0);  /* check that KEY_3 is an integer */
  if (!_cffi_check_int(*o, n, 4U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_3D_MODE(unsigned long long *o)
{
  int n = (KEY_3D_MODE) <= 0;
  *o = (unsigned long long)((KEY_3D_MODE) | 0);  /* check that KEY_3D_MODE is an integer */
  if (!_cffi_check_int(*o, n, 623U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_4(unsigned long long *o)
{
  int n = (KEY_4) <= 0;
  *o = (unsigned long long)((KEY_4) | 0);  /* check that KEY_4 is an integer */
  if (!_cffi_check_int(*o, n, 5U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_5(unsigned long long *o)
{
  int n = (KEY_5) <= 0;
  *o = (unsigned long long)((KEY_5) | 0);  /* check that KEY_5 is an integer */
  if (!_cffi_check_int(*o, n, 6U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_6(unsigned long long *o)
{
  int n = (KEY_6) <= 0;
  *o = (unsigned long long)((KEY_6) | 0);  /* check that KEY_6 is an integer */
  if (!_cffi_check_int(*o, n, 7U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_7(unsigned long long *o)
{
  int n = (KEY_7) <= 0;
  *o = (unsigned long long)((KEY_7) | 0);  /* check that KEY_7 is an integer */
  if (!_cffi_check_int(*o, n, 8U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_8(unsigned long long *o)
{
  int n = (KEY_8) <= 0;
  *o = (unsigned long long)((KEY_8) | 0);  /* check that KEY_8 is an integer */
  if (!_cffi_check_int(*o, n, 9U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_9(unsigned long long *o)
{
  int n = (KEY_9) <= 0;
  *o = (unsigned long long)((KEY_9) | 0);  /* check that KEY_9 is an integer */
  if (!_cffi_check_int(*o, n, 10U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_A(unsigned long long *o)
{
  int n = (KEY_A) <= 0;
  *o = (unsigned long long)((KEY_A) | 0);  /* check that KEY_A is an integer */
  if (!_cffi_check_int(*o, n, 30U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_AB(unsigned long long *o)
{
  int n = (KEY_AB) <= 0;
  *o = (unsigned long long)((KEY_AB) | 0);  /* check that KEY_AB is an integer */
  if (!_cffi_check_int(*o, n, 406U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ADDRESSBOOK(unsigned long long *o)
{
  int n = (KEY_ADDRESSBOOK) <= 0;
  *o = (unsigned long long)((KEY_ADDRESSBOOK) | 0);  /* check that KEY_ADDRESSBOOK is an integer */
  if (!_cffi_check_int(*o, n, 429U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_AGAIN(unsigned long long *o)
{
  int n = (KEY_AGAIN) <= 0;
  *o = (unsigned long long)((KEY_AGAIN) | 0);  /* check that KEY_AGAIN is an integer */
  if (!_cffi_check_int(*o, n, 129U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ALS_TOGGLE(unsigned long long *o)
{
  int n = (KEY_ALS_TOGGLE) <= 0;
  *o = (unsigned long long)((KEY_ALS_TOGGLE) | 0);  /* check that KEY_ALS_TOGGLE is an integer */
  if (!_cffi_check_int(*o, n, 560U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ALTERASE(unsigned long long *o)
{
  int n = (KEY_ALTERASE) <= 0;
  *o = (unsigned long long)((KEY_ALTERASE) | 0);  /* check that KEY_ALTERASE is an integer */
  if (!_cffi_check_int(*o, n, 222U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ANGLE(unsigned long long *o)
{
  int n = (KEY_ANGLE) <= 0;
  *o = (unsigned long long)((KEY_ANGLE) | 0);  /* check that KEY_ANGLE is an integer */
  if (!_cffi_check_int(*o, n, 371U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_APOSTROPHE(unsigned long long *o)
{
  int n = (KEY_APOSTROPHE) <= 0;
  *o = (unsigned long long)((KEY_APOSTROPHE) | 0);  /* check that KEY_APOSTROPHE is an integer */
  if (!_cffi_check_int(*o, n, 40U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_APPSELECT(unsigned long long *o)
{
  int n = (KEY_APPSELECT) <= 0;
  *o = (unsigned long long)((KEY_APPSELECT) | 0);  /* check that KEY_APPSELECT is an integer */
  if (!_cffi_check_int(*o, n, 580U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ARCHIVE(unsigned long long *o)
{
  int n = (KEY_ARCHIVE) <= 0;
  *o = (unsigned long long)((KEY_ARCHIVE) | 0);  /* check that KEY_ARCHIVE is an integer */
  if (!_cffi_check_int(*o, n, 361U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ATTENDANT_OFF(unsigned long long *o)
{
  int n = (KEY_ATTENDANT_OFF) <= 0;
  *o = (unsigned long long)((KEY_ATTENDANT_OFF) | 0);  /* check that KEY_ATTENDANT_OFF is an integer */
  if (!_cffi_check_int(*o, n, 540U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ATTENDANT_ON(unsigned long long *o)
{
  int n = (KEY_ATTENDANT_ON) <= 0;
  *o = (unsigned long long)((KEY_ATTENDANT_ON) | 0);  /* check that KEY_ATTENDANT_ON is an integer */
  if (!_cffi_check_int(*o, n, 539U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ATTENDANT_TOGGLE(unsigned long long *o)
{
  int n = (KEY_ATTENDANT_TOGGLE) <= 0;
  *o = (unsigned long long)((KEY_ATTENDANT_TOGGLE) | 0);  /* check that KEY_ATTENDANT_TOGGLE is an integer */
  if (!_cffi_check_int(*o, n, 541U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_AUDIO(unsigned long long *o)
{
  int n = (KEY_AUDIO) <= 0;
  *o = (unsigned long long)((KEY_AUDIO) | 0);  /* check that KEY_AUDIO is an integer */
  if (!_cffi_check_int(*o, n, 392U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_AUDIO_DESC(unsigned long long *o)
{
  int n = (KEY_AUDIO_DESC) <= 0;
  *o = (unsigned long long)((KEY_AUDIO_DESC) | 0);  /* check that KEY_AUDIO_DESC is an integer */
  if (!_cffi_check_int(*o, n, 622U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_AUX(unsigned long long *o)
{
  int n = (KEY_AUX) <= 0;
  *o = (unsigned long long)((KEY_AUX) | 0);  /* check that KEY_AUX is an integer */
  if (!_cffi_check_int(*o, n, 390U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_B(unsigned long long *o)
{
  int n = (KEY_B) <= 0;
  *o = (unsigned long long)((KEY_B) | 0);  /* check that KEY_B is an integer */
  if (!_cffi_check_int(*o, n, 48U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BACK(unsigned long long *o)
{
  int n = (KEY_BACK) <= 0;
  *o = (unsigned long long)((KEY_BACK) | 0);  /* check that KEY_BACK is an integer */
  if (!_cffi_check_int(*o, n, 158U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BACKSLASH(unsigned long long *o)
{
  int n = (KEY_BACKSLASH) <= 0;
  *o = (unsigned long long)((KEY_BACKSLASH) | 0);  /* check that KEY_BACKSLASH is an integer */
  if (!_cffi_check_int(*o, n, 43U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BACKSPACE(unsigned long long *o)
{
  int n = (KEY_BACKSPACE) <= 0;
  *o = (unsigned long long)((KEY_BACKSPACE) | 0);  /* check that KEY_BACKSPACE is an integer */
  if (!_cffi_check_int(*o, n, 14U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BASSBOOST(unsigned long long *o)
{
  int n = (KEY_BASSBOOST) <= 0;
  *o = (unsigned long long)((KEY_BASSBOOST) | 0);  /* check that KEY_BASSBOOST is an integer */
  if (!_cffi_check_int(*o, n, 209U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BATTERY(unsigned long long *o)
{
  int n = (KEY_BATTERY) <= 0;
  *o = (unsigned long long)((KEY_BATTERY) | 0);  /* check that KEY_BATTERY is an integer */
  if (!_cffi_check_int(*o, n, 236U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BLUE(unsigned long long *o)
{
  int n = (KEY_BLUE) <= 0;
  *o = (unsigned long long)((KEY_BLUE) | 0);  /* check that KEY_BLUE is an integer */
  if (!_cffi_check_int(*o, n, 401U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BLUETOOTH(unsigned long long *o)
{
  int n = (KEY_BLUETOOTH) <= 0;
  *o = (unsigned long long)((KEY_BLUETOOTH) | 0);  /* check that KEY_BLUETOOTH is an integer */
  if (!_cffi_check_int(*o, n, 237U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BOOKMARKS(unsigned long long *o)
{
  int n = (KEY_BOOKMARKS) <= 0;
  *o = (unsigned long long)((KEY_BOOKMARKS) | 0);  /* check that KEY_BOOKMARKS is an integer */
  if (!_cffi_check_int(*o, n, 156U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BREAK(unsigned long long *o)
{
  int n = (KEY_BREAK) <= 0;
  *o = (unsigned long long)((KEY_BREAK) | 0);  /* check that KEY_BREAK is an integer */
  if (!_cffi_check_int(*o, n, 411U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRIGHTNESSDOWN(unsigned long long *o)
{
  int n = (KEY_BRIGHTNESSDOWN) <= 0;
  *o = (unsigned long long)((KEY_BRIGHTNESSDOWN) | 0);  /* check that KEY_BRIGHTNESSDOWN is an integer */
  if (!_cffi_check_int(*o, n, 224U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRIGHTNESSUP(unsigned long long *o)
{
  int n = (KEY_BRIGHTNESSUP) <= 0;
  *o = (unsigned long long)((KEY_BRIGHTNESSUP) | 0);  /* check that KEY_BRIGHTNESSUP is an integer */
  if (!_cffi_check_int(*o, n, 225U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRIGHTNESS_AUTO(unsigned long long *o)
{
  int n = (KEY_BRIGHTNESS_AUTO) <= 0;
  *o = (unsigned long long)((KEY_BRIGHTNESS_AUTO) | 0);  /* check that KEY_BRIGHTNESS_AUTO is an integer */
  if (!_cffi_check_int(*o, n, 244U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRIGHTNESS_CYCLE(unsigned long long *o)
{
  int n = (KEY_BRIGHTNESS_CYCLE) <= 0;
  *o = (unsigned long long)((KEY_BRIGHTNESS_CYCLE) | 0);  /* check that KEY_BRIGHTNESS_CYCLE is an integer */
  if (!_cffi_check_int(*o, n, 243U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRIGHTNESS_MAX(unsigned long long *o)
{
  int n = (KEY_BRIGHTNESS_MAX) <= 0;
  *o = (unsigned long long)((KEY_BRIGHTNESS_MAX) | 0);  /* check that KEY_BRIGHTNESS_MAX is an integer */
  if (!_cffi_check_int(*o, n, 593U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRIGHTNESS_MIN(unsigned long long *o)
{
  int n = (KEY_BRIGHTNESS_MIN) <= 0;
  *o = (unsigned long long)((KEY_BRIGHTNESS_MIN) | 0);  /* check that KEY_BRIGHTNESS_MIN is an integer */
  if (!_cffi_check_int(*o, n, 592U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRL_DOT1(unsigned long long *o)
{
  int n = (KEY_BRL_DOT1) <= 0;
  *o = (unsigned long long)((KEY_BRL_DOT1) | 0);  /* check that KEY_BRL_DOT1 is an integer */
  if (!_cffi_check_int(*o, n, 497U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRL_DOT10(unsigned long long *o)
{
  int n = (KEY_BRL_DOT10) <= 0;
  *o = (unsigned long long)((KEY_BRL_DOT10) | 0);  /* check that KEY_BRL_DOT10 is an integer */
  if (!_cffi_check_int(*o, n, 506U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRL_DOT2(unsigned long long *o)
{
  int n = (KEY_BRL_DOT2) <= 0;
  *o = (unsigned long long)((KEY_BRL_DOT2) | 0);  /* check that KEY_BRL_DOT2 is an integer */
  if (!_cffi_check_int(*o, n, 498U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRL_DOT3(unsigned long long *o)
{
  int n = (KEY_BRL_DOT3) <= 0;
  *o = (unsigned long long)((KEY_BRL_DOT3) | 0);  /* check that KEY_BRL_DOT3 is an integer */
  if (!_cffi_check_int(*o, n, 499U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRL_DOT4(unsigned long long *o)
{
  int n = (KEY_BRL_DOT4) <= 0;
  *o = (unsigned long long)((KEY_BRL_DOT4) | 0);  /* check that KEY_BRL_DOT4 is an integer */
  if (!_cffi_check_int(*o, n, 500U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRL_DOT5(unsigned long long *o)
{
  int n = (KEY_BRL_DOT5) <= 0;
  *o = (unsigned long long)((KEY_BRL_DOT5) | 0);  /* check that KEY_BRL_DOT5 is an integer */
  if (!_cffi_check_int(*o, n, 501U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRL_DOT6(unsigned long long *o)
{
  int n = (KEY_BRL_DOT6) <= 0;
  *o = (unsigned long long)((KEY_BRL_DOT6) | 0);  /* check that KEY_BRL_DOT6 is an integer */
  if (!_cffi_check_int(*o, n, 502U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRL_DOT7(unsigned long long *o)
{
  int n = (KEY_BRL_DOT7) <= 0;
  *o = (unsigned long long)((KEY_BRL_DOT7) | 0);  /* check that KEY_BRL_DOT7 is an integer */
  if (!_cffi_check_int(*o, n, 503U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRL_DOT8(unsigned long long *o)
{
  int n = (KEY_BRL_DOT8) <= 0;
  *o = (unsigned long long)((KEY_BRL_DOT8) | 0);  /* check that KEY_BRL_DOT8 is an integer */
  if (!_cffi_check_int(*o, n, 504U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BRL_DOT9(unsigned long long *o)
{
  int n = (KEY_BRL_DOT9) <= 0;
  *o = (unsigned long long)((KEY_BRL_DOT9) | 0);  /* check that KEY_BRL_DOT9 is an integer */
  if (!_cffi_check_int(*o, n, 505U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_BUTTONCONFIG(unsigned long long *o)
{
  int n = (KEY_BUTTONCONFIG) <= 0;
  *o = (unsigned long long)((KEY_BUTTONCONFIG) | 0);  /* check that KEY_BUTTONCONFIG is an integer */
  if (!_cffi_check_int(*o, n, 576U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_C(unsigned long long *o)
{
  int n = (KEY_C) <= 0;
  *o = (unsigned long long)((KEY_C) | 0);  /* check that KEY_C is an integer */
  if (!_cffi_check_int(*o, n, 46U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CALC(unsigned long long *o)
{
  int n = (KEY_CALC) <= 0;
  *o = (unsigned long long)((KEY_CALC) | 0);  /* check that KEY_CALC is an integer */
  if (!_cffi_check_int(*o, n, 140U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CALENDAR(unsigned long long *o)
{
  int n = (KEY_CALENDAR) <= 0;
  *o = (unsigned long long)((KEY_CALENDAR) | 0);  /* check that KEY_CALENDAR is an integer */
  if (!_cffi_check_int(*o, n, 397U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CAMERA(unsigned long long *o)
{
  int n = (KEY_CAMERA) <= 0;
  *o = (unsigned long long)((KEY_CAMERA) | 0);  /* check that KEY_CAMERA is an integer */
  if (!_cffi_check_int(*o, n, 212U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CAMERA_DOWN(unsigned long long *o)
{
  int n = (KEY_CAMERA_DOWN) <= 0;
  *o = (unsigned long long)((KEY_CAMERA_DOWN) | 0);  /* check that KEY_CAMERA_DOWN is an integer */
  if (!_cffi_check_int(*o, n, 536U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CAMERA_FOCUS(unsigned long long *o)
{
  int n = (KEY_CAMERA_FOCUS) <= 0;
  *o = (unsigned long long)((KEY_CAMERA_FOCUS) | 0);  /* check that KEY_CAMERA_FOCUS is an integer */
  if (!_cffi_check_int(*o, n, 528U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CAMERA_LEFT(unsigned long long *o)
{
  int n = (KEY_CAMERA_LEFT) <= 0;
  *o = (unsigned long long)((KEY_CAMERA_LEFT) | 0);  /* check that KEY_CAMERA_LEFT is an integer */
  if (!_cffi_check_int(*o, n, 537U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CAMERA_RIGHT(unsigned long long *o)
{
  int n = (KEY_CAMERA_RIGHT) <= 0;
  *o = (unsigned long long)((KEY_CAMERA_RIGHT) | 0);  /* check that KEY_CAMERA_RIGHT is an integer */
  if (!_cffi_check_int(*o, n, 538U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CAMERA_UP(unsigned long long *o)
{
  int n = (KEY_CAMERA_UP) <= 0;
  *o = (unsigned long long)((KEY_CAMERA_UP) | 0);  /* check that KEY_CAMERA_UP is an integer */
  if (!_cffi_check_int(*o, n, 535U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CAMERA_ZOOMIN(unsigned long long *o)
{
  int n = (KEY_CAMERA_ZOOMIN) <= 0;
  *o = (unsigned long long)((KEY_CAMERA_ZOOMIN) | 0);  /* check that KEY_CAMERA_ZOOMIN is an integer */
  if (!_cffi_check_int(*o, n, 533U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CAMERA_ZOOMOUT(unsigned long long *o)
{
  int n = (KEY_CAMERA_ZOOMOUT) <= 0;
  *o = (unsigned long long)((KEY_CAMERA_ZOOMOUT) | 0);  /* check that KEY_CAMERA_ZOOMOUT is an integer */
  if (!_cffi_check_int(*o, n, 534U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CANCEL(unsigned long long *o)
{
  int n = (KEY_CANCEL) <= 0;
  *o = (unsigned long long)((KEY_CANCEL) | 0);  /* check that KEY_CANCEL is an integer */
  if (!_cffi_check_int(*o, n, 223U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CAPSLOCK(unsigned long long *o)
{
  int n = (KEY_CAPSLOCK) <= 0;
  *o = (unsigned long long)((KEY_CAPSLOCK) | 0);  /* check that KEY_CAPSLOCK is an integer */
  if (!_cffi_check_int(*o, n, 58U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CD(unsigned long long *o)
{
  int n = (KEY_CD) <= 0;
  *o = (unsigned long long)((KEY_CD) | 0);  /* check that KEY_CD is an integer */
  if (!_cffi_check_int(*o, n, 383U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CHANNEL(unsigned long long *o)
{
  int n = (KEY_CHANNEL) <= 0;
  *o = (unsigned long long)((KEY_CHANNEL) | 0);  /* check that KEY_CHANNEL is an integer */
  if (!_cffi_check_int(*o, n, 363U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CHANNELDOWN(unsigned long long *o)
{
  int n = (KEY_CHANNELDOWN) <= 0;
  *o = (unsigned long long)((KEY_CHANNELDOWN) | 0);  /* check that KEY_CHANNELDOWN is an integer */
  if (!_cffi_check_int(*o, n, 403U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CHANNELUP(unsigned long long *o)
{
  int n = (KEY_CHANNELUP) <= 0;
  *o = (unsigned long long)((KEY_CHANNELUP) | 0);  /* check that KEY_CHANNELUP is an integer */
  if (!_cffi_check_int(*o, n, 402U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CHAT(unsigned long long *o)
{
  int n = (KEY_CHAT) <= 0;
  *o = (unsigned long long)((KEY_CHAT) | 0);  /* check that KEY_CHAT is an integer */
  if (!_cffi_check_int(*o, n, 216U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CLEAR(unsigned long long *o)
{
  int n = (KEY_CLEAR) <= 0;
  *o = (unsigned long long)((KEY_CLEAR) | 0);  /* check that KEY_CLEAR is an integer */
  if (!_cffi_check_int(*o, n, 355U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CLOSE(unsigned long long *o)
{
  int n = (KEY_CLOSE) <= 0;
  *o = (unsigned long long)((KEY_CLOSE) | 0);  /* check that KEY_CLOSE is an integer */
  if (!_cffi_check_int(*o, n, 206U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CLOSECD(unsigned long long *o)
{
  int n = (KEY_CLOSECD) <= 0;
  *o = (unsigned long long)((KEY_CLOSECD) | 0);  /* check that KEY_CLOSECD is an integer */
  if (!_cffi_check_int(*o, n, 160U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_COFFEE(unsigned long long *o)
{
  int n = (KEY_COFFEE) <= 0;
  *o = (unsigned long long)((KEY_COFFEE) | 0);  /* check that KEY_COFFEE is an integer */
  if (!_cffi_check_int(*o, n, 152U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_COMMA(unsigned long long *o)
{
  int n = (KEY_COMMA) <= 0;
  *o = (unsigned long long)((KEY_COMMA) | 0);  /* check that KEY_COMMA is an integer */
  if (!_cffi_check_int(*o, n, 51U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_COMPOSE(unsigned long long *o)
{
  int n = (KEY_COMPOSE) <= 0;
  *o = (unsigned long long)((KEY_COMPOSE) | 0);  /* check that KEY_COMPOSE is an integer */
  if (!_cffi_check_int(*o, n, 127U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_COMPUTER(unsigned long long *o)
{
  int n = (KEY_COMPUTER) <= 0;
  *o = (unsigned long long)((KEY_COMPUTER) | 0);  /* check that KEY_COMPUTER is an integer */
  if (!_cffi_check_int(*o, n, 157U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CONFIG(unsigned long long *o)
{
  int n = (KEY_CONFIG) <= 0;
  *o = (unsigned long long)((KEY_CONFIG) | 0);  /* check that KEY_CONFIG is an integer */
  if (!_cffi_check_int(*o, n, 171U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CONNECT(unsigned long long *o)
{
  int n = (KEY_CONNECT) <= 0;
  *o = (unsigned long long)((KEY_CONNECT) | 0);  /* check that KEY_CONNECT is an integer */
  if (!_cffi_check_int(*o, n, 218U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CONTEXT_MENU(unsigned long long *o)
{
  int n = (KEY_CONTEXT_MENU) <= 0;
  *o = (unsigned long long)((KEY_CONTEXT_MENU) | 0);  /* check that KEY_CONTEXT_MENU is an integer */
  if (!_cffi_check_int(*o, n, 438U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CONTROLPANEL(unsigned long long *o)
{
  int n = (KEY_CONTROLPANEL) <= 0;
  *o = (unsigned long long)((KEY_CONTROLPANEL) | 0);  /* check that KEY_CONTROLPANEL is an integer */
  if (!_cffi_check_int(*o, n, 579U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_COPY(unsigned long long *o)
{
  int n = (KEY_COPY) <= 0;
  *o = (unsigned long long)((KEY_COPY) | 0);  /* check that KEY_COPY is an integer */
  if (!_cffi_check_int(*o, n, 133U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CUT(unsigned long long *o)
{
  int n = (KEY_CUT) <= 0;
  *o = (unsigned long long)((KEY_CUT) | 0);  /* check that KEY_CUT is an integer */
  if (!_cffi_check_int(*o, n, 137U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_CYCLEWINDOWS(unsigned long long *o)
{
  int n = (KEY_CYCLEWINDOWS) <= 0;
  *o = (unsigned long long)((KEY_CYCLEWINDOWS) | 0);  /* check that KEY_CYCLEWINDOWS is an integer */
  if (!_cffi_check_int(*o, n, 154U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_D(unsigned long long *o)
{
  int n = (KEY_D) <= 0;
  *o = (unsigned long long)((KEY_D) | 0);  /* check that KEY_D is an integer */
  if (!_cffi_check_int(*o, n, 32U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DASHBOARD(unsigned long long *o)
{
  int n = (KEY_DASHBOARD) <= 0;
  *o = (unsigned long long)((KEY_DASHBOARD) | 0);  /* check that KEY_DASHBOARD is an integer */
  if (!_cffi_check_int(*o, n, 204U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DATA(unsigned long long *o)
{
  int n = (KEY_DATA) <= 0;
  *o = (unsigned long long)((KEY_DATA) | 0);  /* check that KEY_DATA is an integer */
  if (!_cffi_check_int(*o, n, 631U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DATABASE(unsigned long long *o)
{
  int n = (KEY_DATABASE) <= 0;
  *o = (unsigned long long)((KEY_DATABASE) | 0);  /* check that KEY_DATABASE is an integer */
  if (!_cffi_check_int(*o, n, 426U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DELETE(unsigned long long *o)
{
  int n = (KEY_DELETE) <= 0;
  *o = (unsigned long long)((KEY_DELETE) | 0);  /* check that KEY_DELETE is an integer */
  if (!_cffi_check_int(*o, n, 111U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DELETEFILE(unsigned long long *o)
{
  int n = (KEY_DELETEFILE) <= 0;
  *o = (unsigned long long)((KEY_DELETEFILE) | 0);  /* check that KEY_DELETEFILE is an integer */
  if (!_cffi_check_int(*o, n, 146U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DEL_EOL(unsigned long long *o)
{
  int n = (KEY_DEL_EOL) <= 0;
  *o = (unsigned long long)((KEY_DEL_EOL) | 0);  /* check that KEY_DEL_EOL is an integer */
  if (!_cffi_check_int(*o, n, 448U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DEL_EOS(unsigned long long *o)
{
  int n = (KEY_DEL_EOS) <= 0;
  *o = (unsigned long long)((KEY_DEL_EOS) | 0);  /* check that KEY_DEL_EOS is an integer */
  if (!_cffi_check_int(*o, n, 449U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DEL_LINE(unsigned long long *o)
{
  int n = (KEY_DEL_LINE) <= 0;
  *o = (unsigned long long)((KEY_DEL_LINE) | 0);  /* check that KEY_DEL_LINE is an integer */
  if (!_cffi_check_int(*o, n, 451U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DIGITS(unsigned long long *o)
{
  int n = (KEY_DIGITS) <= 0;
  *o = (unsigned long long)((KEY_DIGITS) | 0);  /* check that KEY_DIGITS is an integer */
  if (!_cffi_check_int(*o, n, 413U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DIRECTORY(unsigned long long *o)
{
  int n = (KEY_DIRECTORY) <= 0;
  *o = (unsigned long long)((KEY_DIRECTORY) | 0);  /* check that KEY_DIRECTORY is an integer */
  if (!_cffi_check_int(*o, n, 394U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DISPLAYTOGGLE(unsigned long long *o)
{
  int n = (KEY_DISPLAYTOGGLE) <= 0;
  *o = (unsigned long long)((KEY_DISPLAYTOGGLE) | 0);  /* check that KEY_DISPLAYTOGGLE is an integer */
  if (!_cffi_check_int(*o, n, 431U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DISPLAY_OFF(unsigned long long *o)
{
  int n = (KEY_DISPLAY_OFF) <= 0;
  *o = (unsigned long long)((KEY_DISPLAY_OFF) | 0);  /* check that KEY_DISPLAY_OFF is an integer */
  if (!_cffi_check_int(*o, n, 245U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DOCUMENTS(unsigned long long *o)
{
  int n = (KEY_DOCUMENTS) <= 0;
  *o = (unsigned long long)((KEY_DOCUMENTS) | 0);  /* check that KEY_DOCUMENTS is an integer */
  if (!_cffi_check_int(*o, n, 235U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DOLLAR(unsigned long long *o)
{
  int n = (KEY_DOLLAR) <= 0;
  *o = (unsigned long long)((KEY_DOLLAR) | 0);  /* check that KEY_DOLLAR is an integer */
  if (!_cffi_check_int(*o, n, 434U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DOT(unsigned long long *o)
{
  int n = (KEY_DOT) <= 0;
  *o = (unsigned long long)((KEY_DOT) | 0);  /* check that KEY_DOT is an integer */
  if (!_cffi_check_int(*o, n, 52U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DOWN(unsigned long long *o)
{
  int n = (KEY_DOWN) <= 0;
  *o = (unsigned long long)((KEY_DOWN) | 0);  /* check that KEY_DOWN is an integer */
  if (!_cffi_check_int(*o, n, 108U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_DVD(unsigned long long *o)
{
  int n = (KEY_DVD) <= 0;
  *o = (unsigned long long)((KEY_DVD) | 0);  /* check that KEY_DVD is an integer */
  if (!_cffi_check_int(*o, n, 389U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_E(unsigned long long *o)
{
  int n = (KEY_E) <= 0;
  *o = (unsigned long long)((KEY_E) | 0);  /* check that KEY_E is an integer */
  if (!_cffi_check_int(*o, n, 18U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_EDIT(unsigned long long *o)
{
  int n = (KEY_EDIT) <= 0;
  *o = (unsigned long long)((KEY_EDIT) | 0);  /* check that KEY_EDIT is an integer */
  if (!_cffi_check_int(*o, n, 176U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_EDITOR(unsigned long long *o)
{
  int n = (KEY_EDITOR) <= 0;
  *o = (unsigned long long)((KEY_EDITOR) | 0);  /* check that KEY_EDITOR is an integer */
  if (!_cffi_check_int(*o, n, 422U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_EJECTCD(unsigned long long *o)
{
  int n = (KEY_EJECTCD) <= 0;
  *o = (unsigned long long)((KEY_EJECTCD) | 0);  /* check that KEY_EJECTCD is an integer */
  if (!_cffi_check_int(*o, n, 161U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_EJECTCLOSECD(unsigned long long *o)
{
  int n = (KEY_EJECTCLOSECD) <= 0;
  *o = (unsigned long long)((KEY_EJECTCLOSECD) | 0);  /* check that KEY_EJECTCLOSECD is an integer */
  if (!_cffi_check_int(*o, n, 162U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_EMAIL(unsigned long long *o)
{
  int n = (KEY_EMAIL) <= 0;
  *o = (unsigned long long)((KEY_EMAIL) | 0);  /* check that KEY_EMAIL is an integer */
  if (!_cffi_check_int(*o, n, 215U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_END(unsigned long long *o)
{
  int n = (KEY_END) <= 0;
  *o = (unsigned long long)((KEY_END) | 0);  /* check that KEY_END is an integer */
  if (!_cffi_check_int(*o, n, 107U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ENTER(unsigned long long *o)
{
  int n = (KEY_ENTER) <= 0;
  *o = (unsigned long long)((KEY_ENTER) | 0);  /* check that KEY_ENTER is an integer */
  if (!_cffi_check_int(*o, n, 28U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_EPG(unsigned long long *o)
{
  int n = (KEY_EPG) <= 0;
  *o = (unsigned long long)((KEY_EPG) | 0);  /* check that KEY_EPG is an integer */
  if (!_cffi_check_int(*o, n, 365U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_EQUAL(unsigned long long *o)
{
  int n = (KEY_EQUAL) <= 0;
  *o = (unsigned long long)((KEY_EQUAL) | 0);  /* check that KEY_EQUAL is an integer */
  if (!_cffi_check_int(*o, n, 13U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ESC(unsigned long long *o)
{
  int n = (KEY_ESC) <= 0;
  *o = (unsigned long long)((KEY_ESC) | 0);  /* check that KEY_ESC is an integer */
  if (!_cffi_check_int(*o, n, 1U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_EURO(unsigned long long *o)
{
  int n = (KEY_EURO) <= 0;
  *o = (unsigned long long)((KEY_EURO) | 0);  /* check that KEY_EURO is an integer */
  if (!_cffi_check_int(*o, n, 435U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_EXIT(unsigned long long *o)
{
  int n = (KEY_EXIT) <= 0;
  *o = (unsigned long long)((KEY_EXIT) | 0);  /* check that KEY_EXIT is an integer */
  if (!_cffi_check_int(*o, n, 174U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F(unsigned long long *o)
{
  int n = (KEY_F) <= 0;
  *o = (unsigned long long)((KEY_F) | 0);  /* check that KEY_F is an integer */
  if (!_cffi_check_int(*o, n, 33U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F1(unsigned long long *o)
{
  int n = (KEY_F1) <= 0;
  *o = (unsigned long long)((KEY_F1) | 0);  /* check that KEY_F1 is an integer */
  if (!_cffi_check_int(*o, n, 59U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F10(unsigned long long *o)
{
  int n = (KEY_F10) <= 0;
  *o = (unsigned long long)((KEY_F10) | 0);  /* check that KEY_F10 is an integer */
  if (!_cffi_check_int(*o, n, 68U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F11(unsigned long long *o)
{
  int n = (KEY_F11) <= 0;
  *o = (unsigned long long)((KEY_F11) | 0);  /* check that KEY_F11 is an integer */
  if (!_cffi_check_int(*o, n, 87U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F12(unsigned long long *o)
{
  int n = (KEY_F12) <= 0;
  *o = (unsigned long long)((KEY_F12) | 0);  /* check that KEY_F12 is an integer */
  if (!_cffi_check_int(*o, n, 88U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F13(unsigned long long *o)
{
  int n = (KEY_F13) <= 0;
  *o = (unsigned long long)((KEY_F13) | 0);  /* check that KEY_F13 is an integer */
  if (!_cffi_check_int(*o, n, 183U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F14(unsigned long long *o)
{
  int n = (KEY_F14) <= 0;
  *o = (unsigned long long)((KEY_F14) | 0);  /* check that KEY_F14 is an integer */
  if (!_cffi_check_int(*o, n, 184U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F15(unsigned long long *o)
{
  int n = (KEY_F15) <= 0;
  *o = (unsigned long long)((KEY_F15) | 0);  /* check that KEY_F15 is an integer */
  if (!_cffi_check_int(*o, n, 185U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F16(unsigned long long *o)
{
  int n = (KEY_F16) <= 0;
  *o = (unsigned long long)((KEY_F16) | 0);  /* check that KEY_F16 is an integer */
  if (!_cffi_check_int(*o, n, 186U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F17(unsigned long long *o)
{
  int n = (KEY_F17) <= 0;
  *o = (unsigned long long)((KEY_F17) | 0);  /* check that KEY_F17 is an integer */
  if (!_cffi_check_int(*o, n, 187U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F18(unsigned long long *o)
{
  int n = (KEY_F18) <= 0;
  *o = (unsigned long long)((KEY_F18) | 0);  /* check that KEY_F18 is an integer */
  if (!_cffi_check_int(*o, n, 188U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F19(unsigned long long *o)
{
  int n = (KEY_F19) <= 0;
  *o = (unsigned long long)((KEY_F19) | 0);  /* check that KEY_F19 is an integer */
  if (!_cffi_check_int(*o, n, 189U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F2(unsigned long long *o)
{
  int n = (KEY_F2) <= 0;
  *o = (unsigned long long)((KEY_F2) | 0);  /* check that KEY_F2 is an integer */
  if (!_cffi_check_int(*o, n, 60U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F20(unsigned long long *o)
{
  int n = (KEY_F20) <= 0;
  *o = (unsigned long long)((KEY_F20) | 0);  /* check that KEY_F20 is an integer */
  if (!_cffi_check_int(*o, n, 190U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F21(unsigned long long *o)
{
  int n = (KEY_F21) <= 0;
  *o = (unsigned long long)((KEY_F21) | 0);  /* check that KEY_F21 is an integer */
  if (!_cffi_check_int(*o, n, 191U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F22(unsigned long long *o)
{
  int n = (KEY_F22) <= 0;
  *o = (unsigned long long)((KEY_F22) | 0);  /* check that KEY_F22 is an integer */
  if (!_cffi_check_int(*o, n, 192U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F23(unsigned long long *o)
{
  int n = (KEY_F23) <= 0;
  *o = (unsigned long long)((KEY_F23) | 0);  /* check that KEY_F23 is an integer */
  if (!_cffi_check_int(*o, n, 193U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F24(unsigned long long *o)
{
  int n = (KEY_F24) <= 0;
  *o = (unsigned long long)((KEY_F24) | 0);  /* check that KEY_F24 is an integer */
  if (!_cffi_check_int(*o, n, 194U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F3(unsigned long long *o)
{
  int n = (KEY_F3) <= 0;
  *o = (unsigned long long)((KEY_F3) | 0);  /* check that KEY_F3 is an integer */
  if (!_cffi_check_int(*o, n, 61U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F4(unsigned long long *o)
{
  int n = (KEY_F4) <= 0;
  *o = (unsigned long long)((KEY_F4) | 0);  /* check that KEY_F4 is an integer */
  if (!_cffi_check_int(*o, n, 62U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F5(unsigned long long *o)
{
  int n = (KEY_F5) <= 0;
  *o = (unsigned long long)((KEY_F5) | 0);  /* check that KEY_F5 is an integer */
  if (!_cffi_check_int(*o, n, 63U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F6(unsigned long long *o)
{
  int n = (KEY_F6) <= 0;
  *o = (unsigned long long)((KEY_F6) | 0);  /* check that KEY_F6 is an integer */
  if (!_cffi_check_int(*o, n, 64U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F7(unsigned long long *o)
{
  int n = (KEY_F7) <= 0;
  *o = (unsigned long long)((KEY_F7) | 0);  /* check that KEY_F7 is an integer */
  if (!_cffi_check_int(*o, n, 65U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F8(unsigned long long *o)
{
  int n = (KEY_F8) <= 0;
  *o = (unsigned long long)((KEY_F8) | 0);  /* check that KEY_F8 is an integer */
  if (!_cffi_check_int(*o, n, 66U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_F9(unsigned long long *o)
{
  int n = (KEY_F9) <= 0;
  *o = (unsigned long long)((KEY_F9) | 0);  /* check that KEY_F9 is an integer */
  if (!_cffi_check_int(*o, n, 67U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FASTFORWARD(unsigned long long *o)
{
  int n = (KEY_FASTFORWARD) <= 0;
  *o = (unsigned long long)((KEY_FASTFORWARD) | 0);  /* check that KEY_FASTFORWARD is an integer */
  if (!_cffi_check_int(*o, n, 208U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FASTREVERSE(unsigned long long *o)
{
  int n = (KEY_FASTREVERSE) <= 0;
  *o = (unsigned long long)((KEY_FASTREVERSE) | 0);  /* check that KEY_FASTREVERSE is an integer */
  if (!_cffi_check_int(*o, n, 629U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FAVORITES(unsigned long long *o)
{
  int n = (KEY_FAVORITES) <= 0;
  *o = (unsigned long long)((KEY_FAVORITES) | 0);  /* check that KEY_FAVORITES is an integer */
  if (!_cffi_check_int(*o, n, 364U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FILE(unsigned long long *o)
{
  int n = (KEY_FILE) <= 0;
  *o = (unsigned long long)((KEY_FILE) | 0);  /* check that KEY_FILE is an integer */
  if (!_cffi_check_int(*o, n, 144U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FINANCE(unsigned long long *o)
{
  int n = (KEY_FINANCE) <= 0;
  *o = (unsigned long long)((KEY_FINANCE) | 0);  /* check that KEY_FINANCE is an integer */
  if (!_cffi_check_int(*o, n, 219U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FIND(unsigned long long *o)
{
  int n = (KEY_FIND) <= 0;
  *o = (unsigned long long)((KEY_FIND) | 0);  /* check that KEY_FIND is an integer */
  if (!_cffi_check_int(*o, n, 136U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FIRST(unsigned long long *o)
{
  int n = (KEY_FIRST) <= 0;
  *o = (unsigned long long)((KEY_FIRST) | 0);  /* check that KEY_FIRST is an integer */
  if (!_cffi_check_int(*o, n, 404U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN(unsigned long long *o)
{
  int n = (KEY_FN) <= 0;
  *o = (unsigned long long)((KEY_FN) | 0);  /* check that KEY_FN is an integer */
  if (!_cffi_check_int(*o, n, 464U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_1(unsigned long long *o)
{
  int n = (KEY_FN_1) <= 0;
  *o = (unsigned long long)((KEY_FN_1) | 0);  /* check that KEY_FN_1 is an integer */
  if (!_cffi_check_int(*o, n, 478U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_2(unsigned long long *o)
{
  int n = (KEY_FN_2) <= 0;
  *o = (unsigned long long)((KEY_FN_2) | 0);  /* check that KEY_FN_2 is an integer */
  if (!_cffi_check_int(*o, n, 479U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_B(unsigned long long *o)
{
  int n = (KEY_FN_B) <= 0;
  *o = (unsigned long long)((KEY_FN_B) | 0);  /* check that KEY_FN_B is an integer */
  if (!_cffi_check_int(*o, n, 484U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_D(unsigned long long *o)
{
  int n = (KEY_FN_D) <= 0;
  *o = (unsigned long long)((KEY_FN_D) | 0);  /* check that KEY_FN_D is an integer */
  if (!_cffi_check_int(*o, n, 480U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_E(unsigned long long *o)
{
  int n = (KEY_FN_E) <= 0;
  *o = (unsigned long long)((KEY_FN_E) | 0);  /* check that KEY_FN_E is an integer */
  if (!_cffi_check_int(*o, n, 481U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_ESC(unsigned long long *o)
{
  int n = (KEY_FN_ESC) <= 0;
  *o = (unsigned long long)((KEY_FN_ESC) | 0);  /* check that KEY_FN_ESC is an integer */
  if (!_cffi_check_int(*o, n, 465U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_F(unsigned long long *o)
{
  int n = (KEY_FN_F) <= 0;
  *o = (unsigned long long)((KEY_FN_F) | 0);  /* check that KEY_FN_F is an integer */
  if (!_cffi_check_int(*o, n, 482U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_F1(unsigned long long *o)
{
  int n = (KEY_FN_F1) <= 0;
  *o = (unsigned long long)((KEY_FN_F1) | 0);  /* check that KEY_FN_F1 is an integer */
  if (!_cffi_check_int(*o, n, 466U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_F10(unsigned long long *o)
{
  int n = (KEY_FN_F10) <= 0;
  *o = (unsigned long long)((KEY_FN_F10) | 0);  /* check that KEY_FN_F10 is an integer */
  if (!_cffi_check_int(*o, n, 475U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_F11(unsigned long long *o)
{
  int n = (KEY_FN_F11) <= 0;
  *o = (unsigned long long)((KEY_FN_F11) | 0);  /* check that KEY_FN_F11 is an integer */
  if (!_cffi_check_int(*o, n, 476U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_F12(unsigned long long *o)
{
  int n = (KEY_FN_F12) <= 0;
  *o = (unsigned long long)((KEY_FN_F12) | 0);  /* check that KEY_FN_F12 is an integer */
  if (!_cffi_check_int(*o, n, 477U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_F2(unsigned long long *o)
{
  int n = (KEY_FN_F2) <= 0;
  *o = (unsigned long long)((KEY_FN_F2) | 0);  /* check that KEY_FN_F2 is an integer */
  if (!_cffi_check_int(*o, n, 467U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_F3(unsigned long long *o)
{
  int n = (KEY_FN_F3) <= 0;
  *o = (unsigned long long)((KEY_FN_F3) | 0);  /* check that KEY_FN_F3 is an integer */
  if (!_cffi_check_int(*o, n, 468U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_F4(unsigned long long *o)
{
  int n = (KEY_FN_F4) <= 0;
  *o = (unsigned long long)((KEY_FN_F4) | 0);  /* check that KEY_FN_F4 is an integer */
  if (!_cffi_check_int(*o, n, 469U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_F5(unsigned long long *o)
{
  int n = (KEY_FN_F5) <= 0;
  *o = (unsigned long long)((KEY_FN_F5) | 0);  /* check that KEY_FN_F5 is an integer */
  if (!_cffi_check_int(*o, n, 470U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_F6(unsigned long long *o)
{
  int n = (KEY_FN_F6) <= 0;
  *o = (unsigned long long)((KEY_FN_F6) | 0);  /* check that KEY_FN_F6 is an integer */
  if (!_cffi_check_int(*o, n, 471U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_F7(unsigned long long *o)
{
  int n = (KEY_FN_F7) <= 0;
  *o = (unsigned long long)((KEY_FN_F7) | 0);  /* check that KEY_FN_F7 is an integer */
  if (!_cffi_check_int(*o, n, 472U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_F8(unsigned long long *o)
{
  int n = (KEY_FN_F8) <= 0;
  *o = (unsigned long long)((KEY_FN_F8) | 0);  /* check that KEY_FN_F8 is an integer */
  if (!_cffi_check_int(*o, n, 473U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_F9(unsigned long long *o)
{
  int n = (KEY_FN_F9) <= 0;
  *o = (unsigned long long)((KEY_FN_F9) | 0);  /* check that KEY_FN_F9 is an integer */
  if (!_cffi_check_int(*o, n, 474U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FN_S(unsigned long long *o)
{
  int n = (KEY_FN_S) <= 0;
  *o = (unsigned long long)((KEY_FN_S) | 0);  /* check that KEY_FN_S is an integer */
  if (!_cffi_check_int(*o, n, 483U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FORWARD(unsigned long long *o)
{
  int n = (KEY_FORWARD) <= 0;
  *o = (unsigned long long)((KEY_FORWARD) | 0);  /* check that KEY_FORWARD is an integer */
  if (!_cffi_check_int(*o, n, 159U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FORWARDMAIL(unsigned long long *o)
{
  int n = (KEY_FORWARDMAIL) <= 0;
  *o = (unsigned long long)((KEY_FORWARDMAIL) | 0);  /* check that KEY_FORWARDMAIL is an integer */
  if (!_cffi_check_int(*o, n, 233U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FRAMEBACK(unsigned long long *o)
{
  int n = (KEY_FRAMEBACK) <= 0;
  *o = (unsigned long long)((KEY_FRAMEBACK) | 0);  /* check that KEY_FRAMEBACK is an integer */
  if (!_cffi_check_int(*o, n, 436U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FRAMEFORWARD(unsigned long long *o)
{
  int n = (KEY_FRAMEFORWARD) <= 0;
  *o = (unsigned long long)((KEY_FRAMEFORWARD) | 0);  /* check that KEY_FRAMEFORWARD is an integer */
  if (!_cffi_check_int(*o, n, 437U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_FRONT(unsigned long long *o)
{
  int n = (KEY_FRONT) <= 0;
  *o = (unsigned long long)((KEY_FRONT) | 0);  /* check that KEY_FRONT is an integer */
  if (!_cffi_check_int(*o, n, 132U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_G(unsigned long long *o)
{
  int n = (KEY_G) <= 0;
  *o = (unsigned long long)((KEY_G) | 0);  /* check that KEY_G is an integer */
  if (!_cffi_check_int(*o, n, 34U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_GAMES(unsigned long long *o)
{
  int n = (KEY_GAMES) <= 0;
  *o = (unsigned long long)((KEY_GAMES) | 0);  /* check that KEY_GAMES is an integer */
  if (!_cffi_check_int(*o, n, 417U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_GOTO(unsigned long long *o)
{
  int n = (KEY_GOTO) <= 0;
  *o = (unsigned long long)((KEY_GOTO) | 0);  /* check that KEY_GOTO is an integer */
  if (!_cffi_check_int(*o, n, 354U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_GRAPHICSEDITOR(unsigned long long *o)
{
  int n = (KEY_GRAPHICSEDITOR) <= 0;
  *o = (unsigned long long)((KEY_GRAPHICSEDITOR) | 0);  /* check that KEY_GRAPHICSEDITOR is an integer */
  if (!_cffi_check_int(*o, n, 424U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_GRAVE(unsigned long long *o)
{
  int n = (KEY_GRAVE) <= 0;
  *o = (unsigned long long)((KEY_GRAVE) | 0);  /* check that KEY_GRAVE is an integer */
  if (!_cffi_check_int(*o, n, 41U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_GREEN(unsigned long long *o)
{
  int n = (KEY_GREEN) <= 0;
  *o = (unsigned long long)((KEY_GREEN) | 0);  /* check that KEY_GREEN is an integer */
  if (!_cffi_check_int(*o, n, 399U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_H(unsigned long long *o)
{
  int n = (KEY_H) <= 0;
  *o = (unsigned long long)((KEY_H) | 0);  /* check that KEY_H is an integer */
  if (!_cffi_check_int(*o, n, 35U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_HANGEUL(unsigned long long *o)
{
  int n = (KEY_HANGEUL) <= 0;
  *o = (unsigned long long)((KEY_HANGEUL) | 0);  /* check that KEY_HANGEUL is an integer */
  if (!_cffi_check_int(*o, n, 122U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_HANJA(unsigned long long *o)
{
  int n = (KEY_HANJA) <= 0;
  *o = (unsigned long long)((KEY_HANJA) | 0);  /* check that KEY_HANJA is an integer */
  if (!_cffi_check_int(*o, n, 123U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_HELP(unsigned long long *o)
{
  int n = (KEY_HELP) <= 0;
  *o = (unsigned long long)((KEY_HELP) | 0);  /* check that KEY_HELP is an integer */
  if (!_cffi_check_int(*o, n, 138U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_HENKAN(unsigned long long *o)
{
  int n = (KEY_HENKAN) <= 0;
  *o = (unsigned long long)((KEY_HENKAN) | 0);  /* check that KEY_HENKAN is an integer */
  if (!_cffi_check_int(*o, n, 92U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_HIRAGANA(unsigned long long *o)
{
  int n = (KEY_HIRAGANA) <= 0;
  *o = (unsigned long long)((KEY_HIRAGANA) | 0);  /* check that KEY_HIRAGANA is an integer */
  if (!_cffi_check_int(*o, n, 91U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_HOME(unsigned long long *o)
{
  int n = (KEY_HOME) <= 0;
  *o = (unsigned long long)((KEY_HOME) | 0);  /* check that KEY_HOME is an integer */
  if (!_cffi_check_int(*o, n, 102U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_HOMEPAGE(unsigned long long *o)
{
  int n = (KEY_HOMEPAGE) <= 0;
  *o = (unsigned long long)((KEY_HOMEPAGE) | 0);  /* check that KEY_HOMEPAGE is an integer */
  if (!_cffi_check_int(*o, n, 172U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_HP(unsigned long long *o)
{
  int n = (KEY_HP) <= 0;
  *o = (unsigned long long)((KEY_HP) | 0);  /* check that KEY_HP is an integer */
  if (!_cffi_check_int(*o, n, 211U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_I(unsigned long long *o)
{
  int n = (KEY_I) <= 0;
  *o = (unsigned long long)((KEY_I) | 0);  /* check that KEY_I is an integer */
  if (!_cffi_check_int(*o, n, 23U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_IMAGES(unsigned long long *o)
{
  int n = (KEY_IMAGES) <= 0;
  *o = (unsigned long long)((KEY_IMAGES) | 0);  /* check that KEY_IMAGES is an integer */
  if (!_cffi_check_int(*o, n, 442U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_INFO(unsigned long long *o)
{
  int n = (KEY_INFO) <= 0;
  *o = (unsigned long long)((KEY_INFO) | 0);  /* check that KEY_INFO is an integer */
  if (!_cffi_check_int(*o, n, 358U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_INSERT(unsigned long long *o)
{
  int n = (KEY_INSERT) <= 0;
  *o = (unsigned long long)((KEY_INSERT) | 0);  /* check that KEY_INSERT is an integer */
  if (!_cffi_check_int(*o, n, 110U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_INS_LINE(unsigned long long *o)
{
  int n = (KEY_INS_LINE) <= 0;
  *o = (unsigned long long)((KEY_INS_LINE) | 0);  /* check that KEY_INS_LINE is an integer */
  if (!_cffi_check_int(*o, n, 450U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ISO(unsigned long long *o)
{
  int n = (KEY_ISO) <= 0;
  *o = (unsigned long long)((KEY_ISO) | 0);  /* check that KEY_ISO is an integer */
  if (!_cffi_check_int(*o, n, 170U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_J(unsigned long long *o)
{
  int n = (KEY_J) <= 0;
  *o = (unsigned long long)((KEY_J) | 0);  /* check that KEY_J is an integer */
  if (!_cffi_check_int(*o, n, 36U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_JOURNAL(unsigned long long *o)
{
  int n = (KEY_JOURNAL) <= 0;
  *o = (unsigned long long)((KEY_JOURNAL) | 0);  /* check that KEY_JOURNAL is an integer */
  if (!_cffi_check_int(*o, n, 578U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_K(unsigned long long *o)
{
  int n = (KEY_K) <= 0;
  *o = (unsigned long long)((KEY_K) | 0);  /* check that KEY_K is an integer */
  if (!_cffi_check_int(*o, n, 37U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KATAKANA(unsigned long long *o)
{
  int n = (KEY_KATAKANA) <= 0;
  *o = (unsigned long long)((KEY_KATAKANA) | 0);  /* check that KEY_KATAKANA is an integer */
  if (!_cffi_check_int(*o, n, 90U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KATAKANAHIRAGANA(unsigned long long *o)
{
  int n = (KEY_KATAKANAHIRAGANA) <= 0;
  *o = (unsigned long long)((KEY_KATAKANAHIRAGANA) | 0);  /* check that KEY_KATAKANAHIRAGANA is an integer */
  if (!_cffi_check_int(*o, n, 93U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KBDILLUMDOWN(unsigned long long *o)
{
  int n = (KEY_KBDILLUMDOWN) <= 0;
  *o = (unsigned long long)((KEY_KBDILLUMDOWN) | 0);  /* check that KEY_KBDILLUMDOWN is an integer */
  if (!_cffi_check_int(*o, n, 229U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KBDILLUMTOGGLE(unsigned long long *o)
{
  int n = (KEY_KBDILLUMTOGGLE) <= 0;
  *o = (unsigned long long)((KEY_KBDILLUMTOGGLE) | 0);  /* check that KEY_KBDILLUMTOGGLE is an integer */
  if (!_cffi_check_int(*o, n, 228U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KBDILLUMUP(unsigned long long *o)
{
  int n = (KEY_KBDILLUMUP) <= 0;
  *o = (unsigned long long)((KEY_KBDILLUMUP) | 0);  /* check that KEY_KBDILLUMUP is an integer */
  if (!_cffi_check_int(*o, n, 230U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KBDINPUTASSIST_ACCEPT(unsigned long long *o)
{
  int n = (KEY_KBDINPUTASSIST_ACCEPT) <= 0;
  *o = (unsigned long long)((KEY_KBDINPUTASSIST_ACCEPT) | 0);  /* check that KEY_KBDINPUTASSIST_ACCEPT is an integer */
  if (!_cffi_check_int(*o, n, 612U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KBDINPUTASSIST_CANCEL(unsigned long long *o)
{
  int n = (KEY_KBDINPUTASSIST_CANCEL) <= 0;
  *o = (unsigned long long)((KEY_KBDINPUTASSIST_CANCEL) | 0);  /* check that KEY_KBDINPUTASSIST_CANCEL is an integer */
  if (!_cffi_check_int(*o, n, 613U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KBDINPUTASSIST_NEXT(unsigned long long *o)
{
  int n = (KEY_KBDINPUTASSIST_NEXT) <= 0;
  *o = (unsigned long long)((KEY_KBDINPUTASSIST_NEXT) | 0);  /* check that KEY_KBDINPUTASSIST_NEXT is an integer */
  if (!_cffi_check_int(*o, n, 609U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KBDINPUTASSIST_NEXTGROUP(unsigned long long *o)
{
  int n = (KEY_KBDINPUTASSIST_NEXTGROUP) <= 0;
  *o = (unsigned long long)((KEY_KBDINPUTASSIST_NEXTGROUP) | 0);  /* check that KEY_KBDINPUTASSIST_NEXTGROUP is an integer */
  if (!_cffi_check_int(*o, n, 611U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KBDINPUTASSIST_PREV(unsigned long long *o)
{
  int n = (KEY_KBDINPUTASSIST_PREV) <= 0;
  *o = (unsigned long long)((KEY_KBDINPUTASSIST_PREV) | 0);  /* check that KEY_KBDINPUTASSIST_PREV is an integer */
  if (!_cffi_check_int(*o, n, 608U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KBDINPUTASSIST_PREVGROUP(unsigned long long *o)
{
  int n = (KEY_KBDINPUTASSIST_PREVGROUP) <= 0;
  *o = (unsigned long long)((KEY_KBDINPUTASSIST_PREVGROUP) | 0);  /* check that KEY_KBDINPUTASSIST_PREVGROUP is an integer */
  if (!_cffi_check_int(*o, n, 610U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KEYBOARD(unsigned long long *o)
{
  int n = (KEY_KEYBOARD) <= 0;
  *o = (unsigned long long)((KEY_KEYBOARD) | 0);  /* check that KEY_KEYBOARD is an integer */
  if (!_cffi_check_int(*o, n, 374U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KP0(unsigned long long *o)
{
  int n = (KEY_KP0) <= 0;
  *o = (unsigned long long)((KEY_KP0) | 0);  /* check that KEY_KP0 is an integer */
  if (!_cffi_check_int(*o, n, 82U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KP1(unsigned long long *o)
{
  int n = (KEY_KP1) <= 0;
  *o = (unsigned long long)((KEY_KP1) | 0);  /* check that KEY_KP1 is an integer */
  if (!_cffi_check_int(*o, n, 79U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KP2(unsigned long long *o)
{
  int n = (KEY_KP2) <= 0;
  *o = (unsigned long long)((KEY_KP2) | 0);  /* check that KEY_KP2 is an integer */
  if (!_cffi_check_int(*o, n, 80U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KP3(unsigned long long *o)
{
  int n = (KEY_KP3) <= 0;
  *o = (unsigned long long)((KEY_KP3) | 0);  /* check that KEY_KP3 is an integer */
  if (!_cffi_check_int(*o, n, 81U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KP4(unsigned long long *o)
{
  int n = (KEY_KP4) <= 0;
  *o = (unsigned long long)((KEY_KP4) | 0);  /* check that KEY_KP4 is an integer */
  if (!_cffi_check_int(*o, n, 75U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KP5(unsigned long long *o)
{
  int n = (KEY_KP5) <= 0;
  *o = (unsigned long long)((KEY_KP5) | 0);  /* check that KEY_KP5 is an integer */
  if (!_cffi_check_int(*o, n, 76U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KP6(unsigned long long *o)
{
  int n = (KEY_KP6) <= 0;
  *o = (unsigned long long)((KEY_KP6) | 0);  /* check that KEY_KP6 is an integer */
  if (!_cffi_check_int(*o, n, 77U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KP7(unsigned long long *o)
{
  int n = (KEY_KP7) <= 0;
  *o = (unsigned long long)((KEY_KP7) | 0);  /* check that KEY_KP7 is an integer */
  if (!_cffi_check_int(*o, n, 71U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KP8(unsigned long long *o)
{
  int n = (KEY_KP8) <= 0;
  *o = (unsigned long long)((KEY_KP8) | 0);  /* check that KEY_KP8 is an integer */
  if (!_cffi_check_int(*o, n, 72U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KP9(unsigned long long *o)
{
  int n = (KEY_KP9) <= 0;
  *o = (unsigned long long)((KEY_KP9) | 0);  /* check that KEY_KP9 is an integer */
  if (!_cffi_check_int(*o, n, 73U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KPASTERISK(unsigned long long *o)
{
  int n = (KEY_KPASTERISK) <= 0;
  *o = (unsigned long long)((KEY_KPASTERISK) | 0);  /* check that KEY_KPASTERISK is an integer */
  if (!_cffi_check_int(*o, n, 55U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KPCOMMA(unsigned long long *o)
{
  int n = (KEY_KPCOMMA) <= 0;
  *o = (unsigned long long)((KEY_KPCOMMA) | 0);  /* check that KEY_KPCOMMA is an integer */
  if (!_cffi_check_int(*o, n, 121U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KPDOT(unsigned long long *o)
{
  int n = (KEY_KPDOT) <= 0;
  *o = (unsigned long long)((KEY_KPDOT) | 0);  /* check that KEY_KPDOT is an integer */
  if (!_cffi_check_int(*o, n, 83U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KPENTER(unsigned long long *o)
{
  int n = (KEY_KPENTER) <= 0;
  *o = (unsigned long long)((KEY_KPENTER) | 0);  /* check that KEY_KPENTER is an integer */
  if (!_cffi_check_int(*o, n, 96U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KPEQUAL(unsigned long long *o)
{
  int n = (KEY_KPEQUAL) <= 0;
  *o = (unsigned long long)((KEY_KPEQUAL) | 0);  /* check that KEY_KPEQUAL is an integer */
  if (!_cffi_check_int(*o, n, 117U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KPJPCOMMA(unsigned long long *o)
{
  int n = (KEY_KPJPCOMMA) <= 0;
  *o = (unsigned long long)((KEY_KPJPCOMMA) | 0);  /* check that KEY_KPJPCOMMA is an integer */
  if (!_cffi_check_int(*o, n, 95U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KPLEFTPAREN(unsigned long long *o)
{
  int n = (KEY_KPLEFTPAREN) <= 0;
  *o = (unsigned long long)((KEY_KPLEFTPAREN) | 0);  /* check that KEY_KPLEFTPAREN is an integer */
  if (!_cffi_check_int(*o, n, 179U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KPMINUS(unsigned long long *o)
{
  int n = (KEY_KPMINUS) <= 0;
  *o = (unsigned long long)((KEY_KPMINUS) | 0);  /* check that KEY_KPMINUS is an integer */
  if (!_cffi_check_int(*o, n, 74U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KPPLUS(unsigned long long *o)
{
  int n = (KEY_KPPLUS) <= 0;
  *o = (unsigned long long)((KEY_KPPLUS) | 0);  /* check that KEY_KPPLUS is an integer */
  if (!_cffi_check_int(*o, n, 78U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KPPLUSMINUS(unsigned long long *o)
{
  int n = (KEY_KPPLUSMINUS) <= 0;
  *o = (unsigned long long)((KEY_KPPLUSMINUS) | 0);  /* check that KEY_KPPLUSMINUS is an integer */
  if (!_cffi_check_int(*o, n, 118U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KPRIGHTPAREN(unsigned long long *o)
{
  int n = (KEY_KPRIGHTPAREN) <= 0;
  *o = (unsigned long long)((KEY_KPRIGHTPAREN) | 0);  /* check that KEY_KPRIGHTPAREN is an integer */
  if (!_cffi_check_int(*o, n, 180U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_KPSLASH(unsigned long long *o)
{
  int n = (KEY_KPSLASH) <= 0;
  *o = (unsigned long long)((KEY_KPSLASH) | 0);  /* check that KEY_KPSLASH is an integer */
  if (!_cffi_check_int(*o, n, 98U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_L(unsigned long long *o)
{
  int n = (KEY_L) <= 0;
  *o = (unsigned long long)((KEY_L) | 0);  /* check that KEY_L is an integer */
  if (!_cffi_check_int(*o, n, 38U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LANGUAGE(unsigned long long *o)
{
  int n = (KEY_LANGUAGE) <= 0;
  *o = (unsigned long long)((KEY_LANGUAGE) | 0);  /* check that KEY_LANGUAGE is an integer */
  if (!_cffi_check_int(*o, n, 368U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LAST(unsigned long long *o)
{
  int n = (KEY_LAST) <= 0;
  *o = (unsigned long long)((KEY_LAST) | 0);  /* check that KEY_LAST is an integer */
  if (!_cffi_check_int(*o, n, 405U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LEFT(unsigned long long *o)
{
  int n = (KEY_LEFT) <= 0;
  *o = (unsigned long long)((KEY_LEFT) | 0);  /* check that KEY_LEFT is an integer */
  if (!_cffi_check_int(*o, n, 105U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LEFTALT(unsigned long long *o)
{
  int n = (KEY_LEFTALT) <= 0;
  *o = (unsigned long long)((KEY_LEFTALT) | 0);  /* check that KEY_LEFTALT is an integer */
  if (!_cffi_check_int(*o, n, 56U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LEFTBRACE(unsigned long long *o)
{
  int n = (KEY_LEFTBRACE) <= 0;
  *o = (unsigned long long)((KEY_LEFTBRACE) | 0);  /* check that KEY_LEFTBRACE is an integer */
  if (!_cffi_check_int(*o, n, 26U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LEFTCTRL(unsigned long long *o)
{
  int n = (KEY_LEFTCTRL) <= 0;
  *o = (unsigned long long)((KEY_LEFTCTRL) | 0);  /* check that KEY_LEFTCTRL is an integer */
  if (!_cffi_check_int(*o, n, 29U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LEFTMETA(unsigned long long *o)
{
  int n = (KEY_LEFTMETA) <= 0;
  *o = (unsigned long long)((KEY_LEFTMETA) | 0);  /* check that KEY_LEFTMETA is an integer */
  if (!_cffi_check_int(*o, n, 125U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LEFTSHIFT(unsigned long long *o)
{
  int n = (KEY_LEFTSHIFT) <= 0;
  *o = (unsigned long long)((KEY_LEFTSHIFT) | 0);  /* check that KEY_LEFTSHIFT is an integer */
  if (!_cffi_check_int(*o, n, 42U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LEFT_DOWN(unsigned long long *o)
{
  int n = (KEY_LEFT_DOWN) <= 0;
  *o = (unsigned long long)((KEY_LEFT_DOWN) | 0);  /* check that KEY_LEFT_DOWN is an integer */
  if (!_cffi_check_int(*o, n, 617U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LEFT_UP(unsigned long long *o)
{
  int n = (KEY_LEFT_UP) <= 0;
  *o = (unsigned long long)((KEY_LEFT_UP) | 0);  /* check that KEY_LEFT_UP is an integer */
  if (!_cffi_check_int(*o, n, 616U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LIGHTS_TOGGLE(unsigned long long *o)
{
  int n = (KEY_LIGHTS_TOGGLE) <= 0;
  *o = (unsigned long long)((KEY_LIGHTS_TOGGLE) | 0);  /* check that KEY_LIGHTS_TOGGLE is an integer */
  if (!_cffi_check_int(*o, n, 542U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LINEFEED(unsigned long long *o)
{
  int n = (KEY_LINEFEED) <= 0;
  *o = (unsigned long long)((KEY_LINEFEED) | 0);  /* check that KEY_LINEFEED is an integer */
  if (!_cffi_check_int(*o, n, 101U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LIST(unsigned long long *o)
{
  int n = (KEY_LIST) <= 0;
  *o = (unsigned long long)((KEY_LIST) | 0);  /* check that KEY_LIST is an integer */
  if (!_cffi_check_int(*o, n, 395U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_LOGOFF(unsigned long long *o)
{
  int n = (KEY_LOGOFF) <= 0;
  *o = (unsigned long long)((KEY_LOGOFF) | 0);  /* check that KEY_LOGOFF is an integer */
  if (!_cffi_check_int(*o, n, 433U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_M(unsigned long long *o)
{
  int n = (KEY_M) <= 0;
  *o = (unsigned long long)((KEY_M) | 0);  /* check that KEY_M is an integer */
  if (!_cffi_check_int(*o, n, 50U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MACRO(unsigned long long *o)
{
  int n = (KEY_MACRO) <= 0;
  *o = (unsigned long long)((KEY_MACRO) | 0);  /* check that KEY_MACRO is an integer */
  if (!_cffi_check_int(*o, n, 112U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MAIL(unsigned long long *o)
{
  int n = (KEY_MAIL) <= 0;
  *o = (unsigned long long)((KEY_MAIL) | 0);  /* check that KEY_MAIL is an integer */
  if (!_cffi_check_int(*o, n, 155U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MAX(unsigned long long *o)
{
  int n = (KEY_MAX) <= 0;
  *o = (unsigned long long)((KEY_MAX) | 0);  /* check that KEY_MAX is an integer */
  if (!_cffi_check_int(*o, n, 767U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MEDIA(unsigned long long *o)
{
  int n = (KEY_MEDIA) <= 0;
  *o = (unsigned long long)((KEY_MEDIA) | 0);  /* check that KEY_MEDIA is an integer */
  if (!_cffi_check_int(*o, n, 226U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MEDIA_REPEAT(unsigned long long *o)
{
  int n = (KEY_MEDIA_REPEAT) <= 0;
  *o = (unsigned long long)((KEY_MEDIA_REPEAT) | 0);  /* check that KEY_MEDIA_REPEAT is an integer */
  if (!_cffi_check_int(*o, n, 439U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MEDIA_TOP_MENU(unsigned long long *o)
{
  int n = (KEY_MEDIA_TOP_MENU) <= 0;
  *o = (unsigned long long)((KEY_MEDIA_TOP_MENU) | 0);  /* check that KEY_MEDIA_TOP_MENU is an integer */
  if (!_cffi_check_int(*o, n, 619U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MEMO(unsigned long long *o)
{
  int n = (KEY_MEMO) <= 0;
  *o = (unsigned long long)((KEY_MEMO) | 0);  /* check that KEY_MEMO is an integer */
  if (!_cffi_check_int(*o, n, 396U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MENU(unsigned long long *o)
{
  int n = (KEY_MENU) <= 0;
  *o = (unsigned long long)((KEY_MENU) | 0);  /* check that KEY_MENU is an integer */
  if (!_cffi_check_int(*o, n, 139U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MESSENGER(unsigned long long *o)
{
  int n = (KEY_MESSENGER) <= 0;
  *o = (unsigned long long)((KEY_MESSENGER) | 0);  /* check that KEY_MESSENGER is an integer */
  if (!_cffi_check_int(*o, n, 430U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MHP(unsigned long long *o)
{
  int n = (KEY_MHP) <= 0;
  *o = (unsigned long long)((KEY_MHP) | 0);  /* check that KEY_MHP is an integer */
  if (!_cffi_check_int(*o, n, 367U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MICMUTE(unsigned long long *o)
{
  int n = (KEY_MICMUTE) <= 0;
  *o = (unsigned long long)((KEY_MICMUTE) | 0);  /* check that KEY_MICMUTE is an integer */
  if (!_cffi_check_int(*o, n, 248U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MINUS(unsigned long long *o)
{
  int n = (KEY_MINUS) <= 0;
  *o = (unsigned long long)((KEY_MINUS) | 0);  /* check that KEY_MINUS is an integer */
  if (!_cffi_check_int(*o, n, 12U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MODE(unsigned long long *o)
{
  int n = (KEY_MODE) <= 0;
  *o = (unsigned long long)((KEY_MODE) | 0);  /* check that KEY_MODE is an integer */
  if (!_cffi_check_int(*o, n, 373U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MOVE(unsigned long long *o)
{
  int n = (KEY_MOVE) <= 0;
  *o = (unsigned long long)((KEY_MOVE) | 0);  /* check that KEY_MOVE is an integer */
  if (!_cffi_check_int(*o, n, 175U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MP3(unsigned long long *o)
{
  int n = (KEY_MP3) <= 0;
  *o = (unsigned long long)((KEY_MP3) | 0);  /* check that KEY_MP3 is an integer */
  if (!_cffi_check_int(*o, n, 391U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MSDOS(unsigned long long *o)
{
  int n = (KEY_MSDOS) <= 0;
  *o = (unsigned long long)((KEY_MSDOS) | 0);  /* check that KEY_MSDOS is an integer */
  if (!_cffi_check_int(*o, n, 151U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MUHENKAN(unsigned long long *o)
{
  int n = (KEY_MUHENKAN) <= 0;
  *o = (unsigned long long)((KEY_MUHENKAN) | 0);  /* check that KEY_MUHENKAN is an integer */
  if (!_cffi_check_int(*o, n, 94U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_MUTE(unsigned long long *o)
{
  int n = (KEY_MUTE) <= 0;
  *o = (unsigned long long)((KEY_MUTE) | 0);  /* check that KEY_MUTE is an integer */
  if (!_cffi_check_int(*o, n, 113U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_N(unsigned long long *o)
{
  int n = (KEY_N) <= 0;
  *o = (unsigned long long)((KEY_N) | 0);  /* check that KEY_N is an integer */
  if (!_cffi_check_int(*o, n, 49U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NEW(unsigned long long *o)
{
  int n = (KEY_NEW) <= 0;
  *o = (unsigned long long)((KEY_NEW) | 0);  /* check that KEY_NEW is an integer */
  if (!_cffi_check_int(*o, n, 181U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NEWS(unsigned long long *o)
{
  int n = (KEY_NEWS) <= 0;
  *o = (unsigned long long)((KEY_NEWS) | 0);  /* check that KEY_NEWS is an integer */
  if (!_cffi_check_int(*o, n, 427U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NEXT(unsigned long long *o)
{
  int n = (KEY_NEXT) <= 0;
  *o = (unsigned long long)((KEY_NEXT) | 0);  /* check that KEY_NEXT is an integer */
  if (!_cffi_check_int(*o, n, 407U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NEXTSONG(unsigned long long *o)
{
  int n = (KEY_NEXTSONG) <= 0;
  *o = (unsigned long long)((KEY_NEXTSONG) | 0);  /* check that KEY_NEXTSONG is an integer */
  if (!_cffi_check_int(*o, n, 163U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NEXT_FAVORITE(unsigned long long *o)
{
  int n = (KEY_NEXT_FAVORITE) <= 0;
  *o = (unsigned long long)((KEY_NEXT_FAVORITE) | 0);  /* check that KEY_NEXT_FAVORITE is an integer */
  if (!_cffi_check_int(*o, n, 624U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_0(unsigned long long *o)
{
  int n = (KEY_NUMERIC_0) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_0) | 0);  /* check that KEY_NUMERIC_0 is an integer */
  if (!_cffi_check_int(*o, n, 512U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_1(unsigned long long *o)
{
  int n = (KEY_NUMERIC_1) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_1) | 0);  /* check that KEY_NUMERIC_1 is an integer */
  if (!_cffi_check_int(*o, n, 513U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_11(unsigned long long *o)
{
  int n = (KEY_NUMERIC_11) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_11) | 0);  /* check that KEY_NUMERIC_11 is an integer */
  if (!_cffi_check_int(*o, n, 620U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_12(unsigned long long *o)
{
  int n = (KEY_NUMERIC_12) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_12) | 0);  /* check that KEY_NUMERIC_12 is an integer */
  if (!_cffi_check_int(*o, n, 621U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_2(unsigned long long *o)
{
  int n = (KEY_NUMERIC_2) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_2) | 0);  /* check that KEY_NUMERIC_2 is an integer */
  if (!_cffi_check_int(*o, n, 514U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_3(unsigned long long *o)
{
  int n = (KEY_NUMERIC_3) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_3) | 0);  /* check that KEY_NUMERIC_3 is an integer */
  if (!_cffi_check_int(*o, n, 515U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_4(unsigned long long *o)
{
  int n = (KEY_NUMERIC_4) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_4) | 0);  /* check that KEY_NUMERIC_4 is an integer */
  if (!_cffi_check_int(*o, n, 516U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_5(unsigned long long *o)
{
  int n = (KEY_NUMERIC_5) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_5) | 0);  /* check that KEY_NUMERIC_5 is an integer */
  if (!_cffi_check_int(*o, n, 517U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_6(unsigned long long *o)
{
  int n = (KEY_NUMERIC_6) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_6) | 0);  /* check that KEY_NUMERIC_6 is an integer */
  if (!_cffi_check_int(*o, n, 518U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_7(unsigned long long *o)
{
  int n = (KEY_NUMERIC_7) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_7) | 0);  /* check that KEY_NUMERIC_7 is an integer */
  if (!_cffi_check_int(*o, n, 519U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_8(unsigned long long *o)
{
  int n = (KEY_NUMERIC_8) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_8) | 0);  /* check that KEY_NUMERIC_8 is an integer */
  if (!_cffi_check_int(*o, n, 520U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_9(unsigned long long *o)
{
  int n = (KEY_NUMERIC_9) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_9) | 0);  /* check that KEY_NUMERIC_9 is an integer */
  if (!_cffi_check_int(*o, n, 521U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_A(unsigned long long *o)
{
  int n = (KEY_NUMERIC_A) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_A) | 0);  /* check that KEY_NUMERIC_A is an integer */
  if (!_cffi_check_int(*o, n, 524U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_B(unsigned long long *o)
{
  int n = (KEY_NUMERIC_B) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_B) | 0);  /* check that KEY_NUMERIC_B is an integer */
  if (!_cffi_check_int(*o, n, 525U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_C(unsigned long long *o)
{
  int n = (KEY_NUMERIC_C) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_C) | 0);  /* check that KEY_NUMERIC_C is an integer */
  if (!_cffi_check_int(*o, n, 526U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_D(unsigned long long *o)
{
  int n = (KEY_NUMERIC_D) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_D) | 0);  /* check that KEY_NUMERIC_D is an integer */
  if (!_cffi_check_int(*o, n, 527U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_POUND(unsigned long long *o)
{
  int n = (KEY_NUMERIC_POUND) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_POUND) | 0);  /* check that KEY_NUMERIC_POUND is an integer */
  if (!_cffi_check_int(*o, n, 523U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMERIC_STAR(unsigned long long *o)
{
  int n = (KEY_NUMERIC_STAR) <= 0;
  *o = (unsigned long long)((KEY_NUMERIC_STAR) | 0);  /* check that KEY_NUMERIC_STAR is an integer */
  if (!_cffi_check_int(*o, n, 522U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_NUMLOCK(unsigned long long *o)
{
  int n = (KEY_NUMLOCK) <= 0;
  *o = (unsigned long long)((KEY_NUMLOCK) | 0);  /* check that KEY_NUMLOCK is an integer */
  if (!_cffi_check_int(*o, n, 69U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_O(unsigned long long *o)
{
  int n = (KEY_O) <= 0;
  *o = (unsigned long long)((KEY_O) | 0);  /* check that KEY_O is an integer */
  if (!_cffi_check_int(*o, n, 24U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_OK(unsigned long long *o)
{
  int n = (KEY_OK) <= 0;
  *o = (unsigned long long)((KEY_OK) | 0);  /* check that KEY_OK is an integer */
  if (!_cffi_check_int(*o, n, 352U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_OPEN(unsigned long long *o)
{
  int n = (KEY_OPEN) <= 0;
  *o = (unsigned long long)((KEY_OPEN) | 0);  /* check that KEY_OPEN is an integer */
  if (!_cffi_check_int(*o, n, 134U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_OPTION(unsigned long long *o)
{
  int n = (KEY_OPTION) <= 0;
  *o = (unsigned long long)((KEY_OPTION) | 0);  /* check that KEY_OPTION is an integer */
  if (!_cffi_check_int(*o, n, 357U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_P(unsigned long long *o)
{
  int n = (KEY_P) <= 0;
  *o = (unsigned long long)((KEY_P) | 0);  /* check that KEY_P is an integer */
  if (!_cffi_check_int(*o, n, 25U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PAGEDOWN(unsigned long long *o)
{
  int n = (KEY_PAGEDOWN) <= 0;
  *o = (unsigned long long)((KEY_PAGEDOWN) | 0);  /* check that KEY_PAGEDOWN is an integer */
  if (!_cffi_check_int(*o, n, 109U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PAGEUP(unsigned long long *o)
{
  int n = (KEY_PAGEUP) <= 0;
  *o = (unsigned long long)((KEY_PAGEUP) | 0);  /* check that KEY_PAGEUP is an integer */
  if (!_cffi_check_int(*o, n, 104U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PASTE(unsigned long long *o)
{
  int n = (KEY_PASTE) <= 0;
  *o = (unsigned long long)((KEY_PASTE) | 0);  /* check that KEY_PASTE is an integer */
  if (!_cffi_check_int(*o, n, 135U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PAUSE(unsigned long long *o)
{
  int n = (KEY_PAUSE) <= 0;
  *o = (unsigned long long)((KEY_PAUSE) | 0);  /* check that KEY_PAUSE is an integer */
  if (!_cffi_check_int(*o, n, 119U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PAUSECD(unsigned long long *o)
{
  int n = (KEY_PAUSECD) <= 0;
  *o = (unsigned long long)((KEY_PAUSECD) | 0);  /* check that KEY_PAUSECD is an integer */
  if (!_cffi_check_int(*o, n, 201U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PAUSE_RECORD(unsigned long long *o)
{
  int n = (KEY_PAUSE_RECORD) <= 0;
  *o = (unsigned long long)((KEY_PAUSE_RECORD) | 0);  /* check that KEY_PAUSE_RECORD is an integer */
  if (!_cffi_check_int(*o, n, 626U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PC(unsigned long long *o)
{
  int n = (KEY_PC) <= 0;
  *o = (unsigned long long)((KEY_PC) | 0);  /* check that KEY_PC is an integer */
  if (!_cffi_check_int(*o, n, 376U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PHONE(unsigned long long *o)
{
  int n = (KEY_PHONE) <= 0;
  *o = (unsigned long long)((KEY_PHONE) | 0);  /* check that KEY_PHONE is an integer */
  if (!_cffi_check_int(*o, n, 169U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PLAY(unsigned long long *o)
{
  int n = (KEY_PLAY) <= 0;
  *o = (unsigned long long)((KEY_PLAY) | 0);  /* check that KEY_PLAY is an integer */
  if (!_cffi_check_int(*o, n, 207U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PLAYCD(unsigned long long *o)
{
  int n = (KEY_PLAYCD) <= 0;
  *o = (unsigned long long)((KEY_PLAYCD) | 0);  /* check that KEY_PLAYCD is an integer */
  if (!_cffi_check_int(*o, n, 200U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PLAYER(unsigned long long *o)
{
  int n = (KEY_PLAYER) <= 0;
  *o = (unsigned long long)((KEY_PLAYER) | 0);  /* check that KEY_PLAYER is an integer */
  if (!_cffi_check_int(*o, n, 387U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PLAYPAUSE(unsigned long long *o)
{
  int n = (KEY_PLAYPAUSE) <= 0;
  *o = (unsigned long long)((KEY_PLAYPAUSE) | 0);  /* check that KEY_PLAYPAUSE is an integer */
  if (!_cffi_check_int(*o, n, 164U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_POWER(unsigned long long *o)
{
  int n = (KEY_POWER) <= 0;
  *o = (unsigned long long)((KEY_POWER) | 0);  /* check that KEY_POWER is an integer */
  if (!_cffi_check_int(*o, n, 116U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_POWER2(unsigned long long *o)
{
  int n = (KEY_POWER2) <= 0;
  *o = (unsigned long long)((KEY_POWER2) | 0);  /* check that KEY_POWER2 is an integer */
  if (!_cffi_check_int(*o, n, 356U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PRESENTATION(unsigned long long *o)
{
  int n = (KEY_PRESENTATION) <= 0;
  *o = (unsigned long long)((KEY_PRESENTATION) | 0);  /* check that KEY_PRESENTATION is an integer */
  if (!_cffi_check_int(*o, n, 425U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PREVIOUS(unsigned long long *o)
{
  int n = (KEY_PREVIOUS) <= 0;
  *o = (unsigned long long)((KEY_PREVIOUS) | 0);  /* check that KEY_PREVIOUS is an integer */
  if (!_cffi_check_int(*o, n, 412U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PREVIOUSSONG(unsigned long long *o)
{
  int n = (KEY_PREVIOUSSONG) <= 0;
  *o = (unsigned long long)((KEY_PREVIOUSSONG) | 0);  /* check that KEY_PREVIOUSSONG is an integer */
  if (!_cffi_check_int(*o, n, 165U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PRINT(unsigned long long *o)
{
  int n = (KEY_PRINT) <= 0;
  *o = (unsigned long long)((KEY_PRINT) | 0);  /* check that KEY_PRINT is an integer */
  if (!_cffi_check_int(*o, n, 210U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PROG1(unsigned long long *o)
{
  int n = (KEY_PROG1) <= 0;
  *o = (unsigned long long)((KEY_PROG1) | 0);  /* check that KEY_PROG1 is an integer */
  if (!_cffi_check_int(*o, n, 148U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PROG2(unsigned long long *o)
{
  int n = (KEY_PROG2) <= 0;
  *o = (unsigned long long)((KEY_PROG2) | 0);  /* check that KEY_PROG2 is an integer */
  if (!_cffi_check_int(*o, n, 149U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PROG3(unsigned long long *o)
{
  int n = (KEY_PROG3) <= 0;
  *o = (unsigned long long)((KEY_PROG3) | 0);  /* check that KEY_PROG3 is an integer */
  if (!_cffi_check_int(*o, n, 202U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PROG4(unsigned long long *o)
{
  int n = (KEY_PROG4) <= 0;
  *o = (unsigned long long)((KEY_PROG4) | 0);  /* check that KEY_PROG4 is an integer */
  if (!_cffi_check_int(*o, n, 203U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PROGRAM(unsigned long long *o)
{
  int n = (KEY_PROGRAM) <= 0;
  *o = (unsigned long long)((KEY_PROGRAM) | 0);  /* check that KEY_PROGRAM is an integer */
  if (!_cffi_check_int(*o, n, 362U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PROPS(unsigned long long *o)
{
  int n = (KEY_PROPS) <= 0;
  *o = (unsigned long long)((KEY_PROPS) | 0);  /* check that KEY_PROPS is an integer */
  if (!_cffi_check_int(*o, n, 130U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_PVR(unsigned long long *o)
{
  int n = (KEY_PVR) <= 0;
  *o = (unsigned long long)((KEY_PVR) | 0);  /* check that KEY_PVR is an integer */
  if (!_cffi_check_int(*o, n, 366U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_Q(unsigned long long *o)
{
  int n = (KEY_Q) <= 0;
  *o = (unsigned long long)((KEY_Q) | 0);  /* check that KEY_Q is an integer */
  if (!_cffi_check_int(*o, n, 16U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_QUESTION(unsigned long long *o)
{
  int n = (KEY_QUESTION) <= 0;
  *o = (unsigned long long)((KEY_QUESTION) | 0);  /* check that KEY_QUESTION is an integer */
  if (!_cffi_check_int(*o, n, 214U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_R(unsigned long long *o)
{
  int n = (KEY_R) <= 0;
  *o = (unsigned long long)((KEY_R) | 0);  /* check that KEY_R is an integer */
  if (!_cffi_check_int(*o, n, 19U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RADIO(unsigned long long *o)
{
  int n = (KEY_RADIO) <= 0;
  *o = (unsigned long long)((KEY_RADIO) | 0);  /* check that KEY_RADIO is an integer */
  if (!_cffi_check_int(*o, n, 385U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RECORD(unsigned long long *o)
{
  int n = (KEY_RECORD) <= 0;
  *o = (unsigned long long)((KEY_RECORD) | 0);  /* check that KEY_RECORD is an integer */
  if (!_cffi_check_int(*o, n, 167U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RED(unsigned long long *o)
{
  int n = (KEY_RED) <= 0;
  *o = (unsigned long long)((KEY_RED) | 0);  /* check that KEY_RED is an integer */
  if (!_cffi_check_int(*o, n, 398U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_REDO(unsigned long long *o)
{
  int n = (KEY_REDO) <= 0;
  *o = (unsigned long long)((KEY_REDO) | 0);  /* check that KEY_REDO is an integer */
  if (!_cffi_check_int(*o, n, 182U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_REFRESH(unsigned long long *o)
{
  int n = (KEY_REFRESH) <= 0;
  *o = (unsigned long long)((KEY_REFRESH) | 0);  /* check that KEY_REFRESH is an integer */
  if (!_cffi_check_int(*o, n, 173U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_REPLY(unsigned long long *o)
{
  int n = (KEY_REPLY) <= 0;
  *o = (unsigned long long)((KEY_REPLY) | 0);  /* check that KEY_REPLY is an integer */
  if (!_cffi_check_int(*o, n, 232U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RESERVED(unsigned long long *o)
{
  int n = (KEY_RESERVED) <= 0;
  *o = (unsigned long long)((KEY_RESERVED) | 0);  /* check that KEY_RESERVED is an integer */
  if (!_cffi_check_int(*o, n, 0))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RESTART(unsigned long long *o)
{
  int n = (KEY_RESTART) <= 0;
  *o = (unsigned long long)((KEY_RESTART) | 0);  /* check that KEY_RESTART is an integer */
  if (!_cffi_check_int(*o, n, 408U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_REWIND(unsigned long long *o)
{
  int n = (KEY_REWIND) <= 0;
  *o = (unsigned long long)((KEY_REWIND) | 0);  /* check that KEY_REWIND is an integer */
  if (!_cffi_check_int(*o, n, 168U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RFKILL(unsigned long long *o)
{
  int n = (KEY_RFKILL) <= 0;
  *o = (unsigned long long)((KEY_RFKILL) | 0);  /* check that KEY_RFKILL is an integer */
  if (!_cffi_check_int(*o, n, 247U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RIGHT(unsigned long long *o)
{
  int n = (KEY_RIGHT) <= 0;
  *o = (unsigned long long)((KEY_RIGHT) | 0);  /* check that KEY_RIGHT is an integer */
  if (!_cffi_check_int(*o, n, 106U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RIGHTALT(unsigned long long *o)
{
  int n = (KEY_RIGHTALT) <= 0;
  *o = (unsigned long long)((KEY_RIGHTALT) | 0);  /* check that KEY_RIGHTALT is an integer */
  if (!_cffi_check_int(*o, n, 100U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RIGHTBRACE(unsigned long long *o)
{
  int n = (KEY_RIGHTBRACE) <= 0;
  *o = (unsigned long long)((KEY_RIGHTBRACE) | 0);  /* check that KEY_RIGHTBRACE is an integer */
  if (!_cffi_check_int(*o, n, 27U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RIGHTCTRL(unsigned long long *o)
{
  int n = (KEY_RIGHTCTRL) <= 0;
  *o = (unsigned long long)((KEY_RIGHTCTRL) | 0);  /* check that KEY_RIGHTCTRL is an integer */
  if (!_cffi_check_int(*o, n, 97U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RIGHTMETA(unsigned long long *o)
{
  int n = (KEY_RIGHTMETA) <= 0;
  *o = (unsigned long long)((KEY_RIGHTMETA) | 0);  /* check that KEY_RIGHTMETA is an integer */
  if (!_cffi_check_int(*o, n, 126U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RIGHTSHIFT(unsigned long long *o)
{
  int n = (KEY_RIGHTSHIFT) <= 0;
  *o = (unsigned long long)((KEY_RIGHTSHIFT) | 0);  /* check that KEY_RIGHTSHIFT is an integer */
  if (!_cffi_check_int(*o, n, 54U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RIGHT_DOWN(unsigned long long *o)
{
  int n = (KEY_RIGHT_DOWN) <= 0;
  *o = (unsigned long long)((KEY_RIGHT_DOWN) | 0);  /* check that KEY_RIGHT_DOWN is an integer */
  if (!_cffi_check_int(*o, n, 615U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RIGHT_UP(unsigned long long *o)
{
  int n = (KEY_RIGHT_UP) <= 0;
  *o = (unsigned long long)((KEY_RIGHT_UP) | 0);  /* check that KEY_RIGHT_UP is an integer */
  if (!_cffi_check_int(*o, n, 614U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_RO(unsigned long long *o)
{
  int n = (KEY_RO) <= 0;
  *o = (unsigned long long)((KEY_RO) | 0);  /* check that KEY_RO is an integer */
  if (!_cffi_check_int(*o, n, 89U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ROOT_MENU(unsigned long long *o)
{
  int n = (KEY_ROOT_MENU) <= 0;
  *o = (unsigned long long)((KEY_ROOT_MENU) | 0);  /* check that KEY_ROOT_MENU is an integer */
  if (!_cffi_check_int(*o, n, 618U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ROTATE_DISPLAY(unsigned long long *o)
{
  int n = (KEY_ROTATE_DISPLAY) <= 0;
  *o = (unsigned long long)((KEY_ROTATE_DISPLAY) | 0);  /* check that KEY_ROTATE_DISPLAY is an integer */
  if (!_cffi_check_int(*o, n, 153U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_S(unsigned long long *o)
{
  int n = (KEY_S) <= 0;
  *o = (unsigned long long)((KEY_S) | 0);  /* check that KEY_S is an integer */
  if (!_cffi_check_int(*o, n, 31U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SAT(unsigned long long *o)
{
  int n = (KEY_SAT) <= 0;
  *o = (unsigned long long)((KEY_SAT) | 0);  /* check that KEY_SAT is an integer */
  if (!_cffi_check_int(*o, n, 381U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SAT2(unsigned long long *o)
{
  int n = (KEY_SAT2) <= 0;
  *o = (unsigned long long)((KEY_SAT2) | 0);  /* check that KEY_SAT2 is an integer */
  if (!_cffi_check_int(*o, n, 382U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SAVE(unsigned long long *o)
{
  int n = (KEY_SAVE) <= 0;
  *o = (unsigned long long)((KEY_SAVE) | 0);  /* check that KEY_SAVE is an integer */
  if (!_cffi_check_int(*o, n, 234U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SCALE(unsigned long long *o)
{
  int n = (KEY_SCALE) <= 0;
  *o = (unsigned long long)((KEY_SCALE) | 0);  /* check that KEY_SCALE is an integer */
  if (!_cffi_check_int(*o, n, 120U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SCREEN(unsigned long long *o)
{
  int n = (KEY_SCREEN) <= 0;
  *o = (unsigned long long)((KEY_SCREEN) | 0);  /* check that KEY_SCREEN is an integer */
  if (!_cffi_check_int(*o, n, 375U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SCREENSAVER(unsigned long long *o)
{
  int n = (KEY_SCREENSAVER) <= 0;
  *o = (unsigned long long)((KEY_SCREENSAVER) | 0);  /* check that KEY_SCREENSAVER is an integer */
  if (!_cffi_check_int(*o, n, 581U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SCROLLDOWN(unsigned long long *o)
{
  int n = (KEY_SCROLLDOWN) <= 0;
  *o = (unsigned long long)((KEY_SCROLLDOWN) | 0);  /* check that KEY_SCROLLDOWN is an integer */
  if (!_cffi_check_int(*o, n, 178U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SCROLLLOCK(unsigned long long *o)
{
  int n = (KEY_SCROLLLOCK) <= 0;
  *o = (unsigned long long)((KEY_SCROLLLOCK) | 0);  /* check that KEY_SCROLLLOCK is an integer */
  if (!_cffi_check_int(*o, n, 70U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SCROLLUP(unsigned long long *o)
{
  int n = (KEY_SCROLLUP) <= 0;
  *o = (unsigned long long)((KEY_SCROLLUP) | 0);  /* check that KEY_SCROLLUP is an integer */
  if (!_cffi_check_int(*o, n, 177U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SEARCH(unsigned long long *o)
{
  int n = (KEY_SEARCH) <= 0;
  *o = (unsigned long long)((KEY_SEARCH) | 0);  /* check that KEY_SEARCH is an integer */
  if (!_cffi_check_int(*o, n, 217U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SELECT(unsigned long long *o)
{
  int n = (KEY_SELECT) <= 0;
  *o = (unsigned long long)((KEY_SELECT) | 0);  /* check that KEY_SELECT is an integer */
  if (!_cffi_check_int(*o, n, 353U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SEMICOLON(unsigned long long *o)
{
  int n = (KEY_SEMICOLON) <= 0;
  *o = (unsigned long long)((KEY_SEMICOLON) | 0);  /* check that KEY_SEMICOLON is an integer */
  if (!_cffi_check_int(*o, n, 39U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SEND(unsigned long long *o)
{
  int n = (KEY_SEND) <= 0;
  *o = (unsigned long long)((KEY_SEND) | 0);  /* check that KEY_SEND is an integer */
  if (!_cffi_check_int(*o, n, 231U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SENDFILE(unsigned long long *o)
{
  int n = (KEY_SENDFILE) <= 0;
  *o = (unsigned long long)((KEY_SENDFILE) | 0);  /* check that KEY_SENDFILE is an integer */
  if (!_cffi_check_int(*o, n, 145U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SETUP(unsigned long long *o)
{
  int n = (KEY_SETUP) <= 0;
  *o = (unsigned long long)((KEY_SETUP) | 0);  /* check that KEY_SETUP is an integer */
  if (!_cffi_check_int(*o, n, 141U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SHOP(unsigned long long *o)
{
  int n = (KEY_SHOP) <= 0;
  *o = (unsigned long long)((KEY_SHOP) | 0);  /* check that KEY_SHOP is an integer */
  if (!_cffi_check_int(*o, n, 221U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SHUFFLE(unsigned long long *o)
{
  int n = (KEY_SHUFFLE) <= 0;
  *o = (unsigned long long)((KEY_SHUFFLE) | 0);  /* check that KEY_SHUFFLE is an integer */
  if (!_cffi_check_int(*o, n, 410U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SLASH(unsigned long long *o)
{
  int n = (KEY_SLASH) <= 0;
  *o = (unsigned long long)((KEY_SLASH) | 0);  /* check that KEY_SLASH is an integer */
  if (!_cffi_check_int(*o, n, 53U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SLEEP(unsigned long long *o)
{
  int n = (KEY_SLEEP) <= 0;
  *o = (unsigned long long)((KEY_SLEEP) | 0);  /* check that KEY_SLEEP is an integer */
  if (!_cffi_check_int(*o, n, 142U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SLOW(unsigned long long *o)
{
  int n = (KEY_SLOW) <= 0;
  *o = (unsigned long long)((KEY_SLOW) | 0);  /* check that KEY_SLOW is an integer */
  if (!_cffi_check_int(*o, n, 409U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SLOWREVERSE(unsigned long long *o)
{
  int n = (KEY_SLOWREVERSE) <= 0;
  *o = (unsigned long long)((KEY_SLOWREVERSE) | 0);  /* check that KEY_SLOWREVERSE is an integer */
  if (!_cffi_check_int(*o, n, 630U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SOUND(unsigned long long *o)
{
  int n = (KEY_SOUND) <= 0;
  *o = (unsigned long long)((KEY_SOUND) | 0);  /* check that KEY_SOUND is an integer */
  if (!_cffi_check_int(*o, n, 213U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SPACE(unsigned long long *o)
{
  int n = (KEY_SPACE) <= 0;
  *o = (unsigned long long)((KEY_SPACE) | 0);  /* check that KEY_SPACE is an integer */
  if (!_cffi_check_int(*o, n, 57U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SPELLCHECK(unsigned long long *o)
{
  int n = (KEY_SPELLCHECK) <= 0;
  *o = (unsigned long long)((KEY_SPELLCHECK) | 0);  /* check that KEY_SPELLCHECK is an integer */
  if (!_cffi_check_int(*o, n, 432U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SPORT(unsigned long long *o)
{
  int n = (KEY_SPORT) <= 0;
  *o = (unsigned long long)((KEY_SPORT) | 0);  /* check that KEY_SPORT is an integer */
  if (!_cffi_check_int(*o, n, 220U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SPREADSHEET(unsigned long long *o)
{
  int n = (KEY_SPREADSHEET) <= 0;
  *o = (unsigned long long)((KEY_SPREADSHEET) | 0);  /* check that KEY_SPREADSHEET is an integer */
  if (!_cffi_check_int(*o, n, 423U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_STOP(unsigned long long *o)
{
  int n = (KEY_STOP) <= 0;
  *o = (unsigned long long)((KEY_STOP) | 0);  /* check that KEY_STOP is an integer */
  if (!_cffi_check_int(*o, n, 128U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_STOPCD(unsigned long long *o)
{
  int n = (KEY_STOPCD) <= 0;
  *o = (unsigned long long)((KEY_STOPCD) | 0);  /* check that KEY_STOPCD is an integer */
  if (!_cffi_check_int(*o, n, 166U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_STOP_RECORD(unsigned long long *o)
{
  int n = (KEY_STOP_RECORD) <= 0;
  *o = (unsigned long long)((KEY_STOP_RECORD) | 0);  /* check that KEY_STOP_RECORD is an integer */
  if (!_cffi_check_int(*o, n, 625U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SUBTITLE(unsigned long long *o)
{
  int n = (KEY_SUBTITLE) <= 0;
  *o = (unsigned long long)((KEY_SUBTITLE) | 0);  /* check that KEY_SUBTITLE is an integer */
  if (!_cffi_check_int(*o, n, 370U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SUSPEND(unsigned long long *o)
{
  int n = (KEY_SUSPEND) <= 0;
  *o = (unsigned long long)((KEY_SUSPEND) | 0);  /* check that KEY_SUSPEND is an integer */
  if (!_cffi_check_int(*o, n, 205U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SWITCHVIDEOMODE(unsigned long long *o)
{
  int n = (KEY_SWITCHVIDEOMODE) <= 0;
  *o = (unsigned long long)((KEY_SWITCHVIDEOMODE) | 0);  /* check that KEY_SWITCHVIDEOMODE is an integer */
  if (!_cffi_check_int(*o, n, 227U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_SYSRQ(unsigned long long *o)
{
  int n = (KEY_SYSRQ) <= 0;
  *o = (unsigned long long)((KEY_SYSRQ) | 0);  /* check that KEY_SYSRQ is an integer */
  if (!_cffi_check_int(*o, n, 99U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_T(unsigned long long *o)
{
  int n = (KEY_T) <= 0;
  *o = (unsigned long long)((KEY_T) | 0);  /* check that KEY_T is an integer */
  if (!_cffi_check_int(*o, n, 20U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TAB(unsigned long long *o)
{
  int n = (KEY_TAB) <= 0;
  *o = (unsigned long long)((KEY_TAB) | 0);  /* check that KEY_TAB is an integer */
  if (!_cffi_check_int(*o, n, 15U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TAPE(unsigned long long *o)
{
  int n = (KEY_TAPE) <= 0;
  *o = (unsigned long long)((KEY_TAPE) | 0);  /* check that KEY_TAPE is an integer */
  if (!_cffi_check_int(*o, n, 384U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TASKMANAGER(unsigned long long *o)
{
  int n = (KEY_TASKMANAGER) <= 0;
  *o = (unsigned long long)((KEY_TASKMANAGER) | 0);  /* check that KEY_TASKMANAGER is an integer */
  if (!_cffi_check_int(*o, n, 577U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TEEN(unsigned long long *o)
{
  int n = (KEY_TEEN) <= 0;
  *o = (unsigned long long)((KEY_TEEN) | 0);  /* check that KEY_TEEN is an integer */
  if (!_cffi_check_int(*o, n, 414U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TEXT(unsigned long long *o)
{
  int n = (KEY_TEXT) <= 0;
  *o = (unsigned long long)((KEY_TEXT) | 0);  /* check that KEY_TEXT is an integer */
  if (!_cffi_check_int(*o, n, 388U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TIME(unsigned long long *o)
{
  int n = (KEY_TIME) <= 0;
  *o = (unsigned long long)((KEY_TIME) | 0);  /* check that KEY_TIME is an integer */
  if (!_cffi_check_int(*o, n, 359U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TITLE(unsigned long long *o)
{
  int n = (KEY_TITLE) <= 0;
  *o = (unsigned long long)((KEY_TITLE) | 0);  /* check that KEY_TITLE is an integer */
  if (!_cffi_check_int(*o, n, 369U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TOUCHPAD_OFF(unsigned long long *o)
{
  int n = (KEY_TOUCHPAD_OFF) <= 0;
  *o = (unsigned long long)((KEY_TOUCHPAD_OFF) | 0);  /* check that KEY_TOUCHPAD_OFF is an integer */
  if (!_cffi_check_int(*o, n, 532U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TOUCHPAD_ON(unsigned long long *o)
{
  int n = (KEY_TOUCHPAD_ON) <= 0;
  *o = (unsigned long long)((KEY_TOUCHPAD_ON) | 0);  /* check that KEY_TOUCHPAD_ON is an integer */
  if (!_cffi_check_int(*o, n, 531U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TOUCHPAD_TOGGLE(unsigned long long *o)
{
  int n = (KEY_TOUCHPAD_TOGGLE) <= 0;
  *o = (unsigned long long)((KEY_TOUCHPAD_TOGGLE) | 0);  /* check that KEY_TOUCHPAD_TOGGLE is an integer */
  if (!_cffi_check_int(*o, n, 530U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TUNER(unsigned long long *o)
{
  int n = (KEY_TUNER) <= 0;
  *o = (unsigned long long)((KEY_TUNER) | 0);  /* check that KEY_TUNER is an integer */
  if (!_cffi_check_int(*o, n, 386U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TV(unsigned long long *o)
{
  int n = (KEY_TV) <= 0;
  *o = (unsigned long long)((KEY_TV) | 0);  /* check that KEY_TV is an integer */
  if (!_cffi_check_int(*o, n, 377U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TV2(unsigned long long *o)
{
  int n = (KEY_TV2) <= 0;
  *o = (unsigned long long)((KEY_TV2) | 0);  /* check that KEY_TV2 is an integer */
  if (!_cffi_check_int(*o, n, 378U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_TWEN(unsigned long long *o)
{
  int n = (KEY_TWEN) <= 0;
  *o = (unsigned long long)((KEY_TWEN) | 0);  /* check that KEY_TWEN is an integer */
  if (!_cffi_check_int(*o, n, 415U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_U(unsigned long long *o)
{
  int n = (KEY_U) <= 0;
  *o = (unsigned long long)((KEY_U) | 0);  /* check that KEY_U is an integer */
  if (!_cffi_check_int(*o, n, 22U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_UNDO(unsigned long long *o)
{
  int n = (KEY_UNDO) <= 0;
  *o = (unsigned long long)((KEY_UNDO) | 0);  /* check that KEY_UNDO is an integer */
  if (!_cffi_check_int(*o, n, 131U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_UNKNOWN(unsigned long long *o)
{
  int n = (KEY_UNKNOWN) <= 0;
  *o = (unsigned long long)((KEY_UNKNOWN) | 0);  /* check that KEY_UNKNOWN is an integer */
  if (!_cffi_check_int(*o, n, 240U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_UNMUTE(unsigned long long *o)
{
  int n = (KEY_UNMUTE) <= 0;
  *o = (unsigned long long)((KEY_UNMUTE) | 0);  /* check that KEY_UNMUTE is an integer */
  if (!_cffi_check_int(*o, n, 628U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_UP(unsigned long long *o)
{
  int n = (KEY_UP) <= 0;
  *o = (unsigned long long)((KEY_UP) | 0);  /* check that KEY_UP is an integer */
  if (!_cffi_check_int(*o, n, 103U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_UWB(unsigned long long *o)
{
  int n = (KEY_UWB) <= 0;
  *o = (unsigned long long)((KEY_UWB) | 0);  /* check that KEY_UWB is an integer */
  if (!_cffi_check_int(*o, n, 239U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_V(unsigned long long *o)
{
  int n = (KEY_V) <= 0;
  *o = (unsigned long long)((KEY_V) | 0);  /* check that KEY_V is an integer */
  if (!_cffi_check_int(*o, n, 47U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_VCR(unsigned long long *o)
{
  int n = (KEY_VCR) <= 0;
  *o = (unsigned long long)((KEY_VCR) | 0);  /* check that KEY_VCR is an integer */
  if (!_cffi_check_int(*o, n, 379U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_VCR2(unsigned long long *o)
{
  int n = (KEY_VCR2) <= 0;
  *o = (unsigned long long)((KEY_VCR2) | 0);  /* check that KEY_VCR2 is an integer */
  if (!_cffi_check_int(*o, n, 380U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_VENDOR(unsigned long long *o)
{
  int n = (KEY_VENDOR) <= 0;
  *o = (unsigned long long)((KEY_VENDOR) | 0);  /* check that KEY_VENDOR is an integer */
  if (!_cffi_check_int(*o, n, 360U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_VIDEO(unsigned long long *o)
{
  int n = (KEY_VIDEO) <= 0;
  *o = (unsigned long long)((KEY_VIDEO) | 0);  /* check that KEY_VIDEO is an integer */
  if (!_cffi_check_int(*o, n, 393U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_VIDEOPHONE(unsigned long long *o)
{
  int n = (KEY_VIDEOPHONE) <= 0;
  *o = (unsigned long long)((KEY_VIDEOPHONE) | 0);  /* check that KEY_VIDEOPHONE is an integer */
  if (!_cffi_check_int(*o, n, 416U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_VIDEO_NEXT(unsigned long long *o)
{
  int n = (KEY_VIDEO_NEXT) <= 0;
  *o = (unsigned long long)((KEY_VIDEO_NEXT) | 0);  /* check that KEY_VIDEO_NEXT is an integer */
  if (!_cffi_check_int(*o, n, 241U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_VIDEO_PREV(unsigned long long *o)
{
  int n = (KEY_VIDEO_PREV) <= 0;
  *o = (unsigned long long)((KEY_VIDEO_PREV) | 0);  /* check that KEY_VIDEO_PREV is an integer */
  if (!_cffi_check_int(*o, n, 242U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_VOD(unsigned long long *o)
{
  int n = (KEY_VOD) <= 0;
  *o = (unsigned long long)((KEY_VOD) | 0);  /* check that KEY_VOD is an integer */
  if (!_cffi_check_int(*o, n, 627U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_VOICECOMMAND(unsigned long long *o)
{
  int n = (KEY_VOICECOMMAND) <= 0;
  *o = (unsigned long long)((KEY_VOICECOMMAND) | 0);  /* check that KEY_VOICECOMMAND is an integer */
  if (!_cffi_check_int(*o, n, 582U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_VOICEMAIL(unsigned long long *o)
{
  int n = (KEY_VOICEMAIL) <= 0;
  *o = (unsigned long long)((KEY_VOICEMAIL) | 0);  /* check that KEY_VOICEMAIL is an integer */
  if (!_cffi_check_int(*o, n, 428U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_VOLUMEDOWN(unsigned long long *o)
{
  int n = (KEY_VOLUMEDOWN) <= 0;
  *o = (unsigned long long)((KEY_VOLUMEDOWN) | 0);  /* check that KEY_VOLUMEDOWN is an integer */
  if (!_cffi_check_int(*o, n, 114U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_VOLUMEUP(unsigned long long *o)
{
  int n = (KEY_VOLUMEUP) <= 0;
  *o = (unsigned long long)((KEY_VOLUMEUP) | 0);  /* check that KEY_VOLUMEUP is an integer */
  if (!_cffi_check_int(*o, n, 115U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_W(unsigned long long *o)
{
  int n = (KEY_W) <= 0;
  *o = (unsigned long long)((KEY_W) | 0);  /* check that KEY_W is an integer */
  if (!_cffi_check_int(*o, n, 17U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_WAKEUP(unsigned long long *o)
{
  int n = (KEY_WAKEUP) <= 0;
  *o = (unsigned long long)((KEY_WAKEUP) | 0);  /* check that KEY_WAKEUP is an integer */
  if (!_cffi_check_int(*o, n, 143U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_WLAN(unsigned long long *o)
{
  int n = (KEY_WLAN) <= 0;
  *o = (unsigned long long)((KEY_WLAN) | 0);  /* check that KEY_WLAN is an integer */
  if (!_cffi_check_int(*o, n, 238U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_WORDPROCESSOR(unsigned long long *o)
{
  int n = (KEY_WORDPROCESSOR) <= 0;
  *o = (unsigned long long)((KEY_WORDPROCESSOR) | 0);  /* check that KEY_WORDPROCESSOR is an integer */
  if (!_cffi_check_int(*o, n, 421U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_WPS_BUTTON(unsigned long long *o)
{
  int n = (KEY_WPS_BUTTON) <= 0;
  *o = (unsigned long long)((KEY_WPS_BUTTON) | 0);  /* check that KEY_WPS_BUTTON is an integer */
  if (!_cffi_check_int(*o, n, 529U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_WWAN(unsigned long long *o)
{
  int n = (KEY_WWAN) <= 0;
  *o = (unsigned long long)((KEY_WWAN) | 0);  /* check that KEY_WWAN is an integer */
  if (!_cffi_check_int(*o, n, 246U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_WWW(unsigned long long *o)
{
  int n = (KEY_WWW) <= 0;
  *o = (unsigned long long)((KEY_WWW) | 0);  /* check that KEY_WWW is an integer */
  if (!_cffi_check_int(*o, n, 150U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_X(unsigned long long *o)
{
  int n = (KEY_X) <= 0;
  *o = (unsigned long long)((KEY_X) | 0);  /* check that KEY_X is an integer */
  if (!_cffi_check_int(*o, n, 45U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_XFER(unsigned long long *o)
{
  int n = (KEY_XFER) <= 0;
  *o = (unsigned long long)((KEY_XFER) | 0);  /* check that KEY_XFER is an integer */
  if (!_cffi_check_int(*o, n, 147U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_Y(unsigned long long *o)
{
  int n = (KEY_Y) <= 0;
  *o = (unsigned long long)((KEY_Y) | 0);  /* check that KEY_Y is an integer */
  if (!_cffi_check_int(*o, n, 21U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_YELLOW(unsigned long long *o)
{
  int n = (KEY_YELLOW) <= 0;
  *o = (unsigned long long)((KEY_YELLOW) | 0);  /* check that KEY_YELLOW is an integer */
  if (!_cffi_check_int(*o, n, 400U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_YEN(unsigned long long *o)
{
  int n = (KEY_YEN) <= 0;
  *o = (unsigned long long)((KEY_YEN) | 0);  /* check that KEY_YEN is an integer */
  if (!_cffi_check_int(*o, n, 124U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_Z(unsigned long long *o)
{
  int n = (KEY_Z) <= 0;
  *o = (unsigned long long)((KEY_Z) | 0);  /* check that KEY_Z is an integer */
  if (!_cffi_check_int(*o, n, 44U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ZENKAKUHANKAKU(unsigned long long *o)
{
  int n = (KEY_ZENKAKUHANKAKU) <= 0;
  *o = (unsigned long long)((KEY_ZENKAKUHANKAKU) | 0);  /* check that KEY_ZENKAKUHANKAKU is an integer */
  if (!_cffi_check_int(*o, n, 85U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ZOOM(unsigned long long *o)
{
  int n = (KEY_ZOOM) <= 0;
  *o = (unsigned long long)((KEY_ZOOM) | 0);  /* check that KEY_ZOOM is an integer */
  if (!_cffi_check_int(*o, n, 372U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ZOOMIN(unsigned long long *o)
{
  int n = (KEY_ZOOMIN) <= 0;
  *o = (unsigned long long)((KEY_ZOOMIN) | 0);  /* check that KEY_ZOOMIN is an integer */
  if (!_cffi_check_int(*o, n, 418U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ZOOMOUT(unsigned long long *o)
{
  int n = (KEY_ZOOMOUT) <= 0;
  *o = (unsigned long long)((KEY_ZOOMOUT) | 0);  /* check that KEY_ZOOMOUT is an integer */
  if (!_cffi_check_int(*o, n, 419U))
    n |= 2;
  return n;
}

static int _cffi_const_KEY_ZOOMRESET(unsigned long long *o)
{
  int n = (KEY_ZOOMRESET) <= 0;
  *o = (unsigned long long)((KEY_ZOOMRESET) | 0);  /* check that KEY_ZOOMRESET is an integer */
  if (!_cffi_check_int(*o, n, 420U))
    n |= 2;
  return n;
}

static int _cffi_const_LED_CAPSL(unsigned long long *o)
{
  int n = (LED_CAPSL) <= 0;
  *o = (unsigned long long)((LED_CAPSL) | 0);  /* check that LED_CAPSL is an integer */
  if (!_cffi_check_int(*o, n, 1U))
    n |= 2;
  return n;
}

static int _cffi_const_LED_CHARGING(unsigned long long *o)
{
  int n = (LED_CHARGING) <= 0;
  *o = (unsigned long long)((LED_CHARGING) | 0);  /* check that LED_CHARGING is an integer */
  if (!_cffi_check_int(*o, n, 10U))
    n |= 2;
  return n;
}

static int _cffi_const_LED_COMPOSE(unsigned long long *o)
{
  int n = (LED_COMPOSE) <= 0;
  *o = (unsigned long long)((LED_COMPOSE) | 0);  /* check that LED_COMPOSE is an integer */
  if (!_cffi_check_int(*o, n, 3U))
    n |= 2;
  return n;
}

static int _cffi_const_LED_KANA(unsigned long long *o)
{
  int n = (LED_KANA) <= 0;
  *o = (unsigned long long)((LED_KANA) | 0);  /* check that LED_KANA is an integer */
  if (!_cffi_check_int(*o, n, 4U))
    n |= 2;
  return n;
}

static int _cffi_const_LED_MAIL(unsigned long long *o)
{
  int n = (LED_MAIL) <= 0;
  *o = (unsigned long long)((LED_MAIL) | 0);  /* check that LED_MAIL is an integer */
  if (!_cffi_check_int(*o, n, 9U))
    n |= 2;
  return n;
}

static int _cffi_const_LED_MAX(unsigned long long *o)
{
  int n = (LED_MAX) <= 0;
  *o = (unsigned long long)((LED_MAX) | 0);  /* check that LED_MAX is an integer */
  if (!_cffi_check_int(*o, n, 15U))
    n |= 2;
  return n;
}

static int _cffi_const_LED_MISC(unsigned long long *o)
{
  int n = (LED_MISC) <= 0;
  *o = (unsigned long long)((LED_MISC) | 0);  /* check that LED_MISC is an integer */
  if (!_cffi_check_int(*o, n, 8U))
    n |= 2;
  return n;
}

static int _cffi_const_LED_MUTE(unsigned long long *o)
{
  int n = (LED_MUTE) <= 0;
  *o = (unsigned long long)((LED_MUTE) | 0);  /* check that LED_MUTE is an integer */
  if (!_cffi_check_int(*o, n, 7U))
    n |= 2;
  return n;
}

static int _cffi_const_LED_NUML(unsigned long long *o)
{
  int n = (LED_NUML) <= 0;
  *o = (unsigned long long)((LED_NUML) | 0);  /* check that LED_NUML is an integer */
  if (!_cffi_check_int(*o, n, 0))
    n |= 2;
  return n;
}

static int _cffi_const_LED_SCROLLL(unsigned long long *o)
{
  int n = (LED_SCROLLL) <= 0;
  *o = (unsigned long long)((LED_SCROLLL) | 0);  /* check that LED_SCROLLL is an integer */
  if (!_cffi_check_int(*o, n, 2U))
    n |= 2;
  return n;
}

static int _cffi_const_LED_SLEEP(unsigned long long *o)
{
  int n = (LED_SLEEP) <= 0;
  *o = (unsigned long long)((LED_SLEEP) | 0);  /* check that LED_SLEEP is an integer */
  if (!_cffi_check_int(*o, n, 5U))
    n |= 2;
  return n;
}

static int _cffi_const_LED_SUSPEND(unsigned long long *o)
{
  int n = (LED_SUSPEND) <= 0;
  *o = (unsigned long long)((LED_SUSPEND) | 0);  /* check that LED_SUSPEND is an integer */
  if (!_cffi_check_int(*o, n, 6U))
    n |= 2;
  return n;
}

static int _cffi_const_MSC_GESTURE(unsigned long long *o)
{
  int n = (MSC_GESTURE) <= 0;
  *o = (unsigned long long)((MSC_GESTURE) | 0);  /* check that MSC_GESTURE is an integer */
  if (!_cffi_check_int(*o, n, 2U))
    n |= 2;
  return n;
}

static int _cffi_const_MSC_MAX(unsigned long long *o)
{
  int n = (MSC_MAX) <= 0;
  *o = (unsigned long long)((MSC_MAX) | 0);  /* check that MSC_MAX is an integer */
  if (!_cffi_check_int(*o, n, 7U))
    n |= 2;
  return n;
}

static int _cffi_const_MSC_PULSELED(unsigned long long *o)
{
  int n = (MSC_PULSELED) <= 0;
  *o = (unsigned long long)((MSC_PULSELED) | 0);  /* check that MSC_PULSELED is an integer */
  if (!_cffi_check_int(*o, n, 1U))
    n |= 2;
  return n;
}

static int _cffi_const_MSC_RAW(unsigned long long *o)
{
  int n = (MSC_RAW) <= 0;
  *o = (unsigned long long)((MSC_RAW) | 0);  /* check that MSC_RAW is an integer */
  if (!_cffi_check_int(*o, n, 3U))
    n |= 2;
  return n;
}

static int _cffi_const_MSC_SCAN(unsigned long long *o)
{
  int n = (MSC_SCAN) <= 0;
  *o = (unsigned long long)((MSC_SCAN) | 0);  /* check that MSC_SCAN is an integer */
  if (!_cffi_check_int(*o, n, 4U))
    n |= 2;
  return n;
}

static int _cffi_const_MSC_SERIAL(unsigned long long *o)
{
  int n = (MSC_SERIAL) <= 0;
  *o = (unsigned long long)((MSC_SERIAL) | 0);  /* check that MSC_SERIAL is an integer */
  if (!_cffi_check_int(*o, n, 0))
    n |= 2;
  return n;
}

static int _cffi_const_MSC_TIMESTAMP(unsigned long long *o)
{
  int n = (MSC_TIMESTAMP) <= 0;
  *o = (unsigned long long)((MSC_TIMESTAMP) | 0);  /* check that MSC_TIMESTAMP is an integer */
  if (!_cffi_check_int(*o, n, 5U))
    n |= 2;
  return n;
}

static int _cffi_const_REL_DIAL(unsigned long long *o)
{
  int n = (REL_DIAL) <= 0;
  *o = (unsigned long long)((REL_DIAL) | 0);  /* check that REL_DIAL is an integer */
  if (!_cffi_check_int(*o, n, 7U))
    n |= 2;
  return n;
}

static int _cffi_const_REL_HWHEEL(unsigned long long *o)
{
  int n = (REL_HWHEEL) <= 0;
  *o = (unsigned long long)((REL_HWHEEL) | 0);  /* check that REL_HWHEEL is an integer */
  if (!_cffi_check_int(*o, n, 6U))
    n |= 2;
  return n;
}

static int _cffi_const_REL_MAX(unsigned long long *o)
{
  int n = (REL_MAX) <= 0;
  *o = (unsigned long long)((REL_MAX) | 0);  /* check that REL_MAX is an integer */
  if (!_cffi_check_int(*o, n, 15U))
    n |= 2;
  return n;
}

static int _cffi_const_REL_MISC(unsigned long long *o)
{
  int n = (REL_MISC) <= 0;
  *o = (unsigned long long)((REL_MISC) | 0);  /* check that REL_MISC is an integer */
  if (!_cffi_check_int(*o, n, 9U))
    n |= 2;
  return n;
}

static int _cffi_const_REL_RX(unsigned long long *o)
{
  int n = (REL_RX) <= 0;
  *o = (unsigned long long)((REL_RX) | 0);  /* check that REL_RX is an integer */
  if (!_cffi_check_int(*o, n, 3U))
    n |= 2;
  return n;
}

static int _cffi_const_REL_RY(unsigned long long *o)
{
  int n = (REL_RY) <= 0;
  *o = (unsigned long long)((REL_RY) | 0);  /* check that REL_RY is an integer */
  if (!_cffi_check_int(*o, n, 4U))
    n |= 2;
  return n;
}

static int _cffi_const_REL_RZ(unsigned long long *o)
{
  int n = (REL_RZ) <= 0;
  *o = (unsigned long long)((REL_RZ) | 0);  /* check that REL_RZ is an integer */
  if (!_cffi_check_int(*o, n, 5U))
    n |= 2;
  return n;
}

static int _cffi_const_REL_WHEEL(unsigned long long *o)
{
  int n = (REL_WHEEL) <= 0;
  *o = (unsigned long long)((REL_WHEEL) | 0);  /* check that REL_WHEEL is an integer */
  if (!_cffi_check_int(*o, n, 8U))
    n |= 2;
  return n;
}

static int _cffi_const_REL_X(unsigned long long *o)
{
  int n = (REL_X) <= 0;
  *o = (unsigned long long)((REL_X) | 0);  /* check that REL_X is an integer */
  if (!_cffi_check_int(*o, n, 0))
    n |= 2;
  return n;
}

static int _cffi_const_REL_Y(unsigned long long *o)
{
  int n = (REL_Y) <= 0;
  *o = (unsigned long long)((REL_Y) | 0);  /* check that REL_Y is an integer */
  if (!_cffi_check_int(*o, n, 1U))
    n |= 2;
  return n;
}

static int _cffi_const_REL_Z(unsigned long long *o)
{
  int n = (REL_Z) <= 0;
  *o = (unsigned long long)((REL_Z) | 0);  /* check that REL_Z is an integer */
  if (!_cffi_check_int(*o, n, 2U))
    n |= 2;
  return n;
}

static int _cffi_const_REP_DELAY(unsigned long long *o)
{
  int n = (REP_DELAY) <= 0;
  *o = (unsigned long long)((REP_DELAY) | 0);  /* check that REP_DELAY is an integer */
  if (!_cffi_check_int(*o, n, 0))
    n |= 2;
  return n;
}

static int _cffi_const_REP_MAX(unsigned long long *o)
{
  int n = (REP_MAX) <= 0;
  *o = (unsigned long long)((REP_MAX) | 0);  /* check that REP_MAX is an integer */
  if (!_cffi_check_int(*o, n, 1U))
    n |= 2;
  return n;
}

static int _cffi_const_REP_PERIOD(unsigned long long *o)
{
  int n = (REP_PERIOD) <= 0;
  *o = (unsigned long long)((REP_PERIOD) | 0);  /* check that REP_PERIOD is an integer */
  if (!_cffi_check_int(*o, n, 1U))
    n |= 2;
  return n;
}

static int _cffi_const_SND_BELL(unsigned long long *o)
{
  int n = (SND_BELL) <= 0;
  *o = (unsigned long long)((SND_BELL) | 0);  /* check that SND_BELL is an integer */
  if (!_cffi_check_int(*o, n, 1U))
    n |= 2;
  return n;
}

static int _cffi_const_SND_CLICK(unsigned long long *o)
{
  int n = (SND_CLICK) <= 0;
  *o = (unsigned long long)((SND_CLICK) | 0);  /* check that SND_CLICK is an integer */
  if (!_cffi_check_int(*o, n, 0))
    n |= 2;
  return n;
}

static int _cffi_const_SND_MAX(unsigned long long *o)
{
  int n = (SND_MAX) <= 0;
  *o = (unsigned long long)((SND_MAX) | 0);  /* check that SND_MAX is an integer */
  if (!_cffi_check_int(*o, n, 7U))
    n |= 2;
  return n;
}

static int _cffi_const_SND_TONE(unsigned long long *o)
{
  int n = (SND_TONE) <= 0;
  *o = (unsigned long long)((SND_TONE) | 0);  /* check that SND_TONE is an integer */
  if (!_cffi_check_int(*o, n, 2U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_CAMERA_LENS_COVER(unsigned long long *o)
{
  int n = (SW_CAMERA_LENS_COVER) <= 0;
  *o = (unsigned long long)((SW_CAMERA_LENS_COVER) | 0);  /* check that SW_CAMERA_LENS_COVER is an integer */
  if (!_cffi_check_int(*o, n, 9U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_DOCK(unsigned long long *o)
{
  int n = (SW_DOCK) <= 0;
  *o = (unsigned long long)((SW_DOCK) | 0);  /* check that SW_DOCK is an integer */
  if (!_cffi_check_int(*o, n, 5U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_FRONT_PROXIMITY(unsigned long long *o)
{
  int n = (SW_FRONT_PROXIMITY) <= 0;
  *o = (unsigned long long)((SW_FRONT_PROXIMITY) | 0);  /* check that SW_FRONT_PROXIMITY is an integer */
  if (!_cffi_check_int(*o, n, 11U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_HEADPHONE_INSERT(unsigned long long *o)
{
  int n = (SW_HEADPHONE_INSERT) <= 0;
  *o = (unsigned long long)((SW_HEADPHONE_INSERT) | 0);  /* check that SW_HEADPHONE_INSERT is an integer */
  if (!_cffi_check_int(*o, n, 2U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_JACK_PHYSICAL_INSERT(unsigned long long *o)
{
  int n = (SW_JACK_PHYSICAL_INSERT) <= 0;
  *o = (unsigned long long)((SW_JACK_PHYSICAL_INSERT) | 0);  /* check that SW_JACK_PHYSICAL_INSERT is an integer */
  if (!_cffi_check_int(*o, n, 7U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_KEYPAD_SLIDE(unsigned long long *o)
{
  int n = (SW_KEYPAD_SLIDE) <= 0;
  *o = (unsigned long long)((SW_KEYPAD_SLIDE) | 0);  /* check that SW_KEYPAD_SLIDE is an integer */
  if (!_cffi_check_int(*o, n, 10U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_LID(unsigned long long *o)
{
  int n = (SW_LID) <= 0;
  *o = (unsigned long long)((SW_LID) | 0);  /* check that SW_LID is an integer */
  if (!_cffi_check_int(*o, n, 0))
    n |= 2;
  return n;
}

static int _cffi_const_SW_LINEIN_INSERT(unsigned long long *o)
{
  int n = (SW_LINEIN_INSERT) <= 0;
  *o = (unsigned long long)((SW_LINEIN_INSERT) | 0);  /* check that SW_LINEIN_INSERT is an integer */
  if (!_cffi_check_int(*o, n, 13U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_LINEOUT_INSERT(unsigned long long *o)
{
  int n = (SW_LINEOUT_INSERT) <= 0;
  *o = (unsigned long long)((SW_LINEOUT_INSERT) | 0);  /* check that SW_LINEOUT_INSERT is an integer */
  if (!_cffi_check_int(*o, n, 6U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_MAX(unsigned long long *o)
{
  int n = (SW_MAX) <= 0;
  *o = (unsigned long long)((SW_MAX) | 0);  /* check that SW_MAX is an integer */
  if (!_cffi_check_int(*o, n, 15U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_MICROPHONE_INSERT(unsigned long long *o)
{
  int n = (SW_MICROPHONE_INSERT) <= 0;
  *o = (unsigned long long)((SW_MICROPHONE_INSERT) | 0);  /* check that SW_MICROPHONE_INSERT is an integer */
  if (!_cffi_check_int(*o, n, 4U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_MUTE_DEVICE(unsigned long long *o)
{
  int n = (SW_MUTE_DEVICE) <= 0;
  *o = (unsigned long long)((SW_MUTE_DEVICE) | 0);  /* check that SW_MUTE_DEVICE is an integer */
  if (!_cffi_check_int(*o, n, 14U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_PEN_INSERTED(unsigned long long *o)
{
  int n = (SW_PEN_INSERTED) <= 0;
  *o = (unsigned long long)((SW_PEN_INSERTED) | 0);  /* check that SW_PEN_INSERTED is an integer */
  if (!_cffi_check_int(*o, n, 15U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_RFKILL_ALL(unsigned long long *o)
{
  int n = (SW_RFKILL_ALL) <= 0;
  *o = (unsigned long long)((SW_RFKILL_ALL) | 0);  /* check that SW_RFKILL_ALL is an integer */
  if (!_cffi_check_int(*o, n, 3U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_ROTATE_LOCK(unsigned long long *o)
{
  int n = (SW_ROTATE_LOCK) <= 0;
  *o = (unsigned long long)((SW_ROTATE_LOCK) | 0);  /* check that SW_ROTATE_LOCK is an integer */
  if (!_cffi_check_int(*o, n, 12U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_TABLET_MODE(unsigned long long *o)
{
  int n = (SW_TABLET_MODE) <= 0;
  *o = (unsigned long long)((SW_TABLET_MODE) | 0);  /* check that SW_TABLET_MODE is an integer */
  if (!_cffi_check_int(*o, n, 1U))
    n |= 2;
  return n;
}

static int _cffi_const_SW_VIDEOOUT_INSERT(unsigned long long *o)
{
  int n = (SW_VIDEOOUT_INSERT) <= 0;
  *o = (unsigned long long)((SW_VIDEOOUT_INSERT) | 0);  /* check that SW_VIDEOOUT_INSERT is an integer */
  if (!_cffi_check_int(*o, n, 8U))
    n |= 2;
  return n;
}

static int _cffi_const_SYN_CONFIG(unsigned long long *o)
{
  int n = (SYN_CONFIG) <= 0;
  *o = (unsigned long long)((SYN_CONFIG) | 0);  /* check that SYN_CONFIG is an integer */
  if (!_cffi_check_int(*o, n, 1U))
    n |= 2;
  return n;
}

static int _cffi_const_SYN_DROPPED(unsigned long long *o)
{
  int n = (SYN_DROPPED) <= 0;
  *o = (unsigned long long)((SYN_DROPPED) | 0);  /* check that SYN_DROPPED is an integer */
  if (!_cffi_check_int(*o, n, 3U))
    n |= 2;
  return n;
}

static int _cffi_const_SYN_MAX(unsigned long long *o)
{
  int n = (SYN_MAX) <= 0;
  *o = (unsigned long long)((SYN_MAX) | 0);  /* check that SYN_MAX is an integer */
  if (!_cffi_check_int(*o, n, 15U))
    n |= 2;
  return n;
}

static int _cffi_const_SYN_MT_REPORT(unsigned long long *o)
{
  int n = (SYN_MT_REPORT) <= 0;
  *o = (unsigned long long)((SYN_MT_REPORT) | 0);  /* check that SYN_MT_REPORT is an integer */
  if (!_cffi_check_int(*o, n, 2U))
    n |= 2;
  return n;
}

static int _cffi_const_SYN_REPORT(unsigned long long *o)
{
  int n = (SYN_REPORT) <= 0;
  *o = (unsigned long long)((SYN_REPORT) | 0);  /* check that SYN_REPORT is an integer */
  if (!_cffi_check_int(*o, n, 0))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_0(unsigned long long *o)
{
  int n = (XKB_KEY_0) <= 0;
  *o = (unsigned long long)((XKB_KEY_0) | 0);  /* check that XKB_KEY_0 is an integer */
  if (!_cffi_check_int(*o, n, 48U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_1(unsigned long long *o)
{
  int n = (XKB_KEY_1) <= 0;
  *o = (unsigned long long)((XKB_KEY_1) | 0);  /* check that XKB_KEY_1 is an integer */
  if (!_cffi_check_int(*o, n, 49U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_2(unsigned long long *o)
{
  int n = (XKB_KEY_2) <= 0;
  *o = (unsigned long long)((XKB_KEY_2) | 0);  /* check that XKB_KEY_2 is an integer */
  if (!_cffi_check_int(*o, n, 50U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3(unsigned long long *o)
{
  int n = (XKB_KEY_3) <= 0;
  *o = (unsigned long long)((XKB_KEY_3) | 0);  /* check that XKB_KEY_3 is an integer */
  if (!_cffi_check_int(*o, n, 51U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_AltCursor(unsigned long long *o)
{
  int n = (XKB_KEY_3270_AltCursor) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_AltCursor) | 0);  /* check that XKB_KEY_3270_AltCursor is an integer */
  if (!_cffi_check_int(*o, n, 64784U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Attn(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Attn) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Attn) | 0);  /* check that XKB_KEY_3270_Attn is an integer */
  if (!_cffi_check_int(*o, n, 64782U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_BackTab(unsigned long long *o)
{
  int n = (XKB_KEY_3270_BackTab) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_BackTab) | 0);  /* check that XKB_KEY_3270_BackTab is an integer */
  if (!_cffi_check_int(*o, n, 64773U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_ChangeScreen(unsigned long long *o)
{
  int n = (XKB_KEY_3270_ChangeScreen) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_ChangeScreen) | 0);  /* check that XKB_KEY_3270_ChangeScreen is an integer */
  if (!_cffi_check_int(*o, n, 64793U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Copy(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Copy) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Copy) | 0);  /* check that XKB_KEY_3270_Copy is an integer */
  if (!_cffi_check_int(*o, n, 64789U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_CursorBlink(unsigned long long *o)
{
  int n = (XKB_KEY_3270_CursorBlink) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_CursorBlink) | 0);  /* check that XKB_KEY_3270_CursorBlink is an integer */
  if (!_cffi_check_int(*o, n, 64783U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_CursorSelect(unsigned long long *o)
{
  int n = (XKB_KEY_3270_CursorSelect) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_CursorSelect) | 0);  /* check that XKB_KEY_3270_CursorSelect is an integer */
  if (!_cffi_check_int(*o, n, 64796U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_DeleteWord(unsigned long long *o)
{
  int n = (XKB_KEY_3270_DeleteWord) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_DeleteWord) | 0);  /* check that XKB_KEY_3270_DeleteWord is an integer */
  if (!_cffi_check_int(*o, n, 64794U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Duplicate(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Duplicate) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Duplicate) | 0);  /* check that XKB_KEY_3270_Duplicate is an integer */
  if (!_cffi_check_int(*o, n, 64769U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Enter(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Enter) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Enter) | 0);  /* check that XKB_KEY_3270_Enter is an integer */
  if (!_cffi_check_int(*o, n, 64798U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_EraseEOF(unsigned long long *o)
{
  int n = (XKB_KEY_3270_EraseEOF) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_EraseEOF) | 0);  /* check that XKB_KEY_3270_EraseEOF is an integer */
  if (!_cffi_check_int(*o, n, 64774U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_EraseInput(unsigned long long *o)
{
  int n = (XKB_KEY_3270_EraseInput) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_EraseInput) | 0);  /* check that XKB_KEY_3270_EraseInput is an integer */
  if (!_cffi_check_int(*o, n, 64775U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_ExSelect(unsigned long long *o)
{
  int n = (XKB_KEY_3270_ExSelect) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_ExSelect) | 0);  /* check that XKB_KEY_3270_ExSelect is an integer */
  if (!_cffi_check_int(*o, n, 64795U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_FieldMark(unsigned long long *o)
{
  int n = (XKB_KEY_3270_FieldMark) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_FieldMark) | 0);  /* check that XKB_KEY_3270_FieldMark is an integer */
  if (!_cffi_check_int(*o, n, 64770U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Ident(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Ident) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Ident) | 0);  /* check that XKB_KEY_3270_Ident is an integer */
  if (!_cffi_check_int(*o, n, 64787U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Jump(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Jump) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Jump) | 0);  /* check that XKB_KEY_3270_Jump is an integer */
  if (!_cffi_check_int(*o, n, 64786U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_KeyClick(unsigned long long *o)
{
  int n = (XKB_KEY_3270_KeyClick) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_KeyClick) | 0);  /* check that XKB_KEY_3270_KeyClick is an integer */
  if (!_cffi_check_int(*o, n, 64785U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Left2(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Left2) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Left2) | 0);  /* check that XKB_KEY_3270_Left2 is an integer */
  if (!_cffi_check_int(*o, n, 64772U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_PA1(unsigned long long *o)
{
  int n = (XKB_KEY_3270_PA1) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_PA1) | 0);  /* check that XKB_KEY_3270_PA1 is an integer */
  if (!_cffi_check_int(*o, n, 64778U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_PA2(unsigned long long *o)
{
  int n = (XKB_KEY_3270_PA2) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_PA2) | 0);  /* check that XKB_KEY_3270_PA2 is an integer */
  if (!_cffi_check_int(*o, n, 64779U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_PA3(unsigned long long *o)
{
  int n = (XKB_KEY_3270_PA3) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_PA3) | 0);  /* check that XKB_KEY_3270_PA3 is an integer */
  if (!_cffi_check_int(*o, n, 64780U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Play(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Play) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Play) | 0);  /* check that XKB_KEY_3270_Play is an integer */
  if (!_cffi_check_int(*o, n, 64790U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_PrintScreen(unsigned long long *o)
{
  int n = (XKB_KEY_3270_PrintScreen) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_PrintScreen) | 0);  /* check that XKB_KEY_3270_PrintScreen is an integer */
  if (!_cffi_check_int(*o, n, 64797U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Quit(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Quit) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Quit) | 0);  /* check that XKB_KEY_3270_Quit is an integer */
  if (!_cffi_check_int(*o, n, 64777U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Record(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Record) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Record) | 0);  /* check that XKB_KEY_3270_Record is an integer */
  if (!_cffi_check_int(*o, n, 64792U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Reset(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Reset) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Reset) | 0);  /* check that XKB_KEY_3270_Reset is an integer */
  if (!_cffi_check_int(*o, n, 64776U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Right2(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Right2) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Right2) | 0);  /* check that XKB_KEY_3270_Right2 is an integer */
  if (!_cffi_check_int(*o, n, 64771U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Rule(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Rule) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Rule) | 0);  /* check that XKB_KEY_3270_Rule is an integer */
  if (!_cffi_check_int(*o, n, 64788U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Setup(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Setup) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Setup) | 0);  /* check that XKB_KEY_3270_Setup is an integer */
  if (!_cffi_check_int(*o, n, 64791U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_3270_Test(unsigned long long *o)
{
  int n = (XKB_KEY_3270_Test) <= 0;
  *o = (unsigned long long)((XKB_KEY_3270_Test) | 0);  /* check that XKB_KEY_3270_Test is an integer */
  if (!_cffi_check_int(*o, n, 64781U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_4(unsigned long long *o)
{
  int n = (XKB_KEY_4) <= 0;
  *o = (unsigned long long)((XKB_KEY_4) | 0);  /* check that XKB_KEY_4 is an integer */
  if (!_cffi_check_int(*o, n, 52U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_5(unsigned long long *o)
{
  int n = (XKB_KEY_5) <= 0;
  *o = (unsigned long long)((XKB_KEY_5) | 0);  /* check that XKB_KEY_5 is an integer */
  if (!_cffi_check_int(*o, n, 53U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_6(unsigned long long *o)
{
  int n = (XKB_KEY_6) <= 0;
  *o = (unsigned long long)((XKB_KEY_6) | 0);  /* check that XKB_KEY_6 is an integer */
  if (!_cffi_check_int(*o, n, 54U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_7(unsigned long long *o)
{
  int n = (XKB_KEY_7) <= 0;
  *o = (unsigned long long)((XKB_KEY_7) | 0);  /* check that XKB_KEY_7 is an integer */
  if (!_cffi_check_int(*o, n, 55U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_8(unsigned long long *o)
{
  int n = (XKB_KEY_8) <= 0;
  *o = (unsigned long long)((XKB_KEY_8) | 0);  /* check that XKB_KEY_8 is an integer */
  if (!_cffi_check_int(*o, n, 56U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_9(unsigned long long *o)
{
  int n = (XKB_KEY_9) <= 0;
  *o = (unsigned long long)((XKB_KEY_9) | 0);  /* check that XKB_KEY_9 is an integer */
  if (!_cffi_check_int(*o, n, 57U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_A(unsigned long long *o)
{
  int n = (XKB_KEY_A) <= 0;
  *o = (unsigned long long)((XKB_KEY_A) | 0);  /* check that XKB_KEY_A is an integer */
  if (!_cffi_check_int(*o, n, 65U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_AE(unsigned long long *o)
{
  int n = (XKB_KEY_AE) <= 0;
  *o = (unsigned long long)((XKB_KEY_AE) | 0);  /* check that XKB_KEY_AE is an integer */
  if (!_cffi_check_int(*o, n, 198U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Aacute(unsigned long long *o)
{
  int n = (XKB_KEY_Aacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Aacute) | 0);  /* check that XKB_KEY_Aacute is an integer */
  if (!_cffi_check_int(*o, n, 193U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Abelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_Abelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Abelowdot) | 0);  /* check that XKB_KEY_Abelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785056U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Abreve(unsigned long long *o)
{
  int n = (XKB_KEY_Abreve) <= 0;
  *o = (unsigned long long)((XKB_KEY_Abreve) | 0);  /* check that XKB_KEY_Abreve is an integer */
  if (!_cffi_check_int(*o, n, 451U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Abreveacute(unsigned long long *o)
{
  int n = (XKB_KEY_Abreveacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Abreveacute) | 0);  /* check that XKB_KEY_Abreveacute is an integer */
  if (!_cffi_check_int(*o, n, 16785070U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Abrevebelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_Abrevebelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Abrevebelowdot) | 0);  /* check that XKB_KEY_Abrevebelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785078U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Abrevegrave(unsigned long long *o)
{
  int n = (XKB_KEY_Abrevegrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_Abrevegrave) | 0);  /* check that XKB_KEY_Abrevegrave is an integer */
  if (!_cffi_check_int(*o, n, 16785072U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Abrevehook(unsigned long long *o)
{
  int n = (XKB_KEY_Abrevehook) <= 0;
  *o = (unsigned long long)((XKB_KEY_Abrevehook) | 0);  /* check that XKB_KEY_Abrevehook is an integer */
  if (!_cffi_check_int(*o, n, 16785074U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Abrevetilde(unsigned long long *o)
{
  int n = (XKB_KEY_Abrevetilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Abrevetilde) | 0);  /* check that XKB_KEY_Abrevetilde is an integer */
  if (!_cffi_check_int(*o, n, 16785076U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_AccessX_Enable(unsigned long long *o)
{
  int n = (XKB_KEY_AccessX_Enable) <= 0;
  *o = (unsigned long long)((XKB_KEY_AccessX_Enable) | 0);  /* check that XKB_KEY_AccessX_Enable is an integer */
  if (!_cffi_check_int(*o, n, 65136U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_AccessX_Feedback_Enable(unsigned long long *o)
{
  int n = (XKB_KEY_AccessX_Feedback_Enable) <= 0;
  *o = (unsigned long long)((XKB_KEY_AccessX_Feedback_Enable) | 0);  /* check that XKB_KEY_AccessX_Feedback_Enable is an integer */
  if (!_cffi_check_int(*o, n, 65137U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Acircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_Acircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_Acircumflex) | 0);  /* check that XKB_KEY_Acircumflex is an integer */
  if (!_cffi_check_int(*o, n, 194U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Acircumflexacute(unsigned long long *o)
{
  int n = (XKB_KEY_Acircumflexacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Acircumflexacute) | 0);  /* check that XKB_KEY_Acircumflexacute is an integer */
  if (!_cffi_check_int(*o, n, 16785060U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Acircumflexbelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_Acircumflexbelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Acircumflexbelowdot) | 0);  /* check that XKB_KEY_Acircumflexbelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785068U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Acircumflexgrave(unsigned long long *o)
{
  int n = (XKB_KEY_Acircumflexgrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_Acircumflexgrave) | 0);  /* check that XKB_KEY_Acircumflexgrave is an integer */
  if (!_cffi_check_int(*o, n, 16785062U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Acircumflexhook(unsigned long long *o)
{
  int n = (XKB_KEY_Acircumflexhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_Acircumflexhook) | 0);  /* check that XKB_KEY_Acircumflexhook is an integer */
  if (!_cffi_check_int(*o, n, 16785064U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Acircumflextilde(unsigned long long *o)
{
  int n = (XKB_KEY_Acircumflextilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Acircumflextilde) | 0);  /* check that XKB_KEY_Acircumflextilde is an integer */
  if (!_cffi_check_int(*o, n, 16785066U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Adiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_Adiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Adiaeresis) | 0);  /* check that XKB_KEY_Adiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 196U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Agrave(unsigned long long *o)
{
  int n = (XKB_KEY_Agrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_Agrave) | 0);  /* check that XKB_KEY_Agrave is an integer */
  if (!_cffi_check_int(*o, n, 192U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ahook(unsigned long long *o)
{
  int n = (XKB_KEY_Ahook) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ahook) | 0);  /* check that XKB_KEY_Ahook is an integer */
  if (!_cffi_check_int(*o, n, 16785058U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Alt_L(unsigned long long *o)
{
  int n = (XKB_KEY_Alt_L) <= 0;
  *o = (unsigned long long)((XKB_KEY_Alt_L) | 0);  /* check that XKB_KEY_Alt_L is an integer */
  if (!_cffi_check_int(*o, n, 65513U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Alt_R(unsigned long long *o)
{
  int n = (XKB_KEY_Alt_R) <= 0;
  *o = (unsigned long long)((XKB_KEY_Alt_R) | 0);  /* check that XKB_KEY_Alt_R is an integer */
  if (!_cffi_check_int(*o, n, 65514U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Amacron(unsigned long long *o)
{
  int n = (XKB_KEY_Amacron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Amacron) | 0);  /* check that XKB_KEY_Amacron is an integer */
  if (!_cffi_check_int(*o, n, 960U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Aogonek(unsigned long long *o)
{
  int n = (XKB_KEY_Aogonek) <= 0;
  *o = (unsigned long long)((XKB_KEY_Aogonek) | 0);  /* check that XKB_KEY_Aogonek is an integer */
  if (!_cffi_check_int(*o, n, 417U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_0(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_0) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_0) | 0);  /* check that XKB_KEY_Arabic_0 is an integer */
  if (!_cffi_check_int(*o, n, 16778848U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_1(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_1) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_1) | 0);  /* check that XKB_KEY_Arabic_1 is an integer */
  if (!_cffi_check_int(*o, n, 16778849U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_2(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_2) | 0);  /* check that XKB_KEY_Arabic_2 is an integer */
  if (!_cffi_check_int(*o, n, 16778850U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_3(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_3) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_3) | 0);  /* check that XKB_KEY_Arabic_3 is an integer */
  if (!_cffi_check_int(*o, n, 16778851U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_4(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_4) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_4) | 0);  /* check that XKB_KEY_Arabic_4 is an integer */
  if (!_cffi_check_int(*o, n, 16778852U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_5(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_5) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_5) | 0);  /* check that XKB_KEY_Arabic_5 is an integer */
  if (!_cffi_check_int(*o, n, 16778853U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_6(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_6) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_6) | 0);  /* check that XKB_KEY_Arabic_6 is an integer */
  if (!_cffi_check_int(*o, n, 16778854U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_7(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_7) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_7) | 0);  /* check that XKB_KEY_Arabic_7 is an integer */
  if (!_cffi_check_int(*o, n, 16778855U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_8(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_8) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_8) | 0);  /* check that XKB_KEY_Arabic_8 is an integer */
  if (!_cffi_check_int(*o, n, 16778856U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_9(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_9) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_9) | 0);  /* check that XKB_KEY_Arabic_9 is an integer */
  if (!_cffi_check_int(*o, n, 16778857U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_ain(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_ain) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_ain) | 0);  /* check that XKB_KEY_Arabic_ain is an integer */
  if (!_cffi_check_int(*o, n, 1497U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_alef(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_alef) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_alef) | 0);  /* check that XKB_KEY_Arabic_alef is an integer */
  if (!_cffi_check_int(*o, n, 1479U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_alefmaksura(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_alefmaksura) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_alefmaksura) | 0);  /* check that XKB_KEY_Arabic_alefmaksura is an integer */
  if (!_cffi_check_int(*o, n, 1513U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_beh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_beh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_beh) | 0);  /* check that XKB_KEY_Arabic_beh is an integer */
  if (!_cffi_check_int(*o, n, 1480U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_comma(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_comma) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_comma) | 0);  /* check that XKB_KEY_Arabic_comma is an integer */
  if (!_cffi_check_int(*o, n, 1452U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_dad(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_dad) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_dad) | 0);  /* check that XKB_KEY_Arabic_dad is an integer */
  if (!_cffi_check_int(*o, n, 1494U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_dal(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_dal) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_dal) | 0);  /* check that XKB_KEY_Arabic_dal is an integer */
  if (!_cffi_check_int(*o, n, 1487U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_damma(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_damma) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_damma) | 0);  /* check that XKB_KEY_Arabic_damma is an integer */
  if (!_cffi_check_int(*o, n, 1519U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_dammatan(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_dammatan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_dammatan) | 0);  /* check that XKB_KEY_Arabic_dammatan is an integer */
  if (!_cffi_check_int(*o, n, 1516U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_ddal(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_ddal) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_ddal) | 0);  /* check that XKB_KEY_Arabic_ddal is an integer */
  if (!_cffi_check_int(*o, n, 16778888U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_farsi_yeh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_farsi_yeh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_farsi_yeh) | 0);  /* check that XKB_KEY_Arabic_farsi_yeh is an integer */
  if (!_cffi_check_int(*o, n, 16778956U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_fatha(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_fatha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_fatha) | 0);  /* check that XKB_KEY_Arabic_fatha is an integer */
  if (!_cffi_check_int(*o, n, 1518U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_fathatan(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_fathatan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_fathatan) | 0);  /* check that XKB_KEY_Arabic_fathatan is an integer */
  if (!_cffi_check_int(*o, n, 1515U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_feh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_feh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_feh) | 0);  /* check that XKB_KEY_Arabic_feh is an integer */
  if (!_cffi_check_int(*o, n, 1505U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_fullstop(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_fullstop) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_fullstop) | 0);  /* check that XKB_KEY_Arabic_fullstop is an integer */
  if (!_cffi_check_int(*o, n, 16778964U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_gaf(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_gaf) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_gaf) | 0);  /* check that XKB_KEY_Arabic_gaf is an integer */
  if (!_cffi_check_int(*o, n, 16778927U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_ghain(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_ghain) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_ghain) | 0);  /* check that XKB_KEY_Arabic_ghain is an integer */
  if (!_cffi_check_int(*o, n, 1498U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_ha(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_ha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_ha) | 0);  /* check that XKB_KEY_Arabic_ha is an integer */
  if (!_cffi_check_int(*o, n, 1511U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_hah(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_hah) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_hah) | 0);  /* check that XKB_KEY_Arabic_hah is an integer */
  if (!_cffi_check_int(*o, n, 1485U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_hamza(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_hamza) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_hamza) | 0);  /* check that XKB_KEY_Arabic_hamza is an integer */
  if (!_cffi_check_int(*o, n, 1473U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_hamza_above(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_hamza_above) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_hamza_above) | 0);  /* check that XKB_KEY_Arabic_hamza_above is an integer */
  if (!_cffi_check_int(*o, n, 16778836U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_hamza_below(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_hamza_below) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_hamza_below) | 0);  /* check that XKB_KEY_Arabic_hamza_below is an integer */
  if (!_cffi_check_int(*o, n, 16778837U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_hamzaonalef(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_hamzaonalef) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_hamzaonalef) | 0);  /* check that XKB_KEY_Arabic_hamzaonalef is an integer */
  if (!_cffi_check_int(*o, n, 1475U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_hamzaonwaw(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_hamzaonwaw) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_hamzaonwaw) | 0);  /* check that XKB_KEY_Arabic_hamzaonwaw is an integer */
  if (!_cffi_check_int(*o, n, 1476U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_hamzaonyeh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_hamzaonyeh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_hamzaonyeh) | 0);  /* check that XKB_KEY_Arabic_hamzaonyeh is an integer */
  if (!_cffi_check_int(*o, n, 1478U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_hamzaunderalef(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_hamzaunderalef) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_hamzaunderalef) | 0);  /* check that XKB_KEY_Arabic_hamzaunderalef is an integer */
  if (!_cffi_check_int(*o, n, 1477U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_heh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_heh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_heh) | 0);  /* check that XKB_KEY_Arabic_heh is an integer */
  if (!_cffi_check_int(*o, n, 1511U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_heh_doachashmee(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_heh_doachashmee) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_heh_doachashmee) | 0);  /* check that XKB_KEY_Arabic_heh_doachashmee is an integer */
  if (!_cffi_check_int(*o, n, 16778942U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_heh_goal(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_heh_goal) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_heh_goal) | 0);  /* check that XKB_KEY_Arabic_heh_goal is an integer */
  if (!_cffi_check_int(*o, n, 16778945U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_jeem(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_jeem) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_jeem) | 0);  /* check that XKB_KEY_Arabic_jeem is an integer */
  if (!_cffi_check_int(*o, n, 1484U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_jeh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_jeh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_jeh) | 0);  /* check that XKB_KEY_Arabic_jeh is an integer */
  if (!_cffi_check_int(*o, n, 16778904U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_kaf(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_kaf) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_kaf) | 0);  /* check that XKB_KEY_Arabic_kaf is an integer */
  if (!_cffi_check_int(*o, n, 1507U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_kasra(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_kasra) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_kasra) | 0);  /* check that XKB_KEY_Arabic_kasra is an integer */
  if (!_cffi_check_int(*o, n, 1520U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_kasratan(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_kasratan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_kasratan) | 0);  /* check that XKB_KEY_Arabic_kasratan is an integer */
  if (!_cffi_check_int(*o, n, 1517U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_keheh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_keheh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_keheh) | 0);  /* check that XKB_KEY_Arabic_keheh is an integer */
  if (!_cffi_check_int(*o, n, 16778921U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_khah(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_khah) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_khah) | 0);  /* check that XKB_KEY_Arabic_khah is an integer */
  if (!_cffi_check_int(*o, n, 1486U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_lam(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_lam) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_lam) | 0);  /* check that XKB_KEY_Arabic_lam is an integer */
  if (!_cffi_check_int(*o, n, 1508U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_madda_above(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_madda_above) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_madda_above) | 0);  /* check that XKB_KEY_Arabic_madda_above is an integer */
  if (!_cffi_check_int(*o, n, 16778835U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_maddaonalef(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_maddaonalef) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_maddaonalef) | 0);  /* check that XKB_KEY_Arabic_maddaonalef is an integer */
  if (!_cffi_check_int(*o, n, 1474U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_meem(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_meem) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_meem) | 0);  /* check that XKB_KEY_Arabic_meem is an integer */
  if (!_cffi_check_int(*o, n, 1509U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_noon(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_noon) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_noon) | 0);  /* check that XKB_KEY_Arabic_noon is an integer */
  if (!_cffi_check_int(*o, n, 1510U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_noon_ghunna(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_noon_ghunna) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_noon_ghunna) | 0);  /* check that XKB_KEY_Arabic_noon_ghunna is an integer */
  if (!_cffi_check_int(*o, n, 16778938U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_peh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_peh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_peh) | 0);  /* check that XKB_KEY_Arabic_peh is an integer */
  if (!_cffi_check_int(*o, n, 16778878U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_percent(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_percent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_percent) | 0);  /* check that XKB_KEY_Arabic_percent is an integer */
  if (!_cffi_check_int(*o, n, 16778858U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_qaf(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_qaf) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_qaf) | 0);  /* check that XKB_KEY_Arabic_qaf is an integer */
  if (!_cffi_check_int(*o, n, 1506U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_question_mark(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_question_mark) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_question_mark) | 0);  /* check that XKB_KEY_Arabic_question_mark is an integer */
  if (!_cffi_check_int(*o, n, 1471U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_ra(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_ra) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_ra) | 0);  /* check that XKB_KEY_Arabic_ra is an integer */
  if (!_cffi_check_int(*o, n, 1489U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_rreh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_rreh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_rreh) | 0);  /* check that XKB_KEY_Arabic_rreh is an integer */
  if (!_cffi_check_int(*o, n, 16778897U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_sad(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_sad) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_sad) | 0);  /* check that XKB_KEY_Arabic_sad is an integer */
  if (!_cffi_check_int(*o, n, 1493U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_seen(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_seen) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_seen) | 0);  /* check that XKB_KEY_Arabic_seen is an integer */
  if (!_cffi_check_int(*o, n, 1491U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_semicolon(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_semicolon) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_semicolon) | 0);  /* check that XKB_KEY_Arabic_semicolon is an integer */
  if (!_cffi_check_int(*o, n, 1467U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_shadda(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_shadda) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_shadda) | 0);  /* check that XKB_KEY_Arabic_shadda is an integer */
  if (!_cffi_check_int(*o, n, 1521U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_sheen(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_sheen) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_sheen) | 0);  /* check that XKB_KEY_Arabic_sheen is an integer */
  if (!_cffi_check_int(*o, n, 1492U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_sukun(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_sukun) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_sukun) | 0);  /* check that XKB_KEY_Arabic_sukun is an integer */
  if (!_cffi_check_int(*o, n, 1522U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_superscript_alef(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_superscript_alef) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_superscript_alef) | 0);  /* check that XKB_KEY_Arabic_superscript_alef is an integer */
  if (!_cffi_check_int(*o, n, 16778864U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_switch(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_switch) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_switch) | 0);  /* check that XKB_KEY_Arabic_switch is an integer */
  if (!_cffi_check_int(*o, n, 65406U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_tah(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_tah) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_tah) | 0);  /* check that XKB_KEY_Arabic_tah is an integer */
  if (!_cffi_check_int(*o, n, 1495U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_tatweel(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_tatweel) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_tatweel) | 0);  /* check that XKB_KEY_Arabic_tatweel is an integer */
  if (!_cffi_check_int(*o, n, 1504U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_tcheh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_tcheh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_tcheh) | 0);  /* check that XKB_KEY_Arabic_tcheh is an integer */
  if (!_cffi_check_int(*o, n, 16778886U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_teh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_teh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_teh) | 0);  /* check that XKB_KEY_Arabic_teh is an integer */
  if (!_cffi_check_int(*o, n, 1482U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_tehmarbuta(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_tehmarbuta) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_tehmarbuta) | 0);  /* check that XKB_KEY_Arabic_tehmarbuta is an integer */
  if (!_cffi_check_int(*o, n, 1481U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_thal(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_thal) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_thal) | 0);  /* check that XKB_KEY_Arabic_thal is an integer */
  if (!_cffi_check_int(*o, n, 1488U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_theh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_theh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_theh) | 0);  /* check that XKB_KEY_Arabic_theh is an integer */
  if (!_cffi_check_int(*o, n, 1483U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_tteh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_tteh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_tteh) | 0);  /* check that XKB_KEY_Arabic_tteh is an integer */
  if (!_cffi_check_int(*o, n, 16778873U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_veh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_veh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_veh) | 0);  /* check that XKB_KEY_Arabic_veh is an integer */
  if (!_cffi_check_int(*o, n, 16778916U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_waw(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_waw) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_waw) | 0);  /* check that XKB_KEY_Arabic_waw is an integer */
  if (!_cffi_check_int(*o, n, 1512U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_yeh(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_yeh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_yeh) | 0);  /* check that XKB_KEY_Arabic_yeh is an integer */
  if (!_cffi_check_int(*o, n, 1514U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_yeh_baree(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_yeh_baree) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_yeh_baree) | 0);  /* check that XKB_KEY_Arabic_yeh_baree is an integer */
  if (!_cffi_check_int(*o, n, 16778962U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_zah(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_zah) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_zah) | 0);  /* check that XKB_KEY_Arabic_zah is an integer */
  if (!_cffi_check_int(*o, n, 1496U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Arabic_zain(unsigned long long *o)
{
  int n = (XKB_KEY_Arabic_zain) <= 0;
  *o = (unsigned long long)((XKB_KEY_Arabic_zain) | 0);  /* check that XKB_KEY_Arabic_zain is an integer */
  if (!_cffi_check_int(*o, n, 1490U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Aring(unsigned long long *o)
{
  int n = (XKB_KEY_Aring) <= 0;
  *o = (unsigned long long)((XKB_KEY_Aring) | 0);  /* check that XKB_KEY_Aring is an integer */
  if (!_cffi_check_int(*o, n, 197U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_AT(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_AT) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_AT) | 0);  /* check that XKB_KEY_Armenian_AT is an integer */
  if (!_cffi_check_int(*o, n, 16778552U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_AYB(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_AYB) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_AYB) | 0);  /* check that XKB_KEY_Armenian_AYB is an integer */
  if (!_cffi_check_int(*o, n, 16778545U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_BEN(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_BEN) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_BEN) | 0);  /* check that XKB_KEY_Armenian_BEN is an integer */
  if (!_cffi_check_int(*o, n, 16778546U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_CHA(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_CHA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_CHA) | 0);  /* check that XKB_KEY_Armenian_CHA is an integer */
  if (!_cffi_check_int(*o, n, 16778569U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_DA(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_DA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_DA) | 0);  /* check that XKB_KEY_Armenian_DA is an integer */
  if (!_cffi_check_int(*o, n, 16778548U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_DZA(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_DZA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_DZA) | 0);  /* check that XKB_KEY_Armenian_DZA is an integer */
  if (!_cffi_check_int(*o, n, 16778561U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_E(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_E) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_E) | 0);  /* check that XKB_KEY_Armenian_E is an integer */
  if (!_cffi_check_int(*o, n, 16778551U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_FE(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_FE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_FE) | 0);  /* check that XKB_KEY_Armenian_FE is an integer */
  if (!_cffi_check_int(*o, n, 16778582U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_GHAT(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_GHAT) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_GHAT) | 0);  /* check that XKB_KEY_Armenian_GHAT is an integer */
  if (!_cffi_check_int(*o, n, 16778562U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_GIM(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_GIM) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_GIM) | 0);  /* check that XKB_KEY_Armenian_GIM is an integer */
  if (!_cffi_check_int(*o, n, 16778547U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_HI(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_HI) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_HI) | 0);  /* check that XKB_KEY_Armenian_HI is an integer */
  if (!_cffi_check_int(*o, n, 16778565U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_HO(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_HO) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_HO) | 0);  /* check that XKB_KEY_Armenian_HO is an integer */
  if (!_cffi_check_int(*o, n, 16778560U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_INI(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_INI) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_INI) | 0);  /* check that XKB_KEY_Armenian_INI is an integer */
  if (!_cffi_check_int(*o, n, 16778555U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_JE(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_JE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_JE) | 0);  /* check that XKB_KEY_Armenian_JE is an integer */
  if (!_cffi_check_int(*o, n, 16778571U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_KE(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_KE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_KE) | 0);  /* check that XKB_KEY_Armenian_KE is an integer */
  if (!_cffi_check_int(*o, n, 16778580U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_KEN(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_KEN) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_KEN) | 0);  /* check that XKB_KEY_Armenian_KEN is an integer */
  if (!_cffi_check_int(*o, n, 16778559U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_KHE(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_KHE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_KHE) | 0);  /* check that XKB_KEY_Armenian_KHE is an integer */
  if (!_cffi_check_int(*o, n, 16778557U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_LYUN(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_LYUN) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_LYUN) | 0);  /* check that XKB_KEY_Armenian_LYUN is an integer */
  if (!_cffi_check_int(*o, n, 16778556U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_MEN(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_MEN) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_MEN) | 0);  /* check that XKB_KEY_Armenian_MEN is an integer */
  if (!_cffi_check_int(*o, n, 16778564U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_NU(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_NU) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_NU) | 0);  /* check that XKB_KEY_Armenian_NU is an integer */
  if (!_cffi_check_int(*o, n, 16778566U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_O(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_O) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_O) | 0);  /* check that XKB_KEY_Armenian_O is an integer */
  if (!_cffi_check_int(*o, n, 16778581U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_PE(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_PE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_PE) | 0);  /* check that XKB_KEY_Armenian_PE is an integer */
  if (!_cffi_check_int(*o, n, 16778570U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_PYUR(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_PYUR) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_PYUR) | 0);  /* check that XKB_KEY_Armenian_PYUR is an integer */
  if (!_cffi_check_int(*o, n, 16778579U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_RA(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_RA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_RA) | 0);  /* check that XKB_KEY_Armenian_RA is an integer */
  if (!_cffi_check_int(*o, n, 16778572U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_RE(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_RE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_RE) | 0);  /* check that XKB_KEY_Armenian_RE is an integer */
  if (!_cffi_check_int(*o, n, 16778576U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_SE(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_SE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_SE) | 0);  /* check that XKB_KEY_Armenian_SE is an integer */
  if (!_cffi_check_int(*o, n, 16778573U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_SHA(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_SHA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_SHA) | 0);  /* check that XKB_KEY_Armenian_SHA is an integer */
  if (!_cffi_check_int(*o, n, 16778567U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_TCHE(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_TCHE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_TCHE) | 0);  /* check that XKB_KEY_Armenian_TCHE is an integer */
  if (!_cffi_check_int(*o, n, 16778563U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_TO(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_TO) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_TO) | 0);  /* check that XKB_KEY_Armenian_TO is an integer */
  if (!_cffi_check_int(*o, n, 16778553U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_TSA(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_TSA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_TSA) | 0);  /* check that XKB_KEY_Armenian_TSA is an integer */
  if (!_cffi_check_int(*o, n, 16778558U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_TSO(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_TSO) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_TSO) | 0);  /* check that XKB_KEY_Armenian_TSO is an integer */
  if (!_cffi_check_int(*o, n, 16778577U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_TYUN(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_TYUN) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_TYUN) | 0);  /* check that XKB_KEY_Armenian_TYUN is an integer */
  if (!_cffi_check_int(*o, n, 16778575U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_VEV(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_VEV) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_VEV) | 0);  /* check that XKB_KEY_Armenian_VEV is an integer */
  if (!_cffi_check_int(*o, n, 16778574U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_VO(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_VO) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_VO) | 0);  /* check that XKB_KEY_Armenian_VO is an integer */
  if (!_cffi_check_int(*o, n, 16778568U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_VYUN(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_VYUN) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_VYUN) | 0);  /* check that XKB_KEY_Armenian_VYUN is an integer */
  if (!_cffi_check_int(*o, n, 16778578U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_YECH(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_YECH) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_YECH) | 0);  /* check that XKB_KEY_Armenian_YECH is an integer */
  if (!_cffi_check_int(*o, n, 16778549U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_ZA(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_ZA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_ZA) | 0);  /* check that XKB_KEY_Armenian_ZA is an integer */
  if (!_cffi_check_int(*o, n, 16778550U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_ZHE(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_ZHE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_ZHE) | 0);  /* check that XKB_KEY_Armenian_ZHE is an integer */
  if (!_cffi_check_int(*o, n, 16778554U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_accent(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_accent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_accent) | 0);  /* check that XKB_KEY_Armenian_accent is an integer */
  if (!_cffi_check_int(*o, n, 16778587U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_amanak(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_amanak) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_amanak) | 0);  /* check that XKB_KEY_Armenian_amanak is an integer */
  if (!_cffi_check_int(*o, n, 16778588U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_apostrophe(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_apostrophe) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_apostrophe) | 0);  /* check that XKB_KEY_Armenian_apostrophe is an integer */
  if (!_cffi_check_int(*o, n, 16778586U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_at(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_at) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_at) | 0);  /* check that XKB_KEY_Armenian_at is an integer */
  if (!_cffi_check_int(*o, n, 16778600U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_ayb(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_ayb) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_ayb) | 0);  /* check that XKB_KEY_Armenian_ayb is an integer */
  if (!_cffi_check_int(*o, n, 16778593U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_ben(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_ben) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_ben) | 0);  /* check that XKB_KEY_Armenian_ben is an integer */
  if (!_cffi_check_int(*o, n, 16778594U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_but(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_but) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_but) | 0);  /* check that XKB_KEY_Armenian_but is an integer */
  if (!_cffi_check_int(*o, n, 16778589U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_cha(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_cha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_cha) | 0);  /* check that XKB_KEY_Armenian_cha is an integer */
  if (!_cffi_check_int(*o, n, 16778617U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_da(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_da) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_da) | 0);  /* check that XKB_KEY_Armenian_da is an integer */
  if (!_cffi_check_int(*o, n, 16778596U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_dza(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_dza) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_dza) | 0);  /* check that XKB_KEY_Armenian_dza is an integer */
  if (!_cffi_check_int(*o, n, 16778609U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_e(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_e) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_e) | 0);  /* check that XKB_KEY_Armenian_e is an integer */
  if (!_cffi_check_int(*o, n, 16778599U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_exclam(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_exclam) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_exclam) | 0);  /* check that XKB_KEY_Armenian_exclam is an integer */
  if (!_cffi_check_int(*o, n, 16778588U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_fe(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_fe) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_fe) | 0);  /* check that XKB_KEY_Armenian_fe is an integer */
  if (!_cffi_check_int(*o, n, 16778630U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_full_stop(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_full_stop) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_full_stop) | 0);  /* check that XKB_KEY_Armenian_full_stop is an integer */
  if (!_cffi_check_int(*o, n, 16778633U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_ghat(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_ghat) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_ghat) | 0);  /* check that XKB_KEY_Armenian_ghat is an integer */
  if (!_cffi_check_int(*o, n, 16778610U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_gim(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_gim) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_gim) | 0);  /* check that XKB_KEY_Armenian_gim is an integer */
  if (!_cffi_check_int(*o, n, 16778595U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_hi(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_hi) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_hi) | 0);  /* check that XKB_KEY_Armenian_hi is an integer */
  if (!_cffi_check_int(*o, n, 16778613U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_ho(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_ho) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_ho) | 0);  /* check that XKB_KEY_Armenian_ho is an integer */
  if (!_cffi_check_int(*o, n, 16778608U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_hyphen(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_hyphen) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_hyphen) | 0);  /* check that XKB_KEY_Armenian_hyphen is an integer */
  if (!_cffi_check_int(*o, n, 16778634U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_ini(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_ini) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_ini) | 0);  /* check that XKB_KEY_Armenian_ini is an integer */
  if (!_cffi_check_int(*o, n, 16778603U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_je(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_je) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_je) | 0);  /* check that XKB_KEY_Armenian_je is an integer */
  if (!_cffi_check_int(*o, n, 16778619U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_ke(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_ke) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_ke) | 0);  /* check that XKB_KEY_Armenian_ke is an integer */
  if (!_cffi_check_int(*o, n, 16778628U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_ken(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_ken) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_ken) | 0);  /* check that XKB_KEY_Armenian_ken is an integer */
  if (!_cffi_check_int(*o, n, 16778607U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_khe(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_khe) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_khe) | 0);  /* check that XKB_KEY_Armenian_khe is an integer */
  if (!_cffi_check_int(*o, n, 16778605U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_ligature_ew(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_ligature_ew) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_ligature_ew) | 0);  /* check that XKB_KEY_Armenian_ligature_ew is an integer */
  if (!_cffi_check_int(*o, n, 16778631U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_lyun(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_lyun) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_lyun) | 0);  /* check that XKB_KEY_Armenian_lyun is an integer */
  if (!_cffi_check_int(*o, n, 16778604U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_men(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_men) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_men) | 0);  /* check that XKB_KEY_Armenian_men is an integer */
  if (!_cffi_check_int(*o, n, 16778612U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_nu(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_nu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_nu) | 0);  /* check that XKB_KEY_Armenian_nu is an integer */
  if (!_cffi_check_int(*o, n, 16778614U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_o(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_o) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_o) | 0);  /* check that XKB_KEY_Armenian_o is an integer */
  if (!_cffi_check_int(*o, n, 16778629U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_paruyk(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_paruyk) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_paruyk) | 0);  /* check that XKB_KEY_Armenian_paruyk is an integer */
  if (!_cffi_check_int(*o, n, 16778590U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_pe(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_pe) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_pe) | 0);  /* check that XKB_KEY_Armenian_pe is an integer */
  if (!_cffi_check_int(*o, n, 16778618U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_pyur(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_pyur) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_pyur) | 0);  /* check that XKB_KEY_Armenian_pyur is an integer */
  if (!_cffi_check_int(*o, n, 16778627U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_question(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_question) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_question) | 0);  /* check that XKB_KEY_Armenian_question is an integer */
  if (!_cffi_check_int(*o, n, 16778590U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_ra(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_ra) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_ra) | 0);  /* check that XKB_KEY_Armenian_ra is an integer */
  if (!_cffi_check_int(*o, n, 16778620U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_re(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_re) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_re) | 0);  /* check that XKB_KEY_Armenian_re is an integer */
  if (!_cffi_check_int(*o, n, 16778624U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_se(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_se) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_se) | 0);  /* check that XKB_KEY_Armenian_se is an integer */
  if (!_cffi_check_int(*o, n, 16778621U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_separation_mark(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_separation_mark) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_separation_mark) | 0);  /* check that XKB_KEY_Armenian_separation_mark is an integer */
  if (!_cffi_check_int(*o, n, 16778589U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_sha(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_sha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_sha) | 0);  /* check that XKB_KEY_Armenian_sha is an integer */
  if (!_cffi_check_int(*o, n, 16778615U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_shesht(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_shesht) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_shesht) | 0);  /* check that XKB_KEY_Armenian_shesht is an integer */
  if (!_cffi_check_int(*o, n, 16778587U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_tche(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_tche) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_tche) | 0);  /* check that XKB_KEY_Armenian_tche is an integer */
  if (!_cffi_check_int(*o, n, 16778611U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_to(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_to) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_to) | 0);  /* check that XKB_KEY_Armenian_to is an integer */
  if (!_cffi_check_int(*o, n, 16778601U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_tsa(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_tsa) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_tsa) | 0);  /* check that XKB_KEY_Armenian_tsa is an integer */
  if (!_cffi_check_int(*o, n, 16778606U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_tso(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_tso) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_tso) | 0);  /* check that XKB_KEY_Armenian_tso is an integer */
  if (!_cffi_check_int(*o, n, 16778625U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_tyun(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_tyun) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_tyun) | 0);  /* check that XKB_KEY_Armenian_tyun is an integer */
  if (!_cffi_check_int(*o, n, 16778623U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_verjaket(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_verjaket) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_verjaket) | 0);  /* check that XKB_KEY_Armenian_verjaket is an integer */
  if (!_cffi_check_int(*o, n, 16778633U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_vev(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_vev) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_vev) | 0);  /* check that XKB_KEY_Armenian_vev is an integer */
  if (!_cffi_check_int(*o, n, 16778622U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_vo(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_vo) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_vo) | 0);  /* check that XKB_KEY_Armenian_vo is an integer */
  if (!_cffi_check_int(*o, n, 16778616U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_vyun(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_vyun) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_vyun) | 0);  /* check that XKB_KEY_Armenian_vyun is an integer */
  if (!_cffi_check_int(*o, n, 16778626U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_yech(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_yech) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_yech) | 0);  /* check that XKB_KEY_Armenian_yech is an integer */
  if (!_cffi_check_int(*o, n, 16778597U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_yentamna(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_yentamna) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_yentamna) | 0);  /* check that XKB_KEY_Armenian_yentamna is an integer */
  if (!_cffi_check_int(*o, n, 16778634U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_za(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_za) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_za) | 0);  /* check that XKB_KEY_Armenian_za is an integer */
  if (!_cffi_check_int(*o, n, 16778598U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Armenian_zhe(unsigned long long *o)
{
  int n = (XKB_KEY_Armenian_zhe) <= 0;
  *o = (unsigned long long)((XKB_KEY_Armenian_zhe) | 0);  /* check that XKB_KEY_Armenian_zhe is an integer */
  if (!_cffi_check_int(*o, n, 16778602U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Atilde(unsigned long long *o)
{
  int n = (XKB_KEY_Atilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Atilde) | 0);  /* check that XKB_KEY_Atilde is an integer */
  if (!_cffi_check_int(*o, n, 195U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_AudibleBell_Enable(unsigned long long *o)
{
  int n = (XKB_KEY_AudibleBell_Enable) <= 0;
  *o = (unsigned long long)((XKB_KEY_AudibleBell_Enable) | 0);  /* check that XKB_KEY_AudibleBell_Enable is an integer */
  if (!_cffi_check_int(*o, n, 65146U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_B(unsigned long long *o)
{
  int n = (XKB_KEY_B) <= 0;
  *o = (unsigned long long)((XKB_KEY_B) | 0);  /* check that XKB_KEY_B is an integer */
  if (!_cffi_check_int(*o, n, 66U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Babovedot(unsigned long long *o)
{
  int n = (XKB_KEY_Babovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Babovedot) | 0);  /* check that XKB_KEY_Babovedot is an integer */
  if (!_cffi_check_int(*o, n, 16784898U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_BackSpace(unsigned long long *o)
{
  int n = (XKB_KEY_BackSpace) <= 0;
  *o = (unsigned long long)((XKB_KEY_BackSpace) | 0);  /* check that XKB_KEY_BackSpace is an integer */
  if (!_cffi_check_int(*o, n, 65288U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_BackTab(unsigned long long *o)
{
  int n = (XKB_KEY_BackTab) <= 0;
  *o = (unsigned long long)((XKB_KEY_BackTab) | 0);  /* check that XKB_KEY_BackTab is an integer */
  if (!_cffi_check_int(*o, n, 268500852U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Begin(unsigned long long *o)
{
  int n = (XKB_KEY_Begin) <= 0;
  *o = (unsigned long long)((XKB_KEY_Begin) | 0);  /* check that XKB_KEY_Begin is an integer */
  if (!_cffi_check_int(*o, n, 65368U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_BounceKeys_Enable(unsigned long long *o)
{
  int n = (XKB_KEY_BounceKeys_Enable) <= 0;
  *o = (unsigned long long)((XKB_KEY_BounceKeys_Enable) | 0);  /* check that XKB_KEY_BounceKeys_Enable is an integer */
  if (!_cffi_check_int(*o, n, 65140U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Break(unsigned long long *o)
{
  int n = (XKB_KEY_Break) <= 0;
  *o = (unsigned long long)((XKB_KEY_Break) | 0);  /* check that XKB_KEY_Break is an integer */
  if (!_cffi_check_int(*o, n, 65387U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Byelorussian_SHORTU(unsigned long long *o)
{
  int n = (XKB_KEY_Byelorussian_SHORTU) <= 0;
  *o = (unsigned long long)((XKB_KEY_Byelorussian_SHORTU) | 0);  /* check that XKB_KEY_Byelorussian_SHORTU is an integer */
  if (!_cffi_check_int(*o, n, 1726U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Byelorussian_shortu(unsigned long long *o)
{
  int n = (XKB_KEY_Byelorussian_shortu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Byelorussian_shortu) | 0);  /* check that XKB_KEY_Byelorussian_shortu is an integer */
  if (!_cffi_check_int(*o, n, 1710U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_C(unsigned long long *o)
{
  int n = (XKB_KEY_C) <= 0;
  *o = (unsigned long long)((XKB_KEY_C) | 0);  /* check that XKB_KEY_C is an integer */
  if (!_cffi_check_int(*o, n, 67U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_CH(unsigned long long *o)
{
  int n = (XKB_KEY_CH) <= 0;
  *o = (unsigned long long)((XKB_KEY_CH) | 0);  /* check that XKB_KEY_CH is an integer */
  if (!_cffi_check_int(*o, n, 65186U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_C_H(unsigned long long *o)
{
  int n = (XKB_KEY_C_H) <= 0;
  *o = (unsigned long long)((XKB_KEY_C_H) | 0);  /* check that XKB_KEY_C_H is an integer */
  if (!_cffi_check_int(*o, n, 65189U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_C_h(unsigned long long *o)
{
  int n = (XKB_KEY_C_h) <= 0;
  *o = (unsigned long long)((XKB_KEY_C_h) | 0);  /* check that XKB_KEY_C_h is an integer */
  if (!_cffi_check_int(*o, n, 65188U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_Cabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cabovedot) | 0);  /* check that XKB_KEY_Cabovedot is an integer */
  if (!_cffi_check_int(*o, n, 709U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cacute(unsigned long long *o)
{
  int n = (XKB_KEY_Cacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cacute) | 0);  /* check that XKB_KEY_Cacute is an integer */
  if (!_cffi_check_int(*o, n, 454U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cancel(unsigned long long *o)
{
  int n = (XKB_KEY_Cancel) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cancel) | 0);  /* check that XKB_KEY_Cancel is an integer */
  if (!_cffi_check_int(*o, n, 65385U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Caps_Lock(unsigned long long *o)
{
  int n = (XKB_KEY_Caps_Lock) <= 0;
  *o = (unsigned long long)((XKB_KEY_Caps_Lock) | 0);  /* check that XKB_KEY_Caps_Lock is an integer */
  if (!_cffi_check_int(*o, n, 65509U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ccaron(unsigned long long *o)
{
  int n = (XKB_KEY_Ccaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ccaron) | 0);  /* check that XKB_KEY_Ccaron is an integer */
  if (!_cffi_check_int(*o, n, 456U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ccedilla(unsigned long long *o)
{
  int n = (XKB_KEY_Ccedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ccedilla) | 0);  /* check that XKB_KEY_Ccedilla is an integer */
  if (!_cffi_check_int(*o, n, 199U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ccircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_Ccircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ccircumflex) | 0);  /* check that XKB_KEY_Ccircumflex is an integer */
  if (!_cffi_check_int(*o, n, 710U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ch(unsigned long long *o)
{
  int n = (XKB_KEY_Ch) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ch) | 0);  /* check that XKB_KEY_Ch is an integer */
  if (!_cffi_check_int(*o, n, 65185U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Clear(unsigned long long *o)
{
  int n = (XKB_KEY_Clear) <= 0;
  *o = (unsigned long long)((XKB_KEY_Clear) | 0);  /* check that XKB_KEY_Clear is an integer */
  if (!_cffi_check_int(*o, n, 65291U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ClearLine(unsigned long long *o)
{
  int n = (XKB_KEY_ClearLine) <= 0;
  *o = (unsigned long long)((XKB_KEY_ClearLine) | 0);  /* check that XKB_KEY_ClearLine is an integer */
  if (!_cffi_check_int(*o, n, 268500847U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Codeinput(unsigned long long *o)
{
  int n = (XKB_KEY_Codeinput) <= 0;
  *o = (unsigned long long)((XKB_KEY_Codeinput) | 0);  /* check that XKB_KEY_Codeinput is an integer */
  if (!_cffi_check_int(*o, n, 65335U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ColonSign(unsigned long long *o)
{
  int n = (XKB_KEY_ColonSign) <= 0;
  *o = (unsigned long long)((XKB_KEY_ColonSign) | 0);  /* check that XKB_KEY_ColonSign is an integer */
  if (!_cffi_check_int(*o, n, 16785569U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Control_L(unsigned long long *o)
{
  int n = (XKB_KEY_Control_L) <= 0;
  *o = (unsigned long long)((XKB_KEY_Control_L) | 0);  /* check that XKB_KEY_Control_L is an integer */
  if (!_cffi_check_int(*o, n, 65507U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Control_R(unsigned long long *o)
{
  int n = (XKB_KEY_Control_R) <= 0;
  *o = (unsigned long long)((XKB_KEY_Control_R) | 0);  /* check that XKB_KEY_Control_R is an integer */
  if (!_cffi_check_int(*o, n, 65508U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_CruzeiroSign(unsigned long long *o)
{
  int n = (XKB_KEY_CruzeiroSign) <= 0;
  *o = (unsigned long long)((XKB_KEY_CruzeiroSign) | 0);  /* check that XKB_KEY_CruzeiroSign is an integer */
  if (!_cffi_check_int(*o, n, 16785570U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_A(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_A) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_A) | 0);  /* check that XKB_KEY_Cyrillic_A is an integer */
  if (!_cffi_check_int(*o, n, 1761U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_BE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_BE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_BE) | 0);  /* check that XKB_KEY_Cyrillic_BE is an integer */
  if (!_cffi_check_int(*o, n, 1762U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_CHE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_CHE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_CHE) | 0);  /* check that XKB_KEY_Cyrillic_CHE is an integer */
  if (!_cffi_check_int(*o, n, 1790U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_CHE_descender(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_CHE_descender) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_CHE_descender) | 0);  /* check that XKB_KEY_Cyrillic_CHE_descender is an integer */
  if (!_cffi_check_int(*o, n, 16778422U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_CHE_vertstroke(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_CHE_vertstroke) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_CHE_vertstroke) | 0);  /* check that XKB_KEY_Cyrillic_CHE_vertstroke is an integer */
  if (!_cffi_check_int(*o, n, 16778424U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_DE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_DE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_DE) | 0);  /* check that XKB_KEY_Cyrillic_DE is an integer */
  if (!_cffi_check_int(*o, n, 1764U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_DZHE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_DZHE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_DZHE) | 0);  /* check that XKB_KEY_Cyrillic_DZHE is an integer */
  if (!_cffi_check_int(*o, n, 1727U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_E(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_E) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_E) | 0);  /* check that XKB_KEY_Cyrillic_E is an integer */
  if (!_cffi_check_int(*o, n, 1788U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_EF(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_EF) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_EF) | 0);  /* check that XKB_KEY_Cyrillic_EF is an integer */
  if (!_cffi_check_int(*o, n, 1766U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_EL(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_EL) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_EL) | 0);  /* check that XKB_KEY_Cyrillic_EL is an integer */
  if (!_cffi_check_int(*o, n, 1772U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_EM(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_EM) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_EM) | 0);  /* check that XKB_KEY_Cyrillic_EM is an integer */
  if (!_cffi_check_int(*o, n, 1773U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_EN(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_EN) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_EN) | 0);  /* check that XKB_KEY_Cyrillic_EN is an integer */
  if (!_cffi_check_int(*o, n, 1774U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_EN_descender(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_EN_descender) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_EN_descender) | 0);  /* check that XKB_KEY_Cyrillic_EN_descender is an integer */
  if (!_cffi_check_int(*o, n, 16778402U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ER(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ER) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ER) | 0);  /* check that XKB_KEY_Cyrillic_ER is an integer */
  if (!_cffi_check_int(*o, n, 1778U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ES(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ES) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ES) | 0);  /* check that XKB_KEY_Cyrillic_ES is an integer */
  if (!_cffi_check_int(*o, n, 1779U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_GHE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_GHE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_GHE) | 0);  /* check that XKB_KEY_Cyrillic_GHE is an integer */
  if (!_cffi_check_int(*o, n, 1767U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_GHE_bar(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_GHE_bar) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_GHE_bar) | 0);  /* check that XKB_KEY_Cyrillic_GHE_bar is an integer */
  if (!_cffi_check_int(*o, n, 16778386U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_HA(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_HA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_HA) | 0);  /* check that XKB_KEY_Cyrillic_HA is an integer */
  if (!_cffi_check_int(*o, n, 1768U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_HARDSIGN(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_HARDSIGN) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_HARDSIGN) | 0);  /* check that XKB_KEY_Cyrillic_HARDSIGN is an integer */
  if (!_cffi_check_int(*o, n, 1791U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_HA_descender(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_HA_descender) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_HA_descender) | 0);  /* check that XKB_KEY_Cyrillic_HA_descender is an integer */
  if (!_cffi_check_int(*o, n, 16778418U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_I(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_I) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_I) | 0);  /* check that XKB_KEY_Cyrillic_I is an integer */
  if (!_cffi_check_int(*o, n, 1769U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_IE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_IE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_IE) | 0);  /* check that XKB_KEY_Cyrillic_IE is an integer */
  if (!_cffi_check_int(*o, n, 1765U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_IO(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_IO) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_IO) | 0);  /* check that XKB_KEY_Cyrillic_IO is an integer */
  if (!_cffi_check_int(*o, n, 1715U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_I_macron(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_I_macron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_I_macron) | 0);  /* check that XKB_KEY_Cyrillic_I_macron is an integer */
  if (!_cffi_check_int(*o, n, 16778466U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_JE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_JE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_JE) | 0);  /* check that XKB_KEY_Cyrillic_JE is an integer */
  if (!_cffi_check_int(*o, n, 1720U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_KA(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_KA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_KA) | 0);  /* check that XKB_KEY_Cyrillic_KA is an integer */
  if (!_cffi_check_int(*o, n, 1771U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_KA_descender(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_KA_descender) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_KA_descender) | 0);  /* check that XKB_KEY_Cyrillic_KA_descender is an integer */
  if (!_cffi_check_int(*o, n, 16778394U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_KA_vertstroke(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_KA_vertstroke) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_KA_vertstroke) | 0);  /* check that XKB_KEY_Cyrillic_KA_vertstroke is an integer */
  if (!_cffi_check_int(*o, n, 16778396U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_LJE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_LJE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_LJE) | 0);  /* check that XKB_KEY_Cyrillic_LJE is an integer */
  if (!_cffi_check_int(*o, n, 1721U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_NJE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_NJE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_NJE) | 0);  /* check that XKB_KEY_Cyrillic_NJE is an integer */
  if (!_cffi_check_int(*o, n, 1722U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_O(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_O) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_O) | 0);  /* check that XKB_KEY_Cyrillic_O is an integer */
  if (!_cffi_check_int(*o, n, 1775U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_O_bar(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_O_bar) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_O_bar) | 0);  /* check that XKB_KEY_Cyrillic_O_bar is an integer */
  if (!_cffi_check_int(*o, n, 16778472U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_PE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_PE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_PE) | 0);  /* check that XKB_KEY_Cyrillic_PE is an integer */
  if (!_cffi_check_int(*o, n, 1776U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_SCHWA(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_SCHWA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_SCHWA) | 0);  /* check that XKB_KEY_Cyrillic_SCHWA is an integer */
  if (!_cffi_check_int(*o, n, 16778456U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_SHA(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_SHA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_SHA) | 0);  /* check that XKB_KEY_Cyrillic_SHA is an integer */
  if (!_cffi_check_int(*o, n, 1787U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_SHCHA(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_SHCHA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_SHCHA) | 0);  /* check that XKB_KEY_Cyrillic_SHCHA is an integer */
  if (!_cffi_check_int(*o, n, 1789U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_SHHA(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_SHHA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_SHHA) | 0);  /* check that XKB_KEY_Cyrillic_SHHA is an integer */
  if (!_cffi_check_int(*o, n, 16778426U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_SHORTI(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_SHORTI) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_SHORTI) | 0);  /* check that XKB_KEY_Cyrillic_SHORTI is an integer */
  if (!_cffi_check_int(*o, n, 1770U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_SOFTSIGN(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_SOFTSIGN) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_SOFTSIGN) | 0);  /* check that XKB_KEY_Cyrillic_SOFTSIGN is an integer */
  if (!_cffi_check_int(*o, n, 1784U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_TE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_TE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_TE) | 0);  /* check that XKB_KEY_Cyrillic_TE is an integer */
  if (!_cffi_check_int(*o, n, 1780U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_TSE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_TSE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_TSE) | 0);  /* check that XKB_KEY_Cyrillic_TSE is an integer */
  if (!_cffi_check_int(*o, n, 1763U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_U(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_U) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_U) | 0);  /* check that XKB_KEY_Cyrillic_U is an integer */
  if (!_cffi_check_int(*o, n, 1781U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_U_macron(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_U_macron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_U_macron) | 0);  /* check that XKB_KEY_Cyrillic_U_macron is an integer */
  if (!_cffi_check_int(*o, n, 16778478U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_U_straight(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_U_straight) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_U_straight) | 0);  /* check that XKB_KEY_Cyrillic_U_straight is an integer */
  if (!_cffi_check_int(*o, n, 16778414U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_U_straight_bar(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_U_straight_bar) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_U_straight_bar) | 0);  /* check that XKB_KEY_Cyrillic_U_straight_bar is an integer */
  if (!_cffi_check_int(*o, n, 16778416U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_VE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_VE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_VE) | 0);  /* check that XKB_KEY_Cyrillic_VE is an integer */
  if (!_cffi_check_int(*o, n, 1783U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_YA(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_YA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_YA) | 0);  /* check that XKB_KEY_Cyrillic_YA is an integer */
  if (!_cffi_check_int(*o, n, 1777U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_YERU(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_YERU) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_YERU) | 0);  /* check that XKB_KEY_Cyrillic_YERU is an integer */
  if (!_cffi_check_int(*o, n, 1785U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_YU(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_YU) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_YU) | 0);  /* check that XKB_KEY_Cyrillic_YU is an integer */
  if (!_cffi_check_int(*o, n, 1760U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ZE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ZE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ZE) | 0);  /* check that XKB_KEY_Cyrillic_ZE is an integer */
  if (!_cffi_check_int(*o, n, 1786U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ZHE(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ZHE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ZHE) | 0);  /* check that XKB_KEY_Cyrillic_ZHE is an integer */
  if (!_cffi_check_int(*o, n, 1782U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ZHE_descender(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ZHE_descender) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ZHE_descender) | 0);  /* check that XKB_KEY_Cyrillic_ZHE_descender is an integer */
  if (!_cffi_check_int(*o, n, 16778390U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_a(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_a) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_a) | 0);  /* check that XKB_KEY_Cyrillic_a is an integer */
  if (!_cffi_check_int(*o, n, 1729U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_be(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_be) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_be) | 0);  /* check that XKB_KEY_Cyrillic_be is an integer */
  if (!_cffi_check_int(*o, n, 1730U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_che(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_che) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_che) | 0);  /* check that XKB_KEY_Cyrillic_che is an integer */
  if (!_cffi_check_int(*o, n, 1758U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_che_descender(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_che_descender) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_che_descender) | 0);  /* check that XKB_KEY_Cyrillic_che_descender is an integer */
  if (!_cffi_check_int(*o, n, 16778423U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_che_vertstroke(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_che_vertstroke) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_che_vertstroke) | 0);  /* check that XKB_KEY_Cyrillic_che_vertstroke is an integer */
  if (!_cffi_check_int(*o, n, 16778425U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_de(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_de) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_de) | 0);  /* check that XKB_KEY_Cyrillic_de is an integer */
  if (!_cffi_check_int(*o, n, 1732U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_dzhe(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_dzhe) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_dzhe) | 0);  /* check that XKB_KEY_Cyrillic_dzhe is an integer */
  if (!_cffi_check_int(*o, n, 1711U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_e(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_e) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_e) | 0);  /* check that XKB_KEY_Cyrillic_e is an integer */
  if (!_cffi_check_int(*o, n, 1756U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ef(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ef) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ef) | 0);  /* check that XKB_KEY_Cyrillic_ef is an integer */
  if (!_cffi_check_int(*o, n, 1734U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_el(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_el) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_el) | 0);  /* check that XKB_KEY_Cyrillic_el is an integer */
  if (!_cffi_check_int(*o, n, 1740U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_em(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_em) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_em) | 0);  /* check that XKB_KEY_Cyrillic_em is an integer */
  if (!_cffi_check_int(*o, n, 1741U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_en(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_en) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_en) | 0);  /* check that XKB_KEY_Cyrillic_en is an integer */
  if (!_cffi_check_int(*o, n, 1742U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_en_descender(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_en_descender) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_en_descender) | 0);  /* check that XKB_KEY_Cyrillic_en_descender is an integer */
  if (!_cffi_check_int(*o, n, 16778403U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_er(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_er) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_er) | 0);  /* check that XKB_KEY_Cyrillic_er is an integer */
  if (!_cffi_check_int(*o, n, 1746U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_es(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_es) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_es) | 0);  /* check that XKB_KEY_Cyrillic_es is an integer */
  if (!_cffi_check_int(*o, n, 1747U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ghe(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ghe) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ghe) | 0);  /* check that XKB_KEY_Cyrillic_ghe is an integer */
  if (!_cffi_check_int(*o, n, 1735U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ghe_bar(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ghe_bar) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ghe_bar) | 0);  /* check that XKB_KEY_Cyrillic_ghe_bar is an integer */
  if (!_cffi_check_int(*o, n, 16778387U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ha(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ha) | 0);  /* check that XKB_KEY_Cyrillic_ha is an integer */
  if (!_cffi_check_int(*o, n, 1736U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ha_descender(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ha_descender) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ha_descender) | 0);  /* check that XKB_KEY_Cyrillic_ha_descender is an integer */
  if (!_cffi_check_int(*o, n, 16778419U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_hardsign(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_hardsign) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_hardsign) | 0);  /* check that XKB_KEY_Cyrillic_hardsign is an integer */
  if (!_cffi_check_int(*o, n, 1759U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_i(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_i) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_i) | 0);  /* check that XKB_KEY_Cyrillic_i is an integer */
  if (!_cffi_check_int(*o, n, 1737U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_i_macron(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_i_macron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_i_macron) | 0);  /* check that XKB_KEY_Cyrillic_i_macron is an integer */
  if (!_cffi_check_int(*o, n, 16778467U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ie(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ie) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ie) | 0);  /* check that XKB_KEY_Cyrillic_ie is an integer */
  if (!_cffi_check_int(*o, n, 1733U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_io(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_io) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_io) | 0);  /* check that XKB_KEY_Cyrillic_io is an integer */
  if (!_cffi_check_int(*o, n, 1699U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_je(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_je) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_je) | 0);  /* check that XKB_KEY_Cyrillic_je is an integer */
  if (!_cffi_check_int(*o, n, 1704U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ka(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ka) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ka) | 0);  /* check that XKB_KEY_Cyrillic_ka is an integer */
  if (!_cffi_check_int(*o, n, 1739U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ka_descender(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ka_descender) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ka_descender) | 0);  /* check that XKB_KEY_Cyrillic_ka_descender is an integer */
  if (!_cffi_check_int(*o, n, 16778395U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ka_vertstroke(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ka_vertstroke) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ka_vertstroke) | 0);  /* check that XKB_KEY_Cyrillic_ka_vertstroke is an integer */
  if (!_cffi_check_int(*o, n, 16778397U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_lje(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_lje) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_lje) | 0);  /* check that XKB_KEY_Cyrillic_lje is an integer */
  if (!_cffi_check_int(*o, n, 1705U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_nje(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_nje) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_nje) | 0);  /* check that XKB_KEY_Cyrillic_nje is an integer */
  if (!_cffi_check_int(*o, n, 1706U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_o(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_o) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_o) | 0);  /* check that XKB_KEY_Cyrillic_o is an integer */
  if (!_cffi_check_int(*o, n, 1743U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_o_bar(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_o_bar) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_o_bar) | 0);  /* check that XKB_KEY_Cyrillic_o_bar is an integer */
  if (!_cffi_check_int(*o, n, 16778473U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_pe(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_pe) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_pe) | 0);  /* check that XKB_KEY_Cyrillic_pe is an integer */
  if (!_cffi_check_int(*o, n, 1744U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_schwa(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_schwa) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_schwa) | 0);  /* check that XKB_KEY_Cyrillic_schwa is an integer */
  if (!_cffi_check_int(*o, n, 16778457U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_sha(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_sha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_sha) | 0);  /* check that XKB_KEY_Cyrillic_sha is an integer */
  if (!_cffi_check_int(*o, n, 1755U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_shcha(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_shcha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_shcha) | 0);  /* check that XKB_KEY_Cyrillic_shcha is an integer */
  if (!_cffi_check_int(*o, n, 1757U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_shha(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_shha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_shha) | 0);  /* check that XKB_KEY_Cyrillic_shha is an integer */
  if (!_cffi_check_int(*o, n, 16778427U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_shorti(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_shorti) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_shorti) | 0);  /* check that XKB_KEY_Cyrillic_shorti is an integer */
  if (!_cffi_check_int(*o, n, 1738U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_softsign(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_softsign) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_softsign) | 0);  /* check that XKB_KEY_Cyrillic_softsign is an integer */
  if (!_cffi_check_int(*o, n, 1752U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_te(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_te) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_te) | 0);  /* check that XKB_KEY_Cyrillic_te is an integer */
  if (!_cffi_check_int(*o, n, 1748U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_tse(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_tse) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_tse) | 0);  /* check that XKB_KEY_Cyrillic_tse is an integer */
  if (!_cffi_check_int(*o, n, 1731U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_u(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_u) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_u) | 0);  /* check that XKB_KEY_Cyrillic_u is an integer */
  if (!_cffi_check_int(*o, n, 1749U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_u_macron(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_u_macron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_u_macron) | 0);  /* check that XKB_KEY_Cyrillic_u_macron is an integer */
  if (!_cffi_check_int(*o, n, 16778479U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_u_straight(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_u_straight) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_u_straight) | 0);  /* check that XKB_KEY_Cyrillic_u_straight is an integer */
  if (!_cffi_check_int(*o, n, 16778415U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_u_straight_bar(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_u_straight_bar) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_u_straight_bar) | 0);  /* check that XKB_KEY_Cyrillic_u_straight_bar is an integer */
  if (!_cffi_check_int(*o, n, 16778417U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ve(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ve) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ve) | 0);  /* check that XKB_KEY_Cyrillic_ve is an integer */
  if (!_cffi_check_int(*o, n, 1751U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ya(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ya) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ya) | 0);  /* check that XKB_KEY_Cyrillic_ya is an integer */
  if (!_cffi_check_int(*o, n, 1745U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_yeru(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_yeru) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_yeru) | 0);  /* check that XKB_KEY_Cyrillic_yeru is an integer */
  if (!_cffi_check_int(*o, n, 1753U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_yu(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_yu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_yu) | 0);  /* check that XKB_KEY_Cyrillic_yu is an integer */
  if (!_cffi_check_int(*o, n, 1728U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_ze(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_ze) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_ze) | 0);  /* check that XKB_KEY_Cyrillic_ze is an integer */
  if (!_cffi_check_int(*o, n, 1754U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_zhe(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_zhe) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_zhe) | 0);  /* check that XKB_KEY_Cyrillic_zhe is an integer */
  if (!_cffi_check_int(*o, n, 1750U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Cyrillic_zhe_descender(unsigned long long *o)
{
  int n = (XKB_KEY_Cyrillic_zhe_descender) <= 0;
  *o = (unsigned long long)((XKB_KEY_Cyrillic_zhe_descender) | 0);  /* check that XKB_KEY_Cyrillic_zhe_descender is an integer */
  if (!_cffi_check_int(*o, n, 16778391U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_D(unsigned long long *o)
{
  int n = (XKB_KEY_D) <= 0;
  *o = (unsigned long long)((XKB_KEY_D) | 0);  /* check that XKB_KEY_D is an integer */
  if (!_cffi_check_int(*o, n, 68U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_DRemove(unsigned long long *o)
{
  int n = (XKB_KEY_DRemove) <= 0;
  *o = (unsigned long long)((XKB_KEY_DRemove) | 0);  /* check that XKB_KEY_DRemove is an integer */
  if (!_cffi_check_int(*o, n, 268500736U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Dabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_Dabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Dabovedot) | 0);  /* check that XKB_KEY_Dabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16784906U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Dacute_accent(unsigned long long *o)
{
  int n = (XKB_KEY_Dacute_accent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Dacute_accent) | 0);  /* check that XKB_KEY_Dacute_accent is an integer */
  if (!_cffi_check_int(*o, n, 268500519U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Dcaron(unsigned long long *o)
{
  int n = (XKB_KEY_Dcaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Dcaron) | 0);  /* check that XKB_KEY_Dcaron is an integer */
  if (!_cffi_check_int(*o, n, 463U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Dcedilla_accent(unsigned long long *o)
{
  int n = (XKB_KEY_Dcedilla_accent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Dcedilla_accent) | 0);  /* check that XKB_KEY_Dcedilla_accent is an integer */
  if (!_cffi_check_int(*o, n, 268500524U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Dcircumflex_accent(unsigned long long *o)
{
  int n = (XKB_KEY_Dcircumflex_accent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Dcircumflex_accent) | 0);  /* check that XKB_KEY_Dcircumflex_accent is an integer */
  if (!_cffi_check_int(*o, n, 268500574U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ddiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_Ddiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ddiaeresis) | 0);  /* check that XKB_KEY_Ddiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 268500514U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Delete(unsigned long long *o)
{
  int n = (XKB_KEY_Delete) <= 0;
  *o = (unsigned long long)((XKB_KEY_Delete) | 0);  /* check that XKB_KEY_Delete is an integer */
  if (!_cffi_check_int(*o, n, 65535U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_DeleteChar(unsigned long long *o)
{
  int n = (XKB_KEY_DeleteChar) <= 0;
  *o = (unsigned long long)((XKB_KEY_DeleteChar) | 0);  /* check that XKB_KEY_DeleteChar is an integer */
  if (!_cffi_check_int(*o, n, 268500851U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_DeleteLine(unsigned long long *o)
{
  int n = (XKB_KEY_DeleteLine) <= 0;
  *o = (unsigned long long)((XKB_KEY_DeleteLine) | 0);  /* check that XKB_KEY_DeleteLine is an integer */
  if (!_cffi_check_int(*o, n, 268500849U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Dgrave_accent(unsigned long long *o)
{
  int n = (XKB_KEY_Dgrave_accent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Dgrave_accent) | 0);  /* check that XKB_KEY_Dgrave_accent is an integer */
  if (!_cffi_check_int(*o, n, 268500576U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_DongSign(unsigned long long *o)
{
  int n = (XKB_KEY_DongSign) <= 0;
  *o = (unsigned long long)((XKB_KEY_DongSign) | 0);  /* check that XKB_KEY_DongSign is an integer */
  if (!_cffi_check_int(*o, n, 16785579U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Down(unsigned long long *o)
{
  int n = (XKB_KEY_Down) <= 0;
  *o = (unsigned long long)((XKB_KEY_Down) | 0);  /* check that XKB_KEY_Down is an integer */
  if (!_cffi_check_int(*o, n, 65364U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Dring_accent(unsigned long long *o)
{
  int n = (XKB_KEY_Dring_accent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Dring_accent) | 0);  /* check that XKB_KEY_Dring_accent is an integer */
  if (!_cffi_check_int(*o, n, 268500656U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Dstroke(unsigned long long *o)
{
  int n = (XKB_KEY_Dstroke) <= 0;
  *o = (unsigned long long)((XKB_KEY_Dstroke) | 0);  /* check that XKB_KEY_Dstroke is an integer */
  if (!_cffi_check_int(*o, n, 464U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Dtilde(unsigned long long *o)
{
  int n = (XKB_KEY_Dtilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Dtilde) | 0);  /* check that XKB_KEY_Dtilde is an integer */
  if (!_cffi_check_int(*o, n, 268500606U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_E(unsigned long long *o)
{
  int n = (XKB_KEY_E) <= 0;
  *o = (unsigned long long)((XKB_KEY_E) | 0);  /* check that XKB_KEY_E is an integer */
  if (!_cffi_check_int(*o, n, 69U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ENG(unsigned long long *o)
{
  int n = (XKB_KEY_ENG) <= 0;
  *o = (unsigned long long)((XKB_KEY_ENG) | 0);  /* check that XKB_KEY_ENG is an integer */
  if (!_cffi_check_int(*o, n, 957U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ETH(unsigned long long *o)
{
  int n = (XKB_KEY_ETH) <= 0;
  *o = (unsigned long long)((XKB_KEY_ETH) | 0);  /* check that XKB_KEY_ETH is an integer */
  if (!_cffi_check_int(*o, n, 208U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_EZH(unsigned long long *o)
{
  int n = (XKB_KEY_EZH) <= 0;
  *o = (unsigned long long)((XKB_KEY_EZH) | 0);  /* check that XKB_KEY_EZH is an integer */
  if (!_cffi_check_int(*o, n, 16777655U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Eabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_Eabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Eabovedot) | 0);  /* check that XKB_KEY_Eabovedot is an integer */
  if (!_cffi_check_int(*o, n, 972U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Eacute(unsigned long long *o)
{
  int n = (XKB_KEY_Eacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Eacute) | 0);  /* check that XKB_KEY_Eacute is an integer */
  if (!_cffi_check_int(*o, n, 201U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ebelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_Ebelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ebelowdot) | 0);  /* check that XKB_KEY_Ebelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785080U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ecaron(unsigned long long *o)
{
  int n = (XKB_KEY_Ecaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ecaron) | 0);  /* check that XKB_KEY_Ecaron is an integer */
  if (!_cffi_check_int(*o, n, 460U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ecircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_Ecircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ecircumflex) | 0);  /* check that XKB_KEY_Ecircumflex is an integer */
  if (!_cffi_check_int(*o, n, 202U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ecircumflexacute(unsigned long long *o)
{
  int n = (XKB_KEY_Ecircumflexacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ecircumflexacute) | 0);  /* check that XKB_KEY_Ecircumflexacute is an integer */
  if (!_cffi_check_int(*o, n, 16785086U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ecircumflexbelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_Ecircumflexbelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ecircumflexbelowdot) | 0);  /* check that XKB_KEY_Ecircumflexbelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785094U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ecircumflexgrave(unsigned long long *o)
{
  int n = (XKB_KEY_Ecircumflexgrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ecircumflexgrave) | 0);  /* check that XKB_KEY_Ecircumflexgrave is an integer */
  if (!_cffi_check_int(*o, n, 16785088U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ecircumflexhook(unsigned long long *o)
{
  int n = (XKB_KEY_Ecircumflexhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ecircumflexhook) | 0);  /* check that XKB_KEY_Ecircumflexhook is an integer */
  if (!_cffi_check_int(*o, n, 16785090U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ecircumflextilde(unsigned long long *o)
{
  int n = (XKB_KEY_Ecircumflextilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ecircumflextilde) | 0);  /* check that XKB_KEY_Ecircumflextilde is an integer */
  if (!_cffi_check_int(*o, n, 16785092U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_EcuSign(unsigned long long *o)
{
  int n = (XKB_KEY_EcuSign) <= 0;
  *o = (unsigned long long)((XKB_KEY_EcuSign) | 0);  /* check that XKB_KEY_EcuSign is an integer */
  if (!_cffi_check_int(*o, n, 16785568U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ediaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_Ediaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ediaeresis) | 0);  /* check that XKB_KEY_Ediaeresis is an integer */
  if (!_cffi_check_int(*o, n, 203U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Egrave(unsigned long long *o)
{
  int n = (XKB_KEY_Egrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_Egrave) | 0);  /* check that XKB_KEY_Egrave is an integer */
  if (!_cffi_check_int(*o, n, 200U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ehook(unsigned long long *o)
{
  int n = (XKB_KEY_Ehook) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ehook) | 0);  /* check that XKB_KEY_Ehook is an integer */
  if (!_cffi_check_int(*o, n, 16785082U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Eisu_Shift(unsigned long long *o)
{
  int n = (XKB_KEY_Eisu_Shift) <= 0;
  *o = (unsigned long long)((XKB_KEY_Eisu_Shift) | 0);  /* check that XKB_KEY_Eisu_Shift is an integer */
  if (!_cffi_check_int(*o, n, 65327U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Eisu_toggle(unsigned long long *o)
{
  int n = (XKB_KEY_Eisu_toggle) <= 0;
  *o = (unsigned long long)((XKB_KEY_Eisu_toggle) | 0);  /* check that XKB_KEY_Eisu_toggle is an integer */
  if (!_cffi_check_int(*o, n, 65328U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Emacron(unsigned long long *o)
{
  int n = (XKB_KEY_Emacron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Emacron) | 0);  /* check that XKB_KEY_Emacron is an integer */
  if (!_cffi_check_int(*o, n, 938U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_End(unsigned long long *o)
{
  int n = (XKB_KEY_End) <= 0;
  *o = (unsigned long long)((XKB_KEY_End) | 0);  /* check that XKB_KEY_End is an integer */
  if (!_cffi_check_int(*o, n, 65367U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Eogonek(unsigned long long *o)
{
  int n = (XKB_KEY_Eogonek) <= 0;
  *o = (unsigned long long)((XKB_KEY_Eogonek) | 0);  /* check that XKB_KEY_Eogonek is an integer */
  if (!_cffi_check_int(*o, n, 458U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Escape(unsigned long long *o)
{
  int n = (XKB_KEY_Escape) <= 0;
  *o = (unsigned long long)((XKB_KEY_Escape) | 0);  /* check that XKB_KEY_Escape is an integer */
  if (!_cffi_check_int(*o, n, 65307U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Eth(unsigned long long *o)
{
  int n = (XKB_KEY_Eth) <= 0;
  *o = (unsigned long long)((XKB_KEY_Eth) | 0);  /* check that XKB_KEY_Eth is an integer */
  if (!_cffi_check_int(*o, n, 208U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Etilde(unsigned long long *o)
{
  int n = (XKB_KEY_Etilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Etilde) | 0);  /* check that XKB_KEY_Etilde is an integer */
  if (!_cffi_check_int(*o, n, 16785084U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_EuroSign(unsigned long long *o)
{
  int n = (XKB_KEY_EuroSign) <= 0;
  *o = (unsigned long long)((XKB_KEY_EuroSign) | 0);  /* check that XKB_KEY_EuroSign is an integer */
  if (!_cffi_check_int(*o, n, 8364U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Execute(unsigned long long *o)
{
  int n = (XKB_KEY_Execute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Execute) | 0);  /* check that XKB_KEY_Execute is an integer */
  if (!_cffi_check_int(*o, n, 65378U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ext16bit_L(unsigned long long *o)
{
  int n = (XKB_KEY_Ext16bit_L) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ext16bit_L) | 0);  /* check that XKB_KEY_Ext16bit_L is an integer */
  if (!_cffi_check_int(*o, n, 268500854U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ext16bit_R(unsigned long long *o)
{
  int n = (XKB_KEY_Ext16bit_R) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ext16bit_R) | 0);  /* check that XKB_KEY_Ext16bit_R is an integer */
  if (!_cffi_check_int(*o, n, 268500855U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F(unsigned long long *o)
{
  int n = (XKB_KEY_F) <= 0;
  *o = (unsigned long long)((XKB_KEY_F) | 0);  /* check that XKB_KEY_F is an integer */
  if (!_cffi_check_int(*o, n, 70U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F1(unsigned long long *o)
{
  int n = (XKB_KEY_F1) <= 0;
  *o = (unsigned long long)((XKB_KEY_F1) | 0);  /* check that XKB_KEY_F1 is an integer */
  if (!_cffi_check_int(*o, n, 65470U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F10(unsigned long long *o)
{
  int n = (XKB_KEY_F10) <= 0;
  *o = (unsigned long long)((XKB_KEY_F10) | 0);  /* check that XKB_KEY_F10 is an integer */
  if (!_cffi_check_int(*o, n, 65479U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F11(unsigned long long *o)
{
  int n = (XKB_KEY_F11) <= 0;
  *o = (unsigned long long)((XKB_KEY_F11) | 0);  /* check that XKB_KEY_F11 is an integer */
  if (!_cffi_check_int(*o, n, 65480U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F12(unsigned long long *o)
{
  int n = (XKB_KEY_F12) <= 0;
  *o = (unsigned long long)((XKB_KEY_F12) | 0);  /* check that XKB_KEY_F12 is an integer */
  if (!_cffi_check_int(*o, n, 65481U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F13(unsigned long long *o)
{
  int n = (XKB_KEY_F13) <= 0;
  *o = (unsigned long long)((XKB_KEY_F13) | 0);  /* check that XKB_KEY_F13 is an integer */
  if (!_cffi_check_int(*o, n, 65482U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F14(unsigned long long *o)
{
  int n = (XKB_KEY_F14) <= 0;
  *o = (unsigned long long)((XKB_KEY_F14) | 0);  /* check that XKB_KEY_F14 is an integer */
  if (!_cffi_check_int(*o, n, 65483U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F15(unsigned long long *o)
{
  int n = (XKB_KEY_F15) <= 0;
  *o = (unsigned long long)((XKB_KEY_F15) | 0);  /* check that XKB_KEY_F15 is an integer */
  if (!_cffi_check_int(*o, n, 65484U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F16(unsigned long long *o)
{
  int n = (XKB_KEY_F16) <= 0;
  *o = (unsigned long long)((XKB_KEY_F16) | 0);  /* check that XKB_KEY_F16 is an integer */
  if (!_cffi_check_int(*o, n, 65485U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F17(unsigned long long *o)
{
  int n = (XKB_KEY_F17) <= 0;
  *o = (unsigned long long)((XKB_KEY_F17) | 0);  /* check that XKB_KEY_F17 is an integer */
  if (!_cffi_check_int(*o, n, 65486U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F18(unsigned long long *o)
{
  int n = (XKB_KEY_F18) <= 0;
  *o = (unsigned long long)((XKB_KEY_F18) | 0);  /* check that XKB_KEY_F18 is an integer */
  if (!_cffi_check_int(*o, n, 65487U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F19(unsigned long long *o)
{
  int n = (XKB_KEY_F19) <= 0;
  *o = (unsigned long long)((XKB_KEY_F19) | 0);  /* check that XKB_KEY_F19 is an integer */
  if (!_cffi_check_int(*o, n, 65488U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F2(unsigned long long *o)
{
  int n = (XKB_KEY_F2) <= 0;
  *o = (unsigned long long)((XKB_KEY_F2) | 0);  /* check that XKB_KEY_F2 is an integer */
  if (!_cffi_check_int(*o, n, 65471U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F20(unsigned long long *o)
{
  int n = (XKB_KEY_F20) <= 0;
  *o = (unsigned long long)((XKB_KEY_F20) | 0);  /* check that XKB_KEY_F20 is an integer */
  if (!_cffi_check_int(*o, n, 65489U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F21(unsigned long long *o)
{
  int n = (XKB_KEY_F21) <= 0;
  *o = (unsigned long long)((XKB_KEY_F21) | 0);  /* check that XKB_KEY_F21 is an integer */
  if (!_cffi_check_int(*o, n, 65490U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F22(unsigned long long *o)
{
  int n = (XKB_KEY_F22) <= 0;
  *o = (unsigned long long)((XKB_KEY_F22) | 0);  /* check that XKB_KEY_F22 is an integer */
  if (!_cffi_check_int(*o, n, 65491U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F23(unsigned long long *o)
{
  int n = (XKB_KEY_F23) <= 0;
  *o = (unsigned long long)((XKB_KEY_F23) | 0);  /* check that XKB_KEY_F23 is an integer */
  if (!_cffi_check_int(*o, n, 65492U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F24(unsigned long long *o)
{
  int n = (XKB_KEY_F24) <= 0;
  *o = (unsigned long long)((XKB_KEY_F24) | 0);  /* check that XKB_KEY_F24 is an integer */
  if (!_cffi_check_int(*o, n, 65493U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F25(unsigned long long *o)
{
  int n = (XKB_KEY_F25) <= 0;
  *o = (unsigned long long)((XKB_KEY_F25) | 0);  /* check that XKB_KEY_F25 is an integer */
  if (!_cffi_check_int(*o, n, 65494U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F26(unsigned long long *o)
{
  int n = (XKB_KEY_F26) <= 0;
  *o = (unsigned long long)((XKB_KEY_F26) | 0);  /* check that XKB_KEY_F26 is an integer */
  if (!_cffi_check_int(*o, n, 65495U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F27(unsigned long long *o)
{
  int n = (XKB_KEY_F27) <= 0;
  *o = (unsigned long long)((XKB_KEY_F27) | 0);  /* check that XKB_KEY_F27 is an integer */
  if (!_cffi_check_int(*o, n, 65496U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F28(unsigned long long *o)
{
  int n = (XKB_KEY_F28) <= 0;
  *o = (unsigned long long)((XKB_KEY_F28) | 0);  /* check that XKB_KEY_F28 is an integer */
  if (!_cffi_check_int(*o, n, 65497U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F29(unsigned long long *o)
{
  int n = (XKB_KEY_F29) <= 0;
  *o = (unsigned long long)((XKB_KEY_F29) | 0);  /* check that XKB_KEY_F29 is an integer */
  if (!_cffi_check_int(*o, n, 65498U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F3(unsigned long long *o)
{
  int n = (XKB_KEY_F3) <= 0;
  *o = (unsigned long long)((XKB_KEY_F3) | 0);  /* check that XKB_KEY_F3 is an integer */
  if (!_cffi_check_int(*o, n, 65472U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F30(unsigned long long *o)
{
  int n = (XKB_KEY_F30) <= 0;
  *o = (unsigned long long)((XKB_KEY_F30) | 0);  /* check that XKB_KEY_F30 is an integer */
  if (!_cffi_check_int(*o, n, 65499U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F31(unsigned long long *o)
{
  int n = (XKB_KEY_F31) <= 0;
  *o = (unsigned long long)((XKB_KEY_F31) | 0);  /* check that XKB_KEY_F31 is an integer */
  if (!_cffi_check_int(*o, n, 65500U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F32(unsigned long long *o)
{
  int n = (XKB_KEY_F32) <= 0;
  *o = (unsigned long long)((XKB_KEY_F32) | 0);  /* check that XKB_KEY_F32 is an integer */
  if (!_cffi_check_int(*o, n, 65501U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F33(unsigned long long *o)
{
  int n = (XKB_KEY_F33) <= 0;
  *o = (unsigned long long)((XKB_KEY_F33) | 0);  /* check that XKB_KEY_F33 is an integer */
  if (!_cffi_check_int(*o, n, 65502U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F34(unsigned long long *o)
{
  int n = (XKB_KEY_F34) <= 0;
  *o = (unsigned long long)((XKB_KEY_F34) | 0);  /* check that XKB_KEY_F34 is an integer */
  if (!_cffi_check_int(*o, n, 65503U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F35(unsigned long long *o)
{
  int n = (XKB_KEY_F35) <= 0;
  *o = (unsigned long long)((XKB_KEY_F35) | 0);  /* check that XKB_KEY_F35 is an integer */
  if (!_cffi_check_int(*o, n, 65504U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F4(unsigned long long *o)
{
  int n = (XKB_KEY_F4) <= 0;
  *o = (unsigned long long)((XKB_KEY_F4) | 0);  /* check that XKB_KEY_F4 is an integer */
  if (!_cffi_check_int(*o, n, 65473U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F5(unsigned long long *o)
{
  int n = (XKB_KEY_F5) <= 0;
  *o = (unsigned long long)((XKB_KEY_F5) | 0);  /* check that XKB_KEY_F5 is an integer */
  if (!_cffi_check_int(*o, n, 65474U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F6(unsigned long long *o)
{
  int n = (XKB_KEY_F6) <= 0;
  *o = (unsigned long long)((XKB_KEY_F6) | 0);  /* check that XKB_KEY_F6 is an integer */
  if (!_cffi_check_int(*o, n, 65475U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F7(unsigned long long *o)
{
  int n = (XKB_KEY_F7) <= 0;
  *o = (unsigned long long)((XKB_KEY_F7) | 0);  /* check that XKB_KEY_F7 is an integer */
  if (!_cffi_check_int(*o, n, 65476U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F8(unsigned long long *o)
{
  int n = (XKB_KEY_F8) <= 0;
  *o = (unsigned long long)((XKB_KEY_F8) | 0);  /* check that XKB_KEY_F8 is an integer */
  if (!_cffi_check_int(*o, n, 65477U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_F9(unsigned long long *o)
{
  int n = (XKB_KEY_F9) <= 0;
  *o = (unsigned long long)((XKB_KEY_F9) | 0);  /* check that XKB_KEY_F9 is an integer */
  if (!_cffi_check_int(*o, n, 65478U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_FFrancSign(unsigned long long *o)
{
  int n = (XKB_KEY_FFrancSign) <= 0;
  *o = (unsigned long long)((XKB_KEY_FFrancSign) | 0);  /* check that XKB_KEY_FFrancSign is an integer */
  if (!_cffi_check_int(*o, n, 16785571U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Fabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_Fabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Fabovedot) | 0);  /* check that XKB_KEY_Fabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16784926U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Farsi_0(unsigned long long *o)
{
  int n = (XKB_KEY_Farsi_0) <= 0;
  *o = (unsigned long long)((XKB_KEY_Farsi_0) | 0);  /* check that XKB_KEY_Farsi_0 is an integer */
  if (!_cffi_check_int(*o, n, 16778992U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Farsi_1(unsigned long long *o)
{
  int n = (XKB_KEY_Farsi_1) <= 0;
  *o = (unsigned long long)((XKB_KEY_Farsi_1) | 0);  /* check that XKB_KEY_Farsi_1 is an integer */
  if (!_cffi_check_int(*o, n, 16778993U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Farsi_2(unsigned long long *o)
{
  int n = (XKB_KEY_Farsi_2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Farsi_2) | 0);  /* check that XKB_KEY_Farsi_2 is an integer */
  if (!_cffi_check_int(*o, n, 16778994U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Farsi_3(unsigned long long *o)
{
  int n = (XKB_KEY_Farsi_3) <= 0;
  *o = (unsigned long long)((XKB_KEY_Farsi_3) | 0);  /* check that XKB_KEY_Farsi_3 is an integer */
  if (!_cffi_check_int(*o, n, 16778995U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Farsi_4(unsigned long long *o)
{
  int n = (XKB_KEY_Farsi_4) <= 0;
  *o = (unsigned long long)((XKB_KEY_Farsi_4) | 0);  /* check that XKB_KEY_Farsi_4 is an integer */
  if (!_cffi_check_int(*o, n, 16778996U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Farsi_5(unsigned long long *o)
{
  int n = (XKB_KEY_Farsi_5) <= 0;
  *o = (unsigned long long)((XKB_KEY_Farsi_5) | 0);  /* check that XKB_KEY_Farsi_5 is an integer */
  if (!_cffi_check_int(*o, n, 16778997U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Farsi_6(unsigned long long *o)
{
  int n = (XKB_KEY_Farsi_6) <= 0;
  *o = (unsigned long long)((XKB_KEY_Farsi_6) | 0);  /* check that XKB_KEY_Farsi_6 is an integer */
  if (!_cffi_check_int(*o, n, 16778998U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Farsi_7(unsigned long long *o)
{
  int n = (XKB_KEY_Farsi_7) <= 0;
  *o = (unsigned long long)((XKB_KEY_Farsi_7) | 0);  /* check that XKB_KEY_Farsi_7 is an integer */
  if (!_cffi_check_int(*o, n, 16778999U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Farsi_8(unsigned long long *o)
{
  int n = (XKB_KEY_Farsi_8) <= 0;
  *o = (unsigned long long)((XKB_KEY_Farsi_8) | 0);  /* check that XKB_KEY_Farsi_8 is an integer */
  if (!_cffi_check_int(*o, n, 16779000U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Farsi_9(unsigned long long *o)
{
  int n = (XKB_KEY_Farsi_9) <= 0;
  *o = (unsigned long long)((XKB_KEY_Farsi_9) | 0);  /* check that XKB_KEY_Farsi_9 is an integer */
  if (!_cffi_check_int(*o, n, 16779001U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Farsi_yeh(unsigned long long *o)
{
  int n = (XKB_KEY_Farsi_yeh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Farsi_yeh) | 0);  /* check that XKB_KEY_Farsi_yeh is an integer */
  if (!_cffi_check_int(*o, n, 16778956U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Find(unsigned long long *o)
{
  int n = (XKB_KEY_Find) <= 0;
  *o = (unsigned long long)((XKB_KEY_Find) | 0);  /* check that XKB_KEY_Find is an integer */
  if (!_cffi_check_int(*o, n, 65384U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_First_Virtual_Screen(unsigned long long *o)
{
  int n = (XKB_KEY_First_Virtual_Screen) <= 0;
  *o = (unsigned long long)((XKB_KEY_First_Virtual_Screen) | 0);  /* check that XKB_KEY_First_Virtual_Screen is an integer */
  if (!_cffi_check_int(*o, n, 65232U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_G(unsigned long long *o)
{
  int n = (XKB_KEY_G) <= 0;
  *o = (unsigned long long)((XKB_KEY_G) | 0);  /* check that XKB_KEY_G is an integer */
  if (!_cffi_check_int(*o, n, 71U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Gabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_Gabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Gabovedot) | 0);  /* check that XKB_KEY_Gabovedot is an integer */
  if (!_cffi_check_int(*o, n, 725U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Gbreve(unsigned long long *o)
{
  int n = (XKB_KEY_Gbreve) <= 0;
  *o = (unsigned long long)((XKB_KEY_Gbreve) | 0);  /* check that XKB_KEY_Gbreve is an integer */
  if (!_cffi_check_int(*o, n, 683U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Gcaron(unsigned long long *o)
{
  int n = (XKB_KEY_Gcaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Gcaron) | 0);  /* check that XKB_KEY_Gcaron is an integer */
  if (!_cffi_check_int(*o, n, 16777702U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Gcedilla(unsigned long long *o)
{
  int n = (XKB_KEY_Gcedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_Gcedilla) | 0);  /* check that XKB_KEY_Gcedilla is an integer */
  if (!_cffi_check_int(*o, n, 939U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Gcircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_Gcircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_Gcircumflex) | 0);  /* check that XKB_KEY_Gcircumflex is an integer */
  if (!_cffi_check_int(*o, n, 728U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_an(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_an) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_an) | 0);  /* check that XKB_KEY_Georgian_an is an integer */
  if (!_cffi_check_int(*o, n, 16781520U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_ban(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_ban) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_ban) | 0);  /* check that XKB_KEY_Georgian_ban is an integer */
  if (!_cffi_check_int(*o, n, 16781521U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_can(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_can) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_can) | 0);  /* check that XKB_KEY_Georgian_can is an integer */
  if (!_cffi_check_int(*o, n, 16781546U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_char(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_char) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_char) | 0);  /* check that XKB_KEY_Georgian_char is an integer */
  if (!_cffi_check_int(*o, n, 16781549U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_chin(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_chin) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_chin) | 0);  /* check that XKB_KEY_Georgian_chin is an integer */
  if (!_cffi_check_int(*o, n, 16781545U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_cil(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_cil) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_cil) | 0);  /* check that XKB_KEY_Georgian_cil is an integer */
  if (!_cffi_check_int(*o, n, 16781548U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_don(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_don) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_don) | 0);  /* check that XKB_KEY_Georgian_don is an integer */
  if (!_cffi_check_int(*o, n, 16781523U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_en(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_en) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_en) | 0);  /* check that XKB_KEY_Georgian_en is an integer */
  if (!_cffi_check_int(*o, n, 16781524U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_fi(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_fi) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_fi) | 0);  /* check that XKB_KEY_Georgian_fi is an integer */
  if (!_cffi_check_int(*o, n, 16781558U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_gan(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_gan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_gan) | 0);  /* check that XKB_KEY_Georgian_gan is an integer */
  if (!_cffi_check_int(*o, n, 16781522U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_ghan(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_ghan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_ghan) | 0);  /* check that XKB_KEY_Georgian_ghan is an integer */
  if (!_cffi_check_int(*o, n, 16781542U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_hae(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_hae) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_hae) | 0);  /* check that XKB_KEY_Georgian_hae is an integer */
  if (!_cffi_check_int(*o, n, 16781552U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_har(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_har) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_har) | 0);  /* check that XKB_KEY_Georgian_har is an integer */
  if (!_cffi_check_int(*o, n, 16781556U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_he(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_he) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_he) | 0);  /* check that XKB_KEY_Georgian_he is an integer */
  if (!_cffi_check_int(*o, n, 16781553U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_hie(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_hie) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_hie) | 0);  /* check that XKB_KEY_Georgian_hie is an integer */
  if (!_cffi_check_int(*o, n, 16781554U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_hoe(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_hoe) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_hoe) | 0);  /* check that XKB_KEY_Georgian_hoe is an integer */
  if (!_cffi_check_int(*o, n, 16781557U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_in(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_in) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_in) | 0);  /* check that XKB_KEY_Georgian_in is an integer */
  if (!_cffi_check_int(*o, n, 16781528U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_jhan(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_jhan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_jhan) | 0);  /* check that XKB_KEY_Georgian_jhan is an integer */
  if (!_cffi_check_int(*o, n, 16781551U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_jil(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_jil) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_jil) | 0);  /* check that XKB_KEY_Georgian_jil is an integer */
  if (!_cffi_check_int(*o, n, 16781547U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_kan(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_kan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_kan) | 0);  /* check that XKB_KEY_Georgian_kan is an integer */
  if (!_cffi_check_int(*o, n, 16781529U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_khar(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_khar) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_khar) | 0);  /* check that XKB_KEY_Georgian_khar is an integer */
  if (!_cffi_check_int(*o, n, 16781541U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_las(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_las) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_las) | 0);  /* check that XKB_KEY_Georgian_las is an integer */
  if (!_cffi_check_int(*o, n, 16781530U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_man(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_man) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_man) | 0);  /* check that XKB_KEY_Georgian_man is an integer */
  if (!_cffi_check_int(*o, n, 16781531U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_nar(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_nar) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_nar) | 0);  /* check that XKB_KEY_Georgian_nar is an integer */
  if (!_cffi_check_int(*o, n, 16781532U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_on(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_on) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_on) | 0);  /* check that XKB_KEY_Georgian_on is an integer */
  if (!_cffi_check_int(*o, n, 16781533U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_par(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_par) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_par) | 0);  /* check that XKB_KEY_Georgian_par is an integer */
  if (!_cffi_check_int(*o, n, 16781534U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_phar(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_phar) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_phar) | 0);  /* check that XKB_KEY_Georgian_phar is an integer */
  if (!_cffi_check_int(*o, n, 16781540U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_qar(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_qar) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_qar) | 0);  /* check that XKB_KEY_Georgian_qar is an integer */
  if (!_cffi_check_int(*o, n, 16781543U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_rae(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_rae) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_rae) | 0);  /* check that XKB_KEY_Georgian_rae is an integer */
  if (!_cffi_check_int(*o, n, 16781536U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_san(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_san) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_san) | 0);  /* check that XKB_KEY_Georgian_san is an integer */
  if (!_cffi_check_int(*o, n, 16781537U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_shin(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_shin) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_shin) | 0);  /* check that XKB_KEY_Georgian_shin is an integer */
  if (!_cffi_check_int(*o, n, 16781544U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_tan(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_tan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_tan) | 0);  /* check that XKB_KEY_Georgian_tan is an integer */
  if (!_cffi_check_int(*o, n, 16781527U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_tar(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_tar) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_tar) | 0);  /* check that XKB_KEY_Georgian_tar is an integer */
  if (!_cffi_check_int(*o, n, 16781538U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_un(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_un) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_un) | 0);  /* check that XKB_KEY_Georgian_un is an integer */
  if (!_cffi_check_int(*o, n, 16781539U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_vin(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_vin) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_vin) | 0);  /* check that XKB_KEY_Georgian_vin is an integer */
  if (!_cffi_check_int(*o, n, 16781525U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_we(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_we) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_we) | 0);  /* check that XKB_KEY_Georgian_we is an integer */
  if (!_cffi_check_int(*o, n, 16781555U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_xan(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_xan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_xan) | 0);  /* check that XKB_KEY_Georgian_xan is an integer */
  if (!_cffi_check_int(*o, n, 16781550U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_zen(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_zen) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_zen) | 0);  /* check that XKB_KEY_Georgian_zen is an integer */
  if (!_cffi_check_int(*o, n, 16781526U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Georgian_zhar(unsigned long long *o)
{
  int n = (XKB_KEY_Georgian_zhar) <= 0;
  *o = (unsigned long long)((XKB_KEY_Georgian_zhar) | 0);  /* check that XKB_KEY_Georgian_zhar is an integer */
  if (!_cffi_check_int(*o, n, 16781535U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_ALPHA(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_ALPHA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_ALPHA) | 0);  /* check that XKB_KEY_Greek_ALPHA is an integer */
  if (!_cffi_check_int(*o, n, 1985U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_ALPHAaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_ALPHAaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_ALPHAaccent) | 0);  /* check that XKB_KEY_Greek_ALPHAaccent is an integer */
  if (!_cffi_check_int(*o, n, 1953U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_BETA(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_BETA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_BETA) | 0);  /* check that XKB_KEY_Greek_BETA is an integer */
  if (!_cffi_check_int(*o, n, 1986U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_CHI(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_CHI) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_CHI) | 0);  /* check that XKB_KEY_Greek_CHI is an integer */
  if (!_cffi_check_int(*o, n, 2007U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_DELTA(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_DELTA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_DELTA) | 0);  /* check that XKB_KEY_Greek_DELTA is an integer */
  if (!_cffi_check_int(*o, n, 1988U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_EPSILON(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_EPSILON) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_EPSILON) | 0);  /* check that XKB_KEY_Greek_EPSILON is an integer */
  if (!_cffi_check_int(*o, n, 1989U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_EPSILONaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_EPSILONaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_EPSILONaccent) | 0);  /* check that XKB_KEY_Greek_EPSILONaccent is an integer */
  if (!_cffi_check_int(*o, n, 1954U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_ETA(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_ETA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_ETA) | 0);  /* check that XKB_KEY_Greek_ETA is an integer */
  if (!_cffi_check_int(*o, n, 1991U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_ETAaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_ETAaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_ETAaccent) | 0);  /* check that XKB_KEY_Greek_ETAaccent is an integer */
  if (!_cffi_check_int(*o, n, 1955U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_GAMMA(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_GAMMA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_GAMMA) | 0);  /* check that XKB_KEY_Greek_GAMMA is an integer */
  if (!_cffi_check_int(*o, n, 1987U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_IOTA(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_IOTA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_IOTA) | 0);  /* check that XKB_KEY_Greek_IOTA is an integer */
  if (!_cffi_check_int(*o, n, 1993U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_IOTAaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_IOTAaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_IOTAaccent) | 0);  /* check that XKB_KEY_Greek_IOTAaccent is an integer */
  if (!_cffi_check_int(*o, n, 1956U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_IOTAdiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_IOTAdiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_IOTAdiaeresis) | 0);  /* check that XKB_KEY_Greek_IOTAdiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 1957U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_IOTAdieresis(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_IOTAdieresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_IOTAdieresis) | 0);  /* check that XKB_KEY_Greek_IOTAdieresis is an integer */
  if (!_cffi_check_int(*o, n, 1957U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_KAPPA(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_KAPPA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_KAPPA) | 0);  /* check that XKB_KEY_Greek_KAPPA is an integer */
  if (!_cffi_check_int(*o, n, 1994U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_LAMBDA(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_LAMBDA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_LAMBDA) | 0);  /* check that XKB_KEY_Greek_LAMBDA is an integer */
  if (!_cffi_check_int(*o, n, 1995U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_LAMDA(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_LAMDA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_LAMDA) | 0);  /* check that XKB_KEY_Greek_LAMDA is an integer */
  if (!_cffi_check_int(*o, n, 1995U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_MU(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_MU) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_MU) | 0);  /* check that XKB_KEY_Greek_MU is an integer */
  if (!_cffi_check_int(*o, n, 1996U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_NU(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_NU) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_NU) | 0);  /* check that XKB_KEY_Greek_NU is an integer */
  if (!_cffi_check_int(*o, n, 1997U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_OMEGA(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_OMEGA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_OMEGA) | 0);  /* check that XKB_KEY_Greek_OMEGA is an integer */
  if (!_cffi_check_int(*o, n, 2009U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_OMEGAaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_OMEGAaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_OMEGAaccent) | 0);  /* check that XKB_KEY_Greek_OMEGAaccent is an integer */
  if (!_cffi_check_int(*o, n, 1963U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_OMICRON(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_OMICRON) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_OMICRON) | 0);  /* check that XKB_KEY_Greek_OMICRON is an integer */
  if (!_cffi_check_int(*o, n, 1999U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_OMICRONaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_OMICRONaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_OMICRONaccent) | 0);  /* check that XKB_KEY_Greek_OMICRONaccent is an integer */
  if (!_cffi_check_int(*o, n, 1959U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_PHI(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_PHI) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_PHI) | 0);  /* check that XKB_KEY_Greek_PHI is an integer */
  if (!_cffi_check_int(*o, n, 2006U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_PI(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_PI) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_PI) | 0);  /* check that XKB_KEY_Greek_PI is an integer */
  if (!_cffi_check_int(*o, n, 2000U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_PSI(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_PSI) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_PSI) | 0);  /* check that XKB_KEY_Greek_PSI is an integer */
  if (!_cffi_check_int(*o, n, 2008U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_RHO(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_RHO) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_RHO) | 0);  /* check that XKB_KEY_Greek_RHO is an integer */
  if (!_cffi_check_int(*o, n, 2001U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_SIGMA(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_SIGMA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_SIGMA) | 0);  /* check that XKB_KEY_Greek_SIGMA is an integer */
  if (!_cffi_check_int(*o, n, 2002U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_TAU(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_TAU) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_TAU) | 0);  /* check that XKB_KEY_Greek_TAU is an integer */
  if (!_cffi_check_int(*o, n, 2004U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_THETA(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_THETA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_THETA) | 0);  /* check that XKB_KEY_Greek_THETA is an integer */
  if (!_cffi_check_int(*o, n, 1992U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_UPSILON(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_UPSILON) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_UPSILON) | 0);  /* check that XKB_KEY_Greek_UPSILON is an integer */
  if (!_cffi_check_int(*o, n, 2005U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_UPSILONaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_UPSILONaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_UPSILONaccent) | 0);  /* check that XKB_KEY_Greek_UPSILONaccent is an integer */
  if (!_cffi_check_int(*o, n, 1960U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_UPSILONdieresis(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_UPSILONdieresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_UPSILONdieresis) | 0);  /* check that XKB_KEY_Greek_UPSILONdieresis is an integer */
  if (!_cffi_check_int(*o, n, 1961U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_XI(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_XI) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_XI) | 0);  /* check that XKB_KEY_Greek_XI is an integer */
  if (!_cffi_check_int(*o, n, 1998U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_ZETA(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_ZETA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_ZETA) | 0);  /* check that XKB_KEY_Greek_ZETA is an integer */
  if (!_cffi_check_int(*o, n, 1990U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_accentdieresis(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_accentdieresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_accentdieresis) | 0);  /* check that XKB_KEY_Greek_accentdieresis is an integer */
  if (!_cffi_check_int(*o, n, 1966U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_alpha(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_alpha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_alpha) | 0);  /* check that XKB_KEY_Greek_alpha is an integer */
  if (!_cffi_check_int(*o, n, 2017U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_alphaaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_alphaaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_alphaaccent) | 0);  /* check that XKB_KEY_Greek_alphaaccent is an integer */
  if (!_cffi_check_int(*o, n, 1969U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_beta(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_beta) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_beta) | 0);  /* check that XKB_KEY_Greek_beta is an integer */
  if (!_cffi_check_int(*o, n, 2018U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_chi(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_chi) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_chi) | 0);  /* check that XKB_KEY_Greek_chi is an integer */
  if (!_cffi_check_int(*o, n, 2039U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_delta(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_delta) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_delta) | 0);  /* check that XKB_KEY_Greek_delta is an integer */
  if (!_cffi_check_int(*o, n, 2020U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_epsilon(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_epsilon) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_epsilon) | 0);  /* check that XKB_KEY_Greek_epsilon is an integer */
  if (!_cffi_check_int(*o, n, 2021U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_epsilonaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_epsilonaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_epsilonaccent) | 0);  /* check that XKB_KEY_Greek_epsilonaccent is an integer */
  if (!_cffi_check_int(*o, n, 1970U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_eta(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_eta) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_eta) | 0);  /* check that XKB_KEY_Greek_eta is an integer */
  if (!_cffi_check_int(*o, n, 2023U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_etaaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_etaaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_etaaccent) | 0);  /* check that XKB_KEY_Greek_etaaccent is an integer */
  if (!_cffi_check_int(*o, n, 1971U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_finalsmallsigma(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_finalsmallsigma) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_finalsmallsigma) | 0);  /* check that XKB_KEY_Greek_finalsmallsigma is an integer */
  if (!_cffi_check_int(*o, n, 2035U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_gamma(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_gamma) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_gamma) | 0);  /* check that XKB_KEY_Greek_gamma is an integer */
  if (!_cffi_check_int(*o, n, 2019U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_horizbar(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_horizbar) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_horizbar) | 0);  /* check that XKB_KEY_Greek_horizbar is an integer */
  if (!_cffi_check_int(*o, n, 1967U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_iota(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_iota) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_iota) | 0);  /* check that XKB_KEY_Greek_iota is an integer */
  if (!_cffi_check_int(*o, n, 2025U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_iotaaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_iotaaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_iotaaccent) | 0);  /* check that XKB_KEY_Greek_iotaaccent is an integer */
  if (!_cffi_check_int(*o, n, 1972U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_iotaaccentdieresis(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_iotaaccentdieresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_iotaaccentdieresis) | 0);  /* check that XKB_KEY_Greek_iotaaccentdieresis is an integer */
  if (!_cffi_check_int(*o, n, 1974U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_iotadieresis(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_iotadieresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_iotadieresis) | 0);  /* check that XKB_KEY_Greek_iotadieresis is an integer */
  if (!_cffi_check_int(*o, n, 1973U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_kappa(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_kappa) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_kappa) | 0);  /* check that XKB_KEY_Greek_kappa is an integer */
  if (!_cffi_check_int(*o, n, 2026U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_lambda(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_lambda) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_lambda) | 0);  /* check that XKB_KEY_Greek_lambda is an integer */
  if (!_cffi_check_int(*o, n, 2027U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_lamda(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_lamda) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_lamda) | 0);  /* check that XKB_KEY_Greek_lamda is an integer */
  if (!_cffi_check_int(*o, n, 2027U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_mu(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_mu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_mu) | 0);  /* check that XKB_KEY_Greek_mu is an integer */
  if (!_cffi_check_int(*o, n, 2028U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_nu(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_nu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_nu) | 0);  /* check that XKB_KEY_Greek_nu is an integer */
  if (!_cffi_check_int(*o, n, 2029U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_omega(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_omega) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_omega) | 0);  /* check that XKB_KEY_Greek_omega is an integer */
  if (!_cffi_check_int(*o, n, 2041U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_omegaaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_omegaaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_omegaaccent) | 0);  /* check that XKB_KEY_Greek_omegaaccent is an integer */
  if (!_cffi_check_int(*o, n, 1979U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_omicron(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_omicron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_omicron) | 0);  /* check that XKB_KEY_Greek_omicron is an integer */
  if (!_cffi_check_int(*o, n, 2031U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_omicronaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_omicronaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_omicronaccent) | 0);  /* check that XKB_KEY_Greek_omicronaccent is an integer */
  if (!_cffi_check_int(*o, n, 1975U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_phi(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_phi) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_phi) | 0);  /* check that XKB_KEY_Greek_phi is an integer */
  if (!_cffi_check_int(*o, n, 2038U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_pi(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_pi) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_pi) | 0);  /* check that XKB_KEY_Greek_pi is an integer */
  if (!_cffi_check_int(*o, n, 2032U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_psi(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_psi) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_psi) | 0);  /* check that XKB_KEY_Greek_psi is an integer */
  if (!_cffi_check_int(*o, n, 2040U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_rho(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_rho) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_rho) | 0);  /* check that XKB_KEY_Greek_rho is an integer */
  if (!_cffi_check_int(*o, n, 2033U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_sigma(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_sigma) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_sigma) | 0);  /* check that XKB_KEY_Greek_sigma is an integer */
  if (!_cffi_check_int(*o, n, 2034U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_switch(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_switch) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_switch) | 0);  /* check that XKB_KEY_Greek_switch is an integer */
  if (!_cffi_check_int(*o, n, 65406U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_tau(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_tau) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_tau) | 0);  /* check that XKB_KEY_Greek_tau is an integer */
  if (!_cffi_check_int(*o, n, 2036U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_theta(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_theta) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_theta) | 0);  /* check that XKB_KEY_Greek_theta is an integer */
  if (!_cffi_check_int(*o, n, 2024U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_upsilon(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_upsilon) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_upsilon) | 0);  /* check that XKB_KEY_Greek_upsilon is an integer */
  if (!_cffi_check_int(*o, n, 2037U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_upsilonaccent(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_upsilonaccent) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_upsilonaccent) | 0);  /* check that XKB_KEY_Greek_upsilonaccent is an integer */
  if (!_cffi_check_int(*o, n, 1976U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_upsilonaccentdieresis(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_upsilonaccentdieresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_upsilonaccentdieresis) | 0);  /* check that XKB_KEY_Greek_upsilonaccentdieresis is an integer */
  if (!_cffi_check_int(*o, n, 1978U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_upsilondieresis(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_upsilondieresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_upsilondieresis) | 0);  /* check that XKB_KEY_Greek_upsilondieresis is an integer */
  if (!_cffi_check_int(*o, n, 1977U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_xi(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_xi) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_xi) | 0);  /* check that XKB_KEY_Greek_xi is an integer */
  if (!_cffi_check_int(*o, n, 2030U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Greek_zeta(unsigned long long *o)
{
  int n = (XKB_KEY_Greek_zeta) <= 0;
  *o = (unsigned long long)((XKB_KEY_Greek_zeta) | 0);  /* check that XKB_KEY_Greek_zeta is an integer */
  if (!_cffi_check_int(*o, n, 2022U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_H(unsigned long long *o)
{
  int n = (XKB_KEY_H) <= 0;
  *o = (unsigned long long)((XKB_KEY_H) | 0);  /* check that XKB_KEY_H is an integer */
  if (!_cffi_check_int(*o, n, 72U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul) | 0);  /* check that XKB_KEY_Hangul is an integer */
  if (!_cffi_check_int(*o, n, 65329U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_A(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_A) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_A) | 0);  /* check that XKB_KEY_Hangul_A is an integer */
  if (!_cffi_check_int(*o, n, 3775U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_AE(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_AE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_AE) | 0);  /* check that XKB_KEY_Hangul_AE is an integer */
  if (!_cffi_check_int(*o, n, 3776U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_AraeA(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_AraeA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_AraeA) | 0);  /* check that XKB_KEY_Hangul_AraeA is an integer */
  if (!_cffi_check_int(*o, n, 3830U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_AraeAE(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_AraeAE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_AraeAE) | 0);  /* check that XKB_KEY_Hangul_AraeAE is an integer */
  if (!_cffi_check_int(*o, n, 3831U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Banja(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Banja) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Banja) | 0);  /* check that XKB_KEY_Hangul_Banja is an integer */
  if (!_cffi_check_int(*o, n, 65337U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Cieuc(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Cieuc) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Cieuc) | 0);  /* check that XKB_KEY_Hangul_Cieuc is an integer */
  if (!_cffi_check_int(*o, n, 3770U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Codeinput(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Codeinput) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Codeinput) | 0);  /* check that XKB_KEY_Hangul_Codeinput is an integer */
  if (!_cffi_check_int(*o, n, 65335U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Dikeud(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Dikeud) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Dikeud) | 0);  /* check that XKB_KEY_Hangul_Dikeud is an integer */
  if (!_cffi_check_int(*o, n, 3751U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_E(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_E) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_E) | 0);  /* check that XKB_KEY_Hangul_E is an integer */
  if (!_cffi_check_int(*o, n, 3780U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_EO(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_EO) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_EO) | 0);  /* check that XKB_KEY_Hangul_EO is an integer */
  if (!_cffi_check_int(*o, n, 3779U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_EU(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_EU) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_EU) | 0);  /* check that XKB_KEY_Hangul_EU is an integer */
  if (!_cffi_check_int(*o, n, 3793U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_End(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_End) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_End) | 0);  /* check that XKB_KEY_Hangul_End is an integer */
  if (!_cffi_check_int(*o, n, 65331U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Hanja(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Hanja) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Hanja) | 0);  /* check that XKB_KEY_Hangul_Hanja is an integer */
  if (!_cffi_check_int(*o, n, 65332U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Hieuh(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Hieuh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Hieuh) | 0);  /* check that XKB_KEY_Hangul_Hieuh is an integer */
  if (!_cffi_check_int(*o, n, 3774U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_I(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_I) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_I) | 0);  /* check that XKB_KEY_Hangul_I is an integer */
  if (!_cffi_check_int(*o, n, 3795U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Ieung(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Ieung) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Ieung) | 0);  /* check that XKB_KEY_Hangul_Ieung is an integer */
  if (!_cffi_check_int(*o, n, 3767U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Cieuc(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Cieuc) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Cieuc) | 0);  /* check that XKB_KEY_Hangul_J_Cieuc is an integer */
  if (!_cffi_check_int(*o, n, 3818U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Dikeud(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Dikeud) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Dikeud) | 0);  /* check that XKB_KEY_Hangul_J_Dikeud is an integer */
  if (!_cffi_check_int(*o, n, 3802U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Hieuh(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Hieuh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Hieuh) | 0);  /* check that XKB_KEY_Hangul_J_Hieuh is an integer */
  if (!_cffi_check_int(*o, n, 3822U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Ieung(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Ieung) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Ieung) | 0);  /* check that XKB_KEY_Hangul_J_Ieung is an integer */
  if (!_cffi_check_int(*o, n, 3816U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Jieuj(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Jieuj) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Jieuj) | 0);  /* check that XKB_KEY_Hangul_J_Jieuj is an integer */
  if (!_cffi_check_int(*o, n, 3817U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Khieuq(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Khieuq) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Khieuq) | 0);  /* check that XKB_KEY_Hangul_J_Khieuq is an integer */
  if (!_cffi_check_int(*o, n, 3819U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Kiyeog(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Kiyeog) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Kiyeog) | 0);  /* check that XKB_KEY_Hangul_J_Kiyeog is an integer */
  if (!_cffi_check_int(*o, n, 3796U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_KiyeogSios(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_KiyeogSios) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_KiyeogSios) | 0);  /* check that XKB_KEY_Hangul_J_KiyeogSios is an integer */
  if (!_cffi_check_int(*o, n, 3798U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_KkogjiDalrinIeung(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_KkogjiDalrinIeung) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_KkogjiDalrinIeung) | 0);  /* check that XKB_KEY_Hangul_J_KkogjiDalrinIeung is an integer */
  if (!_cffi_check_int(*o, n, 3833U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Mieum(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Mieum) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Mieum) | 0);  /* check that XKB_KEY_Hangul_J_Mieum is an integer */
  if (!_cffi_check_int(*o, n, 3811U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Nieun(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Nieun) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Nieun) | 0);  /* check that XKB_KEY_Hangul_J_Nieun is an integer */
  if (!_cffi_check_int(*o, n, 3799U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_NieunHieuh(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_NieunHieuh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_NieunHieuh) | 0);  /* check that XKB_KEY_Hangul_J_NieunHieuh is an integer */
  if (!_cffi_check_int(*o, n, 3801U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_NieunJieuj(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_NieunJieuj) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_NieunJieuj) | 0);  /* check that XKB_KEY_Hangul_J_NieunJieuj is an integer */
  if (!_cffi_check_int(*o, n, 3800U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_PanSios(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_PanSios) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_PanSios) | 0);  /* check that XKB_KEY_Hangul_J_PanSios is an integer */
  if (!_cffi_check_int(*o, n, 3832U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Phieuf(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Phieuf) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Phieuf) | 0);  /* check that XKB_KEY_Hangul_J_Phieuf is an integer */
  if (!_cffi_check_int(*o, n, 3821U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Pieub(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Pieub) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Pieub) | 0);  /* check that XKB_KEY_Hangul_J_Pieub is an integer */
  if (!_cffi_check_int(*o, n, 3812U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_PieubSios(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_PieubSios) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_PieubSios) | 0);  /* check that XKB_KEY_Hangul_J_PieubSios is an integer */
  if (!_cffi_check_int(*o, n, 3813U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Rieul(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Rieul) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Rieul) | 0);  /* check that XKB_KEY_Hangul_J_Rieul is an integer */
  if (!_cffi_check_int(*o, n, 3803U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_RieulHieuh(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_RieulHieuh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_RieulHieuh) | 0);  /* check that XKB_KEY_Hangul_J_RieulHieuh is an integer */
  if (!_cffi_check_int(*o, n, 3810U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_RieulKiyeog(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_RieulKiyeog) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_RieulKiyeog) | 0);  /* check that XKB_KEY_Hangul_J_RieulKiyeog is an integer */
  if (!_cffi_check_int(*o, n, 3804U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_RieulMieum(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_RieulMieum) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_RieulMieum) | 0);  /* check that XKB_KEY_Hangul_J_RieulMieum is an integer */
  if (!_cffi_check_int(*o, n, 3805U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_RieulPhieuf(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_RieulPhieuf) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_RieulPhieuf) | 0);  /* check that XKB_KEY_Hangul_J_RieulPhieuf is an integer */
  if (!_cffi_check_int(*o, n, 3809U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_RieulPieub(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_RieulPieub) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_RieulPieub) | 0);  /* check that XKB_KEY_Hangul_J_RieulPieub is an integer */
  if (!_cffi_check_int(*o, n, 3806U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_RieulSios(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_RieulSios) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_RieulSios) | 0);  /* check that XKB_KEY_Hangul_J_RieulSios is an integer */
  if (!_cffi_check_int(*o, n, 3807U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_RieulTieut(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_RieulTieut) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_RieulTieut) | 0);  /* check that XKB_KEY_Hangul_J_RieulTieut is an integer */
  if (!_cffi_check_int(*o, n, 3808U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Sios(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Sios) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Sios) | 0);  /* check that XKB_KEY_Hangul_J_Sios is an integer */
  if (!_cffi_check_int(*o, n, 3814U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_SsangKiyeog(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_SsangKiyeog) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_SsangKiyeog) | 0);  /* check that XKB_KEY_Hangul_J_SsangKiyeog is an integer */
  if (!_cffi_check_int(*o, n, 3797U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_SsangSios(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_SsangSios) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_SsangSios) | 0);  /* check that XKB_KEY_Hangul_J_SsangSios is an integer */
  if (!_cffi_check_int(*o, n, 3815U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_Tieut(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_Tieut) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_Tieut) | 0);  /* check that XKB_KEY_Hangul_J_Tieut is an integer */
  if (!_cffi_check_int(*o, n, 3820U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_J_YeorinHieuh(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_J_YeorinHieuh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_J_YeorinHieuh) | 0);  /* check that XKB_KEY_Hangul_J_YeorinHieuh is an integer */
  if (!_cffi_check_int(*o, n, 3834U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Jamo(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Jamo) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Jamo) | 0);  /* check that XKB_KEY_Hangul_Jamo is an integer */
  if (!_cffi_check_int(*o, n, 65333U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Jeonja(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Jeonja) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Jeonja) | 0);  /* check that XKB_KEY_Hangul_Jeonja is an integer */
  if (!_cffi_check_int(*o, n, 65336U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Jieuj(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Jieuj) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Jieuj) | 0);  /* check that XKB_KEY_Hangul_Jieuj is an integer */
  if (!_cffi_check_int(*o, n, 3768U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Khieuq(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Khieuq) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Khieuq) | 0);  /* check that XKB_KEY_Hangul_Khieuq is an integer */
  if (!_cffi_check_int(*o, n, 3771U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Kiyeog(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Kiyeog) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Kiyeog) | 0);  /* check that XKB_KEY_Hangul_Kiyeog is an integer */
  if (!_cffi_check_int(*o, n, 3745U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_KiyeogSios(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_KiyeogSios) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_KiyeogSios) | 0);  /* check that XKB_KEY_Hangul_KiyeogSios is an integer */
  if (!_cffi_check_int(*o, n, 3747U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_KkogjiDalrinIeung(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_KkogjiDalrinIeung) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_KkogjiDalrinIeung) | 0);  /* check that XKB_KEY_Hangul_KkogjiDalrinIeung is an integer */
  if (!_cffi_check_int(*o, n, 3827U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Mieum(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Mieum) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Mieum) | 0);  /* check that XKB_KEY_Hangul_Mieum is an integer */
  if (!_cffi_check_int(*o, n, 3761U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_MultipleCandidate(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_MultipleCandidate) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_MultipleCandidate) | 0);  /* check that XKB_KEY_Hangul_MultipleCandidate is an integer */
  if (!_cffi_check_int(*o, n, 65341U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Nieun(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Nieun) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Nieun) | 0);  /* check that XKB_KEY_Hangul_Nieun is an integer */
  if (!_cffi_check_int(*o, n, 3748U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_NieunHieuh(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_NieunHieuh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_NieunHieuh) | 0);  /* check that XKB_KEY_Hangul_NieunHieuh is an integer */
  if (!_cffi_check_int(*o, n, 3750U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_NieunJieuj(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_NieunJieuj) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_NieunJieuj) | 0);  /* check that XKB_KEY_Hangul_NieunJieuj is an integer */
  if (!_cffi_check_int(*o, n, 3749U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_O(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_O) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_O) | 0);  /* check that XKB_KEY_Hangul_O is an integer */
  if (!_cffi_check_int(*o, n, 3783U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_OE(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_OE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_OE) | 0);  /* check that XKB_KEY_Hangul_OE is an integer */
  if (!_cffi_check_int(*o, n, 3786U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_PanSios(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_PanSios) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_PanSios) | 0);  /* check that XKB_KEY_Hangul_PanSios is an integer */
  if (!_cffi_check_int(*o, n, 3826U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Phieuf(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Phieuf) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Phieuf) | 0);  /* check that XKB_KEY_Hangul_Phieuf is an integer */
  if (!_cffi_check_int(*o, n, 3773U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Pieub(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Pieub) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Pieub) | 0);  /* check that XKB_KEY_Hangul_Pieub is an integer */
  if (!_cffi_check_int(*o, n, 3762U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_PieubSios(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_PieubSios) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_PieubSios) | 0);  /* check that XKB_KEY_Hangul_PieubSios is an integer */
  if (!_cffi_check_int(*o, n, 3764U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_PostHanja(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_PostHanja) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_PostHanja) | 0);  /* check that XKB_KEY_Hangul_PostHanja is an integer */
  if (!_cffi_check_int(*o, n, 65339U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_PreHanja(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_PreHanja) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_PreHanja) | 0);  /* check that XKB_KEY_Hangul_PreHanja is an integer */
  if (!_cffi_check_int(*o, n, 65338U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_PreviousCandidate(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_PreviousCandidate) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_PreviousCandidate) | 0);  /* check that XKB_KEY_Hangul_PreviousCandidate is an integer */
  if (!_cffi_check_int(*o, n, 65342U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Rieul(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Rieul) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Rieul) | 0);  /* check that XKB_KEY_Hangul_Rieul is an integer */
  if (!_cffi_check_int(*o, n, 3753U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_RieulHieuh(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_RieulHieuh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_RieulHieuh) | 0);  /* check that XKB_KEY_Hangul_RieulHieuh is an integer */
  if (!_cffi_check_int(*o, n, 3760U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_RieulKiyeog(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_RieulKiyeog) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_RieulKiyeog) | 0);  /* check that XKB_KEY_Hangul_RieulKiyeog is an integer */
  if (!_cffi_check_int(*o, n, 3754U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_RieulMieum(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_RieulMieum) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_RieulMieum) | 0);  /* check that XKB_KEY_Hangul_RieulMieum is an integer */
  if (!_cffi_check_int(*o, n, 3755U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_RieulPhieuf(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_RieulPhieuf) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_RieulPhieuf) | 0);  /* check that XKB_KEY_Hangul_RieulPhieuf is an integer */
  if (!_cffi_check_int(*o, n, 3759U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_RieulPieub(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_RieulPieub) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_RieulPieub) | 0);  /* check that XKB_KEY_Hangul_RieulPieub is an integer */
  if (!_cffi_check_int(*o, n, 3756U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_RieulSios(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_RieulSios) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_RieulSios) | 0);  /* check that XKB_KEY_Hangul_RieulSios is an integer */
  if (!_cffi_check_int(*o, n, 3757U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_RieulTieut(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_RieulTieut) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_RieulTieut) | 0);  /* check that XKB_KEY_Hangul_RieulTieut is an integer */
  if (!_cffi_check_int(*o, n, 3758U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_RieulYeorinHieuh(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_RieulYeorinHieuh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_RieulYeorinHieuh) | 0);  /* check that XKB_KEY_Hangul_RieulYeorinHieuh is an integer */
  if (!_cffi_check_int(*o, n, 3823U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Romaja(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Romaja) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Romaja) | 0);  /* check that XKB_KEY_Hangul_Romaja is an integer */
  if (!_cffi_check_int(*o, n, 65334U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_SingleCandidate(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_SingleCandidate) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_SingleCandidate) | 0);  /* check that XKB_KEY_Hangul_SingleCandidate is an integer */
  if (!_cffi_check_int(*o, n, 65340U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Sios(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Sios) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Sios) | 0);  /* check that XKB_KEY_Hangul_Sios is an integer */
  if (!_cffi_check_int(*o, n, 3765U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Special(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Special) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Special) | 0);  /* check that XKB_KEY_Hangul_Special is an integer */
  if (!_cffi_check_int(*o, n, 65343U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_SsangDikeud(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_SsangDikeud) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_SsangDikeud) | 0);  /* check that XKB_KEY_Hangul_SsangDikeud is an integer */
  if (!_cffi_check_int(*o, n, 3752U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_SsangJieuj(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_SsangJieuj) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_SsangJieuj) | 0);  /* check that XKB_KEY_Hangul_SsangJieuj is an integer */
  if (!_cffi_check_int(*o, n, 3769U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_SsangKiyeog(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_SsangKiyeog) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_SsangKiyeog) | 0);  /* check that XKB_KEY_Hangul_SsangKiyeog is an integer */
  if (!_cffi_check_int(*o, n, 3746U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_SsangPieub(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_SsangPieub) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_SsangPieub) | 0);  /* check that XKB_KEY_Hangul_SsangPieub is an integer */
  if (!_cffi_check_int(*o, n, 3763U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_SsangSios(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_SsangSios) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_SsangSios) | 0);  /* check that XKB_KEY_Hangul_SsangSios is an integer */
  if (!_cffi_check_int(*o, n, 3766U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Start(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Start) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Start) | 0);  /* check that XKB_KEY_Hangul_Start is an integer */
  if (!_cffi_check_int(*o, n, 65330U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_SunkyeongeumMieum(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_SunkyeongeumMieum) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_SunkyeongeumMieum) | 0);  /* check that XKB_KEY_Hangul_SunkyeongeumMieum is an integer */
  if (!_cffi_check_int(*o, n, 3824U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_SunkyeongeumPhieuf(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_SunkyeongeumPhieuf) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_SunkyeongeumPhieuf) | 0);  /* check that XKB_KEY_Hangul_SunkyeongeumPhieuf is an integer */
  if (!_cffi_check_int(*o, n, 3828U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_SunkyeongeumPieub(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_SunkyeongeumPieub) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_SunkyeongeumPieub) | 0);  /* check that XKB_KEY_Hangul_SunkyeongeumPieub is an integer */
  if (!_cffi_check_int(*o, n, 3825U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_Tieut(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_Tieut) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_Tieut) | 0);  /* check that XKB_KEY_Hangul_Tieut is an integer */
  if (!_cffi_check_int(*o, n, 3772U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_U(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_U) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_U) | 0);  /* check that XKB_KEY_Hangul_U is an integer */
  if (!_cffi_check_int(*o, n, 3788U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_WA(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_WA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_WA) | 0);  /* check that XKB_KEY_Hangul_WA is an integer */
  if (!_cffi_check_int(*o, n, 3784U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_WAE(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_WAE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_WAE) | 0);  /* check that XKB_KEY_Hangul_WAE is an integer */
  if (!_cffi_check_int(*o, n, 3785U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_WE(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_WE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_WE) | 0);  /* check that XKB_KEY_Hangul_WE is an integer */
  if (!_cffi_check_int(*o, n, 3790U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_WEO(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_WEO) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_WEO) | 0);  /* check that XKB_KEY_Hangul_WEO is an integer */
  if (!_cffi_check_int(*o, n, 3789U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_WI(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_WI) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_WI) | 0);  /* check that XKB_KEY_Hangul_WI is an integer */
  if (!_cffi_check_int(*o, n, 3791U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_YA(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_YA) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_YA) | 0);  /* check that XKB_KEY_Hangul_YA is an integer */
  if (!_cffi_check_int(*o, n, 3777U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_YAE(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_YAE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_YAE) | 0);  /* check that XKB_KEY_Hangul_YAE is an integer */
  if (!_cffi_check_int(*o, n, 3778U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_YE(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_YE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_YE) | 0);  /* check that XKB_KEY_Hangul_YE is an integer */
  if (!_cffi_check_int(*o, n, 3782U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_YEO(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_YEO) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_YEO) | 0);  /* check that XKB_KEY_Hangul_YEO is an integer */
  if (!_cffi_check_int(*o, n, 3781U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_YI(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_YI) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_YI) | 0);  /* check that XKB_KEY_Hangul_YI is an integer */
  if (!_cffi_check_int(*o, n, 3794U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_YO(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_YO) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_YO) | 0);  /* check that XKB_KEY_Hangul_YO is an integer */
  if (!_cffi_check_int(*o, n, 3787U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_YU(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_YU) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_YU) | 0);  /* check that XKB_KEY_Hangul_YU is an integer */
  if (!_cffi_check_int(*o, n, 3792U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_YeorinHieuh(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_YeorinHieuh) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_YeorinHieuh) | 0);  /* check that XKB_KEY_Hangul_YeorinHieuh is an integer */
  if (!_cffi_check_int(*o, n, 3829U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hangul_switch(unsigned long long *o)
{
  int n = (XKB_KEY_Hangul_switch) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hangul_switch) | 0);  /* check that XKB_KEY_Hangul_switch is an integer */
  if (!_cffi_check_int(*o, n, 65406U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hankaku(unsigned long long *o)
{
  int n = (XKB_KEY_Hankaku) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hankaku) | 0);  /* check that XKB_KEY_Hankaku is an integer */
  if (!_cffi_check_int(*o, n, 65321U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hcircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_Hcircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hcircumflex) | 0);  /* check that XKB_KEY_Hcircumflex is an integer */
  if (!_cffi_check_int(*o, n, 678U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hebrew_switch(unsigned long long *o)
{
  int n = (XKB_KEY_Hebrew_switch) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hebrew_switch) | 0);  /* check that XKB_KEY_Hebrew_switch is an integer */
  if (!_cffi_check_int(*o, n, 65406U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Help(unsigned long long *o)
{
  int n = (XKB_KEY_Help) <= 0;
  *o = (unsigned long long)((XKB_KEY_Help) | 0);  /* check that XKB_KEY_Help is an integer */
  if (!_cffi_check_int(*o, n, 65386U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Henkan(unsigned long long *o)
{
  int n = (XKB_KEY_Henkan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Henkan) | 0);  /* check that XKB_KEY_Henkan is an integer */
  if (!_cffi_check_int(*o, n, 65315U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Henkan_Mode(unsigned long long *o)
{
  int n = (XKB_KEY_Henkan_Mode) <= 0;
  *o = (unsigned long long)((XKB_KEY_Henkan_Mode) | 0);  /* check that XKB_KEY_Henkan_Mode is an integer */
  if (!_cffi_check_int(*o, n, 65315U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hiragana(unsigned long long *o)
{
  int n = (XKB_KEY_Hiragana) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hiragana) | 0);  /* check that XKB_KEY_Hiragana is an integer */
  if (!_cffi_check_int(*o, n, 65317U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hiragana_Katakana(unsigned long long *o)
{
  int n = (XKB_KEY_Hiragana_Katakana) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hiragana_Katakana) | 0);  /* check that XKB_KEY_Hiragana_Katakana is an integer */
  if (!_cffi_check_int(*o, n, 65319U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Home(unsigned long long *o)
{
  int n = (XKB_KEY_Home) <= 0;
  *o = (unsigned long long)((XKB_KEY_Home) | 0);  /* check that XKB_KEY_Home is an integer */
  if (!_cffi_check_int(*o, n, 65360U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hstroke(unsigned long long *o)
{
  int n = (XKB_KEY_Hstroke) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hstroke) | 0);  /* check that XKB_KEY_Hstroke is an integer */
  if (!_cffi_check_int(*o, n, 673U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hyper_L(unsigned long long *o)
{
  int n = (XKB_KEY_Hyper_L) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hyper_L) | 0);  /* check that XKB_KEY_Hyper_L is an integer */
  if (!_cffi_check_int(*o, n, 65517U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Hyper_R(unsigned long long *o)
{
  int n = (XKB_KEY_Hyper_R) <= 0;
  *o = (unsigned long long)((XKB_KEY_Hyper_R) | 0);  /* check that XKB_KEY_Hyper_R is an integer */
  if (!_cffi_check_int(*o, n, 65518U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_I(unsigned long long *o)
{
  int n = (XKB_KEY_I) <= 0;
  *o = (unsigned long long)((XKB_KEY_I) | 0);  /* check that XKB_KEY_I is an integer */
  if (!_cffi_check_int(*o, n, 73U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_IO(unsigned long long *o)
{
  int n = (XKB_KEY_IO) <= 0;
  *o = (unsigned long long)((XKB_KEY_IO) | 0);  /* check that XKB_KEY_IO is an integer */
  if (!_cffi_check_int(*o, n, 268435694U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Center_Object(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Center_Object) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Center_Object) | 0);  /* check that XKB_KEY_ISO_Center_Object is an integer */
  if (!_cffi_check_int(*o, n, 65075U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Continuous_Underline(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Continuous_Underline) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Continuous_Underline) | 0);  /* check that XKB_KEY_ISO_Continuous_Underline is an integer */
  if (!_cffi_check_int(*o, n, 65072U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Discontinuous_Underline(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Discontinuous_Underline) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Discontinuous_Underline) | 0);  /* check that XKB_KEY_ISO_Discontinuous_Underline is an integer */
  if (!_cffi_check_int(*o, n, 65073U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Emphasize(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Emphasize) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Emphasize) | 0);  /* check that XKB_KEY_ISO_Emphasize is an integer */
  if (!_cffi_check_int(*o, n, 65074U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Enter(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Enter) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Enter) | 0);  /* check that XKB_KEY_ISO_Enter is an integer */
  if (!_cffi_check_int(*o, n, 65076U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Fast_Cursor_Down(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Fast_Cursor_Down) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Fast_Cursor_Down) | 0);  /* check that XKB_KEY_ISO_Fast_Cursor_Down is an integer */
  if (!_cffi_check_int(*o, n, 65071U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Fast_Cursor_Left(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Fast_Cursor_Left) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Fast_Cursor_Left) | 0);  /* check that XKB_KEY_ISO_Fast_Cursor_Left is an integer */
  if (!_cffi_check_int(*o, n, 65068U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Fast_Cursor_Right(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Fast_Cursor_Right) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Fast_Cursor_Right) | 0);  /* check that XKB_KEY_ISO_Fast_Cursor_Right is an integer */
  if (!_cffi_check_int(*o, n, 65069U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Fast_Cursor_Up(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Fast_Cursor_Up) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Fast_Cursor_Up) | 0);  /* check that XKB_KEY_ISO_Fast_Cursor_Up is an integer */
  if (!_cffi_check_int(*o, n, 65070U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_First_Group(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_First_Group) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_First_Group) | 0);  /* check that XKB_KEY_ISO_First_Group is an integer */
  if (!_cffi_check_int(*o, n, 65036U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_First_Group_Lock(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_First_Group_Lock) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_First_Group_Lock) | 0);  /* check that XKB_KEY_ISO_First_Group_Lock is an integer */
  if (!_cffi_check_int(*o, n, 65037U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Group_Latch(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Group_Latch) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Group_Latch) | 0);  /* check that XKB_KEY_ISO_Group_Latch is an integer */
  if (!_cffi_check_int(*o, n, 65030U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Group_Lock(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Group_Lock) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Group_Lock) | 0);  /* check that XKB_KEY_ISO_Group_Lock is an integer */
  if (!_cffi_check_int(*o, n, 65031U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Group_Shift(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Group_Shift) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Group_Shift) | 0);  /* check that XKB_KEY_ISO_Group_Shift is an integer */
  if (!_cffi_check_int(*o, n, 65406U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Last_Group(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Last_Group) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Last_Group) | 0);  /* check that XKB_KEY_ISO_Last_Group is an integer */
  if (!_cffi_check_int(*o, n, 65038U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Last_Group_Lock(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Last_Group_Lock) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Last_Group_Lock) | 0);  /* check that XKB_KEY_ISO_Last_Group_Lock is an integer */
  if (!_cffi_check_int(*o, n, 65039U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Left_Tab(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Left_Tab) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Left_Tab) | 0);  /* check that XKB_KEY_ISO_Left_Tab is an integer */
  if (!_cffi_check_int(*o, n, 65056U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Level2_Latch(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Level2_Latch) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Level2_Latch) | 0);  /* check that XKB_KEY_ISO_Level2_Latch is an integer */
  if (!_cffi_check_int(*o, n, 65026U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Level3_Latch(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Level3_Latch) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Level3_Latch) | 0);  /* check that XKB_KEY_ISO_Level3_Latch is an integer */
  if (!_cffi_check_int(*o, n, 65028U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Level3_Lock(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Level3_Lock) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Level3_Lock) | 0);  /* check that XKB_KEY_ISO_Level3_Lock is an integer */
  if (!_cffi_check_int(*o, n, 65029U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Level3_Shift(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Level3_Shift) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Level3_Shift) | 0);  /* check that XKB_KEY_ISO_Level3_Shift is an integer */
  if (!_cffi_check_int(*o, n, 65027U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Level5_Latch(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Level5_Latch) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Level5_Latch) | 0);  /* check that XKB_KEY_ISO_Level5_Latch is an integer */
  if (!_cffi_check_int(*o, n, 65042U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Level5_Lock(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Level5_Lock) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Level5_Lock) | 0);  /* check that XKB_KEY_ISO_Level5_Lock is an integer */
  if (!_cffi_check_int(*o, n, 65043U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Level5_Shift(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Level5_Shift) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Level5_Shift) | 0);  /* check that XKB_KEY_ISO_Level5_Shift is an integer */
  if (!_cffi_check_int(*o, n, 65041U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Lock(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Lock) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Lock) | 0);  /* check that XKB_KEY_ISO_Lock is an integer */
  if (!_cffi_check_int(*o, n, 65025U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Move_Line_Down(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Move_Line_Down) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Move_Line_Down) | 0);  /* check that XKB_KEY_ISO_Move_Line_Down is an integer */
  if (!_cffi_check_int(*o, n, 65058U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Move_Line_Up(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Move_Line_Up) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Move_Line_Up) | 0);  /* check that XKB_KEY_ISO_Move_Line_Up is an integer */
  if (!_cffi_check_int(*o, n, 65057U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Next_Group(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Next_Group) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Next_Group) | 0);  /* check that XKB_KEY_ISO_Next_Group is an integer */
  if (!_cffi_check_int(*o, n, 65032U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Next_Group_Lock(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Next_Group_Lock) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Next_Group_Lock) | 0);  /* check that XKB_KEY_ISO_Next_Group_Lock is an integer */
  if (!_cffi_check_int(*o, n, 65033U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Partial_Line_Down(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Partial_Line_Down) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Partial_Line_Down) | 0);  /* check that XKB_KEY_ISO_Partial_Line_Down is an integer */
  if (!_cffi_check_int(*o, n, 65060U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Partial_Line_Up(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Partial_Line_Up) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Partial_Line_Up) | 0);  /* check that XKB_KEY_ISO_Partial_Line_Up is an integer */
  if (!_cffi_check_int(*o, n, 65059U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Partial_Space_Left(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Partial_Space_Left) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Partial_Space_Left) | 0);  /* check that XKB_KEY_ISO_Partial_Space_Left is an integer */
  if (!_cffi_check_int(*o, n, 65061U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Partial_Space_Right(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Partial_Space_Right) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Partial_Space_Right) | 0);  /* check that XKB_KEY_ISO_Partial_Space_Right is an integer */
  if (!_cffi_check_int(*o, n, 65062U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Prev_Group(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Prev_Group) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Prev_Group) | 0);  /* check that XKB_KEY_ISO_Prev_Group is an integer */
  if (!_cffi_check_int(*o, n, 65034U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Prev_Group_Lock(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Prev_Group_Lock) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Prev_Group_Lock) | 0);  /* check that XKB_KEY_ISO_Prev_Group_Lock is an integer */
  if (!_cffi_check_int(*o, n, 65035U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Release_Both_Margins(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Release_Both_Margins) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Release_Both_Margins) | 0);  /* check that XKB_KEY_ISO_Release_Both_Margins is an integer */
  if (!_cffi_check_int(*o, n, 65067U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Release_Margin_Left(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Release_Margin_Left) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Release_Margin_Left) | 0);  /* check that XKB_KEY_ISO_Release_Margin_Left is an integer */
  if (!_cffi_check_int(*o, n, 65065U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Release_Margin_Right(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Release_Margin_Right) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Release_Margin_Right) | 0);  /* check that XKB_KEY_ISO_Release_Margin_Right is an integer */
  if (!_cffi_check_int(*o, n, 65066U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Set_Margin_Left(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Set_Margin_Left) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Set_Margin_Left) | 0);  /* check that XKB_KEY_ISO_Set_Margin_Left is an integer */
  if (!_cffi_check_int(*o, n, 65063U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ISO_Set_Margin_Right(unsigned long long *o)
{
  int n = (XKB_KEY_ISO_Set_Margin_Right) <= 0;
  *o = (unsigned long long)((XKB_KEY_ISO_Set_Margin_Right) | 0);  /* check that XKB_KEY_ISO_Set_Margin_Right is an integer */
  if (!_cffi_check_int(*o, n, 65064U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Iabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_Iabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Iabovedot) | 0);  /* check that XKB_KEY_Iabovedot is an integer */
  if (!_cffi_check_int(*o, n, 681U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Iacute(unsigned long long *o)
{
  int n = (XKB_KEY_Iacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Iacute) | 0);  /* check that XKB_KEY_Iacute is an integer */
  if (!_cffi_check_int(*o, n, 205U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ibelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_Ibelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ibelowdot) | 0);  /* check that XKB_KEY_Ibelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785098U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ibreve(unsigned long long *o)
{
  int n = (XKB_KEY_Ibreve) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ibreve) | 0);  /* check that XKB_KEY_Ibreve is an integer */
  if (!_cffi_check_int(*o, n, 16777516U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Icircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_Icircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_Icircumflex) | 0);  /* check that XKB_KEY_Icircumflex is an integer */
  if (!_cffi_check_int(*o, n, 206U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Idiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_Idiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Idiaeresis) | 0);  /* check that XKB_KEY_Idiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 207U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Igrave(unsigned long long *o)
{
  int n = (XKB_KEY_Igrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_Igrave) | 0);  /* check that XKB_KEY_Igrave is an integer */
  if (!_cffi_check_int(*o, n, 204U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ihook(unsigned long long *o)
{
  int n = (XKB_KEY_Ihook) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ihook) | 0);  /* check that XKB_KEY_Ihook is an integer */
  if (!_cffi_check_int(*o, n, 16785096U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Imacron(unsigned long long *o)
{
  int n = (XKB_KEY_Imacron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Imacron) | 0);  /* check that XKB_KEY_Imacron is an integer */
  if (!_cffi_check_int(*o, n, 975U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Insert(unsigned long long *o)
{
  int n = (XKB_KEY_Insert) <= 0;
  *o = (unsigned long long)((XKB_KEY_Insert) | 0);  /* check that XKB_KEY_Insert is an integer */
  if (!_cffi_check_int(*o, n, 65379U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_InsertChar(unsigned long long *o)
{
  int n = (XKB_KEY_InsertChar) <= 0;
  *o = (unsigned long long)((XKB_KEY_InsertChar) | 0);  /* check that XKB_KEY_InsertChar is an integer */
  if (!_cffi_check_int(*o, n, 268500850U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_InsertLine(unsigned long long *o)
{
  int n = (XKB_KEY_InsertLine) <= 0;
  *o = (unsigned long long)((XKB_KEY_InsertLine) | 0);  /* check that XKB_KEY_InsertLine is an integer */
  if (!_cffi_check_int(*o, n, 268500848U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Iogonek(unsigned long long *o)
{
  int n = (XKB_KEY_Iogonek) <= 0;
  *o = (unsigned long long)((XKB_KEY_Iogonek) | 0);  /* check that XKB_KEY_Iogonek is an integer */
  if (!_cffi_check_int(*o, n, 967U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Itilde(unsigned long long *o)
{
  int n = (XKB_KEY_Itilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Itilde) | 0);  /* check that XKB_KEY_Itilde is an integer */
  if (!_cffi_check_int(*o, n, 933U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_J(unsigned long long *o)
{
  int n = (XKB_KEY_J) <= 0;
  *o = (unsigned long long)((XKB_KEY_J) | 0);  /* check that XKB_KEY_J is an integer */
  if (!_cffi_check_int(*o, n, 74U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Jcircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_Jcircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_Jcircumflex) | 0);  /* check that XKB_KEY_Jcircumflex is an integer */
  if (!_cffi_check_int(*o, n, 684U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_K(unsigned long long *o)
{
  int n = (XKB_KEY_K) <= 0;
  *o = (unsigned long long)((XKB_KEY_K) | 0);  /* check that XKB_KEY_K is an integer */
  if (!_cffi_check_int(*o, n, 75U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_0(unsigned long long *o)
{
  int n = (XKB_KEY_KP_0) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_0) | 0);  /* check that XKB_KEY_KP_0 is an integer */
  if (!_cffi_check_int(*o, n, 65456U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_1(unsigned long long *o)
{
  int n = (XKB_KEY_KP_1) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_1) | 0);  /* check that XKB_KEY_KP_1 is an integer */
  if (!_cffi_check_int(*o, n, 65457U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_2(unsigned long long *o)
{
  int n = (XKB_KEY_KP_2) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_2) | 0);  /* check that XKB_KEY_KP_2 is an integer */
  if (!_cffi_check_int(*o, n, 65458U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_3(unsigned long long *o)
{
  int n = (XKB_KEY_KP_3) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_3) | 0);  /* check that XKB_KEY_KP_3 is an integer */
  if (!_cffi_check_int(*o, n, 65459U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_4(unsigned long long *o)
{
  int n = (XKB_KEY_KP_4) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_4) | 0);  /* check that XKB_KEY_KP_4 is an integer */
  if (!_cffi_check_int(*o, n, 65460U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_5(unsigned long long *o)
{
  int n = (XKB_KEY_KP_5) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_5) | 0);  /* check that XKB_KEY_KP_5 is an integer */
  if (!_cffi_check_int(*o, n, 65461U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_6(unsigned long long *o)
{
  int n = (XKB_KEY_KP_6) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_6) | 0);  /* check that XKB_KEY_KP_6 is an integer */
  if (!_cffi_check_int(*o, n, 65462U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_7(unsigned long long *o)
{
  int n = (XKB_KEY_KP_7) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_7) | 0);  /* check that XKB_KEY_KP_7 is an integer */
  if (!_cffi_check_int(*o, n, 65463U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_8(unsigned long long *o)
{
  int n = (XKB_KEY_KP_8) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_8) | 0);  /* check that XKB_KEY_KP_8 is an integer */
  if (!_cffi_check_int(*o, n, 65464U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_9(unsigned long long *o)
{
  int n = (XKB_KEY_KP_9) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_9) | 0);  /* check that XKB_KEY_KP_9 is an integer */
  if (!_cffi_check_int(*o, n, 65465U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Add(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Add) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Add) | 0);  /* check that XKB_KEY_KP_Add is an integer */
  if (!_cffi_check_int(*o, n, 65451U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_BackTab(unsigned long long *o)
{
  int n = (XKB_KEY_KP_BackTab) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_BackTab) | 0);  /* check that XKB_KEY_KP_BackTab is an integer */
  if (!_cffi_check_int(*o, n, 268500853U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Begin(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Begin) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Begin) | 0);  /* check that XKB_KEY_KP_Begin is an integer */
  if (!_cffi_check_int(*o, n, 65437U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Decimal(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Decimal) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Decimal) | 0);  /* check that XKB_KEY_KP_Decimal is an integer */
  if (!_cffi_check_int(*o, n, 65454U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Delete(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Delete) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Delete) | 0);  /* check that XKB_KEY_KP_Delete is an integer */
  if (!_cffi_check_int(*o, n, 65439U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Divide(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Divide) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Divide) | 0);  /* check that XKB_KEY_KP_Divide is an integer */
  if (!_cffi_check_int(*o, n, 65455U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Down(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Down) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Down) | 0);  /* check that XKB_KEY_KP_Down is an integer */
  if (!_cffi_check_int(*o, n, 65433U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_End(unsigned long long *o)
{
  int n = (XKB_KEY_KP_End) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_End) | 0);  /* check that XKB_KEY_KP_End is an integer */
  if (!_cffi_check_int(*o, n, 65436U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Enter(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Enter) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Enter) | 0);  /* check that XKB_KEY_KP_Enter is an integer */
  if (!_cffi_check_int(*o, n, 65421U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Equal(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Equal) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Equal) | 0);  /* check that XKB_KEY_KP_Equal is an integer */
  if (!_cffi_check_int(*o, n, 65469U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_F1(unsigned long long *o)
{
  int n = (XKB_KEY_KP_F1) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_F1) | 0);  /* check that XKB_KEY_KP_F1 is an integer */
  if (!_cffi_check_int(*o, n, 65425U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_F2(unsigned long long *o)
{
  int n = (XKB_KEY_KP_F2) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_F2) | 0);  /* check that XKB_KEY_KP_F2 is an integer */
  if (!_cffi_check_int(*o, n, 65426U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_F3(unsigned long long *o)
{
  int n = (XKB_KEY_KP_F3) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_F3) | 0);  /* check that XKB_KEY_KP_F3 is an integer */
  if (!_cffi_check_int(*o, n, 65427U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_F4(unsigned long long *o)
{
  int n = (XKB_KEY_KP_F4) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_F4) | 0);  /* check that XKB_KEY_KP_F4 is an integer */
  if (!_cffi_check_int(*o, n, 65428U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Home(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Home) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Home) | 0);  /* check that XKB_KEY_KP_Home is an integer */
  if (!_cffi_check_int(*o, n, 65429U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Insert(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Insert) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Insert) | 0);  /* check that XKB_KEY_KP_Insert is an integer */
  if (!_cffi_check_int(*o, n, 65438U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Left(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Left) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Left) | 0);  /* check that XKB_KEY_KP_Left is an integer */
  if (!_cffi_check_int(*o, n, 65430U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Multiply(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Multiply) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Multiply) | 0);  /* check that XKB_KEY_KP_Multiply is an integer */
  if (!_cffi_check_int(*o, n, 65450U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Next(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Next) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Next) | 0);  /* check that XKB_KEY_KP_Next is an integer */
  if (!_cffi_check_int(*o, n, 65435U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Page_Down(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Page_Down) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Page_Down) | 0);  /* check that XKB_KEY_KP_Page_Down is an integer */
  if (!_cffi_check_int(*o, n, 65435U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Page_Up(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Page_Up) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Page_Up) | 0);  /* check that XKB_KEY_KP_Page_Up is an integer */
  if (!_cffi_check_int(*o, n, 65434U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Prior(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Prior) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Prior) | 0);  /* check that XKB_KEY_KP_Prior is an integer */
  if (!_cffi_check_int(*o, n, 65434U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Right(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Right) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Right) | 0);  /* check that XKB_KEY_KP_Right is an integer */
  if (!_cffi_check_int(*o, n, 65432U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Separator(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Separator) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Separator) | 0);  /* check that XKB_KEY_KP_Separator is an integer */
  if (!_cffi_check_int(*o, n, 65452U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Space(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Space) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Space) | 0);  /* check that XKB_KEY_KP_Space is an integer */
  if (!_cffi_check_int(*o, n, 65408U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Subtract(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Subtract) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Subtract) | 0);  /* check that XKB_KEY_KP_Subtract is an integer */
  if (!_cffi_check_int(*o, n, 65453U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Tab(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Tab) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Tab) | 0);  /* check that XKB_KEY_KP_Tab is an integer */
  if (!_cffi_check_int(*o, n, 65417U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_KP_Up(unsigned long long *o)
{
  int n = (XKB_KEY_KP_Up) <= 0;
  *o = (unsigned long long)((XKB_KEY_KP_Up) | 0);  /* check that XKB_KEY_KP_Up is an integer */
  if (!_cffi_check_int(*o, n, 65431U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Kana_Lock(unsigned long long *o)
{
  int n = (XKB_KEY_Kana_Lock) <= 0;
  *o = (unsigned long long)((XKB_KEY_Kana_Lock) | 0);  /* check that XKB_KEY_Kana_Lock is an integer */
  if (!_cffi_check_int(*o, n, 65325U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Kana_Shift(unsigned long long *o)
{
  int n = (XKB_KEY_Kana_Shift) <= 0;
  *o = (unsigned long long)((XKB_KEY_Kana_Shift) | 0);  /* check that XKB_KEY_Kana_Shift is an integer */
  if (!_cffi_check_int(*o, n, 65326U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Kanji(unsigned long long *o)
{
  int n = (XKB_KEY_Kanji) <= 0;
  *o = (unsigned long long)((XKB_KEY_Kanji) | 0);  /* check that XKB_KEY_Kanji is an integer */
  if (!_cffi_check_int(*o, n, 65313U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Kanji_Bangou(unsigned long long *o)
{
  int n = (XKB_KEY_Kanji_Bangou) <= 0;
  *o = (unsigned long long)((XKB_KEY_Kanji_Bangou) | 0);  /* check that XKB_KEY_Kanji_Bangou is an integer */
  if (!_cffi_check_int(*o, n, 65335U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Katakana(unsigned long long *o)
{
  int n = (XKB_KEY_Katakana) <= 0;
  *o = (unsigned long long)((XKB_KEY_Katakana) | 0);  /* check that XKB_KEY_Katakana is an integer */
  if (!_cffi_check_int(*o, n, 65318U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Kcedilla(unsigned long long *o)
{
  int n = (XKB_KEY_Kcedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_Kcedilla) | 0);  /* check that XKB_KEY_Kcedilla is an integer */
  if (!_cffi_check_int(*o, n, 979U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Korean_Won(unsigned long long *o)
{
  int n = (XKB_KEY_Korean_Won) <= 0;
  *o = (unsigned long long)((XKB_KEY_Korean_Won) | 0);  /* check that XKB_KEY_Korean_Won is an integer */
  if (!_cffi_check_int(*o, n, 3839U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_L(unsigned long long *o)
{
  int n = (XKB_KEY_L) <= 0;
  *o = (unsigned long long)((XKB_KEY_L) | 0);  /* check that XKB_KEY_L is an integer */
  if (!_cffi_check_int(*o, n, 76U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_L1(unsigned long long *o)
{
  int n = (XKB_KEY_L1) <= 0;
  *o = (unsigned long long)((XKB_KEY_L1) | 0);  /* check that XKB_KEY_L1 is an integer */
  if (!_cffi_check_int(*o, n, 65480U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_L10(unsigned long long *o)
{
  int n = (XKB_KEY_L10) <= 0;
  *o = (unsigned long long)((XKB_KEY_L10) | 0);  /* check that XKB_KEY_L10 is an integer */
  if (!_cffi_check_int(*o, n, 65489U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_L2(unsigned long long *o)
{
  int n = (XKB_KEY_L2) <= 0;
  *o = (unsigned long long)((XKB_KEY_L2) | 0);  /* check that XKB_KEY_L2 is an integer */
  if (!_cffi_check_int(*o, n, 65481U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_L3(unsigned long long *o)
{
  int n = (XKB_KEY_L3) <= 0;
  *o = (unsigned long long)((XKB_KEY_L3) | 0);  /* check that XKB_KEY_L3 is an integer */
  if (!_cffi_check_int(*o, n, 65482U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_L4(unsigned long long *o)
{
  int n = (XKB_KEY_L4) <= 0;
  *o = (unsigned long long)((XKB_KEY_L4) | 0);  /* check that XKB_KEY_L4 is an integer */
  if (!_cffi_check_int(*o, n, 65483U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_L5(unsigned long long *o)
{
  int n = (XKB_KEY_L5) <= 0;
  *o = (unsigned long long)((XKB_KEY_L5) | 0);  /* check that XKB_KEY_L5 is an integer */
  if (!_cffi_check_int(*o, n, 65484U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_L6(unsigned long long *o)
{
  int n = (XKB_KEY_L6) <= 0;
  *o = (unsigned long long)((XKB_KEY_L6) | 0);  /* check that XKB_KEY_L6 is an integer */
  if (!_cffi_check_int(*o, n, 65485U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_L7(unsigned long long *o)
{
  int n = (XKB_KEY_L7) <= 0;
  *o = (unsigned long long)((XKB_KEY_L7) | 0);  /* check that XKB_KEY_L7 is an integer */
  if (!_cffi_check_int(*o, n, 65486U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_L8(unsigned long long *o)
{
  int n = (XKB_KEY_L8) <= 0;
  *o = (unsigned long long)((XKB_KEY_L8) | 0);  /* check that XKB_KEY_L8 is an integer */
  if (!_cffi_check_int(*o, n, 65487U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_L9(unsigned long long *o)
{
  int n = (XKB_KEY_L9) <= 0;
  *o = (unsigned long long)((XKB_KEY_L9) | 0);  /* check that XKB_KEY_L9 is an integer */
  if (!_cffi_check_int(*o, n, 65488U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Lacute(unsigned long long *o)
{
  int n = (XKB_KEY_Lacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Lacute) | 0);  /* check that XKB_KEY_Lacute is an integer */
  if (!_cffi_check_int(*o, n, 453U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Last_Virtual_Screen(unsigned long long *o)
{
  int n = (XKB_KEY_Last_Virtual_Screen) <= 0;
  *o = (unsigned long long)((XKB_KEY_Last_Virtual_Screen) | 0);  /* check that XKB_KEY_Last_Virtual_Screen is an integer */
  if (!_cffi_check_int(*o, n, 65236U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Lbelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_Lbelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Lbelowdot) | 0);  /* check that XKB_KEY_Lbelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16784950U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Lcaron(unsigned long long *o)
{
  int n = (XKB_KEY_Lcaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Lcaron) | 0);  /* check that XKB_KEY_Lcaron is an integer */
  if (!_cffi_check_int(*o, n, 421U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Lcedilla(unsigned long long *o)
{
  int n = (XKB_KEY_Lcedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_Lcedilla) | 0);  /* check that XKB_KEY_Lcedilla is an integer */
  if (!_cffi_check_int(*o, n, 934U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Left(unsigned long long *o)
{
  int n = (XKB_KEY_Left) <= 0;
  *o = (unsigned long long)((XKB_KEY_Left) | 0);  /* check that XKB_KEY_Left is an integer */
  if (!_cffi_check_int(*o, n, 65361U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Linefeed(unsigned long long *o)
{
  int n = (XKB_KEY_Linefeed) <= 0;
  *o = (unsigned long long)((XKB_KEY_Linefeed) | 0);  /* check that XKB_KEY_Linefeed is an integer */
  if (!_cffi_check_int(*o, n, 65290U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_LiraSign(unsigned long long *o)
{
  int n = (XKB_KEY_LiraSign) <= 0;
  *o = (unsigned long long)((XKB_KEY_LiraSign) | 0);  /* check that XKB_KEY_LiraSign is an integer */
  if (!_cffi_check_int(*o, n, 16785572U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Lstroke(unsigned long long *o)
{
  int n = (XKB_KEY_Lstroke) <= 0;
  *o = (unsigned long long)((XKB_KEY_Lstroke) | 0);  /* check that XKB_KEY_Lstroke is an integer */
  if (!_cffi_check_int(*o, n, 419U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_M(unsigned long long *o)
{
  int n = (XKB_KEY_M) <= 0;
  *o = (unsigned long long)((XKB_KEY_M) | 0);  /* check that XKB_KEY_M is an integer */
  if (!_cffi_check_int(*o, n, 77U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Mabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_Mabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Mabovedot) | 0);  /* check that XKB_KEY_Mabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16784960U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Macedonia_DSE(unsigned long long *o)
{
  int n = (XKB_KEY_Macedonia_DSE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Macedonia_DSE) | 0);  /* check that XKB_KEY_Macedonia_DSE is an integer */
  if (!_cffi_check_int(*o, n, 1717U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Macedonia_GJE(unsigned long long *o)
{
  int n = (XKB_KEY_Macedonia_GJE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Macedonia_GJE) | 0);  /* check that XKB_KEY_Macedonia_GJE is an integer */
  if (!_cffi_check_int(*o, n, 1714U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Macedonia_KJE(unsigned long long *o)
{
  int n = (XKB_KEY_Macedonia_KJE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Macedonia_KJE) | 0);  /* check that XKB_KEY_Macedonia_KJE is an integer */
  if (!_cffi_check_int(*o, n, 1724U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Macedonia_dse(unsigned long long *o)
{
  int n = (XKB_KEY_Macedonia_dse) <= 0;
  *o = (unsigned long long)((XKB_KEY_Macedonia_dse) | 0);  /* check that XKB_KEY_Macedonia_dse is an integer */
  if (!_cffi_check_int(*o, n, 1701U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Macedonia_gje(unsigned long long *o)
{
  int n = (XKB_KEY_Macedonia_gje) <= 0;
  *o = (unsigned long long)((XKB_KEY_Macedonia_gje) | 0);  /* check that XKB_KEY_Macedonia_gje is an integer */
  if (!_cffi_check_int(*o, n, 1698U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Macedonia_kje(unsigned long long *o)
{
  int n = (XKB_KEY_Macedonia_kje) <= 0;
  *o = (unsigned long long)((XKB_KEY_Macedonia_kje) | 0);  /* check that XKB_KEY_Macedonia_kje is an integer */
  if (!_cffi_check_int(*o, n, 1708U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Mae_Koho(unsigned long long *o)
{
  int n = (XKB_KEY_Mae_Koho) <= 0;
  *o = (unsigned long long)((XKB_KEY_Mae_Koho) | 0);  /* check that XKB_KEY_Mae_Koho is an integer */
  if (!_cffi_check_int(*o, n, 65342U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Massyo(unsigned long long *o)
{
  int n = (XKB_KEY_Massyo) <= 0;
  *o = (unsigned long long)((XKB_KEY_Massyo) | 0);  /* check that XKB_KEY_Massyo is an integer */
  if (!_cffi_check_int(*o, n, 65324U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Menu(unsigned long long *o)
{
  int n = (XKB_KEY_Menu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Menu) | 0);  /* check that XKB_KEY_Menu is an integer */
  if (!_cffi_check_int(*o, n, 65383U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Meta_L(unsigned long long *o)
{
  int n = (XKB_KEY_Meta_L) <= 0;
  *o = (unsigned long long)((XKB_KEY_Meta_L) | 0);  /* check that XKB_KEY_Meta_L is an integer */
  if (!_cffi_check_int(*o, n, 65511U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Meta_R(unsigned long long *o)
{
  int n = (XKB_KEY_Meta_R) <= 0;
  *o = (unsigned long long)((XKB_KEY_Meta_R) | 0);  /* check that XKB_KEY_Meta_R is an integer */
  if (!_cffi_check_int(*o, n, 65512U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_MillSign(unsigned long long *o)
{
  int n = (XKB_KEY_MillSign) <= 0;
  *o = (unsigned long long)((XKB_KEY_MillSign) | 0);  /* check that XKB_KEY_MillSign is an integer */
  if (!_cffi_check_int(*o, n, 16785573U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Mode_switch(unsigned long long *o)
{
  int n = (XKB_KEY_Mode_switch) <= 0;
  *o = (unsigned long long)((XKB_KEY_Mode_switch) | 0);  /* check that XKB_KEY_Mode_switch is an integer */
  if (!_cffi_check_int(*o, n, 65406U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_MouseKeys_Accel_Enable(unsigned long long *o)
{
  int n = (XKB_KEY_MouseKeys_Accel_Enable) <= 0;
  *o = (unsigned long long)((XKB_KEY_MouseKeys_Accel_Enable) | 0);  /* check that XKB_KEY_MouseKeys_Accel_Enable is an integer */
  if (!_cffi_check_int(*o, n, 65143U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_MouseKeys_Enable(unsigned long long *o)
{
  int n = (XKB_KEY_MouseKeys_Enable) <= 0;
  *o = (unsigned long long)((XKB_KEY_MouseKeys_Enable) | 0);  /* check that XKB_KEY_MouseKeys_Enable is an integer */
  if (!_cffi_check_int(*o, n, 65142U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Muhenkan(unsigned long long *o)
{
  int n = (XKB_KEY_Muhenkan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Muhenkan) | 0);  /* check that XKB_KEY_Muhenkan is an integer */
  if (!_cffi_check_int(*o, n, 65314U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Multi_key(unsigned long long *o)
{
  int n = (XKB_KEY_Multi_key) <= 0;
  *o = (unsigned long long)((XKB_KEY_Multi_key) | 0);  /* check that XKB_KEY_Multi_key is an integer */
  if (!_cffi_check_int(*o, n, 65312U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_MultipleCandidate(unsigned long long *o)
{
  int n = (XKB_KEY_MultipleCandidate) <= 0;
  *o = (unsigned long long)((XKB_KEY_MultipleCandidate) | 0);  /* check that XKB_KEY_MultipleCandidate is an integer */
  if (!_cffi_check_int(*o, n, 65341U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_N(unsigned long long *o)
{
  int n = (XKB_KEY_N) <= 0;
  *o = (unsigned long long)((XKB_KEY_N) | 0);  /* check that XKB_KEY_N is an integer */
  if (!_cffi_check_int(*o, n, 78U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Nacute(unsigned long long *o)
{
  int n = (XKB_KEY_Nacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Nacute) | 0);  /* check that XKB_KEY_Nacute is an integer */
  if (!_cffi_check_int(*o, n, 465U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_NairaSign(unsigned long long *o)
{
  int n = (XKB_KEY_NairaSign) <= 0;
  *o = (unsigned long long)((XKB_KEY_NairaSign) | 0);  /* check that XKB_KEY_NairaSign is an integer */
  if (!_cffi_check_int(*o, n, 16785574U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ncaron(unsigned long long *o)
{
  int n = (XKB_KEY_Ncaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ncaron) | 0);  /* check that XKB_KEY_Ncaron is an integer */
  if (!_cffi_check_int(*o, n, 466U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ncedilla(unsigned long long *o)
{
  int n = (XKB_KEY_Ncedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ncedilla) | 0);  /* check that XKB_KEY_Ncedilla is an integer */
  if (!_cffi_check_int(*o, n, 977U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_NewSheqelSign(unsigned long long *o)
{
  int n = (XKB_KEY_NewSheqelSign) <= 0;
  *o = (unsigned long long)((XKB_KEY_NewSheqelSign) | 0);  /* check that XKB_KEY_NewSheqelSign is an integer */
  if (!_cffi_check_int(*o, n, 16785578U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Next(unsigned long long *o)
{
  int n = (XKB_KEY_Next) <= 0;
  *o = (unsigned long long)((XKB_KEY_Next) | 0);  /* check that XKB_KEY_Next is an integer */
  if (!_cffi_check_int(*o, n, 65366U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Next_Virtual_Screen(unsigned long long *o)
{
  int n = (XKB_KEY_Next_Virtual_Screen) <= 0;
  *o = (unsigned long long)((XKB_KEY_Next_Virtual_Screen) | 0);  /* check that XKB_KEY_Next_Virtual_Screen is an integer */
  if (!_cffi_check_int(*o, n, 65234U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_NoSymbol(unsigned long long *o)
{
  int n = (XKB_KEY_NoSymbol) <= 0;
  *o = (unsigned long long)((XKB_KEY_NoSymbol) | 0);  /* check that XKB_KEY_NoSymbol is an integer */
  if (!_cffi_check_int(*o, n, 0))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ntilde(unsigned long long *o)
{
  int n = (XKB_KEY_Ntilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ntilde) | 0);  /* check that XKB_KEY_Ntilde is an integer */
  if (!_cffi_check_int(*o, n, 209U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Num_Lock(unsigned long long *o)
{
  int n = (XKB_KEY_Num_Lock) <= 0;
  *o = (unsigned long long)((XKB_KEY_Num_Lock) | 0);  /* check that XKB_KEY_Num_Lock is an integer */
  if (!_cffi_check_int(*o, n, 65407U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_O(unsigned long long *o)
{
  int n = (XKB_KEY_O) <= 0;
  *o = (unsigned long long)((XKB_KEY_O) | 0);  /* check that XKB_KEY_O is an integer */
  if (!_cffi_check_int(*o, n, 79U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_OE(unsigned long long *o)
{
  int n = (XKB_KEY_OE) <= 0;
  *o = (unsigned long long)((XKB_KEY_OE) | 0);  /* check that XKB_KEY_OE is an integer */
  if (!_cffi_check_int(*o, n, 5052U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Oacute(unsigned long long *o)
{
  int n = (XKB_KEY_Oacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Oacute) | 0);  /* check that XKB_KEY_Oacute is an integer */
  if (!_cffi_check_int(*o, n, 211U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Obarred(unsigned long long *o)
{
  int n = (XKB_KEY_Obarred) <= 0;
  *o = (unsigned long long)((XKB_KEY_Obarred) | 0);  /* check that XKB_KEY_Obarred is an integer */
  if (!_cffi_check_int(*o, n, 16777631U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Obelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_Obelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Obelowdot) | 0);  /* check that XKB_KEY_Obelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785100U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ocaron(unsigned long long *o)
{
  int n = (XKB_KEY_Ocaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ocaron) | 0);  /* check that XKB_KEY_Ocaron is an integer */
  if (!_cffi_check_int(*o, n, 16777681U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ocircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_Ocircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ocircumflex) | 0);  /* check that XKB_KEY_Ocircumflex is an integer */
  if (!_cffi_check_int(*o, n, 212U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ocircumflexacute(unsigned long long *o)
{
  int n = (XKB_KEY_Ocircumflexacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ocircumflexacute) | 0);  /* check that XKB_KEY_Ocircumflexacute is an integer */
  if (!_cffi_check_int(*o, n, 16785104U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ocircumflexbelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_Ocircumflexbelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ocircumflexbelowdot) | 0);  /* check that XKB_KEY_Ocircumflexbelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785112U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ocircumflexgrave(unsigned long long *o)
{
  int n = (XKB_KEY_Ocircumflexgrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ocircumflexgrave) | 0);  /* check that XKB_KEY_Ocircumflexgrave is an integer */
  if (!_cffi_check_int(*o, n, 16785106U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ocircumflexhook(unsigned long long *o)
{
  int n = (XKB_KEY_Ocircumflexhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ocircumflexhook) | 0);  /* check that XKB_KEY_Ocircumflexhook is an integer */
  if (!_cffi_check_int(*o, n, 16785108U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ocircumflextilde(unsigned long long *o)
{
  int n = (XKB_KEY_Ocircumflextilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ocircumflextilde) | 0);  /* check that XKB_KEY_Ocircumflextilde is an integer */
  if (!_cffi_check_int(*o, n, 16785110U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Odiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_Odiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Odiaeresis) | 0);  /* check that XKB_KEY_Odiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 214U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Odoubleacute(unsigned long long *o)
{
  int n = (XKB_KEY_Odoubleacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Odoubleacute) | 0);  /* check that XKB_KEY_Odoubleacute is an integer */
  if (!_cffi_check_int(*o, n, 469U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ograve(unsigned long long *o)
{
  int n = (XKB_KEY_Ograve) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ograve) | 0);  /* check that XKB_KEY_Ograve is an integer */
  if (!_cffi_check_int(*o, n, 210U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ohook(unsigned long long *o)
{
  int n = (XKB_KEY_Ohook) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ohook) | 0);  /* check that XKB_KEY_Ohook is an integer */
  if (!_cffi_check_int(*o, n, 16785102U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ohorn(unsigned long long *o)
{
  int n = (XKB_KEY_Ohorn) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ohorn) | 0);  /* check that XKB_KEY_Ohorn is an integer */
  if (!_cffi_check_int(*o, n, 16777632U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ohornacute(unsigned long long *o)
{
  int n = (XKB_KEY_Ohornacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ohornacute) | 0);  /* check that XKB_KEY_Ohornacute is an integer */
  if (!_cffi_check_int(*o, n, 16785114U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ohornbelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_Ohornbelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ohornbelowdot) | 0);  /* check that XKB_KEY_Ohornbelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785122U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ohorngrave(unsigned long long *o)
{
  int n = (XKB_KEY_Ohorngrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ohorngrave) | 0);  /* check that XKB_KEY_Ohorngrave is an integer */
  if (!_cffi_check_int(*o, n, 16785116U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ohornhook(unsigned long long *o)
{
  int n = (XKB_KEY_Ohornhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ohornhook) | 0);  /* check that XKB_KEY_Ohornhook is an integer */
  if (!_cffi_check_int(*o, n, 16785118U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ohorntilde(unsigned long long *o)
{
  int n = (XKB_KEY_Ohorntilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ohorntilde) | 0);  /* check that XKB_KEY_Ohorntilde is an integer */
  if (!_cffi_check_int(*o, n, 16785120U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Omacron(unsigned long long *o)
{
  int n = (XKB_KEY_Omacron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Omacron) | 0);  /* check that XKB_KEY_Omacron is an integer */
  if (!_cffi_check_int(*o, n, 978U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ooblique(unsigned long long *o)
{
  int n = (XKB_KEY_Ooblique) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ooblique) | 0);  /* check that XKB_KEY_Ooblique is an integer */
  if (!_cffi_check_int(*o, n, 216U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Oslash(unsigned long long *o)
{
  int n = (XKB_KEY_Oslash) <= 0;
  *o = (unsigned long long)((XKB_KEY_Oslash) | 0);  /* check that XKB_KEY_Oslash is an integer */
  if (!_cffi_check_int(*o, n, 216U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Otilde(unsigned long long *o)
{
  int n = (XKB_KEY_Otilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Otilde) | 0);  /* check that XKB_KEY_Otilde is an integer */
  if (!_cffi_check_int(*o, n, 213U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Overlay1_Enable(unsigned long long *o)
{
  int n = (XKB_KEY_Overlay1_Enable) <= 0;
  *o = (unsigned long long)((XKB_KEY_Overlay1_Enable) | 0);  /* check that XKB_KEY_Overlay1_Enable is an integer */
  if (!_cffi_check_int(*o, n, 65144U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Overlay2_Enable(unsigned long long *o)
{
  int n = (XKB_KEY_Overlay2_Enable) <= 0;
  *o = (unsigned long long)((XKB_KEY_Overlay2_Enable) | 0);  /* check that XKB_KEY_Overlay2_Enable is an integer */
  if (!_cffi_check_int(*o, n, 65145U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_P(unsigned long long *o)
{
  int n = (XKB_KEY_P) <= 0;
  *o = (unsigned long long)((XKB_KEY_P) | 0);  /* check that XKB_KEY_P is an integer */
  if (!_cffi_check_int(*o, n, 80U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_Pabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pabovedot) | 0);  /* check that XKB_KEY_Pabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16784982U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Page_Down(unsigned long long *o)
{
  int n = (XKB_KEY_Page_Down) <= 0;
  *o = (unsigned long long)((XKB_KEY_Page_Down) | 0);  /* check that XKB_KEY_Page_Down is an integer */
  if (!_cffi_check_int(*o, n, 65366U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Page_Up(unsigned long long *o)
{
  int n = (XKB_KEY_Page_Up) <= 0;
  *o = (unsigned long long)((XKB_KEY_Page_Up) | 0);  /* check that XKB_KEY_Page_Up is an integer */
  if (!_cffi_check_int(*o, n, 65365U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pause(unsigned long long *o)
{
  int n = (XKB_KEY_Pause) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pause) | 0);  /* check that XKB_KEY_Pause is an integer */
  if (!_cffi_check_int(*o, n, 65299U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_PesetaSign(unsigned long long *o)
{
  int n = (XKB_KEY_PesetaSign) <= 0;
  *o = (unsigned long long)((XKB_KEY_PesetaSign) | 0);  /* check that XKB_KEY_PesetaSign is an integer */
  if (!_cffi_check_int(*o, n, 16785575U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Accelerate(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Accelerate) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Accelerate) | 0);  /* check that XKB_KEY_Pointer_Accelerate is an integer */
  if (!_cffi_check_int(*o, n, 65274U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Button1(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Button1) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Button1) | 0);  /* check that XKB_KEY_Pointer_Button1 is an integer */
  if (!_cffi_check_int(*o, n, 65257U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Button2(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Button2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Button2) | 0);  /* check that XKB_KEY_Pointer_Button2 is an integer */
  if (!_cffi_check_int(*o, n, 65258U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Button3(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Button3) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Button3) | 0);  /* check that XKB_KEY_Pointer_Button3 is an integer */
  if (!_cffi_check_int(*o, n, 65259U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Button4(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Button4) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Button4) | 0);  /* check that XKB_KEY_Pointer_Button4 is an integer */
  if (!_cffi_check_int(*o, n, 65260U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Button5(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Button5) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Button5) | 0);  /* check that XKB_KEY_Pointer_Button5 is an integer */
  if (!_cffi_check_int(*o, n, 65261U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Button_Dflt(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Button_Dflt) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Button_Dflt) | 0);  /* check that XKB_KEY_Pointer_Button_Dflt is an integer */
  if (!_cffi_check_int(*o, n, 65256U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_DblClick1(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_DblClick1) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_DblClick1) | 0);  /* check that XKB_KEY_Pointer_DblClick1 is an integer */
  if (!_cffi_check_int(*o, n, 65263U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_DblClick2(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_DblClick2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_DblClick2) | 0);  /* check that XKB_KEY_Pointer_DblClick2 is an integer */
  if (!_cffi_check_int(*o, n, 65264U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_DblClick3(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_DblClick3) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_DblClick3) | 0);  /* check that XKB_KEY_Pointer_DblClick3 is an integer */
  if (!_cffi_check_int(*o, n, 65265U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_DblClick4(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_DblClick4) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_DblClick4) | 0);  /* check that XKB_KEY_Pointer_DblClick4 is an integer */
  if (!_cffi_check_int(*o, n, 65266U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_DblClick5(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_DblClick5) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_DblClick5) | 0);  /* check that XKB_KEY_Pointer_DblClick5 is an integer */
  if (!_cffi_check_int(*o, n, 65267U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_DblClick_Dflt(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_DblClick_Dflt) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_DblClick_Dflt) | 0);  /* check that XKB_KEY_Pointer_DblClick_Dflt is an integer */
  if (!_cffi_check_int(*o, n, 65262U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_DfltBtnNext(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_DfltBtnNext) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_DfltBtnNext) | 0);  /* check that XKB_KEY_Pointer_DfltBtnNext is an integer */
  if (!_cffi_check_int(*o, n, 65275U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_DfltBtnPrev(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_DfltBtnPrev) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_DfltBtnPrev) | 0);  /* check that XKB_KEY_Pointer_DfltBtnPrev is an integer */
  if (!_cffi_check_int(*o, n, 65276U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Down(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Down) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Down) | 0);  /* check that XKB_KEY_Pointer_Down is an integer */
  if (!_cffi_check_int(*o, n, 65251U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_DownLeft(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_DownLeft) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_DownLeft) | 0);  /* check that XKB_KEY_Pointer_DownLeft is an integer */
  if (!_cffi_check_int(*o, n, 65254U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_DownRight(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_DownRight) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_DownRight) | 0);  /* check that XKB_KEY_Pointer_DownRight is an integer */
  if (!_cffi_check_int(*o, n, 65255U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Drag1(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Drag1) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Drag1) | 0);  /* check that XKB_KEY_Pointer_Drag1 is an integer */
  if (!_cffi_check_int(*o, n, 65269U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Drag2(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Drag2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Drag2) | 0);  /* check that XKB_KEY_Pointer_Drag2 is an integer */
  if (!_cffi_check_int(*o, n, 65270U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Drag3(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Drag3) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Drag3) | 0);  /* check that XKB_KEY_Pointer_Drag3 is an integer */
  if (!_cffi_check_int(*o, n, 65271U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Drag4(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Drag4) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Drag4) | 0);  /* check that XKB_KEY_Pointer_Drag4 is an integer */
  if (!_cffi_check_int(*o, n, 65272U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Drag5(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Drag5) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Drag5) | 0);  /* check that XKB_KEY_Pointer_Drag5 is an integer */
  if (!_cffi_check_int(*o, n, 65277U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Drag_Dflt(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Drag_Dflt) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Drag_Dflt) | 0);  /* check that XKB_KEY_Pointer_Drag_Dflt is an integer */
  if (!_cffi_check_int(*o, n, 65268U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_EnableKeys(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_EnableKeys) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_EnableKeys) | 0);  /* check that XKB_KEY_Pointer_EnableKeys is an integer */
  if (!_cffi_check_int(*o, n, 65273U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Left(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Left) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Left) | 0);  /* check that XKB_KEY_Pointer_Left is an integer */
  if (!_cffi_check_int(*o, n, 65248U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Right(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Right) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Right) | 0);  /* check that XKB_KEY_Pointer_Right is an integer */
  if (!_cffi_check_int(*o, n, 65249U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_Up(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_Up) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_Up) | 0);  /* check that XKB_KEY_Pointer_Up is an integer */
  if (!_cffi_check_int(*o, n, 65250U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_UpLeft(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_UpLeft) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_UpLeft) | 0);  /* check that XKB_KEY_Pointer_UpLeft is an integer */
  if (!_cffi_check_int(*o, n, 65252U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Pointer_UpRight(unsigned long long *o)
{
  int n = (XKB_KEY_Pointer_UpRight) <= 0;
  *o = (unsigned long long)((XKB_KEY_Pointer_UpRight) | 0);  /* check that XKB_KEY_Pointer_UpRight is an integer */
  if (!_cffi_check_int(*o, n, 65253U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Prev_Virtual_Screen(unsigned long long *o)
{
  int n = (XKB_KEY_Prev_Virtual_Screen) <= 0;
  *o = (unsigned long long)((XKB_KEY_Prev_Virtual_Screen) | 0);  /* check that XKB_KEY_Prev_Virtual_Screen is an integer */
  if (!_cffi_check_int(*o, n, 65233U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_PreviousCandidate(unsigned long long *o)
{
  int n = (XKB_KEY_PreviousCandidate) <= 0;
  *o = (unsigned long long)((XKB_KEY_PreviousCandidate) | 0);  /* check that XKB_KEY_PreviousCandidate is an integer */
  if (!_cffi_check_int(*o, n, 65342U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Print(unsigned long long *o)
{
  int n = (XKB_KEY_Print) <= 0;
  *o = (unsigned long long)((XKB_KEY_Print) | 0);  /* check that XKB_KEY_Print is an integer */
  if (!_cffi_check_int(*o, n, 65377U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Prior(unsigned long long *o)
{
  int n = (XKB_KEY_Prior) <= 0;
  *o = (unsigned long long)((XKB_KEY_Prior) | 0);  /* check that XKB_KEY_Prior is an integer */
  if (!_cffi_check_int(*o, n, 65365U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Q(unsigned long long *o)
{
  int n = (XKB_KEY_Q) <= 0;
  *o = (unsigned long long)((XKB_KEY_Q) | 0);  /* check that XKB_KEY_Q is an integer */
  if (!_cffi_check_int(*o, n, 81U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R(unsigned long long *o)
{
  int n = (XKB_KEY_R) <= 0;
  *o = (unsigned long long)((XKB_KEY_R) | 0);  /* check that XKB_KEY_R is an integer */
  if (!_cffi_check_int(*o, n, 82U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R1(unsigned long long *o)
{
  int n = (XKB_KEY_R1) <= 0;
  *o = (unsigned long long)((XKB_KEY_R1) | 0);  /* check that XKB_KEY_R1 is an integer */
  if (!_cffi_check_int(*o, n, 65490U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R10(unsigned long long *o)
{
  int n = (XKB_KEY_R10) <= 0;
  *o = (unsigned long long)((XKB_KEY_R10) | 0);  /* check that XKB_KEY_R10 is an integer */
  if (!_cffi_check_int(*o, n, 65499U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R11(unsigned long long *o)
{
  int n = (XKB_KEY_R11) <= 0;
  *o = (unsigned long long)((XKB_KEY_R11) | 0);  /* check that XKB_KEY_R11 is an integer */
  if (!_cffi_check_int(*o, n, 65500U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R12(unsigned long long *o)
{
  int n = (XKB_KEY_R12) <= 0;
  *o = (unsigned long long)((XKB_KEY_R12) | 0);  /* check that XKB_KEY_R12 is an integer */
  if (!_cffi_check_int(*o, n, 65501U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R13(unsigned long long *o)
{
  int n = (XKB_KEY_R13) <= 0;
  *o = (unsigned long long)((XKB_KEY_R13) | 0);  /* check that XKB_KEY_R13 is an integer */
  if (!_cffi_check_int(*o, n, 65502U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R14(unsigned long long *o)
{
  int n = (XKB_KEY_R14) <= 0;
  *o = (unsigned long long)((XKB_KEY_R14) | 0);  /* check that XKB_KEY_R14 is an integer */
  if (!_cffi_check_int(*o, n, 65503U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R15(unsigned long long *o)
{
  int n = (XKB_KEY_R15) <= 0;
  *o = (unsigned long long)((XKB_KEY_R15) | 0);  /* check that XKB_KEY_R15 is an integer */
  if (!_cffi_check_int(*o, n, 65504U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R2(unsigned long long *o)
{
  int n = (XKB_KEY_R2) <= 0;
  *o = (unsigned long long)((XKB_KEY_R2) | 0);  /* check that XKB_KEY_R2 is an integer */
  if (!_cffi_check_int(*o, n, 65491U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R3(unsigned long long *o)
{
  int n = (XKB_KEY_R3) <= 0;
  *o = (unsigned long long)((XKB_KEY_R3) | 0);  /* check that XKB_KEY_R3 is an integer */
  if (!_cffi_check_int(*o, n, 65492U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R4(unsigned long long *o)
{
  int n = (XKB_KEY_R4) <= 0;
  *o = (unsigned long long)((XKB_KEY_R4) | 0);  /* check that XKB_KEY_R4 is an integer */
  if (!_cffi_check_int(*o, n, 65493U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R5(unsigned long long *o)
{
  int n = (XKB_KEY_R5) <= 0;
  *o = (unsigned long long)((XKB_KEY_R5) | 0);  /* check that XKB_KEY_R5 is an integer */
  if (!_cffi_check_int(*o, n, 65494U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R6(unsigned long long *o)
{
  int n = (XKB_KEY_R6) <= 0;
  *o = (unsigned long long)((XKB_KEY_R6) | 0);  /* check that XKB_KEY_R6 is an integer */
  if (!_cffi_check_int(*o, n, 65495U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R7(unsigned long long *o)
{
  int n = (XKB_KEY_R7) <= 0;
  *o = (unsigned long long)((XKB_KEY_R7) | 0);  /* check that XKB_KEY_R7 is an integer */
  if (!_cffi_check_int(*o, n, 65496U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R8(unsigned long long *o)
{
  int n = (XKB_KEY_R8) <= 0;
  *o = (unsigned long long)((XKB_KEY_R8) | 0);  /* check that XKB_KEY_R8 is an integer */
  if (!_cffi_check_int(*o, n, 65497U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_R9(unsigned long long *o)
{
  int n = (XKB_KEY_R9) <= 0;
  *o = (unsigned long long)((XKB_KEY_R9) | 0);  /* check that XKB_KEY_R9 is an integer */
  if (!_cffi_check_int(*o, n, 65498U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Racute(unsigned long long *o)
{
  int n = (XKB_KEY_Racute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Racute) | 0);  /* check that XKB_KEY_Racute is an integer */
  if (!_cffi_check_int(*o, n, 448U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Rcaron(unsigned long long *o)
{
  int n = (XKB_KEY_Rcaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Rcaron) | 0);  /* check that XKB_KEY_Rcaron is an integer */
  if (!_cffi_check_int(*o, n, 472U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Rcedilla(unsigned long long *o)
{
  int n = (XKB_KEY_Rcedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_Rcedilla) | 0);  /* check that XKB_KEY_Rcedilla is an integer */
  if (!_cffi_check_int(*o, n, 931U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Redo(unsigned long long *o)
{
  int n = (XKB_KEY_Redo) <= 0;
  *o = (unsigned long long)((XKB_KEY_Redo) | 0);  /* check that XKB_KEY_Redo is an integer */
  if (!_cffi_check_int(*o, n, 65382U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_RepeatKeys_Enable(unsigned long long *o)
{
  int n = (XKB_KEY_RepeatKeys_Enable) <= 0;
  *o = (unsigned long long)((XKB_KEY_RepeatKeys_Enable) | 0);  /* check that XKB_KEY_RepeatKeys_Enable is an integer */
  if (!_cffi_check_int(*o, n, 65138U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Reset(unsigned long long *o)
{
  int n = (XKB_KEY_Reset) <= 0;
  *o = (unsigned long long)((XKB_KEY_Reset) | 0);  /* check that XKB_KEY_Reset is an integer */
  if (!_cffi_check_int(*o, n, 268500844U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Return(unsigned long long *o)
{
  int n = (XKB_KEY_Return) <= 0;
  *o = (unsigned long long)((XKB_KEY_Return) | 0);  /* check that XKB_KEY_Return is an integer */
  if (!_cffi_check_int(*o, n, 65293U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Right(unsigned long long *o)
{
  int n = (XKB_KEY_Right) <= 0;
  *o = (unsigned long long)((XKB_KEY_Right) | 0);  /* check that XKB_KEY_Right is an integer */
  if (!_cffi_check_int(*o, n, 65363U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Romaji(unsigned long long *o)
{
  int n = (XKB_KEY_Romaji) <= 0;
  *o = (unsigned long long)((XKB_KEY_Romaji) | 0);  /* check that XKB_KEY_Romaji is an integer */
  if (!_cffi_check_int(*o, n, 65316U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_RupeeSign(unsigned long long *o)
{
  int n = (XKB_KEY_RupeeSign) <= 0;
  *o = (unsigned long long)((XKB_KEY_RupeeSign) | 0);  /* check that XKB_KEY_RupeeSign is an integer */
  if (!_cffi_check_int(*o, n, 16785576U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_S(unsigned long long *o)
{
  int n = (XKB_KEY_S) <= 0;
  *o = (unsigned long long)((XKB_KEY_S) | 0);  /* check that XKB_KEY_S is an integer */
  if (!_cffi_check_int(*o, n, 83U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SCHWA(unsigned long long *o)
{
  int n = (XKB_KEY_SCHWA) <= 0;
  *o = (unsigned long long)((XKB_KEY_SCHWA) | 0);  /* check that XKB_KEY_SCHWA is an integer */
  if (!_cffi_check_int(*o, n, 16777615U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_Sabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sabovedot) | 0);  /* check that XKB_KEY_Sabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16784992U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sacute(unsigned long long *o)
{
  int n = (XKB_KEY_Sacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sacute) | 0);  /* check that XKB_KEY_Sacute is an integer */
  if (!_cffi_check_int(*o, n, 422U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Scaron(unsigned long long *o)
{
  int n = (XKB_KEY_Scaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Scaron) | 0);  /* check that XKB_KEY_Scaron is an integer */
  if (!_cffi_check_int(*o, n, 425U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Scedilla(unsigned long long *o)
{
  int n = (XKB_KEY_Scedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_Scedilla) | 0);  /* check that XKB_KEY_Scedilla is an integer */
  if (!_cffi_check_int(*o, n, 426U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Scircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_Scircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_Scircumflex) | 0);  /* check that XKB_KEY_Scircumflex is an integer */
  if (!_cffi_check_int(*o, n, 734U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Scroll_Lock(unsigned long long *o)
{
  int n = (XKB_KEY_Scroll_Lock) <= 0;
  *o = (unsigned long long)((XKB_KEY_Scroll_Lock) | 0);  /* check that XKB_KEY_Scroll_Lock is an integer */
  if (!_cffi_check_int(*o, n, 65300U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Select(unsigned long long *o)
{
  int n = (XKB_KEY_Select) <= 0;
  *o = (unsigned long long)((XKB_KEY_Select) | 0);  /* check that XKB_KEY_Select is an integer */
  if (!_cffi_check_int(*o, n, 65376U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Serbian_DJE(unsigned long long *o)
{
  int n = (XKB_KEY_Serbian_DJE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Serbian_DJE) | 0);  /* check that XKB_KEY_Serbian_DJE is an integer */
  if (!_cffi_check_int(*o, n, 1713U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Serbian_DZE(unsigned long long *o)
{
  int n = (XKB_KEY_Serbian_DZE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Serbian_DZE) | 0);  /* check that XKB_KEY_Serbian_DZE is an integer */
  if (!_cffi_check_int(*o, n, 1727U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Serbian_JE(unsigned long long *o)
{
  int n = (XKB_KEY_Serbian_JE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Serbian_JE) | 0);  /* check that XKB_KEY_Serbian_JE is an integer */
  if (!_cffi_check_int(*o, n, 1720U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Serbian_LJE(unsigned long long *o)
{
  int n = (XKB_KEY_Serbian_LJE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Serbian_LJE) | 0);  /* check that XKB_KEY_Serbian_LJE is an integer */
  if (!_cffi_check_int(*o, n, 1721U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Serbian_NJE(unsigned long long *o)
{
  int n = (XKB_KEY_Serbian_NJE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Serbian_NJE) | 0);  /* check that XKB_KEY_Serbian_NJE is an integer */
  if (!_cffi_check_int(*o, n, 1722U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Serbian_TSHE(unsigned long long *o)
{
  int n = (XKB_KEY_Serbian_TSHE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Serbian_TSHE) | 0);  /* check that XKB_KEY_Serbian_TSHE is an integer */
  if (!_cffi_check_int(*o, n, 1723U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Serbian_dje(unsigned long long *o)
{
  int n = (XKB_KEY_Serbian_dje) <= 0;
  *o = (unsigned long long)((XKB_KEY_Serbian_dje) | 0);  /* check that XKB_KEY_Serbian_dje is an integer */
  if (!_cffi_check_int(*o, n, 1697U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Serbian_dze(unsigned long long *o)
{
  int n = (XKB_KEY_Serbian_dze) <= 0;
  *o = (unsigned long long)((XKB_KEY_Serbian_dze) | 0);  /* check that XKB_KEY_Serbian_dze is an integer */
  if (!_cffi_check_int(*o, n, 1711U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Serbian_je(unsigned long long *o)
{
  int n = (XKB_KEY_Serbian_je) <= 0;
  *o = (unsigned long long)((XKB_KEY_Serbian_je) | 0);  /* check that XKB_KEY_Serbian_je is an integer */
  if (!_cffi_check_int(*o, n, 1704U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Serbian_lje(unsigned long long *o)
{
  int n = (XKB_KEY_Serbian_lje) <= 0;
  *o = (unsigned long long)((XKB_KEY_Serbian_lje) | 0);  /* check that XKB_KEY_Serbian_lje is an integer */
  if (!_cffi_check_int(*o, n, 1705U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Serbian_nje(unsigned long long *o)
{
  int n = (XKB_KEY_Serbian_nje) <= 0;
  *o = (unsigned long long)((XKB_KEY_Serbian_nje) | 0);  /* check that XKB_KEY_Serbian_nje is an integer */
  if (!_cffi_check_int(*o, n, 1706U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Serbian_tshe(unsigned long long *o)
{
  int n = (XKB_KEY_Serbian_tshe) <= 0;
  *o = (unsigned long long)((XKB_KEY_Serbian_tshe) | 0);  /* check that XKB_KEY_Serbian_tshe is an integer */
  if (!_cffi_check_int(*o, n, 1707U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Shift_L(unsigned long long *o)
{
  int n = (XKB_KEY_Shift_L) <= 0;
  *o = (unsigned long long)((XKB_KEY_Shift_L) | 0);  /* check that XKB_KEY_Shift_L is an integer */
  if (!_cffi_check_int(*o, n, 65505U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Shift_Lock(unsigned long long *o)
{
  int n = (XKB_KEY_Shift_Lock) <= 0;
  *o = (unsigned long long)((XKB_KEY_Shift_Lock) | 0);  /* check that XKB_KEY_Shift_Lock is an integer */
  if (!_cffi_check_int(*o, n, 65510U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Shift_R(unsigned long long *o)
{
  int n = (XKB_KEY_Shift_R) <= 0;
  *o = (unsigned long long)((XKB_KEY_Shift_R) | 0);  /* check that XKB_KEY_Shift_R is an integer */
  if (!_cffi_check_int(*o, n, 65506U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SingleCandidate(unsigned long long *o)
{
  int n = (XKB_KEY_SingleCandidate) <= 0;
  *o = (unsigned long long)((XKB_KEY_SingleCandidate) | 0);  /* check that XKB_KEY_SingleCandidate is an integer */
  if (!_cffi_check_int(*o, n, 65340U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_a(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_a) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_a) | 0);  /* check that XKB_KEY_Sinh_a is an integer */
  if (!_cffi_check_int(*o, n, 16780677U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_aa(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_aa) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_aa) | 0);  /* check that XKB_KEY_Sinh_aa is an integer */
  if (!_cffi_check_int(*o, n, 16780678U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_aa2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_aa2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_aa2) | 0);  /* check that XKB_KEY_Sinh_aa2 is an integer */
  if (!_cffi_check_int(*o, n, 16780751U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ae(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ae) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ae) | 0);  /* check that XKB_KEY_Sinh_ae is an integer */
  if (!_cffi_check_int(*o, n, 16780679U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ae2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ae2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ae2) | 0);  /* check that XKB_KEY_Sinh_ae2 is an integer */
  if (!_cffi_check_int(*o, n, 16780752U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_aee(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_aee) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_aee) | 0);  /* check that XKB_KEY_Sinh_aee is an integer */
  if (!_cffi_check_int(*o, n, 16780680U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_aee2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_aee2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_aee2) | 0);  /* check that XKB_KEY_Sinh_aee2 is an integer */
  if (!_cffi_check_int(*o, n, 16780753U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ai(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ai) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ai) | 0);  /* check that XKB_KEY_Sinh_ai is an integer */
  if (!_cffi_check_int(*o, n, 16780691U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ai2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ai2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ai2) | 0);  /* check that XKB_KEY_Sinh_ai2 is an integer */
  if (!_cffi_check_int(*o, n, 16780763U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_al(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_al) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_al) | 0);  /* check that XKB_KEY_Sinh_al is an integer */
  if (!_cffi_check_int(*o, n, 16780746U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_au(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_au) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_au) | 0);  /* check that XKB_KEY_Sinh_au is an integer */
  if (!_cffi_check_int(*o, n, 16780694U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_au2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_au2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_au2) | 0);  /* check that XKB_KEY_Sinh_au2 is an integer */
  if (!_cffi_check_int(*o, n, 16780766U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ba(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ba) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ba) | 0);  /* check that XKB_KEY_Sinh_ba is an integer */
  if (!_cffi_check_int(*o, n, 16780726U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_bha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_bha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_bha) | 0);  /* check that XKB_KEY_Sinh_bha is an integer */
  if (!_cffi_check_int(*o, n, 16780727U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ca(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ca) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ca) | 0);  /* check that XKB_KEY_Sinh_ca is an integer */
  if (!_cffi_check_int(*o, n, 16780704U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_cha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_cha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_cha) | 0);  /* check that XKB_KEY_Sinh_cha is an integer */
  if (!_cffi_check_int(*o, n, 16780705U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_dda(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_dda) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_dda) | 0);  /* check that XKB_KEY_Sinh_dda is an integer */
  if (!_cffi_check_int(*o, n, 16780713U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ddha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ddha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ddha) | 0);  /* check that XKB_KEY_Sinh_ddha is an integer */
  if (!_cffi_check_int(*o, n, 16780714U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_dha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_dha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_dha) | 0);  /* check that XKB_KEY_Sinh_dha is an integer */
  if (!_cffi_check_int(*o, n, 16780719U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_dhha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_dhha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_dhha) | 0);  /* check that XKB_KEY_Sinh_dhha is an integer */
  if (!_cffi_check_int(*o, n, 16780720U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_e(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_e) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_e) | 0);  /* check that XKB_KEY_Sinh_e is an integer */
  if (!_cffi_check_int(*o, n, 16780689U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_e2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_e2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_e2) | 0);  /* check that XKB_KEY_Sinh_e2 is an integer */
  if (!_cffi_check_int(*o, n, 16780761U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ee(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ee) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ee) | 0);  /* check that XKB_KEY_Sinh_ee is an integer */
  if (!_cffi_check_int(*o, n, 16780690U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ee2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ee2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ee2) | 0);  /* check that XKB_KEY_Sinh_ee2 is an integer */
  if (!_cffi_check_int(*o, n, 16780762U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_fa(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_fa) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_fa) | 0);  /* check that XKB_KEY_Sinh_fa is an integer */
  if (!_cffi_check_int(*o, n, 16780742U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ga(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ga) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ga) | 0);  /* check that XKB_KEY_Sinh_ga is an integer */
  if (!_cffi_check_int(*o, n, 16780700U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_gha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_gha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_gha) | 0);  /* check that XKB_KEY_Sinh_gha is an integer */
  if (!_cffi_check_int(*o, n, 16780701U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_h2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_h2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_h2) | 0);  /* check that XKB_KEY_Sinh_h2 is an integer */
  if (!_cffi_check_int(*o, n, 16780675U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ha) | 0);  /* check that XKB_KEY_Sinh_ha is an integer */
  if (!_cffi_check_int(*o, n, 16780740U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_i(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_i) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_i) | 0);  /* check that XKB_KEY_Sinh_i is an integer */
  if (!_cffi_check_int(*o, n, 16780681U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_i2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_i2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_i2) | 0);  /* check that XKB_KEY_Sinh_i2 is an integer */
  if (!_cffi_check_int(*o, n, 16780754U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ii(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ii) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ii) | 0);  /* check that XKB_KEY_Sinh_ii is an integer */
  if (!_cffi_check_int(*o, n, 16780682U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ii2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ii2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ii2) | 0);  /* check that XKB_KEY_Sinh_ii2 is an integer */
  if (!_cffi_check_int(*o, n, 16780755U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ja(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ja) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ja) | 0);  /* check that XKB_KEY_Sinh_ja is an integer */
  if (!_cffi_check_int(*o, n, 16780706U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_jha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_jha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_jha) | 0);  /* check that XKB_KEY_Sinh_jha is an integer */
  if (!_cffi_check_int(*o, n, 16780707U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_jnya(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_jnya) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_jnya) | 0);  /* check that XKB_KEY_Sinh_jnya is an integer */
  if (!_cffi_check_int(*o, n, 16780709U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ka(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ka) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ka) | 0);  /* check that XKB_KEY_Sinh_ka is an integer */
  if (!_cffi_check_int(*o, n, 16780698U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_kha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_kha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_kha) | 0);  /* check that XKB_KEY_Sinh_kha is an integer */
  if (!_cffi_check_int(*o, n, 16780699U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_kunddaliya(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_kunddaliya) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_kunddaliya) | 0);  /* check that XKB_KEY_Sinh_kunddaliya is an integer */
  if (!_cffi_check_int(*o, n, 16780788U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_la(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_la) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_la) | 0);  /* check that XKB_KEY_Sinh_la is an integer */
  if (!_cffi_check_int(*o, n, 16780733U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_lla(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_lla) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_lla) | 0);  /* check that XKB_KEY_Sinh_lla is an integer */
  if (!_cffi_check_int(*o, n, 16780741U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_lu(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_lu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_lu) | 0);  /* check that XKB_KEY_Sinh_lu is an integer */
  if (!_cffi_check_int(*o, n, 16780687U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_lu2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_lu2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_lu2) | 0);  /* check that XKB_KEY_Sinh_lu2 is an integer */
  if (!_cffi_check_int(*o, n, 16780767U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_luu(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_luu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_luu) | 0);  /* check that XKB_KEY_Sinh_luu is an integer */
  if (!_cffi_check_int(*o, n, 16780688U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_luu2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_luu2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_luu2) | 0);  /* check that XKB_KEY_Sinh_luu2 is an integer */
  if (!_cffi_check_int(*o, n, 16780787U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ma(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ma) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ma) | 0);  /* check that XKB_KEY_Sinh_ma is an integer */
  if (!_cffi_check_int(*o, n, 16780728U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_mba(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_mba) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_mba) | 0);  /* check that XKB_KEY_Sinh_mba is an integer */
  if (!_cffi_check_int(*o, n, 16780729U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_na(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_na) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_na) | 0);  /* check that XKB_KEY_Sinh_na is an integer */
  if (!_cffi_check_int(*o, n, 16780721U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ndda(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ndda) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ndda) | 0);  /* check that XKB_KEY_Sinh_ndda is an integer */
  if (!_cffi_check_int(*o, n, 16780716U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ndha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ndha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ndha) | 0);  /* check that XKB_KEY_Sinh_ndha is an integer */
  if (!_cffi_check_int(*o, n, 16780723U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ng(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ng) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ng) | 0);  /* check that XKB_KEY_Sinh_ng is an integer */
  if (!_cffi_check_int(*o, n, 16780674U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ng2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ng2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ng2) | 0);  /* check that XKB_KEY_Sinh_ng2 is an integer */
  if (!_cffi_check_int(*o, n, 16780702U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_nga(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_nga) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_nga) | 0);  /* check that XKB_KEY_Sinh_nga is an integer */
  if (!_cffi_check_int(*o, n, 16780703U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_nja(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_nja) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_nja) | 0);  /* check that XKB_KEY_Sinh_nja is an integer */
  if (!_cffi_check_int(*o, n, 16780710U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_nna(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_nna) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_nna) | 0);  /* check that XKB_KEY_Sinh_nna is an integer */
  if (!_cffi_check_int(*o, n, 16780715U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_nya(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_nya) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_nya) | 0);  /* check that XKB_KEY_Sinh_nya is an integer */
  if (!_cffi_check_int(*o, n, 16780708U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_o(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_o) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_o) | 0);  /* check that XKB_KEY_Sinh_o is an integer */
  if (!_cffi_check_int(*o, n, 16780692U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_o2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_o2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_o2) | 0);  /* check that XKB_KEY_Sinh_o2 is an integer */
  if (!_cffi_check_int(*o, n, 16780764U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_oo(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_oo) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_oo) | 0);  /* check that XKB_KEY_Sinh_oo is an integer */
  if (!_cffi_check_int(*o, n, 16780693U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_oo2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_oo2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_oo2) | 0);  /* check that XKB_KEY_Sinh_oo2 is an integer */
  if (!_cffi_check_int(*o, n, 16780765U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_pa(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_pa) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_pa) | 0);  /* check that XKB_KEY_Sinh_pa is an integer */
  if (!_cffi_check_int(*o, n, 16780724U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_pha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_pha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_pha) | 0);  /* check that XKB_KEY_Sinh_pha is an integer */
  if (!_cffi_check_int(*o, n, 16780725U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ra(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ra) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ra) | 0);  /* check that XKB_KEY_Sinh_ra is an integer */
  if (!_cffi_check_int(*o, n, 16780731U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ri(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ri) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ri) | 0);  /* check that XKB_KEY_Sinh_ri is an integer */
  if (!_cffi_check_int(*o, n, 16780685U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_rii(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_rii) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_rii) | 0);  /* check that XKB_KEY_Sinh_rii is an integer */
  if (!_cffi_check_int(*o, n, 16780686U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ru2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ru2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ru2) | 0);  /* check that XKB_KEY_Sinh_ru2 is an integer */
  if (!_cffi_check_int(*o, n, 16780760U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ruu2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ruu2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ruu2) | 0);  /* check that XKB_KEY_Sinh_ruu2 is an integer */
  if (!_cffi_check_int(*o, n, 16780786U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_sa(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_sa) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_sa) | 0);  /* check that XKB_KEY_Sinh_sa is an integer */
  if (!_cffi_check_int(*o, n, 16780739U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_sha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_sha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_sha) | 0);  /* check that XKB_KEY_Sinh_sha is an integer */
  if (!_cffi_check_int(*o, n, 16780737U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ssha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ssha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ssha) | 0);  /* check that XKB_KEY_Sinh_ssha is an integer */
  if (!_cffi_check_int(*o, n, 16780738U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_tha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_tha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_tha) | 0);  /* check that XKB_KEY_Sinh_tha is an integer */
  if (!_cffi_check_int(*o, n, 16780717U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_thha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_thha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_thha) | 0);  /* check that XKB_KEY_Sinh_thha is an integer */
  if (!_cffi_check_int(*o, n, 16780718U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_tta(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_tta) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_tta) | 0);  /* check that XKB_KEY_Sinh_tta is an integer */
  if (!_cffi_check_int(*o, n, 16780711U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ttha(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ttha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ttha) | 0);  /* check that XKB_KEY_Sinh_ttha is an integer */
  if (!_cffi_check_int(*o, n, 16780712U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_u(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_u) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_u) | 0);  /* check that XKB_KEY_Sinh_u is an integer */
  if (!_cffi_check_int(*o, n, 16780683U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_u2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_u2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_u2) | 0);  /* check that XKB_KEY_Sinh_u2 is an integer */
  if (!_cffi_check_int(*o, n, 16780756U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_uu(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_uu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_uu) | 0);  /* check that XKB_KEY_Sinh_uu is an integer */
  if (!_cffi_check_int(*o, n, 16780684U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_uu2(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_uu2) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_uu2) | 0);  /* check that XKB_KEY_Sinh_uu2 is an integer */
  if (!_cffi_check_int(*o, n, 16780758U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_va(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_va) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_va) | 0);  /* check that XKB_KEY_Sinh_va is an integer */
  if (!_cffi_check_int(*o, n, 16780736U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sinh_ya(unsigned long long *o)
{
  int n = (XKB_KEY_Sinh_ya) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sinh_ya) | 0);  /* check that XKB_KEY_Sinh_ya is an integer */
  if (!_cffi_check_int(*o, n, 16780730U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SlowKeys_Enable(unsigned long long *o)
{
  int n = (XKB_KEY_SlowKeys_Enable) <= 0;
  *o = (unsigned long long)((XKB_KEY_SlowKeys_Enable) | 0);  /* check that XKB_KEY_SlowKeys_Enable is an integer */
  if (!_cffi_check_int(*o, n, 65139U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_StickyKeys_Enable(unsigned long long *o)
{
  int n = (XKB_KEY_StickyKeys_Enable) <= 0;
  *o = (unsigned long long)((XKB_KEY_StickyKeys_Enable) | 0);  /* check that XKB_KEY_StickyKeys_Enable is an integer */
  if (!_cffi_check_int(*o, n, 65141U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunAgain(unsigned long long *o)
{
  int n = (XKB_KEY_SunAgain) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunAgain) | 0);  /* check that XKB_KEY_SunAgain is an integer */
  if (!_cffi_check_int(*o, n, 65382U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunAltGraph(unsigned long long *o)
{
  int n = (XKB_KEY_SunAltGraph) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunAltGraph) | 0);  /* check that XKB_KEY_SunAltGraph is an integer */
  if (!_cffi_check_int(*o, n, 65406U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunAudioLowerVolume(unsigned long long *o)
{
  int n = (XKB_KEY_SunAudioLowerVolume) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunAudioLowerVolume) | 0);  /* check that XKB_KEY_SunAudioLowerVolume is an integer */
  if (!_cffi_check_int(*o, n, 268828535U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunAudioMute(unsigned long long *o)
{
  int n = (XKB_KEY_SunAudioMute) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunAudioMute) | 0);  /* check that XKB_KEY_SunAudioMute is an integer */
  if (!_cffi_check_int(*o, n, 268828536U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunAudioRaiseVolume(unsigned long long *o)
{
  int n = (XKB_KEY_SunAudioRaiseVolume) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunAudioRaiseVolume) | 0);  /* check that XKB_KEY_SunAudioRaiseVolume is an integer */
  if (!_cffi_check_int(*o, n, 268828537U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunCompose(unsigned long long *o)
{
  int n = (XKB_KEY_SunCompose) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunCompose) | 0);  /* check that XKB_KEY_SunCompose is an integer */
  if (!_cffi_check_int(*o, n, 65312U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunCopy(unsigned long long *o)
{
  int n = (XKB_KEY_SunCopy) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunCopy) | 0);  /* check that XKB_KEY_SunCopy is an integer */
  if (!_cffi_check_int(*o, n, 268828530U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunCut(unsigned long long *o)
{
  int n = (XKB_KEY_SunCut) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunCut) | 0);  /* check that XKB_KEY_SunCut is an integer */
  if (!_cffi_check_int(*o, n, 268828533U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunF36(unsigned long long *o)
{
  int n = (XKB_KEY_SunF36) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunF36) | 0);  /* check that XKB_KEY_SunF36 is an integer */
  if (!_cffi_check_int(*o, n, 268828432U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunF37(unsigned long long *o)
{
  int n = (XKB_KEY_SunF37) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunF37) | 0);  /* check that XKB_KEY_SunF37 is an integer */
  if (!_cffi_check_int(*o, n, 268828433U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunFA_Acute(unsigned long long *o)
{
  int n = (XKB_KEY_SunFA_Acute) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunFA_Acute) | 0);  /* check that XKB_KEY_SunFA_Acute is an integer */
  if (!_cffi_check_int(*o, n, 268828419U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunFA_Cedilla(unsigned long long *o)
{
  int n = (XKB_KEY_SunFA_Cedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunFA_Cedilla) | 0);  /* check that XKB_KEY_SunFA_Cedilla is an integer */
  if (!_cffi_check_int(*o, n, 268828421U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunFA_Circum(unsigned long long *o)
{
  int n = (XKB_KEY_SunFA_Circum) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunFA_Circum) | 0);  /* check that XKB_KEY_SunFA_Circum is an integer */
  if (!_cffi_check_int(*o, n, 268828417U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunFA_Diaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_SunFA_Diaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunFA_Diaeresis) | 0);  /* check that XKB_KEY_SunFA_Diaeresis is an integer */
  if (!_cffi_check_int(*o, n, 268828420U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunFA_Grave(unsigned long long *o)
{
  int n = (XKB_KEY_SunFA_Grave) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunFA_Grave) | 0);  /* check that XKB_KEY_SunFA_Grave is an integer */
  if (!_cffi_check_int(*o, n, 268828416U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunFA_Tilde(unsigned long long *o)
{
  int n = (XKB_KEY_SunFA_Tilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunFA_Tilde) | 0);  /* check that XKB_KEY_SunFA_Tilde is an integer */
  if (!_cffi_check_int(*o, n, 268828418U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunFind(unsigned long long *o)
{
  int n = (XKB_KEY_SunFind) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunFind) | 0);  /* check that XKB_KEY_SunFind is an integer */
  if (!_cffi_check_int(*o, n, 65384U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunFront(unsigned long long *o)
{
  int n = (XKB_KEY_SunFront) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunFront) | 0);  /* check that XKB_KEY_SunFront is an integer */
  if (!_cffi_check_int(*o, n, 268828529U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunOpen(unsigned long long *o)
{
  int n = (XKB_KEY_SunOpen) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunOpen) | 0);  /* check that XKB_KEY_SunOpen is an integer */
  if (!_cffi_check_int(*o, n, 268828531U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunPageDown(unsigned long long *o)
{
  int n = (XKB_KEY_SunPageDown) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunPageDown) | 0);  /* check that XKB_KEY_SunPageDown is an integer */
  if (!_cffi_check_int(*o, n, 65366U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunPageUp(unsigned long long *o)
{
  int n = (XKB_KEY_SunPageUp) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunPageUp) | 0);  /* check that XKB_KEY_SunPageUp is an integer */
  if (!_cffi_check_int(*o, n, 65365U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunPaste(unsigned long long *o)
{
  int n = (XKB_KEY_SunPaste) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunPaste) | 0);  /* check that XKB_KEY_SunPaste is an integer */
  if (!_cffi_check_int(*o, n, 268828532U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunPowerSwitch(unsigned long long *o)
{
  int n = (XKB_KEY_SunPowerSwitch) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunPowerSwitch) | 0);  /* check that XKB_KEY_SunPowerSwitch is an integer */
  if (!_cffi_check_int(*o, n, 268828534U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunPowerSwitchShift(unsigned long long *o)
{
  int n = (XKB_KEY_SunPowerSwitchShift) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunPowerSwitchShift) | 0);  /* check that XKB_KEY_SunPowerSwitchShift is an integer */
  if (!_cffi_check_int(*o, n, 268828541U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunPrint_Screen(unsigned long long *o)
{
  int n = (XKB_KEY_SunPrint_Screen) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunPrint_Screen) | 0);  /* check that XKB_KEY_SunPrint_Screen is an integer */
  if (!_cffi_check_int(*o, n, 65377U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunProps(unsigned long long *o)
{
  int n = (XKB_KEY_SunProps) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunProps) | 0);  /* check that XKB_KEY_SunProps is an integer */
  if (!_cffi_check_int(*o, n, 268828528U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunStop(unsigned long long *o)
{
  int n = (XKB_KEY_SunStop) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunStop) | 0);  /* check that XKB_KEY_SunStop is an integer */
  if (!_cffi_check_int(*o, n, 65385U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunSys_Req(unsigned long long *o)
{
  int n = (XKB_KEY_SunSys_Req) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunSys_Req) | 0);  /* check that XKB_KEY_SunSys_Req is an integer */
  if (!_cffi_check_int(*o, n, 268828512U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunUndo(unsigned long long *o)
{
  int n = (XKB_KEY_SunUndo) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunUndo) | 0);  /* check that XKB_KEY_SunUndo is an integer */
  if (!_cffi_check_int(*o, n, 65381U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunVideoDegauss(unsigned long long *o)
{
  int n = (XKB_KEY_SunVideoDegauss) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunVideoDegauss) | 0);  /* check that XKB_KEY_SunVideoDegauss is an integer */
  if (!_cffi_check_int(*o, n, 268828538U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunVideoLowerBrightness(unsigned long long *o)
{
  int n = (XKB_KEY_SunVideoLowerBrightness) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunVideoLowerBrightness) | 0);  /* check that XKB_KEY_SunVideoLowerBrightness is an integer */
  if (!_cffi_check_int(*o, n, 268828539U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_SunVideoRaiseBrightness(unsigned long long *o)
{
  int n = (XKB_KEY_SunVideoRaiseBrightness) <= 0;
  *o = (unsigned long long)((XKB_KEY_SunVideoRaiseBrightness) | 0);  /* check that XKB_KEY_SunVideoRaiseBrightness is an integer */
  if (!_cffi_check_int(*o, n, 268828540U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Super_L(unsigned long long *o)
{
  int n = (XKB_KEY_Super_L) <= 0;
  *o = (unsigned long long)((XKB_KEY_Super_L) | 0);  /* check that XKB_KEY_Super_L is an integer */
  if (!_cffi_check_int(*o, n, 65515U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Super_R(unsigned long long *o)
{
  int n = (XKB_KEY_Super_R) <= 0;
  *o = (unsigned long long)((XKB_KEY_Super_R) | 0);  /* check that XKB_KEY_Super_R is an integer */
  if (!_cffi_check_int(*o, n, 65516U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Sys_Req(unsigned long long *o)
{
  int n = (XKB_KEY_Sys_Req) <= 0;
  *o = (unsigned long long)((XKB_KEY_Sys_Req) | 0);  /* check that XKB_KEY_Sys_Req is an integer */
  if (!_cffi_check_int(*o, n, 65301U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_System(unsigned long long *o)
{
  int n = (XKB_KEY_System) <= 0;
  *o = (unsigned long long)((XKB_KEY_System) | 0);  /* check that XKB_KEY_System is an integer */
  if (!_cffi_check_int(*o, n, 268500845U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_T(unsigned long long *o)
{
  int n = (XKB_KEY_T) <= 0;
  *o = (unsigned long long)((XKB_KEY_T) | 0);  /* check that XKB_KEY_T is an integer */
  if (!_cffi_check_int(*o, n, 84U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_THORN(unsigned long long *o)
{
  int n = (XKB_KEY_THORN) <= 0;
  *o = (unsigned long long)((XKB_KEY_THORN) | 0);  /* check that XKB_KEY_THORN is an integer */
  if (!_cffi_check_int(*o, n, 222U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Tab(unsigned long long *o)
{
  int n = (XKB_KEY_Tab) <= 0;
  *o = (unsigned long long)((XKB_KEY_Tab) | 0);  /* check that XKB_KEY_Tab is an integer */
  if (!_cffi_check_int(*o, n, 65289U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Tabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_Tabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Tabovedot) | 0);  /* check that XKB_KEY_Tabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16785002U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Tcaron(unsigned long long *o)
{
  int n = (XKB_KEY_Tcaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Tcaron) | 0);  /* check that XKB_KEY_Tcaron is an integer */
  if (!_cffi_check_int(*o, n, 427U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Tcedilla(unsigned long long *o)
{
  int n = (XKB_KEY_Tcedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_Tcedilla) | 0);  /* check that XKB_KEY_Tcedilla is an integer */
  if (!_cffi_check_int(*o, n, 478U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Terminate_Server(unsigned long long *o)
{
  int n = (XKB_KEY_Terminate_Server) <= 0;
  *o = (unsigned long long)((XKB_KEY_Terminate_Server) | 0);  /* check that XKB_KEY_Terminate_Server is an integer */
  if (!_cffi_check_int(*o, n, 65237U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_baht(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_baht) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_baht) | 0);  /* check that XKB_KEY_Thai_baht is an integer */
  if (!_cffi_check_int(*o, n, 3551U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_bobaimai(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_bobaimai) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_bobaimai) | 0);  /* check that XKB_KEY_Thai_bobaimai is an integer */
  if (!_cffi_check_int(*o, n, 3514U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_chochan(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_chochan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_chochan) | 0);  /* check that XKB_KEY_Thai_chochan is an integer */
  if (!_cffi_check_int(*o, n, 3496U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_chochang(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_chochang) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_chochang) | 0);  /* check that XKB_KEY_Thai_chochang is an integer */
  if (!_cffi_check_int(*o, n, 3498U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_choching(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_choching) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_choching) | 0);  /* check that XKB_KEY_Thai_choching is an integer */
  if (!_cffi_check_int(*o, n, 3497U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_chochoe(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_chochoe) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_chochoe) | 0);  /* check that XKB_KEY_Thai_chochoe is an integer */
  if (!_cffi_check_int(*o, n, 3500U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_dochada(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_dochada) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_dochada) | 0);  /* check that XKB_KEY_Thai_dochada is an integer */
  if (!_cffi_check_int(*o, n, 3502U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_dodek(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_dodek) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_dodek) | 0);  /* check that XKB_KEY_Thai_dodek is an integer */
  if (!_cffi_check_int(*o, n, 3508U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_fofa(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_fofa) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_fofa) | 0);  /* check that XKB_KEY_Thai_fofa is an integer */
  if (!_cffi_check_int(*o, n, 3517U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_fofan(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_fofan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_fofan) | 0);  /* check that XKB_KEY_Thai_fofan is an integer */
  if (!_cffi_check_int(*o, n, 3519U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_hohip(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_hohip) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_hohip) | 0);  /* check that XKB_KEY_Thai_hohip is an integer */
  if (!_cffi_check_int(*o, n, 3531U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_honokhuk(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_honokhuk) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_honokhuk) | 0);  /* check that XKB_KEY_Thai_honokhuk is an integer */
  if (!_cffi_check_int(*o, n, 3534U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_khokhai(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_khokhai) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_khokhai) | 0);  /* check that XKB_KEY_Thai_khokhai is an integer */
  if (!_cffi_check_int(*o, n, 3490U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_khokhon(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_khokhon) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_khokhon) | 0);  /* check that XKB_KEY_Thai_khokhon is an integer */
  if (!_cffi_check_int(*o, n, 3493U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_khokhuat(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_khokhuat) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_khokhuat) | 0);  /* check that XKB_KEY_Thai_khokhuat is an integer */
  if (!_cffi_check_int(*o, n, 3491U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_khokhwai(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_khokhwai) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_khokhwai) | 0);  /* check that XKB_KEY_Thai_khokhwai is an integer */
  if (!_cffi_check_int(*o, n, 3492U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_khorakhang(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_khorakhang) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_khorakhang) | 0);  /* check that XKB_KEY_Thai_khorakhang is an integer */
  if (!_cffi_check_int(*o, n, 3494U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_kokai(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_kokai) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_kokai) | 0);  /* check that XKB_KEY_Thai_kokai is an integer */
  if (!_cffi_check_int(*o, n, 3489U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_lakkhangyao(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_lakkhangyao) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_lakkhangyao) | 0);  /* check that XKB_KEY_Thai_lakkhangyao is an integer */
  if (!_cffi_check_int(*o, n, 3557U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_lekchet(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_lekchet) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_lekchet) | 0);  /* check that XKB_KEY_Thai_lekchet is an integer */
  if (!_cffi_check_int(*o, n, 3575U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_lekha(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_lekha) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_lekha) | 0);  /* check that XKB_KEY_Thai_lekha is an integer */
  if (!_cffi_check_int(*o, n, 3573U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_lekhok(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_lekhok) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_lekhok) | 0);  /* check that XKB_KEY_Thai_lekhok is an integer */
  if (!_cffi_check_int(*o, n, 3574U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_lekkao(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_lekkao) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_lekkao) | 0);  /* check that XKB_KEY_Thai_lekkao is an integer */
  if (!_cffi_check_int(*o, n, 3577U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_leknung(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_leknung) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_leknung) | 0);  /* check that XKB_KEY_Thai_leknung is an integer */
  if (!_cffi_check_int(*o, n, 3569U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_lekpaet(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_lekpaet) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_lekpaet) | 0);  /* check that XKB_KEY_Thai_lekpaet is an integer */
  if (!_cffi_check_int(*o, n, 3576U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_leksam(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_leksam) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_leksam) | 0);  /* check that XKB_KEY_Thai_leksam is an integer */
  if (!_cffi_check_int(*o, n, 3571U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_leksi(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_leksi) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_leksi) | 0);  /* check that XKB_KEY_Thai_leksi is an integer */
  if (!_cffi_check_int(*o, n, 3572U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_leksong(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_leksong) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_leksong) | 0);  /* check that XKB_KEY_Thai_leksong is an integer */
  if (!_cffi_check_int(*o, n, 3570U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_leksun(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_leksun) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_leksun) | 0);  /* check that XKB_KEY_Thai_leksun is an integer */
  if (!_cffi_check_int(*o, n, 3568U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_lochula(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_lochula) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_lochula) | 0);  /* check that XKB_KEY_Thai_lochula is an integer */
  if (!_cffi_check_int(*o, n, 3532U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_loling(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_loling) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_loling) | 0);  /* check that XKB_KEY_Thai_loling is an integer */
  if (!_cffi_check_int(*o, n, 3525U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_lu(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_lu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_lu) | 0);  /* check that XKB_KEY_Thai_lu is an integer */
  if (!_cffi_check_int(*o, n, 3526U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_maichattawa(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_maichattawa) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_maichattawa) | 0);  /* check that XKB_KEY_Thai_maichattawa is an integer */
  if (!_cffi_check_int(*o, n, 3563U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_maiek(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_maiek) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_maiek) | 0);  /* check that XKB_KEY_Thai_maiek is an integer */
  if (!_cffi_check_int(*o, n, 3560U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_maihanakat(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_maihanakat) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_maihanakat) | 0);  /* check that XKB_KEY_Thai_maihanakat is an integer */
  if (!_cffi_check_int(*o, n, 3537U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_maihanakat_maitho(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_maihanakat_maitho) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_maihanakat_maitho) | 0);  /* check that XKB_KEY_Thai_maihanakat_maitho is an integer */
  if (!_cffi_check_int(*o, n, 3550U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_maitaikhu(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_maitaikhu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_maitaikhu) | 0);  /* check that XKB_KEY_Thai_maitaikhu is an integer */
  if (!_cffi_check_int(*o, n, 3559U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_maitho(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_maitho) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_maitho) | 0);  /* check that XKB_KEY_Thai_maitho is an integer */
  if (!_cffi_check_int(*o, n, 3561U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_maitri(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_maitri) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_maitri) | 0);  /* check that XKB_KEY_Thai_maitri is an integer */
  if (!_cffi_check_int(*o, n, 3562U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_maiyamok(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_maiyamok) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_maiyamok) | 0);  /* check that XKB_KEY_Thai_maiyamok is an integer */
  if (!_cffi_check_int(*o, n, 3558U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_moma(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_moma) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_moma) | 0);  /* check that XKB_KEY_Thai_moma is an integer */
  if (!_cffi_check_int(*o, n, 3521U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_ngongu(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_ngongu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_ngongu) | 0);  /* check that XKB_KEY_Thai_ngongu is an integer */
  if (!_cffi_check_int(*o, n, 3495U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_nikhahit(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_nikhahit) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_nikhahit) | 0);  /* check that XKB_KEY_Thai_nikhahit is an integer */
  if (!_cffi_check_int(*o, n, 3565U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_nonen(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_nonen) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_nonen) | 0);  /* check that XKB_KEY_Thai_nonen is an integer */
  if (!_cffi_check_int(*o, n, 3507U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_nonu(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_nonu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_nonu) | 0);  /* check that XKB_KEY_Thai_nonu is an integer */
  if (!_cffi_check_int(*o, n, 3513U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_oang(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_oang) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_oang) | 0);  /* check that XKB_KEY_Thai_oang is an integer */
  if (!_cffi_check_int(*o, n, 3533U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_paiyannoi(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_paiyannoi) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_paiyannoi) | 0);  /* check that XKB_KEY_Thai_paiyannoi is an integer */
  if (!_cffi_check_int(*o, n, 3535U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_phinthu(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_phinthu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_phinthu) | 0);  /* check that XKB_KEY_Thai_phinthu is an integer */
  if (!_cffi_check_int(*o, n, 3546U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_phophan(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_phophan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_phophan) | 0);  /* check that XKB_KEY_Thai_phophan is an integer */
  if (!_cffi_check_int(*o, n, 3518U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_phophung(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_phophung) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_phophung) | 0);  /* check that XKB_KEY_Thai_phophung is an integer */
  if (!_cffi_check_int(*o, n, 3516U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_phosamphao(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_phosamphao) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_phosamphao) | 0);  /* check that XKB_KEY_Thai_phosamphao is an integer */
  if (!_cffi_check_int(*o, n, 3520U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_popla(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_popla) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_popla) | 0);  /* check that XKB_KEY_Thai_popla is an integer */
  if (!_cffi_check_int(*o, n, 3515U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_rorua(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_rorua) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_rorua) | 0);  /* check that XKB_KEY_Thai_rorua is an integer */
  if (!_cffi_check_int(*o, n, 3523U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_ru(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_ru) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_ru) | 0);  /* check that XKB_KEY_Thai_ru is an integer */
  if (!_cffi_check_int(*o, n, 3524U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_saraa(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_saraa) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_saraa) | 0);  /* check that XKB_KEY_Thai_saraa is an integer */
  if (!_cffi_check_int(*o, n, 3536U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_saraaa(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_saraaa) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_saraaa) | 0);  /* check that XKB_KEY_Thai_saraaa is an integer */
  if (!_cffi_check_int(*o, n, 3538U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_saraae(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_saraae) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_saraae) | 0);  /* check that XKB_KEY_Thai_saraae is an integer */
  if (!_cffi_check_int(*o, n, 3553U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_saraaimaimalai(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_saraaimaimalai) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_saraaimaimalai) | 0);  /* check that XKB_KEY_Thai_saraaimaimalai is an integer */
  if (!_cffi_check_int(*o, n, 3556U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_saraaimaimuan(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_saraaimaimuan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_saraaimaimuan) | 0);  /* check that XKB_KEY_Thai_saraaimaimuan is an integer */
  if (!_cffi_check_int(*o, n, 3555U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_saraam(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_saraam) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_saraam) | 0);  /* check that XKB_KEY_Thai_saraam is an integer */
  if (!_cffi_check_int(*o, n, 3539U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_sarae(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_sarae) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_sarae) | 0);  /* check that XKB_KEY_Thai_sarae is an integer */
  if (!_cffi_check_int(*o, n, 3552U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_sarai(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_sarai) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_sarai) | 0);  /* check that XKB_KEY_Thai_sarai is an integer */
  if (!_cffi_check_int(*o, n, 3540U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_saraii(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_saraii) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_saraii) | 0);  /* check that XKB_KEY_Thai_saraii is an integer */
  if (!_cffi_check_int(*o, n, 3541U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_sarao(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_sarao) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_sarao) | 0);  /* check that XKB_KEY_Thai_sarao is an integer */
  if (!_cffi_check_int(*o, n, 3554U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_sarau(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_sarau) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_sarau) | 0);  /* check that XKB_KEY_Thai_sarau is an integer */
  if (!_cffi_check_int(*o, n, 3544U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_saraue(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_saraue) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_saraue) | 0);  /* check that XKB_KEY_Thai_saraue is an integer */
  if (!_cffi_check_int(*o, n, 3542U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_sarauee(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_sarauee) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_sarauee) | 0);  /* check that XKB_KEY_Thai_sarauee is an integer */
  if (!_cffi_check_int(*o, n, 3543U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_sarauu(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_sarauu) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_sarauu) | 0);  /* check that XKB_KEY_Thai_sarauu is an integer */
  if (!_cffi_check_int(*o, n, 3545U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_sorusi(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_sorusi) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_sorusi) | 0);  /* check that XKB_KEY_Thai_sorusi is an integer */
  if (!_cffi_check_int(*o, n, 3529U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_sosala(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_sosala) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_sosala) | 0);  /* check that XKB_KEY_Thai_sosala is an integer */
  if (!_cffi_check_int(*o, n, 3528U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_soso(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_soso) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_soso) | 0);  /* check that XKB_KEY_Thai_soso is an integer */
  if (!_cffi_check_int(*o, n, 3499U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_sosua(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_sosua) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_sosua) | 0);  /* check that XKB_KEY_Thai_sosua is an integer */
  if (!_cffi_check_int(*o, n, 3530U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_thanthakhat(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_thanthakhat) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_thanthakhat) | 0);  /* check that XKB_KEY_Thai_thanthakhat is an integer */
  if (!_cffi_check_int(*o, n, 3564U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_thonangmontho(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_thonangmontho) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_thonangmontho) | 0);  /* check that XKB_KEY_Thai_thonangmontho is an integer */
  if (!_cffi_check_int(*o, n, 3505U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_thophuthao(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_thophuthao) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_thophuthao) | 0);  /* check that XKB_KEY_Thai_thophuthao is an integer */
  if (!_cffi_check_int(*o, n, 3506U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_thothahan(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_thothahan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_thothahan) | 0);  /* check that XKB_KEY_Thai_thothahan is an integer */
  if (!_cffi_check_int(*o, n, 3511U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_thothan(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_thothan) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_thothan) | 0);  /* check that XKB_KEY_Thai_thothan is an integer */
  if (!_cffi_check_int(*o, n, 3504U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_thothong(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_thothong) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_thothong) | 0);  /* check that XKB_KEY_Thai_thothong is an integer */
  if (!_cffi_check_int(*o, n, 3512U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_thothung(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_thothung) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_thothung) | 0);  /* check that XKB_KEY_Thai_thothung is an integer */
  if (!_cffi_check_int(*o, n, 3510U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_topatak(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_topatak) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_topatak) | 0);  /* check that XKB_KEY_Thai_topatak is an integer */
  if (!_cffi_check_int(*o, n, 3503U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_totao(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_totao) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_totao) | 0);  /* check that XKB_KEY_Thai_totao is an integer */
  if (!_cffi_check_int(*o, n, 3509U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_wowaen(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_wowaen) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_wowaen) | 0);  /* check that XKB_KEY_Thai_wowaen is an integer */
  if (!_cffi_check_int(*o, n, 3527U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_yoyak(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_yoyak) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_yoyak) | 0);  /* check that XKB_KEY_Thai_yoyak is an integer */
  if (!_cffi_check_int(*o, n, 3522U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thai_yoying(unsigned long long *o)
{
  int n = (XKB_KEY_Thai_yoying) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thai_yoying) | 0);  /* check that XKB_KEY_Thai_yoying is an integer */
  if (!_cffi_check_int(*o, n, 3501U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Thorn(unsigned long long *o)
{
  int n = (XKB_KEY_Thorn) <= 0;
  *o = (unsigned long long)((XKB_KEY_Thorn) | 0);  /* check that XKB_KEY_Thorn is an integer */
  if (!_cffi_check_int(*o, n, 222U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Touroku(unsigned long long *o)
{
  int n = (XKB_KEY_Touroku) <= 0;
  *o = (unsigned long long)((XKB_KEY_Touroku) | 0);  /* check that XKB_KEY_Touroku is an integer */
  if (!_cffi_check_int(*o, n, 65323U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Tslash(unsigned long long *o)
{
  int n = (XKB_KEY_Tslash) <= 0;
  *o = (unsigned long long)((XKB_KEY_Tslash) | 0);  /* check that XKB_KEY_Tslash is an integer */
  if (!_cffi_check_int(*o, n, 940U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_U(unsigned long long *o)
{
  int n = (XKB_KEY_U) <= 0;
  *o = (unsigned long long)((XKB_KEY_U) | 0);  /* check that XKB_KEY_U is an integer */
  if (!_cffi_check_int(*o, n, 85U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Uacute(unsigned long long *o)
{
  int n = (XKB_KEY_Uacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Uacute) | 0);  /* check that XKB_KEY_Uacute is an integer */
  if (!_cffi_check_int(*o, n, 218U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ubelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_Ubelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ubelowdot) | 0);  /* check that XKB_KEY_Ubelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785124U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ubreve(unsigned long long *o)
{
  int n = (XKB_KEY_Ubreve) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ubreve) | 0);  /* check that XKB_KEY_Ubreve is an integer */
  if (!_cffi_check_int(*o, n, 733U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ucircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_Ucircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ucircumflex) | 0);  /* check that XKB_KEY_Ucircumflex is an integer */
  if (!_cffi_check_int(*o, n, 219U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Udiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_Udiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Udiaeresis) | 0);  /* check that XKB_KEY_Udiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 220U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Udoubleacute(unsigned long long *o)
{
  int n = (XKB_KEY_Udoubleacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Udoubleacute) | 0);  /* check that XKB_KEY_Udoubleacute is an integer */
  if (!_cffi_check_int(*o, n, 475U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ugrave(unsigned long long *o)
{
  int n = (XKB_KEY_Ugrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ugrave) | 0);  /* check that XKB_KEY_Ugrave is an integer */
  if (!_cffi_check_int(*o, n, 217U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Uhook(unsigned long long *o)
{
  int n = (XKB_KEY_Uhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_Uhook) | 0);  /* check that XKB_KEY_Uhook is an integer */
  if (!_cffi_check_int(*o, n, 16785126U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Uhorn(unsigned long long *o)
{
  int n = (XKB_KEY_Uhorn) <= 0;
  *o = (unsigned long long)((XKB_KEY_Uhorn) | 0);  /* check that XKB_KEY_Uhorn is an integer */
  if (!_cffi_check_int(*o, n, 16777647U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Uhornacute(unsigned long long *o)
{
  int n = (XKB_KEY_Uhornacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Uhornacute) | 0);  /* check that XKB_KEY_Uhornacute is an integer */
  if (!_cffi_check_int(*o, n, 16785128U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Uhornbelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_Uhornbelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Uhornbelowdot) | 0);  /* check that XKB_KEY_Uhornbelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785136U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Uhorngrave(unsigned long long *o)
{
  int n = (XKB_KEY_Uhorngrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_Uhorngrave) | 0);  /* check that XKB_KEY_Uhorngrave is an integer */
  if (!_cffi_check_int(*o, n, 16785130U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Uhornhook(unsigned long long *o)
{
  int n = (XKB_KEY_Uhornhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_Uhornhook) | 0);  /* check that XKB_KEY_Uhornhook is an integer */
  if (!_cffi_check_int(*o, n, 16785132U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Uhorntilde(unsigned long long *o)
{
  int n = (XKB_KEY_Uhorntilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Uhorntilde) | 0);  /* check that XKB_KEY_Uhorntilde is an integer */
  if (!_cffi_check_int(*o, n, 16785134U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukrainian_GHE_WITH_UPTURN(unsigned long long *o)
{
  int n = (XKB_KEY_Ukrainian_GHE_WITH_UPTURN) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukrainian_GHE_WITH_UPTURN) | 0);  /* check that XKB_KEY_Ukrainian_GHE_WITH_UPTURN is an integer */
  if (!_cffi_check_int(*o, n, 1725U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukrainian_I(unsigned long long *o)
{
  int n = (XKB_KEY_Ukrainian_I) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukrainian_I) | 0);  /* check that XKB_KEY_Ukrainian_I is an integer */
  if (!_cffi_check_int(*o, n, 1718U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukrainian_IE(unsigned long long *o)
{
  int n = (XKB_KEY_Ukrainian_IE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukrainian_IE) | 0);  /* check that XKB_KEY_Ukrainian_IE is an integer */
  if (!_cffi_check_int(*o, n, 1716U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukrainian_YI(unsigned long long *o)
{
  int n = (XKB_KEY_Ukrainian_YI) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukrainian_YI) | 0);  /* check that XKB_KEY_Ukrainian_YI is an integer */
  if (!_cffi_check_int(*o, n, 1719U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukrainian_ghe_with_upturn(unsigned long long *o)
{
  int n = (XKB_KEY_Ukrainian_ghe_with_upturn) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukrainian_ghe_with_upturn) | 0);  /* check that XKB_KEY_Ukrainian_ghe_with_upturn is an integer */
  if (!_cffi_check_int(*o, n, 1709U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukrainian_i(unsigned long long *o)
{
  int n = (XKB_KEY_Ukrainian_i) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukrainian_i) | 0);  /* check that XKB_KEY_Ukrainian_i is an integer */
  if (!_cffi_check_int(*o, n, 1702U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukrainian_ie(unsigned long long *o)
{
  int n = (XKB_KEY_Ukrainian_ie) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukrainian_ie) | 0);  /* check that XKB_KEY_Ukrainian_ie is an integer */
  if (!_cffi_check_int(*o, n, 1700U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukrainian_yi(unsigned long long *o)
{
  int n = (XKB_KEY_Ukrainian_yi) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukrainian_yi) | 0);  /* check that XKB_KEY_Ukrainian_yi is an integer */
  if (!_cffi_check_int(*o, n, 1703U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukranian_I(unsigned long long *o)
{
  int n = (XKB_KEY_Ukranian_I) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukranian_I) | 0);  /* check that XKB_KEY_Ukranian_I is an integer */
  if (!_cffi_check_int(*o, n, 1718U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukranian_JE(unsigned long long *o)
{
  int n = (XKB_KEY_Ukranian_JE) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukranian_JE) | 0);  /* check that XKB_KEY_Ukranian_JE is an integer */
  if (!_cffi_check_int(*o, n, 1716U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukranian_YI(unsigned long long *o)
{
  int n = (XKB_KEY_Ukranian_YI) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukranian_YI) | 0);  /* check that XKB_KEY_Ukranian_YI is an integer */
  if (!_cffi_check_int(*o, n, 1719U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukranian_i(unsigned long long *o)
{
  int n = (XKB_KEY_Ukranian_i) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukranian_i) | 0);  /* check that XKB_KEY_Ukranian_i is an integer */
  if (!_cffi_check_int(*o, n, 1702U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukranian_je(unsigned long long *o)
{
  int n = (XKB_KEY_Ukranian_je) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukranian_je) | 0);  /* check that XKB_KEY_Ukranian_je is an integer */
  if (!_cffi_check_int(*o, n, 1700U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ukranian_yi(unsigned long long *o)
{
  int n = (XKB_KEY_Ukranian_yi) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ukranian_yi) | 0);  /* check that XKB_KEY_Ukranian_yi is an integer */
  if (!_cffi_check_int(*o, n, 1703U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Umacron(unsigned long long *o)
{
  int n = (XKB_KEY_Umacron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Umacron) | 0);  /* check that XKB_KEY_Umacron is an integer */
  if (!_cffi_check_int(*o, n, 990U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Undo(unsigned long long *o)
{
  int n = (XKB_KEY_Undo) <= 0;
  *o = (unsigned long long)((XKB_KEY_Undo) | 0);  /* check that XKB_KEY_Undo is an integer */
  if (!_cffi_check_int(*o, n, 65381U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Uogonek(unsigned long long *o)
{
  int n = (XKB_KEY_Uogonek) <= 0;
  *o = (unsigned long long)((XKB_KEY_Uogonek) | 0);  /* check that XKB_KEY_Uogonek is an integer */
  if (!_cffi_check_int(*o, n, 985U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Up(unsigned long long *o)
{
  int n = (XKB_KEY_Up) <= 0;
  *o = (unsigned long long)((XKB_KEY_Up) | 0);  /* check that XKB_KEY_Up is an integer */
  if (!_cffi_check_int(*o, n, 65362U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Uring(unsigned long long *o)
{
  int n = (XKB_KEY_Uring) <= 0;
  *o = (unsigned long long)((XKB_KEY_Uring) | 0);  /* check that XKB_KEY_Uring is an integer */
  if (!_cffi_check_int(*o, n, 473U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_User(unsigned long long *o)
{
  int n = (XKB_KEY_User) <= 0;
  *o = (unsigned long long)((XKB_KEY_User) | 0);  /* check that XKB_KEY_User is an integer */
  if (!_cffi_check_int(*o, n, 268500846U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Utilde(unsigned long long *o)
{
  int n = (XKB_KEY_Utilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Utilde) | 0);  /* check that XKB_KEY_Utilde is an integer */
  if (!_cffi_check_int(*o, n, 989U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_V(unsigned long long *o)
{
  int n = (XKB_KEY_V) <= 0;
  *o = (unsigned long long)((XKB_KEY_V) | 0);  /* check that XKB_KEY_V is an integer */
  if (!_cffi_check_int(*o, n, 86U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_VoidSymbol(unsigned long long *o)
{
  int n = (XKB_KEY_VoidSymbol) <= 0;
  *o = (unsigned long long)((XKB_KEY_VoidSymbol) | 0);  /* check that XKB_KEY_VoidSymbol is an integer */
  if (!_cffi_check_int(*o, n, 16777215U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_W(unsigned long long *o)
{
  int n = (XKB_KEY_W) <= 0;
  *o = (unsigned long long)((XKB_KEY_W) | 0);  /* check that XKB_KEY_W is an integer */
  if (!_cffi_check_int(*o, n, 87U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Wacute(unsigned long long *o)
{
  int n = (XKB_KEY_Wacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Wacute) | 0);  /* check that XKB_KEY_Wacute is an integer */
  if (!_cffi_check_int(*o, n, 16785026U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Wcircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_Wcircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_Wcircumflex) | 0);  /* check that XKB_KEY_Wcircumflex is an integer */
  if (!_cffi_check_int(*o, n, 16777588U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Wdiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_Wdiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Wdiaeresis) | 0);  /* check that XKB_KEY_Wdiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 16785028U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Wgrave(unsigned long long *o)
{
  int n = (XKB_KEY_Wgrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_Wgrave) | 0);  /* check that XKB_KEY_Wgrave is an integer */
  if (!_cffi_check_int(*o, n, 16785024U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_WonSign(unsigned long long *o)
{
  int n = (XKB_KEY_WonSign) <= 0;
  *o = (unsigned long long)((XKB_KEY_WonSign) | 0);  /* check that XKB_KEY_WonSign is an integer */
  if (!_cffi_check_int(*o, n, 16785577U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_X(unsigned long long *o)
{
  int n = (XKB_KEY_X) <= 0;
  *o = (unsigned long long)((XKB_KEY_X) | 0);  /* check that XKB_KEY_X is an integer */
  if (!_cffi_check_int(*o, n, 88U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AddFavorite(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AddFavorite) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AddFavorite) | 0);  /* check that XKB_KEY_XF86AddFavorite is an integer */
  if (!_cffi_check_int(*o, n, 269025081U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86ApplicationLeft(unsigned long long *o)
{
  int n = (XKB_KEY_XF86ApplicationLeft) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86ApplicationLeft) | 0);  /* check that XKB_KEY_XF86ApplicationLeft is an integer */
  if (!_cffi_check_int(*o, n, 269025104U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86ApplicationRight(unsigned long long *o)
{
  int n = (XKB_KEY_XF86ApplicationRight) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86ApplicationRight) | 0);  /* check that XKB_KEY_XF86ApplicationRight is an integer */
  if (!_cffi_check_int(*o, n, 269025105U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioCycleTrack(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioCycleTrack) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioCycleTrack) | 0);  /* check that XKB_KEY_XF86AudioCycleTrack is an integer */
  if (!_cffi_check_int(*o, n, 269025179U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioForward(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioForward) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioForward) | 0);  /* check that XKB_KEY_XF86AudioForward is an integer */
  if (!_cffi_check_int(*o, n, 269025175U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioLowerVolume(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioLowerVolume) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioLowerVolume) | 0);  /* check that XKB_KEY_XF86AudioLowerVolume is an integer */
  if (!_cffi_check_int(*o, n, 269025041U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioMedia(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioMedia) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioMedia) | 0);  /* check that XKB_KEY_XF86AudioMedia is an integer */
  if (!_cffi_check_int(*o, n, 269025074U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioMicMute(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioMicMute) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioMicMute) | 0);  /* check that XKB_KEY_XF86AudioMicMute is an integer */
  if (!_cffi_check_int(*o, n, 269025202U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioMute(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioMute) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioMute) | 0);  /* check that XKB_KEY_XF86AudioMute is an integer */
  if (!_cffi_check_int(*o, n, 269025042U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioNext(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioNext) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioNext) | 0);  /* check that XKB_KEY_XF86AudioNext is an integer */
  if (!_cffi_check_int(*o, n, 269025047U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioPause(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioPause) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioPause) | 0);  /* check that XKB_KEY_XF86AudioPause is an integer */
  if (!_cffi_check_int(*o, n, 269025073U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioPlay(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioPlay) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioPlay) | 0);  /* check that XKB_KEY_XF86AudioPlay is an integer */
  if (!_cffi_check_int(*o, n, 269025044U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioPrev(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioPrev) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioPrev) | 0);  /* check that XKB_KEY_XF86AudioPrev is an integer */
  if (!_cffi_check_int(*o, n, 269025046U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioRaiseVolume(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioRaiseVolume) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioRaiseVolume) | 0);  /* check that XKB_KEY_XF86AudioRaiseVolume is an integer */
  if (!_cffi_check_int(*o, n, 269025043U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioRandomPlay(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioRandomPlay) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioRandomPlay) | 0);  /* check that XKB_KEY_XF86AudioRandomPlay is an integer */
  if (!_cffi_check_int(*o, n, 269025177U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioRecord(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioRecord) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioRecord) | 0);  /* check that XKB_KEY_XF86AudioRecord is an integer */
  if (!_cffi_check_int(*o, n, 269025052U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioRepeat(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioRepeat) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioRepeat) | 0);  /* check that XKB_KEY_XF86AudioRepeat is an integer */
  if (!_cffi_check_int(*o, n, 269025176U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioRewind(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioRewind) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioRewind) | 0);  /* check that XKB_KEY_XF86AudioRewind is an integer */
  if (!_cffi_check_int(*o, n, 269025086U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86AudioStop(unsigned long long *o)
{
  int n = (XKB_KEY_XF86AudioStop) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86AudioStop) | 0);  /* check that XKB_KEY_XF86AudioStop is an integer */
  if (!_cffi_check_int(*o, n, 269025045U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Away(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Away) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Away) | 0);  /* check that XKB_KEY_XF86Away is an integer */
  if (!_cffi_check_int(*o, n, 269025165U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Back(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Back) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Back) | 0);  /* check that XKB_KEY_XF86Back is an integer */
  if (!_cffi_check_int(*o, n, 269025062U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86BackForward(unsigned long long *o)
{
  int n = (XKB_KEY_XF86BackForward) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86BackForward) | 0);  /* check that XKB_KEY_XF86BackForward is an integer */
  if (!_cffi_check_int(*o, n, 269025087U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Battery(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Battery) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Battery) | 0);  /* check that XKB_KEY_XF86Battery is an integer */
  if (!_cffi_check_int(*o, n, 269025171U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Blue(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Blue) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Blue) | 0);  /* check that XKB_KEY_XF86Blue is an integer */
  if (!_cffi_check_int(*o, n, 269025190U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Bluetooth(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Bluetooth) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Bluetooth) | 0);  /* check that XKB_KEY_XF86Bluetooth is an integer */
  if (!_cffi_check_int(*o, n, 269025172U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Book(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Book) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Book) | 0);  /* check that XKB_KEY_XF86Book is an integer */
  if (!_cffi_check_int(*o, n, 269025106U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86BrightnessAdjust(unsigned long long *o)
{
  int n = (XKB_KEY_XF86BrightnessAdjust) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86BrightnessAdjust) | 0);  /* check that XKB_KEY_XF86BrightnessAdjust is an integer */
  if (!_cffi_check_int(*o, n, 269025083U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86CD(unsigned long long *o)
{
  int n = (XKB_KEY_XF86CD) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86CD) | 0);  /* check that XKB_KEY_XF86CD is an integer */
  if (!_cffi_check_int(*o, n, 269025107U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Calculater(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Calculater) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Calculater) | 0);  /* check that XKB_KEY_XF86Calculater is an integer */
  if (!_cffi_check_int(*o, n, 269025108U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Calculator(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Calculator) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Calculator) | 0);  /* check that XKB_KEY_XF86Calculator is an integer */
  if (!_cffi_check_int(*o, n, 269025053U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Calendar(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Calendar) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Calendar) | 0);  /* check that XKB_KEY_XF86Calendar is an integer */
  if (!_cffi_check_int(*o, n, 269025056U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Clear(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Clear) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Clear) | 0);  /* check that XKB_KEY_XF86Clear is an integer */
  if (!_cffi_check_int(*o, n, 269025109U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86ClearGrab(unsigned long long *o)
{
  int n = (XKB_KEY_XF86ClearGrab) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86ClearGrab) | 0);  /* check that XKB_KEY_XF86ClearGrab is an integer */
  if (!_cffi_check_int(*o, n, 269024801U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Close(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Close) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Close) | 0);  /* check that XKB_KEY_XF86Close is an integer */
  if (!_cffi_check_int(*o, n, 269025110U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Community(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Community) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Community) | 0);  /* check that XKB_KEY_XF86Community is an integer */
  if (!_cffi_check_int(*o, n, 269025085U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86ContrastAdjust(unsigned long long *o)
{
  int n = (XKB_KEY_XF86ContrastAdjust) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86ContrastAdjust) | 0);  /* check that XKB_KEY_XF86ContrastAdjust is an integer */
  if (!_cffi_check_int(*o, n, 269025058U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Copy(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Copy) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Copy) | 0);  /* check that XKB_KEY_XF86Copy is an integer */
  if (!_cffi_check_int(*o, n, 269025111U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Cut(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Cut) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Cut) | 0);  /* check that XKB_KEY_XF86Cut is an integer */
  if (!_cffi_check_int(*o, n, 269025112U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86CycleAngle(unsigned long long *o)
{
  int n = (XKB_KEY_XF86CycleAngle) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86CycleAngle) | 0);  /* check that XKB_KEY_XF86CycleAngle is an integer */
  if (!_cffi_check_int(*o, n, 269025180U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86DOS(unsigned long long *o)
{
  int n = (XKB_KEY_XF86DOS) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86DOS) | 0);  /* check that XKB_KEY_XF86DOS is an integer */
  if (!_cffi_check_int(*o, n, 269025114U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Display(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Display) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Display) | 0);  /* check that XKB_KEY_XF86Display is an integer */
  if (!_cffi_check_int(*o, n, 269025113U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Documents(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Documents) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Documents) | 0);  /* check that XKB_KEY_XF86Documents is an integer */
  if (!_cffi_check_int(*o, n, 269025115U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Eject(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Eject) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Eject) | 0);  /* check that XKB_KEY_XF86Eject is an integer */
  if (!_cffi_check_int(*o, n, 269025068U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Excel(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Excel) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Excel) | 0);  /* check that XKB_KEY_XF86Excel is an integer */
  if (!_cffi_check_int(*o, n, 269025116U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Explorer(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Explorer) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Explorer) | 0);  /* check that XKB_KEY_XF86Explorer is an integer */
  if (!_cffi_check_int(*o, n, 269025117U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Favorites(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Favorites) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Favorites) | 0);  /* check that XKB_KEY_XF86Favorites is an integer */
  if (!_cffi_check_int(*o, n, 269025072U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Finance(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Finance) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Finance) | 0);  /* check that XKB_KEY_XF86Finance is an integer */
  if (!_cffi_check_int(*o, n, 269025084U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Forward(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Forward) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Forward) | 0);  /* check that XKB_KEY_XF86Forward is an integer */
  if (!_cffi_check_int(*o, n, 269025063U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86FrameBack(unsigned long long *o)
{
  int n = (XKB_KEY_XF86FrameBack) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86FrameBack) | 0);  /* check that XKB_KEY_XF86FrameBack is an integer */
  if (!_cffi_check_int(*o, n, 269025181U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86FrameForward(unsigned long long *o)
{
  int n = (XKB_KEY_XF86FrameForward) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86FrameForward) | 0);  /* check that XKB_KEY_XF86FrameForward is an integer */
  if (!_cffi_check_int(*o, n, 269025182U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Game(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Game) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Game) | 0);  /* check that XKB_KEY_XF86Game is an integer */
  if (!_cffi_check_int(*o, n, 269025118U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Go(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Go) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Go) | 0);  /* check that XKB_KEY_XF86Go is an integer */
  if (!_cffi_check_int(*o, n, 269025119U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Green(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Green) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Green) | 0);  /* check that XKB_KEY_XF86Green is an integer */
  if (!_cffi_check_int(*o, n, 269025188U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Hibernate(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Hibernate) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Hibernate) | 0);  /* check that XKB_KEY_XF86Hibernate is an integer */
  if (!_cffi_check_int(*o, n, 269025192U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86History(unsigned long long *o)
{
  int n = (XKB_KEY_XF86History) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86History) | 0);  /* check that XKB_KEY_XF86History is an integer */
  if (!_cffi_check_int(*o, n, 269025079U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86HomePage(unsigned long long *o)
{
  int n = (XKB_KEY_XF86HomePage) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86HomePage) | 0);  /* check that XKB_KEY_XF86HomePage is an integer */
  if (!_cffi_check_int(*o, n, 269025048U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86HotLinks(unsigned long long *o)
{
  int n = (XKB_KEY_XF86HotLinks) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86HotLinks) | 0);  /* check that XKB_KEY_XF86HotLinks is an integer */
  if (!_cffi_check_int(*o, n, 269025082U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86KbdBrightnessDown(unsigned long long *o)
{
  int n = (XKB_KEY_XF86KbdBrightnessDown) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86KbdBrightnessDown) | 0);  /* check that XKB_KEY_XF86KbdBrightnessDown is an integer */
  if (!_cffi_check_int(*o, n, 269025030U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86KbdBrightnessUp(unsigned long long *o)
{
  int n = (XKB_KEY_XF86KbdBrightnessUp) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86KbdBrightnessUp) | 0);  /* check that XKB_KEY_XF86KbdBrightnessUp is an integer */
  if (!_cffi_check_int(*o, n, 269025029U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86KbdLightOnOff(unsigned long long *o)
{
  int n = (XKB_KEY_XF86KbdLightOnOff) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86KbdLightOnOff) | 0);  /* check that XKB_KEY_XF86KbdLightOnOff is an integer */
  if (!_cffi_check_int(*o, n, 269025028U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Launch0(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Launch0) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Launch0) | 0);  /* check that XKB_KEY_XF86Launch0 is an integer */
  if (!_cffi_check_int(*o, n, 269025088U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Launch1(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Launch1) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Launch1) | 0);  /* check that XKB_KEY_XF86Launch1 is an integer */
  if (!_cffi_check_int(*o, n, 269025089U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Launch2(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Launch2) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Launch2) | 0);  /* check that XKB_KEY_XF86Launch2 is an integer */
  if (!_cffi_check_int(*o, n, 269025090U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Launch3(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Launch3) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Launch3) | 0);  /* check that XKB_KEY_XF86Launch3 is an integer */
  if (!_cffi_check_int(*o, n, 269025091U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Launch4(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Launch4) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Launch4) | 0);  /* check that XKB_KEY_XF86Launch4 is an integer */
  if (!_cffi_check_int(*o, n, 269025092U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Launch5(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Launch5) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Launch5) | 0);  /* check that XKB_KEY_XF86Launch5 is an integer */
  if (!_cffi_check_int(*o, n, 269025093U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Launch6(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Launch6) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Launch6) | 0);  /* check that XKB_KEY_XF86Launch6 is an integer */
  if (!_cffi_check_int(*o, n, 269025094U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Launch7(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Launch7) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Launch7) | 0);  /* check that XKB_KEY_XF86Launch7 is an integer */
  if (!_cffi_check_int(*o, n, 269025095U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Launch8(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Launch8) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Launch8) | 0);  /* check that XKB_KEY_XF86Launch8 is an integer */
  if (!_cffi_check_int(*o, n, 269025096U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Launch9(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Launch9) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Launch9) | 0);  /* check that XKB_KEY_XF86Launch9 is an integer */
  if (!_cffi_check_int(*o, n, 269025097U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86LaunchA(unsigned long long *o)
{
  int n = (XKB_KEY_XF86LaunchA) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86LaunchA) | 0);  /* check that XKB_KEY_XF86LaunchA is an integer */
  if (!_cffi_check_int(*o, n, 269025098U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86LaunchB(unsigned long long *o)
{
  int n = (XKB_KEY_XF86LaunchB) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86LaunchB) | 0);  /* check that XKB_KEY_XF86LaunchB is an integer */
  if (!_cffi_check_int(*o, n, 269025099U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86LaunchC(unsigned long long *o)
{
  int n = (XKB_KEY_XF86LaunchC) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86LaunchC) | 0);  /* check that XKB_KEY_XF86LaunchC is an integer */
  if (!_cffi_check_int(*o, n, 269025100U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86LaunchD(unsigned long long *o)
{
  int n = (XKB_KEY_XF86LaunchD) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86LaunchD) | 0);  /* check that XKB_KEY_XF86LaunchD is an integer */
  if (!_cffi_check_int(*o, n, 269025101U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86LaunchE(unsigned long long *o)
{
  int n = (XKB_KEY_XF86LaunchE) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86LaunchE) | 0);  /* check that XKB_KEY_XF86LaunchE is an integer */
  if (!_cffi_check_int(*o, n, 269025102U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86LaunchF(unsigned long long *o)
{
  int n = (XKB_KEY_XF86LaunchF) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86LaunchF) | 0);  /* check that XKB_KEY_XF86LaunchF is an integer */
  if (!_cffi_check_int(*o, n, 269025103U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86LightBulb(unsigned long long *o)
{
  int n = (XKB_KEY_XF86LightBulb) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86LightBulb) | 0);  /* check that XKB_KEY_XF86LightBulb is an integer */
  if (!_cffi_check_int(*o, n, 269025077U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86LogGrabInfo(unsigned long long *o)
{
  int n = (XKB_KEY_XF86LogGrabInfo) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86LogGrabInfo) | 0);  /* check that XKB_KEY_XF86LogGrabInfo is an integer */
  if (!_cffi_check_int(*o, n, 269024805U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86LogOff(unsigned long long *o)
{
  int n = (XKB_KEY_XF86LogOff) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86LogOff) | 0);  /* check that XKB_KEY_XF86LogOff is an integer */
  if (!_cffi_check_int(*o, n, 269025121U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86LogWindowTree(unsigned long long *o)
{
  int n = (XKB_KEY_XF86LogWindowTree) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86LogWindowTree) | 0);  /* check that XKB_KEY_XF86LogWindowTree is an integer */
  if (!_cffi_check_int(*o, n, 269024804U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Mail(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Mail) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Mail) | 0);  /* check that XKB_KEY_XF86Mail is an integer */
  if (!_cffi_check_int(*o, n, 269025049U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86MailForward(unsigned long long *o)
{
  int n = (XKB_KEY_XF86MailForward) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86MailForward) | 0);  /* check that XKB_KEY_XF86MailForward is an integer */
  if (!_cffi_check_int(*o, n, 269025168U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Market(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Market) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Market) | 0);  /* check that XKB_KEY_XF86Market is an integer */
  if (!_cffi_check_int(*o, n, 269025122U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Meeting(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Meeting) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Meeting) | 0);  /* check that XKB_KEY_XF86Meeting is an integer */
  if (!_cffi_check_int(*o, n, 269025123U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Memo(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Memo) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Memo) | 0);  /* check that XKB_KEY_XF86Memo is an integer */
  if (!_cffi_check_int(*o, n, 269025054U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86MenuKB(unsigned long long *o)
{
  int n = (XKB_KEY_XF86MenuKB) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86MenuKB) | 0);  /* check that XKB_KEY_XF86MenuKB is an integer */
  if (!_cffi_check_int(*o, n, 269025125U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86MenuPB(unsigned long long *o)
{
  int n = (XKB_KEY_XF86MenuPB) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86MenuPB) | 0);  /* check that XKB_KEY_XF86MenuPB is an integer */
  if (!_cffi_check_int(*o, n, 269025126U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Messenger(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Messenger) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Messenger) | 0);  /* check that XKB_KEY_XF86Messenger is an integer */
  if (!_cffi_check_int(*o, n, 269025166U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86ModeLock(unsigned long long *o)
{
  int n = (XKB_KEY_XF86ModeLock) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86ModeLock) | 0);  /* check that XKB_KEY_XF86ModeLock is an integer */
  if (!_cffi_check_int(*o, n, 269025025U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86MonBrightnessDown(unsigned long long *o)
{
  int n = (XKB_KEY_XF86MonBrightnessDown) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86MonBrightnessDown) | 0);  /* check that XKB_KEY_XF86MonBrightnessDown is an integer */
  if (!_cffi_check_int(*o, n, 269025027U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86MonBrightnessUp(unsigned long long *o)
{
  int n = (XKB_KEY_XF86MonBrightnessUp) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86MonBrightnessUp) | 0);  /* check that XKB_KEY_XF86MonBrightnessUp is an integer */
  if (!_cffi_check_int(*o, n, 269025026U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Music(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Music) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Music) | 0);  /* check that XKB_KEY_XF86Music is an integer */
  if (!_cffi_check_int(*o, n, 269025170U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86MyComputer(unsigned long long *o)
{
  int n = (XKB_KEY_XF86MyComputer) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86MyComputer) | 0);  /* check that XKB_KEY_XF86MyComputer is an integer */
  if (!_cffi_check_int(*o, n, 269025075U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86MySites(unsigned long long *o)
{
  int n = (XKB_KEY_XF86MySites) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86MySites) | 0);  /* check that XKB_KEY_XF86MySites is an integer */
  if (!_cffi_check_int(*o, n, 269025127U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86New(unsigned long long *o)
{
  int n = (XKB_KEY_XF86New) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86New) | 0);  /* check that XKB_KEY_XF86New is an integer */
  if (!_cffi_check_int(*o, n, 269025128U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86News(unsigned long long *o)
{
  int n = (XKB_KEY_XF86News) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86News) | 0);  /* check that XKB_KEY_XF86News is an integer */
  if (!_cffi_check_int(*o, n, 269025129U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Next_VMode(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Next_VMode) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Next_VMode) | 0);  /* check that XKB_KEY_XF86Next_VMode is an integer */
  if (!_cffi_check_int(*o, n, 269024802U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86OfficeHome(unsigned long long *o)
{
  int n = (XKB_KEY_XF86OfficeHome) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86OfficeHome) | 0);  /* check that XKB_KEY_XF86OfficeHome is an integer */
  if (!_cffi_check_int(*o, n, 269025130U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Open(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Open) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Open) | 0);  /* check that XKB_KEY_XF86Open is an integer */
  if (!_cffi_check_int(*o, n, 269025131U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86OpenURL(unsigned long long *o)
{
  int n = (XKB_KEY_XF86OpenURL) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86OpenURL) | 0);  /* check that XKB_KEY_XF86OpenURL is an integer */
  if (!_cffi_check_int(*o, n, 269025080U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Option(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Option) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Option) | 0);  /* check that XKB_KEY_XF86Option is an integer */
  if (!_cffi_check_int(*o, n, 269025132U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Paste(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Paste) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Paste) | 0);  /* check that XKB_KEY_XF86Paste is an integer */
  if (!_cffi_check_int(*o, n, 269025133U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Phone(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Phone) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Phone) | 0);  /* check that XKB_KEY_XF86Phone is an integer */
  if (!_cffi_check_int(*o, n, 269025134U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Pictures(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Pictures) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Pictures) | 0);  /* check that XKB_KEY_XF86Pictures is an integer */
  if (!_cffi_check_int(*o, n, 269025169U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86PowerDown(unsigned long long *o)
{
  int n = (XKB_KEY_XF86PowerDown) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86PowerDown) | 0);  /* check that XKB_KEY_XF86PowerDown is an integer */
  if (!_cffi_check_int(*o, n, 269025057U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86PowerOff(unsigned long long *o)
{
  int n = (XKB_KEY_XF86PowerOff) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86PowerOff) | 0);  /* check that XKB_KEY_XF86PowerOff is an integer */
  if (!_cffi_check_int(*o, n, 269025066U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Prev_VMode(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Prev_VMode) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Prev_VMode) | 0);  /* check that XKB_KEY_XF86Prev_VMode is an integer */
  if (!_cffi_check_int(*o, n, 269024803U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Q(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Q) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Q) | 0);  /* check that XKB_KEY_XF86Q is an integer */
  if (!_cffi_check_int(*o, n, 269025136U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Red(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Red) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Red) | 0);  /* check that XKB_KEY_XF86Red is an integer */
  if (!_cffi_check_int(*o, n, 269025187U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Refresh(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Refresh) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Refresh) | 0);  /* check that XKB_KEY_XF86Refresh is an integer */
  if (!_cffi_check_int(*o, n, 269025065U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Reload(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Reload) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Reload) | 0);  /* check that XKB_KEY_XF86Reload is an integer */
  if (!_cffi_check_int(*o, n, 269025139U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Reply(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Reply) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Reply) | 0);  /* check that XKB_KEY_XF86Reply is an integer */
  if (!_cffi_check_int(*o, n, 269025138U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86RockerDown(unsigned long long *o)
{
  int n = (XKB_KEY_XF86RockerDown) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86RockerDown) | 0);  /* check that XKB_KEY_XF86RockerDown is an integer */
  if (!_cffi_check_int(*o, n, 269025060U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86RockerEnter(unsigned long long *o)
{
  int n = (XKB_KEY_XF86RockerEnter) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86RockerEnter) | 0);  /* check that XKB_KEY_XF86RockerEnter is an integer */
  if (!_cffi_check_int(*o, n, 269025061U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86RockerUp(unsigned long long *o)
{
  int n = (XKB_KEY_XF86RockerUp) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86RockerUp) | 0);  /* check that XKB_KEY_XF86RockerUp is an integer */
  if (!_cffi_check_int(*o, n, 269025059U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86RotateWindows(unsigned long long *o)
{
  int n = (XKB_KEY_XF86RotateWindows) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86RotateWindows) | 0);  /* check that XKB_KEY_XF86RotateWindows is an integer */
  if (!_cffi_check_int(*o, n, 269025140U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86RotationKB(unsigned long long *o)
{
  int n = (XKB_KEY_XF86RotationKB) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86RotationKB) | 0);  /* check that XKB_KEY_XF86RotationKB is an integer */
  if (!_cffi_check_int(*o, n, 269025142U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86RotationPB(unsigned long long *o)
{
  int n = (XKB_KEY_XF86RotationPB) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86RotationPB) | 0);  /* check that XKB_KEY_XF86RotationPB is an integer */
  if (!_cffi_check_int(*o, n, 269025141U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Save(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Save) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Save) | 0);  /* check that XKB_KEY_XF86Save is an integer */
  if (!_cffi_check_int(*o, n, 269025143U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86ScreenSaver(unsigned long long *o)
{
  int n = (XKB_KEY_XF86ScreenSaver) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86ScreenSaver) | 0);  /* check that XKB_KEY_XF86ScreenSaver is an integer */
  if (!_cffi_check_int(*o, n, 269025069U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86ScrollClick(unsigned long long *o)
{
  int n = (XKB_KEY_XF86ScrollClick) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86ScrollClick) | 0);  /* check that XKB_KEY_XF86ScrollClick is an integer */
  if (!_cffi_check_int(*o, n, 269025146U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86ScrollDown(unsigned long long *o)
{
  int n = (XKB_KEY_XF86ScrollDown) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86ScrollDown) | 0);  /* check that XKB_KEY_XF86ScrollDown is an integer */
  if (!_cffi_check_int(*o, n, 269025145U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86ScrollUp(unsigned long long *o)
{
  int n = (XKB_KEY_XF86ScrollUp) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86ScrollUp) | 0);  /* check that XKB_KEY_XF86ScrollUp is an integer */
  if (!_cffi_check_int(*o, n, 269025144U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Search(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Search) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Search) | 0);  /* check that XKB_KEY_XF86Search is an integer */
  if (!_cffi_check_int(*o, n, 269025051U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Select(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Select) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Select) | 0);  /* check that XKB_KEY_XF86Select is an integer */
  if (!_cffi_check_int(*o, n, 269025184U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Send(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Send) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Send) | 0);  /* check that XKB_KEY_XF86Send is an integer */
  if (!_cffi_check_int(*o, n, 269025147U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Shop(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Shop) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Shop) | 0);  /* check that XKB_KEY_XF86Shop is an integer */
  if (!_cffi_check_int(*o, n, 269025078U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Sleep(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Sleep) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Sleep) | 0);  /* check that XKB_KEY_XF86Sleep is an integer */
  if (!_cffi_check_int(*o, n, 269025071U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Spell(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Spell) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Spell) | 0);  /* check that XKB_KEY_XF86Spell is an integer */
  if (!_cffi_check_int(*o, n, 269025148U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86SplitScreen(unsigned long long *o)
{
  int n = (XKB_KEY_XF86SplitScreen) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86SplitScreen) | 0);  /* check that XKB_KEY_XF86SplitScreen is an integer */
  if (!_cffi_check_int(*o, n, 269025149U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Standby(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Standby) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Standby) | 0);  /* check that XKB_KEY_XF86Standby is an integer */
  if (!_cffi_check_int(*o, n, 269025040U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Start(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Start) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Start) | 0);  /* check that XKB_KEY_XF86Start is an integer */
  if (!_cffi_check_int(*o, n, 269025050U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Stop(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Stop) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Stop) | 0);  /* check that XKB_KEY_XF86Stop is an integer */
  if (!_cffi_check_int(*o, n, 269025064U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Subtitle(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Subtitle) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Subtitle) | 0);  /* check that XKB_KEY_XF86Subtitle is an integer */
  if (!_cffi_check_int(*o, n, 269025178U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Support(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Support) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Support) | 0);  /* check that XKB_KEY_XF86Support is an integer */
  if (!_cffi_check_int(*o, n, 269025150U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Suspend(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Suspend) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Suspend) | 0);  /* check that XKB_KEY_XF86Suspend is an integer */
  if (!_cffi_check_int(*o, n, 269025191U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Switch_VT_1(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Switch_VT_1) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Switch_VT_1) | 0);  /* check that XKB_KEY_XF86Switch_VT_1 is an integer */
  if (!_cffi_check_int(*o, n, 269024769U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Switch_VT_10(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Switch_VT_10) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Switch_VT_10) | 0);  /* check that XKB_KEY_XF86Switch_VT_10 is an integer */
  if (!_cffi_check_int(*o, n, 269024778U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Switch_VT_11(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Switch_VT_11) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Switch_VT_11) | 0);  /* check that XKB_KEY_XF86Switch_VT_11 is an integer */
  if (!_cffi_check_int(*o, n, 269024779U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Switch_VT_12(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Switch_VT_12) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Switch_VT_12) | 0);  /* check that XKB_KEY_XF86Switch_VT_12 is an integer */
  if (!_cffi_check_int(*o, n, 269024780U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Switch_VT_2(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Switch_VT_2) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Switch_VT_2) | 0);  /* check that XKB_KEY_XF86Switch_VT_2 is an integer */
  if (!_cffi_check_int(*o, n, 269024770U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Switch_VT_3(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Switch_VT_3) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Switch_VT_3) | 0);  /* check that XKB_KEY_XF86Switch_VT_3 is an integer */
  if (!_cffi_check_int(*o, n, 269024771U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Switch_VT_4(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Switch_VT_4) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Switch_VT_4) | 0);  /* check that XKB_KEY_XF86Switch_VT_4 is an integer */
  if (!_cffi_check_int(*o, n, 269024772U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Switch_VT_5(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Switch_VT_5) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Switch_VT_5) | 0);  /* check that XKB_KEY_XF86Switch_VT_5 is an integer */
  if (!_cffi_check_int(*o, n, 269024773U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Switch_VT_6(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Switch_VT_6) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Switch_VT_6) | 0);  /* check that XKB_KEY_XF86Switch_VT_6 is an integer */
  if (!_cffi_check_int(*o, n, 269024774U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Switch_VT_7(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Switch_VT_7) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Switch_VT_7) | 0);  /* check that XKB_KEY_XF86Switch_VT_7 is an integer */
  if (!_cffi_check_int(*o, n, 269024775U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Switch_VT_8(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Switch_VT_8) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Switch_VT_8) | 0);  /* check that XKB_KEY_XF86Switch_VT_8 is an integer */
  if (!_cffi_check_int(*o, n, 269024776U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Switch_VT_9(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Switch_VT_9) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Switch_VT_9) | 0);  /* check that XKB_KEY_XF86Switch_VT_9 is an integer */
  if (!_cffi_check_int(*o, n, 269024777U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86TaskPane(unsigned long long *o)
{
  int n = (XKB_KEY_XF86TaskPane) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86TaskPane) | 0);  /* check that XKB_KEY_XF86TaskPane is an integer */
  if (!_cffi_check_int(*o, n, 269025151U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Terminal(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Terminal) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Terminal) | 0);  /* check that XKB_KEY_XF86Terminal is an integer */
  if (!_cffi_check_int(*o, n, 269025152U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Time(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Time) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Time) | 0);  /* check that XKB_KEY_XF86Time is an integer */
  if (!_cffi_check_int(*o, n, 269025183U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86ToDoList(unsigned long long *o)
{
  int n = (XKB_KEY_XF86ToDoList) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86ToDoList) | 0);  /* check that XKB_KEY_XF86ToDoList is an integer */
  if (!_cffi_check_int(*o, n, 269025055U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Tools(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Tools) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Tools) | 0);  /* check that XKB_KEY_XF86Tools is an integer */
  if (!_cffi_check_int(*o, n, 269025153U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86TopMenu(unsigned long long *o)
{
  int n = (XKB_KEY_XF86TopMenu) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86TopMenu) | 0);  /* check that XKB_KEY_XF86TopMenu is an integer */
  if (!_cffi_check_int(*o, n, 269025186U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86TouchpadOff(unsigned long long *o)
{
  int n = (XKB_KEY_XF86TouchpadOff) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86TouchpadOff) | 0);  /* check that XKB_KEY_XF86TouchpadOff is an integer */
  if (!_cffi_check_int(*o, n, 269025201U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86TouchpadOn(unsigned long long *o)
{
  int n = (XKB_KEY_XF86TouchpadOn) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86TouchpadOn) | 0);  /* check that XKB_KEY_XF86TouchpadOn is an integer */
  if (!_cffi_check_int(*o, n, 269025200U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86TouchpadToggle(unsigned long long *o)
{
  int n = (XKB_KEY_XF86TouchpadToggle) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86TouchpadToggle) | 0);  /* check that XKB_KEY_XF86TouchpadToggle is an integer */
  if (!_cffi_check_int(*o, n, 269025193U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Travel(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Travel) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Travel) | 0);  /* check that XKB_KEY_XF86Travel is an integer */
  if (!_cffi_check_int(*o, n, 269025154U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86UWB(unsigned long long *o)
{
  int n = (XKB_KEY_XF86UWB) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86UWB) | 0);  /* check that XKB_KEY_XF86UWB is an integer */
  if (!_cffi_check_int(*o, n, 269025174U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Ungrab(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Ungrab) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Ungrab) | 0);  /* check that XKB_KEY_XF86Ungrab is an integer */
  if (!_cffi_check_int(*o, n, 269024800U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86User1KB(unsigned long long *o)
{
  int n = (XKB_KEY_XF86User1KB) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86User1KB) | 0);  /* check that XKB_KEY_XF86User1KB is an integer */
  if (!_cffi_check_int(*o, n, 269025157U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86User2KB(unsigned long long *o)
{
  int n = (XKB_KEY_XF86User2KB) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86User2KB) | 0);  /* check that XKB_KEY_XF86User2KB is an integer */
  if (!_cffi_check_int(*o, n, 269025158U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86UserPB(unsigned long long *o)
{
  int n = (XKB_KEY_XF86UserPB) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86UserPB) | 0);  /* check that XKB_KEY_XF86UserPB is an integer */
  if (!_cffi_check_int(*o, n, 269025156U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86VendorHome(unsigned long long *o)
{
  int n = (XKB_KEY_XF86VendorHome) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86VendorHome) | 0);  /* check that XKB_KEY_XF86VendorHome is an integer */
  if (!_cffi_check_int(*o, n, 269025076U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Video(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Video) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Video) | 0);  /* check that XKB_KEY_XF86Video is an integer */
  if (!_cffi_check_int(*o, n, 269025159U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86View(unsigned long long *o)
{
  int n = (XKB_KEY_XF86View) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86View) | 0);  /* check that XKB_KEY_XF86View is an integer */
  if (!_cffi_check_int(*o, n, 269025185U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86WLAN(unsigned long long *o)
{
  int n = (XKB_KEY_XF86WLAN) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86WLAN) | 0);  /* check that XKB_KEY_XF86WLAN is an integer */
  if (!_cffi_check_int(*o, n, 269025173U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86WWW(unsigned long long *o)
{
  int n = (XKB_KEY_XF86WWW) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86WWW) | 0);  /* check that XKB_KEY_XF86WWW is an integer */
  if (!_cffi_check_int(*o, n, 269025070U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86WakeUp(unsigned long long *o)
{
  int n = (XKB_KEY_XF86WakeUp) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86WakeUp) | 0);  /* check that XKB_KEY_XF86WakeUp is an integer */
  if (!_cffi_check_int(*o, n, 269025067U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86WebCam(unsigned long long *o)
{
  int n = (XKB_KEY_XF86WebCam) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86WebCam) | 0);  /* check that XKB_KEY_XF86WebCam is an integer */
  if (!_cffi_check_int(*o, n, 269025167U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86WheelButton(unsigned long long *o)
{
  int n = (XKB_KEY_XF86WheelButton) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86WheelButton) | 0);  /* check that XKB_KEY_XF86WheelButton is an integer */
  if (!_cffi_check_int(*o, n, 269025160U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Word(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Word) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Word) | 0);  /* check that XKB_KEY_XF86Word is an integer */
  if (!_cffi_check_int(*o, n, 269025161U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Xfer(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Xfer) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Xfer) | 0);  /* check that XKB_KEY_XF86Xfer is an integer */
  if (!_cffi_check_int(*o, n, 269025162U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86Yellow(unsigned long long *o)
{
  int n = (XKB_KEY_XF86Yellow) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86Yellow) | 0);  /* check that XKB_KEY_XF86Yellow is an integer */
  if (!_cffi_check_int(*o, n, 269025189U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86ZoomIn(unsigned long long *o)
{
  int n = (XKB_KEY_XF86ZoomIn) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86ZoomIn) | 0);  /* check that XKB_KEY_XF86ZoomIn is an integer */
  if (!_cffi_check_int(*o, n, 269025163U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86ZoomOut(unsigned long long *o)
{
  int n = (XKB_KEY_XF86ZoomOut) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86ZoomOut) | 0);  /* check that XKB_KEY_XF86ZoomOut is an integer */
  if (!_cffi_check_int(*o, n, 269025164U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_XF86iTouch(unsigned long long *o)
{
  int n = (XKB_KEY_XF86iTouch) <= 0;
  *o = (unsigned long long)((XKB_KEY_XF86iTouch) | 0);  /* check that XKB_KEY_XF86iTouch is an integer */
  if (!_cffi_check_int(*o, n, 269025120U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Xabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_Xabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Xabovedot) | 0);  /* check that XKB_KEY_Xabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16785034U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Y(unsigned long long *o)
{
  int n = (XKB_KEY_Y) <= 0;
  *o = (unsigned long long)((XKB_KEY_Y) | 0);  /* check that XKB_KEY_Y is an integer */
  if (!_cffi_check_int(*o, n, 89U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Yacute(unsigned long long *o)
{
  int n = (XKB_KEY_Yacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Yacute) | 0);  /* check that XKB_KEY_Yacute is an integer */
  if (!_cffi_check_int(*o, n, 221U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ybelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_Ybelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ybelowdot) | 0);  /* check that XKB_KEY_Ybelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785140U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ycircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_Ycircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ycircumflex) | 0);  /* check that XKB_KEY_Ycircumflex is an integer */
  if (!_cffi_check_int(*o, n, 16777590U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ydiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_Ydiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ydiaeresis) | 0);  /* check that XKB_KEY_Ydiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 5054U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ygrave(unsigned long long *o)
{
  int n = (XKB_KEY_Ygrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ygrave) | 0);  /* check that XKB_KEY_Ygrave is an integer */
  if (!_cffi_check_int(*o, n, 16785138U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Yhook(unsigned long long *o)
{
  int n = (XKB_KEY_Yhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_Yhook) | 0);  /* check that XKB_KEY_Yhook is an integer */
  if (!_cffi_check_int(*o, n, 16785142U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Ytilde(unsigned long long *o)
{
  int n = (XKB_KEY_Ytilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_Ytilde) | 0);  /* check that XKB_KEY_Ytilde is an integer */
  if (!_cffi_check_int(*o, n, 16785144U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Z(unsigned long long *o)
{
  int n = (XKB_KEY_Z) <= 0;
  *o = (unsigned long long)((XKB_KEY_Z) | 0);  /* check that XKB_KEY_Z is an integer */
  if (!_cffi_check_int(*o, n, 90U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Zabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_Zabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_Zabovedot) | 0);  /* check that XKB_KEY_Zabovedot is an integer */
  if (!_cffi_check_int(*o, n, 431U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Zacute(unsigned long long *o)
{
  int n = (XKB_KEY_Zacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_Zacute) | 0);  /* check that XKB_KEY_Zacute is an integer */
  if (!_cffi_check_int(*o, n, 428U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Zcaron(unsigned long long *o)
{
  int n = (XKB_KEY_Zcaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_Zcaron) | 0);  /* check that XKB_KEY_Zcaron is an integer */
  if (!_cffi_check_int(*o, n, 430U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Zen_Koho(unsigned long long *o)
{
  int n = (XKB_KEY_Zen_Koho) <= 0;
  *o = (unsigned long long)((XKB_KEY_Zen_Koho) | 0);  /* check that XKB_KEY_Zen_Koho is an integer */
  if (!_cffi_check_int(*o, n, 65341U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Zenkaku(unsigned long long *o)
{
  int n = (XKB_KEY_Zenkaku) <= 0;
  *o = (unsigned long long)((XKB_KEY_Zenkaku) | 0);  /* check that XKB_KEY_Zenkaku is an integer */
  if (!_cffi_check_int(*o, n, 65320U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Zenkaku_Hankaku(unsigned long long *o)
{
  int n = (XKB_KEY_Zenkaku_Hankaku) <= 0;
  *o = (unsigned long long)((XKB_KEY_Zenkaku_Hankaku) | 0);  /* check that XKB_KEY_Zenkaku_Hankaku is an integer */
  if (!_cffi_check_int(*o, n, 65322U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_Zstroke(unsigned long long *o)
{
  int n = (XKB_KEY_Zstroke) <= 0;
  *o = (unsigned long long)((XKB_KEY_Zstroke) | 0);  /* check that XKB_KEY_Zstroke is an integer */
  if (!_cffi_check_int(*o, n, 16777653U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_a(unsigned long long *o)
{
  int n = (XKB_KEY_a) <= 0;
  *o = (unsigned long long)((XKB_KEY_a) | 0);  /* check that XKB_KEY_a is an integer */
  if (!_cffi_check_int(*o, n, 97U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_aacute(unsigned long long *o)
{
  int n = (XKB_KEY_aacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_aacute) | 0);  /* check that XKB_KEY_aacute is an integer */
  if (!_cffi_check_int(*o, n, 225U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_abelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_abelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_abelowdot) | 0);  /* check that XKB_KEY_abelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785057U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_abovedot(unsigned long long *o)
{
  int n = (XKB_KEY_abovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_abovedot) | 0);  /* check that XKB_KEY_abovedot is an integer */
  if (!_cffi_check_int(*o, n, 511U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_abreve(unsigned long long *o)
{
  int n = (XKB_KEY_abreve) <= 0;
  *o = (unsigned long long)((XKB_KEY_abreve) | 0);  /* check that XKB_KEY_abreve is an integer */
  if (!_cffi_check_int(*o, n, 483U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_abreveacute(unsigned long long *o)
{
  int n = (XKB_KEY_abreveacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_abreveacute) | 0);  /* check that XKB_KEY_abreveacute is an integer */
  if (!_cffi_check_int(*o, n, 16785071U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_abrevebelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_abrevebelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_abrevebelowdot) | 0);  /* check that XKB_KEY_abrevebelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785079U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_abrevegrave(unsigned long long *o)
{
  int n = (XKB_KEY_abrevegrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_abrevegrave) | 0);  /* check that XKB_KEY_abrevegrave is an integer */
  if (!_cffi_check_int(*o, n, 16785073U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_abrevehook(unsigned long long *o)
{
  int n = (XKB_KEY_abrevehook) <= 0;
  *o = (unsigned long long)((XKB_KEY_abrevehook) | 0);  /* check that XKB_KEY_abrevehook is an integer */
  if (!_cffi_check_int(*o, n, 16785075U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_abrevetilde(unsigned long long *o)
{
  int n = (XKB_KEY_abrevetilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_abrevetilde) | 0);  /* check that XKB_KEY_abrevetilde is an integer */
  if (!_cffi_check_int(*o, n, 16785077U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_acircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_acircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_acircumflex) | 0);  /* check that XKB_KEY_acircumflex is an integer */
  if (!_cffi_check_int(*o, n, 226U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_acircumflexacute(unsigned long long *o)
{
  int n = (XKB_KEY_acircumflexacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_acircumflexacute) | 0);  /* check that XKB_KEY_acircumflexacute is an integer */
  if (!_cffi_check_int(*o, n, 16785061U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_acircumflexbelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_acircumflexbelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_acircumflexbelowdot) | 0);  /* check that XKB_KEY_acircumflexbelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785069U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_acircumflexgrave(unsigned long long *o)
{
  int n = (XKB_KEY_acircumflexgrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_acircumflexgrave) | 0);  /* check that XKB_KEY_acircumflexgrave is an integer */
  if (!_cffi_check_int(*o, n, 16785063U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_acircumflexhook(unsigned long long *o)
{
  int n = (XKB_KEY_acircumflexhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_acircumflexhook) | 0);  /* check that XKB_KEY_acircumflexhook is an integer */
  if (!_cffi_check_int(*o, n, 16785065U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_acircumflextilde(unsigned long long *o)
{
  int n = (XKB_KEY_acircumflextilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_acircumflextilde) | 0);  /* check that XKB_KEY_acircumflextilde is an integer */
  if (!_cffi_check_int(*o, n, 16785067U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_acute(unsigned long long *o)
{
  int n = (XKB_KEY_acute) <= 0;
  *o = (unsigned long long)((XKB_KEY_acute) | 0);  /* check that XKB_KEY_acute is an integer */
  if (!_cffi_check_int(*o, n, 180U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_adiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_adiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_adiaeresis) | 0);  /* check that XKB_KEY_adiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 228U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ae(unsigned long long *o)
{
  int n = (XKB_KEY_ae) <= 0;
  *o = (unsigned long long)((XKB_KEY_ae) | 0);  /* check that XKB_KEY_ae is an integer */
  if (!_cffi_check_int(*o, n, 230U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_agrave(unsigned long long *o)
{
  int n = (XKB_KEY_agrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_agrave) | 0);  /* check that XKB_KEY_agrave is an integer */
  if (!_cffi_check_int(*o, n, 224U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ahook(unsigned long long *o)
{
  int n = (XKB_KEY_ahook) <= 0;
  *o = (unsigned long long)((XKB_KEY_ahook) | 0);  /* check that XKB_KEY_ahook is an integer */
  if (!_cffi_check_int(*o, n, 16785059U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_amacron(unsigned long long *o)
{
  int n = (XKB_KEY_amacron) <= 0;
  *o = (unsigned long long)((XKB_KEY_amacron) | 0);  /* check that XKB_KEY_amacron is an integer */
  if (!_cffi_check_int(*o, n, 992U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ampersand(unsigned long long *o)
{
  int n = (XKB_KEY_ampersand) <= 0;
  *o = (unsigned long long)((XKB_KEY_ampersand) | 0);  /* check that XKB_KEY_ampersand is an integer */
  if (!_cffi_check_int(*o, n, 38U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_aogonek(unsigned long long *o)
{
  int n = (XKB_KEY_aogonek) <= 0;
  *o = (unsigned long long)((XKB_KEY_aogonek) | 0);  /* check that XKB_KEY_aogonek is an integer */
  if (!_cffi_check_int(*o, n, 433U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_apostrophe(unsigned long long *o)
{
  int n = (XKB_KEY_apostrophe) <= 0;
  *o = (unsigned long long)((XKB_KEY_apostrophe) | 0);  /* check that XKB_KEY_apostrophe is an integer */
  if (!_cffi_check_int(*o, n, 39U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_approxeq(unsigned long long *o)
{
  int n = (XKB_KEY_approxeq) <= 0;
  *o = (unsigned long long)((XKB_KEY_approxeq) | 0);  /* check that XKB_KEY_approxeq is an integer */
  if (!_cffi_check_int(*o, n, 16785992U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_approximate(unsigned long long *o)
{
  int n = (XKB_KEY_approximate) <= 0;
  *o = (unsigned long long)((XKB_KEY_approximate) | 0);  /* check that XKB_KEY_approximate is an integer */
  if (!_cffi_check_int(*o, n, 2248U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_aring(unsigned long long *o)
{
  int n = (XKB_KEY_aring) <= 0;
  *o = (unsigned long long)((XKB_KEY_aring) | 0);  /* check that XKB_KEY_aring is an integer */
  if (!_cffi_check_int(*o, n, 229U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_asciicircum(unsigned long long *o)
{
  int n = (XKB_KEY_asciicircum) <= 0;
  *o = (unsigned long long)((XKB_KEY_asciicircum) | 0);  /* check that XKB_KEY_asciicircum is an integer */
  if (!_cffi_check_int(*o, n, 94U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_asciitilde(unsigned long long *o)
{
  int n = (XKB_KEY_asciitilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_asciitilde) | 0);  /* check that XKB_KEY_asciitilde is an integer */
  if (!_cffi_check_int(*o, n, 126U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_asterisk(unsigned long long *o)
{
  int n = (XKB_KEY_asterisk) <= 0;
  *o = (unsigned long long)((XKB_KEY_asterisk) | 0);  /* check that XKB_KEY_asterisk is an integer */
  if (!_cffi_check_int(*o, n, 42U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_at(unsigned long long *o)
{
  int n = (XKB_KEY_at) <= 0;
  *o = (unsigned long long)((XKB_KEY_at) | 0);  /* check that XKB_KEY_at is an integer */
  if (!_cffi_check_int(*o, n, 64U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_atilde(unsigned long long *o)
{
  int n = (XKB_KEY_atilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_atilde) | 0);  /* check that XKB_KEY_atilde is an integer */
  if (!_cffi_check_int(*o, n, 227U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_b(unsigned long long *o)
{
  int n = (XKB_KEY_b) <= 0;
  *o = (unsigned long long)((XKB_KEY_b) | 0);  /* check that XKB_KEY_b is an integer */
  if (!_cffi_check_int(*o, n, 98U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_babovedot(unsigned long long *o)
{
  int n = (XKB_KEY_babovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_babovedot) | 0);  /* check that XKB_KEY_babovedot is an integer */
  if (!_cffi_check_int(*o, n, 16784899U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_backslash(unsigned long long *o)
{
  int n = (XKB_KEY_backslash) <= 0;
  *o = (unsigned long long)((XKB_KEY_backslash) | 0);  /* check that XKB_KEY_backslash is an integer */
  if (!_cffi_check_int(*o, n, 92U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ballotcross(unsigned long long *o)
{
  int n = (XKB_KEY_ballotcross) <= 0;
  *o = (unsigned long long)((XKB_KEY_ballotcross) | 0);  /* check that XKB_KEY_ballotcross is an integer */
  if (!_cffi_check_int(*o, n, 2804U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_bar(unsigned long long *o)
{
  int n = (XKB_KEY_bar) <= 0;
  *o = (unsigned long long)((XKB_KEY_bar) | 0);  /* check that XKB_KEY_bar is an integer */
  if (!_cffi_check_int(*o, n, 124U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_because(unsigned long long *o)
{
  int n = (XKB_KEY_because) <= 0;
  *o = (unsigned long long)((XKB_KEY_because) | 0);  /* check that XKB_KEY_because is an integer */
  if (!_cffi_check_int(*o, n, 16785973U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_blank(unsigned long long *o)
{
  int n = (XKB_KEY_blank) <= 0;
  *o = (unsigned long long)((XKB_KEY_blank) | 0);  /* check that XKB_KEY_blank is an integer */
  if (!_cffi_check_int(*o, n, 2527U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_block(unsigned long long *o)
{
  int n = (XKB_KEY_block) <= 0;
  *o = (unsigned long long)((XKB_KEY_block) | 0);  /* check that XKB_KEY_block is an integer */
  if (!_cffi_check_int(*o, n, 268435708U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_botintegral(unsigned long long *o)
{
  int n = (XKB_KEY_botintegral) <= 0;
  *o = (unsigned long long)((XKB_KEY_botintegral) | 0);  /* check that XKB_KEY_botintegral is an integer */
  if (!_cffi_check_int(*o, n, 2213U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_botleftparens(unsigned long long *o)
{
  int n = (XKB_KEY_botleftparens) <= 0;
  *o = (unsigned long long)((XKB_KEY_botleftparens) | 0);  /* check that XKB_KEY_botleftparens is an integer */
  if (!_cffi_check_int(*o, n, 2220U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_botleftsqbracket(unsigned long long *o)
{
  int n = (XKB_KEY_botleftsqbracket) <= 0;
  *o = (unsigned long long)((XKB_KEY_botleftsqbracket) | 0);  /* check that XKB_KEY_botleftsqbracket is an integer */
  if (!_cffi_check_int(*o, n, 2216U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_botleftsummation(unsigned long long *o)
{
  int n = (XKB_KEY_botleftsummation) <= 0;
  *o = (unsigned long long)((XKB_KEY_botleftsummation) | 0);  /* check that XKB_KEY_botleftsummation is an integer */
  if (!_cffi_check_int(*o, n, 2226U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_botrightparens(unsigned long long *o)
{
  int n = (XKB_KEY_botrightparens) <= 0;
  *o = (unsigned long long)((XKB_KEY_botrightparens) | 0);  /* check that XKB_KEY_botrightparens is an integer */
  if (!_cffi_check_int(*o, n, 2222U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_botrightsqbracket(unsigned long long *o)
{
  int n = (XKB_KEY_botrightsqbracket) <= 0;
  *o = (unsigned long long)((XKB_KEY_botrightsqbracket) | 0);  /* check that XKB_KEY_botrightsqbracket is an integer */
  if (!_cffi_check_int(*o, n, 2218U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_botrightsummation(unsigned long long *o)
{
  int n = (XKB_KEY_botrightsummation) <= 0;
  *o = (unsigned long long)((XKB_KEY_botrightsummation) | 0);  /* check that XKB_KEY_botrightsummation is an integer */
  if (!_cffi_check_int(*o, n, 2230U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_bott(unsigned long long *o)
{
  int n = (XKB_KEY_bott) <= 0;
  *o = (unsigned long long)((XKB_KEY_bott) | 0);  /* check that XKB_KEY_bott is an integer */
  if (!_cffi_check_int(*o, n, 2550U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_botvertsummationconnector(unsigned long long *o)
{
  int n = (XKB_KEY_botvertsummationconnector) <= 0;
  *o = (unsigned long long)((XKB_KEY_botvertsummationconnector) | 0);  /* check that XKB_KEY_botvertsummationconnector is an integer */
  if (!_cffi_check_int(*o, n, 2228U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braceleft(unsigned long long *o)
{
  int n = (XKB_KEY_braceleft) <= 0;
  *o = (unsigned long long)((XKB_KEY_braceleft) | 0);  /* check that XKB_KEY_braceleft is an integer */
  if (!_cffi_check_int(*o, n, 123U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braceright(unsigned long long *o)
{
  int n = (XKB_KEY_braceright) <= 0;
  *o = (unsigned long long)((XKB_KEY_braceright) | 0);  /* check that XKB_KEY_braceright is an integer */
  if (!_cffi_check_int(*o, n, 125U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_bracketleft(unsigned long long *o)
{
  int n = (XKB_KEY_bracketleft) <= 0;
  *o = (unsigned long long)((XKB_KEY_bracketleft) | 0);  /* check that XKB_KEY_bracketleft is an integer */
  if (!_cffi_check_int(*o, n, 91U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_bracketright(unsigned long long *o)
{
  int n = (XKB_KEY_bracketright) <= 0;
  *o = (unsigned long long)((XKB_KEY_bracketright) | 0);  /* check that XKB_KEY_bracketright is an integer */
  if (!_cffi_check_int(*o, n, 93U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_blank(unsigned long long *o)
{
  int n = (XKB_KEY_braille_blank) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_blank) | 0);  /* check that XKB_KEY_braille_blank is an integer */
  if (!_cffi_check_int(*o, n, 16787456U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dot_1(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dot_1) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dot_1) | 0);  /* check that XKB_KEY_braille_dot_1 is an integer */
  if (!_cffi_check_int(*o, n, 65521U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dot_10(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dot_10) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dot_10) | 0);  /* check that XKB_KEY_braille_dot_10 is an integer */
  if (!_cffi_check_int(*o, n, 65530U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dot_2(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dot_2) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dot_2) | 0);  /* check that XKB_KEY_braille_dot_2 is an integer */
  if (!_cffi_check_int(*o, n, 65522U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dot_3(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dot_3) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dot_3) | 0);  /* check that XKB_KEY_braille_dot_3 is an integer */
  if (!_cffi_check_int(*o, n, 65523U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dot_4(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dot_4) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dot_4) | 0);  /* check that XKB_KEY_braille_dot_4 is an integer */
  if (!_cffi_check_int(*o, n, 65524U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dot_5(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dot_5) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dot_5) | 0);  /* check that XKB_KEY_braille_dot_5 is an integer */
  if (!_cffi_check_int(*o, n, 65525U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dot_6(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dot_6) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dot_6) | 0);  /* check that XKB_KEY_braille_dot_6 is an integer */
  if (!_cffi_check_int(*o, n, 65526U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dot_7(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dot_7) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dot_7) | 0);  /* check that XKB_KEY_braille_dot_7 is an integer */
  if (!_cffi_check_int(*o, n, 65527U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dot_8(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dot_8) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dot_8) | 0);  /* check that XKB_KEY_braille_dot_8 is an integer */
  if (!_cffi_check_int(*o, n, 65528U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dot_9(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dot_9) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dot_9) | 0);  /* check that XKB_KEY_braille_dot_9 is an integer */
  if (!_cffi_check_int(*o, n, 65529U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1) | 0);  /* check that XKB_KEY_braille_dots_1 is an integer */
  if (!_cffi_check_int(*o, n, 16787457U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12) | 0);  /* check that XKB_KEY_braille_dots_12 is an integer */
  if (!_cffi_check_int(*o, n, 16787459U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_123(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_123) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_123) | 0);  /* check that XKB_KEY_braille_dots_123 is an integer */
  if (!_cffi_check_int(*o, n, 16787463U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1234(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1234) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1234) | 0);  /* check that XKB_KEY_braille_dots_1234 is an integer */
  if (!_cffi_check_int(*o, n, 16787471U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12345(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12345) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12345) | 0);  /* check that XKB_KEY_braille_dots_12345 is an integer */
  if (!_cffi_check_int(*o, n, 16787487U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_123456(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_123456) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_123456) | 0);  /* check that XKB_KEY_braille_dots_123456 is an integer */
  if (!_cffi_check_int(*o, n, 16787519U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1234567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1234567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1234567) | 0);  /* check that XKB_KEY_braille_dots_1234567 is an integer */
  if (!_cffi_check_int(*o, n, 16787583U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12345678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12345678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12345678) | 0);  /* check that XKB_KEY_braille_dots_12345678 is an integer */
  if (!_cffi_check_int(*o, n, 16787711U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1234568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1234568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1234568) | 0);  /* check that XKB_KEY_braille_dots_1234568 is an integer */
  if (!_cffi_check_int(*o, n, 16787647U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_123457(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_123457) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_123457) | 0);  /* check that XKB_KEY_braille_dots_123457 is an integer */
  if (!_cffi_check_int(*o, n, 16787551U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1234578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1234578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1234578) | 0);  /* check that XKB_KEY_braille_dots_1234578 is an integer */
  if (!_cffi_check_int(*o, n, 16787679U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_123458(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_123458) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_123458) | 0);  /* check that XKB_KEY_braille_dots_123458 is an integer */
  if (!_cffi_check_int(*o, n, 16787615U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12346(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12346) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12346) | 0);  /* check that XKB_KEY_braille_dots_12346 is an integer */
  if (!_cffi_check_int(*o, n, 16787503U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_123467(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_123467) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_123467) | 0);  /* check that XKB_KEY_braille_dots_123467 is an integer */
  if (!_cffi_check_int(*o, n, 16787567U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1234678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1234678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1234678) | 0);  /* check that XKB_KEY_braille_dots_1234678 is an integer */
  if (!_cffi_check_int(*o, n, 16787695U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_123468(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_123468) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_123468) | 0);  /* check that XKB_KEY_braille_dots_123468 is an integer */
  if (!_cffi_check_int(*o, n, 16787631U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12347(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12347) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12347) | 0);  /* check that XKB_KEY_braille_dots_12347 is an integer */
  if (!_cffi_check_int(*o, n, 16787535U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_123478(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_123478) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_123478) | 0);  /* check that XKB_KEY_braille_dots_123478 is an integer */
  if (!_cffi_check_int(*o, n, 16787663U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12348(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12348) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12348) | 0);  /* check that XKB_KEY_braille_dots_12348 is an integer */
  if (!_cffi_check_int(*o, n, 16787599U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1235(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1235) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1235) | 0);  /* check that XKB_KEY_braille_dots_1235 is an integer */
  if (!_cffi_check_int(*o, n, 16787479U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12356(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12356) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12356) | 0);  /* check that XKB_KEY_braille_dots_12356 is an integer */
  if (!_cffi_check_int(*o, n, 16787511U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_123567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_123567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_123567) | 0);  /* check that XKB_KEY_braille_dots_123567 is an integer */
  if (!_cffi_check_int(*o, n, 16787575U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1235678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1235678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1235678) | 0);  /* check that XKB_KEY_braille_dots_1235678 is an integer */
  if (!_cffi_check_int(*o, n, 16787703U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_123568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_123568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_123568) | 0);  /* check that XKB_KEY_braille_dots_123568 is an integer */
  if (!_cffi_check_int(*o, n, 16787639U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12357(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12357) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12357) | 0);  /* check that XKB_KEY_braille_dots_12357 is an integer */
  if (!_cffi_check_int(*o, n, 16787543U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_123578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_123578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_123578) | 0);  /* check that XKB_KEY_braille_dots_123578 is an integer */
  if (!_cffi_check_int(*o, n, 16787671U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12358(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12358) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12358) | 0);  /* check that XKB_KEY_braille_dots_12358 is an integer */
  if (!_cffi_check_int(*o, n, 16787607U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1236(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1236) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1236) | 0);  /* check that XKB_KEY_braille_dots_1236 is an integer */
  if (!_cffi_check_int(*o, n, 16787495U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12367(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12367) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12367) | 0);  /* check that XKB_KEY_braille_dots_12367 is an integer */
  if (!_cffi_check_int(*o, n, 16787559U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_123678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_123678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_123678) | 0);  /* check that XKB_KEY_braille_dots_123678 is an integer */
  if (!_cffi_check_int(*o, n, 16787687U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12368(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12368) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12368) | 0);  /* check that XKB_KEY_braille_dots_12368 is an integer */
  if (!_cffi_check_int(*o, n, 16787623U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1237(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1237) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1237) | 0);  /* check that XKB_KEY_braille_dots_1237 is an integer */
  if (!_cffi_check_int(*o, n, 16787527U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12378(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12378) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12378) | 0);  /* check that XKB_KEY_braille_dots_12378 is an integer */
  if (!_cffi_check_int(*o, n, 16787655U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1238(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1238) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1238) | 0);  /* check that XKB_KEY_braille_dots_1238 is an integer */
  if (!_cffi_check_int(*o, n, 16787591U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_124(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_124) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_124) | 0);  /* check that XKB_KEY_braille_dots_124 is an integer */
  if (!_cffi_check_int(*o, n, 16787467U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1245(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1245) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1245) | 0);  /* check that XKB_KEY_braille_dots_1245 is an integer */
  if (!_cffi_check_int(*o, n, 16787483U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12456(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12456) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12456) | 0);  /* check that XKB_KEY_braille_dots_12456 is an integer */
  if (!_cffi_check_int(*o, n, 16787515U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_124567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_124567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_124567) | 0);  /* check that XKB_KEY_braille_dots_124567 is an integer */
  if (!_cffi_check_int(*o, n, 16787579U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1245678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1245678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1245678) | 0);  /* check that XKB_KEY_braille_dots_1245678 is an integer */
  if (!_cffi_check_int(*o, n, 16787707U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_124568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_124568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_124568) | 0);  /* check that XKB_KEY_braille_dots_124568 is an integer */
  if (!_cffi_check_int(*o, n, 16787643U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12457(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12457) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12457) | 0);  /* check that XKB_KEY_braille_dots_12457 is an integer */
  if (!_cffi_check_int(*o, n, 16787547U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_124578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_124578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_124578) | 0);  /* check that XKB_KEY_braille_dots_124578 is an integer */
  if (!_cffi_check_int(*o, n, 16787675U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12458(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12458) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12458) | 0);  /* check that XKB_KEY_braille_dots_12458 is an integer */
  if (!_cffi_check_int(*o, n, 16787611U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1246(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1246) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1246) | 0);  /* check that XKB_KEY_braille_dots_1246 is an integer */
  if (!_cffi_check_int(*o, n, 16787499U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12467(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12467) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12467) | 0);  /* check that XKB_KEY_braille_dots_12467 is an integer */
  if (!_cffi_check_int(*o, n, 16787563U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_124678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_124678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_124678) | 0);  /* check that XKB_KEY_braille_dots_124678 is an integer */
  if (!_cffi_check_int(*o, n, 16787691U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12468(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12468) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12468) | 0);  /* check that XKB_KEY_braille_dots_12468 is an integer */
  if (!_cffi_check_int(*o, n, 16787627U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1247(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1247) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1247) | 0);  /* check that XKB_KEY_braille_dots_1247 is an integer */
  if (!_cffi_check_int(*o, n, 16787531U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12478(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12478) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12478) | 0);  /* check that XKB_KEY_braille_dots_12478 is an integer */
  if (!_cffi_check_int(*o, n, 16787659U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1248(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1248) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1248) | 0);  /* check that XKB_KEY_braille_dots_1248 is an integer */
  if (!_cffi_check_int(*o, n, 16787595U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_125(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_125) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_125) | 0);  /* check that XKB_KEY_braille_dots_125 is an integer */
  if (!_cffi_check_int(*o, n, 16787475U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1256(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1256) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1256) | 0);  /* check that XKB_KEY_braille_dots_1256 is an integer */
  if (!_cffi_check_int(*o, n, 16787507U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12567) | 0);  /* check that XKB_KEY_braille_dots_12567 is an integer */
  if (!_cffi_check_int(*o, n, 16787571U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_125678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_125678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_125678) | 0);  /* check that XKB_KEY_braille_dots_125678 is an integer */
  if (!_cffi_check_int(*o, n, 16787699U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12568) | 0);  /* check that XKB_KEY_braille_dots_12568 is an integer */
  if (!_cffi_check_int(*o, n, 16787635U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1257(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1257) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1257) | 0);  /* check that XKB_KEY_braille_dots_1257 is an integer */
  if (!_cffi_check_int(*o, n, 16787539U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12578) | 0);  /* check that XKB_KEY_braille_dots_12578 is an integer */
  if (!_cffi_check_int(*o, n, 16787667U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1258(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1258) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1258) | 0);  /* check that XKB_KEY_braille_dots_1258 is an integer */
  if (!_cffi_check_int(*o, n, 16787603U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_126(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_126) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_126) | 0);  /* check that XKB_KEY_braille_dots_126 is an integer */
  if (!_cffi_check_int(*o, n, 16787491U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1267(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1267) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1267) | 0);  /* check that XKB_KEY_braille_dots_1267 is an integer */
  if (!_cffi_check_int(*o, n, 16787555U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_12678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_12678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_12678) | 0);  /* check that XKB_KEY_braille_dots_12678 is an integer */
  if (!_cffi_check_int(*o, n, 16787683U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1268(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1268) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1268) | 0);  /* check that XKB_KEY_braille_dots_1268 is an integer */
  if (!_cffi_check_int(*o, n, 16787619U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_127(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_127) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_127) | 0);  /* check that XKB_KEY_braille_dots_127 is an integer */
  if (!_cffi_check_int(*o, n, 16787523U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1278(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1278) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1278) | 0);  /* check that XKB_KEY_braille_dots_1278 is an integer */
  if (!_cffi_check_int(*o, n, 16787651U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_128(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_128) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_128) | 0);  /* check that XKB_KEY_braille_dots_128 is an integer */
  if (!_cffi_check_int(*o, n, 16787587U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_13(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_13) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_13) | 0);  /* check that XKB_KEY_braille_dots_13 is an integer */
  if (!_cffi_check_int(*o, n, 16787461U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_134(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_134) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_134) | 0);  /* check that XKB_KEY_braille_dots_134 is an integer */
  if (!_cffi_check_int(*o, n, 16787469U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1345(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1345) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1345) | 0);  /* check that XKB_KEY_braille_dots_1345 is an integer */
  if (!_cffi_check_int(*o, n, 16787485U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_13456(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_13456) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_13456) | 0);  /* check that XKB_KEY_braille_dots_13456 is an integer */
  if (!_cffi_check_int(*o, n, 16787517U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_134567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_134567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_134567) | 0);  /* check that XKB_KEY_braille_dots_134567 is an integer */
  if (!_cffi_check_int(*o, n, 16787581U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1345678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1345678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1345678) | 0);  /* check that XKB_KEY_braille_dots_1345678 is an integer */
  if (!_cffi_check_int(*o, n, 16787709U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_134568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_134568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_134568) | 0);  /* check that XKB_KEY_braille_dots_134568 is an integer */
  if (!_cffi_check_int(*o, n, 16787645U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_13457(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_13457) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_13457) | 0);  /* check that XKB_KEY_braille_dots_13457 is an integer */
  if (!_cffi_check_int(*o, n, 16787549U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_134578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_134578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_134578) | 0);  /* check that XKB_KEY_braille_dots_134578 is an integer */
  if (!_cffi_check_int(*o, n, 16787677U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_13458(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_13458) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_13458) | 0);  /* check that XKB_KEY_braille_dots_13458 is an integer */
  if (!_cffi_check_int(*o, n, 16787613U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1346(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1346) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1346) | 0);  /* check that XKB_KEY_braille_dots_1346 is an integer */
  if (!_cffi_check_int(*o, n, 16787501U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_13467(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_13467) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_13467) | 0);  /* check that XKB_KEY_braille_dots_13467 is an integer */
  if (!_cffi_check_int(*o, n, 16787565U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_134678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_134678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_134678) | 0);  /* check that XKB_KEY_braille_dots_134678 is an integer */
  if (!_cffi_check_int(*o, n, 16787693U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_13468(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_13468) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_13468) | 0);  /* check that XKB_KEY_braille_dots_13468 is an integer */
  if (!_cffi_check_int(*o, n, 16787629U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1347(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1347) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1347) | 0);  /* check that XKB_KEY_braille_dots_1347 is an integer */
  if (!_cffi_check_int(*o, n, 16787533U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_13478(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_13478) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_13478) | 0);  /* check that XKB_KEY_braille_dots_13478 is an integer */
  if (!_cffi_check_int(*o, n, 16787661U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1348(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1348) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1348) | 0);  /* check that XKB_KEY_braille_dots_1348 is an integer */
  if (!_cffi_check_int(*o, n, 16787597U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_135(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_135) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_135) | 0);  /* check that XKB_KEY_braille_dots_135 is an integer */
  if (!_cffi_check_int(*o, n, 16787477U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1356(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1356) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1356) | 0);  /* check that XKB_KEY_braille_dots_1356 is an integer */
  if (!_cffi_check_int(*o, n, 16787509U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_13567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_13567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_13567) | 0);  /* check that XKB_KEY_braille_dots_13567 is an integer */
  if (!_cffi_check_int(*o, n, 16787573U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_135678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_135678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_135678) | 0);  /* check that XKB_KEY_braille_dots_135678 is an integer */
  if (!_cffi_check_int(*o, n, 16787701U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_13568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_13568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_13568) | 0);  /* check that XKB_KEY_braille_dots_13568 is an integer */
  if (!_cffi_check_int(*o, n, 16787637U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1357(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1357) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1357) | 0);  /* check that XKB_KEY_braille_dots_1357 is an integer */
  if (!_cffi_check_int(*o, n, 16787541U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_13578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_13578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_13578) | 0);  /* check that XKB_KEY_braille_dots_13578 is an integer */
  if (!_cffi_check_int(*o, n, 16787669U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1358(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1358) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1358) | 0);  /* check that XKB_KEY_braille_dots_1358 is an integer */
  if (!_cffi_check_int(*o, n, 16787605U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_136(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_136) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_136) | 0);  /* check that XKB_KEY_braille_dots_136 is an integer */
  if (!_cffi_check_int(*o, n, 16787493U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1367(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1367) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1367) | 0);  /* check that XKB_KEY_braille_dots_1367 is an integer */
  if (!_cffi_check_int(*o, n, 16787557U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_13678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_13678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_13678) | 0);  /* check that XKB_KEY_braille_dots_13678 is an integer */
  if (!_cffi_check_int(*o, n, 16787685U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1368(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1368) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1368) | 0);  /* check that XKB_KEY_braille_dots_1368 is an integer */
  if (!_cffi_check_int(*o, n, 16787621U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_137(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_137) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_137) | 0);  /* check that XKB_KEY_braille_dots_137 is an integer */
  if (!_cffi_check_int(*o, n, 16787525U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1378(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1378) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1378) | 0);  /* check that XKB_KEY_braille_dots_1378 is an integer */
  if (!_cffi_check_int(*o, n, 16787653U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_138(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_138) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_138) | 0);  /* check that XKB_KEY_braille_dots_138 is an integer */
  if (!_cffi_check_int(*o, n, 16787589U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_14(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_14) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_14) | 0);  /* check that XKB_KEY_braille_dots_14 is an integer */
  if (!_cffi_check_int(*o, n, 16787465U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_145(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_145) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_145) | 0);  /* check that XKB_KEY_braille_dots_145 is an integer */
  if (!_cffi_check_int(*o, n, 16787481U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1456(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1456) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1456) | 0);  /* check that XKB_KEY_braille_dots_1456 is an integer */
  if (!_cffi_check_int(*o, n, 16787513U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_14567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_14567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_14567) | 0);  /* check that XKB_KEY_braille_dots_14567 is an integer */
  if (!_cffi_check_int(*o, n, 16787577U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_145678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_145678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_145678) | 0);  /* check that XKB_KEY_braille_dots_145678 is an integer */
  if (!_cffi_check_int(*o, n, 16787705U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_14568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_14568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_14568) | 0);  /* check that XKB_KEY_braille_dots_14568 is an integer */
  if (!_cffi_check_int(*o, n, 16787641U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1457(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1457) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1457) | 0);  /* check that XKB_KEY_braille_dots_1457 is an integer */
  if (!_cffi_check_int(*o, n, 16787545U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_14578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_14578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_14578) | 0);  /* check that XKB_KEY_braille_dots_14578 is an integer */
  if (!_cffi_check_int(*o, n, 16787673U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1458(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1458) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1458) | 0);  /* check that XKB_KEY_braille_dots_1458 is an integer */
  if (!_cffi_check_int(*o, n, 16787609U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_146(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_146) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_146) | 0);  /* check that XKB_KEY_braille_dots_146 is an integer */
  if (!_cffi_check_int(*o, n, 16787497U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1467(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1467) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1467) | 0);  /* check that XKB_KEY_braille_dots_1467 is an integer */
  if (!_cffi_check_int(*o, n, 16787561U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_14678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_14678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_14678) | 0);  /* check that XKB_KEY_braille_dots_14678 is an integer */
  if (!_cffi_check_int(*o, n, 16787689U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1468(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1468) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1468) | 0);  /* check that XKB_KEY_braille_dots_1468 is an integer */
  if (!_cffi_check_int(*o, n, 16787625U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_147(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_147) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_147) | 0);  /* check that XKB_KEY_braille_dots_147 is an integer */
  if (!_cffi_check_int(*o, n, 16787529U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1478(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1478) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1478) | 0);  /* check that XKB_KEY_braille_dots_1478 is an integer */
  if (!_cffi_check_int(*o, n, 16787657U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_148(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_148) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_148) | 0);  /* check that XKB_KEY_braille_dots_148 is an integer */
  if (!_cffi_check_int(*o, n, 16787593U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_15(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_15) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_15) | 0);  /* check that XKB_KEY_braille_dots_15 is an integer */
  if (!_cffi_check_int(*o, n, 16787473U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_156(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_156) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_156) | 0);  /* check that XKB_KEY_braille_dots_156 is an integer */
  if (!_cffi_check_int(*o, n, 16787505U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1567) | 0);  /* check that XKB_KEY_braille_dots_1567 is an integer */
  if (!_cffi_check_int(*o, n, 16787569U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_15678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_15678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_15678) | 0);  /* check that XKB_KEY_braille_dots_15678 is an integer */
  if (!_cffi_check_int(*o, n, 16787697U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1568) | 0);  /* check that XKB_KEY_braille_dots_1568 is an integer */
  if (!_cffi_check_int(*o, n, 16787633U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_157(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_157) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_157) | 0);  /* check that XKB_KEY_braille_dots_157 is an integer */
  if (!_cffi_check_int(*o, n, 16787537U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1578) | 0);  /* check that XKB_KEY_braille_dots_1578 is an integer */
  if (!_cffi_check_int(*o, n, 16787665U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_158(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_158) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_158) | 0);  /* check that XKB_KEY_braille_dots_158 is an integer */
  if (!_cffi_check_int(*o, n, 16787601U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_16(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_16) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_16) | 0);  /* check that XKB_KEY_braille_dots_16 is an integer */
  if (!_cffi_check_int(*o, n, 16787489U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_167(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_167) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_167) | 0);  /* check that XKB_KEY_braille_dots_167 is an integer */
  if (!_cffi_check_int(*o, n, 16787553U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_1678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_1678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_1678) | 0);  /* check that XKB_KEY_braille_dots_1678 is an integer */
  if (!_cffi_check_int(*o, n, 16787681U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_168(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_168) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_168) | 0);  /* check that XKB_KEY_braille_dots_168 is an integer */
  if (!_cffi_check_int(*o, n, 16787617U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_17(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_17) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_17) | 0);  /* check that XKB_KEY_braille_dots_17 is an integer */
  if (!_cffi_check_int(*o, n, 16787521U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_178(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_178) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_178) | 0);  /* check that XKB_KEY_braille_dots_178 is an integer */
  if (!_cffi_check_int(*o, n, 16787649U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_18(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_18) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_18) | 0);  /* check that XKB_KEY_braille_dots_18 is an integer */
  if (!_cffi_check_int(*o, n, 16787585U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2) | 0);  /* check that XKB_KEY_braille_dots_2 is an integer */
  if (!_cffi_check_int(*o, n, 16787458U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_23(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_23) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_23) | 0);  /* check that XKB_KEY_braille_dots_23 is an integer */
  if (!_cffi_check_int(*o, n, 16787462U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_234(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_234) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_234) | 0);  /* check that XKB_KEY_braille_dots_234 is an integer */
  if (!_cffi_check_int(*o, n, 16787470U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2345(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2345) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2345) | 0);  /* check that XKB_KEY_braille_dots_2345 is an integer */
  if (!_cffi_check_int(*o, n, 16787486U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_23456(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_23456) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_23456) | 0);  /* check that XKB_KEY_braille_dots_23456 is an integer */
  if (!_cffi_check_int(*o, n, 16787518U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_234567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_234567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_234567) | 0);  /* check that XKB_KEY_braille_dots_234567 is an integer */
  if (!_cffi_check_int(*o, n, 16787582U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2345678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2345678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2345678) | 0);  /* check that XKB_KEY_braille_dots_2345678 is an integer */
  if (!_cffi_check_int(*o, n, 16787710U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_234568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_234568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_234568) | 0);  /* check that XKB_KEY_braille_dots_234568 is an integer */
  if (!_cffi_check_int(*o, n, 16787646U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_23457(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_23457) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_23457) | 0);  /* check that XKB_KEY_braille_dots_23457 is an integer */
  if (!_cffi_check_int(*o, n, 16787550U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_234578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_234578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_234578) | 0);  /* check that XKB_KEY_braille_dots_234578 is an integer */
  if (!_cffi_check_int(*o, n, 16787678U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_23458(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_23458) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_23458) | 0);  /* check that XKB_KEY_braille_dots_23458 is an integer */
  if (!_cffi_check_int(*o, n, 16787614U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2346(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2346) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2346) | 0);  /* check that XKB_KEY_braille_dots_2346 is an integer */
  if (!_cffi_check_int(*o, n, 16787502U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_23467(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_23467) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_23467) | 0);  /* check that XKB_KEY_braille_dots_23467 is an integer */
  if (!_cffi_check_int(*o, n, 16787566U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_234678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_234678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_234678) | 0);  /* check that XKB_KEY_braille_dots_234678 is an integer */
  if (!_cffi_check_int(*o, n, 16787694U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_23468(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_23468) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_23468) | 0);  /* check that XKB_KEY_braille_dots_23468 is an integer */
  if (!_cffi_check_int(*o, n, 16787630U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2347(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2347) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2347) | 0);  /* check that XKB_KEY_braille_dots_2347 is an integer */
  if (!_cffi_check_int(*o, n, 16787534U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_23478(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_23478) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_23478) | 0);  /* check that XKB_KEY_braille_dots_23478 is an integer */
  if (!_cffi_check_int(*o, n, 16787662U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2348(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2348) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2348) | 0);  /* check that XKB_KEY_braille_dots_2348 is an integer */
  if (!_cffi_check_int(*o, n, 16787598U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_235(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_235) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_235) | 0);  /* check that XKB_KEY_braille_dots_235 is an integer */
  if (!_cffi_check_int(*o, n, 16787478U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2356(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2356) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2356) | 0);  /* check that XKB_KEY_braille_dots_2356 is an integer */
  if (!_cffi_check_int(*o, n, 16787510U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_23567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_23567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_23567) | 0);  /* check that XKB_KEY_braille_dots_23567 is an integer */
  if (!_cffi_check_int(*o, n, 16787574U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_235678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_235678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_235678) | 0);  /* check that XKB_KEY_braille_dots_235678 is an integer */
  if (!_cffi_check_int(*o, n, 16787702U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_23568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_23568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_23568) | 0);  /* check that XKB_KEY_braille_dots_23568 is an integer */
  if (!_cffi_check_int(*o, n, 16787638U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2357(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2357) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2357) | 0);  /* check that XKB_KEY_braille_dots_2357 is an integer */
  if (!_cffi_check_int(*o, n, 16787542U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_23578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_23578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_23578) | 0);  /* check that XKB_KEY_braille_dots_23578 is an integer */
  if (!_cffi_check_int(*o, n, 16787670U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2358(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2358) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2358) | 0);  /* check that XKB_KEY_braille_dots_2358 is an integer */
  if (!_cffi_check_int(*o, n, 16787606U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_236(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_236) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_236) | 0);  /* check that XKB_KEY_braille_dots_236 is an integer */
  if (!_cffi_check_int(*o, n, 16787494U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2367(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2367) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2367) | 0);  /* check that XKB_KEY_braille_dots_2367 is an integer */
  if (!_cffi_check_int(*o, n, 16787558U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_23678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_23678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_23678) | 0);  /* check that XKB_KEY_braille_dots_23678 is an integer */
  if (!_cffi_check_int(*o, n, 16787686U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2368(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2368) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2368) | 0);  /* check that XKB_KEY_braille_dots_2368 is an integer */
  if (!_cffi_check_int(*o, n, 16787622U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_237(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_237) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_237) | 0);  /* check that XKB_KEY_braille_dots_237 is an integer */
  if (!_cffi_check_int(*o, n, 16787526U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2378(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2378) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2378) | 0);  /* check that XKB_KEY_braille_dots_2378 is an integer */
  if (!_cffi_check_int(*o, n, 16787654U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_238(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_238) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_238) | 0);  /* check that XKB_KEY_braille_dots_238 is an integer */
  if (!_cffi_check_int(*o, n, 16787590U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_24(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_24) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_24) | 0);  /* check that XKB_KEY_braille_dots_24 is an integer */
  if (!_cffi_check_int(*o, n, 16787466U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_245(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_245) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_245) | 0);  /* check that XKB_KEY_braille_dots_245 is an integer */
  if (!_cffi_check_int(*o, n, 16787482U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2456(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2456) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2456) | 0);  /* check that XKB_KEY_braille_dots_2456 is an integer */
  if (!_cffi_check_int(*o, n, 16787514U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_24567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_24567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_24567) | 0);  /* check that XKB_KEY_braille_dots_24567 is an integer */
  if (!_cffi_check_int(*o, n, 16787578U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_245678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_245678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_245678) | 0);  /* check that XKB_KEY_braille_dots_245678 is an integer */
  if (!_cffi_check_int(*o, n, 16787706U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_24568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_24568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_24568) | 0);  /* check that XKB_KEY_braille_dots_24568 is an integer */
  if (!_cffi_check_int(*o, n, 16787642U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2457(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2457) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2457) | 0);  /* check that XKB_KEY_braille_dots_2457 is an integer */
  if (!_cffi_check_int(*o, n, 16787546U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_24578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_24578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_24578) | 0);  /* check that XKB_KEY_braille_dots_24578 is an integer */
  if (!_cffi_check_int(*o, n, 16787674U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2458(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2458) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2458) | 0);  /* check that XKB_KEY_braille_dots_2458 is an integer */
  if (!_cffi_check_int(*o, n, 16787610U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_246(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_246) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_246) | 0);  /* check that XKB_KEY_braille_dots_246 is an integer */
  if (!_cffi_check_int(*o, n, 16787498U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2467(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2467) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2467) | 0);  /* check that XKB_KEY_braille_dots_2467 is an integer */
  if (!_cffi_check_int(*o, n, 16787562U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_24678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_24678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_24678) | 0);  /* check that XKB_KEY_braille_dots_24678 is an integer */
  if (!_cffi_check_int(*o, n, 16787690U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2468(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2468) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2468) | 0);  /* check that XKB_KEY_braille_dots_2468 is an integer */
  if (!_cffi_check_int(*o, n, 16787626U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_247(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_247) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_247) | 0);  /* check that XKB_KEY_braille_dots_247 is an integer */
  if (!_cffi_check_int(*o, n, 16787530U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2478(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2478) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2478) | 0);  /* check that XKB_KEY_braille_dots_2478 is an integer */
  if (!_cffi_check_int(*o, n, 16787658U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_248(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_248) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_248) | 0);  /* check that XKB_KEY_braille_dots_248 is an integer */
  if (!_cffi_check_int(*o, n, 16787594U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_25(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_25) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_25) | 0);  /* check that XKB_KEY_braille_dots_25 is an integer */
  if (!_cffi_check_int(*o, n, 16787474U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_256(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_256) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_256) | 0);  /* check that XKB_KEY_braille_dots_256 is an integer */
  if (!_cffi_check_int(*o, n, 16787506U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2567) | 0);  /* check that XKB_KEY_braille_dots_2567 is an integer */
  if (!_cffi_check_int(*o, n, 16787570U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_25678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_25678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_25678) | 0);  /* check that XKB_KEY_braille_dots_25678 is an integer */
  if (!_cffi_check_int(*o, n, 16787698U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2568) | 0);  /* check that XKB_KEY_braille_dots_2568 is an integer */
  if (!_cffi_check_int(*o, n, 16787634U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_257(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_257) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_257) | 0);  /* check that XKB_KEY_braille_dots_257 is an integer */
  if (!_cffi_check_int(*o, n, 16787538U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2578) | 0);  /* check that XKB_KEY_braille_dots_2578 is an integer */
  if (!_cffi_check_int(*o, n, 16787666U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_258(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_258) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_258) | 0);  /* check that XKB_KEY_braille_dots_258 is an integer */
  if (!_cffi_check_int(*o, n, 16787602U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_26(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_26) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_26) | 0);  /* check that XKB_KEY_braille_dots_26 is an integer */
  if (!_cffi_check_int(*o, n, 16787490U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_267(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_267) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_267) | 0);  /* check that XKB_KEY_braille_dots_267 is an integer */
  if (!_cffi_check_int(*o, n, 16787554U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_2678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_2678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_2678) | 0);  /* check that XKB_KEY_braille_dots_2678 is an integer */
  if (!_cffi_check_int(*o, n, 16787682U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_268(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_268) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_268) | 0);  /* check that XKB_KEY_braille_dots_268 is an integer */
  if (!_cffi_check_int(*o, n, 16787618U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_27(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_27) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_27) | 0);  /* check that XKB_KEY_braille_dots_27 is an integer */
  if (!_cffi_check_int(*o, n, 16787522U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_278(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_278) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_278) | 0);  /* check that XKB_KEY_braille_dots_278 is an integer */
  if (!_cffi_check_int(*o, n, 16787650U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_28(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_28) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_28) | 0);  /* check that XKB_KEY_braille_dots_28 is an integer */
  if (!_cffi_check_int(*o, n, 16787586U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_3(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_3) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_3) | 0);  /* check that XKB_KEY_braille_dots_3 is an integer */
  if (!_cffi_check_int(*o, n, 16787460U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_34(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_34) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_34) | 0);  /* check that XKB_KEY_braille_dots_34 is an integer */
  if (!_cffi_check_int(*o, n, 16787468U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_345(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_345) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_345) | 0);  /* check that XKB_KEY_braille_dots_345 is an integer */
  if (!_cffi_check_int(*o, n, 16787484U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_3456(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_3456) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_3456) | 0);  /* check that XKB_KEY_braille_dots_3456 is an integer */
  if (!_cffi_check_int(*o, n, 16787516U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_34567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_34567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_34567) | 0);  /* check that XKB_KEY_braille_dots_34567 is an integer */
  if (!_cffi_check_int(*o, n, 16787580U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_345678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_345678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_345678) | 0);  /* check that XKB_KEY_braille_dots_345678 is an integer */
  if (!_cffi_check_int(*o, n, 16787708U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_34568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_34568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_34568) | 0);  /* check that XKB_KEY_braille_dots_34568 is an integer */
  if (!_cffi_check_int(*o, n, 16787644U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_3457(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_3457) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_3457) | 0);  /* check that XKB_KEY_braille_dots_3457 is an integer */
  if (!_cffi_check_int(*o, n, 16787548U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_34578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_34578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_34578) | 0);  /* check that XKB_KEY_braille_dots_34578 is an integer */
  if (!_cffi_check_int(*o, n, 16787676U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_3458(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_3458) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_3458) | 0);  /* check that XKB_KEY_braille_dots_3458 is an integer */
  if (!_cffi_check_int(*o, n, 16787612U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_346(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_346) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_346) | 0);  /* check that XKB_KEY_braille_dots_346 is an integer */
  if (!_cffi_check_int(*o, n, 16787500U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_3467(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_3467) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_3467) | 0);  /* check that XKB_KEY_braille_dots_3467 is an integer */
  if (!_cffi_check_int(*o, n, 16787564U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_34678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_34678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_34678) | 0);  /* check that XKB_KEY_braille_dots_34678 is an integer */
  if (!_cffi_check_int(*o, n, 16787692U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_3468(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_3468) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_3468) | 0);  /* check that XKB_KEY_braille_dots_3468 is an integer */
  if (!_cffi_check_int(*o, n, 16787628U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_347(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_347) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_347) | 0);  /* check that XKB_KEY_braille_dots_347 is an integer */
  if (!_cffi_check_int(*o, n, 16787532U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_3478(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_3478) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_3478) | 0);  /* check that XKB_KEY_braille_dots_3478 is an integer */
  if (!_cffi_check_int(*o, n, 16787660U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_348(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_348) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_348) | 0);  /* check that XKB_KEY_braille_dots_348 is an integer */
  if (!_cffi_check_int(*o, n, 16787596U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_35(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_35) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_35) | 0);  /* check that XKB_KEY_braille_dots_35 is an integer */
  if (!_cffi_check_int(*o, n, 16787476U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_356(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_356) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_356) | 0);  /* check that XKB_KEY_braille_dots_356 is an integer */
  if (!_cffi_check_int(*o, n, 16787508U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_3567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_3567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_3567) | 0);  /* check that XKB_KEY_braille_dots_3567 is an integer */
  if (!_cffi_check_int(*o, n, 16787572U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_35678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_35678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_35678) | 0);  /* check that XKB_KEY_braille_dots_35678 is an integer */
  if (!_cffi_check_int(*o, n, 16787700U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_3568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_3568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_3568) | 0);  /* check that XKB_KEY_braille_dots_3568 is an integer */
  if (!_cffi_check_int(*o, n, 16787636U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_357(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_357) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_357) | 0);  /* check that XKB_KEY_braille_dots_357 is an integer */
  if (!_cffi_check_int(*o, n, 16787540U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_3578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_3578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_3578) | 0);  /* check that XKB_KEY_braille_dots_3578 is an integer */
  if (!_cffi_check_int(*o, n, 16787668U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_358(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_358) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_358) | 0);  /* check that XKB_KEY_braille_dots_358 is an integer */
  if (!_cffi_check_int(*o, n, 16787604U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_36(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_36) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_36) | 0);  /* check that XKB_KEY_braille_dots_36 is an integer */
  if (!_cffi_check_int(*o, n, 16787492U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_367(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_367) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_367) | 0);  /* check that XKB_KEY_braille_dots_367 is an integer */
  if (!_cffi_check_int(*o, n, 16787556U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_3678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_3678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_3678) | 0);  /* check that XKB_KEY_braille_dots_3678 is an integer */
  if (!_cffi_check_int(*o, n, 16787684U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_368(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_368) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_368) | 0);  /* check that XKB_KEY_braille_dots_368 is an integer */
  if (!_cffi_check_int(*o, n, 16787620U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_37(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_37) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_37) | 0);  /* check that XKB_KEY_braille_dots_37 is an integer */
  if (!_cffi_check_int(*o, n, 16787524U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_378(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_378) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_378) | 0);  /* check that XKB_KEY_braille_dots_378 is an integer */
  if (!_cffi_check_int(*o, n, 16787652U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_38(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_38) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_38) | 0);  /* check that XKB_KEY_braille_dots_38 is an integer */
  if (!_cffi_check_int(*o, n, 16787588U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_4(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_4) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_4) | 0);  /* check that XKB_KEY_braille_dots_4 is an integer */
  if (!_cffi_check_int(*o, n, 16787464U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_45(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_45) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_45) | 0);  /* check that XKB_KEY_braille_dots_45 is an integer */
  if (!_cffi_check_int(*o, n, 16787480U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_456(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_456) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_456) | 0);  /* check that XKB_KEY_braille_dots_456 is an integer */
  if (!_cffi_check_int(*o, n, 16787512U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_4567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_4567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_4567) | 0);  /* check that XKB_KEY_braille_dots_4567 is an integer */
  if (!_cffi_check_int(*o, n, 16787576U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_45678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_45678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_45678) | 0);  /* check that XKB_KEY_braille_dots_45678 is an integer */
  if (!_cffi_check_int(*o, n, 16787704U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_4568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_4568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_4568) | 0);  /* check that XKB_KEY_braille_dots_4568 is an integer */
  if (!_cffi_check_int(*o, n, 16787640U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_457(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_457) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_457) | 0);  /* check that XKB_KEY_braille_dots_457 is an integer */
  if (!_cffi_check_int(*o, n, 16787544U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_4578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_4578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_4578) | 0);  /* check that XKB_KEY_braille_dots_4578 is an integer */
  if (!_cffi_check_int(*o, n, 16787672U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_458(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_458) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_458) | 0);  /* check that XKB_KEY_braille_dots_458 is an integer */
  if (!_cffi_check_int(*o, n, 16787608U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_46(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_46) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_46) | 0);  /* check that XKB_KEY_braille_dots_46 is an integer */
  if (!_cffi_check_int(*o, n, 16787496U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_467(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_467) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_467) | 0);  /* check that XKB_KEY_braille_dots_467 is an integer */
  if (!_cffi_check_int(*o, n, 16787560U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_4678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_4678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_4678) | 0);  /* check that XKB_KEY_braille_dots_4678 is an integer */
  if (!_cffi_check_int(*o, n, 16787688U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_468(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_468) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_468) | 0);  /* check that XKB_KEY_braille_dots_468 is an integer */
  if (!_cffi_check_int(*o, n, 16787624U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_47(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_47) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_47) | 0);  /* check that XKB_KEY_braille_dots_47 is an integer */
  if (!_cffi_check_int(*o, n, 16787528U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_478(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_478) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_478) | 0);  /* check that XKB_KEY_braille_dots_478 is an integer */
  if (!_cffi_check_int(*o, n, 16787656U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_48(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_48) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_48) | 0);  /* check that XKB_KEY_braille_dots_48 is an integer */
  if (!_cffi_check_int(*o, n, 16787592U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_5(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_5) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_5) | 0);  /* check that XKB_KEY_braille_dots_5 is an integer */
  if (!_cffi_check_int(*o, n, 16787472U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_56(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_56) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_56) | 0);  /* check that XKB_KEY_braille_dots_56 is an integer */
  if (!_cffi_check_int(*o, n, 16787504U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_567(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_567) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_567) | 0);  /* check that XKB_KEY_braille_dots_567 is an integer */
  if (!_cffi_check_int(*o, n, 16787568U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_5678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_5678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_5678) | 0);  /* check that XKB_KEY_braille_dots_5678 is an integer */
  if (!_cffi_check_int(*o, n, 16787696U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_568(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_568) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_568) | 0);  /* check that XKB_KEY_braille_dots_568 is an integer */
  if (!_cffi_check_int(*o, n, 16787632U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_57(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_57) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_57) | 0);  /* check that XKB_KEY_braille_dots_57 is an integer */
  if (!_cffi_check_int(*o, n, 16787536U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_578(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_578) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_578) | 0);  /* check that XKB_KEY_braille_dots_578 is an integer */
  if (!_cffi_check_int(*o, n, 16787664U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_58(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_58) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_58) | 0);  /* check that XKB_KEY_braille_dots_58 is an integer */
  if (!_cffi_check_int(*o, n, 16787600U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_6(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_6) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_6) | 0);  /* check that XKB_KEY_braille_dots_6 is an integer */
  if (!_cffi_check_int(*o, n, 16787488U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_67(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_67) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_67) | 0);  /* check that XKB_KEY_braille_dots_67 is an integer */
  if (!_cffi_check_int(*o, n, 16787552U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_678(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_678) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_678) | 0);  /* check that XKB_KEY_braille_dots_678 is an integer */
  if (!_cffi_check_int(*o, n, 16787680U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_68(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_68) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_68) | 0);  /* check that XKB_KEY_braille_dots_68 is an integer */
  if (!_cffi_check_int(*o, n, 16787616U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_7(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_7) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_7) | 0);  /* check that XKB_KEY_braille_dots_7 is an integer */
  if (!_cffi_check_int(*o, n, 16787520U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_78(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_78) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_78) | 0);  /* check that XKB_KEY_braille_dots_78 is an integer */
  if (!_cffi_check_int(*o, n, 16787648U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_braille_dots_8(unsigned long long *o)
{
  int n = (XKB_KEY_braille_dots_8) <= 0;
  *o = (unsigned long long)((XKB_KEY_braille_dots_8) | 0);  /* check that XKB_KEY_braille_dots_8 is an integer */
  if (!_cffi_check_int(*o, n, 16787584U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_breve(unsigned long long *o)
{
  int n = (XKB_KEY_breve) <= 0;
  *o = (unsigned long long)((XKB_KEY_breve) | 0);  /* check that XKB_KEY_breve is an integer */
  if (!_cffi_check_int(*o, n, 418U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_brokenbar(unsigned long long *o)
{
  int n = (XKB_KEY_brokenbar) <= 0;
  *o = (unsigned long long)((XKB_KEY_brokenbar) | 0);  /* check that XKB_KEY_brokenbar is an integer */
  if (!_cffi_check_int(*o, n, 166U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_c(unsigned long long *o)
{
  int n = (XKB_KEY_c) <= 0;
  *o = (unsigned long long)((XKB_KEY_c) | 0);  /* check that XKB_KEY_c is an integer */
  if (!_cffi_check_int(*o, n, 99U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_c_h(unsigned long long *o)
{
  int n = (XKB_KEY_c_h) <= 0;
  *o = (unsigned long long)((XKB_KEY_c_h) | 0);  /* check that XKB_KEY_c_h is an integer */
  if (!_cffi_check_int(*o, n, 65187U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_cabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_cabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_cabovedot) | 0);  /* check that XKB_KEY_cabovedot is an integer */
  if (!_cffi_check_int(*o, n, 741U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_cacute(unsigned long long *o)
{
  int n = (XKB_KEY_cacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_cacute) | 0);  /* check that XKB_KEY_cacute is an integer */
  if (!_cffi_check_int(*o, n, 486U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_careof(unsigned long long *o)
{
  int n = (XKB_KEY_careof) <= 0;
  *o = (unsigned long long)((XKB_KEY_careof) | 0);  /* check that XKB_KEY_careof is an integer */
  if (!_cffi_check_int(*o, n, 2744U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_caret(unsigned long long *o)
{
  int n = (XKB_KEY_caret) <= 0;
  *o = (unsigned long long)((XKB_KEY_caret) | 0);  /* check that XKB_KEY_caret is an integer */
  if (!_cffi_check_int(*o, n, 2812U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_caron(unsigned long long *o)
{
  int n = (XKB_KEY_caron) <= 0;
  *o = (unsigned long long)((XKB_KEY_caron) | 0);  /* check that XKB_KEY_caron is an integer */
  if (!_cffi_check_int(*o, n, 439U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ccaron(unsigned long long *o)
{
  int n = (XKB_KEY_ccaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_ccaron) | 0);  /* check that XKB_KEY_ccaron is an integer */
  if (!_cffi_check_int(*o, n, 488U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ccedilla(unsigned long long *o)
{
  int n = (XKB_KEY_ccedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_ccedilla) | 0);  /* check that XKB_KEY_ccedilla is an integer */
  if (!_cffi_check_int(*o, n, 231U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ccircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_ccircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_ccircumflex) | 0);  /* check that XKB_KEY_ccircumflex is an integer */
  if (!_cffi_check_int(*o, n, 742U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_cedilla(unsigned long long *o)
{
  int n = (XKB_KEY_cedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_cedilla) | 0);  /* check that XKB_KEY_cedilla is an integer */
  if (!_cffi_check_int(*o, n, 184U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_cent(unsigned long long *o)
{
  int n = (XKB_KEY_cent) <= 0;
  *o = (unsigned long long)((XKB_KEY_cent) | 0);  /* check that XKB_KEY_cent is an integer */
  if (!_cffi_check_int(*o, n, 162U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ch(unsigned long long *o)
{
  int n = (XKB_KEY_ch) <= 0;
  *o = (unsigned long long)((XKB_KEY_ch) | 0);  /* check that XKB_KEY_ch is an integer */
  if (!_cffi_check_int(*o, n, 65184U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_checkerboard(unsigned long long *o)
{
  int n = (XKB_KEY_checkerboard) <= 0;
  *o = (unsigned long long)((XKB_KEY_checkerboard) | 0);  /* check that XKB_KEY_checkerboard is an integer */
  if (!_cffi_check_int(*o, n, 2529U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_checkmark(unsigned long long *o)
{
  int n = (XKB_KEY_checkmark) <= 0;
  *o = (unsigned long long)((XKB_KEY_checkmark) | 0);  /* check that XKB_KEY_checkmark is an integer */
  if (!_cffi_check_int(*o, n, 2803U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_circle(unsigned long long *o)
{
  int n = (XKB_KEY_circle) <= 0;
  *o = (unsigned long long)((XKB_KEY_circle) | 0);  /* check that XKB_KEY_circle is an integer */
  if (!_cffi_check_int(*o, n, 3023U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_club(unsigned long long *o)
{
  int n = (XKB_KEY_club) <= 0;
  *o = (unsigned long long)((XKB_KEY_club) | 0);  /* check that XKB_KEY_club is an integer */
  if (!_cffi_check_int(*o, n, 2796U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_colon(unsigned long long *o)
{
  int n = (XKB_KEY_colon) <= 0;
  *o = (unsigned long long)((XKB_KEY_colon) | 0);  /* check that XKB_KEY_colon is an integer */
  if (!_cffi_check_int(*o, n, 58U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_comma(unsigned long long *o)
{
  int n = (XKB_KEY_comma) <= 0;
  *o = (unsigned long long)((XKB_KEY_comma) | 0);  /* check that XKB_KEY_comma is an integer */
  if (!_cffi_check_int(*o, n, 44U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_containsas(unsigned long long *o)
{
  int n = (XKB_KEY_containsas) <= 0;
  *o = (unsigned long long)((XKB_KEY_containsas) | 0);  /* check that XKB_KEY_containsas is an integer */
  if (!_cffi_check_int(*o, n, 16785931U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_copyright(unsigned long long *o)
{
  int n = (XKB_KEY_copyright) <= 0;
  *o = (unsigned long long)((XKB_KEY_copyright) | 0);  /* check that XKB_KEY_copyright is an integer */
  if (!_cffi_check_int(*o, n, 169U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_cr(unsigned long long *o)
{
  int n = (XKB_KEY_cr) <= 0;
  *o = (unsigned long long)((XKB_KEY_cr) | 0);  /* check that XKB_KEY_cr is an integer */
  if (!_cffi_check_int(*o, n, 2532U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_crossinglines(unsigned long long *o)
{
  int n = (XKB_KEY_crossinglines) <= 0;
  *o = (unsigned long long)((XKB_KEY_crossinglines) | 0);  /* check that XKB_KEY_crossinglines is an integer */
  if (!_cffi_check_int(*o, n, 2542U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_cuberoot(unsigned long long *o)
{
  int n = (XKB_KEY_cuberoot) <= 0;
  *o = (unsigned long long)((XKB_KEY_cuberoot) | 0);  /* check that XKB_KEY_cuberoot is an integer */
  if (!_cffi_check_int(*o, n, 16785947U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_currency(unsigned long long *o)
{
  int n = (XKB_KEY_currency) <= 0;
  *o = (unsigned long long)((XKB_KEY_currency) | 0);  /* check that XKB_KEY_currency is an integer */
  if (!_cffi_check_int(*o, n, 164U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_cursor(unsigned long long *o)
{
  int n = (XKB_KEY_cursor) <= 0;
  *o = (unsigned long long)((XKB_KEY_cursor) | 0);  /* check that XKB_KEY_cursor is an integer */
  if (!_cffi_check_int(*o, n, 2815U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_d(unsigned long long *o)
{
  int n = (XKB_KEY_d) <= 0;
  *o = (unsigned long long)((XKB_KEY_d) | 0);  /* check that XKB_KEY_d is an integer */
  if (!_cffi_check_int(*o, n, 100U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_dabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_dabovedot) | 0);  /* check that XKB_KEY_dabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16784907U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dagger(unsigned long long *o)
{
  int n = (XKB_KEY_dagger) <= 0;
  *o = (unsigned long long)((XKB_KEY_dagger) | 0);  /* check that XKB_KEY_dagger is an integer */
  if (!_cffi_check_int(*o, n, 2801U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dcaron(unsigned long long *o)
{
  int n = (XKB_KEY_dcaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_dcaron) | 0);  /* check that XKB_KEY_dcaron is an integer */
  if (!_cffi_check_int(*o, n, 495U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_A(unsigned long long *o)
{
  int n = (XKB_KEY_dead_A) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_A) | 0);  /* check that XKB_KEY_dead_A is an integer */
  if (!_cffi_check_int(*o, n, 65153U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_E(unsigned long long *o)
{
  int n = (XKB_KEY_dead_E) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_E) | 0);  /* check that XKB_KEY_dead_E is an integer */
  if (!_cffi_check_int(*o, n, 65155U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_I(unsigned long long *o)
{
  int n = (XKB_KEY_dead_I) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_I) | 0);  /* check that XKB_KEY_dead_I is an integer */
  if (!_cffi_check_int(*o, n, 65157U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_O(unsigned long long *o)
{
  int n = (XKB_KEY_dead_O) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_O) | 0);  /* check that XKB_KEY_dead_O is an integer */
  if (!_cffi_check_int(*o, n, 65159U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_U(unsigned long long *o)
{
  int n = (XKB_KEY_dead_U) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_U) | 0);  /* check that XKB_KEY_dead_U is an integer */
  if (!_cffi_check_int(*o, n, 65161U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_a(unsigned long long *o)
{
  int n = (XKB_KEY_dead_a) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_a) | 0);  /* check that XKB_KEY_dead_a is an integer */
  if (!_cffi_check_int(*o, n, 65152U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_abovecomma(unsigned long long *o)
{
  int n = (XKB_KEY_dead_abovecomma) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_abovecomma) | 0);  /* check that XKB_KEY_dead_abovecomma is an integer */
  if (!_cffi_check_int(*o, n, 65124U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_abovedot(unsigned long long *o)
{
  int n = (XKB_KEY_dead_abovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_abovedot) | 0);  /* check that XKB_KEY_dead_abovedot is an integer */
  if (!_cffi_check_int(*o, n, 65110U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_abovereversedcomma(unsigned long long *o)
{
  int n = (XKB_KEY_dead_abovereversedcomma) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_abovereversedcomma) | 0);  /* check that XKB_KEY_dead_abovereversedcomma is an integer */
  if (!_cffi_check_int(*o, n, 65125U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_abovering(unsigned long long *o)
{
  int n = (XKB_KEY_dead_abovering) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_abovering) | 0);  /* check that XKB_KEY_dead_abovering is an integer */
  if (!_cffi_check_int(*o, n, 65112U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_aboveverticalline(unsigned long long *o)
{
  int n = (XKB_KEY_dead_aboveverticalline) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_aboveverticalline) | 0);  /* check that XKB_KEY_dead_aboveverticalline is an integer */
  if (!_cffi_check_int(*o, n, 65169U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_acute(unsigned long long *o)
{
  int n = (XKB_KEY_dead_acute) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_acute) | 0);  /* check that XKB_KEY_dead_acute is an integer */
  if (!_cffi_check_int(*o, n, 65105U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_belowbreve(unsigned long long *o)
{
  int n = (XKB_KEY_dead_belowbreve) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_belowbreve) | 0);  /* check that XKB_KEY_dead_belowbreve is an integer */
  if (!_cffi_check_int(*o, n, 65131U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_belowcircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_dead_belowcircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_belowcircumflex) | 0);  /* check that XKB_KEY_dead_belowcircumflex is an integer */
  if (!_cffi_check_int(*o, n, 65129U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_belowcomma(unsigned long long *o)
{
  int n = (XKB_KEY_dead_belowcomma) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_belowcomma) | 0);  /* check that XKB_KEY_dead_belowcomma is an integer */
  if (!_cffi_check_int(*o, n, 65134U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_belowdiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_dead_belowdiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_belowdiaeresis) | 0);  /* check that XKB_KEY_dead_belowdiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 65132U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_belowdot(unsigned long long *o)
{
  int n = (XKB_KEY_dead_belowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_belowdot) | 0);  /* check that XKB_KEY_dead_belowdot is an integer */
  if (!_cffi_check_int(*o, n, 65120U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_belowmacron(unsigned long long *o)
{
  int n = (XKB_KEY_dead_belowmacron) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_belowmacron) | 0);  /* check that XKB_KEY_dead_belowmacron is an integer */
  if (!_cffi_check_int(*o, n, 65128U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_belowring(unsigned long long *o)
{
  int n = (XKB_KEY_dead_belowring) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_belowring) | 0);  /* check that XKB_KEY_dead_belowring is an integer */
  if (!_cffi_check_int(*o, n, 65127U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_belowtilde(unsigned long long *o)
{
  int n = (XKB_KEY_dead_belowtilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_belowtilde) | 0);  /* check that XKB_KEY_dead_belowtilde is an integer */
  if (!_cffi_check_int(*o, n, 65130U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_belowverticalline(unsigned long long *o)
{
  int n = (XKB_KEY_dead_belowverticalline) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_belowverticalline) | 0);  /* check that XKB_KEY_dead_belowverticalline is an integer */
  if (!_cffi_check_int(*o, n, 65170U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_breve(unsigned long long *o)
{
  int n = (XKB_KEY_dead_breve) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_breve) | 0);  /* check that XKB_KEY_dead_breve is an integer */
  if (!_cffi_check_int(*o, n, 65109U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_capital_schwa(unsigned long long *o)
{
  int n = (XKB_KEY_dead_capital_schwa) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_capital_schwa) | 0);  /* check that XKB_KEY_dead_capital_schwa is an integer */
  if (!_cffi_check_int(*o, n, 65163U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_caron(unsigned long long *o)
{
  int n = (XKB_KEY_dead_caron) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_caron) | 0);  /* check that XKB_KEY_dead_caron is an integer */
  if (!_cffi_check_int(*o, n, 65114U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_cedilla(unsigned long long *o)
{
  int n = (XKB_KEY_dead_cedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_cedilla) | 0);  /* check that XKB_KEY_dead_cedilla is an integer */
  if (!_cffi_check_int(*o, n, 65115U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_circumflex(unsigned long long *o)
{
  int n = (XKB_KEY_dead_circumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_circumflex) | 0);  /* check that XKB_KEY_dead_circumflex is an integer */
  if (!_cffi_check_int(*o, n, 65106U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_currency(unsigned long long *o)
{
  int n = (XKB_KEY_dead_currency) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_currency) | 0);  /* check that XKB_KEY_dead_currency is an integer */
  if (!_cffi_check_int(*o, n, 65135U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_dasia(unsigned long long *o)
{
  int n = (XKB_KEY_dead_dasia) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_dasia) | 0);  /* check that XKB_KEY_dead_dasia is an integer */
  if (!_cffi_check_int(*o, n, 65125U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_diaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_dead_diaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_diaeresis) | 0);  /* check that XKB_KEY_dead_diaeresis is an integer */
  if (!_cffi_check_int(*o, n, 65111U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_doubleacute(unsigned long long *o)
{
  int n = (XKB_KEY_dead_doubleacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_doubleacute) | 0);  /* check that XKB_KEY_dead_doubleacute is an integer */
  if (!_cffi_check_int(*o, n, 65113U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_doublegrave(unsigned long long *o)
{
  int n = (XKB_KEY_dead_doublegrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_doublegrave) | 0);  /* check that XKB_KEY_dead_doublegrave is an integer */
  if (!_cffi_check_int(*o, n, 65126U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_e(unsigned long long *o)
{
  int n = (XKB_KEY_dead_e) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_e) | 0);  /* check that XKB_KEY_dead_e is an integer */
  if (!_cffi_check_int(*o, n, 65154U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_grave(unsigned long long *o)
{
  int n = (XKB_KEY_dead_grave) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_grave) | 0);  /* check that XKB_KEY_dead_grave is an integer */
  if (!_cffi_check_int(*o, n, 65104U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_greek(unsigned long long *o)
{
  int n = (XKB_KEY_dead_greek) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_greek) | 0);  /* check that XKB_KEY_dead_greek is an integer */
  if (!_cffi_check_int(*o, n, 65164U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_hook(unsigned long long *o)
{
  int n = (XKB_KEY_dead_hook) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_hook) | 0);  /* check that XKB_KEY_dead_hook is an integer */
  if (!_cffi_check_int(*o, n, 65121U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_horn(unsigned long long *o)
{
  int n = (XKB_KEY_dead_horn) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_horn) | 0);  /* check that XKB_KEY_dead_horn is an integer */
  if (!_cffi_check_int(*o, n, 65122U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_i(unsigned long long *o)
{
  int n = (XKB_KEY_dead_i) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_i) | 0);  /* check that XKB_KEY_dead_i is an integer */
  if (!_cffi_check_int(*o, n, 65156U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_invertedbreve(unsigned long long *o)
{
  int n = (XKB_KEY_dead_invertedbreve) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_invertedbreve) | 0);  /* check that XKB_KEY_dead_invertedbreve is an integer */
  if (!_cffi_check_int(*o, n, 65133U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_iota(unsigned long long *o)
{
  int n = (XKB_KEY_dead_iota) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_iota) | 0);  /* check that XKB_KEY_dead_iota is an integer */
  if (!_cffi_check_int(*o, n, 65117U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_longsolidusoverlay(unsigned long long *o)
{
  int n = (XKB_KEY_dead_longsolidusoverlay) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_longsolidusoverlay) | 0);  /* check that XKB_KEY_dead_longsolidusoverlay is an integer */
  if (!_cffi_check_int(*o, n, 65171U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_lowline(unsigned long long *o)
{
  int n = (XKB_KEY_dead_lowline) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_lowline) | 0);  /* check that XKB_KEY_dead_lowline is an integer */
  if (!_cffi_check_int(*o, n, 65168U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_macron(unsigned long long *o)
{
  int n = (XKB_KEY_dead_macron) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_macron) | 0);  /* check that XKB_KEY_dead_macron is an integer */
  if (!_cffi_check_int(*o, n, 65108U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_o(unsigned long long *o)
{
  int n = (XKB_KEY_dead_o) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_o) | 0);  /* check that XKB_KEY_dead_o is an integer */
  if (!_cffi_check_int(*o, n, 65158U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_ogonek(unsigned long long *o)
{
  int n = (XKB_KEY_dead_ogonek) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_ogonek) | 0);  /* check that XKB_KEY_dead_ogonek is an integer */
  if (!_cffi_check_int(*o, n, 65116U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_perispomeni(unsigned long long *o)
{
  int n = (XKB_KEY_dead_perispomeni) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_perispomeni) | 0);  /* check that XKB_KEY_dead_perispomeni is an integer */
  if (!_cffi_check_int(*o, n, 65107U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_psili(unsigned long long *o)
{
  int n = (XKB_KEY_dead_psili) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_psili) | 0);  /* check that XKB_KEY_dead_psili is an integer */
  if (!_cffi_check_int(*o, n, 65124U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_semivoiced_sound(unsigned long long *o)
{
  int n = (XKB_KEY_dead_semivoiced_sound) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_semivoiced_sound) | 0);  /* check that XKB_KEY_dead_semivoiced_sound is an integer */
  if (!_cffi_check_int(*o, n, 65119U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_small_schwa(unsigned long long *o)
{
  int n = (XKB_KEY_dead_small_schwa) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_small_schwa) | 0);  /* check that XKB_KEY_dead_small_schwa is an integer */
  if (!_cffi_check_int(*o, n, 65162U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_stroke(unsigned long long *o)
{
  int n = (XKB_KEY_dead_stroke) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_stroke) | 0);  /* check that XKB_KEY_dead_stroke is an integer */
  if (!_cffi_check_int(*o, n, 65123U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_tilde(unsigned long long *o)
{
  int n = (XKB_KEY_dead_tilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_tilde) | 0);  /* check that XKB_KEY_dead_tilde is an integer */
  if (!_cffi_check_int(*o, n, 65107U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_u(unsigned long long *o)
{
  int n = (XKB_KEY_dead_u) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_u) | 0);  /* check that XKB_KEY_dead_u is an integer */
  if (!_cffi_check_int(*o, n, 65160U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dead_voiced_sound(unsigned long long *o)
{
  int n = (XKB_KEY_dead_voiced_sound) <= 0;
  *o = (unsigned long long)((XKB_KEY_dead_voiced_sound) | 0);  /* check that XKB_KEY_dead_voiced_sound is an integer */
  if (!_cffi_check_int(*o, n, 65118U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_decimalpoint(unsigned long long *o)
{
  int n = (XKB_KEY_decimalpoint) <= 0;
  *o = (unsigned long long)((XKB_KEY_decimalpoint) | 0);  /* check that XKB_KEY_decimalpoint is an integer */
  if (!_cffi_check_int(*o, n, 2749U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_degree(unsigned long long *o)
{
  int n = (XKB_KEY_degree) <= 0;
  *o = (unsigned long long)((XKB_KEY_degree) | 0);  /* check that XKB_KEY_degree is an integer */
  if (!_cffi_check_int(*o, n, 176U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_diaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_diaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_diaeresis) | 0);  /* check that XKB_KEY_diaeresis is an integer */
  if (!_cffi_check_int(*o, n, 168U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_diamond(unsigned long long *o)
{
  int n = (XKB_KEY_diamond) <= 0;
  *o = (unsigned long long)((XKB_KEY_diamond) | 0);  /* check that XKB_KEY_diamond is an integer */
  if (!_cffi_check_int(*o, n, 2797U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_digitspace(unsigned long long *o)
{
  int n = (XKB_KEY_digitspace) <= 0;
  *o = (unsigned long long)((XKB_KEY_digitspace) | 0);  /* check that XKB_KEY_digitspace is an integer */
  if (!_cffi_check_int(*o, n, 2725U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dintegral(unsigned long long *o)
{
  int n = (XKB_KEY_dintegral) <= 0;
  *o = (unsigned long long)((XKB_KEY_dintegral) | 0);  /* check that XKB_KEY_dintegral is an integer */
  if (!_cffi_check_int(*o, n, 16785964U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_division(unsigned long long *o)
{
  int n = (XKB_KEY_division) <= 0;
  *o = (unsigned long long)((XKB_KEY_division) | 0);  /* check that XKB_KEY_division is an integer */
  if (!_cffi_check_int(*o, n, 247U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dollar(unsigned long long *o)
{
  int n = (XKB_KEY_dollar) <= 0;
  *o = (unsigned long long)((XKB_KEY_dollar) | 0);  /* check that XKB_KEY_dollar is an integer */
  if (!_cffi_check_int(*o, n, 36U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_doubbaselinedot(unsigned long long *o)
{
  int n = (XKB_KEY_doubbaselinedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_doubbaselinedot) | 0);  /* check that XKB_KEY_doubbaselinedot is an integer */
  if (!_cffi_check_int(*o, n, 2735U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_doubleacute(unsigned long long *o)
{
  int n = (XKB_KEY_doubleacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_doubleacute) | 0);  /* check that XKB_KEY_doubleacute is an integer */
  if (!_cffi_check_int(*o, n, 445U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_doubledagger(unsigned long long *o)
{
  int n = (XKB_KEY_doubledagger) <= 0;
  *o = (unsigned long long)((XKB_KEY_doubledagger) | 0);  /* check that XKB_KEY_doubledagger is an integer */
  if (!_cffi_check_int(*o, n, 2802U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_doublelowquotemark(unsigned long long *o)
{
  int n = (XKB_KEY_doublelowquotemark) <= 0;
  *o = (unsigned long long)((XKB_KEY_doublelowquotemark) | 0);  /* check that XKB_KEY_doublelowquotemark is an integer */
  if (!_cffi_check_int(*o, n, 2814U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_downarrow(unsigned long long *o)
{
  int n = (XKB_KEY_downarrow) <= 0;
  *o = (unsigned long long)((XKB_KEY_downarrow) | 0);  /* check that XKB_KEY_downarrow is an integer */
  if (!_cffi_check_int(*o, n, 2302U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_downcaret(unsigned long long *o)
{
  int n = (XKB_KEY_downcaret) <= 0;
  *o = (unsigned long long)((XKB_KEY_downcaret) | 0);  /* check that XKB_KEY_downcaret is an integer */
  if (!_cffi_check_int(*o, n, 2984U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_downshoe(unsigned long long *o)
{
  int n = (XKB_KEY_downshoe) <= 0;
  *o = (unsigned long long)((XKB_KEY_downshoe) | 0);  /* check that XKB_KEY_downshoe is an integer */
  if (!_cffi_check_int(*o, n, 3030U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_downstile(unsigned long long *o)
{
  int n = (XKB_KEY_downstile) <= 0;
  *o = (unsigned long long)((XKB_KEY_downstile) | 0);  /* check that XKB_KEY_downstile is an integer */
  if (!_cffi_check_int(*o, n, 3012U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_downtack(unsigned long long *o)
{
  int n = (XKB_KEY_downtack) <= 0;
  *o = (unsigned long long)((XKB_KEY_downtack) | 0);  /* check that XKB_KEY_downtack is an integer */
  if (!_cffi_check_int(*o, n, 3010U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_dstroke(unsigned long long *o)
{
  int n = (XKB_KEY_dstroke) <= 0;
  *o = (unsigned long long)((XKB_KEY_dstroke) | 0);  /* check that XKB_KEY_dstroke is an integer */
  if (!_cffi_check_int(*o, n, 496U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_e(unsigned long long *o)
{
  int n = (XKB_KEY_e) <= 0;
  *o = (unsigned long long)((XKB_KEY_e) | 0);  /* check that XKB_KEY_e is an integer */
  if (!_cffi_check_int(*o, n, 101U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_eabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_eabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_eabovedot) | 0);  /* check that XKB_KEY_eabovedot is an integer */
  if (!_cffi_check_int(*o, n, 1004U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_eacute(unsigned long long *o)
{
  int n = (XKB_KEY_eacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_eacute) | 0);  /* check that XKB_KEY_eacute is an integer */
  if (!_cffi_check_int(*o, n, 233U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ebelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_ebelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_ebelowdot) | 0);  /* check that XKB_KEY_ebelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785081U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ecaron(unsigned long long *o)
{
  int n = (XKB_KEY_ecaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_ecaron) | 0);  /* check that XKB_KEY_ecaron is an integer */
  if (!_cffi_check_int(*o, n, 492U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ecircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_ecircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_ecircumflex) | 0);  /* check that XKB_KEY_ecircumflex is an integer */
  if (!_cffi_check_int(*o, n, 234U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ecircumflexacute(unsigned long long *o)
{
  int n = (XKB_KEY_ecircumflexacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_ecircumflexacute) | 0);  /* check that XKB_KEY_ecircumflexacute is an integer */
  if (!_cffi_check_int(*o, n, 16785087U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ecircumflexbelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_ecircumflexbelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_ecircumflexbelowdot) | 0);  /* check that XKB_KEY_ecircumflexbelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785095U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ecircumflexgrave(unsigned long long *o)
{
  int n = (XKB_KEY_ecircumflexgrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_ecircumflexgrave) | 0);  /* check that XKB_KEY_ecircumflexgrave is an integer */
  if (!_cffi_check_int(*o, n, 16785089U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ecircumflexhook(unsigned long long *o)
{
  int n = (XKB_KEY_ecircumflexhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_ecircumflexhook) | 0);  /* check that XKB_KEY_ecircumflexhook is an integer */
  if (!_cffi_check_int(*o, n, 16785091U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ecircumflextilde(unsigned long long *o)
{
  int n = (XKB_KEY_ecircumflextilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_ecircumflextilde) | 0);  /* check that XKB_KEY_ecircumflextilde is an integer */
  if (!_cffi_check_int(*o, n, 16785093U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ediaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_ediaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_ediaeresis) | 0);  /* check that XKB_KEY_ediaeresis is an integer */
  if (!_cffi_check_int(*o, n, 235U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_egrave(unsigned long long *o)
{
  int n = (XKB_KEY_egrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_egrave) | 0);  /* check that XKB_KEY_egrave is an integer */
  if (!_cffi_check_int(*o, n, 232U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ehook(unsigned long long *o)
{
  int n = (XKB_KEY_ehook) <= 0;
  *o = (unsigned long long)((XKB_KEY_ehook) | 0);  /* check that XKB_KEY_ehook is an integer */
  if (!_cffi_check_int(*o, n, 16785083U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_eightsubscript(unsigned long long *o)
{
  int n = (XKB_KEY_eightsubscript) <= 0;
  *o = (unsigned long long)((XKB_KEY_eightsubscript) | 0);  /* check that XKB_KEY_eightsubscript is an integer */
  if (!_cffi_check_int(*o, n, 16785544U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_eightsuperior(unsigned long long *o)
{
  int n = (XKB_KEY_eightsuperior) <= 0;
  *o = (unsigned long long)((XKB_KEY_eightsuperior) | 0);  /* check that XKB_KEY_eightsuperior is an integer */
  if (!_cffi_check_int(*o, n, 16785528U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_elementof(unsigned long long *o)
{
  int n = (XKB_KEY_elementof) <= 0;
  *o = (unsigned long long)((XKB_KEY_elementof) | 0);  /* check that XKB_KEY_elementof is an integer */
  if (!_cffi_check_int(*o, n, 16785928U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ellipsis(unsigned long long *o)
{
  int n = (XKB_KEY_ellipsis) <= 0;
  *o = (unsigned long long)((XKB_KEY_ellipsis) | 0);  /* check that XKB_KEY_ellipsis is an integer */
  if (!_cffi_check_int(*o, n, 2734U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_em3space(unsigned long long *o)
{
  int n = (XKB_KEY_em3space) <= 0;
  *o = (unsigned long long)((XKB_KEY_em3space) | 0);  /* check that XKB_KEY_em3space is an integer */
  if (!_cffi_check_int(*o, n, 2723U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_em4space(unsigned long long *o)
{
  int n = (XKB_KEY_em4space) <= 0;
  *o = (unsigned long long)((XKB_KEY_em4space) | 0);  /* check that XKB_KEY_em4space is an integer */
  if (!_cffi_check_int(*o, n, 2724U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_emacron(unsigned long long *o)
{
  int n = (XKB_KEY_emacron) <= 0;
  *o = (unsigned long long)((XKB_KEY_emacron) | 0);  /* check that XKB_KEY_emacron is an integer */
  if (!_cffi_check_int(*o, n, 954U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_emdash(unsigned long long *o)
{
  int n = (XKB_KEY_emdash) <= 0;
  *o = (unsigned long long)((XKB_KEY_emdash) | 0);  /* check that XKB_KEY_emdash is an integer */
  if (!_cffi_check_int(*o, n, 2729U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_emfilledcircle(unsigned long long *o)
{
  int n = (XKB_KEY_emfilledcircle) <= 0;
  *o = (unsigned long long)((XKB_KEY_emfilledcircle) | 0);  /* check that XKB_KEY_emfilledcircle is an integer */
  if (!_cffi_check_int(*o, n, 2782U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_emfilledrect(unsigned long long *o)
{
  int n = (XKB_KEY_emfilledrect) <= 0;
  *o = (unsigned long long)((XKB_KEY_emfilledrect) | 0);  /* check that XKB_KEY_emfilledrect is an integer */
  if (!_cffi_check_int(*o, n, 2783U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_emopencircle(unsigned long long *o)
{
  int n = (XKB_KEY_emopencircle) <= 0;
  *o = (unsigned long long)((XKB_KEY_emopencircle) | 0);  /* check that XKB_KEY_emopencircle is an integer */
  if (!_cffi_check_int(*o, n, 2766U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_emopenrectangle(unsigned long long *o)
{
  int n = (XKB_KEY_emopenrectangle) <= 0;
  *o = (unsigned long long)((XKB_KEY_emopenrectangle) | 0);  /* check that XKB_KEY_emopenrectangle is an integer */
  if (!_cffi_check_int(*o, n, 2767U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_emptyset(unsigned long long *o)
{
  int n = (XKB_KEY_emptyset) <= 0;
  *o = (unsigned long long)((XKB_KEY_emptyset) | 0);  /* check that XKB_KEY_emptyset is an integer */
  if (!_cffi_check_int(*o, n, 16785925U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_emspace(unsigned long long *o)
{
  int n = (XKB_KEY_emspace) <= 0;
  *o = (unsigned long long)((XKB_KEY_emspace) | 0);  /* check that XKB_KEY_emspace is an integer */
  if (!_cffi_check_int(*o, n, 2721U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_endash(unsigned long long *o)
{
  int n = (XKB_KEY_endash) <= 0;
  *o = (unsigned long long)((XKB_KEY_endash) | 0);  /* check that XKB_KEY_endash is an integer */
  if (!_cffi_check_int(*o, n, 2730U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_enfilledcircbullet(unsigned long long *o)
{
  int n = (XKB_KEY_enfilledcircbullet) <= 0;
  *o = (unsigned long long)((XKB_KEY_enfilledcircbullet) | 0);  /* check that XKB_KEY_enfilledcircbullet is an integer */
  if (!_cffi_check_int(*o, n, 2790U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_enfilledsqbullet(unsigned long long *o)
{
  int n = (XKB_KEY_enfilledsqbullet) <= 0;
  *o = (unsigned long long)((XKB_KEY_enfilledsqbullet) | 0);  /* check that XKB_KEY_enfilledsqbullet is an integer */
  if (!_cffi_check_int(*o, n, 2791U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_eng(unsigned long long *o)
{
  int n = (XKB_KEY_eng) <= 0;
  *o = (unsigned long long)((XKB_KEY_eng) | 0);  /* check that XKB_KEY_eng is an integer */
  if (!_cffi_check_int(*o, n, 959U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_enopencircbullet(unsigned long long *o)
{
  int n = (XKB_KEY_enopencircbullet) <= 0;
  *o = (unsigned long long)((XKB_KEY_enopencircbullet) | 0);  /* check that XKB_KEY_enopencircbullet is an integer */
  if (!_cffi_check_int(*o, n, 2784U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_enopensquarebullet(unsigned long long *o)
{
  int n = (XKB_KEY_enopensquarebullet) <= 0;
  *o = (unsigned long long)((XKB_KEY_enopensquarebullet) | 0);  /* check that XKB_KEY_enopensquarebullet is an integer */
  if (!_cffi_check_int(*o, n, 2785U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_enspace(unsigned long long *o)
{
  int n = (XKB_KEY_enspace) <= 0;
  *o = (unsigned long long)((XKB_KEY_enspace) | 0);  /* check that XKB_KEY_enspace is an integer */
  if (!_cffi_check_int(*o, n, 2722U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_eogonek(unsigned long long *o)
{
  int n = (XKB_KEY_eogonek) <= 0;
  *o = (unsigned long long)((XKB_KEY_eogonek) | 0);  /* check that XKB_KEY_eogonek is an integer */
  if (!_cffi_check_int(*o, n, 490U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_equal(unsigned long long *o)
{
  int n = (XKB_KEY_equal) <= 0;
  *o = (unsigned long long)((XKB_KEY_equal) | 0);  /* check that XKB_KEY_equal is an integer */
  if (!_cffi_check_int(*o, n, 61U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_eth(unsigned long long *o)
{
  int n = (XKB_KEY_eth) <= 0;
  *o = (unsigned long long)((XKB_KEY_eth) | 0);  /* check that XKB_KEY_eth is an integer */
  if (!_cffi_check_int(*o, n, 240U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_etilde(unsigned long long *o)
{
  int n = (XKB_KEY_etilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_etilde) | 0);  /* check that XKB_KEY_etilde is an integer */
  if (!_cffi_check_int(*o, n, 16785085U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_exclam(unsigned long long *o)
{
  int n = (XKB_KEY_exclam) <= 0;
  *o = (unsigned long long)((XKB_KEY_exclam) | 0);  /* check that XKB_KEY_exclam is an integer */
  if (!_cffi_check_int(*o, n, 33U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_exclamdown(unsigned long long *o)
{
  int n = (XKB_KEY_exclamdown) <= 0;
  *o = (unsigned long long)((XKB_KEY_exclamdown) | 0);  /* check that XKB_KEY_exclamdown is an integer */
  if (!_cffi_check_int(*o, n, 161U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ezh(unsigned long long *o)
{
  int n = (XKB_KEY_ezh) <= 0;
  *o = (unsigned long long)((XKB_KEY_ezh) | 0);  /* check that XKB_KEY_ezh is an integer */
  if (!_cffi_check_int(*o, n, 16777874U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_f(unsigned long long *o)
{
  int n = (XKB_KEY_f) <= 0;
  *o = (unsigned long long)((XKB_KEY_f) | 0);  /* check that XKB_KEY_f is an integer */
  if (!_cffi_check_int(*o, n, 102U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_fabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_fabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_fabovedot) | 0);  /* check that XKB_KEY_fabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16784927U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_femalesymbol(unsigned long long *o)
{
  int n = (XKB_KEY_femalesymbol) <= 0;
  *o = (unsigned long long)((XKB_KEY_femalesymbol) | 0);  /* check that XKB_KEY_femalesymbol is an integer */
  if (!_cffi_check_int(*o, n, 2808U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ff(unsigned long long *o)
{
  int n = (XKB_KEY_ff) <= 0;
  *o = (unsigned long long)((XKB_KEY_ff) | 0);  /* check that XKB_KEY_ff is an integer */
  if (!_cffi_check_int(*o, n, 2531U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_figdash(unsigned long long *o)
{
  int n = (XKB_KEY_figdash) <= 0;
  *o = (unsigned long long)((XKB_KEY_figdash) | 0);  /* check that XKB_KEY_figdash is an integer */
  if (!_cffi_check_int(*o, n, 2747U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_filledlefttribullet(unsigned long long *o)
{
  int n = (XKB_KEY_filledlefttribullet) <= 0;
  *o = (unsigned long long)((XKB_KEY_filledlefttribullet) | 0);  /* check that XKB_KEY_filledlefttribullet is an integer */
  if (!_cffi_check_int(*o, n, 2780U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_filledrectbullet(unsigned long long *o)
{
  int n = (XKB_KEY_filledrectbullet) <= 0;
  *o = (unsigned long long)((XKB_KEY_filledrectbullet) | 0);  /* check that XKB_KEY_filledrectbullet is an integer */
  if (!_cffi_check_int(*o, n, 2779U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_filledrighttribullet(unsigned long long *o)
{
  int n = (XKB_KEY_filledrighttribullet) <= 0;
  *o = (unsigned long long)((XKB_KEY_filledrighttribullet) | 0);  /* check that XKB_KEY_filledrighttribullet is an integer */
  if (!_cffi_check_int(*o, n, 2781U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_filledtribulletdown(unsigned long long *o)
{
  int n = (XKB_KEY_filledtribulletdown) <= 0;
  *o = (unsigned long long)((XKB_KEY_filledtribulletdown) | 0);  /* check that XKB_KEY_filledtribulletdown is an integer */
  if (!_cffi_check_int(*o, n, 2793U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_filledtribulletup(unsigned long long *o)
{
  int n = (XKB_KEY_filledtribulletup) <= 0;
  *o = (unsigned long long)((XKB_KEY_filledtribulletup) | 0);  /* check that XKB_KEY_filledtribulletup is an integer */
  if (!_cffi_check_int(*o, n, 2792U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_fiveeighths(unsigned long long *o)
{
  int n = (XKB_KEY_fiveeighths) <= 0;
  *o = (unsigned long long)((XKB_KEY_fiveeighths) | 0);  /* check that XKB_KEY_fiveeighths is an integer */
  if (!_cffi_check_int(*o, n, 2757U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_fivesixths(unsigned long long *o)
{
  int n = (XKB_KEY_fivesixths) <= 0;
  *o = (unsigned long long)((XKB_KEY_fivesixths) | 0);  /* check that XKB_KEY_fivesixths is an integer */
  if (!_cffi_check_int(*o, n, 2743U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_fivesubscript(unsigned long long *o)
{
  int n = (XKB_KEY_fivesubscript) <= 0;
  *o = (unsigned long long)((XKB_KEY_fivesubscript) | 0);  /* check that XKB_KEY_fivesubscript is an integer */
  if (!_cffi_check_int(*o, n, 16785541U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_fivesuperior(unsigned long long *o)
{
  int n = (XKB_KEY_fivesuperior) <= 0;
  *o = (unsigned long long)((XKB_KEY_fivesuperior) | 0);  /* check that XKB_KEY_fivesuperior is an integer */
  if (!_cffi_check_int(*o, n, 16785525U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_fourfifths(unsigned long long *o)
{
  int n = (XKB_KEY_fourfifths) <= 0;
  *o = (unsigned long long)((XKB_KEY_fourfifths) | 0);  /* check that XKB_KEY_fourfifths is an integer */
  if (!_cffi_check_int(*o, n, 2741U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_foursubscript(unsigned long long *o)
{
  int n = (XKB_KEY_foursubscript) <= 0;
  *o = (unsigned long long)((XKB_KEY_foursubscript) | 0);  /* check that XKB_KEY_foursubscript is an integer */
  if (!_cffi_check_int(*o, n, 16785540U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_foursuperior(unsigned long long *o)
{
  int n = (XKB_KEY_foursuperior) <= 0;
  *o = (unsigned long long)((XKB_KEY_foursuperior) | 0);  /* check that XKB_KEY_foursuperior is an integer */
  if (!_cffi_check_int(*o, n, 16785524U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_fourthroot(unsigned long long *o)
{
  int n = (XKB_KEY_fourthroot) <= 0;
  *o = (unsigned long long)((XKB_KEY_fourthroot) | 0);  /* check that XKB_KEY_fourthroot is an integer */
  if (!_cffi_check_int(*o, n, 16785948U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_function(unsigned long long *o)
{
  int n = (XKB_KEY_function) <= 0;
  *o = (unsigned long long)((XKB_KEY_function) | 0);  /* check that XKB_KEY_function is an integer */
  if (!_cffi_check_int(*o, n, 2294U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_g(unsigned long long *o)
{
  int n = (XKB_KEY_g) <= 0;
  *o = (unsigned long long)((XKB_KEY_g) | 0);  /* check that XKB_KEY_g is an integer */
  if (!_cffi_check_int(*o, n, 103U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_gabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_gabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_gabovedot) | 0);  /* check that XKB_KEY_gabovedot is an integer */
  if (!_cffi_check_int(*o, n, 757U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_gbreve(unsigned long long *o)
{
  int n = (XKB_KEY_gbreve) <= 0;
  *o = (unsigned long long)((XKB_KEY_gbreve) | 0);  /* check that XKB_KEY_gbreve is an integer */
  if (!_cffi_check_int(*o, n, 699U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_gcaron(unsigned long long *o)
{
  int n = (XKB_KEY_gcaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_gcaron) | 0);  /* check that XKB_KEY_gcaron is an integer */
  if (!_cffi_check_int(*o, n, 16777703U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_gcedilla(unsigned long long *o)
{
  int n = (XKB_KEY_gcedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_gcedilla) | 0);  /* check that XKB_KEY_gcedilla is an integer */
  if (!_cffi_check_int(*o, n, 955U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_gcircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_gcircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_gcircumflex) | 0);  /* check that XKB_KEY_gcircumflex is an integer */
  if (!_cffi_check_int(*o, n, 760U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_grave(unsigned long long *o)
{
  int n = (XKB_KEY_grave) <= 0;
  *o = (unsigned long long)((XKB_KEY_grave) | 0);  /* check that XKB_KEY_grave is an integer */
  if (!_cffi_check_int(*o, n, 96U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_greater(unsigned long long *o)
{
  int n = (XKB_KEY_greater) <= 0;
  *o = (unsigned long long)((XKB_KEY_greater) | 0);  /* check that XKB_KEY_greater is an integer */
  if (!_cffi_check_int(*o, n, 62U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_greaterthanequal(unsigned long long *o)
{
  int n = (XKB_KEY_greaterthanequal) <= 0;
  *o = (unsigned long long)((XKB_KEY_greaterthanequal) | 0);  /* check that XKB_KEY_greaterthanequal is an integer */
  if (!_cffi_check_int(*o, n, 2238U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_guilder(unsigned long long *o)
{
  int n = (XKB_KEY_guilder) <= 0;
  *o = (unsigned long long)((XKB_KEY_guilder) | 0);  /* check that XKB_KEY_guilder is an integer */
  if (!_cffi_check_int(*o, n, 268435646U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_guillemotleft(unsigned long long *o)
{
  int n = (XKB_KEY_guillemotleft) <= 0;
  *o = (unsigned long long)((XKB_KEY_guillemotleft) | 0);  /* check that XKB_KEY_guillemotleft is an integer */
  if (!_cffi_check_int(*o, n, 171U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_guillemotright(unsigned long long *o)
{
  int n = (XKB_KEY_guillemotright) <= 0;
  *o = (unsigned long long)((XKB_KEY_guillemotright) | 0);  /* check that XKB_KEY_guillemotright is an integer */
  if (!_cffi_check_int(*o, n, 187U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_h(unsigned long long *o)
{
  int n = (XKB_KEY_h) <= 0;
  *o = (unsigned long long)((XKB_KEY_h) | 0);  /* check that XKB_KEY_h is an integer */
  if (!_cffi_check_int(*o, n, 104U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hairspace(unsigned long long *o)
{
  int n = (XKB_KEY_hairspace) <= 0;
  *o = (unsigned long long)((XKB_KEY_hairspace) | 0);  /* check that XKB_KEY_hairspace is an integer */
  if (!_cffi_check_int(*o, n, 2728U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hcircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_hcircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_hcircumflex) | 0);  /* check that XKB_KEY_hcircumflex is an integer */
  if (!_cffi_check_int(*o, n, 694U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_heart(unsigned long long *o)
{
  int n = (XKB_KEY_heart) <= 0;
  *o = (unsigned long long)((XKB_KEY_heart) | 0);  /* check that XKB_KEY_heart is an integer */
  if (!_cffi_check_int(*o, n, 2798U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_aleph(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_aleph) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_aleph) | 0);  /* check that XKB_KEY_hebrew_aleph is an integer */
  if (!_cffi_check_int(*o, n, 3296U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_ayin(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_ayin) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_ayin) | 0);  /* check that XKB_KEY_hebrew_ayin is an integer */
  if (!_cffi_check_int(*o, n, 3314U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_bet(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_bet) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_bet) | 0);  /* check that XKB_KEY_hebrew_bet is an integer */
  if (!_cffi_check_int(*o, n, 3297U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_beth(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_beth) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_beth) | 0);  /* check that XKB_KEY_hebrew_beth is an integer */
  if (!_cffi_check_int(*o, n, 3297U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_chet(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_chet) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_chet) | 0);  /* check that XKB_KEY_hebrew_chet is an integer */
  if (!_cffi_check_int(*o, n, 3303U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_dalet(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_dalet) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_dalet) | 0);  /* check that XKB_KEY_hebrew_dalet is an integer */
  if (!_cffi_check_int(*o, n, 3299U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_daleth(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_daleth) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_daleth) | 0);  /* check that XKB_KEY_hebrew_daleth is an integer */
  if (!_cffi_check_int(*o, n, 3299U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_doublelowline(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_doublelowline) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_doublelowline) | 0);  /* check that XKB_KEY_hebrew_doublelowline is an integer */
  if (!_cffi_check_int(*o, n, 3295U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_finalkaph(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_finalkaph) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_finalkaph) | 0);  /* check that XKB_KEY_hebrew_finalkaph is an integer */
  if (!_cffi_check_int(*o, n, 3306U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_finalmem(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_finalmem) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_finalmem) | 0);  /* check that XKB_KEY_hebrew_finalmem is an integer */
  if (!_cffi_check_int(*o, n, 3309U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_finalnun(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_finalnun) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_finalnun) | 0);  /* check that XKB_KEY_hebrew_finalnun is an integer */
  if (!_cffi_check_int(*o, n, 3311U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_finalpe(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_finalpe) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_finalpe) | 0);  /* check that XKB_KEY_hebrew_finalpe is an integer */
  if (!_cffi_check_int(*o, n, 3315U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_finalzade(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_finalzade) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_finalzade) | 0);  /* check that XKB_KEY_hebrew_finalzade is an integer */
  if (!_cffi_check_int(*o, n, 3317U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_finalzadi(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_finalzadi) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_finalzadi) | 0);  /* check that XKB_KEY_hebrew_finalzadi is an integer */
  if (!_cffi_check_int(*o, n, 3317U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_gimel(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_gimel) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_gimel) | 0);  /* check that XKB_KEY_hebrew_gimel is an integer */
  if (!_cffi_check_int(*o, n, 3298U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_gimmel(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_gimmel) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_gimmel) | 0);  /* check that XKB_KEY_hebrew_gimmel is an integer */
  if (!_cffi_check_int(*o, n, 3298U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_he(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_he) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_he) | 0);  /* check that XKB_KEY_hebrew_he is an integer */
  if (!_cffi_check_int(*o, n, 3300U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_het(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_het) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_het) | 0);  /* check that XKB_KEY_hebrew_het is an integer */
  if (!_cffi_check_int(*o, n, 3303U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_kaph(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_kaph) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_kaph) | 0);  /* check that XKB_KEY_hebrew_kaph is an integer */
  if (!_cffi_check_int(*o, n, 3307U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_kuf(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_kuf) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_kuf) | 0);  /* check that XKB_KEY_hebrew_kuf is an integer */
  if (!_cffi_check_int(*o, n, 3319U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_lamed(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_lamed) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_lamed) | 0);  /* check that XKB_KEY_hebrew_lamed is an integer */
  if (!_cffi_check_int(*o, n, 3308U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_mem(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_mem) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_mem) | 0);  /* check that XKB_KEY_hebrew_mem is an integer */
  if (!_cffi_check_int(*o, n, 3310U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_nun(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_nun) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_nun) | 0);  /* check that XKB_KEY_hebrew_nun is an integer */
  if (!_cffi_check_int(*o, n, 3312U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_pe(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_pe) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_pe) | 0);  /* check that XKB_KEY_hebrew_pe is an integer */
  if (!_cffi_check_int(*o, n, 3316U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_qoph(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_qoph) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_qoph) | 0);  /* check that XKB_KEY_hebrew_qoph is an integer */
  if (!_cffi_check_int(*o, n, 3319U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_resh(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_resh) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_resh) | 0);  /* check that XKB_KEY_hebrew_resh is an integer */
  if (!_cffi_check_int(*o, n, 3320U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_samech(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_samech) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_samech) | 0);  /* check that XKB_KEY_hebrew_samech is an integer */
  if (!_cffi_check_int(*o, n, 3313U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_samekh(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_samekh) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_samekh) | 0);  /* check that XKB_KEY_hebrew_samekh is an integer */
  if (!_cffi_check_int(*o, n, 3313U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_shin(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_shin) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_shin) | 0);  /* check that XKB_KEY_hebrew_shin is an integer */
  if (!_cffi_check_int(*o, n, 3321U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_taf(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_taf) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_taf) | 0);  /* check that XKB_KEY_hebrew_taf is an integer */
  if (!_cffi_check_int(*o, n, 3322U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_taw(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_taw) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_taw) | 0);  /* check that XKB_KEY_hebrew_taw is an integer */
  if (!_cffi_check_int(*o, n, 3322U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_tet(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_tet) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_tet) | 0);  /* check that XKB_KEY_hebrew_tet is an integer */
  if (!_cffi_check_int(*o, n, 3304U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_teth(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_teth) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_teth) | 0);  /* check that XKB_KEY_hebrew_teth is an integer */
  if (!_cffi_check_int(*o, n, 3304U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_waw(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_waw) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_waw) | 0);  /* check that XKB_KEY_hebrew_waw is an integer */
  if (!_cffi_check_int(*o, n, 3301U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_yod(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_yod) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_yod) | 0);  /* check that XKB_KEY_hebrew_yod is an integer */
  if (!_cffi_check_int(*o, n, 3305U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_zade(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_zade) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_zade) | 0);  /* check that XKB_KEY_hebrew_zade is an integer */
  if (!_cffi_check_int(*o, n, 3318U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_zadi(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_zadi) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_zadi) | 0);  /* check that XKB_KEY_hebrew_zadi is an integer */
  if (!_cffi_check_int(*o, n, 3318U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_zain(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_zain) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_zain) | 0);  /* check that XKB_KEY_hebrew_zain is an integer */
  if (!_cffi_check_int(*o, n, 3302U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hebrew_zayin(unsigned long long *o)
{
  int n = (XKB_KEY_hebrew_zayin) <= 0;
  *o = (unsigned long long)((XKB_KEY_hebrew_zayin) | 0);  /* check that XKB_KEY_hebrew_zayin is an integer */
  if (!_cffi_check_int(*o, n, 3302U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hexagram(unsigned long long *o)
{
  int n = (XKB_KEY_hexagram) <= 0;
  *o = (unsigned long long)((XKB_KEY_hexagram) | 0);  /* check that XKB_KEY_hexagram is an integer */
  if (!_cffi_check_int(*o, n, 2778U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_horizconnector(unsigned long long *o)
{
  int n = (XKB_KEY_horizconnector) <= 0;
  *o = (unsigned long long)((XKB_KEY_horizconnector) | 0);  /* check that XKB_KEY_horizconnector is an integer */
  if (!_cffi_check_int(*o, n, 2211U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_horizlinescan1(unsigned long long *o)
{
  int n = (XKB_KEY_horizlinescan1) <= 0;
  *o = (unsigned long long)((XKB_KEY_horizlinescan1) | 0);  /* check that XKB_KEY_horizlinescan1 is an integer */
  if (!_cffi_check_int(*o, n, 2543U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_horizlinescan3(unsigned long long *o)
{
  int n = (XKB_KEY_horizlinescan3) <= 0;
  *o = (unsigned long long)((XKB_KEY_horizlinescan3) | 0);  /* check that XKB_KEY_horizlinescan3 is an integer */
  if (!_cffi_check_int(*o, n, 2544U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_horizlinescan5(unsigned long long *o)
{
  int n = (XKB_KEY_horizlinescan5) <= 0;
  *o = (unsigned long long)((XKB_KEY_horizlinescan5) | 0);  /* check that XKB_KEY_horizlinescan5 is an integer */
  if (!_cffi_check_int(*o, n, 2545U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_horizlinescan7(unsigned long long *o)
{
  int n = (XKB_KEY_horizlinescan7) <= 0;
  *o = (unsigned long long)((XKB_KEY_horizlinescan7) | 0);  /* check that XKB_KEY_horizlinescan7 is an integer */
  if (!_cffi_check_int(*o, n, 2546U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_horizlinescan9(unsigned long long *o)
{
  int n = (XKB_KEY_horizlinescan9) <= 0;
  *o = (unsigned long long)((XKB_KEY_horizlinescan9) | 0);  /* check that XKB_KEY_horizlinescan9 is an integer */
  if (!_cffi_check_int(*o, n, 2547U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpBackTab(unsigned long long *o)
{
  int n = (XKB_KEY_hpBackTab) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpBackTab) | 0);  /* check that XKB_KEY_hpBackTab is an integer */
  if (!_cffi_check_int(*o, n, 268500852U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpClearLine(unsigned long long *o)
{
  int n = (XKB_KEY_hpClearLine) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpClearLine) | 0);  /* check that XKB_KEY_hpClearLine is an integer */
  if (!_cffi_check_int(*o, n, 268500847U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpDeleteChar(unsigned long long *o)
{
  int n = (XKB_KEY_hpDeleteChar) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpDeleteChar) | 0);  /* check that XKB_KEY_hpDeleteChar is an integer */
  if (!_cffi_check_int(*o, n, 268500851U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpDeleteLine(unsigned long long *o)
{
  int n = (XKB_KEY_hpDeleteLine) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpDeleteLine) | 0);  /* check that XKB_KEY_hpDeleteLine is an integer */
  if (!_cffi_check_int(*o, n, 268500849U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpIO(unsigned long long *o)
{
  int n = (XKB_KEY_hpIO) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpIO) | 0);  /* check that XKB_KEY_hpIO is an integer */
  if (!_cffi_check_int(*o, n, 268435694U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpInsertChar(unsigned long long *o)
{
  int n = (XKB_KEY_hpInsertChar) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpInsertChar) | 0);  /* check that XKB_KEY_hpInsertChar is an integer */
  if (!_cffi_check_int(*o, n, 268500850U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpInsertLine(unsigned long long *o)
{
  int n = (XKB_KEY_hpInsertLine) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpInsertLine) | 0);  /* check that XKB_KEY_hpInsertLine is an integer */
  if (!_cffi_check_int(*o, n, 268500848U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpKP_BackTab(unsigned long long *o)
{
  int n = (XKB_KEY_hpKP_BackTab) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpKP_BackTab) | 0);  /* check that XKB_KEY_hpKP_BackTab is an integer */
  if (!_cffi_check_int(*o, n, 268500853U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpModelock1(unsigned long long *o)
{
  int n = (XKB_KEY_hpModelock1) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpModelock1) | 0);  /* check that XKB_KEY_hpModelock1 is an integer */
  if (!_cffi_check_int(*o, n, 268500808U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpModelock2(unsigned long long *o)
{
  int n = (XKB_KEY_hpModelock2) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpModelock2) | 0);  /* check that XKB_KEY_hpModelock2 is an integer */
  if (!_cffi_check_int(*o, n, 268500809U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpReset(unsigned long long *o)
{
  int n = (XKB_KEY_hpReset) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpReset) | 0);  /* check that XKB_KEY_hpReset is an integer */
  if (!_cffi_check_int(*o, n, 268500844U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpSystem(unsigned long long *o)
{
  int n = (XKB_KEY_hpSystem) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpSystem) | 0);  /* check that XKB_KEY_hpSystem is an integer */
  if (!_cffi_check_int(*o, n, 268500845U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpUser(unsigned long long *o)
{
  int n = (XKB_KEY_hpUser) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpUser) | 0);  /* check that XKB_KEY_hpUser is an integer */
  if (!_cffi_check_int(*o, n, 268500846U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpYdiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_hpYdiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpYdiaeresis) | 0);  /* check that XKB_KEY_hpYdiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 268435694U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpblock(unsigned long long *o)
{
  int n = (XKB_KEY_hpblock) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpblock) | 0);  /* check that XKB_KEY_hpblock is an integer */
  if (!_cffi_check_int(*o, n, 268435708U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpguilder(unsigned long long *o)
{
  int n = (XKB_KEY_hpguilder) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpguilder) | 0);  /* check that XKB_KEY_hpguilder is an integer */
  if (!_cffi_check_int(*o, n, 268435646U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hplira(unsigned long long *o)
{
  int n = (XKB_KEY_hplira) <= 0;
  *o = (unsigned long long)((XKB_KEY_hplira) | 0);  /* check that XKB_KEY_hplira is an integer */
  if (!_cffi_check_int(*o, n, 268435631U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hplongminus(unsigned long long *o)
{
  int n = (XKB_KEY_hplongminus) <= 0;
  *o = (unsigned long long)((XKB_KEY_hplongminus) | 0);  /* check that XKB_KEY_hplongminus is an integer */
  if (!_cffi_check_int(*o, n, 268435702U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpmute_acute(unsigned long long *o)
{
  int n = (XKB_KEY_hpmute_acute) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpmute_acute) | 0);  /* check that XKB_KEY_hpmute_acute is an integer */
  if (!_cffi_check_int(*o, n, 268435624U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpmute_asciicircum(unsigned long long *o)
{
  int n = (XKB_KEY_hpmute_asciicircum) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpmute_asciicircum) | 0);  /* check that XKB_KEY_hpmute_asciicircum is an integer */
  if (!_cffi_check_int(*o, n, 268435626U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpmute_asciitilde(unsigned long long *o)
{
  int n = (XKB_KEY_hpmute_asciitilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpmute_asciitilde) | 0);  /* check that XKB_KEY_hpmute_asciitilde is an integer */
  if (!_cffi_check_int(*o, n, 268435628U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpmute_diaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_hpmute_diaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpmute_diaeresis) | 0);  /* check that XKB_KEY_hpmute_diaeresis is an integer */
  if (!_cffi_check_int(*o, n, 268435627U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hpmute_grave(unsigned long long *o)
{
  int n = (XKB_KEY_hpmute_grave) <= 0;
  *o = (unsigned long long)((XKB_KEY_hpmute_grave) | 0);  /* check that XKB_KEY_hpmute_grave is an integer */
  if (!_cffi_check_int(*o, n, 268435625U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hstroke(unsigned long long *o)
{
  int n = (XKB_KEY_hstroke) <= 0;
  *o = (unsigned long long)((XKB_KEY_hstroke) | 0);  /* check that XKB_KEY_hstroke is an integer */
  if (!_cffi_check_int(*o, n, 689U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ht(unsigned long long *o)
{
  int n = (XKB_KEY_ht) <= 0;
  *o = (unsigned long long)((XKB_KEY_ht) | 0);  /* check that XKB_KEY_ht is an integer */
  if (!_cffi_check_int(*o, n, 2530U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_hyphen(unsigned long long *o)
{
  int n = (XKB_KEY_hyphen) <= 0;
  *o = (unsigned long long)((XKB_KEY_hyphen) | 0);  /* check that XKB_KEY_hyphen is an integer */
  if (!_cffi_check_int(*o, n, 173U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_i(unsigned long long *o)
{
  int n = (XKB_KEY_i) <= 0;
  *o = (unsigned long long)((XKB_KEY_i) | 0);  /* check that XKB_KEY_i is an integer */
  if (!_cffi_check_int(*o, n, 105U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_iacute(unsigned long long *o)
{
  int n = (XKB_KEY_iacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_iacute) | 0);  /* check that XKB_KEY_iacute is an integer */
  if (!_cffi_check_int(*o, n, 237U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ibelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_ibelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_ibelowdot) | 0);  /* check that XKB_KEY_ibelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785099U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ibreve(unsigned long long *o)
{
  int n = (XKB_KEY_ibreve) <= 0;
  *o = (unsigned long long)((XKB_KEY_ibreve) | 0);  /* check that XKB_KEY_ibreve is an integer */
  if (!_cffi_check_int(*o, n, 16777517U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_icircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_icircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_icircumflex) | 0);  /* check that XKB_KEY_icircumflex is an integer */
  if (!_cffi_check_int(*o, n, 238U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_identical(unsigned long long *o)
{
  int n = (XKB_KEY_identical) <= 0;
  *o = (unsigned long long)((XKB_KEY_identical) | 0);  /* check that XKB_KEY_identical is an integer */
  if (!_cffi_check_int(*o, n, 2255U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_idiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_idiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_idiaeresis) | 0);  /* check that XKB_KEY_idiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 239U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_idotless(unsigned long long *o)
{
  int n = (XKB_KEY_idotless) <= 0;
  *o = (unsigned long long)((XKB_KEY_idotless) | 0);  /* check that XKB_KEY_idotless is an integer */
  if (!_cffi_check_int(*o, n, 697U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ifonlyif(unsigned long long *o)
{
  int n = (XKB_KEY_ifonlyif) <= 0;
  *o = (unsigned long long)((XKB_KEY_ifonlyif) | 0);  /* check that XKB_KEY_ifonlyif is an integer */
  if (!_cffi_check_int(*o, n, 2253U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_igrave(unsigned long long *o)
{
  int n = (XKB_KEY_igrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_igrave) | 0);  /* check that XKB_KEY_igrave is an integer */
  if (!_cffi_check_int(*o, n, 236U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ihook(unsigned long long *o)
{
  int n = (XKB_KEY_ihook) <= 0;
  *o = (unsigned long long)((XKB_KEY_ihook) | 0);  /* check that XKB_KEY_ihook is an integer */
  if (!_cffi_check_int(*o, n, 16785097U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_imacron(unsigned long long *o)
{
  int n = (XKB_KEY_imacron) <= 0;
  *o = (unsigned long long)((XKB_KEY_imacron) | 0);  /* check that XKB_KEY_imacron is an integer */
  if (!_cffi_check_int(*o, n, 1007U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_implies(unsigned long long *o)
{
  int n = (XKB_KEY_implies) <= 0;
  *o = (unsigned long long)((XKB_KEY_implies) | 0);  /* check that XKB_KEY_implies is an integer */
  if (!_cffi_check_int(*o, n, 2254U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_includedin(unsigned long long *o)
{
  int n = (XKB_KEY_includedin) <= 0;
  *o = (unsigned long long)((XKB_KEY_includedin) | 0);  /* check that XKB_KEY_includedin is an integer */
  if (!_cffi_check_int(*o, n, 2266U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_includes(unsigned long long *o)
{
  int n = (XKB_KEY_includes) <= 0;
  *o = (unsigned long long)((XKB_KEY_includes) | 0);  /* check that XKB_KEY_includes is an integer */
  if (!_cffi_check_int(*o, n, 2267U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_infinity(unsigned long long *o)
{
  int n = (XKB_KEY_infinity) <= 0;
  *o = (unsigned long long)((XKB_KEY_infinity) | 0);  /* check that XKB_KEY_infinity is an integer */
  if (!_cffi_check_int(*o, n, 2242U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_integral(unsigned long long *o)
{
  int n = (XKB_KEY_integral) <= 0;
  *o = (unsigned long long)((XKB_KEY_integral) | 0);  /* check that XKB_KEY_integral is an integer */
  if (!_cffi_check_int(*o, n, 2239U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_intersection(unsigned long long *o)
{
  int n = (XKB_KEY_intersection) <= 0;
  *o = (unsigned long long)((XKB_KEY_intersection) | 0);  /* check that XKB_KEY_intersection is an integer */
  if (!_cffi_check_int(*o, n, 2268U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_iogonek(unsigned long long *o)
{
  int n = (XKB_KEY_iogonek) <= 0;
  *o = (unsigned long long)((XKB_KEY_iogonek) | 0);  /* check that XKB_KEY_iogonek is an integer */
  if (!_cffi_check_int(*o, n, 999U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_itilde(unsigned long long *o)
{
  int n = (XKB_KEY_itilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_itilde) | 0);  /* check that XKB_KEY_itilde is an integer */
  if (!_cffi_check_int(*o, n, 949U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_j(unsigned long long *o)
{
  int n = (XKB_KEY_j) <= 0;
  *o = (unsigned long long)((XKB_KEY_j) | 0);  /* check that XKB_KEY_j is an integer */
  if (!_cffi_check_int(*o, n, 106U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_jcircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_jcircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_jcircumflex) | 0);  /* check that XKB_KEY_jcircumflex is an integer */
  if (!_cffi_check_int(*o, n, 700U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_jot(unsigned long long *o)
{
  int n = (XKB_KEY_jot) <= 0;
  *o = (unsigned long long)((XKB_KEY_jot) | 0);  /* check that XKB_KEY_jot is an integer */
  if (!_cffi_check_int(*o, n, 3018U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_k(unsigned long long *o)
{
  int n = (XKB_KEY_k) <= 0;
  *o = (unsigned long long)((XKB_KEY_k) | 0);  /* check that XKB_KEY_k is an integer */
  if (!_cffi_check_int(*o, n, 107U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_A(unsigned long long *o)
{
  int n = (XKB_KEY_kana_A) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_A) | 0);  /* check that XKB_KEY_kana_A is an integer */
  if (!_cffi_check_int(*o, n, 1201U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_CHI(unsigned long long *o)
{
  int n = (XKB_KEY_kana_CHI) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_CHI) | 0);  /* check that XKB_KEY_kana_CHI is an integer */
  if (!_cffi_check_int(*o, n, 1217U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_E(unsigned long long *o)
{
  int n = (XKB_KEY_kana_E) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_E) | 0);  /* check that XKB_KEY_kana_E is an integer */
  if (!_cffi_check_int(*o, n, 1204U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_FU(unsigned long long *o)
{
  int n = (XKB_KEY_kana_FU) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_FU) | 0);  /* check that XKB_KEY_kana_FU is an integer */
  if (!_cffi_check_int(*o, n, 1228U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_HA(unsigned long long *o)
{
  int n = (XKB_KEY_kana_HA) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_HA) | 0);  /* check that XKB_KEY_kana_HA is an integer */
  if (!_cffi_check_int(*o, n, 1226U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_HE(unsigned long long *o)
{
  int n = (XKB_KEY_kana_HE) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_HE) | 0);  /* check that XKB_KEY_kana_HE is an integer */
  if (!_cffi_check_int(*o, n, 1229U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_HI(unsigned long long *o)
{
  int n = (XKB_KEY_kana_HI) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_HI) | 0);  /* check that XKB_KEY_kana_HI is an integer */
  if (!_cffi_check_int(*o, n, 1227U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_HO(unsigned long long *o)
{
  int n = (XKB_KEY_kana_HO) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_HO) | 0);  /* check that XKB_KEY_kana_HO is an integer */
  if (!_cffi_check_int(*o, n, 1230U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_HU(unsigned long long *o)
{
  int n = (XKB_KEY_kana_HU) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_HU) | 0);  /* check that XKB_KEY_kana_HU is an integer */
  if (!_cffi_check_int(*o, n, 1228U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_I(unsigned long long *o)
{
  int n = (XKB_KEY_kana_I) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_I) | 0);  /* check that XKB_KEY_kana_I is an integer */
  if (!_cffi_check_int(*o, n, 1202U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_KA(unsigned long long *o)
{
  int n = (XKB_KEY_kana_KA) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_KA) | 0);  /* check that XKB_KEY_kana_KA is an integer */
  if (!_cffi_check_int(*o, n, 1206U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_KE(unsigned long long *o)
{
  int n = (XKB_KEY_kana_KE) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_KE) | 0);  /* check that XKB_KEY_kana_KE is an integer */
  if (!_cffi_check_int(*o, n, 1209U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_KI(unsigned long long *o)
{
  int n = (XKB_KEY_kana_KI) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_KI) | 0);  /* check that XKB_KEY_kana_KI is an integer */
  if (!_cffi_check_int(*o, n, 1207U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_KO(unsigned long long *o)
{
  int n = (XKB_KEY_kana_KO) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_KO) | 0);  /* check that XKB_KEY_kana_KO is an integer */
  if (!_cffi_check_int(*o, n, 1210U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_KU(unsigned long long *o)
{
  int n = (XKB_KEY_kana_KU) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_KU) | 0);  /* check that XKB_KEY_kana_KU is an integer */
  if (!_cffi_check_int(*o, n, 1208U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_MA(unsigned long long *o)
{
  int n = (XKB_KEY_kana_MA) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_MA) | 0);  /* check that XKB_KEY_kana_MA is an integer */
  if (!_cffi_check_int(*o, n, 1231U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_ME(unsigned long long *o)
{
  int n = (XKB_KEY_kana_ME) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_ME) | 0);  /* check that XKB_KEY_kana_ME is an integer */
  if (!_cffi_check_int(*o, n, 1234U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_MI(unsigned long long *o)
{
  int n = (XKB_KEY_kana_MI) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_MI) | 0);  /* check that XKB_KEY_kana_MI is an integer */
  if (!_cffi_check_int(*o, n, 1232U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_MO(unsigned long long *o)
{
  int n = (XKB_KEY_kana_MO) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_MO) | 0);  /* check that XKB_KEY_kana_MO is an integer */
  if (!_cffi_check_int(*o, n, 1235U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_MU(unsigned long long *o)
{
  int n = (XKB_KEY_kana_MU) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_MU) | 0);  /* check that XKB_KEY_kana_MU is an integer */
  if (!_cffi_check_int(*o, n, 1233U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_N(unsigned long long *o)
{
  int n = (XKB_KEY_kana_N) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_N) | 0);  /* check that XKB_KEY_kana_N is an integer */
  if (!_cffi_check_int(*o, n, 1245U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_NA(unsigned long long *o)
{
  int n = (XKB_KEY_kana_NA) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_NA) | 0);  /* check that XKB_KEY_kana_NA is an integer */
  if (!_cffi_check_int(*o, n, 1221U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_NE(unsigned long long *o)
{
  int n = (XKB_KEY_kana_NE) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_NE) | 0);  /* check that XKB_KEY_kana_NE is an integer */
  if (!_cffi_check_int(*o, n, 1224U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_NI(unsigned long long *o)
{
  int n = (XKB_KEY_kana_NI) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_NI) | 0);  /* check that XKB_KEY_kana_NI is an integer */
  if (!_cffi_check_int(*o, n, 1222U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_NO(unsigned long long *o)
{
  int n = (XKB_KEY_kana_NO) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_NO) | 0);  /* check that XKB_KEY_kana_NO is an integer */
  if (!_cffi_check_int(*o, n, 1225U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_NU(unsigned long long *o)
{
  int n = (XKB_KEY_kana_NU) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_NU) | 0);  /* check that XKB_KEY_kana_NU is an integer */
  if (!_cffi_check_int(*o, n, 1223U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_O(unsigned long long *o)
{
  int n = (XKB_KEY_kana_O) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_O) | 0);  /* check that XKB_KEY_kana_O is an integer */
  if (!_cffi_check_int(*o, n, 1205U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_RA(unsigned long long *o)
{
  int n = (XKB_KEY_kana_RA) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_RA) | 0);  /* check that XKB_KEY_kana_RA is an integer */
  if (!_cffi_check_int(*o, n, 1239U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_RE(unsigned long long *o)
{
  int n = (XKB_KEY_kana_RE) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_RE) | 0);  /* check that XKB_KEY_kana_RE is an integer */
  if (!_cffi_check_int(*o, n, 1242U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_RI(unsigned long long *o)
{
  int n = (XKB_KEY_kana_RI) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_RI) | 0);  /* check that XKB_KEY_kana_RI is an integer */
  if (!_cffi_check_int(*o, n, 1240U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_RO(unsigned long long *o)
{
  int n = (XKB_KEY_kana_RO) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_RO) | 0);  /* check that XKB_KEY_kana_RO is an integer */
  if (!_cffi_check_int(*o, n, 1243U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_RU(unsigned long long *o)
{
  int n = (XKB_KEY_kana_RU) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_RU) | 0);  /* check that XKB_KEY_kana_RU is an integer */
  if (!_cffi_check_int(*o, n, 1241U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_SA(unsigned long long *o)
{
  int n = (XKB_KEY_kana_SA) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_SA) | 0);  /* check that XKB_KEY_kana_SA is an integer */
  if (!_cffi_check_int(*o, n, 1211U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_SE(unsigned long long *o)
{
  int n = (XKB_KEY_kana_SE) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_SE) | 0);  /* check that XKB_KEY_kana_SE is an integer */
  if (!_cffi_check_int(*o, n, 1214U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_SHI(unsigned long long *o)
{
  int n = (XKB_KEY_kana_SHI) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_SHI) | 0);  /* check that XKB_KEY_kana_SHI is an integer */
  if (!_cffi_check_int(*o, n, 1212U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_SO(unsigned long long *o)
{
  int n = (XKB_KEY_kana_SO) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_SO) | 0);  /* check that XKB_KEY_kana_SO is an integer */
  if (!_cffi_check_int(*o, n, 1215U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_SU(unsigned long long *o)
{
  int n = (XKB_KEY_kana_SU) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_SU) | 0);  /* check that XKB_KEY_kana_SU is an integer */
  if (!_cffi_check_int(*o, n, 1213U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_TA(unsigned long long *o)
{
  int n = (XKB_KEY_kana_TA) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_TA) | 0);  /* check that XKB_KEY_kana_TA is an integer */
  if (!_cffi_check_int(*o, n, 1216U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_TE(unsigned long long *o)
{
  int n = (XKB_KEY_kana_TE) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_TE) | 0);  /* check that XKB_KEY_kana_TE is an integer */
  if (!_cffi_check_int(*o, n, 1219U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_TI(unsigned long long *o)
{
  int n = (XKB_KEY_kana_TI) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_TI) | 0);  /* check that XKB_KEY_kana_TI is an integer */
  if (!_cffi_check_int(*o, n, 1217U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_TO(unsigned long long *o)
{
  int n = (XKB_KEY_kana_TO) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_TO) | 0);  /* check that XKB_KEY_kana_TO is an integer */
  if (!_cffi_check_int(*o, n, 1220U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_TSU(unsigned long long *o)
{
  int n = (XKB_KEY_kana_TSU) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_TSU) | 0);  /* check that XKB_KEY_kana_TSU is an integer */
  if (!_cffi_check_int(*o, n, 1218U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_TU(unsigned long long *o)
{
  int n = (XKB_KEY_kana_TU) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_TU) | 0);  /* check that XKB_KEY_kana_TU is an integer */
  if (!_cffi_check_int(*o, n, 1218U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_U(unsigned long long *o)
{
  int n = (XKB_KEY_kana_U) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_U) | 0);  /* check that XKB_KEY_kana_U is an integer */
  if (!_cffi_check_int(*o, n, 1203U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_WA(unsigned long long *o)
{
  int n = (XKB_KEY_kana_WA) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_WA) | 0);  /* check that XKB_KEY_kana_WA is an integer */
  if (!_cffi_check_int(*o, n, 1244U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_WO(unsigned long long *o)
{
  int n = (XKB_KEY_kana_WO) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_WO) | 0);  /* check that XKB_KEY_kana_WO is an integer */
  if (!_cffi_check_int(*o, n, 1190U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_YA(unsigned long long *o)
{
  int n = (XKB_KEY_kana_YA) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_YA) | 0);  /* check that XKB_KEY_kana_YA is an integer */
  if (!_cffi_check_int(*o, n, 1236U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_YO(unsigned long long *o)
{
  int n = (XKB_KEY_kana_YO) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_YO) | 0);  /* check that XKB_KEY_kana_YO is an integer */
  if (!_cffi_check_int(*o, n, 1238U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_YU(unsigned long long *o)
{
  int n = (XKB_KEY_kana_YU) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_YU) | 0);  /* check that XKB_KEY_kana_YU is an integer */
  if (!_cffi_check_int(*o, n, 1237U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_a(unsigned long long *o)
{
  int n = (XKB_KEY_kana_a) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_a) | 0);  /* check that XKB_KEY_kana_a is an integer */
  if (!_cffi_check_int(*o, n, 1191U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_closingbracket(unsigned long long *o)
{
  int n = (XKB_KEY_kana_closingbracket) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_closingbracket) | 0);  /* check that XKB_KEY_kana_closingbracket is an integer */
  if (!_cffi_check_int(*o, n, 1187U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_comma(unsigned long long *o)
{
  int n = (XKB_KEY_kana_comma) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_comma) | 0);  /* check that XKB_KEY_kana_comma is an integer */
  if (!_cffi_check_int(*o, n, 1188U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_conjunctive(unsigned long long *o)
{
  int n = (XKB_KEY_kana_conjunctive) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_conjunctive) | 0);  /* check that XKB_KEY_kana_conjunctive is an integer */
  if (!_cffi_check_int(*o, n, 1189U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_e(unsigned long long *o)
{
  int n = (XKB_KEY_kana_e) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_e) | 0);  /* check that XKB_KEY_kana_e is an integer */
  if (!_cffi_check_int(*o, n, 1194U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_fullstop(unsigned long long *o)
{
  int n = (XKB_KEY_kana_fullstop) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_fullstop) | 0);  /* check that XKB_KEY_kana_fullstop is an integer */
  if (!_cffi_check_int(*o, n, 1185U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_i(unsigned long long *o)
{
  int n = (XKB_KEY_kana_i) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_i) | 0);  /* check that XKB_KEY_kana_i is an integer */
  if (!_cffi_check_int(*o, n, 1192U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_middledot(unsigned long long *o)
{
  int n = (XKB_KEY_kana_middledot) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_middledot) | 0);  /* check that XKB_KEY_kana_middledot is an integer */
  if (!_cffi_check_int(*o, n, 1189U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_o(unsigned long long *o)
{
  int n = (XKB_KEY_kana_o) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_o) | 0);  /* check that XKB_KEY_kana_o is an integer */
  if (!_cffi_check_int(*o, n, 1195U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_openingbracket(unsigned long long *o)
{
  int n = (XKB_KEY_kana_openingbracket) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_openingbracket) | 0);  /* check that XKB_KEY_kana_openingbracket is an integer */
  if (!_cffi_check_int(*o, n, 1186U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_switch(unsigned long long *o)
{
  int n = (XKB_KEY_kana_switch) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_switch) | 0);  /* check that XKB_KEY_kana_switch is an integer */
  if (!_cffi_check_int(*o, n, 65406U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_tsu(unsigned long long *o)
{
  int n = (XKB_KEY_kana_tsu) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_tsu) | 0);  /* check that XKB_KEY_kana_tsu is an integer */
  if (!_cffi_check_int(*o, n, 1199U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_tu(unsigned long long *o)
{
  int n = (XKB_KEY_kana_tu) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_tu) | 0);  /* check that XKB_KEY_kana_tu is an integer */
  if (!_cffi_check_int(*o, n, 1199U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_u(unsigned long long *o)
{
  int n = (XKB_KEY_kana_u) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_u) | 0);  /* check that XKB_KEY_kana_u is an integer */
  if (!_cffi_check_int(*o, n, 1193U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_ya(unsigned long long *o)
{
  int n = (XKB_KEY_kana_ya) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_ya) | 0);  /* check that XKB_KEY_kana_ya is an integer */
  if (!_cffi_check_int(*o, n, 1196U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_yo(unsigned long long *o)
{
  int n = (XKB_KEY_kana_yo) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_yo) | 0);  /* check that XKB_KEY_kana_yo is an integer */
  if (!_cffi_check_int(*o, n, 1198U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kana_yu(unsigned long long *o)
{
  int n = (XKB_KEY_kana_yu) <= 0;
  *o = (unsigned long long)((XKB_KEY_kana_yu) | 0);  /* check that XKB_KEY_kana_yu is an integer */
  if (!_cffi_check_int(*o, n, 1197U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kappa(unsigned long long *o)
{
  int n = (XKB_KEY_kappa) <= 0;
  *o = (unsigned long long)((XKB_KEY_kappa) | 0);  /* check that XKB_KEY_kappa is an integer */
  if (!_cffi_check_int(*o, n, 930U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kcedilla(unsigned long long *o)
{
  int n = (XKB_KEY_kcedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_kcedilla) | 0);  /* check that XKB_KEY_kcedilla is an integer */
  if (!_cffi_check_int(*o, n, 1011U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_kra(unsigned long long *o)
{
  int n = (XKB_KEY_kra) <= 0;
  *o = (unsigned long long)((XKB_KEY_kra) | 0);  /* check that XKB_KEY_kra is an integer */
  if (!_cffi_check_int(*o, n, 930U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_l(unsigned long long *o)
{
  int n = (XKB_KEY_l) <= 0;
  *o = (unsigned long long)((XKB_KEY_l) | 0);  /* check that XKB_KEY_l is an integer */
  if (!_cffi_check_int(*o, n, 108U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_lacute(unsigned long long *o)
{
  int n = (XKB_KEY_lacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_lacute) | 0);  /* check that XKB_KEY_lacute is an integer */
  if (!_cffi_check_int(*o, n, 485U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_latincross(unsigned long long *o)
{
  int n = (XKB_KEY_latincross) <= 0;
  *o = (unsigned long long)((XKB_KEY_latincross) | 0);  /* check that XKB_KEY_latincross is an integer */
  if (!_cffi_check_int(*o, n, 2777U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_lbelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_lbelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_lbelowdot) | 0);  /* check that XKB_KEY_lbelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16784951U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_lcaron(unsigned long long *o)
{
  int n = (XKB_KEY_lcaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_lcaron) | 0);  /* check that XKB_KEY_lcaron is an integer */
  if (!_cffi_check_int(*o, n, 437U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_lcedilla(unsigned long long *o)
{
  int n = (XKB_KEY_lcedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_lcedilla) | 0);  /* check that XKB_KEY_lcedilla is an integer */
  if (!_cffi_check_int(*o, n, 950U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_leftanglebracket(unsigned long long *o)
{
  int n = (XKB_KEY_leftanglebracket) <= 0;
  *o = (unsigned long long)((XKB_KEY_leftanglebracket) | 0);  /* check that XKB_KEY_leftanglebracket is an integer */
  if (!_cffi_check_int(*o, n, 2748U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_leftarrow(unsigned long long *o)
{
  int n = (XKB_KEY_leftarrow) <= 0;
  *o = (unsigned long long)((XKB_KEY_leftarrow) | 0);  /* check that XKB_KEY_leftarrow is an integer */
  if (!_cffi_check_int(*o, n, 2299U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_leftcaret(unsigned long long *o)
{
  int n = (XKB_KEY_leftcaret) <= 0;
  *o = (unsigned long long)((XKB_KEY_leftcaret) | 0);  /* check that XKB_KEY_leftcaret is an integer */
  if (!_cffi_check_int(*o, n, 2979U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_leftdoublequotemark(unsigned long long *o)
{
  int n = (XKB_KEY_leftdoublequotemark) <= 0;
  *o = (unsigned long long)((XKB_KEY_leftdoublequotemark) | 0);  /* check that XKB_KEY_leftdoublequotemark is an integer */
  if (!_cffi_check_int(*o, n, 2770U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_leftmiddlecurlybrace(unsigned long long *o)
{
  int n = (XKB_KEY_leftmiddlecurlybrace) <= 0;
  *o = (unsigned long long)((XKB_KEY_leftmiddlecurlybrace) | 0);  /* check that XKB_KEY_leftmiddlecurlybrace is an integer */
  if (!_cffi_check_int(*o, n, 2223U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_leftopentriangle(unsigned long long *o)
{
  int n = (XKB_KEY_leftopentriangle) <= 0;
  *o = (unsigned long long)((XKB_KEY_leftopentriangle) | 0);  /* check that XKB_KEY_leftopentriangle is an integer */
  if (!_cffi_check_int(*o, n, 2764U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_leftpointer(unsigned long long *o)
{
  int n = (XKB_KEY_leftpointer) <= 0;
  *o = (unsigned long long)((XKB_KEY_leftpointer) | 0);  /* check that XKB_KEY_leftpointer is an integer */
  if (!_cffi_check_int(*o, n, 2794U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_leftradical(unsigned long long *o)
{
  int n = (XKB_KEY_leftradical) <= 0;
  *o = (unsigned long long)((XKB_KEY_leftradical) | 0);  /* check that XKB_KEY_leftradical is an integer */
  if (!_cffi_check_int(*o, n, 2209U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_leftshoe(unsigned long long *o)
{
  int n = (XKB_KEY_leftshoe) <= 0;
  *o = (unsigned long long)((XKB_KEY_leftshoe) | 0);  /* check that XKB_KEY_leftshoe is an integer */
  if (!_cffi_check_int(*o, n, 3034U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_leftsinglequotemark(unsigned long long *o)
{
  int n = (XKB_KEY_leftsinglequotemark) <= 0;
  *o = (unsigned long long)((XKB_KEY_leftsinglequotemark) | 0);  /* check that XKB_KEY_leftsinglequotemark is an integer */
  if (!_cffi_check_int(*o, n, 2768U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_leftt(unsigned long long *o)
{
  int n = (XKB_KEY_leftt) <= 0;
  *o = (unsigned long long)((XKB_KEY_leftt) | 0);  /* check that XKB_KEY_leftt is an integer */
  if (!_cffi_check_int(*o, n, 2548U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_lefttack(unsigned long long *o)
{
  int n = (XKB_KEY_lefttack) <= 0;
  *o = (unsigned long long)((XKB_KEY_lefttack) | 0);  /* check that XKB_KEY_lefttack is an integer */
  if (!_cffi_check_int(*o, n, 3036U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_less(unsigned long long *o)
{
  int n = (XKB_KEY_less) <= 0;
  *o = (unsigned long long)((XKB_KEY_less) | 0);  /* check that XKB_KEY_less is an integer */
  if (!_cffi_check_int(*o, n, 60U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_lessthanequal(unsigned long long *o)
{
  int n = (XKB_KEY_lessthanequal) <= 0;
  *o = (unsigned long long)((XKB_KEY_lessthanequal) | 0);  /* check that XKB_KEY_lessthanequal is an integer */
  if (!_cffi_check_int(*o, n, 2236U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_lf(unsigned long long *o)
{
  int n = (XKB_KEY_lf) <= 0;
  *o = (unsigned long long)((XKB_KEY_lf) | 0);  /* check that XKB_KEY_lf is an integer */
  if (!_cffi_check_int(*o, n, 2533U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_lira(unsigned long long *o)
{
  int n = (XKB_KEY_lira) <= 0;
  *o = (unsigned long long)((XKB_KEY_lira) | 0);  /* check that XKB_KEY_lira is an integer */
  if (!_cffi_check_int(*o, n, 268435631U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_logicaland(unsigned long long *o)
{
  int n = (XKB_KEY_logicaland) <= 0;
  *o = (unsigned long long)((XKB_KEY_logicaland) | 0);  /* check that XKB_KEY_logicaland is an integer */
  if (!_cffi_check_int(*o, n, 2270U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_logicalor(unsigned long long *o)
{
  int n = (XKB_KEY_logicalor) <= 0;
  *o = (unsigned long long)((XKB_KEY_logicalor) | 0);  /* check that XKB_KEY_logicalor is an integer */
  if (!_cffi_check_int(*o, n, 2271U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_longminus(unsigned long long *o)
{
  int n = (XKB_KEY_longminus) <= 0;
  *o = (unsigned long long)((XKB_KEY_longminus) | 0);  /* check that XKB_KEY_longminus is an integer */
  if (!_cffi_check_int(*o, n, 268435702U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_lowleftcorner(unsigned long long *o)
{
  int n = (XKB_KEY_lowleftcorner) <= 0;
  *o = (unsigned long long)((XKB_KEY_lowleftcorner) | 0);  /* check that XKB_KEY_lowleftcorner is an integer */
  if (!_cffi_check_int(*o, n, 2541U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_lowrightcorner(unsigned long long *o)
{
  int n = (XKB_KEY_lowrightcorner) <= 0;
  *o = (unsigned long long)((XKB_KEY_lowrightcorner) | 0);  /* check that XKB_KEY_lowrightcorner is an integer */
  if (!_cffi_check_int(*o, n, 2538U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_lstroke(unsigned long long *o)
{
  int n = (XKB_KEY_lstroke) <= 0;
  *o = (unsigned long long)((XKB_KEY_lstroke) | 0);  /* check that XKB_KEY_lstroke is an integer */
  if (!_cffi_check_int(*o, n, 435U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_m(unsigned long long *o)
{
  int n = (XKB_KEY_m) <= 0;
  *o = (unsigned long long)((XKB_KEY_m) | 0);  /* check that XKB_KEY_m is an integer */
  if (!_cffi_check_int(*o, n, 109U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_mabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_mabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_mabovedot) | 0);  /* check that XKB_KEY_mabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16784961U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_macron(unsigned long long *o)
{
  int n = (XKB_KEY_macron) <= 0;
  *o = (unsigned long long)((XKB_KEY_macron) | 0);  /* check that XKB_KEY_macron is an integer */
  if (!_cffi_check_int(*o, n, 175U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_malesymbol(unsigned long long *o)
{
  int n = (XKB_KEY_malesymbol) <= 0;
  *o = (unsigned long long)((XKB_KEY_malesymbol) | 0);  /* check that XKB_KEY_malesymbol is an integer */
  if (!_cffi_check_int(*o, n, 2807U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_maltesecross(unsigned long long *o)
{
  int n = (XKB_KEY_maltesecross) <= 0;
  *o = (unsigned long long)((XKB_KEY_maltesecross) | 0);  /* check that XKB_KEY_maltesecross is an integer */
  if (!_cffi_check_int(*o, n, 2800U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_marker(unsigned long long *o)
{
  int n = (XKB_KEY_marker) <= 0;
  *o = (unsigned long long)((XKB_KEY_marker) | 0);  /* check that XKB_KEY_marker is an integer */
  if (!_cffi_check_int(*o, n, 2751U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_masculine(unsigned long long *o)
{
  int n = (XKB_KEY_masculine) <= 0;
  *o = (unsigned long long)((XKB_KEY_masculine) | 0);  /* check that XKB_KEY_masculine is an integer */
  if (!_cffi_check_int(*o, n, 186U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_minus(unsigned long long *o)
{
  int n = (XKB_KEY_minus) <= 0;
  *o = (unsigned long long)((XKB_KEY_minus) | 0);  /* check that XKB_KEY_minus is an integer */
  if (!_cffi_check_int(*o, n, 45U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_minutes(unsigned long long *o)
{
  int n = (XKB_KEY_minutes) <= 0;
  *o = (unsigned long long)((XKB_KEY_minutes) | 0);  /* check that XKB_KEY_minutes is an integer */
  if (!_cffi_check_int(*o, n, 2774U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_mu(unsigned long long *o)
{
  int n = (XKB_KEY_mu) <= 0;
  *o = (unsigned long long)((XKB_KEY_mu) | 0);  /* check that XKB_KEY_mu is an integer */
  if (!_cffi_check_int(*o, n, 181U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_multiply(unsigned long long *o)
{
  int n = (XKB_KEY_multiply) <= 0;
  *o = (unsigned long long)((XKB_KEY_multiply) | 0);  /* check that XKB_KEY_multiply is an integer */
  if (!_cffi_check_int(*o, n, 215U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_musicalflat(unsigned long long *o)
{
  int n = (XKB_KEY_musicalflat) <= 0;
  *o = (unsigned long long)((XKB_KEY_musicalflat) | 0);  /* check that XKB_KEY_musicalflat is an integer */
  if (!_cffi_check_int(*o, n, 2806U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_musicalsharp(unsigned long long *o)
{
  int n = (XKB_KEY_musicalsharp) <= 0;
  *o = (unsigned long long)((XKB_KEY_musicalsharp) | 0);  /* check that XKB_KEY_musicalsharp is an integer */
  if (!_cffi_check_int(*o, n, 2805U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_mute_acute(unsigned long long *o)
{
  int n = (XKB_KEY_mute_acute) <= 0;
  *o = (unsigned long long)((XKB_KEY_mute_acute) | 0);  /* check that XKB_KEY_mute_acute is an integer */
  if (!_cffi_check_int(*o, n, 268435624U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_mute_asciicircum(unsigned long long *o)
{
  int n = (XKB_KEY_mute_asciicircum) <= 0;
  *o = (unsigned long long)((XKB_KEY_mute_asciicircum) | 0);  /* check that XKB_KEY_mute_asciicircum is an integer */
  if (!_cffi_check_int(*o, n, 268435626U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_mute_asciitilde(unsigned long long *o)
{
  int n = (XKB_KEY_mute_asciitilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_mute_asciitilde) | 0);  /* check that XKB_KEY_mute_asciitilde is an integer */
  if (!_cffi_check_int(*o, n, 268435628U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_mute_diaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_mute_diaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_mute_diaeresis) | 0);  /* check that XKB_KEY_mute_diaeresis is an integer */
  if (!_cffi_check_int(*o, n, 268435627U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_mute_grave(unsigned long long *o)
{
  int n = (XKB_KEY_mute_grave) <= 0;
  *o = (unsigned long long)((XKB_KEY_mute_grave) | 0);  /* check that XKB_KEY_mute_grave is an integer */
  if (!_cffi_check_int(*o, n, 268435625U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_n(unsigned long long *o)
{
  int n = (XKB_KEY_n) <= 0;
  *o = (unsigned long long)((XKB_KEY_n) | 0);  /* check that XKB_KEY_n is an integer */
  if (!_cffi_check_int(*o, n, 110U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_nabla(unsigned long long *o)
{
  int n = (XKB_KEY_nabla) <= 0;
  *o = (unsigned long long)((XKB_KEY_nabla) | 0);  /* check that XKB_KEY_nabla is an integer */
  if (!_cffi_check_int(*o, n, 2245U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_nacute(unsigned long long *o)
{
  int n = (XKB_KEY_nacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_nacute) | 0);  /* check that XKB_KEY_nacute is an integer */
  if (!_cffi_check_int(*o, n, 497U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ncaron(unsigned long long *o)
{
  int n = (XKB_KEY_ncaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_ncaron) | 0);  /* check that XKB_KEY_ncaron is an integer */
  if (!_cffi_check_int(*o, n, 498U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ncedilla(unsigned long long *o)
{
  int n = (XKB_KEY_ncedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_ncedilla) | 0);  /* check that XKB_KEY_ncedilla is an integer */
  if (!_cffi_check_int(*o, n, 1009U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ninesubscript(unsigned long long *o)
{
  int n = (XKB_KEY_ninesubscript) <= 0;
  *o = (unsigned long long)((XKB_KEY_ninesubscript) | 0);  /* check that XKB_KEY_ninesubscript is an integer */
  if (!_cffi_check_int(*o, n, 16785545U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ninesuperior(unsigned long long *o)
{
  int n = (XKB_KEY_ninesuperior) <= 0;
  *o = (unsigned long long)((XKB_KEY_ninesuperior) | 0);  /* check that XKB_KEY_ninesuperior is an integer */
  if (!_cffi_check_int(*o, n, 16785529U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_nl(unsigned long long *o)
{
  int n = (XKB_KEY_nl) <= 0;
  *o = (unsigned long long)((XKB_KEY_nl) | 0);  /* check that XKB_KEY_nl is an integer */
  if (!_cffi_check_int(*o, n, 2536U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_nobreakspace(unsigned long long *o)
{
  int n = (XKB_KEY_nobreakspace) <= 0;
  *o = (unsigned long long)((XKB_KEY_nobreakspace) | 0);  /* check that XKB_KEY_nobreakspace is an integer */
  if (!_cffi_check_int(*o, n, 160U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_notapproxeq(unsigned long long *o)
{
  int n = (XKB_KEY_notapproxeq) <= 0;
  *o = (unsigned long long)((XKB_KEY_notapproxeq) | 0);  /* check that XKB_KEY_notapproxeq is an integer */
  if (!_cffi_check_int(*o, n, 16785991U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_notelementof(unsigned long long *o)
{
  int n = (XKB_KEY_notelementof) <= 0;
  *o = (unsigned long long)((XKB_KEY_notelementof) | 0);  /* check that XKB_KEY_notelementof is an integer */
  if (!_cffi_check_int(*o, n, 16785929U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_notequal(unsigned long long *o)
{
  int n = (XKB_KEY_notequal) <= 0;
  *o = (unsigned long long)((XKB_KEY_notequal) | 0);  /* check that XKB_KEY_notequal is an integer */
  if (!_cffi_check_int(*o, n, 2237U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_notidentical(unsigned long long *o)
{
  int n = (XKB_KEY_notidentical) <= 0;
  *o = (unsigned long long)((XKB_KEY_notidentical) | 0);  /* check that XKB_KEY_notidentical is an integer */
  if (!_cffi_check_int(*o, n, 16786018U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_notsign(unsigned long long *o)
{
  int n = (XKB_KEY_notsign) <= 0;
  *o = (unsigned long long)((XKB_KEY_notsign) | 0);  /* check that XKB_KEY_notsign is an integer */
  if (!_cffi_check_int(*o, n, 172U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ntilde(unsigned long long *o)
{
  int n = (XKB_KEY_ntilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_ntilde) | 0);  /* check that XKB_KEY_ntilde is an integer */
  if (!_cffi_check_int(*o, n, 241U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_numbersign(unsigned long long *o)
{
  int n = (XKB_KEY_numbersign) <= 0;
  *o = (unsigned long long)((XKB_KEY_numbersign) | 0);  /* check that XKB_KEY_numbersign is an integer */
  if (!_cffi_check_int(*o, n, 35U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_numerosign(unsigned long long *o)
{
  int n = (XKB_KEY_numerosign) <= 0;
  *o = (unsigned long long)((XKB_KEY_numerosign) | 0);  /* check that XKB_KEY_numerosign is an integer */
  if (!_cffi_check_int(*o, n, 1712U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_o(unsigned long long *o)
{
  int n = (XKB_KEY_o) <= 0;
  *o = (unsigned long long)((XKB_KEY_o) | 0);  /* check that XKB_KEY_o is an integer */
  if (!_cffi_check_int(*o, n, 111U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_oacute(unsigned long long *o)
{
  int n = (XKB_KEY_oacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_oacute) | 0);  /* check that XKB_KEY_oacute is an integer */
  if (!_cffi_check_int(*o, n, 243U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_obarred(unsigned long long *o)
{
  int n = (XKB_KEY_obarred) <= 0;
  *o = (unsigned long long)((XKB_KEY_obarred) | 0);  /* check that XKB_KEY_obarred is an integer */
  if (!_cffi_check_int(*o, n, 16777845U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_obelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_obelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_obelowdot) | 0);  /* check that XKB_KEY_obelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785101U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ocaron(unsigned long long *o)
{
  int n = (XKB_KEY_ocaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_ocaron) | 0);  /* check that XKB_KEY_ocaron is an integer */
  if (!_cffi_check_int(*o, n, 16777682U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ocircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_ocircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_ocircumflex) | 0);  /* check that XKB_KEY_ocircumflex is an integer */
  if (!_cffi_check_int(*o, n, 244U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ocircumflexacute(unsigned long long *o)
{
  int n = (XKB_KEY_ocircumflexacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_ocircumflexacute) | 0);  /* check that XKB_KEY_ocircumflexacute is an integer */
  if (!_cffi_check_int(*o, n, 16785105U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ocircumflexbelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_ocircumflexbelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_ocircumflexbelowdot) | 0);  /* check that XKB_KEY_ocircumflexbelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785113U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ocircumflexgrave(unsigned long long *o)
{
  int n = (XKB_KEY_ocircumflexgrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_ocircumflexgrave) | 0);  /* check that XKB_KEY_ocircumflexgrave is an integer */
  if (!_cffi_check_int(*o, n, 16785107U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ocircumflexhook(unsigned long long *o)
{
  int n = (XKB_KEY_ocircumflexhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_ocircumflexhook) | 0);  /* check that XKB_KEY_ocircumflexhook is an integer */
  if (!_cffi_check_int(*o, n, 16785109U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ocircumflextilde(unsigned long long *o)
{
  int n = (XKB_KEY_ocircumflextilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_ocircumflextilde) | 0);  /* check that XKB_KEY_ocircumflextilde is an integer */
  if (!_cffi_check_int(*o, n, 16785111U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_odiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_odiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_odiaeresis) | 0);  /* check that XKB_KEY_odiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 246U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_odoubleacute(unsigned long long *o)
{
  int n = (XKB_KEY_odoubleacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_odoubleacute) | 0);  /* check that XKB_KEY_odoubleacute is an integer */
  if (!_cffi_check_int(*o, n, 501U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_oe(unsigned long long *o)
{
  int n = (XKB_KEY_oe) <= 0;
  *o = (unsigned long long)((XKB_KEY_oe) | 0);  /* check that XKB_KEY_oe is an integer */
  if (!_cffi_check_int(*o, n, 5053U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ogonek(unsigned long long *o)
{
  int n = (XKB_KEY_ogonek) <= 0;
  *o = (unsigned long long)((XKB_KEY_ogonek) | 0);  /* check that XKB_KEY_ogonek is an integer */
  if (!_cffi_check_int(*o, n, 434U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ograve(unsigned long long *o)
{
  int n = (XKB_KEY_ograve) <= 0;
  *o = (unsigned long long)((XKB_KEY_ograve) | 0);  /* check that XKB_KEY_ograve is an integer */
  if (!_cffi_check_int(*o, n, 242U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ohook(unsigned long long *o)
{
  int n = (XKB_KEY_ohook) <= 0;
  *o = (unsigned long long)((XKB_KEY_ohook) | 0);  /* check that XKB_KEY_ohook is an integer */
  if (!_cffi_check_int(*o, n, 16785103U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ohorn(unsigned long long *o)
{
  int n = (XKB_KEY_ohorn) <= 0;
  *o = (unsigned long long)((XKB_KEY_ohorn) | 0);  /* check that XKB_KEY_ohorn is an integer */
  if (!_cffi_check_int(*o, n, 16777633U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ohornacute(unsigned long long *o)
{
  int n = (XKB_KEY_ohornacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_ohornacute) | 0);  /* check that XKB_KEY_ohornacute is an integer */
  if (!_cffi_check_int(*o, n, 16785115U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ohornbelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_ohornbelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_ohornbelowdot) | 0);  /* check that XKB_KEY_ohornbelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785123U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ohorngrave(unsigned long long *o)
{
  int n = (XKB_KEY_ohorngrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_ohorngrave) | 0);  /* check that XKB_KEY_ohorngrave is an integer */
  if (!_cffi_check_int(*o, n, 16785117U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ohornhook(unsigned long long *o)
{
  int n = (XKB_KEY_ohornhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_ohornhook) | 0);  /* check that XKB_KEY_ohornhook is an integer */
  if (!_cffi_check_int(*o, n, 16785119U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ohorntilde(unsigned long long *o)
{
  int n = (XKB_KEY_ohorntilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_ohorntilde) | 0);  /* check that XKB_KEY_ohorntilde is an integer */
  if (!_cffi_check_int(*o, n, 16785121U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_omacron(unsigned long long *o)
{
  int n = (XKB_KEY_omacron) <= 0;
  *o = (unsigned long long)((XKB_KEY_omacron) | 0);  /* check that XKB_KEY_omacron is an integer */
  if (!_cffi_check_int(*o, n, 1010U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_oneeighth(unsigned long long *o)
{
  int n = (XKB_KEY_oneeighth) <= 0;
  *o = (unsigned long long)((XKB_KEY_oneeighth) | 0);  /* check that XKB_KEY_oneeighth is an integer */
  if (!_cffi_check_int(*o, n, 2755U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_onefifth(unsigned long long *o)
{
  int n = (XKB_KEY_onefifth) <= 0;
  *o = (unsigned long long)((XKB_KEY_onefifth) | 0);  /* check that XKB_KEY_onefifth is an integer */
  if (!_cffi_check_int(*o, n, 2738U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_onehalf(unsigned long long *o)
{
  int n = (XKB_KEY_onehalf) <= 0;
  *o = (unsigned long long)((XKB_KEY_onehalf) | 0);  /* check that XKB_KEY_onehalf is an integer */
  if (!_cffi_check_int(*o, n, 189U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_onequarter(unsigned long long *o)
{
  int n = (XKB_KEY_onequarter) <= 0;
  *o = (unsigned long long)((XKB_KEY_onequarter) | 0);  /* check that XKB_KEY_onequarter is an integer */
  if (!_cffi_check_int(*o, n, 188U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_onesixth(unsigned long long *o)
{
  int n = (XKB_KEY_onesixth) <= 0;
  *o = (unsigned long long)((XKB_KEY_onesixth) | 0);  /* check that XKB_KEY_onesixth is an integer */
  if (!_cffi_check_int(*o, n, 2742U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_onesubscript(unsigned long long *o)
{
  int n = (XKB_KEY_onesubscript) <= 0;
  *o = (unsigned long long)((XKB_KEY_onesubscript) | 0);  /* check that XKB_KEY_onesubscript is an integer */
  if (!_cffi_check_int(*o, n, 16785537U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_onesuperior(unsigned long long *o)
{
  int n = (XKB_KEY_onesuperior) <= 0;
  *o = (unsigned long long)((XKB_KEY_onesuperior) | 0);  /* check that XKB_KEY_onesuperior is an integer */
  if (!_cffi_check_int(*o, n, 185U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_onethird(unsigned long long *o)
{
  int n = (XKB_KEY_onethird) <= 0;
  *o = (unsigned long long)((XKB_KEY_onethird) | 0);  /* check that XKB_KEY_onethird is an integer */
  if (!_cffi_check_int(*o, n, 2736U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ooblique(unsigned long long *o)
{
  int n = (XKB_KEY_ooblique) <= 0;
  *o = (unsigned long long)((XKB_KEY_ooblique) | 0);  /* check that XKB_KEY_ooblique is an integer */
  if (!_cffi_check_int(*o, n, 248U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_openrectbullet(unsigned long long *o)
{
  int n = (XKB_KEY_openrectbullet) <= 0;
  *o = (unsigned long long)((XKB_KEY_openrectbullet) | 0);  /* check that XKB_KEY_openrectbullet is an integer */
  if (!_cffi_check_int(*o, n, 2786U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_openstar(unsigned long long *o)
{
  int n = (XKB_KEY_openstar) <= 0;
  *o = (unsigned long long)((XKB_KEY_openstar) | 0);  /* check that XKB_KEY_openstar is an integer */
  if (!_cffi_check_int(*o, n, 2789U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_opentribulletdown(unsigned long long *o)
{
  int n = (XKB_KEY_opentribulletdown) <= 0;
  *o = (unsigned long long)((XKB_KEY_opentribulletdown) | 0);  /* check that XKB_KEY_opentribulletdown is an integer */
  if (!_cffi_check_int(*o, n, 2788U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_opentribulletup(unsigned long long *o)
{
  int n = (XKB_KEY_opentribulletup) <= 0;
  *o = (unsigned long long)((XKB_KEY_opentribulletup) | 0);  /* check that XKB_KEY_opentribulletup is an integer */
  if (!_cffi_check_int(*o, n, 2787U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ordfeminine(unsigned long long *o)
{
  int n = (XKB_KEY_ordfeminine) <= 0;
  *o = (unsigned long long)((XKB_KEY_ordfeminine) | 0);  /* check that XKB_KEY_ordfeminine is an integer */
  if (!_cffi_check_int(*o, n, 170U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfActivate(unsigned long long *o)
{
  int n = (XKB_KEY_osfActivate) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfActivate) | 0);  /* check that XKB_KEY_osfActivate is an integer */
  if (!_cffi_check_int(*o, n, 268762948U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfAddMode(unsigned long long *o)
{
  int n = (XKB_KEY_osfAddMode) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfAddMode) | 0);  /* check that XKB_KEY_osfAddMode is an integer */
  if (!_cffi_check_int(*o, n, 268762929U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfBackSpace(unsigned long long *o)
{
  int n = (XKB_KEY_osfBackSpace) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfBackSpace) | 0);  /* check that XKB_KEY_osfBackSpace is an integer */
  if (!_cffi_check_int(*o, n, 268762888U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfBackTab(unsigned long long *o)
{
  int n = (XKB_KEY_osfBackTab) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfBackTab) | 0);  /* check that XKB_KEY_osfBackTab is an integer */
  if (!_cffi_check_int(*o, n, 268762887U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfBeginData(unsigned long long *o)
{
  int n = (XKB_KEY_osfBeginData) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfBeginData) | 0);  /* check that XKB_KEY_osfBeginData is an integer */
  if (!_cffi_check_int(*o, n, 268762970U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfBeginLine(unsigned long long *o)
{
  int n = (XKB_KEY_osfBeginLine) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfBeginLine) | 0);  /* check that XKB_KEY_osfBeginLine is an integer */
  if (!_cffi_check_int(*o, n, 268762968U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfCancel(unsigned long long *o)
{
  int n = (XKB_KEY_osfCancel) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfCancel) | 0);  /* check that XKB_KEY_osfCancel is an integer */
  if (!_cffi_check_int(*o, n, 268762985U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfClear(unsigned long long *o)
{
  int n = (XKB_KEY_osfClear) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfClear) | 0);  /* check that XKB_KEY_osfClear is an integer */
  if (!_cffi_check_int(*o, n, 268762891U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfCopy(unsigned long long *o)
{
  int n = (XKB_KEY_osfCopy) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfCopy) | 0);  /* check that XKB_KEY_osfCopy is an integer */
  if (!_cffi_check_int(*o, n, 268762882U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfCut(unsigned long long *o)
{
  int n = (XKB_KEY_osfCut) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfCut) | 0);  /* check that XKB_KEY_osfCut is an integer */
  if (!_cffi_check_int(*o, n, 268762883U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfDelete(unsigned long long *o)
{
  int n = (XKB_KEY_osfDelete) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfDelete) | 0);  /* check that XKB_KEY_osfDelete is an integer */
  if (!_cffi_check_int(*o, n, 268763135U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfDeselectAll(unsigned long long *o)
{
  int n = (XKB_KEY_osfDeselectAll) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfDeselectAll) | 0);  /* check that XKB_KEY_osfDeselectAll is an integer */
  if (!_cffi_check_int(*o, n, 268762994U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfDown(unsigned long long *o)
{
  int n = (XKB_KEY_osfDown) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfDown) | 0);  /* check that XKB_KEY_osfDown is an integer */
  if (!_cffi_check_int(*o, n, 268762964U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfEndData(unsigned long long *o)
{
  int n = (XKB_KEY_osfEndData) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfEndData) | 0);  /* check that XKB_KEY_osfEndData is an integer */
  if (!_cffi_check_int(*o, n, 268762969U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfEndLine(unsigned long long *o)
{
  int n = (XKB_KEY_osfEndLine) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfEndLine) | 0);  /* check that XKB_KEY_osfEndLine is an integer */
  if (!_cffi_check_int(*o, n, 268762967U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfEscape(unsigned long long *o)
{
  int n = (XKB_KEY_osfEscape) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfEscape) | 0);  /* check that XKB_KEY_osfEscape is an integer */
  if (!_cffi_check_int(*o, n, 268762907U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfExtend(unsigned long long *o)
{
  int n = (XKB_KEY_osfExtend) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfExtend) | 0);  /* check that XKB_KEY_osfExtend is an integer */
  if (!_cffi_check_int(*o, n, 268762996U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfHelp(unsigned long long *o)
{
  int n = (XKB_KEY_osfHelp) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfHelp) | 0);  /* check that XKB_KEY_osfHelp is an integer */
  if (!_cffi_check_int(*o, n, 268762986U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfInsert(unsigned long long *o)
{
  int n = (XKB_KEY_osfInsert) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfInsert) | 0);  /* check that XKB_KEY_osfInsert is an integer */
  if (!_cffi_check_int(*o, n, 268762979U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfLeft(unsigned long long *o)
{
  int n = (XKB_KEY_osfLeft) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfLeft) | 0);  /* check that XKB_KEY_osfLeft is an integer */
  if (!_cffi_check_int(*o, n, 268762961U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfMenu(unsigned long long *o)
{
  int n = (XKB_KEY_osfMenu) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfMenu) | 0);  /* check that XKB_KEY_osfMenu is an integer */
  if (!_cffi_check_int(*o, n, 268762983U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfMenuBar(unsigned long long *o)
{
  int n = (XKB_KEY_osfMenuBar) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfMenuBar) | 0);  /* check that XKB_KEY_osfMenuBar is an integer */
  if (!_cffi_check_int(*o, n, 268762949U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfNextField(unsigned long long *o)
{
  int n = (XKB_KEY_osfNextField) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfNextField) | 0);  /* check that XKB_KEY_osfNextField is an integer */
  if (!_cffi_check_int(*o, n, 268762974U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfNextMenu(unsigned long long *o)
{
  int n = (XKB_KEY_osfNextMenu) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfNextMenu) | 0);  /* check that XKB_KEY_osfNextMenu is an integer */
  if (!_cffi_check_int(*o, n, 268762972U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfPageDown(unsigned long long *o)
{
  int n = (XKB_KEY_osfPageDown) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfPageDown) | 0);  /* check that XKB_KEY_osfPageDown is an integer */
  if (!_cffi_check_int(*o, n, 268762946U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfPageLeft(unsigned long long *o)
{
  int n = (XKB_KEY_osfPageLeft) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfPageLeft) | 0);  /* check that XKB_KEY_osfPageLeft is an integer */
  if (!_cffi_check_int(*o, n, 268762944U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfPageRight(unsigned long long *o)
{
  int n = (XKB_KEY_osfPageRight) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfPageRight) | 0);  /* check that XKB_KEY_osfPageRight is an integer */
  if (!_cffi_check_int(*o, n, 268762947U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfPageUp(unsigned long long *o)
{
  int n = (XKB_KEY_osfPageUp) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfPageUp) | 0);  /* check that XKB_KEY_osfPageUp is an integer */
  if (!_cffi_check_int(*o, n, 268762945U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfPaste(unsigned long long *o)
{
  int n = (XKB_KEY_osfPaste) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfPaste) | 0);  /* check that XKB_KEY_osfPaste is an integer */
  if (!_cffi_check_int(*o, n, 268762884U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfPrevField(unsigned long long *o)
{
  int n = (XKB_KEY_osfPrevField) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfPrevField) | 0);  /* check that XKB_KEY_osfPrevField is an integer */
  if (!_cffi_check_int(*o, n, 268762973U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfPrevMenu(unsigned long long *o)
{
  int n = (XKB_KEY_osfPrevMenu) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfPrevMenu) | 0);  /* check that XKB_KEY_osfPrevMenu is an integer */
  if (!_cffi_check_int(*o, n, 268762971U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfPrimaryPaste(unsigned long long *o)
{
  int n = (XKB_KEY_osfPrimaryPaste) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfPrimaryPaste) | 0);  /* check that XKB_KEY_osfPrimaryPaste is an integer */
  if (!_cffi_check_int(*o, n, 268762930U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfQuickPaste(unsigned long long *o)
{
  int n = (XKB_KEY_osfQuickPaste) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfQuickPaste) | 0);  /* check that XKB_KEY_osfQuickPaste is an integer */
  if (!_cffi_check_int(*o, n, 268762931U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfReselect(unsigned long long *o)
{
  int n = (XKB_KEY_osfReselect) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfReselect) | 0);  /* check that XKB_KEY_osfReselect is an integer */
  if (!_cffi_check_int(*o, n, 268762995U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfRestore(unsigned long long *o)
{
  int n = (XKB_KEY_osfRestore) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfRestore) | 0);  /* check that XKB_KEY_osfRestore is an integer */
  if (!_cffi_check_int(*o, n, 268763000U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfRight(unsigned long long *o)
{
  int n = (XKB_KEY_osfRight) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfRight) | 0);  /* check that XKB_KEY_osfRight is an integer */
  if (!_cffi_check_int(*o, n, 268762963U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfSelect(unsigned long long *o)
{
  int n = (XKB_KEY_osfSelect) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfSelect) | 0);  /* check that XKB_KEY_osfSelect is an integer */
  if (!_cffi_check_int(*o, n, 268762976U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfSelectAll(unsigned long long *o)
{
  int n = (XKB_KEY_osfSelectAll) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfSelectAll) | 0);  /* check that XKB_KEY_osfSelectAll is an integer */
  if (!_cffi_check_int(*o, n, 268762993U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfUndo(unsigned long long *o)
{
  int n = (XKB_KEY_osfUndo) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfUndo) | 0);  /* check that XKB_KEY_osfUndo is an integer */
  if (!_cffi_check_int(*o, n, 268762981U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_osfUp(unsigned long long *o)
{
  int n = (XKB_KEY_osfUp) <= 0;
  *o = (unsigned long long)((XKB_KEY_osfUp) | 0);  /* check that XKB_KEY_osfUp is an integer */
  if (!_cffi_check_int(*o, n, 268762962U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_oslash(unsigned long long *o)
{
  int n = (XKB_KEY_oslash) <= 0;
  *o = (unsigned long long)((XKB_KEY_oslash) | 0);  /* check that XKB_KEY_oslash is an integer */
  if (!_cffi_check_int(*o, n, 248U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_otilde(unsigned long long *o)
{
  int n = (XKB_KEY_otilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_otilde) | 0);  /* check that XKB_KEY_otilde is an integer */
  if (!_cffi_check_int(*o, n, 245U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_overbar(unsigned long long *o)
{
  int n = (XKB_KEY_overbar) <= 0;
  *o = (unsigned long long)((XKB_KEY_overbar) | 0);  /* check that XKB_KEY_overbar is an integer */
  if (!_cffi_check_int(*o, n, 3008U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_overline(unsigned long long *o)
{
  int n = (XKB_KEY_overline) <= 0;
  *o = (unsigned long long)((XKB_KEY_overline) | 0);  /* check that XKB_KEY_overline is an integer */
  if (!_cffi_check_int(*o, n, 1150U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_p(unsigned long long *o)
{
  int n = (XKB_KEY_p) <= 0;
  *o = (unsigned long long)((XKB_KEY_p) | 0);  /* check that XKB_KEY_p is an integer */
  if (!_cffi_check_int(*o, n, 112U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_pabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_pabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_pabovedot) | 0);  /* check that XKB_KEY_pabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16784983U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_paragraph(unsigned long long *o)
{
  int n = (XKB_KEY_paragraph) <= 0;
  *o = (unsigned long long)((XKB_KEY_paragraph) | 0);  /* check that XKB_KEY_paragraph is an integer */
  if (!_cffi_check_int(*o, n, 182U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_parenleft(unsigned long long *o)
{
  int n = (XKB_KEY_parenleft) <= 0;
  *o = (unsigned long long)((XKB_KEY_parenleft) | 0);  /* check that XKB_KEY_parenleft is an integer */
  if (!_cffi_check_int(*o, n, 40U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_parenright(unsigned long long *o)
{
  int n = (XKB_KEY_parenright) <= 0;
  *o = (unsigned long long)((XKB_KEY_parenright) | 0);  /* check that XKB_KEY_parenright is an integer */
  if (!_cffi_check_int(*o, n, 41U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_partdifferential(unsigned long long *o)
{
  int n = (XKB_KEY_partdifferential) <= 0;
  *o = (unsigned long long)((XKB_KEY_partdifferential) | 0);  /* check that XKB_KEY_partdifferential is an integer */
  if (!_cffi_check_int(*o, n, 16785922U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_partialderivative(unsigned long long *o)
{
  int n = (XKB_KEY_partialderivative) <= 0;
  *o = (unsigned long long)((XKB_KEY_partialderivative) | 0);  /* check that XKB_KEY_partialderivative is an integer */
  if (!_cffi_check_int(*o, n, 2287U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_percent(unsigned long long *o)
{
  int n = (XKB_KEY_percent) <= 0;
  *o = (unsigned long long)((XKB_KEY_percent) | 0);  /* check that XKB_KEY_percent is an integer */
  if (!_cffi_check_int(*o, n, 37U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_period(unsigned long long *o)
{
  int n = (XKB_KEY_period) <= 0;
  *o = (unsigned long long)((XKB_KEY_period) | 0);  /* check that XKB_KEY_period is an integer */
  if (!_cffi_check_int(*o, n, 46U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_periodcentered(unsigned long long *o)
{
  int n = (XKB_KEY_periodcentered) <= 0;
  *o = (unsigned long long)((XKB_KEY_periodcentered) | 0);  /* check that XKB_KEY_periodcentered is an integer */
  if (!_cffi_check_int(*o, n, 183U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_permille(unsigned long long *o)
{
  int n = (XKB_KEY_permille) <= 0;
  *o = (unsigned long long)((XKB_KEY_permille) | 0);  /* check that XKB_KEY_permille is an integer */
  if (!_cffi_check_int(*o, n, 2773U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_phonographcopyright(unsigned long long *o)
{
  int n = (XKB_KEY_phonographcopyright) <= 0;
  *o = (unsigned long long)((XKB_KEY_phonographcopyright) | 0);  /* check that XKB_KEY_phonographcopyright is an integer */
  if (!_cffi_check_int(*o, n, 2811U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_plus(unsigned long long *o)
{
  int n = (XKB_KEY_plus) <= 0;
  *o = (unsigned long long)((XKB_KEY_plus) | 0);  /* check that XKB_KEY_plus is an integer */
  if (!_cffi_check_int(*o, n, 43U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_plusminus(unsigned long long *o)
{
  int n = (XKB_KEY_plusminus) <= 0;
  *o = (unsigned long long)((XKB_KEY_plusminus) | 0);  /* check that XKB_KEY_plusminus is an integer */
  if (!_cffi_check_int(*o, n, 177U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_prescription(unsigned long long *o)
{
  int n = (XKB_KEY_prescription) <= 0;
  *o = (unsigned long long)((XKB_KEY_prescription) | 0);  /* check that XKB_KEY_prescription is an integer */
  if (!_cffi_check_int(*o, n, 2772U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_prolongedsound(unsigned long long *o)
{
  int n = (XKB_KEY_prolongedsound) <= 0;
  *o = (unsigned long long)((XKB_KEY_prolongedsound) | 0);  /* check that XKB_KEY_prolongedsound is an integer */
  if (!_cffi_check_int(*o, n, 1200U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_punctspace(unsigned long long *o)
{
  int n = (XKB_KEY_punctspace) <= 0;
  *o = (unsigned long long)((XKB_KEY_punctspace) | 0);  /* check that XKB_KEY_punctspace is an integer */
  if (!_cffi_check_int(*o, n, 2726U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_q(unsigned long long *o)
{
  int n = (XKB_KEY_q) <= 0;
  *o = (unsigned long long)((XKB_KEY_q) | 0);  /* check that XKB_KEY_q is an integer */
  if (!_cffi_check_int(*o, n, 113U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_quad(unsigned long long *o)
{
  int n = (XKB_KEY_quad) <= 0;
  *o = (unsigned long long)((XKB_KEY_quad) | 0);  /* check that XKB_KEY_quad is an integer */
  if (!_cffi_check_int(*o, n, 3020U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_question(unsigned long long *o)
{
  int n = (XKB_KEY_question) <= 0;
  *o = (unsigned long long)((XKB_KEY_question) | 0);  /* check that XKB_KEY_question is an integer */
  if (!_cffi_check_int(*o, n, 63U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_questiondown(unsigned long long *o)
{
  int n = (XKB_KEY_questiondown) <= 0;
  *o = (unsigned long long)((XKB_KEY_questiondown) | 0);  /* check that XKB_KEY_questiondown is an integer */
  if (!_cffi_check_int(*o, n, 191U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_quotedbl(unsigned long long *o)
{
  int n = (XKB_KEY_quotedbl) <= 0;
  *o = (unsigned long long)((XKB_KEY_quotedbl) | 0);  /* check that XKB_KEY_quotedbl is an integer */
  if (!_cffi_check_int(*o, n, 34U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_quoteleft(unsigned long long *o)
{
  int n = (XKB_KEY_quoteleft) <= 0;
  *o = (unsigned long long)((XKB_KEY_quoteleft) | 0);  /* check that XKB_KEY_quoteleft is an integer */
  if (!_cffi_check_int(*o, n, 96U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_quoteright(unsigned long long *o)
{
  int n = (XKB_KEY_quoteright) <= 0;
  *o = (unsigned long long)((XKB_KEY_quoteright) | 0);  /* check that XKB_KEY_quoteright is an integer */
  if (!_cffi_check_int(*o, n, 39U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_r(unsigned long long *o)
{
  int n = (XKB_KEY_r) <= 0;
  *o = (unsigned long long)((XKB_KEY_r) | 0);  /* check that XKB_KEY_r is an integer */
  if (!_cffi_check_int(*o, n, 114U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_racute(unsigned long long *o)
{
  int n = (XKB_KEY_racute) <= 0;
  *o = (unsigned long long)((XKB_KEY_racute) | 0);  /* check that XKB_KEY_racute is an integer */
  if (!_cffi_check_int(*o, n, 480U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_radical(unsigned long long *o)
{
  int n = (XKB_KEY_radical) <= 0;
  *o = (unsigned long long)((XKB_KEY_radical) | 0);  /* check that XKB_KEY_radical is an integer */
  if (!_cffi_check_int(*o, n, 2262U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_rcaron(unsigned long long *o)
{
  int n = (XKB_KEY_rcaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_rcaron) | 0);  /* check that XKB_KEY_rcaron is an integer */
  if (!_cffi_check_int(*o, n, 504U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_rcedilla(unsigned long long *o)
{
  int n = (XKB_KEY_rcedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_rcedilla) | 0);  /* check that XKB_KEY_rcedilla is an integer */
  if (!_cffi_check_int(*o, n, 947U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_registered(unsigned long long *o)
{
  int n = (XKB_KEY_registered) <= 0;
  *o = (unsigned long long)((XKB_KEY_registered) | 0);  /* check that XKB_KEY_registered is an integer */
  if (!_cffi_check_int(*o, n, 174U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_rightanglebracket(unsigned long long *o)
{
  int n = (XKB_KEY_rightanglebracket) <= 0;
  *o = (unsigned long long)((XKB_KEY_rightanglebracket) | 0);  /* check that XKB_KEY_rightanglebracket is an integer */
  if (!_cffi_check_int(*o, n, 2750U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_rightarrow(unsigned long long *o)
{
  int n = (XKB_KEY_rightarrow) <= 0;
  *o = (unsigned long long)((XKB_KEY_rightarrow) | 0);  /* check that XKB_KEY_rightarrow is an integer */
  if (!_cffi_check_int(*o, n, 2301U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_rightcaret(unsigned long long *o)
{
  int n = (XKB_KEY_rightcaret) <= 0;
  *o = (unsigned long long)((XKB_KEY_rightcaret) | 0);  /* check that XKB_KEY_rightcaret is an integer */
  if (!_cffi_check_int(*o, n, 2982U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_rightdoublequotemark(unsigned long long *o)
{
  int n = (XKB_KEY_rightdoublequotemark) <= 0;
  *o = (unsigned long long)((XKB_KEY_rightdoublequotemark) | 0);  /* check that XKB_KEY_rightdoublequotemark is an integer */
  if (!_cffi_check_int(*o, n, 2771U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_rightmiddlecurlybrace(unsigned long long *o)
{
  int n = (XKB_KEY_rightmiddlecurlybrace) <= 0;
  *o = (unsigned long long)((XKB_KEY_rightmiddlecurlybrace) | 0);  /* check that XKB_KEY_rightmiddlecurlybrace is an integer */
  if (!_cffi_check_int(*o, n, 2224U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_rightmiddlesummation(unsigned long long *o)
{
  int n = (XKB_KEY_rightmiddlesummation) <= 0;
  *o = (unsigned long long)((XKB_KEY_rightmiddlesummation) | 0);  /* check that XKB_KEY_rightmiddlesummation is an integer */
  if (!_cffi_check_int(*o, n, 2231U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_rightopentriangle(unsigned long long *o)
{
  int n = (XKB_KEY_rightopentriangle) <= 0;
  *o = (unsigned long long)((XKB_KEY_rightopentriangle) | 0);  /* check that XKB_KEY_rightopentriangle is an integer */
  if (!_cffi_check_int(*o, n, 2765U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_rightpointer(unsigned long long *o)
{
  int n = (XKB_KEY_rightpointer) <= 0;
  *o = (unsigned long long)((XKB_KEY_rightpointer) | 0);  /* check that XKB_KEY_rightpointer is an integer */
  if (!_cffi_check_int(*o, n, 2795U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_rightshoe(unsigned long long *o)
{
  int n = (XKB_KEY_rightshoe) <= 0;
  *o = (unsigned long long)((XKB_KEY_rightshoe) | 0);  /* check that XKB_KEY_rightshoe is an integer */
  if (!_cffi_check_int(*o, n, 3032U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_rightsinglequotemark(unsigned long long *o)
{
  int n = (XKB_KEY_rightsinglequotemark) <= 0;
  *o = (unsigned long long)((XKB_KEY_rightsinglequotemark) | 0);  /* check that XKB_KEY_rightsinglequotemark is an integer */
  if (!_cffi_check_int(*o, n, 2769U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_rightt(unsigned long long *o)
{
  int n = (XKB_KEY_rightt) <= 0;
  *o = (unsigned long long)((XKB_KEY_rightt) | 0);  /* check that XKB_KEY_rightt is an integer */
  if (!_cffi_check_int(*o, n, 2549U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_righttack(unsigned long long *o)
{
  int n = (XKB_KEY_righttack) <= 0;
  *o = (unsigned long long)((XKB_KEY_righttack) | 0);  /* check that XKB_KEY_righttack is an integer */
  if (!_cffi_check_int(*o, n, 3068U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_s(unsigned long long *o)
{
  int n = (XKB_KEY_s) <= 0;
  *o = (unsigned long long)((XKB_KEY_s) | 0);  /* check that XKB_KEY_s is an integer */
  if (!_cffi_check_int(*o, n, 115U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_sabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_sabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_sabovedot) | 0);  /* check that XKB_KEY_sabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16784993U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_sacute(unsigned long long *o)
{
  int n = (XKB_KEY_sacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_sacute) | 0);  /* check that XKB_KEY_sacute is an integer */
  if (!_cffi_check_int(*o, n, 438U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_scaron(unsigned long long *o)
{
  int n = (XKB_KEY_scaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_scaron) | 0);  /* check that XKB_KEY_scaron is an integer */
  if (!_cffi_check_int(*o, n, 441U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_scedilla(unsigned long long *o)
{
  int n = (XKB_KEY_scedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_scedilla) | 0);  /* check that XKB_KEY_scedilla is an integer */
  if (!_cffi_check_int(*o, n, 442U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_schwa(unsigned long long *o)
{
  int n = (XKB_KEY_schwa) <= 0;
  *o = (unsigned long long)((XKB_KEY_schwa) | 0);  /* check that XKB_KEY_schwa is an integer */
  if (!_cffi_check_int(*o, n, 16777817U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_scircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_scircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_scircumflex) | 0);  /* check that XKB_KEY_scircumflex is an integer */
  if (!_cffi_check_int(*o, n, 766U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_script_switch(unsigned long long *o)
{
  int n = (XKB_KEY_script_switch) <= 0;
  *o = (unsigned long long)((XKB_KEY_script_switch) | 0);  /* check that XKB_KEY_script_switch is an integer */
  if (!_cffi_check_int(*o, n, 65406U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_seconds(unsigned long long *o)
{
  int n = (XKB_KEY_seconds) <= 0;
  *o = (unsigned long long)((XKB_KEY_seconds) | 0);  /* check that XKB_KEY_seconds is an integer */
  if (!_cffi_check_int(*o, n, 2775U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_section(unsigned long long *o)
{
  int n = (XKB_KEY_section) <= 0;
  *o = (unsigned long long)((XKB_KEY_section) | 0);  /* check that XKB_KEY_section is an integer */
  if (!_cffi_check_int(*o, n, 167U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_semicolon(unsigned long long *o)
{
  int n = (XKB_KEY_semicolon) <= 0;
  *o = (unsigned long long)((XKB_KEY_semicolon) | 0);  /* check that XKB_KEY_semicolon is an integer */
  if (!_cffi_check_int(*o, n, 59U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_semivoicedsound(unsigned long long *o)
{
  int n = (XKB_KEY_semivoicedsound) <= 0;
  *o = (unsigned long long)((XKB_KEY_semivoicedsound) | 0);  /* check that XKB_KEY_semivoicedsound is an integer */
  if (!_cffi_check_int(*o, n, 1247U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_seveneighths(unsigned long long *o)
{
  int n = (XKB_KEY_seveneighths) <= 0;
  *o = (unsigned long long)((XKB_KEY_seveneighths) | 0);  /* check that XKB_KEY_seveneighths is an integer */
  if (!_cffi_check_int(*o, n, 2758U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_sevensubscript(unsigned long long *o)
{
  int n = (XKB_KEY_sevensubscript) <= 0;
  *o = (unsigned long long)((XKB_KEY_sevensubscript) | 0);  /* check that XKB_KEY_sevensubscript is an integer */
  if (!_cffi_check_int(*o, n, 16785543U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_sevensuperior(unsigned long long *o)
{
  int n = (XKB_KEY_sevensuperior) <= 0;
  *o = (unsigned long long)((XKB_KEY_sevensuperior) | 0);  /* check that XKB_KEY_sevensuperior is an integer */
  if (!_cffi_check_int(*o, n, 16785527U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_signaturemark(unsigned long long *o)
{
  int n = (XKB_KEY_signaturemark) <= 0;
  *o = (unsigned long long)((XKB_KEY_signaturemark) | 0);  /* check that XKB_KEY_signaturemark is an integer */
  if (!_cffi_check_int(*o, n, 2762U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_signifblank(unsigned long long *o)
{
  int n = (XKB_KEY_signifblank) <= 0;
  *o = (unsigned long long)((XKB_KEY_signifblank) | 0);  /* check that XKB_KEY_signifblank is an integer */
  if (!_cffi_check_int(*o, n, 2732U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_similarequal(unsigned long long *o)
{
  int n = (XKB_KEY_similarequal) <= 0;
  *o = (unsigned long long)((XKB_KEY_similarequal) | 0);  /* check that XKB_KEY_similarequal is an integer */
  if (!_cffi_check_int(*o, n, 2249U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_singlelowquotemark(unsigned long long *o)
{
  int n = (XKB_KEY_singlelowquotemark) <= 0;
  *o = (unsigned long long)((XKB_KEY_singlelowquotemark) | 0);  /* check that XKB_KEY_singlelowquotemark is an integer */
  if (!_cffi_check_int(*o, n, 2813U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_sixsubscript(unsigned long long *o)
{
  int n = (XKB_KEY_sixsubscript) <= 0;
  *o = (unsigned long long)((XKB_KEY_sixsubscript) | 0);  /* check that XKB_KEY_sixsubscript is an integer */
  if (!_cffi_check_int(*o, n, 16785542U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_sixsuperior(unsigned long long *o)
{
  int n = (XKB_KEY_sixsuperior) <= 0;
  *o = (unsigned long long)((XKB_KEY_sixsuperior) | 0);  /* check that XKB_KEY_sixsuperior is an integer */
  if (!_cffi_check_int(*o, n, 16785526U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_slash(unsigned long long *o)
{
  int n = (XKB_KEY_slash) <= 0;
  *o = (unsigned long long)((XKB_KEY_slash) | 0);  /* check that XKB_KEY_slash is an integer */
  if (!_cffi_check_int(*o, n, 47U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_soliddiamond(unsigned long long *o)
{
  int n = (XKB_KEY_soliddiamond) <= 0;
  *o = (unsigned long long)((XKB_KEY_soliddiamond) | 0);  /* check that XKB_KEY_soliddiamond is an integer */
  if (!_cffi_check_int(*o, n, 2528U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_space(unsigned long long *o)
{
  int n = (XKB_KEY_space) <= 0;
  *o = (unsigned long long)((XKB_KEY_space) | 0);  /* check that XKB_KEY_space is an integer */
  if (!_cffi_check_int(*o, n, 32U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_squareroot(unsigned long long *o)
{
  int n = (XKB_KEY_squareroot) <= 0;
  *o = (unsigned long long)((XKB_KEY_squareroot) | 0);  /* check that XKB_KEY_squareroot is an integer */
  if (!_cffi_check_int(*o, n, 16785946U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ssharp(unsigned long long *o)
{
  int n = (XKB_KEY_ssharp) <= 0;
  *o = (unsigned long long)((XKB_KEY_ssharp) | 0);  /* check that XKB_KEY_ssharp is an integer */
  if (!_cffi_check_int(*o, n, 223U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_sterling(unsigned long long *o)
{
  int n = (XKB_KEY_sterling) <= 0;
  *o = (unsigned long long)((XKB_KEY_sterling) | 0);  /* check that XKB_KEY_sterling is an integer */
  if (!_cffi_check_int(*o, n, 163U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_stricteq(unsigned long long *o)
{
  int n = (XKB_KEY_stricteq) <= 0;
  *o = (unsigned long long)((XKB_KEY_stricteq) | 0);  /* check that XKB_KEY_stricteq is an integer */
  if (!_cffi_check_int(*o, n, 16786019U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_t(unsigned long long *o)
{
  int n = (XKB_KEY_t) <= 0;
  *o = (unsigned long long)((XKB_KEY_t) | 0);  /* check that XKB_KEY_t is an integer */
  if (!_cffi_check_int(*o, n, 116U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_tabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_tabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_tabovedot) | 0);  /* check that XKB_KEY_tabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16785003U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_tcaron(unsigned long long *o)
{
  int n = (XKB_KEY_tcaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_tcaron) | 0);  /* check that XKB_KEY_tcaron is an integer */
  if (!_cffi_check_int(*o, n, 443U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_tcedilla(unsigned long long *o)
{
  int n = (XKB_KEY_tcedilla) <= 0;
  *o = (unsigned long long)((XKB_KEY_tcedilla) | 0);  /* check that XKB_KEY_tcedilla is an integer */
  if (!_cffi_check_int(*o, n, 510U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_telephone(unsigned long long *o)
{
  int n = (XKB_KEY_telephone) <= 0;
  *o = (unsigned long long)((XKB_KEY_telephone) | 0);  /* check that XKB_KEY_telephone is an integer */
  if (!_cffi_check_int(*o, n, 2809U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_telephonerecorder(unsigned long long *o)
{
  int n = (XKB_KEY_telephonerecorder) <= 0;
  *o = (unsigned long long)((XKB_KEY_telephonerecorder) | 0);  /* check that XKB_KEY_telephonerecorder is an integer */
  if (!_cffi_check_int(*o, n, 2810U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_therefore(unsigned long long *o)
{
  int n = (XKB_KEY_therefore) <= 0;
  *o = (unsigned long long)((XKB_KEY_therefore) | 0);  /* check that XKB_KEY_therefore is an integer */
  if (!_cffi_check_int(*o, n, 2240U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_thinspace(unsigned long long *o)
{
  int n = (XKB_KEY_thinspace) <= 0;
  *o = (unsigned long long)((XKB_KEY_thinspace) | 0);  /* check that XKB_KEY_thinspace is an integer */
  if (!_cffi_check_int(*o, n, 2727U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_thorn(unsigned long long *o)
{
  int n = (XKB_KEY_thorn) <= 0;
  *o = (unsigned long long)((XKB_KEY_thorn) | 0);  /* check that XKB_KEY_thorn is an integer */
  if (!_cffi_check_int(*o, n, 254U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_threeeighths(unsigned long long *o)
{
  int n = (XKB_KEY_threeeighths) <= 0;
  *o = (unsigned long long)((XKB_KEY_threeeighths) | 0);  /* check that XKB_KEY_threeeighths is an integer */
  if (!_cffi_check_int(*o, n, 2756U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_threefifths(unsigned long long *o)
{
  int n = (XKB_KEY_threefifths) <= 0;
  *o = (unsigned long long)((XKB_KEY_threefifths) | 0);  /* check that XKB_KEY_threefifths is an integer */
  if (!_cffi_check_int(*o, n, 2740U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_threequarters(unsigned long long *o)
{
  int n = (XKB_KEY_threequarters) <= 0;
  *o = (unsigned long long)((XKB_KEY_threequarters) | 0);  /* check that XKB_KEY_threequarters is an integer */
  if (!_cffi_check_int(*o, n, 190U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_threesubscript(unsigned long long *o)
{
  int n = (XKB_KEY_threesubscript) <= 0;
  *o = (unsigned long long)((XKB_KEY_threesubscript) | 0);  /* check that XKB_KEY_threesubscript is an integer */
  if (!_cffi_check_int(*o, n, 16785539U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_threesuperior(unsigned long long *o)
{
  int n = (XKB_KEY_threesuperior) <= 0;
  *o = (unsigned long long)((XKB_KEY_threesuperior) | 0);  /* check that XKB_KEY_threesuperior is an integer */
  if (!_cffi_check_int(*o, n, 179U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_tintegral(unsigned long long *o)
{
  int n = (XKB_KEY_tintegral) <= 0;
  *o = (unsigned long long)((XKB_KEY_tintegral) | 0);  /* check that XKB_KEY_tintegral is an integer */
  if (!_cffi_check_int(*o, n, 16785965U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_topintegral(unsigned long long *o)
{
  int n = (XKB_KEY_topintegral) <= 0;
  *o = (unsigned long long)((XKB_KEY_topintegral) | 0);  /* check that XKB_KEY_topintegral is an integer */
  if (!_cffi_check_int(*o, n, 2212U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_topleftparens(unsigned long long *o)
{
  int n = (XKB_KEY_topleftparens) <= 0;
  *o = (unsigned long long)((XKB_KEY_topleftparens) | 0);  /* check that XKB_KEY_topleftparens is an integer */
  if (!_cffi_check_int(*o, n, 2219U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_topleftradical(unsigned long long *o)
{
  int n = (XKB_KEY_topleftradical) <= 0;
  *o = (unsigned long long)((XKB_KEY_topleftradical) | 0);  /* check that XKB_KEY_topleftradical is an integer */
  if (!_cffi_check_int(*o, n, 2210U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_topleftsqbracket(unsigned long long *o)
{
  int n = (XKB_KEY_topleftsqbracket) <= 0;
  *o = (unsigned long long)((XKB_KEY_topleftsqbracket) | 0);  /* check that XKB_KEY_topleftsqbracket is an integer */
  if (!_cffi_check_int(*o, n, 2215U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_topleftsummation(unsigned long long *o)
{
  int n = (XKB_KEY_topleftsummation) <= 0;
  *o = (unsigned long long)((XKB_KEY_topleftsummation) | 0);  /* check that XKB_KEY_topleftsummation is an integer */
  if (!_cffi_check_int(*o, n, 2225U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_toprightparens(unsigned long long *o)
{
  int n = (XKB_KEY_toprightparens) <= 0;
  *o = (unsigned long long)((XKB_KEY_toprightparens) | 0);  /* check that XKB_KEY_toprightparens is an integer */
  if (!_cffi_check_int(*o, n, 2221U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_toprightsqbracket(unsigned long long *o)
{
  int n = (XKB_KEY_toprightsqbracket) <= 0;
  *o = (unsigned long long)((XKB_KEY_toprightsqbracket) | 0);  /* check that XKB_KEY_toprightsqbracket is an integer */
  if (!_cffi_check_int(*o, n, 2217U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_toprightsummation(unsigned long long *o)
{
  int n = (XKB_KEY_toprightsummation) <= 0;
  *o = (unsigned long long)((XKB_KEY_toprightsummation) | 0);  /* check that XKB_KEY_toprightsummation is an integer */
  if (!_cffi_check_int(*o, n, 2229U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_topt(unsigned long long *o)
{
  int n = (XKB_KEY_topt) <= 0;
  *o = (unsigned long long)((XKB_KEY_topt) | 0);  /* check that XKB_KEY_topt is an integer */
  if (!_cffi_check_int(*o, n, 2551U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_topvertsummationconnector(unsigned long long *o)
{
  int n = (XKB_KEY_topvertsummationconnector) <= 0;
  *o = (unsigned long long)((XKB_KEY_topvertsummationconnector) | 0);  /* check that XKB_KEY_topvertsummationconnector is an integer */
  if (!_cffi_check_int(*o, n, 2227U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_trademark(unsigned long long *o)
{
  int n = (XKB_KEY_trademark) <= 0;
  *o = (unsigned long long)((XKB_KEY_trademark) | 0);  /* check that XKB_KEY_trademark is an integer */
  if (!_cffi_check_int(*o, n, 2761U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_trademarkincircle(unsigned long long *o)
{
  int n = (XKB_KEY_trademarkincircle) <= 0;
  *o = (unsigned long long)((XKB_KEY_trademarkincircle) | 0);  /* check that XKB_KEY_trademarkincircle is an integer */
  if (!_cffi_check_int(*o, n, 2763U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_tslash(unsigned long long *o)
{
  int n = (XKB_KEY_tslash) <= 0;
  *o = (unsigned long long)((XKB_KEY_tslash) | 0);  /* check that XKB_KEY_tslash is an integer */
  if (!_cffi_check_int(*o, n, 956U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_twofifths(unsigned long long *o)
{
  int n = (XKB_KEY_twofifths) <= 0;
  *o = (unsigned long long)((XKB_KEY_twofifths) | 0);  /* check that XKB_KEY_twofifths is an integer */
  if (!_cffi_check_int(*o, n, 2739U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_twosubscript(unsigned long long *o)
{
  int n = (XKB_KEY_twosubscript) <= 0;
  *o = (unsigned long long)((XKB_KEY_twosubscript) | 0);  /* check that XKB_KEY_twosubscript is an integer */
  if (!_cffi_check_int(*o, n, 16785538U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_twosuperior(unsigned long long *o)
{
  int n = (XKB_KEY_twosuperior) <= 0;
  *o = (unsigned long long)((XKB_KEY_twosuperior) | 0);  /* check that XKB_KEY_twosuperior is an integer */
  if (!_cffi_check_int(*o, n, 178U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_twothirds(unsigned long long *o)
{
  int n = (XKB_KEY_twothirds) <= 0;
  *o = (unsigned long long)((XKB_KEY_twothirds) | 0);  /* check that XKB_KEY_twothirds is an integer */
  if (!_cffi_check_int(*o, n, 2737U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_u(unsigned long long *o)
{
  int n = (XKB_KEY_u) <= 0;
  *o = (unsigned long long)((XKB_KEY_u) | 0);  /* check that XKB_KEY_u is an integer */
  if (!_cffi_check_int(*o, n, 117U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_uacute(unsigned long long *o)
{
  int n = (XKB_KEY_uacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_uacute) | 0);  /* check that XKB_KEY_uacute is an integer */
  if (!_cffi_check_int(*o, n, 250U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ubelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_ubelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_ubelowdot) | 0);  /* check that XKB_KEY_ubelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785125U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ubreve(unsigned long long *o)
{
  int n = (XKB_KEY_ubreve) <= 0;
  *o = (unsigned long long)((XKB_KEY_ubreve) | 0);  /* check that XKB_KEY_ubreve is an integer */
  if (!_cffi_check_int(*o, n, 765U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ucircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_ucircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_ucircumflex) | 0);  /* check that XKB_KEY_ucircumflex is an integer */
  if (!_cffi_check_int(*o, n, 251U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_udiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_udiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_udiaeresis) | 0);  /* check that XKB_KEY_udiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 252U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_udoubleacute(unsigned long long *o)
{
  int n = (XKB_KEY_udoubleacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_udoubleacute) | 0);  /* check that XKB_KEY_udoubleacute is an integer */
  if (!_cffi_check_int(*o, n, 507U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ugrave(unsigned long long *o)
{
  int n = (XKB_KEY_ugrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_ugrave) | 0);  /* check that XKB_KEY_ugrave is an integer */
  if (!_cffi_check_int(*o, n, 249U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_uhook(unsigned long long *o)
{
  int n = (XKB_KEY_uhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_uhook) | 0);  /* check that XKB_KEY_uhook is an integer */
  if (!_cffi_check_int(*o, n, 16785127U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_uhorn(unsigned long long *o)
{
  int n = (XKB_KEY_uhorn) <= 0;
  *o = (unsigned long long)((XKB_KEY_uhorn) | 0);  /* check that XKB_KEY_uhorn is an integer */
  if (!_cffi_check_int(*o, n, 16777648U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_uhornacute(unsigned long long *o)
{
  int n = (XKB_KEY_uhornacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_uhornacute) | 0);  /* check that XKB_KEY_uhornacute is an integer */
  if (!_cffi_check_int(*o, n, 16785129U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_uhornbelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_uhornbelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_uhornbelowdot) | 0);  /* check that XKB_KEY_uhornbelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785137U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_uhorngrave(unsigned long long *o)
{
  int n = (XKB_KEY_uhorngrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_uhorngrave) | 0);  /* check that XKB_KEY_uhorngrave is an integer */
  if (!_cffi_check_int(*o, n, 16785131U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_uhornhook(unsigned long long *o)
{
  int n = (XKB_KEY_uhornhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_uhornhook) | 0);  /* check that XKB_KEY_uhornhook is an integer */
  if (!_cffi_check_int(*o, n, 16785133U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_uhorntilde(unsigned long long *o)
{
  int n = (XKB_KEY_uhorntilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_uhorntilde) | 0);  /* check that XKB_KEY_uhorntilde is an integer */
  if (!_cffi_check_int(*o, n, 16785135U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_umacron(unsigned long long *o)
{
  int n = (XKB_KEY_umacron) <= 0;
  *o = (unsigned long long)((XKB_KEY_umacron) | 0);  /* check that XKB_KEY_umacron is an integer */
  if (!_cffi_check_int(*o, n, 1022U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_underbar(unsigned long long *o)
{
  int n = (XKB_KEY_underbar) <= 0;
  *o = (unsigned long long)((XKB_KEY_underbar) | 0);  /* check that XKB_KEY_underbar is an integer */
  if (!_cffi_check_int(*o, n, 3014U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_underscore(unsigned long long *o)
{
  int n = (XKB_KEY_underscore) <= 0;
  *o = (unsigned long long)((XKB_KEY_underscore) | 0);  /* check that XKB_KEY_underscore is an integer */
  if (!_cffi_check_int(*o, n, 95U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_union(unsigned long long *o)
{
  int n = (XKB_KEY_union) <= 0;
  *o = (unsigned long long)((XKB_KEY_union) | 0);  /* check that XKB_KEY_union is an integer */
  if (!_cffi_check_int(*o, n, 2269U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_uogonek(unsigned long long *o)
{
  int n = (XKB_KEY_uogonek) <= 0;
  *o = (unsigned long long)((XKB_KEY_uogonek) | 0);  /* check that XKB_KEY_uogonek is an integer */
  if (!_cffi_check_int(*o, n, 1017U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_uparrow(unsigned long long *o)
{
  int n = (XKB_KEY_uparrow) <= 0;
  *o = (unsigned long long)((XKB_KEY_uparrow) | 0);  /* check that XKB_KEY_uparrow is an integer */
  if (!_cffi_check_int(*o, n, 2300U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_upcaret(unsigned long long *o)
{
  int n = (XKB_KEY_upcaret) <= 0;
  *o = (unsigned long long)((XKB_KEY_upcaret) | 0);  /* check that XKB_KEY_upcaret is an integer */
  if (!_cffi_check_int(*o, n, 2985U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_upleftcorner(unsigned long long *o)
{
  int n = (XKB_KEY_upleftcorner) <= 0;
  *o = (unsigned long long)((XKB_KEY_upleftcorner) | 0);  /* check that XKB_KEY_upleftcorner is an integer */
  if (!_cffi_check_int(*o, n, 2540U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_uprightcorner(unsigned long long *o)
{
  int n = (XKB_KEY_uprightcorner) <= 0;
  *o = (unsigned long long)((XKB_KEY_uprightcorner) | 0);  /* check that XKB_KEY_uprightcorner is an integer */
  if (!_cffi_check_int(*o, n, 2539U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_upshoe(unsigned long long *o)
{
  int n = (XKB_KEY_upshoe) <= 0;
  *o = (unsigned long long)((XKB_KEY_upshoe) | 0);  /* check that XKB_KEY_upshoe is an integer */
  if (!_cffi_check_int(*o, n, 3011U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_upstile(unsigned long long *o)
{
  int n = (XKB_KEY_upstile) <= 0;
  *o = (unsigned long long)((XKB_KEY_upstile) | 0);  /* check that XKB_KEY_upstile is an integer */
  if (!_cffi_check_int(*o, n, 3027U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_uptack(unsigned long long *o)
{
  int n = (XKB_KEY_uptack) <= 0;
  *o = (unsigned long long)((XKB_KEY_uptack) | 0);  /* check that XKB_KEY_uptack is an integer */
  if (!_cffi_check_int(*o, n, 3022U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_uring(unsigned long long *o)
{
  int n = (XKB_KEY_uring) <= 0;
  *o = (unsigned long long)((XKB_KEY_uring) | 0);  /* check that XKB_KEY_uring is an integer */
  if (!_cffi_check_int(*o, n, 505U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_utilde(unsigned long long *o)
{
  int n = (XKB_KEY_utilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_utilde) | 0);  /* check that XKB_KEY_utilde is an integer */
  if (!_cffi_check_int(*o, n, 1021U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_v(unsigned long long *o)
{
  int n = (XKB_KEY_v) <= 0;
  *o = (unsigned long long)((XKB_KEY_v) | 0);  /* check that XKB_KEY_v is an integer */
  if (!_cffi_check_int(*o, n, 118U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_variation(unsigned long long *o)
{
  int n = (XKB_KEY_variation) <= 0;
  *o = (unsigned long long)((XKB_KEY_variation) | 0);  /* check that XKB_KEY_variation is an integer */
  if (!_cffi_check_int(*o, n, 2241U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_vertbar(unsigned long long *o)
{
  int n = (XKB_KEY_vertbar) <= 0;
  *o = (unsigned long long)((XKB_KEY_vertbar) | 0);  /* check that XKB_KEY_vertbar is an integer */
  if (!_cffi_check_int(*o, n, 2552U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_vertconnector(unsigned long long *o)
{
  int n = (XKB_KEY_vertconnector) <= 0;
  *o = (unsigned long long)((XKB_KEY_vertconnector) | 0);  /* check that XKB_KEY_vertconnector is an integer */
  if (!_cffi_check_int(*o, n, 2214U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_voicedsound(unsigned long long *o)
{
  int n = (XKB_KEY_voicedsound) <= 0;
  *o = (unsigned long long)((XKB_KEY_voicedsound) | 0);  /* check that XKB_KEY_voicedsound is an integer */
  if (!_cffi_check_int(*o, n, 1246U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_vt(unsigned long long *o)
{
  int n = (XKB_KEY_vt) <= 0;
  *o = (unsigned long long)((XKB_KEY_vt) | 0);  /* check that XKB_KEY_vt is an integer */
  if (!_cffi_check_int(*o, n, 2537U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_w(unsigned long long *o)
{
  int n = (XKB_KEY_w) <= 0;
  *o = (unsigned long long)((XKB_KEY_w) | 0);  /* check that XKB_KEY_w is an integer */
  if (!_cffi_check_int(*o, n, 119U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_wacute(unsigned long long *o)
{
  int n = (XKB_KEY_wacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_wacute) | 0);  /* check that XKB_KEY_wacute is an integer */
  if (!_cffi_check_int(*o, n, 16785027U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_wcircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_wcircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_wcircumflex) | 0);  /* check that XKB_KEY_wcircumflex is an integer */
  if (!_cffi_check_int(*o, n, 16777589U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_wdiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_wdiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_wdiaeresis) | 0);  /* check that XKB_KEY_wdiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 16785029U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_wgrave(unsigned long long *o)
{
  int n = (XKB_KEY_wgrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_wgrave) | 0);  /* check that XKB_KEY_wgrave is an integer */
  if (!_cffi_check_int(*o, n, 16785025U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_x(unsigned long long *o)
{
  int n = (XKB_KEY_x) <= 0;
  *o = (unsigned long long)((XKB_KEY_x) | 0);  /* check that XKB_KEY_x is an integer */
  if (!_cffi_check_int(*o, n, 120U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_xabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_xabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_xabovedot) | 0);  /* check that XKB_KEY_xabovedot is an integer */
  if (!_cffi_check_int(*o, n, 16785035U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_y(unsigned long long *o)
{
  int n = (XKB_KEY_y) <= 0;
  *o = (unsigned long long)((XKB_KEY_y) | 0);  /* check that XKB_KEY_y is an integer */
  if (!_cffi_check_int(*o, n, 121U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_yacute(unsigned long long *o)
{
  int n = (XKB_KEY_yacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_yacute) | 0);  /* check that XKB_KEY_yacute is an integer */
  if (!_cffi_check_int(*o, n, 253U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ybelowdot(unsigned long long *o)
{
  int n = (XKB_KEY_ybelowdot) <= 0;
  *o = (unsigned long long)((XKB_KEY_ybelowdot) | 0);  /* check that XKB_KEY_ybelowdot is an integer */
  if (!_cffi_check_int(*o, n, 16785141U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ycircumflex(unsigned long long *o)
{
  int n = (XKB_KEY_ycircumflex) <= 0;
  *o = (unsigned long long)((XKB_KEY_ycircumflex) | 0);  /* check that XKB_KEY_ycircumflex is an integer */
  if (!_cffi_check_int(*o, n, 16777591U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ydiaeresis(unsigned long long *o)
{
  int n = (XKB_KEY_ydiaeresis) <= 0;
  *o = (unsigned long long)((XKB_KEY_ydiaeresis) | 0);  /* check that XKB_KEY_ydiaeresis is an integer */
  if (!_cffi_check_int(*o, n, 255U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_yen(unsigned long long *o)
{
  int n = (XKB_KEY_yen) <= 0;
  *o = (unsigned long long)((XKB_KEY_yen) | 0);  /* check that XKB_KEY_yen is an integer */
  if (!_cffi_check_int(*o, n, 165U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ygrave(unsigned long long *o)
{
  int n = (XKB_KEY_ygrave) <= 0;
  *o = (unsigned long long)((XKB_KEY_ygrave) | 0);  /* check that XKB_KEY_ygrave is an integer */
  if (!_cffi_check_int(*o, n, 16785139U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_yhook(unsigned long long *o)
{
  int n = (XKB_KEY_yhook) <= 0;
  *o = (unsigned long long)((XKB_KEY_yhook) | 0);  /* check that XKB_KEY_yhook is an integer */
  if (!_cffi_check_int(*o, n, 16785143U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_ytilde(unsigned long long *o)
{
  int n = (XKB_KEY_ytilde) <= 0;
  *o = (unsigned long long)((XKB_KEY_ytilde) | 0);  /* check that XKB_KEY_ytilde is an integer */
  if (!_cffi_check_int(*o, n, 16785145U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_z(unsigned long long *o)
{
  int n = (XKB_KEY_z) <= 0;
  *o = (unsigned long long)((XKB_KEY_z) | 0);  /* check that XKB_KEY_z is an integer */
  if (!_cffi_check_int(*o, n, 122U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_zabovedot(unsigned long long *o)
{
  int n = (XKB_KEY_zabovedot) <= 0;
  *o = (unsigned long long)((XKB_KEY_zabovedot) | 0);  /* check that XKB_KEY_zabovedot is an integer */
  if (!_cffi_check_int(*o, n, 447U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_zacute(unsigned long long *o)
{
  int n = (XKB_KEY_zacute) <= 0;
  *o = (unsigned long long)((XKB_KEY_zacute) | 0);  /* check that XKB_KEY_zacute is an integer */
  if (!_cffi_check_int(*o, n, 444U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_zcaron(unsigned long long *o)
{
  int n = (XKB_KEY_zcaron) <= 0;
  *o = (unsigned long long)((XKB_KEY_zcaron) | 0);  /* check that XKB_KEY_zcaron is an integer */
  if (!_cffi_check_int(*o, n, 446U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_zerosubscript(unsigned long long *o)
{
  int n = (XKB_KEY_zerosubscript) <= 0;
  *o = (unsigned long long)((XKB_KEY_zerosubscript) | 0);  /* check that XKB_KEY_zerosubscript is an integer */
  if (!_cffi_check_int(*o, n, 16785536U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_zerosuperior(unsigned long long *o)
{
  int n = (XKB_KEY_zerosuperior) <= 0;
  *o = (unsigned long long)((XKB_KEY_zerosuperior) | 0);  /* check that XKB_KEY_zerosuperior is an integer */
  if (!_cffi_check_int(*o, n, 16785520U))
    n |= 2;
  return n;
}

static int _cffi_const_XKB_KEY_zstroke(unsigned long long *o)
{
  int n = (XKB_KEY_zstroke) <= 0;
  *o = (unsigned long long)((XKB_KEY_zstroke) | 0);  /* check that XKB_KEY_zstroke is an integer */
  if (!_cffi_check_int(*o, n, 16777654U))
    n |= 2;
  return n;
}

_CFFI_UNUSED_FN
static void _cffi_checkfld_struct_wlc_geometry(struct wlc_geometry *p)
{
  /* only to generate compile-time warnings or errors */
  (void)p;
  { struct wlc_point *tmp = &p->origin; (void)tmp; }
  { struct wlc_size *tmp = &p->size; (void)tmp; }
}
struct _cffi_align_struct_wlc_geometry { char x; struct wlc_geometry y; };

_CFFI_UNUSED_FN
static void _cffi_checkfld_struct_wlc_modifiers(struct wlc_modifiers *p)
{
  /* only to generate compile-time warnings or errors */
  (void)p;
  (void)((p->leds) | 0);  /* check that 'struct wlc_modifiers.leds' is an integer */
  (void)((p->mods) | 0);  /* check that 'struct wlc_modifiers.mods' is an integer */
}
struct _cffi_align_struct_wlc_modifiers { char x; struct wlc_modifiers y; };

_CFFI_UNUSED_FN
static void _cffi_checkfld_struct_wlc_point(struct wlc_point *p)
{
  /* only to generate compile-time warnings or errors */
  (void)p;
  (void)((p->x) | 0);  /* check that 'struct wlc_point.x' is an integer */
  (void)((p->y) | 0);  /* check that 'struct wlc_point.y' is an integer */
}
struct _cffi_align_struct_wlc_point { char x; struct wlc_point y; };

_CFFI_UNUSED_FN
static void _cffi_checkfld_struct_wlc_size(struct wlc_size *p)
{
  /* only to generate compile-time warnings or errors */
  (void)p;
  (void)((p->w) | 0);  /* check that 'struct wlc_size.w' is an integer */
  (void)((p->h) | 0);  /* check that 'struct wlc_size.h' is an integer */
}
struct _cffi_align_struct_wlc_size { char x; struct wlc_size y; };

static const struct _cffi_global_s _cffi_globals[] = {
  { "ABS_BRAKE", (void *)_cffi_const_ABS_BRAKE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_DISTANCE", (void *)_cffi_const_ABS_DISTANCE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_GAS", (void *)_cffi_const_ABS_GAS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_HAT0X", (void *)_cffi_const_ABS_HAT0X, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_HAT0Y", (void *)_cffi_const_ABS_HAT0Y, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_HAT1X", (void *)_cffi_const_ABS_HAT1X, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_HAT1Y", (void *)_cffi_const_ABS_HAT1Y, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_HAT2X", (void *)_cffi_const_ABS_HAT2X, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_HAT2Y", (void *)_cffi_const_ABS_HAT2Y, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_HAT3X", (void *)_cffi_const_ABS_HAT3X, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_HAT3Y", (void *)_cffi_const_ABS_HAT3Y, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MAX", (void *)_cffi_const_ABS_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MISC", (void *)_cffi_const_ABS_MISC, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_BLOB_ID", (void *)_cffi_const_ABS_MT_BLOB_ID, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_DISTANCE", (void *)_cffi_const_ABS_MT_DISTANCE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_ORIENTATION", (void *)_cffi_const_ABS_MT_ORIENTATION, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_POSITION_X", (void *)_cffi_const_ABS_MT_POSITION_X, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_POSITION_Y", (void *)_cffi_const_ABS_MT_POSITION_Y, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_PRESSURE", (void *)_cffi_const_ABS_MT_PRESSURE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_SLOT", (void *)_cffi_const_ABS_MT_SLOT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_TOOL_TYPE", (void *)_cffi_const_ABS_MT_TOOL_TYPE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_TOOL_X", (void *)_cffi_const_ABS_MT_TOOL_X, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_TOOL_Y", (void *)_cffi_const_ABS_MT_TOOL_Y, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_TOUCH_MAJOR", (void *)_cffi_const_ABS_MT_TOUCH_MAJOR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_TOUCH_MINOR", (void *)_cffi_const_ABS_MT_TOUCH_MINOR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_TRACKING_ID", (void *)_cffi_const_ABS_MT_TRACKING_ID, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_WIDTH_MAJOR", (void *)_cffi_const_ABS_MT_WIDTH_MAJOR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_MT_WIDTH_MINOR", (void *)_cffi_const_ABS_MT_WIDTH_MINOR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_PRESSURE", (void *)_cffi_const_ABS_PRESSURE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_RUDDER", (void *)_cffi_const_ABS_RUDDER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_RX", (void *)_cffi_const_ABS_RX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_RY", (void *)_cffi_const_ABS_RY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_RZ", (void *)_cffi_const_ABS_RZ, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_THROTTLE", (void *)_cffi_const_ABS_THROTTLE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_TILT_X", (void *)_cffi_const_ABS_TILT_X, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_TILT_Y", (void *)_cffi_const_ABS_TILT_Y, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_TOOL_WIDTH", (void *)_cffi_const_ABS_TOOL_WIDTH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_VOLUME", (void *)_cffi_const_ABS_VOLUME, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_WHEEL", (void *)_cffi_const_ABS_WHEEL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_X", (void *)_cffi_const_ABS_X, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_Y", (void *)_cffi_const_ABS_Y, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "ABS_Z", (void *)_cffi_const_ABS_Z, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_0", (void *)_cffi_const_BTN_0, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_1", (void *)_cffi_const_BTN_1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_2", (void *)_cffi_const_BTN_2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_3", (void *)_cffi_const_BTN_3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_4", (void *)_cffi_const_BTN_4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_5", (void *)_cffi_const_BTN_5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_6", (void *)_cffi_const_BTN_6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_7", (void *)_cffi_const_BTN_7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_8", (void *)_cffi_const_BTN_8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_9", (void *)_cffi_const_BTN_9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_BACK", (void *)_cffi_const_BTN_BACK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_BASE", (void *)_cffi_const_BTN_BASE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_BASE2", (void *)_cffi_const_BTN_BASE2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_BASE3", (void *)_cffi_const_BTN_BASE3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_BASE4", (void *)_cffi_const_BTN_BASE4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_BASE5", (void *)_cffi_const_BTN_BASE5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_BASE6", (void *)_cffi_const_BTN_BASE6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_C", (void *)_cffi_const_BTN_C, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_DEAD", (void *)_cffi_const_BTN_DEAD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_DIGI", (void *)_cffi_const_BTN_DIGI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_DPAD_DOWN", (void *)_cffi_const_BTN_DPAD_DOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_DPAD_LEFT", (void *)_cffi_const_BTN_DPAD_LEFT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_DPAD_RIGHT", (void *)_cffi_const_BTN_DPAD_RIGHT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_DPAD_UP", (void *)_cffi_const_BTN_DPAD_UP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_EAST", (void *)_cffi_const_BTN_EAST, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_EXTRA", (void *)_cffi_const_BTN_EXTRA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_FORWARD", (void *)_cffi_const_BTN_FORWARD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_GAMEPAD", (void *)_cffi_const_BTN_GAMEPAD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_GEAR_DOWN", (void *)_cffi_const_BTN_GEAR_DOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_GEAR_UP", (void *)_cffi_const_BTN_GEAR_UP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_JOYSTICK", (void *)_cffi_const_BTN_JOYSTICK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_LEFT", (void *)_cffi_const_BTN_LEFT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_MIDDLE", (void *)_cffi_const_BTN_MIDDLE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_MISC", (void *)_cffi_const_BTN_MISC, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_MODE", (void *)_cffi_const_BTN_MODE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_MOUSE", (void *)_cffi_const_BTN_MOUSE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_NORTH", (void *)_cffi_const_BTN_NORTH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_PINKIE", (void *)_cffi_const_BTN_PINKIE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_RIGHT", (void *)_cffi_const_BTN_RIGHT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_SELECT", (void *)_cffi_const_BTN_SELECT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_SIDE", (void *)_cffi_const_BTN_SIDE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_SOUTH", (void *)_cffi_const_BTN_SOUTH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_START", (void *)_cffi_const_BTN_START, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_STYLUS", (void *)_cffi_const_BTN_STYLUS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_STYLUS2", (void *)_cffi_const_BTN_STYLUS2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TASK", (void *)_cffi_const_BTN_TASK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_THUMB", (void *)_cffi_const_BTN_THUMB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_THUMB2", (void *)_cffi_const_BTN_THUMB2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_THUMBL", (void *)_cffi_const_BTN_THUMBL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_THUMBR", (void *)_cffi_const_BTN_THUMBR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TL", (void *)_cffi_const_BTN_TL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TL2", (void *)_cffi_const_BTN_TL2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOOL_AIRBRUSH", (void *)_cffi_const_BTN_TOOL_AIRBRUSH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOOL_BRUSH", (void *)_cffi_const_BTN_TOOL_BRUSH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOOL_DOUBLETAP", (void *)_cffi_const_BTN_TOOL_DOUBLETAP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOOL_FINGER", (void *)_cffi_const_BTN_TOOL_FINGER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOOL_LENS", (void *)_cffi_const_BTN_TOOL_LENS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOOL_MOUSE", (void *)_cffi_const_BTN_TOOL_MOUSE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOOL_PEN", (void *)_cffi_const_BTN_TOOL_PEN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOOL_PENCIL", (void *)_cffi_const_BTN_TOOL_PENCIL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOOL_QUADTAP", (void *)_cffi_const_BTN_TOOL_QUADTAP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOOL_QUINTTAP", (void *)_cffi_const_BTN_TOOL_QUINTTAP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOOL_RUBBER", (void *)_cffi_const_BTN_TOOL_RUBBER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOOL_TRIPLETAP", (void *)_cffi_const_BTN_TOOL_TRIPLETAP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOP", (void *)_cffi_const_BTN_TOP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOP2", (void *)_cffi_const_BTN_TOP2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TOUCH", (void *)_cffi_const_BTN_TOUCH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TR", (void *)_cffi_const_BTN_TR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TR2", (void *)_cffi_const_BTN_TR2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER", (void *)_cffi_const_BTN_TRIGGER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY", (void *)_cffi_const_BTN_TRIGGER_HAPPY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY1", (void *)_cffi_const_BTN_TRIGGER_HAPPY1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY10", (void *)_cffi_const_BTN_TRIGGER_HAPPY10, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY11", (void *)_cffi_const_BTN_TRIGGER_HAPPY11, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY12", (void *)_cffi_const_BTN_TRIGGER_HAPPY12, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY13", (void *)_cffi_const_BTN_TRIGGER_HAPPY13, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY14", (void *)_cffi_const_BTN_TRIGGER_HAPPY14, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY15", (void *)_cffi_const_BTN_TRIGGER_HAPPY15, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY16", (void *)_cffi_const_BTN_TRIGGER_HAPPY16, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY17", (void *)_cffi_const_BTN_TRIGGER_HAPPY17, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY18", (void *)_cffi_const_BTN_TRIGGER_HAPPY18, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY19", (void *)_cffi_const_BTN_TRIGGER_HAPPY19, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY2", (void *)_cffi_const_BTN_TRIGGER_HAPPY2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY20", (void *)_cffi_const_BTN_TRIGGER_HAPPY20, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY21", (void *)_cffi_const_BTN_TRIGGER_HAPPY21, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY22", (void *)_cffi_const_BTN_TRIGGER_HAPPY22, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY23", (void *)_cffi_const_BTN_TRIGGER_HAPPY23, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY24", (void *)_cffi_const_BTN_TRIGGER_HAPPY24, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY25", (void *)_cffi_const_BTN_TRIGGER_HAPPY25, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY26", (void *)_cffi_const_BTN_TRIGGER_HAPPY26, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY27", (void *)_cffi_const_BTN_TRIGGER_HAPPY27, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY28", (void *)_cffi_const_BTN_TRIGGER_HAPPY28, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY29", (void *)_cffi_const_BTN_TRIGGER_HAPPY29, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY3", (void *)_cffi_const_BTN_TRIGGER_HAPPY3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY30", (void *)_cffi_const_BTN_TRIGGER_HAPPY30, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY31", (void *)_cffi_const_BTN_TRIGGER_HAPPY31, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY32", (void *)_cffi_const_BTN_TRIGGER_HAPPY32, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY33", (void *)_cffi_const_BTN_TRIGGER_HAPPY33, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY34", (void *)_cffi_const_BTN_TRIGGER_HAPPY34, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY35", (void *)_cffi_const_BTN_TRIGGER_HAPPY35, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY36", (void *)_cffi_const_BTN_TRIGGER_HAPPY36, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY37", (void *)_cffi_const_BTN_TRIGGER_HAPPY37, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY38", (void *)_cffi_const_BTN_TRIGGER_HAPPY38, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY39", (void *)_cffi_const_BTN_TRIGGER_HAPPY39, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY4", (void *)_cffi_const_BTN_TRIGGER_HAPPY4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY40", (void *)_cffi_const_BTN_TRIGGER_HAPPY40, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY5", (void *)_cffi_const_BTN_TRIGGER_HAPPY5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY6", (void *)_cffi_const_BTN_TRIGGER_HAPPY6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY7", (void *)_cffi_const_BTN_TRIGGER_HAPPY7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY8", (void *)_cffi_const_BTN_TRIGGER_HAPPY8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_TRIGGER_HAPPY9", (void *)_cffi_const_BTN_TRIGGER_HAPPY9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_WEST", (void *)_cffi_const_BTN_WEST, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_WHEEL", (void *)_cffi_const_BTN_WHEEL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "BTN_Z", (void *)_cffi_const_BTN_Z, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "EV_ABS", (void *)_cffi_const_EV_ABS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "EV_FF", (void *)_cffi_const_EV_FF, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "EV_FF_STATUS", (void *)_cffi_const_EV_FF_STATUS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "EV_KEY", (void *)_cffi_const_EV_KEY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "EV_LED", (void *)_cffi_const_EV_LED, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "EV_MAX", (void *)_cffi_const_EV_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "EV_MSC", (void *)_cffi_const_EV_MSC, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "EV_PWR", (void *)_cffi_const_EV_PWR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "EV_REL", (void *)_cffi_const_EV_REL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "EV_REP", (void *)_cffi_const_EV_REP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "EV_SND", (void *)_cffi_const_EV_SND, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "EV_SW", (void *)_cffi_const_EV_SW, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "EV_SYN", (void *)_cffi_const_EV_SYN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "INPUT_PROP_ACCELEROMETER", (void *)_cffi_const_INPUT_PROP_ACCELEROMETER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "INPUT_PROP_BUTTONPAD", (void *)_cffi_const_INPUT_PROP_BUTTONPAD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "INPUT_PROP_DIRECT", (void *)_cffi_const_INPUT_PROP_DIRECT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "INPUT_PROP_MAX", (void *)_cffi_const_INPUT_PROP_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "INPUT_PROP_POINTER", (void *)_cffi_const_INPUT_PROP_POINTER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "INPUT_PROP_POINTING_STICK", (void *)_cffi_const_INPUT_PROP_POINTING_STICK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "INPUT_PROP_SEMI_MT", (void *)_cffi_const_INPUT_PROP_SEMI_MT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "INPUT_PROP_TOPBUTTONPAD", (void *)_cffi_const_INPUT_PROP_TOPBUTTONPAD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_0", (void *)_cffi_const_KEY_0, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_1", (void *)_cffi_const_KEY_1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_102ND", (void *)_cffi_const_KEY_102ND, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_10CHANNELSDOWN", (void *)_cffi_const_KEY_10CHANNELSDOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_10CHANNELSUP", (void *)_cffi_const_KEY_10CHANNELSUP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_2", (void *)_cffi_const_KEY_2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_3", (void *)_cffi_const_KEY_3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_3D_MODE", (void *)_cffi_const_KEY_3D_MODE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_4", (void *)_cffi_const_KEY_4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_5", (void *)_cffi_const_KEY_5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_6", (void *)_cffi_const_KEY_6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_7", (void *)_cffi_const_KEY_7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_8", (void *)_cffi_const_KEY_8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_9", (void *)_cffi_const_KEY_9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_A", (void *)_cffi_const_KEY_A, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_AB", (void *)_cffi_const_KEY_AB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ADDRESSBOOK", (void *)_cffi_const_KEY_ADDRESSBOOK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_AGAIN", (void *)_cffi_const_KEY_AGAIN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ALS_TOGGLE", (void *)_cffi_const_KEY_ALS_TOGGLE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ALTERASE", (void *)_cffi_const_KEY_ALTERASE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ANGLE", (void *)_cffi_const_KEY_ANGLE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_APOSTROPHE", (void *)_cffi_const_KEY_APOSTROPHE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_APPSELECT", (void *)_cffi_const_KEY_APPSELECT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ARCHIVE", (void *)_cffi_const_KEY_ARCHIVE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ATTENDANT_OFF", (void *)_cffi_const_KEY_ATTENDANT_OFF, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ATTENDANT_ON", (void *)_cffi_const_KEY_ATTENDANT_ON, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ATTENDANT_TOGGLE", (void *)_cffi_const_KEY_ATTENDANT_TOGGLE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_AUDIO", (void *)_cffi_const_KEY_AUDIO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_AUDIO_DESC", (void *)_cffi_const_KEY_AUDIO_DESC, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_AUX", (void *)_cffi_const_KEY_AUX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_B", (void *)_cffi_const_KEY_B, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BACK", (void *)_cffi_const_KEY_BACK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BACKSLASH", (void *)_cffi_const_KEY_BACKSLASH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BACKSPACE", (void *)_cffi_const_KEY_BACKSPACE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BASSBOOST", (void *)_cffi_const_KEY_BASSBOOST, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BATTERY", (void *)_cffi_const_KEY_BATTERY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BLUE", (void *)_cffi_const_KEY_BLUE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BLUETOOTH", (void *)_cffi_const_KEY_BLUETOOTH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BOOKMARKS", (void *)_cffi_const_KEY_BOOKMARKS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BREAK", (void *)_cffi_const_KEY_BREAK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRIGHTNESSDOWN", (void *)_cffi_const_KEY_BRIGHTNESSDOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRIGHTNESSUP", (void *)_cffi_const_KEY_BRIGHTNESSUP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRIGHTNESS_AUTO", (void *)_cffi_const_KEY_BRIGHTNESS_AUTO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRIGHTNESS_CYCLE", (void *)_cffi_const_KEY_BRIGHTNESS_CYCLE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRIGHTNESS_MAX", (void *)_cffi_const_KEY_BRIGHTNESS_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRIGHTNESS_MIN", (void *)_cffi_const_KEY_BRIGHTNESS_MIN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRL_DOT1", (void *)_cffi_const_KEY_BRL_DOT1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRL_DOT10", (void *)_cffi_const_KEY_BRL_DOT10, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRL_DOT2", (void *)_cffi_const_KEY_BRL_DOT2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRL_DOT3", (void *)_cffi_const_KEY_BRL_DOT3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRL_DOT4", (void *)_cffi_const_KEY_BRL_DOT4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRL_DOT5", (void *)_cffi_const_KEY_BRL_DOT5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRL_DOT6", (void *)_cffi_const_KEY_BRL_DOT6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRL_DOT7", (void *)_cffi_const_KEY_BRL_DOT7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRL_DOT8", (void *)_cffi_const_KEY_BRL_DOT8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BRL_DOT9", (void *)_cffi_const_KEY_BRL_DOT9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_BUTTONCONFIG", (void *)_cffi_const_KEY_BUTTONCONFIG, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_C", (void *)_cffi_const_KEY_C, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CALC", (void *)_cffi_const_KEY_CALC, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CALENDAR", (void *)_cffi_const_KEY_CALENDAR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CAMERA", (void *)_cffi_const_KEY_CAMERA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CAMERA_DOWN", (void *)_cffi_const_KEY_CAMERA_DOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CAMERA_FOCUS", (void *)_cffi_const_KEY_CAMERA_FOCUS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CAMERA_LEFT", (void *)_cffi_const_KEY_CAMERA_LEFT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CAMERA_RIGHT", (void *)_cffi_const_KEY_CAMERA_RIGHT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CAMERA_UP", (void *)_cffi_const_KEY_CAMERA_UP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CAMERA_ZOOMIN", (void *)_cffi_const_KEY_CAMERA_ZOOMIN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CAMERA_ZOOMOUT", (void *)_cffi_const_KEY_CAMERA_ZOOMOUT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CANCEL", (void *)_cffi_const_KEY_CANCEL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CAPSLOCK", (void *)_cffi_const_KEY_CAPSLOCK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CD", (void *)_cffi_const_KEY_CD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CHANNEL", (void *)_cffi_const_KEY_CHANNEL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CHANNELDOWN", (void *)_cffi_const_KEY_CHANNELDOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CHANNELUP", (void *)_cffi_const_KEY_CHANNELUP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CHAT", (void *)_cffi_const_KEY_CHAT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CLEAR", (void *)_cffi_const_KEY_CLEAR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CLOSE", (void *)_cffi_const_KEY_CLOSE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CLOSECD", (void *)_cffi_const_KEY_CLOSECD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_COFFEE", (void *)_cffi_const_KEY_COFFEE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_COMMA", (void *)_cffi_const_KEY_COMMA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_COMPOSE", (void *)_cffi_const_KEY_COMPOSE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_COMPUTER", (void *)_cffi_const_KEY_COMPUTER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CONFIG", (void *)_cffi_const_KEY_CONFIG, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CONNECT", (void *)_cffi_const_KEY_CONNECT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CONTEXT_MENU", (void *)_cffi_const_KEY_CONTEXT_MENU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CONTROLPANEL", (void *)_cffi_const_KEY_CONTROLPANEL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_COPY", (void *)_cffi_const_KEY_COPY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CUT", (void *)_cffi_const_KEY_CUT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_CYCLEWINDOWS", (void *)_cffi_const_KEY_CYCLEWINDOWS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_D", (void *)_cffi_const_KEY_D, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DASHBOARD", (void *)_cffi_const_KEY_DASHBOARD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DATA", (void *)_cffi_const_KEY_DATA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DATABASE", (void *)_cffi_const_KEY_DATABASE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DELETE", (void *)_cffi_const_KEY_DELETE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DELETEFILE", (void *)_cffi_const_KEY_DELETEFILE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DEL_EOL", (void *)_cffi_const_KEY_DEL_EOL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DEL_EOS", (void *)_cffi_const_KEY_DEL_EOS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DEL_LINE", (void *)_cffi_const_KEY_DEL_LINE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DIGITS", (void *)_cffi_const_KEY_DIGITS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DIRECTORY", (void *)_cffi_const_KEY_DIRECTORY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DISPLAYTOGGLE", (void *)_cffi_const_KEY_DISPLAYTOGGLE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DISPLAY_OFF", (void *)_cffi_const_KEY_DISPLAY_OFF, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DOCUMENTS", (void *)_cffi_const_KEY_DOCUMENTS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DOLLAR", (void *)_cffi_const_KEY_DOLLAR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DOT", (void *)_cffi_const_KEY_DOT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DOWN", (void *)_cffi_const_KEY_DOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_DVD", (void *)_cffi_const_KEY_DVD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_E", (void *)_cffi_const_KEY_E, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_EDIT", (void *)_cffi_const_KEY_EDIT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_EDITOR", (void *)_cffi_const_KEY_EDITOR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_EJECTCD", (void *)_cffi_const_KEY_EJECTCD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_EJECTCLOSECD", (void *)_cffi_const_KEY_EJECTCLOSECD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_EMAIL", (void *)_cffi_const_KEY_EMAIL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_END", (void *)_cffi_const_KEY_END, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ENTER", (void *)_cffi_const_KEY_ENTER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_EPG", (void *)_cffi_const_KEY_EPG, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_EQUAL", (void *)_cffi_const_KEY_EQUAL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ESC", (void *)_cffi_const_KEY_ESC, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_EURO", (void *)_cffi_const_KEY_EURO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_EXIT", (void *)_cffi_const_KEY_EXIT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F", (void *)_cffi_const_KEY_F, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F1", (void *)_cffi_const_KEY_F1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F10", (void *)_cffi_const_KEY_F10, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F11", (void *)_cffi_const_KEY_F11, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F12", (void *)_cffi_const_KEY_F12, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F13", (void *)_cffi_const_KEY_F13, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F14", (void *)_cffi_const_KEY_F14, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F15", (void *)_cffi_const_KEY_F15, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F16", (void *)_cffi_const_KEY_F16, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F17", (void *)_cffi_const_KEY_F17, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F18", (void *)_cffi_const_KEY_F18, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F19", (void *)_cffi_const_KEY_F19, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F2", (void *)_cffi_const_KEY_F2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F20", (void *)_cffi_const_KEY_F20, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F21", (void *)_cffi_const_KEY_F21, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F22", (void *)_cffi_const_KEY_F22, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F23", (void *)_cffi_const_KEY_F23, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F24", (void *)_cffi_const_KEY_F24, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F3", (void *)_cffi_const_KEY_F3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F4", (void *)_cffi_const_KEY_F4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F5", (void *)_cffi_const_KEY_F5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F6", (void *)_cffi_const_KEY_F6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F7", (void *)_cffi_const_KEY_F7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F8", (void *)_cffi_const_KEY_F8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_F9", (void *)_cffi_const_KEY_F9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FASTFORWARD", (void *)_cffi_const_KEY_FASTFORWARD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FASTREVERSE", (void *)_cffi_const_KEY_FASTREVERSE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FAVORITES", (void *)_cffi_const_KEY_FAVORITES, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FILE", (void *)_cffi_const_KEY_FILE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FINANCE", (void *)_cffi_const_KEY_FINANCE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FIND", (void *)_cffi_const_KEY_FIND, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FIRST", (void *)_cffi_const_KEY_FIRST, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN", (void *)_cffi_const_KEY_FN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_1", (void *)_cffi_const_KEY_FN_1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_2", (void *)_cffi_const_KEY_FN_2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_B", (void *)_cffi_const_KEY_FN_B, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_D", (void *)_cffi_const_KEY_FN_D, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_E", (void *)_cffi_const_KEY_FN_E, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_ESC", (void *)_cffi_const_KEY_FN_ESC, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_F", (void *)_cffi_const_KEY_FN_F, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_F1", (void *)_cffi_const_KEY_FN_F1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_F10", (void *)_cffi_const_KEY_FN_F10, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_F11", (void *)_cffi_const_KEY_FN_F11, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_F12", (void *)_cffi_const_KEY_FN_F12, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_F2", (void *)_cffi_const_KEY_FN_F2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_F3", (void *)_cffi_const_KEY_FN_F3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_F4", (void *)_cffi_const_KEY_FN_F4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_F5", (void *)_cffi_const_KEY_FN_F5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_F6", (void *)_cffi_const_KEY_FN_F6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_F7", (void *)_cffi_const_KEY_FN_F7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_F8", (void *)_cffi_const_KEY_FN_F8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_F9", (void *)_cffi_const_KEY_FN_F9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FN_S", (void *)_cffi_const_KEY_FN_S, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FORWARD", (void *)_cffi_const_KEY_FORWARD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FORWARDMAIL", (void *)_cffi_const_KEY_FORWARDMAIL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FRAMEBACK", (void *)_cffi_const_KEY_FRAMEBACK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FRAMEFORWARD", (void *)_cffi_const_KEY_FRAMEFORWARD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_FRONT", (void *)_cffi_const_KEY_FRONT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_G", (void *)_cffi_const_KEY_G, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_GAMES", (void *)_cffi_const_KEY_GAMES, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_GOTO", (void *)_cffi_const_KEY_GOTO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_GRAPHICSEDITOR", (void *)_cffi_const_KEY_GRAPHICSEDITOR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_GRAVE", (void *)_cffi_const_KEY_GRAVE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_GREEN", (void *)_cffi_const_KEY_GREEN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_H", (void *)_cffi_const_KEY_H, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_HANGEUL", (void *)_cffi_const_KEY_HANGEUL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_HANJA", (void *)_cffi_const_KEY_HANJA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_HELP", (void *)_cffi_const_KEY_HELP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_HENKAN", (void *)_cffi_const_KEY_HENKAN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_HIRAGANA", (void *)_cffi_const_KEY_HIRAGANA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_HOME", (void *)_cffi_const_KEY_HOME, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_HOMEPAGE", (void *)_cffi_const_KEY_HOMEPAGE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_HP", (void *)_cffi_const_KEY_HP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_I", (void *)_cffi_const_KEY_I, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_IMAGES", (void *)_cffi_const_KEY_IMAGES, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_INFO", (void *)_cffi_const_KEY_INFO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_INSERT", (void *)_cffi_const_KEY_INSERT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_INS_LINE", (void *)_cffi_const_KEY_INS_LINE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ISO", (void *)_cffi_const_KEY_ISO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_J", (void *)_cffi_const_KEY_J, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_JOURNAL", (void *)_cffi_const_KEY_JOURNAL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_K", (void *)_cffi_const_KEY_K, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KATAKANA", (void *)_cffi_const_KEY_KATAKANA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KATAKANAHIRAGANA", (void *)_cffi_const_KEY_KATAKANAHIRAGANA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KBDILLUMDOWN", (void *)_cffi_const_KEY_KBDILLUMDOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KBDILLUMTOGGLE", (void *)_cffi_const_KEY_KBDILLUMTOGGLE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KBDILLUMUP", (void *)_cffi_const_KEY_KBDILLUMUP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KBDINPUTASSIST_ACCEPT", (void *)_cffi_const_KEY_KBDINPUTASSIST_ACCEPT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KBDINPUTASSIST_CANCEL", (void *)_cffi_const_KEY_KBDINPUTASSIST_CANCEL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KBDINPUTASSIST_NEXT", (void *)_cffi_const_KEY_KBDINPUTASSIST_NEXT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KBDINPUTASSIST_NEXTGROUP", (void *)_cffi_const_KEY_KBDINPUTASSIST_NEXTGROUP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KBDINPUTASSIST_PREV", (void *)_cffi_const_KEY_KBDINPUTASSIST_PREV, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KBDINPUTASSIST_PREVGROUP", (void *)_cffi_const_KEY_KBDINPUTASSIST_PREVGROUP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KEYBOARD", (void *)_cffi_const_KEY_KEYBOARD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KP0", (void *)_cffi_const_KEY_KP0, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KP1", (void *)_cffi_const_KEY_KP1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KP2", (void *)_cffi_const_KEY_KP2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KP3", (void *)_cffi_const_KEY_KP3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KP4", (void *)_cffi_const_KEY_KP4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KP5", (void *)_cffi_const_KEY_KP5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KP6", (void *)_cffi_const_KEY_KP6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KP7", (void *)_cffi_const_KEY_KP7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KP8", (void *)_cffi_const_KEY_KP8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KP9", (void *)_cffi_const_KEY_KP9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KPASTERISK", (void *)_cffi_const_KEY_KPASTERISK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KPCOMMA", (void *)_cffi_const_KEY_KPCOMMA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KPDOT", (void *)_cffi_const_KEY_KPDOT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KPENTER", (void *)_cffi_const_KEY_KPENTER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KPEQUAL", (void *)_cffi_const_KEY_KPEQUAL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KPJPCOMMA", (void *)_cffi_const_KEY_KPJPCOMMA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KPLEFTPAREN", (void *)_cffi_const_KEY_KPLEFTPAREN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KPMINUS", (void *)_cffi_const_KEY_KPMINUS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KPPLUS", (void *)_cffi_const_KEY_KPPLUS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KPPLUSMINUS", (void *)_cffi_const_KEY_KPPLUSMINUS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KPRIGHTPAREN", (void *)_cffi_const_KEY_KPRIGHTPAREN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_KPSLASH", (void *)_cffi_const_KEY_KPSLASH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_L", (void *)_cffi_const_KEY_L, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LANGUAGE", (void *)_cffi_const_KEY_LANGUAGE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LAST", (void *)_cffi_const_KEY_LAST, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LEFT", (void *)_cffi_const_KEY_LEFT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LEFTALT", (void *)_cffi_const_KEY_LEFTALT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LEFTBRACE", (void *)_cffi_const_KEY_LEFTBRACE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LEFTCTRL", (void *)_cffi_const_KEY_LEFTCTRL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LEFTMETA", (void *)_cffi_const_KEY_LEFTMETA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LEFTSHIFT", (void *)_cffi_const_KEY_LEFTSHIFT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LEFT_DOWN", (void *)_cffi_const_KEY_LEFT_DOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LEFT_UP", (void *)_cffi_const_KEY_LEFT_UP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LIGHTS_TOGGLE", (void *)_cffi_const_KEY_LIGHTS_TOGGLE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LINEFEED", (void *)_cffi_const_KEY_LINEFEED, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LIST", (void *)_cffi_const_KEY_LIST, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_LOGOFF", (void *)_cffi_const_KEY_LOGOFF, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_M", (void *)_cffi_const_KEY_M, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MACRO", (void *)_cffi_const_KEY_MACRO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MAIL", (void *)_cffi_const_KEY_MAIL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MAX", (void *)_cffi_const_KEY_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MEDIA", (void *)_cffi_const_KEY_MEDIA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MEDIA_REPEAT", (void *)_cffi_const_KEY_MEDIA_REPEAT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MEDIA_TOP_MENU", (void *)_cffi_const_KEY_MEDIA_TOP_MENU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MEMO", (void *)_cffi_const_KEY_MEMO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MENU", (void *)_cffi_const_KEY_MENU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MESSENGER", (void *)_cffi_const_KEY_MESSENGER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MHP", (void *)_cffi_const_KEY_MHP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MICMUTE", (void *)_cffi_const_KEY_MICMUTE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MINUS", (void *)_cffi_const_KEY_MINUS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MODE", (void *)_cffi_const_KEY_MODE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MOVE", (void *)_cffi_const_KEY_MOVE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MP3", (void *)_cffi_const_KEY_MP3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MSDOS", (void *)_cffi_const_KEY_MSDOS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MUHENKAN", (void *)_cffi_const_KEY_MUHENKAN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_MUTE", (void *)_cffi_const_KEY_MUTE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_N", (void *)_cffi_const_KEY_N, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NEW", (void *)_cffi_const_KEY_NEW, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NEWS", (void *)_cffi_const_KEY_NEWS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NEXT", (void *)_cffi_const_KEY_NEXT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NEXTSONG", (void *)_cffi_const_KEY_NEXTSONG, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NEXT_FAVORITE", (void *)_cffi_const_KEY_NEXT_FAVORITE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_0", (void *)_cffi_const_KEY_NUMERIC_0, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_1", (void *)_cffi_const_KEY_NUMERIC_1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_11", (void *)_cffi_const_KEY_NUMERIC_11, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_12", (void *)_cffi_const_KEY_NUMERIC_12, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_2", (void *)_cffi_const_KEY_NUMERIC_2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_3", (void *)_cffi_const_KEY_NUMERIC_3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_4", (void *)_cffi_const_KEY_NUMERIC_4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_5", (void *)_cffi_const_KEY_NUMERIC_5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_6", (void *)_cffi_const_KEY_NUMERIC_6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_7", (void *)_cffi_const_KEY_NUMERIC_7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_8", (void *)_cffi_const_KEY_NUMERIC_8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_9", (void *)_cffi_const_KEY_NUMERIC_9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_A", (void *)_cffi_const_KEY_NUMERIC_A, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_B", (void *)_cffi_const_KEY_NUMERIC_B, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_C", (void *)_cffi_const_KEY_NUMERIC_C, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_D", (void *)_cffi_const_KEY_NUMERIC_D, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_POUND", (void *)_cffi_const_KEY_NUMERIC_POUND, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMERIC_STAR", (void *)_cffi_const_KEY_NUMERIC_STAR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_NUMLOCK", (void *)_cffi_const_KEY_NUMLOCK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_O", (void *)_cffi_const_KEY_O, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_OK", (void *)_cffi_const_KEY_OK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_OPEN", (void *)_cffi_const_KEY_OPEN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_OPTION", (void *)_cffi_const_KEY_OPTION, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_P", (void *)_cffi_const_KEY_P, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PAGEDOWN", (void *)_cffi_const_KEY_PAGEDOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PAGEUP", (void *)_cffi_const_KEY_PAGEUP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PASTE", (void *)_cffi_const_KEY_PASTE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PAUSE", (void *)_cffi_const_KEY_PAUSE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PAUSECD", (void *)_cffi_const_KEY_PAUSECD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PAUSE_RECORD", (void *)_cffi_const_KEY_PAUSE_RECORD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PC", (void *)_cffi_const_KEY_PC, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PHONE", (void *)_cffi_const_KEY_PHONE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PLAY", (void *)_cffi_const_KEY_PLAY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PLAYCD", (void *)_cffi_const_KEY_PLAYCD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PLAYER", (void *)_cffi_const_KEY_PLAYER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PLAYPAUSE", (void *)_cffi_const_KEY_PLAYPAUSE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_POWER", (void *)_cffi_const_KEY_POWER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_POWER2", (void *)_cffi_const_KEY_POWER2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PRESENTATION", (void *)_cffi_const_KEY_PRESENTATION, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PREVIOUS", (void *)_cffi_const_KEY_PREVIOUS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PREVIOUSSONG", (void *)_cffi_const_KEY_PREVIOUSSONG, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PRINT", (void *)_cffi_const_KEY_PRINT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PROG1", (void *)_cffi_const_KEY_PROG1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PROG2", (void *)_cffi_const_KEY_PROG2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PROG3", (void *)_cffi_const_KEY_PROG3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PROG4", (void *)_cffi_const_KEY_PROG4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PROGRAM", (void *)_cffi_const_KEY_PROGRAM, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PROPS", (void *)_cffi_const_KEY_PROPS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_PVR", (void *)_cffi_const_KEY_PVR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_Q", (void *)_cffi_const_KEY_Q, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_QUESTION", (void *)_cffi_const_KEY_QUESTION, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_R", (void *)_cffi_const_KEY_R, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RADIO", (void *)_cffi_const_KEY_RADIO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RECORD", (void *)_cffi_const_KEY_RECORD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RED", (void *)_cffi_const_KEY_RED, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_REDO", (void *)_cffi_const_KEY_REDO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_REFRESH", (void *)_cffi_const_KEY_REFRESH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_REPLY", (void *)_cffi_const_KEY_REPLY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RESERVED", (void *)_cffi_const_KEY_RESERVED, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RESTART", (void *)_cffi_const_KEY_RESTART, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_REWIND", (void *)_cffi_const_KEY_REWIND, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RFKILL", (void *)_cffi_const_KEY_RFKILL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RIGHT", (void *)_cffi_const_KEY_RIGHT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RIGHTALT", (void *)_cffi_const_KEY_RIGHTALT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RIGHTBRACE", (void *)_cffi_const_KEY_RIGHTBRACE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RIGHTCTRL", (void *)_cffi_const_KEY_RIGHTCTRL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RIGHTMETA", (void *)_cffi_const_KEY_RIGHTMETA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RIGHTSHIFT", (void *)_cffi_const_KEY_RIGHTSHIFT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RIGHT_DOWN", (void *)_cffi_const_KEY_RIGHT_DOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RIGHT_UP", (void *)_cffi_const_KEY_RIGHT_UP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_RO", (void *)_cffi_const_KEY_RO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ROOT_MENU", (void *)_cffi_const_KEY_ROOT_MENU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ROTATE_DISPLAY", (void *)_cffi_const_KEY_ROTATE_DISPLAY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_S", (void *)_cffi_const_KEY_S, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SAT", (void *)_cffi_const_KEY_SAT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SAT2", (void *)_cffi_const_KEY_SAT2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SAVE", (void *)_cffi_const_KEY_SAVE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SCALE", (void *)_cffi_const_KEY_SCALE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SCREEN", (void *)_cffi_const_KEY_SCREEN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SCREENSAVER", (void *)_cffi_const_KEY_SCREENSAVER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SCROLLDOWN", (void *)_cffi_const_KEY_SCROLLDOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SCROLLLOCK", (void *)_cffi_const_KEY_SCROLLLOCK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SCROLLUP", (void *)_cffi_const_KEY_SCROLLUP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SEARCH", (void *)_cffi_const_KEY_SEARCH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SELECT", (void *)_cffi_const_KEY_SELECT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SEMICOLON", (void *)_cffi_const_KEY_SEMICOLON, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SEND", (void *)_cffi_const_KEY_SEND, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SENDFILE", (void *)_cffi_const_KEY_SENDFILE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SETUP", (void *)_cffi_const_KEY_SETUP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SHOP", (void *)_cffi_const_KEY_SHOP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SHUFFLE", (void *)_cffi_const_KEY_SHUFFLE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SLASH", (void *)_cffi_const_KEY_SLASH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SLEEP", (void *)_cffi_const_KEY_SLEEP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SLOW", (void *)_cffi_const_KEY_SLOW, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SLOWREVERSE", (void *)_cffi_const_KEY_SLOWREVERSE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SOUND", (void *)_cffi_const_KEY_SOUND, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SPACE", (void *)_cffi_const_KEY_SPACE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SPELLCHECK", (void *)_cffi_const_KEY_SPELLCHECK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SPORT", (void *)_cffi_const_KEY_SPORT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SPREADSHEET", (void *)_cffi_const_KEY_SPREADSHEET, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_STOP", (void *)_cffi_const_KEY_STOP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_STOPCD", (void *)_cffi_const_KEY_STOPCD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_STOP_RECORD", (void *)_cffi_const_KEY_STOP_RECORD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SUBTITLE", (void *)_cffi_const_KEY_SUBTITLE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SUSPEND", (void *)_cffi_const_KEY_SUSPEND, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SWITCHVIDEOMODE", (void *)_cffi_const_KEY_SWITCHVIDEOMODE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_SYSRQ", (void *)_cffi_const_KEY_SYSRQ, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_T", (void *)_cffi_const_KEY_T, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TAB", (void *)_cffi_const_KEY_TAB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TAPE", (void *)_cffi_const_KEY_TAPE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TASKMANAGER", (void *)_cffi_const_KEY_TASKMANAGER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TEEN", (void *)_cffi_const_KEY_TEEN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TEXT", (void *)_cffi_const_KEY_TEXT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TIME", (void *)_cffi_const_KEY_TIME, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TITLE", (void *)_cffi_const_KEY_TITLE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TOUCHPAD_OFF", (void *)_cffi_const_KEY_TOUCHPAD_OFF, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TOUCHPAD_ON", (void *)_cffi_const_KEY_TOUCHPAD_ON, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TOUCHPAD_TOGGLE", (void *)_cffi_const_KEY_TOUCHPAD_TOGGLE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TUNER", (void *)_cffi_const_KEY_TUNER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TV", (void *)_cffi_const_KEY_TV, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TV2", (void *)_cffi_const_KEY_TV2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_TWEN", (void *)_cffi_const_KEY_TWEN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_U", (void *)_cffi_const_KEY_U, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_UNDO", (void *)_cffi_const_KEY_UNDO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_UNKNOWN", (void *)_cffi_const_KEY_UNKNOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_UNMUTE", (void *)_cffi_const_KEY_UNMUTE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_UP", (void *)_cffi_const_KEY_UP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_UWB", (void *)_cffi_const_KEY_UWB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_V", (void *)_cffi_const_KEY_V, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_VCR", (void *)_cffi_const_KEY_VCR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_VCR2", (void *)_cffi_const_KEY_VCR2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_VENDOR", (void *)_cffi_const_KEY_VENDOR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_VIDEO", (void *)_cffi_const_KEY_VIDEO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_VIDEOPHONE", (void *)_cffi_const_KEY_VIDEOPHONE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_VIDEO_NEXT", (void *)_cffi_const_KEY_VIDEO_NEXT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_VIDEO_PREV", (void *)_cffi_const_KEY_VIDEO_PREV, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_VOD", (void *)_cffi_const_KEY_VOD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_VOICECOMMAND", (void *)_cffi_const_KEY_VOICECOMMAND, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_VOICEMAIL", (void *)_cffi_const_KEY_VOICEMAIL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_VOLUMEDOWN", (void *)_cffi_const_KEY_VOLUMEDOWN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_VOLUMEUP", (void *)_cffi_const_KEY_VOLUMEUP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_W", (void *)_cffi_const_KEY_W, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_WAKEUP", (void *)_cffi_const_KEY_WAKEUP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_WLAN", (void *)_cffi_const_KEY_WLAN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_WORDPROCESSOR", (void *)_cffi_const_KEY_WORDPROCESSOR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_WPS_BUTTON", (void *)_cffi_const_KEY_WPS_BUTTON, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_WWAN", (void *)_cffi_const_KEY_WWAN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_WWW", (void *)_cffi_const_KEY_WWW, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_X", (void *)_cffi_const_KEY_X, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_XFER", (void *)_cffi_const_KEY_XFER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_Y", (void *)_cffi_const_KEY_Y, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_YELLOW", (void *)_cffi_const_KEY_YELLOW, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_YEN", (void *)_cffi_const_KEY_YEN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_Z", (void *)_cffi_const_KEY_Z, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ZENKAKUHANKAKU", (void *)_cffi_const_KEY_ZENKAKUHANKAKU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ZOOM", (void *)_cffi_const_KEY_ZOOM, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ZOOMIN", (void *)_cffi_const_KEY_ZOOMIN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ZOOMOUT", (void *)_cffi_const_KEY_ZOOMOUT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "KEY_ZOOMRESET", (void *)_cffi_const_KEY_ZOOMRESET, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "LED_CAPSL", (void *)_cffi_const_LED_CAPSL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "LED_CHARGING", (void *)_cffi_const_LED_CHARGING, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "LED_COMPOSE", (void *)_cffi_const_LED_COMPOSE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "LED_KANA", (void *)_cffi_const_LED_KANA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "LED_MAIL", (void *)_cffi_const_LED_MAIL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "LED_MAX", (void *)_cffi_const_LED_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "LED_MISC", (void *)_cffi_const_LED_MISC, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "LED_MUTE", (void *)_cffi_const_LED_MUTE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "LED_NUML", (void *)_cffi_const_LED_NUML, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "LED_SCROLLL", (void *)_cffi_const_LED_SCROLLL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "LED_SLEEP", (void *)_cffi_const_LED_SLEEP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "LED_SUSPEND", (void *)_cffi_const_LED_SUSPEND, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "MSC_GESTURE", (void *)_cffi_const_MSC_GESTURE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "MSC_MAX", (void *)_cffi_const_MSC_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "MSC_PULSELED", (void *)_cffi_const_MSC_PULSELED, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "MSC_RAW", (void *)_cffi_const_MSC_RAW, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "MSC_SCAN", (void *)_cffi_const_MSC_SCAN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "MSC_SERIAL", (void *)_cffi_const_MSC_SERIAL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "MSC_TIMESTAMP", (void *)_cffi_const_MSC_TIMESTAMP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REL_DIAL", (void *)_cffi_const_REL_DIAL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REL_HWHEEL", (void *)_cffi_const_REL_HWHEEL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REL_MAX", (void *)_cffi_const_REL_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REL_MISC", (void *)_cffi_const_REL_MISC, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REL_RX", (void *)_cffi_const_REL_RX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REL_RY", (void *)_cffi_const_REL_RY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REL_RZ", (void *)_cffi_const_REL_RZ, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REL_WHEEL", (void *)_cffi_const_REL_WHEEL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REL_X", (void *)_cffi_const_REL_X, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REL_Y", (void *)_cffi_const_REL_Y, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REL_Z", (void *)_cffi_const_REL_Z, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REP_DELAY", (void *)_cffi_const_REP_DELAY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REP_MAX", (void *)_cffi_const_REP_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "REP_PERIOD", (void *)_cffi_const_REP_PERIOD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SND_BELL", (void *)_cffi_const_SND_BELL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SND_CLICK", (void *)_cffi_const_SND_CLICK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SND_MAX", (void *)_cffi_const_SND_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SND_TONE", (void *)_cffi_const_SND_TONE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_CAMERA_LENS_COVER", (void *)_cffi_const_SW_CAMERA_LENS_COVER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_DOCK", (void *)_cffi_const_SW_DOCK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_FRONT_PROXIMITY", (void *)_cffi_const_SW_FRONT_PROXIMITY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_HEADPHONE_INSERT", (void *)_cffi_const_SW_HEADPHONE_INSERT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_JACK_PHYSICAL_INSERT", (void *)_cffi_const_SW_JACK_PHYSICAL_INSERT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_KEYPAD_SLIDE", (void *)_cffi_const_SW_KEYPAD_SLIDE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_LID", (void *)_cffi_const_SW_LID, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_LINEIN_INSERT", (void *)_cffi_const_SW_LINEIN_INSERT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_LINEOUT_INSERT", (void *)_cffi_const_SW_LINEOUT_INSERT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_MAX", (void *)_cffi_const_SW_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_MICROPHONE_INSERT", (void *)_cffi_const_SW_MICROPHONE_INSERT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_MUTE_DEVICE", (void *)_cffi_const_SW_MUTE_DEVICE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_PEN_INSERTED", (void *)_cffi_const_SW_PEN_INSERTED, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_RFKILL_ALL", (void *)_cffi_const_SW_RFKILL_ALL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_ROTATE_LOCK", (void *)_cffi_const_SW_ROTATE_LOCK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_TABLET_MODE", (void *)_cffi_const_SW_TABLET_MODE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SW_VIDEOOUT_INSERT", (void *)_cffi_const_SW_VIDEOOUT_INSERT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SYN_CONFIG", (void *)_cffi_const_SYN_CONFIG, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SYN_DROPPED", (void *)_cffi_const_SYN_DROPPED, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SYN_MAX", (void *)_cffi_const_SYN_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SYN_MT_REPORT", (void *)_cffi_const_SYN_MT_REPORT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "SYN_REPORT", (void *)_cffi_const_SYN_REPORT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "WLC_BACKEND_DRM", (void *)_cffi_const_WLC_BACKEND_DRM, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BACKEND_NONE", (void *)_cffi_const_WLC_BACKEND_NONE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BACKEND_X11", (void *)_cffi_const_WLC_BACKEND_X11, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_ACTIVATED", (void *)_cffi_const_WLC_BIT_ACTIVATED, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_ANCHOR_BOTTOM", (void *)_cffi_const_WLC_BIT_ANCHOR_BOTTOM, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_ANCHOR_LEFT", (void *)_cffi_const_WLC_BIT_ANCHOR_LEFT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_ANCHOR_NONE", (void *)_cffi_const_WLC_BIT_ANCHOR_NONE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_ANCHOR_RIGHT", (void *)_cffi_const_WLC_BIT_ANCHOR_RIGHT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_ANCHOR_TOP", (void *)_cffi_const_WLC_BIT_ANCHOR_TOP, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_X", (void *)_cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_X, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_Y", (void *)_cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_Y, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_CONSTRAINT_ADJUSTMENT_NONE", (void *)_cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_NONE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_X", (void *)_cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_X, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_Y", (void *)_cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_Y, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_X", (void *)_cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_X, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_Y", (void *)_cffi_const_WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_Y, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_FULLSCREEN", (void *)_cffi_const_WLC_BIT_FULLSCREEN, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_GRAVITY_BOTTOM", (void *)_cffi_const_WLC_BIT_GRAVITY_BOTTOM, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_GRAVITY_LEFT", (void *)_cffi_const_WLC_BIT_GRAVITY_LEFT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_GRAVITY_NONE", (void *)_cffi_const_WLC_BIT_GRAVITY_NONE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_GRAVITY_RIGHT", (void *)_cffi_const_WLC_BIT_GRAVITY_RIGHT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_GRAVITY_TOP", (void *)_cffi_const_WLC_BIT_GRAVITY_TOP, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_LED_CAPS", (void *)_cffi_const_WLC_BIT_LED_CAPS, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_LED_NUM", (void *)_cffi_const_WLC_BIT_LED_NUM, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_LED_SCROLL", (void *)_cffi_const_WLC_BIT_LED_SCROLL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_MAXIMIZED", (void *)_cffi_const_WLC_BIT_MAXIMIZED, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_MODAL", (void *)_cffi_const_WLC_BIT_MODAL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_MOD_ALT", (void *)_cffi_const_WLC_BIT_MOD_ALT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_MOD_CAPS", (void *)_cffi_const_WLC_BIT_MOD_CAPS, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_MOD_CTRL", (void *)_cffi_const_WLC_BIT_MOD_CTRL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_MOD_LOGO", (void *)_cffi_const_WLC_BIT_MOD_LOGO, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_MOD_MOD2", (void *)_cffi_const_WLC_BIT_MOD_MOD2, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_MOD_MOD3", (void *)_cffi_const_WLC_BIT_MOD_MOD3, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_MOD_MOD5", (void *)_cffi_const_WLC_BIT_MOD_MOD5, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_MOD_SHIFT", (void *)_cffi_const_WLC_BIT_MOD_SHIFT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_MOVING", (void *)_cffi_const_WLC_BIT_MOVING, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_OVERRIDE_REDIRECT", (void *)_cffi_const_WLC_BIT_OVERRIDE_REDIRECT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_POPUP", (void *)_cffi_const_WLC_BIT_POPUP, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_PROPERTY_APP_ID", (void *)_cffi_const_WLC_BIT_PROPERTY_APP_ID, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_PROPERTY_CLASS", (void *)_cffi_const_WLC_BIT_PROPERTY_CLASS, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_PROPERTY_PID", (void *)_cffi_const_WLC_BIT_PROPERTY_PID, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_PROPERTY_TITLE", (void *)_cffi_const_WLC_BIT_PROPERTY_TITLE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_RESIZING", (void *)_cffi_const_WLC_BIT_RESIZING, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_SPLASH", (void *)_cffi_const_WLC_BIT_SPLASH, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BIT_UNMANAGED", (void *)_cffi_const_WLC_BIT_UNMANAGED, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BUTTON_STATE_PRESSED", (void *)_cffi_const_WLC_BUTTON_STATE_PRESSED, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_BUTTON_STATE_RELEASED", (void *)_cffi_const_WLC_BUTTON_STATE_RELEASED, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_EVENT_ERROR", (void *)_cffi_const_WLC_EVENT_ERROR, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_EVENT_HANGUP", (void *)_cffi_const_WLC_EVENT_HANGUP, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_EVENT_READABLE", (void *)_cffi_const_WLC_EVENT_READABLE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_EVENT_WRITABLE", (void *)_cffi_const_WLC_EVENT_WRITABLE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_KEY_STATE_PRESSED", (void *)_cffi_const_WLC_KEY_STATE_PRESSED, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_KEY_STATE_RELEASED", (void *)_cffi_const_WLC_KEY_STATE_RELEASED, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_LOG_ERROR", (void *)_cffi_const_WLC_LOG_ERROR, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_LOG_INFO", (void *)_cffi_const_WLC_LOG_INFO, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_LOG_WARN", (void *)_cffi_const_WLC_LOG_WARN, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_LOG_WAYLAND", (void *)_cffi_const_WLC_LOG_WAYLAND, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_RESIZE_EDGE_BOTTOM", (void *)_cffi_const_WLC_RESIZE_EDGE_BOTTOM, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_RESIZE_EDGE_BOTTOM_LEFT", (void *)_cffi_const_WLC_RESIZE_EDGE_BOTTOM_LEFT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_RESIZE_EDGE_BOTTOM_RIGHT", (void *)_cffi_const_WLC_RESIZE_EDGE_BOTTOM_RIGHT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_RESIZE_EDGE_LEFT", (void *)_cffi_const_WLC_RESIZE_EDGE_LEFT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_RESIZE_EDGE_NONE", (void *)_cffi_const_WLC_RESIZE_EDGE_NONE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_RESIZE_EDGE_RIGHT", (void *)_cffi_const_WLC_RESIZE_EDGE_RIGHT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_RESIZE_EDGE_TOP", (void *)_cffi_const_WLC_RESIZE_EDGE_TOP, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_RESIZE_EDGE_TOP_LEFT", (void *)_cffi_const_WLC_RESIZE_EDGE_TOP_LEFT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_RESIZE_EDGE_TOP_RIGHT", (void *)_cffi_const_WLC_RESIZE_EDGE_TOP_RIGHT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_SCROLL_AXIS_HORIZONTAL", (void *)_cffi_const_WLC_SCROLL_AXIS_HORIZONTAL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_SCROLL_AXIS_VERTICAL", (void *)_cffi_const_WLC_SCROLL_AXIS_VERTICAL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_TOUCH_CANCEL", (void *)_cffi_const_WLC_TOUCH_CANCEL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_TOUCH_DOWN", (void *)_cffi_const_WLC_TOUCH_DOWN, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_TOUCH_FRAME", (void *)_cffi_const_WLC_TOUCH_FRAME, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_TOUCH_MOTION", (void *)_cffi_const_WLC_TOUCH_MOTION, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "WLC_TOUCH_UP", (void *)_cffi_const_WLC_TOUCH_UP, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "XKB_KEY_0", (void *)_cffi_const_XKB_KEY_0, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_1", (void *)_cffi_const_XKB_KEY_1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_2", (void *)_cffi_const_XKB_KEY_2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3", (void *)_cffi_const_XKB_KEY_3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_AltCursor", (void *)_cffi_const_XKB_KEY_3270_AltCursor, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Attn", (void *)_cffi_const_XKB_KEY_3270_Attn, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_BackTab", (void *)_cffi_const_XKB_KEY_3270_BackTab, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_ChangeScreen", (void *)_cffi_const_XKB_KEY_3270_ChangeScreen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Copy", (void *)_cffi_const_XKB_KEY_3270_Copy, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_CursorBlink", (void *)_cffi_const_XKB_KEY_3270_CursorBlink, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_CursorSelect", (void *)_cffi_const_XKB_KEY_3270_CursorSelect, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_DeleteWord", (void *)_cffi_const_XKB_KEY_3270_DeleteWord, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Duplicate", (void *)_cffi_const_XKB_KEY_3270_Duplicate, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Enter", (void *)_cffi_const_XKB_KEY_3270_Enter, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_EraseEOF", (void *)_cffi_const_XKB_KEY_3270_EraseEOF, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_EraseInput", (void *)_cffi_const_XKB_KEY_3270_EraseInput, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_ExSelect", (void *)_cffi_const_XKB_KEY_3270_ExSelect, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_FieldMark", (void *)_cffi_const_XKB_KEY_3270_FieldMark, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Ident", (void *)_cffi_const_XKB_KEY_3270_Ident, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Jump", (void *)_cffi_const_XKB_KEY_3270_Jump, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_KeyClick", (void *)_cffi_const_XKB_KEY_3270_KeyClick, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Left2", (void *)_cffi_const_XKB_KEY_3270_Left2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_PA1", (void *)_cffi_const_XKB_KEY_3270_PA1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_PA2", (void *)_cffi_const_XKB_KEY_3270_PA2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_PA3", (void *)_cffi_const_XKB_KEY_3270_PA3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Play", (void *)_cffi_const_XKB_KEY_3270_Play, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_PrintScreen", (void *)_cffi_const_XKB_KEY_3270_PrintScreen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Quit", (void *)_cffi_const_XKB_KEY_3270_Quit, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Record", (void *)_cffi_const_XKB_KEY_3270_Record, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Reset", (void *)_cffi_const_XKB_KEY_3270_Reset, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Right2", (void *)_cffi_const_XKB_KEY_3270_Right2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Rule", (void *)_cffi_const_XKB_KEY_3270_Rule, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Setup", (void *)_cffi_const_XKB_KEY_3270_Setup, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_3270_Test", (void *)_cffi_const_XKB_KEY_3270_Test, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_4", (void *)_cffi_const_XKB_KEY_4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_5", (void *)_cffi_const_XKB_KEY_5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_6", (void *)_cffi_const_XKB_KEY_6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_7", (void *)_cffi_const_XKB_KEY_7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_8", (void *)_cffi_const_XKB_KEY_8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_9", (void *)_cffi_const_XKB_KEY_9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_A", (void *)_cffi_const_XKB_KEY_A, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_AE", (void *)_cffi_const_XKB_KEY_AE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Aacute", (void *)_cffi_const_XKB_KEY_Aacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Abelowdot", (void *)_cffi_const_XKB_KEY_Abelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Abreve", (void *)_cffi_const_XKB_KEY_Abreve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Abreveacute", (void *)_cffi_const_XKB_KEY_Abreveacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Abrevebelowdot", (void *)_cffi_const_XKB_KEY_Abrevebelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Abrevegrave", (void *)_cffi_const_XKB_KEY_Abrevegrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Abrevehook", (void *)_cffi_const_XKB_KEY_Abrevehook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Abrevetilde", (void *)_cffi_const_XKB_KEY_Abrevetilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_AccessX_Enable", (void *)_cffi_const_XKB_KEY_AccessX_Enable, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_AccessX_Feedback_Enable", (void *)_cffi_const_XKB_KEY_AccessX_Feedback_Enable, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Acircumflex", (void *)_cffi_const_XKB_KEY_Acircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Acircumflexacute", (void *)_cffi_const_XKB_KEY_Acircumflexacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Acircumflexbelowdot", (void *)_cffi_const_XKB_KEY_Acircumflexbelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Acircumflexgrave", (void *)_cffi_const_XKB_KEY_Acircumflexgrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Acircumflexhook", (void *)_cffi_const_XKB_KEY_Acircumflexhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Acircumflextilde", (void *)_cffi_const_XKB_KEY_Acircumflextilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Adiaeresis", (void *)_cffi_const_XKB_KEY_Adiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Agrave", (void *)_cffi_const_XKB_KEY_Agrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ahook", (void *)_cffi_const_XKB_KEY_Ahook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Alt_L", (void *)_cffi_const_XKB_KEY_Alt_L, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Alt_R", (void *)_cffi_const_XKB_KEY_Alt_R, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Amacron", (void *)_cffi_const_XKB_KEY_Amacron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Aogonek", (void *)_cffi_const_XKB_KEY_Aogonek, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_0", (void *)_cffi_const_XKB_KEY_Arabic_0, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_1", (void *)_cffi_const_XKB_KEY_Arabic_1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_2", (void *)_cffi_const_XKB_KEY_Arabic_2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_3", (void *)_cffi_const_XKB_KEY_Arabic_3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_4", (void *)_cffi_const_XKB_KEY_Arabic_4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_5", (void *)_cffi_const_XKB_KEY_Arabic_5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_6", (void *)_cffi_const_XKB_KEY_Arabic_6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_7", (void *)_cffi_const_XKB_KEY_Arabic_7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_8", (void *)_cffi_const_XKB_KEY_Arabic_8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_9", (void *)_cffi_const_XKB_KEY_Arabic_9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_ain", (void *)_cffi_const_XKB_KEY_Arabic_ain, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_alef", (void *)_cffi_const_XKB_KEY_Arabic_alef, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_alefmaksura", (void *)_cffi_const_XKB_KEY_Arabic_alefmaksura, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_beh", (void *)_cffi_const_XKB_KEY_Arabic_beh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_comma", (void *)_cffi_const_XKB_KEY_Arabic_comma, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_dad", (void *)_cffi_const_XKB_KEY_Arabic_dad, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_dal", (void *)_cffi_const_XKB_KEY_Arabic_dal, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_damma", (void *)_cffi_const_XKB_KEY_Arabic_damma, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_dammatan", (void *)_cffi_const_XKB_KEY_Arabic_dammatan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_ddal", (void *)_cffi_const_XKB_KEY_Arabic_ddal, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_farsi_yeh", (void *)_cffi_const_XKB_KEY_Arabic_farsi_yeh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_fatha", (void *)_cffi_const_XKB_KEY_Arabic_fatha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_fathatan", (void *)_cffi_const_XKB_KEY_Arabic_fathatan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_feh", (void *)_cffi_const_XKB_KEY_Arabic_feh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_fullstop", (void *)_cffi_const_XKB_KEY_Arabic_fullstop, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_gaf", (void *)_cffi_const_XKB_KEY_Arabic_gaf, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_ghain", (void *)_cffi_const_XKB_KEY_Arabic_ghain, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_ha", (void *)_cffi_const_XKB_KEY_Arabic_ha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_hah", (void *)_cffi_const_XKB_KEY_Arabic_hah, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_hamza", (void *)_cffi_const_XKB_KEY_Arabic_hamza, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_hamza_above", (void *)_cffi_const_XKB_KEY_Arabic_hamza_above, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_hamza_below", (void *)_cffi_const_XKB_KEY_Arabic_hamza_below, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_hamzaonalef", (void *)_cffi_const_XKB_KEY_Arabic_hamzaonalef, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_hamzaonwaw", (void *)_cffi_const_XKB_KEY_Arabic_hamzaonwaw, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_hamzaonyeh", (void *)_cffi_const_XKB_KEY_Arabic_hamzaonyeh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_hamzaunderalef", (void *)_cffi_const_XKB_KEY_Arabic_hamzaunderalef, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_heh", (void *)_cffi_const_XKB_KEY_Arabic_heh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_heh_doachashmee", (void *)_cffi_const_XKB_KEY_Arabic_heh_doachashmee, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_heh_goal", (void *)_cffi_const_XKB_KEY_Arabic_heh_goal, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_jeem", (void *)_cffi_const_XKB_KEY_Arabic_jeem, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_jeh", (void *)_cffi_const_XKB_KEY_Arabic_jeh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_kaf", (void *)_cffi_const_XKB_KEY_Arabic_kaf, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_kasra", (void *)_cffi_const_XKB_KEY_Arabic_kasra, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_kasratan", (void *)_cffi_const_XKB_KEY_Arabic_kasratan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_keheh", (void *)_cffi_const_XKB_KEY_Arabic_keheh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_khah", (void *)_cffi_const_XKB_KEY_Arabic_khah, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_lam", (void *)_cffi_const_XKB_KEY_Arabic_lam, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_madda_above", (void *)_cffi_const_XKB_KEY_Arabic_madda_above, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_maddaonalef", (void *)_cffi_const_XKB_KEY_Arabic_maddaonalef, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_meem", (void *)_cffi_const_XKB_KEY_Arabic_meem, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_noon", (void *)_cffi_const_XKB_KEY_Arabic_noon, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_noon_ghunna", (void *)_cffi_const_XKB_KEY_Arabic_noon_ghunna, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_peh", (void *)_cffi_const_XKB_KEY_Arabic_peh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_percent", (void *)_cffi_const_XKB_KEY_Arabic_percent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_qaf", (void *)_cffi_const_XKB_KEY_Arabic_qaf, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_question_mark", (void *)_cffi_const_XKB_KEY_Arabic_question_mark, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_ra", (void *)_cffi_const_XKB_KEY_Arabic_ra, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_rreh", (void *)_cffi_const_XKB_KEY_Arabic_rreh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_sad", (void *)_cffi_const_XKB_KEY_Arabic_sad, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_seen", (void *)_cffi_const_XKB_KEY_Arabic_seen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_semicolon", (void *)_cffi_const_XKB_KEY_Arabic_semicolon, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_shadda", (void *)_cffi_const_XKB_KEY_Arabic_shadda, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_sheen", (void *)_cffi_const_XKB_KEY_Arabic_sheen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_sukun", (void *)_cffi_const_XKB_KEY_Arabic_sukun, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_superscript_alef", (void *)_cffi_const_XKB_KEY_Arabic_superscript_alef, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_switch", (void *)_cffi_const_XKB_KEY_Arabic_switch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_tah", (void *)_cffi_const_XKB_KEY_Arabic_tah, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_tatweel", (void *)_cffi_const_XKB_KEY_Arabic_tatweel, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_tcheh", (void *)_cffi_const_XKB_KEY_Arabic_tcheh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_teh", (void *)_cffi_const_XKB_KEY_Arabic_teh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_tehmarbuta", (void *)_cffi_const_XKB_KEY_Arabic_tehmarbuta, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_thal", (void *)_cffi_const_XKB_KEY_Arabic_thal, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_theh", (void *)_cffi_const_XKB_KEY_Arabic_theh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_tteh", (void *)_cffi_const_XKB_KEY_Arabic_tteh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_veh", (void *)_cffi_const_XKB_KEY_Arabic_veh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_waw", (void *)_cffi_const_XKB_KEY_Arabic_waw, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_yeh", (void *)_cffi_const_XKB_KEY_Arabic_yeh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_yeh_baree", (void *)_cffi_const_XKB_KEY_Arabic_yeh_baree, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_zah", (void *)_cffi_const_XKB_KEY_Arabic_zah, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Arabic_zain", (void *)_cffi_const_XKB_KEY_Arabic_zain, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Aring", (void *)_cffi_const_XKB_KEY_Aring, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_AT", (void *)_cffi_const_XKB_KEY_Armenian_AT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_AYB", (void *)_cffi_const_XKB_KEY_Armenian_AYB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_BEN", (void *)_cffi_const_XKB_KEY_Armenian_BEN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_CHA", (void *)_cffi_const_XKB_KEY_Armenian_CHA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_DA", (void *)_cffi_const_XKB_KEY_Armenian_DA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_DZA", (void *)_cffi_const_XKB_KEY_Armenian_DZA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_E", (void *)_cffi_const_XKB_KEY_Armenian_E, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_FE", (void *)_cffi_const_XKB_KEY_Armenian_FE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_GHAT", (void *)_cffi_const_XKB_KEY_Armenian_GHAT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_GIM", (void *)_cffi_const_XKB_KEY_Armenian_GIM, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_HI", (void *)_cffi_const_XKB_KEY_Armenian_HI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_HO", (void *)_cffi_const_XKB_KEY_Armenian_HO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_INI", (void *)_cffi_const_XKB_KEY_Armenian_INI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_JE", (void *)_cffi_const_XKB_KEY_Armenian_JE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_KE", (void *)_cffi_const_XKB_KEY_Armenian_KE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_KEN", (void *)_cffi_const_XKB_KEY_Armenian_KEN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_KHE", (void *)_cffi_const_XKB_KEY_Armenian_KHE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_LYUN", (void *)_cffi_const_XKB_KEY_Armenian_LYUN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_MEN", (void *)_cffi_const_XKB_KEY_Armenian_MEN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_NU", (void *)_cffi_const_XKB_KEY_Armenian_NU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_O", (void *)_cffi_const_XKB_KEY_Armenian_O, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_PE", (void *)_cffi_const_XKB_KEY_Armenian_PE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_PYUR", (void *)_cffi_const_XKB_KEY_Armenian_PYUR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_RA", (void *)_cffi_const_XKB_KEY_Armenian_RA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_RE", (void *)_cffi_const_XKB_KEY_Armenian_RE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_SE", (void *)_cffi_const_XKB_KEY_Armenian_SE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_SHA", (void *)_cffi_const_XKB_KEY_Armenian_SHA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_TCHE", (void *)_cffi_const_XKB_KEY_Armenian_TCHE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_TO", (void *)_cffi_const_XKB_KEY_Armenian_TO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_TSA", (void *)_cffi_const_XKB_KEY_Armenian_TSA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_TSO", (void *)_cffi_const_XKB_KEY_Armenian_TSO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_TYUN", (void *)_cffi_const_XKB_KEY_Armenian_TYUN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_VEV", (void *)_cffi_const_XKB_KEY_Armenian_VEV, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_VO", (void *)_cffi_const_XKB_KEY_Armenian_VO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_VYUN", (void *)_cffi_const_XKB_KEY_Armenian_VYUN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_YECH", (void *)_cffi_const_XKB_KEY_Armenian_YECH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_ZA", (void *)_cffi_const_XKB_KEY_Armenian_ZA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_ZHE", (void *)_cffi_const_XKB_KEY_Armenian_ZHE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_accent", (void *)_cffi_const_XKB_KEY_Armenian_accent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_amanak", (void *)_cffi_const_XKB_KEY_Armenian_amanak, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_apostrophe", (void *)_cffi_const_XKB_KEY_Armenian_apostrophe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_at", (void *)_cffi_const_XKB_KEY_Armenian_at, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_ayb", (void *)_cffi_const_XKB_KEY_Armenian_ayb, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_ben", (void *)_cffi_const_XKB_KEY_Armenian_ben, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_but", (void *)_cffi_const_XKB_KEY_Armenian_but, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_cha", (void *)_cffi_const_XKB_KEY_Armenian_cha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_da", (void *)_cffi_const_XKB_KEY_Armenian_da, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_dza", (void *)_cffi_const_XKB_KEY_Armenian_dza, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_e", (void *)_cffi_const_XKB_KEY_Armenian_e, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_exclam", (void *)_cffi_const_XKB_KEY_Armenian_exclam, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_fe", (void *)_cffi_const_XKB_KEY_Armenian_fe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_full_stop", (void *)_cffi_const_XKB_KEY_Armenian_full_stop, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_ghat", (void *)_cffi_const_XKB_KEY_Armenian_ghat, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_gim", (void *)_cffi_const_XKB_KEY_Armenian_gim, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_hi", (void *)_cffi_const_XKB_KEY_Armenian_hi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_ho", (void *)_cffi_const_XKB_KEY_Armenian_ho, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_hyphen", (void *)_cffi_const_XKB_KEY_Armenian_hyphen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_ini", (void *)_cffi_const_XKB_KEY_Armenian_ini, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_je", (void *)_cffi_const_XKB_KEY_Armenian_je, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_ke", (void *)_cffi_const_XKB_KEY_Armenian_ke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_ken", (void *)_cffi_const_XKB_KEY_Armenian_ken, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_khe", (void *)_cffi_const_XKB_KEY_Armenian_khe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_ligature_ew", (void *)_cffi_const_XKB_KEY_Armenian_ligature_ew, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_lyun", (void *)_cffi_const_XKB_KEY_Armenian_lyun, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_men", (void *)_cffi_const_XKB_KEY_Armenian_men, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_nu", (void *)_cffi_const_XKB_KEY_Armenian_nu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_o", (void *)_cffi_const_XKB_KEY_Armenian_o, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_paruyk", (void *)_cffi_const_XKB_KEY_Armenian_paruyk, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_pe", (void *)_cffi_const_XKB_KEY_Armenian_pe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_pyur", (void *)_cffi_const_XKB_KEY_Armenian_pyur, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_question", (void *)_cffi_const_XKB_KEY_Armenian_question, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_ra", (void *)_cffi_const_XKB_KEY_Armenian_ra, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_re", (void *)_cffi_const_XKB_KEY_Armenian_re, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_se", (void *)_cffi_const_XKB_KEY_Armenian_se, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_separation_mark", (void *)_cffi_const_XKB_KEY_Armenian_separation_mark, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_sha", (void *)_cffi_const_XKB_KEY_Armenian_sha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_shesht", (void *)_cffi_const_XKB_KEY_Armenian_shesht, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_tche", (void *)_cffi_const_XKB_KEY_Armenian_tche, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_to", (void *)_cffi_const_XKB_KEY_Armenian_to, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_tsa", (void *)_cffi_const_XKB_KEY_Armenian_tsa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_tso", (void *)_cffi_const_XKB_KEY_Armenian_tso, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_tyun", (void *)_cffi_const_XKB_KEY_Armenian_tyun, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_verjaket", (void *)_cffi_const_XKB_KEY_Armenian_verjaket, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_vev", (void *)_cffi_const_XKB_KEY_Armenian_vev, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_vo", (void *)_cffi_const_XKB_KEY_Armenian_vo, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_vyun", (void *)_cffi_const_XKB_KEY_Armenian_vyun, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_yech", (void *)_cffi_const_XKB_KEY_Armenian_yech, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_yentamna", (void *)_cffi_const_XKB_KEY_Armenian_yentamna, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_za", (void *)_cffi_const_XKB_KEY_Armenian_za, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Armenian_zhe", (void *)_cffi_const_XKB_KEY_Armenian_zhe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Atilde", (void *)_cffi_const_XKB_KEY_Atilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_AudibleBell_Enable", (void *)_cffi_const_XKB_KEY_AudibleBell_Enable, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_B", (void *)_cffi_const_XKB_KEY_B, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Babovedot", (void *)_cffi_const_XKB_KEY_Babovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_BackSpace", (void *)_cffi_const_XKB_KEY_BackSpace, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_BackTab", (void *)_cffi_const_XKB_KEY_BackTab, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Begin", (void *)_cffi_const_XKB_KEY_Begin, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_BounceKeys_Enable", (void *)_cffi_const_XKB_KEY_BounceKeys_Enable, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Break", (void *)_cffi_const_XKB_KEY_Break, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Byelorussian_SHORTU", (void *)_cffi_const_XKB_KEY_Byelorussian_SHORTU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Byelorussian_shortu", (void *)_cffi_const_XKB_KEY_Byelorussian_shortu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_C", (void *)_cffi_const_XKB_KEY_C, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_CH", (void *)_cffi_const_XKB_KEY_CH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_C_H", (void *)_cffi_const_XKB_KEY_C_H, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_C_h", (void *)_cffi_const_XKB_KEY_C_h, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cabovedot", (void *)_cffi_const_XKB_KEY_Cabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cacute", (void *)_cffi_const_XKB_KEY_Cacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cancel", (void *)_cffi_const_XKB_KEY_Cancel, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Caps_Lock", (void *)_cffi_const_XKB_KEY_Caps_Lock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ccaron", (void *)_cffi_const_XKB_KEY_Ccaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ccedilla", (void *)_cffi_const_XKB_KEY_Ccedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ccircumflex", (void *)_cffi_const_XKB_KEY_Ccircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ch", (void *)_cffi_const_XKB_KEY_Ch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Clear", (void *)_cffi_const_XKB_KEY_Clear, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ClearLine", (void *)_cffi_const_XKB_KEY_ClearLine, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Codeinput", (void *)_cffi_const_XKB_KEY_Codeinput, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ColonSign", (void *)_cffi_const_XKB_KEY_ColonSign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Control_L", (void *)_cffi_const_XKB_KEY_Control_L, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Control_R", (void *)_cffi_const_XKB_KEY_Control_R, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_CruzeiroSign", (void *)_cffi_const_XKB_KEY_CruzeiroSign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_A", (void *)_cffi_const_XKB_KEY_Cyrillic_A, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_BE", (void *)_cffi_const_XKB_KEY_Cyrillic_BE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_CHE", (void *)_cffi_const_XKB_KEY_Cyrillic_CHE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_CHE_descender", (void *)_cffi_const_XKB_KEY_Cyrillic_CHE_descender, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_CHE_vertstroke", (void *)_cffi_const_XKB_KEY_Cyrillic_CHE_vertstroke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_DE", (void *)_cffi_const_XKB_KEY_Cyrillic_DE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_DZHE", (void *)_cffi_const_XKB_KEY_Cyrillic_DZHE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_E", (void *)_cffi_const_XKB_KEY_Cyrillic_E, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_EF", (void *)_cffi_const_XKB_KEY_Cyrillic_EF, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_EL", (void *)_cffi_const_XKB_KEY_Cyrillic_EL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_EM", (void *)_cffi_const_XKB_KEY_Cyrillic_EM, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_EN", (void *)_cffi_const_XKB_KEY_Cyrillic_EN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_EN_descender", (void *)_cffi_const_XKB_KEY_Cyrillic_EN_descender, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ER", (void *)_cffi_const_XKB_KEY_Cyrillic_ER, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ES", (void *)_cffi_const_XKB_KEY_Cyrillic_ES, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_GHE", (void *)_cffi_const_XKB_KEY_Cyrillic_GHE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_GHE_bar", (void *)_cffi_const_XKB_KEY_Cyrillic_GHE_bar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_HA", (void *)_cffi_const_XKB_KEY_Cyrillic_HA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_HARDSIGN", (void *)_cffi_const_XKB_KEY_Cyrillic_HARDSIGN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_HA_descender", (void *)_cffi_const_XKB_KEY_Cyrillic_HA_descender, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_I", (void *)_cffi_const_XKB_KEY_Cyrillic_I, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_IE", (void *)_cffi_const_XKB_KEY_Cyrillic_IE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_IO", (void *)_cffi_const_XKB_KEY_Cyrillic_IO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_I_macron", (void *)_cffi_const_XKB_KEY_Cyrillic_I_macron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_JE", (void *)_cffi_const_XKB_KEY_Cyrillic_JE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_KA", (void *)_cffi_const_XKB_KEY_Cyrillic_KA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_KA_descender", (void *)_cffi_const_XKB_KEY_Cyrillic_KA_descender, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_KA_vertstroke", (void *)_cffi_const_XKB_KEY_Cyrillic_KA_vertstroke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_LJE", (void *)_cffi_const_XKB_KEY_Cyrillic_LJE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_NJE", (void *)_cffi_const_XKB_KEY_Cyrillic_NJE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_O", (void *)_cffi_const_XKB_KEY_Cyrillic_O, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_O_bar", (void *)_cffi_const_XKB_KEY_Cyrillic_O_bar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_PE", (void *)_cffi_const_XKB_KEY_Cyrillic_PE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_SCHWA", (void *)_cffi_const_XKB_KEY_Cyrillic_SCHWA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_SHA", (void *)_cffi_const_XKB_KEY_Cyrillic_SHA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_SHCHA", (void *)_cffi_const_XKB_KEY_Cyrillic_SHCHA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_SHHA", (void *)_cffi_const_XKB_KEY_Cyrillic_SHHA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_SHORTI", (void *)_cffi_const_XKB_KEY_Cyrillic_SHORTI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_SOFTSIGN", (void *)_cffi_const_XKB_KEY_Cyrillic_SOFTSIGN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_TE", (void *)_cffi_const_XKB_KEY_Cyrillic_TE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_TSE", (void *)_cffi_const_XKB_KEY_Cyrillic_TSE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_U", (void *)_cffi_const_XKB_KEY_Cyrillic_U, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_U_macron", (void *)_cffi_const_XKB_KEY_Cyrillic_U_macron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_U_straight", (void *)_cffi_const_XKB_KEY_Cyrillic_U_straight, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_U_straight_bar", (void *)_cffi_const_XKB_KEY_Cyrillic_U_straight_bar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_VE", (void *)_cffi_const_XKB_KEY_Cyrillic_VE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_YA", (void *)_cffi_const_XKB_KEY_Cyrillic_YA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_YERU", (void *)_cffi_const_XKB_KEY_Cyrillic_YERU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_YU", (void *)_cffi_const_XKB_KEY_Cyrillic_YU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ZE", (void *)_cffi_const_XKB_KEY_Cyrillic_ZE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ZHE", (void *)_cffi_const_XKB_KEY_Cyrillic_ZHE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ZHE_descender", (void *)_cffi_const_XKB_KEY_Cyrillic_ZHE_descender, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_a", (void *)_cffi_const_XKB_KEY_Cyrillic_a, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_be", (void *)_cffi_const_XKB_KEY_Cyrillic_be, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_che", (void *)_cffi_const_XKB_KEY_Cyrillic_che, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_che_descender", (void *)_cffi_const_XKB_KEY_Cyrillic_che_descender, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_che_vertstroke", (void *)_cffi_const_XKB_KEY_Cyrillic_che_vertstroke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_de", (void *)_cffi_const_XKB_KEY_Cyrillic_de, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_dzhe", (void *)_cffi_const_XKB_KEY_Cyrillic_dzhe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_e", (void *)_cffi_const_XKB_KEY_Cyrillic_e, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ef", (void *)_cffi_const_XKB_KEY_Cyrillic_ef, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_el", (void *)_cffi_const_XKB_KEY_Cyrillic_el, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_em", (void *)_cffi_const_XKB_KEY_Cyrillic_em, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_en", (void *)_cffi_const_XKB_KEY_Cyrillic_en, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_en_descender", (void *)_cffi_const_XKB_KEY_Cyrillic_en_descender, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_er", (void *)_cffi_const_XKB_KEY_Cyrillic_er, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_es", (void *)_cffi_const_XKB_KEY_Cyrillic_es, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ghe", (void *)_cffi_const_XKB_KEY_Cyrillic_ghe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ghe_bar", (void *)_cffi_const_XKB_KEY_Cyrillic_ghe_bar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ha", (void *)_cffi_const_XKB_KEY_Cyrillic_ha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ha_descender", (void *)_cffi_const_XKB_KEY_Cyrillic_ha_descender, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_hardsign", (void *)_cffi_const_XKB_KEY_Cyrillic_hardsign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_i", (void *)_cffi_const_XKB_KEY_Cyrillic_i, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_i_macron", (void *)_cffi_const_XKB_KEY_Cyrillic_i_macron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ie", (void *)_cffi_const_XKB_KEY_Cyrillic_ie, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_io", (void *)_cffi_const_XKB_KEY_Cyrillic_io, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_je", (void *)_cffi_const_XKB_KEY_Cyrillic_je, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ka", (void *)_cffi_const_XKB_KEY_Cyrillic_ka, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ka_descender", (void *)_cffi_const_XKB_KEY_Cyrillic_ka_descender, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ka_vertstroke", (void *)_cffi_const_XKB_KEY_Cyrillic_ka_vertstroke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_lje", (void *)_cffi_const_XKB_KEY_Cyrillic_lje, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_nje", (void *)_cffi_const_XKB_KEY_Cyrillic_nje, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_o", (void *)_cffi_const_XKB_KEY_Cyrillic_o, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_o_bar", (void *)_cffi_const_XKB_KEY_Cyrillic_o_bar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_pe", (void *)_cffi_const_XKB_KEY_Cyrillic_pe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_schwa", (void *)_cffi_const_XKB_KEY_Cyrillic_schwa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_sha", (void *)_cffi_const_XKB_KEY_Cyrillic_sha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_shcha", (void *)_cffi_const_XKB_KEY_Cyrillic_shcha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_shha", (void *)_cffi_const_XKB_KEY_Cyrillic_shha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_shorti", (void *)_cffi_const_XKB_KEY_Cyrillic_shorti, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_softsign", (void *)_cffi_const_XKB_KEY_Cyrillic_softsign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_te", (void *)_cffi_const_XKB_KEY_Cyrillic_te, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_tse", (void *)_cffi_const_XKB_KEY_Cyrillic_tse, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_u", (void *)_cffi_const_XKB_KEY_Cyrillic_u, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_u_macron", (void *)_cffi_const_XKB_KEY_Cyrillic_u_macron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_u_straight", (void *)_cffi_const_XKB_KEY_Cyrillic_u_straight, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_u_straight_bar", (void *)_cffi_const_XKB_KEY_Cyrillic_u_straight_bar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ve", (void *)_cffi_const_XKB_KEY_Cyrillic_ve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ya", (void *)_cffi_const_XKB_KEY_Cyrillic_ya, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_yeru", (void *)_cffi_const_XKB_KEY_Cyrillic_yeru, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_yu", (void *)_cffi_const_XKB_KEY_Cyrillic_yu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_ze", (void *)_cffi_const_XKB_KEY_Cyrillic_ze, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_zhe", (void *)_cffi_const_XKB_KEY_Cyrillic_zhe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Cyrillic_zhe_descender", (void *)_cffi_const_XKB_KEY_Cyrillic_zhe_descender, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_D", (void *)_cffi_const_XKB_KEY_D, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_DRemove", (void *)_cffi_const_XKB_KEY_DRemove, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Dabovedot", (void *)_cffi_const_XKB_KEY_Dabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Dacute_accent", (void *)_cffi_const_XKB_KEY_Dacute_accent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Dcaron", (void *)_cffi_const_XKB_KEY_Dcaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Dcedilla_accent", (void *)_cffi_const_XKB_KEY_Dcedilla_accent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Dcircumflex_accent", (void *)_cffi_const_XKB_KEY_Dcircumflex_accent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ddiaeresis", (void *)_cffi_const_XKB_KEY_Ddiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Delete", (void *)_cffi_const_XKB_KEY_Delete, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_DeleteChar", (void *)_cffi_const_XKB_KEY_DeleteChar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_DeleteLine", (void *)_cffi_const_XKB_KEY_DeleteLine, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Dgrave_accent", (void *)_cffi_const_XKB_KEY_Dgrave_accent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_DongSign", (void *)_cffi_const_XKB_KEY_DongSign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Down", (void *)_cffi_const_XKB_KEY_Down, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Dring_accent", (void *)_cffi_const_XKB_KEY_Dring_accent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Dstroke", (void *)_cffi_const_XKB_KEY_Dstroke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Dtilde", (void *)_cffi_const_XKB_KEY_Dtilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_E", (void *)_cffi_const_XKB_KEY_E, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ENG", (void *)_cffi_const_XKB_KEY_ENG, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ETH", (void *)_cffi_const_XKB_KEY_ETH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_EZH", (void *)_cffi_const_XKB_KEY_EZH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Eabovedot", (void *)_cffi_const_XKB_KEY_Eabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Eacute", (void *)_cffi_const_XKB_KEY_Eacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ebelowdot", (void *)_cffi_const_XKB_KEY_Ebelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ecaron", (void *)_cffi_const_XKB_KEY_Ecaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ecircumflex", (void *)_cffi_const_XKB_KEY_Ecircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ecircumflexacute", (void *)_cffi_const_XKB_KEY_Ecircumflexacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ecircumflexbelowdot", (void *)_cffi_const_XKB_KEY_Ecircumflexbelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ecircumflexgrave", (void *)_cffi_const_XKB_KEY_Ecircumflexgrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ecircumflexhook", (void *)_cffi_const_XKB_KEY_Ecircumflexhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ecircumflextilde", (void *)_cffi_const_XKB_KEY_Ecircumflextilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_EcuSign", (void *)_cffi_const_XKB_KEY_EcuSign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ediaeresis", (void *)_cffi_const_XKB_KEY_Ediaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Egrave", (void *)_cffi_const_XKB_KEY_Egrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ehook", (void *)_cffi_const_XKB_KEY_Ehook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Eisu_Shift", (void *)_cffi_const_XKB_KEY_Eisu_Shift, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Eisu_toggle", (void *)_cffi_const_XKB_KEY_Eisu_toggle, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Emacron", (void *)_cffi_const_XKB_KEY_Emacron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_End", (void *)_cffi_const_XKB_KEY_End, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Eogonek", (void *)_cffi_const_XKB_KEY_Eogonek, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Escape", (void *)_cffi_const_XKB_KEY_Escape, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Eth", (void *)_cffi_const_XKB_KEY_Eth, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Etilde", (void *)_cffi_const_XKB_KEY_Etilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_EuroSign", (void *)_cffi_const_XKB_KEY_EuroSign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Execute", (void *)_cffi_const_XKB_KEY_Execute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ext16bit_L", (void *)_cffi_const_XKB_KEY_Ext16bit_L, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ext16bit_R", (void *)_cffi_const_XKB_KEY_Ext16bit_R, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F", (void *)_cffi_const_XKB_KEY_F, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F1", (void *)_cffi_const_XKB_KEY_F1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F10", (void *)_cffi_const_XKB_KEY_F10, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F11", (void *)_cffi_const_XKB_KEY_F11, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F12", (void *)_cffi_const_XKB_KEY_F12, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F13", (void *)_cffi_const_XKB_KEY_F13, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F14", (void *)_cffi_const_XKB_KEY_F14, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F15", (void *)_cffi_const_XKB_KEY_F15, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F16", (void *)_cffi_const_XKB_KEY_F16, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F17", (void *)_cffi_const_XKB_KEY_F17, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F18", (void *)_cffi_const_XKB_KEY_F18, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F19", (void *)_cffi_const_XKB_KEY_F19, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F2", (void *)_cffi_const_XKB_KEY_F2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F20", (void *)_cffi_const_XKB_KEY_F20, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F21", (void *)_cffi_const_XKB_KEY_F21, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F22", (void *)_cffi_const_XKB_KEY_F22, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F23", (void *)_cffi_const_XKB_KEY_F23, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F24", (void *)_cffi_const_XKB_KEY_F24, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F25", (void *)_cffi_const_XKB_KEY_F25, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F26", (void *)_cffi_const_XKB_KEY_F26, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F27", (void *)_cffi_const_XKB_KEY_F27, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F28", (void *)_cffi_const_XKB_KEY_F28, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F29", (void *)_cffi_const_XKB_KEY_F29, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F3", (void *)_cffi_const_XKB_KEY_F3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F30", (void *)_cffi_const_XKB_KEY_F30, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F31", (void *)_cffi_const_XKB_KEY_F31, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F32", (void *)_cffi_const_XKB_KEY_F32, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F33", (void *)_cffi_const_XKB_KEY_F33, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F34", (void *)_cffi_const_XKB_KEY_F34, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F35", (void *)_cffi_const_XKB_KEY_F35, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F4", (void *)_cffi_const_XKB_KEY_F4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F5", (void *)_cffi_const_XKB_KEY_F5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F6", (void *)_cffi_const_XKB_KEY_F6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F7", (void *)_cffi_const_XKB_KEY_F7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F8", (void *)_cffi_const_XKB_KEY_F8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_F9", (void *)_cffi_const_XKB_KEY_F9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_FFrancSign", (void *)_cffi_const_XKB_KEY_FFrancSign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Fabovedot", (void *)_cffi_const_XKB_KEY_Fabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Farsi_0", (void *)_cffi_const_XKB_KEY_Farsi_0, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Farsi_1", (void *)_cffi_const_XKB_KEY_Farsi_1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Farsi_2", (void *)_cffi_const_XKB_KEY_Farsi_2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Farsi_3", (void *)_cffi_const_XKB_KEY_Farsi_3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Farsi_4", (void *)_cffi_const_XKB_KEY_Farsi_4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Farsi_5", (void *)_cffi_const_XKB_KEY_Farsi_5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Farsi_6", (void *)_cffi_const_XKB_KEY_Farsi_6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Farsi_7", (void *)_cffi_const_XKB_KEY_Farsi_7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Farsi_8", (void *)_cffi_const_XKB_KEY_Farsi_8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Farsi_9", (void *)_cffi_const_XKB_KEY_Farsi_9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Farsi_yeh", (void *)_cffi_const_XKB_KEY_Farsi_yeh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Find", (void *)_cffi_const_XKB_KEY_Find, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_First_Virtual_Screen", (void *)_cffi_const_XKB_KEY_First_Virtual_Screen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_G", (void *)_cffi_const_XKB_KEY_G, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Gabovedot", (void *)_cffi_const_XKB_KEY_Gabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Gbreve", (void *)_cffi_const_XKB_KEY_Gbreve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Gcaron", (void *)_cffi_const_XKB_KEY_Gcaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Gcedilla", (void *)_cffi_const_XKB_KEY_Gcedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Gcircumflex", (void *)_cffi_const_XKB_KEY_Gcircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_an", (void *)_cffi_const_XKB_KEY_Georgian_an, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_ban", (void *)_cffi_const_XKB_KEY_Georgian_ban, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_can", (void *)_cffi_const_XKB_KEY_Georgian_can, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_char", (void *)_cffi_const_XKB_KEY_Georgian_char, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_chin", (void *)_cffi_const_XKB_KEY_Georgian_chin, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_cil", (void *)_cffi_const_XKB_KEY_Georgian_cil, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_don", (void *)_cffi_const_XKB_KEY_Georgian_don, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_en", (void *)_cffi_const_XKB_KEY_Georgian_en, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_fi", (void *)_cffi_const_XKB_KEY_Georgian_fi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_gan", (void *)_cffi_const_XKB_KEY_Georgian_gan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_ghan", (void *)_cffi_const_XKB_KEY_Georgian_ghan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_hae", (void *)_cffi_const_XKB_KEY_Georgian_hae, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_har", (void *)_cffi_const_XKB_KEY_Georgian_har, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_he", (void *)_cffi_const_XKB_KEY_Georgian_he, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_hie", (void *)_cffi_const_XKB_KEY_Georgian_hie, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_hoe", (void *)_cffi_const_XKB_KEY_Georgian_hoe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_in", (void *)_cffi_const_XKB_KEY_Georgian_in, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_jhan", (void *)_cffi_const_XKB_KEY_Georgian_jhan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_jil", (void *)_cffi_const_XKB_KEY_Georgian_jil, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_kan", (void *)_cffi_const_XKB_KEY_Georgian_kan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_khar", (void *)_cffi_const_XKB_KEY_Georgian_khar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_las", (void *)_cffi_const_XKB_KEY_Georgian_las, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_man", (void *)_cffi_const_XKB_KEY_Georgian_man, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_nar", (void *)_cffi_const_XKB_KEY_Georgian_nar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_on", (void *)_cffi_const_XKB_KEY_Georgian_on, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_par", (void *)_cffi_const_XKB_KEY_Georgian_par, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_phar", (void *)_cffi_const_XKB_KEY_Georgian_phar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_qar", (void *)_cffi_const_XKB_KEY_Georgian_qar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_rae", (void *)_cffi_const_XKB_KEY_Georgian_rae, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_san", (void *)_cffi_const_XKB_KEY_Georgian_san, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_shin", (void *)_cffi_const_XKB_KEY_Georgian_shin, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_tan", (void *)_cffi_const_XKB_KEY_Georgian_tan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_tar", (void *)_cffi_const_XKB_KEY_Georgian_tar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_un", (void *)_cffi_const_XKB_KEY_Georgian_un, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_vin", (void *)_cffi_const_XKB_KEY_Georgian_vin, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_we", (void *)_cffi_const_XKB_KEY_Georgian_we, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_xan", (void *)_cffi_const_XKB_KEY_Georgian_xan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_zen", (void *)_cffi_const_XKB_KEY_Georgian_zen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Georgian_zhar", (void *)_cffi_const_XKB_KEY_Georgian_zhar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_ALPHA", (void *)_cffi_const_XKB_KEY_Greek_ALPHA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_ALPHAaccent", (void *)_cffi_const_XKB_KEY_Greek_ALPHAaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_BETA", (void *)_cffi_const_XKB_KEY_Greek_BETA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_CHI", (void *)_cffi_const_XKB_KEY_Greek_CHI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_DELTA", (void *)_cffi_const_XKB_KEY_Greek_DELTA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_EPSILON", (void *)_cffi_const_XKB_KEY_Greek_EPSILON, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_EPSILONaccent", (void *)_cffi_const_XKB_KEY_Greek_EPSILONaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_ETA", (void *)_cffi_const_XKB_KEY_Greek_ETA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_ETAaccent", (void *)_cffi_const_XKB_KEY_Greek_ETAaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_GAMMA", (void *)_cffi_const_XKB_KEY_Greek_GAMMA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_IOTA", (void *)_cffi_const_XKB_KEY_Greek_IOTA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_IOTAaccent", (void *)_cffi_const_XKB_KEY_Greek_IOTAaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_IOTAdiaeresis", (void *)_cffi_const_XKB_KEY_Greek_IOTAdiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_IOTAdieresis", (void *)_cffi_const_XKB_KEY_Greek_IOTAdieresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_KAPPA", (void *)_cffi_const_XKB_KEY_Greek_KAPPA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_LAMBDA", (void *)_cffi_const_XKB_KEY_Greek_LAMBDA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_LAMDA", (void *)_cffi_const_XKB_KEY_Greek_LAMDA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_MU", (void *)_cffi_const_XKB_KEY_Greek_MU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_NU", (void *)_cffi_const_XKB_KEY_Greek_NU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_OMEGA", (void *)_cffi_const_XKB_KEY_Greek_OMEGA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_OMEGAaccent", (void *)_cffi_const_XKB_KEY_Greek_OMEGAaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_OMICRON", (void *)_cffi_const_XKB_KEY_Greek_OMICRON, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_OMICRONaccent", (void *)_cffi_const_XKB_KEY_Greek_OMICRONaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_PHI", (void *)_cffi_const_XKB_KEY_Greek_PHI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_PI", (void *)_cffi_const_XKB_KEY_Greek_PI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_PSI", (void *)_cffi_const_XKB_KEY_Greek_PSI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_RHO", (void *)_cffi_const_XKB_KEY_Greek_RHO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_SIGMA", (void *)_cffi_const_XKB_KEY_Greek_SIGMA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_TAU", (void *)_cffi_const_XKB_KEY_Greek_TAU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_THETA", (void *)_cffi_const_XKB_KEY_Greek_THETA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_UPSILON", (void *)_cffi_const_XKB_KEY_Greek_UPSILON, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_UPSILONaccent", (void *)_cffi_const_XKB_KEY_Greek_UPSILONaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_UPSILONdieresis", (void *)_cffi_const_XKB_KEY_Greek_UPSILONdieresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_XI", (void *)_cffi_const_XKB_KEY_Greek_XI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_ZETA", (void *)_cffi_const_XKB_KEY_Greek_ZETA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_accentdieresis", (void *)_cffi_const_XKB_KEY_Greek_accentdieresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_alpha", (void *)_cffi_const_XKB_KEY_Greek_alpha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_alphaaccent", (void *)_cffi_const_XKB_KEY_Greek_alphaaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_beta", (void *)_cffi_const_XKB_KEY_Greek_beta, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_chi", (void *)_cffi_const_XKB_KEY_Greek_chi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_delta", (void *)_cffi_const_XKB_KEY_Greek_delta, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_epsilon", (void *)_cffi_const_XKB_KEY_Greek_epsilon, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_epsilonaccent", (void *)_cffi_const_XKB_KEY_Greek_epsilonaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_eta", (void *)_cffi_const_XKB_KEY_Greek_eta, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_etaaccent", (void *)_cffi_const_XKB_KEY_Greek_etaaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_finalsmallsigma", (void *)_cffi_const_XKB_KEY_Greek_finalsmallsigma, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_gamma", (void *)_cffi_const_XKB_KEY_Greek_gamma, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_horizbar", (void *)_cffi_const_XKB_KEY_Greek_horizbar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_iota", (void *)_cffi_const_XKB_KEY_Greek_iota, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_iotaaccent", (void *)_cffi_const_XKB_KEY_Greek_iotaaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_iotaaccentdieresis", (void *)_cffi_const_XKB_KEY_Greek_iotaaccentdieresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_iotadieresis", (void *)_cffi_const_XKB_KEY_Greek_iotadieresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_kappa", (void *)_cffi_const_XKB_KEY_Greek_kappa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_lambda", (void *)_cffi_const_XKB_KEY_Greek_lambda, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_lamda", (void *)_cffi_const_XKB_KEY_Greek_lamda, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_mu", (void *)_cffi_const_XKB_KEY_Greek_mu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_nu", (void *)_cffi_const_XKB_KEY_Greek_nu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_omega", (void *)_cffi_const_XKB_KEY_Greek_omega, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_omegaaccent", (void *)_cffi_const_XKB_KEY_Greek_omegaaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_omicron", (void *)_cffi_const_XKB_KEY_Greek_omicron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_omicronaccent", (void *)_cffi_const_XKB_KEY_Greek_omicronaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_phi", (void *)_cffi_const_XKB_KEY_Greek_phi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_pi", (void *)_cffi_const_XKB_KEY_Greek_pi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_psi", (void *)_cffi_const_XKB_KEY_Greek_psi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_rho", (void *)_cffi_const_XKB_KEY_Greek_rho, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_sigma", (void *)_cffi_const_XKB_KEY_Greek_sigma, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_switch", (void *)_cffi_const_XKB_KEY_Greek_switch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_tau", (void *)_cffi_const_XKB_KEY_Greek_tau, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_theta", (void *)_cffi_const_XKB_KEY_Greek_theta, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_upsilon", (void *)_cffi_const_XKB_KEY_Greek_upsilon, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_upsilonaccent", (void *)_cffi_const_XKB_KEY_Greek_upsilonaccent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_upsilonaccentdieresis", (void *)_cffi_const_XKB_KEY_Greek_upsilonaccentdieresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_upsilondieresis", (void *)_cffi_const_XKB_KEY_Greek_upsilondieresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_xi", (void *)_cffi_const_XKB_KEY_Greek_xi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Greek_zeta", (void *)_cffi_const_XKB_KEY_Greek_zeta, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_H", (void *)_cffi_const_XKB_KEY_H, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul", (void *)_cffi_const_XKB_KEY_Hangul, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_A", (void *)_cffi_const_XKB_KEY_Hangul_A, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_AE", (void *)_cffi_const_XKB_KEY_Hangul_AE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_AraeA", (void *)_cffi_const_XKB_KEY_Hangul_AraeA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_AraeAE", (void *)_cffi_const_XKB_KEY_Hangul_AraeAE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Banja", (void *)_cffi_const_XKB_KEY_Hangul_Banja, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Cieuc", (void *)_cffi_const_XKB_KEY_Hangul_Cieuc, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Codeinput", (void *)_cffi_const_XKB_KEY_Hangul_Codeinput, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Dikeud", (void *)_cffi_const_XKB_KEY_Hangul_Dikeud, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_E", (void *)_cffi_const_XKB_KEY_Hangul_E, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_EO", (void *)_cffi_const_XKB_KEY_Hangul_EO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_EU", (void *)_cffi_const_XKB_KEY_Hangul_EU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_End", (void *)_cffi_const_XKB_KEY_Hangul_End, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Hanja", (void *)_cffi_const_XKB_KEY_Hangul_Hanja, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Hieuh", (void *)_cffi_const_XKB_KEY_Hangul_Hieuh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_I", (void *)_cffi_const_XKB_KEY_Hangul_I, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Ieung", (void *)_cffi_const_XKB_KEY_Hangul_Ieung, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Cieuc", (void *)_cffi_const_XKB_KEY_Hangul_J_Cieuc, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Dikeud", (void *)_cffi_const_XKB_KEY_Hangul_J_Dikeud, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Hieuh", (void *)_cffi_const_XKB_KEY_Hangul_J_Hieuh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Ieung", (void *)_cffi_const_XKB_KEY_Hangul_J_Ieung, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Jieuj", (void *)_cffi_const_XKB_KEY_Hangul_J_Jieuj, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Khieuq", (void *)_cffi_const_XKB_KEY_Hangul_J_Khieuq, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Kiyeog", (void *)_cffi_const_XKB_KEY_Hangul_J_Kiyeog, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_KiyeogSios", (void *)_cffi_const_XKB_KEY_Hangul_J_KiyeogSios, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_KkogjiDalrinIeung", (void *)_cffi_const_XKB_KEY_Hangul_J_KkogjiDalrinIeung, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Mieum", (void *)_cffi_const_XKB_KEY_Hangul_J_Mieum, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Nieun", (void *)_cffi_const_XKB_KEY_Hangul_J_Nieun, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_NieunHieuh", (void *)_cffi_const_XKB_KEY_Hangul_J_NieunHieuh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_NieunJieuj", (void *)_cffi_const_XKB_KEY_Hangul_J_NieunJieuj, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_PanSios", (void *)_cffi_const_XKB_KEY_Hangul_J_PanSios, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Phieuf", (void *)_cffi_const_XKB_KEY_Hangul_J_Phieuf, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Pieub", (void *)_cffi_const_XKB_KEY_Hangul_J_Pieub, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_PieubSios", (void *)_cffi_const_XKB_KEY_Hangul_J_PieubSios, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Rieul", (void *)_cffi_const_XKB_KEY_Hangul_J_Rieul, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_RieulHieuh", (void *)_cffi_const_XKB_KEY_Hangul_J_RieulHieuh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_RieulKiyeog", (void *)_cffi_const_XKB_KEY_Hangul_J_RieulKiyeog, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_RieulMieum", (void *)_cffi_const_XKB_KEY_Hangul_J_RieulMieum, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_RieulPhieuf", (void *)_cffi_const_XKB_KEY_Hangul_J_RieulPhieuf, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_RieulPieub", (void *)_cffi_const_XKB_KEY_Hangul_J_RieulPieub, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_RieulSios", (void *)_cffi_const_XKB_KEY_Hangul_J_RieulSios, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_RieulTieut", (void *)_cffi_const_XKB_KEY_Hangul_J_RieulTieut, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Sios", (void *)_cffi_const_XKB_KEY_Hangul_J_Sios, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_SsangKiyeog", (void *)_cffi_const_XKB_KEY_Hangul_J_SsangKiyeog, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_SsangSios", (void *)_cffi_const_XKB_KEY_Hangul_J_SsangSios, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_Tieut", (void *)_cffi_const_XKB_KEY_Hangul_J_Tieut, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_J_YeorinHieuh", (void *)_cffi_const_XKB_KEY_Hangul_J_YeorinHieuh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Jamo", (void *)_cffi_const_XKB_KEY_Hangul_Jamo, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Jeonja", (void *)_cffi_const_XKB_KEY_Hangul_Jeonja, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Jieuj", (void *)_cffi_const_XKB_KEY_Hangul_Jieuj, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Khieuq", (void *)_cffi_const_XKB_KEY_Hangul_Khieuq, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Kiyeog", (void *)_cffi_const_XKB_KEY_Hangul_Kiyeog, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_KiyeogSios", (void *)_cffi_const_XKB_KEY_Hangul_KiyeogSios, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_KkogjiDalrinIeung", (void *)_cffi_const_XKB_KEY_Hangul_KkogjiDalrinIeung, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Mieum", (void *)_cffi_const_XKB_KEY_Hangul_Mieum, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_MultipleCandidate", (void *)_cffi_const_XKB_KEY_Hangul_MultipleCandidate, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Nieun", (void *)_cffi_const_XKB_KEY_Hangul_Nieun, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_NieunHieuh", (void *)_cffi_const_XKB_KEY_Hangul_NieunHieuh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_NieunJieuj", (void *)_cffi_const_XKB_KEY_Hangul_NieunJieuj, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_O", (void *)_cffi_const_XKB_KEY_Hangul_O, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_OE", (void *)_cffi_const_XKB_KEY_Hangul_OE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_PanSios", (void *)_cffi_const_XKB_KEY_Hangul_PanSios, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Phieuf", (void *)_cffi_const_XKB_KEY_Hangul_Phieuf, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Pieub", (void *)_cffi_const_XKB_KEY_Hangul_Pieub, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_PieubSios", (void *)_cffi_const_XKB_KEY_Hangul_PieubSios, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_PostHanja", (void *)_cffi_const_XKB_KEY_Hangul_PostHanja, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_PreHanja", (void *)_cffi_const_XKB_KEY_Hangul_PreHanja, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_PreviousCandidate", (void *)_cffi_const_XKB_KEY_Hangul_PreviousCandidate, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Rieul", (void *)_cffi_const_XKB_KEY_Hangul_Rieul, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_RieulHieuh", (void *)_cffi_const_XKB_KEY_Hangul_RieulHieuh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_RieulKiyeog", (void *)_cffi_const_XKB_KEY_Hangul_RieulKiyeog, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_RieulMieum", (void *)_cffi_const_XKB_KEY_Hangul_RieulMieum, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_RieulPhieuf", (void *)_cffi_const_XKB_KEY_Hangul_RieulPhieuf, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_RieulPieub", (void *)_cffi_const_XKB_KEY_Hangul_RieulPieub, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_RieulSios", (void *)_cffi_const_XKB_KEY_Hangul_RieulSios, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_RieulTieut", (void *)_cffi_const_XKB_KEY_Hangul_RieulTieut, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_RieulYeorinHieuh", (void *)_cffi_const_XKB_KEY_Hangul_RieulYeorinHieuh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Romaja", (void *)_cffi_const_XKB_KEY_Hangul_Romaja, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_SingleCandidate", (void *)_cffi_const_XKB_KEY_Hangul_SingleCandidate, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Sios", (void *)_cffi_const_XKB_KEY_Hangul_Sios, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Special", (void *)_cffi_const_XKB_KEY_Hangul_Special, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_SsangDikeud", (void *)_cffi_const_XKB_KEY_Hangul_SsangDikeud, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_SsangJieuj", (void *)_cffi_const_XKB_KEY_Hangul_SsangJieuj, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_SsangKiyeog", (void *)_cffi_const_XKB_KEY_Hangul_SsangKiyeog, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_SsangPieub", (void *)_cffi_const_XKB_KEY_Hangul_SsangPieub, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_SsangSios", (void *)_cffi_const_XKB_KEY_Hangul_SsangSios, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Start", (void *)_cffi_const_XKB_KEY_Hangul_Start, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_SunkyeongeumMieum", (void *)_cffi_const_XKB_KEY_Hangul_SunkyeongeumMieum, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_SunkyeongeumPhieuf", (void *)_cffi_const_XKB_KEY_Hangul_SunkyeongeumPhieuf, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_SunkyeongeumPieub", (void *)_cffi_const_XKB_KEY_Hangul_SunkyeongeumPieub, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_Tieut", (void *)_cffi_const_XKB_KEY_Hangul_Tieut, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_U", (void *)_cffi_const_XKB_KEY_Hangul_U, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_WA", (void *)_cffi_const_XKB_KEY_Hangul_WA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_WAE", (void *)_cffi_const_XKB_KEY_Hangul_WAE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_WE", (void *)_cffi_const_XKB_KEY_Hangul_WE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_WEO", (void *)_cffi_const_XKB_KEY_Hangul_WEO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_WI", (void *)_cffi_const_XKB_KEY_Hangul_WI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_YA", (void *)_cffi_const_XKB_KEY_Hangul_YA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_YAE", (void *)_cffi_const_XKB_KEY_Hangul_YAE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_YE", (void *)_cffi_const_XKB_KEY_Hangul_YE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_YEO", (void *)_cffi_const_XKB_KEY_Hangul_YEO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_YI", (void *)_cffi_const_XKB_KEY_Hangul_YI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_YO", (void *)_cffi_const_XKB_KEY_Hangul_YO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_YU", (void *)_cffi_const_XKB_KEY_Hangul_YU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_YeorinHieuh", (void *)_cffi_const_XKB_KEY_Hangul_YeorinHieuh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hangul_switch", (void *)_cffi_const_XKB_KEY_Hangul_switch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hankaku", (void *)_cffi_const_XKB_KEY_Hankaku, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hcircumflex", (void *)_cffi_const_XKB_KEY_Hcircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hebrew_switch", (void *)_cffi_const_XKB_KEY_Hebrew_switch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Help", (void *)_cffi_const_XKB_KEY_Help, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Henkan", (void *)_cffi_const_XKB_KEY_Henkan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Henkan_Mode", (void *)_cffi_const_XKB_KEY_Henkan_Mode, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hiragana", (void *)_cffi_const_XKB_KEY_Hiragana, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hiragana_Katakana", (void *)_cffi_const_XKB_KEY_Hiragana_Katakana, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Home", (void *)_cffi_const_XKB_KEY_Home, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hstroke", (void *)_cffi_const_XKB_KEY_Hstroke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hyper_L", (void *)_cffi_const_XKB_KEY_Hyper_L, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Hyper_R", (void *)_cffi_const_XKB_KEY_Hyper_R, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_I", (void *)_cffi_const_XKB_KEY_I, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_IO", (void *)_cffi_const_XKB_KEY_IO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Center_Object", (void *)_cffi_const_XKB_KEY_ISO_Center_Object, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Continuous_Underline", (void *)_cffi_const_XKB_KEY_ISO_Continuous_Underline, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Discontinuous_Underline", (void *)_cffi_const_XKB_KEY_ISO_Discontinuous_Underline, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Emphasize", (void *)_cffi_const_XKB_KEY_ISO_Emphasize, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Enter", (void *)_cffi_const_XKB_KEY_ISO_Enter, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Fast_Cursor_Down", (void *)_cffi_const_XKB_KEY_ISO_Fast_Cursor_Down, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Fast_Cursor_Left", (void *)_cffi_const_XKB_KEY_ISO_Fast_Cursor_Left, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Fast_Cursor_Right", (void *)_cffi_const_XKB_KEY_ISO_Fast_Cursor_Right, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Fast_Cursor_Up", (void *)_cffi_const_XKB_KEY_ISO_Fast_Cursor_Up, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_First_Group", (void *)_cffi_const_XKB_KEY_ISO_First_Group, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_First_Group_Lock", (void *)_cffi_const_XKB_KEY_ISO_First_Group_Lock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Group_Latch", (void *)_cffi_const_XKB_KEY_ISO_Group_Latch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Group_Lock", (void *)_cffi_const_XKB_KEY_ISO_Group_Lock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Group_Shift", (void *)_cffi_const_XKB_KEY_ISO_Group_Shift, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Last_Group", (void *)_cffi_const_XKB_KEY_ISO_Last_Group, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Last_Group_Lock", (void *)_cffi_const_XKB_KEY_ISO_Last_Group_Lock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Left_Tab", (void *)_cffi_const_XKB_KEY_ISO_Left_Tab, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Level2_Latch", (void *)_cffi_const_XKB_KEY_ISO_Level2_Latch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Level3_Latch", (void *)_cffi_const_XKB_KEY_ISO_Level3_Latch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Level3_Lock", (void *)_cffi_const_XKB_KEY_ISO_Level3_Lock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Level3_Shift", (void *)_cffi_const_XKB_KEY_ISO_Level3_Shift, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Level5_Latch", (void *)_cffi_const_XKB_KEY_ISO_Level5_Latch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Level5_Lock", (void *)_cffi_const_XKB_KEY_ISO_Level5_Lock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Level5_Shift", (void *)_cffi_const_XKB_KEY_ISO_Level5_Shift, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Lock", (void *)_cffi_const_XKB_KEY_ISO_Lock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Move_Line_Down", (void *)_cffi_const_XKB_KEY_ISO_Move_Line_Down, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Move_Line_Up", (void *)_cffi_const_XKB_KEY_ISO_Move_Line_Up, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Next_Group", (void *)_cffi_const_XKB_KEY_ISO_Next_Group, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Next_Group_Lock", (void *)_cffi_const_XKB_KEY_ISO_Next_Group_Lock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Partial_Line_Down", (void *)_cffi_const_XKB_KEY_ISO_Partial_Line_Down, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Partial_Line_Up", (void *)_cffi_const_XKB_KEY_ISO_Partial_Line_Up, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Partial_Space_Left", (void *)_cffi_const_XKB_KEY_ISO_Partial_Space_Left, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Partial_Space_Right", (void *)_cffi_const_XKB_KEY_ISO_Partial_Space_Right, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Prev_Group", (void *)_cffi_const_XKB_KEY_ISO_Prev_Group, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Prev_Group_Lock", (void *)_cffi_const_XKB_KEY_ISO_Prev_Group_Lock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Release_Both_Margins", (void *)_cffi_const_XKB_KEY_ISO_Release_Both_Margins, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Release_Margin_Left", (void *)_cffi_const_XKB_KEY_ISO_Release_Margin_Left, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Release_Margin_Right", (void *)_cffi_const_XKB_KEY_ISO_Release_Margin_Right, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Set_Margin_Left", (void *)_cffi_const_XKB_KEY_ISO_Set_Margin_Left, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ISO_Set_Margin_Right", (void *)_cffi_const_XKB_KEY_ISO_Set_Margin_Right, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Iabovedot", (void *)_cffi_const_XKB_KEY_Iabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Iacute", (void *)_cffi_const_XKB_KEY_Iacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ibelowdot", (void *)_cffi_const_XKB_KEY_Ibelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ibreve", (void *)_cffi_const_XKB_KEY_Ibreve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Icircumflex", (void *)_cffi_const_XKB_KEY_Icircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Idiaeresis", (void *)_cffi_const_XKB_KEY_Idiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Igrave", (void *)_cffi_const_XKB_KEY_Igrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ihook", (void *)_cffi_const_XKB_KEY_Ihook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Imacron", (void *)_cffi_const_XKB_KEY_Imacron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Insert", (void *)_cffi_const_XKB_KEY_Insert, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_InsertChar", (void *)_cffi_const_XKB_KEY_InsertChar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_InsertLine", (void *)_cffi_const_XKB_KEY_InsertLine, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Iogonek", (void *)_cffi_const_XKB_KEY_Iogonek, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Itilde", (void *)_cffi_const_XKB_KEY_Itilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_J", (void *)_cffi_const_XKB_KEY_J, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Jcircumflex", (void *)_cffi_const_XKB_KEY_Jcircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_K", (void *)_cffi_const_XKB_KEY_K, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_0", (void *)_cffi_const_XKB_KEY_KP_0, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_1", (void *)_cffi_const_XKB_KEY_KP_1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_2", (void *)_cffi_const_XKB_KEY_KP_2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_3", (void *)_cffi_const_XKB_KEY_KP_3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_4", (void *)_cffi_const_XKB_KEY_KP_4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_5", (void *)_cffi_const_XKB_KEY_KP_5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_6", (void *)_cffi_const_XKB_KEY_KP_6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_7", (void *)_cffi_const_XKB_KEY_KP_7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_8", (void *)_cffi_const_XKB_KEY_KP_8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_9", (void *)_cffi_const_XKB_KEY_KP_9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Add", (void *)_cffi_const_XKB_KEY_KP_Add, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_BackTab", (void *)_cffi_const_XKB_KEY_KP_BackTab, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Begin", (void *)_cffi_const_XKB_KEY_KP_Begin, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Decimal", (void *)_cffi_const_XKB_KEY_KP_Decimal, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Delete", (void *)_cffi_const_XKB_KEY_KP_Delete, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Divide", (void *)_cffi_const_XKB_KEY_KP_Divide, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Down", (void *)_cffi_const_XKB_KEY_KP_Down, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_End", (void *)_cffi_const_XKB_KEY_KP_End, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Enter", (void *)_cffi_const_XKB_KEY_KP_Enter, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Equal", (void *)_cffi_const_XKB_KEY_KP_Equal, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_F1", (void *)_cffi_const_XKB_KEY_KP_F1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_F2", (void *)_cffi_const_XKB_KEY_KP_F2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_F3", (void *)_cffi_const_XKB_KEY_KP_F3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_F4", (void *)_cffi_const_XKB_KEY_KP_F4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Home", (void *)_cffi_const_XKB_KEY_KP_Home, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Insert", (void *)_cffi_const_XKB_KEY_KP_Insert, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Left", (void *)_cffi_const_XKB_KEY_KP_Left, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Multiply", (void *)_cffi_const_XKB_KEY_KP_Multiply, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Next", (void *)_cffi_const_XKB_KEY_KP_Next, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Page_Down", (void *)_cffi_const_XKB_KEY_KP_Page_Down, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Page_Up", (void *)_cffi_const_XKB_KEY_KP_Page_Up, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Prior", (void *)_cffi_const_XKB_KEY_KP_Prior, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Right", (void *)_cffi_const_XKB_KEY_KP_Right, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Separator", (void *)_cffi_const_XKB_KEY_KP_Separator, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Space", (void *)_cffi_const_XKB_KEY_KP_Space, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Subtract", (void *)_cffi_const_XKB_KEY_KP_Subtract, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Tab", (void *)_cffi_const_XKB_KEY_KP_Tab, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_KP_Up", (void *)_cffi_const_XKB_KEY_KP_Up, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Kana_Lock", (void *)_cffi_const_XKB_KEY_Kana_Lock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Kana_Shift", (void *)_cffi_const_XKB_KEY_Kana_Shift, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Kanji", (void *)_cffi_const_XKB_KEY_Kanji, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Kanji_Bangou", (void *)_cffi_const_XKB_KEY_Kanji_Bangou, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Katakana", (void *)_cffi_const_XKB_KEY_Katakana, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Kcedilla", (void *)_cffi_const_XKB_KEY_Kcedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Korean_Won", (void *)_cffi_const_XKB_KEY_Korean_Won, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_L", (void *)_cffi_const_XKB_KEY_L, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_L1", (void *)_cffi_const_XKB_KEY_L1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_L10", (void *)_cffi_const_XKB_KEY_L10, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_L2", (void *)_cffi_const_XKB_KEY_L2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_L3", (void *)_cffi_const_XKB_KEY_L3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_L4", (void *)_cffi_const_XKB_KEY_L4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_L5", (void *)_cffi_const_XKB_KEY_L5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_L6", (void *)_cffi_const_XKB_KEY_L6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_L7", (void *)_cffi_const_XKB_KEY_L7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_L8", (void *)_cffi_const_XKB_KEY_L8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_L9", (void *)_cffi_const_XKB_KEY_L9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Lacute", (void *)_cffi_const_XKB_KEY_Lacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Last_Virtual_Screen", (void *)_cffi_const_XKB_KEY_Last_Virtual_Screen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Lbelowdot", (void *)_cffi_const_XKB_KEY_Lbelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Lcaron", (void *)_cffi_const_XKB_KEY_Lcaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Lcedilla", (void *)_cffi_const_XKB_KEY_Lcedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Left", (void *)_cffi_const_XKB_KEY_Left, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Linefeed", (void *)_cffi_const_XKB_KEY_Linefeed, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_LiraSign", (void *)_cffi_const_XKB_KEY_LiraSign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Lstroke", (void *)_cffi_const_XKB_KEY_Lstroke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_M", (void *)_cffi_const_XKB_KEY_M, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Mabovedot", (void *)_cffi_const_XKB_KEY_Mabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Macedonia_DSE", (void *)_cffi_const_XKB_KEY_Macedonia_DSE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Macedonia_GJE", (void *)_cffi_const_XKB_KEY_Macedonia_GJE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Macedonia_KJE", (void *)_cffi_const_XKB_KEY_Macedonia_KJE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Macedonia_dse", (void *)_cffi_const_XKB_KEY_Macedonia_dse, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Macedonia_gje", (void *)_cffi_const_XKB_KEY_Macedonia_gje, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Macedonia_kje", (void *)_cffi_const_XKB_KEY_Macedonia_kje, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Mae_Koho", (void *)_cffi_const_XKB_KEY_Mae_Koho, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Massyo", (void *)_cffi_const_XKB_KEY_Massyo, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Menu", (void *)_cffi_const_XKB_KEY_Menu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Meta_L", (void *)_cffi_const_XKB_KEY_Meta_L, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Meta_R", (void *)_cffi_const_XKB_KEY_Meta_R, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_MillSign", (void *)_cffi_const_XKB_KEY_MillSign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Mode_switch", (void *)_cffi_const_XKB_KEY_Mode_switch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_MouseKeys_Accel_Enable", (void *)_cffi_const_XKB_KEY_MouseKeys_Accel_Enable, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_MouseKeys_Enable", (void *)_cffi_const_XKB_KEY_MouseKeys_Enable, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Muhenkan", (void *)_cffi_const_XKB_KEY_Muhenkan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Multi_key", (void *)_cffi_const_XKB_KEY_Multi_key, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_MultipleCandidate", (void *)_cffi_const_XKB_KEY_MultipleCandidate, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_N", (void *)_cffi_const_XKB_KEY_N, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Nacute", (void *)_cffi_const_XKB_KEY_Nacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_NairaSign", (void *)_cffi_const_XKB_KEY_NairaSign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ncaron", (void *)_cffi_const_XKB_KEY_Ncaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ncedilla", (void *)_cffi_const_XKB_KEY_Ncedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_NewSheqelSign", (void *)_cffi_const_XKB_KEY_NewSheqelSign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Next", (void *)_cffi_const_XKB_KEY_Next, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Next_Virtual_Screen", (void *)_cffi_const_XKB_KEY_Next_Virtual_Screen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_NoSymbol", (void *)_cffi_const_XKB_KEY_NoSymbol, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ntilde", (void *)_cffi_const_XKB_KEY_Ntilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Num_Lock", (void *)_cffi_const_XKB_KEY_Num_Lock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_O", (void *)_cffi_const_XKB_KEY_O, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_OE", (void *)_cffi_const_XKB_KEY_OE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Oacute", (void *)_cffi_const_XKB_KEY_Oacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Obarred", (void *)_cffi_const_XKB_KEY_Obarred, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Obelowdot", (void *)_cffi_const_XKB_KEY_Obelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ocaron", (void *)_cffi_const_XKB_KEY_Ocaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ocircumflex", (void *)_cffi_const_XKB_KEY_Ocircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ocircumflexacute", (void *)_cffi_const_XKB_KEY_Ocircumflexacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ocircumflexbelowdot", (void *)_cffi_const_XKB_KEY_Ocircumflexbelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ocircumflexgrave", (void *)_cffi_const_XKB_KEY_Ocircumflexgrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ocircumflexhook", (void *)_cffi_const_XKB_KEY_Ocircumflexhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ocircumflextilde", (void *)_cffi_const_XKB_KEY_Ocircumflextilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Odiaeresis", (void *)_cffi_const_XKB_KEY_Odiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Odoubleacute", (void *)_cffi_const_XKB_KEY_Odoubleacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ograve", (void *)_cffi_const_XKB_KEY_Ograve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ohook", (void *)_cffi_const_XKB_KEY_Ohook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ohorn", (void *)_cffi_const_XKB_KEY_Ohorn, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ohornacute", (void *)_cffi_const_XKB_KEY_Ohornacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ohornbelowdot", (void *)_cffi_const_XKB_KEY_Ohornbelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ohorngrave", (void *)_cffi_const_XKB_KEY_Ohorngrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ohornhook", (void *)_cffi_const_XKB_KEY_Ohornhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ohorntilde", (void *)_cffi_const_XKB_KEY_Ohorntilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Omacron", (void *)_cffi_const_XKB_KEY_Omacron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ooblique", (void *)_cffi_const_XKB_KEY_Ooblique, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Oslash", (void *)_cffi_const_XKB_KEY_Oslash, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Otilde", (void *)_cffi_const_XKB_KEY_Otilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Overlay1_Enable", (void *)_cffi_const_XKB_KEY_Overlay1_Enable, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Overlay2_Enable", (void *)_cffi_const_XKB_KEY_Overlay2_Enable, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_P", (void *)_cffi_const_XKB_KEY_P, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pabovedot", (void *)_cffi_const_XKB_KEY_Pabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Page_Down", (void *)_cffi_const_XKB_KEY_Page_Down, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Page_Up", (void *)_cffi_const_XKB_KEY_Page_Up, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pause", (void *)_cffi_const_XKB_KEY_Pause, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_PesetaSign", (void *)_cffi_const_XKB_KEY_PesetaSign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Accelerate", (void *)_cffi_const_XKB_KEY_Pointer_Accelerate, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Button1", (void *)_cffi_const_XKB_KEY_Pointer_Button1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Button2", (void *)_cffi_const_XKB_KEY_Pointer_Button2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Button3", (void *)_cffi_const_XKB_KEY_Pointer_Button3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Button4", (void *)_cffi_const_XKB_KEY_Pointer_Button4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Button5", (void *)_cffi_const_XKB_KEY_Pointer_Button5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Button_Dflt", (void *)_cffi_const_XKB_KEY_Pointer_Button_Dflt, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_DblClick1", (void *)_cffi_const_XKB_KEY_Pointer_DblClick1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_DblClick2", (void *)_cffi_const_XKB_KEY_Pointer_DblClick2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_DblClick3", (void *)_cffi_const_XKB_KEY_Pointer_DblClick3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_DblClick4", (void *)_cffi_const_XKB_KEY_Pointer_DblClick4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_DblClick5", (void *)_cffi_const_XKB_KEY_Pointer_DblClick5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_DblClick_Dflt", (void *)_cffi_const_XKB_KEY_Pointer_DblClick_Dflt, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_DfltBtnNext", (void *)_cffi_const_XKB_KEY_Pointer_DfltBtnNext, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_DfltBtnPrev", (void *)_cffi_const_XKB_KEY_Pointer_DfltBtnPrev, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Down", (void *)_cffi_const_XKB_KEY_Pointer_Down, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_DownLeft", (void *)_cffi_const_XKB_KEY_Pointer_DownLeft, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_DownRight", (void *)_cffi_const_XKB_KEY_Pointer_DownRight, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Drag1", (void *)_cffi_const_XKB_KEY_Pointer_Drag1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Drag2", (void *)_cffi_const_XKB_KEY_Pointer_Drag2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Drag3", (void *)_cffi_const_XKB_KEY_Pointer_Drag3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Drag4", (void *)_cffi_const_XKB_KEY_Pointer_Drag4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Drag5", (void *)_cffi_const_XKB_KEY_Pointer_Drag5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Drag_Dflt", (void *)_cffi_const_XKB_KEY_Pointer_Drag_Dflt, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_EnableKeys", (void *)_cffi_const_XKB_KEY_Pointer_EnableKeys, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Left", (void *)_cffi_const_XKB_KEY_Pointer_Left, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Right", (void *)_cffi_const_XKB_KEY_Pointer_Right, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_Up", (void *)_cffi_const_XKB_KEY_Pointer_Up, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_UpLeft", (void *)_cffi_const_XKB_KEY_Pointer_UpLeft, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Pointer_UpRight", (void *)_cffi_const_XKB_KEY_Pointer_UpRight, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Prev_Virtual_Screen", (void *)_cffi_const_XKB_KEY_Prev_Virtual_Screen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_PreviousCandidate", (void *)_cffi_const_XKB_KEY_PreviousCandidate, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Print", (void *)_cffi_const_XKB_KEY_Print, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Prior", (void *)_cffi_const_XKB_KEY_Prior, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Q", (void *)_cffi_const_XKB_KEY_Q, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R", (void *)_cffi_const_XKB_KEY_R, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R1", (void *)_cffi_const_XKB_KEY_R1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R10", (void *)_cffi_const_XKB_KEY_R10, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R11", (void *)_cffi_const_XKB_KEY_R11, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R12", (void *)_cffi_const_XKB_KEY_R12, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R13", (void *)_cffi_const_XKB_KEY_R13, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R14", (void *)_cffi_const_XKB_KEY_R14, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R15", (void *)_cffi_const_XKB_KEY_R15, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R2", (void *)_cffi_const_XKB_KEY_R2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R3", (void *)_cffi_const_XKB_KEY_R3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R4", (void *)_cffi_const_XKB_KEY_R4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R5", (void *)_cffi_const_XKB_KEY_R5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R6", (void *)_cffi_const_XKB_KEY_R6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R7", (void *)_cffi_const_XKB_KEY_R7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R8", (void *)_cffi_const_XKB_KEY_R8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_R9", (void *)_cffi_const_XKB_KEY_R9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Racute", (void *)_cffi_const_XKB_KEY_Racute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Rcaron", (void *)_cffi_const_XKB_KEY_Rcaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Rcedilla", (void *)_cffi_const_XKB_KEY_Rcedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Redo", (void *)_cffi_const_XKB_KEY_Redo, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_RepeatKeys_Enable", (void *)_cffi_const_XKB_KEY_RepeatKeys_Enable, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Reset", (void *)_cffi_const_XKB_KEY_Reset, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Return", (void *)_cffi_const_XKB_KEY_Return, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Right", (void *)_cffi_const_XKB_KEY_Right, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Romaji", (void *)_cffi_const_XKB_KEY_Romaji, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_RupeeSign", (void *)_cffi_const_XKB_KEY_RupeeSign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_S", (void *)_cffi_const_XKB_KEY_S, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SCHWA", (void *)_cffi_const_XKB_KEY_SCHWA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sabovedot", (void *)_cffi_const_XKB_KEY_Sabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sacute", (void *)_cffi_const_XKB_KEY_Sacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Scaron", (void *)_cffi_const_XKB_KEY_Scaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Scedilla", (void *)_cffi_const_XKB_KEY_Scedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Scircumflex", (void *)_cffi_const_XKB_KEY_Scircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Scroll_Lock", (void *)_cffi_const_XKB_KEY_Scroll_Lock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Select", (void *)_cffi_const_XKB_KEY_Select, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Serbian_DJE", (void *)_cffi_const_XKB_KEY_Serbian_DJE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Serbian_DZE", (void *)_cffi_const_XKB_KEY_Serbian_DZE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Serbian_JE", (void *)_cffi_const_XKB_KEY_Serbian_JE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Serbian_LJE", (void *)_cffi_const_XKB_KEY_Serbian_LJE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Serbian_NJE", (void *)_cffi_const_XKB_KEY_Serbian_NJE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Serbian_TSHE", (void *)_cffi_const_XKB_KEY_Serbian_TSHE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Serbian_dje", (void *)_cffi_const_XKB_KEY_Serbian_dje, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Serbian_dze", (void *)_cffi_const_XKB_KEY_Serbian_dze, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Serbian_je", (void *)_cffi_const_XKB_KEY_Serbian_je, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Serbian_lje", (void *)_cffi_const_XKB_KEY_Serbian_lje, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Serbian_nje", (void *)_cffi_const_XKB_KEY_Serbian_nje, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Serbian_tshe", (void *)_cffi_const_XKB_KEY_Serbian_tshe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Shift_L", (void *)_cffi_const_XKB_KEY_Shift_L, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Shift_Lock", (void *)_cffi_const_XKB_KEY_Shift_Lock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Shift_R", (void *)_cffi_const_XKB_KEY_Shift_R, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SingleCandidate", (void *)_cffi_const_XKB_KEY_SingleCandidate, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_a", (void *)_cffi_const_XKB_KEY_Sinh_a, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_aa", (void *)_cffi_const_XKB_KEY_Sinh_aa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_aa2", (void *)_cffi_const_XKB_KEY_Sinh_aa2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ae", (void *)_cffi_const_XKB_KEY_Sinh_ae, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ae2", (void *)_cffi_const_XKB_KEY_Sinh_ae2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_aee", (void *)_cffi_const_XKB_KEY_Sinh_aee, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_aee2", (void *)_cffi_const_XKB_KEY_Sinh_aee2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ai", (void *)_cffi_const_XKB_KEY_Sinh_ai, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ai2", (void *)_cffi_const_XKB_KEY_Sinh_ai2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_al", (void *)_cffi_const_XKB_KEY_Sinh_al, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_au", (void *)_cffi_const_XKB_KEY_Sinh_au, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_au2", (void *)_cffi_const_XKB_KEY_Sinh_au2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ba", (void *)_cffi_const_XKB_KEY_Sinh_ba, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_bha", (void *)_cffi_const_XKB_KEY_Sinh_bha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ca", (void *)_cffi_const_XKB_KEY_Sinh_ca, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_cha", (void *)_cffi_const_XKB_KEY_Sinh_cha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_dda", (void *)_cffi_const_XKB_KEY_Sinh_dda, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ddha", (void *)_cffi_const_XKB_KEY_Sinh_ddha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_dha", (void *)_cffi_const_XKB_KEY_Sinh_dha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_dhha", (void *)_cffi_const_XKB_KEY_Sinh_dhha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_e", (void *)_cffi_const_XKB_KEY_Sinh_e, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_e2", (void *)_cffi_const_XKB_KEY_Sinh_e2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ee", (void *)_cffi_const_XKB_KEY_Sinh_ee, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ee2", (void *)_cffi_const_XKB_KEY_Sinh_ee2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_fa", (void *)_cffi_const_XKB_KEY_Sinh_fa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ga", (void *)_cffi_const_XKB_KEY_Sinh_ga, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_gha", (void *)_cffi_const_XKB_KEY_Sinh_gha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_h2", (void *)_cffi_const_XKB_KEY_Sinh_h2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ha", (void *)_cffi_const_XKB_KEY_Sinh_ha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_i", (void *)_cffi_const_XKB_KEY_Sinh_i, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_i2", (void *)_cffi_const_XKB_KEY_Sinh_i2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ii", (void *)_cffi_const_XKB_KEY_Sinh_ii, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ii2", (void *)_cffi_const_XKB_KEY_Sinh_ii2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ja", (void *)_cffi_const_XKB_KEY_Sinh_ja, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_jha", (void *)_cffi_const_XKB_KEY_Sinh_jha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_jnya", (void *)_cffi_const_XKB_KEY_Sinh_jnya, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ka", (void *)_cffi_const_XKB_KEY_Sinh_ka, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_kha", (void *)_cffi_const_XKB_KEY_Sinh_kha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_kunddaliya", (void *)_cffi_const_XKB_KEY_Sinh_kunddaliya, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_la", (void *)_cffi_const_XKB_KEY_Sinh_la, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_lla", (void *)_cffi_const_XKB_KEY_Sinh_lla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_lu", (void *)_cffi_const_XKB_KEY_Sinh_lu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_lu2", (void *)_cffi_const_XKB_KEY_Sinh_lu2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_luu", (void *)_cffi_const_XKB_KEY_Sinh_luu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_luu2", (void *)_cffi_const_XKB_KEY_Sinh_luu2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ma", (void *)_cffi_const_XKB_KEY_Sinh_ma, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_mba", (void *)_cffi_const_XKB_KEY_Sinh_mba, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_na", (void *)_cffi_const_XKB_KEY_Sinh_na, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ndda", (void *)_cffi_const_XKB_KEY_Sinh_ndda, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ndha", (void *)_cffi_const_XKB_KEY_Sinh_ndha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ng", (void *)_cffi_const_XKB_KEY_Sinh_ng, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ng2", (void *)_cffi_const_XKB_KEY_Sinh_ng2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_nga", (void *)_cffi_const_XKB_KEY_Sinh_nga, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_nja", (void *)_cffi_const_XKB_KEY_Sinh_nja, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_nna", (void *)_cffi_const_XKB_KEY_Sinh_nna, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_nya", (void *)_cffi_const_XKB_KEY_Sinh_nya, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_o", (void *)_cffi_const_XKB_KEY_Sinh_o, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_o2", (void *)_cffi_const_XKB_KEY_Sinh_o2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_oo", (void *)_cffi_const_XKB_KEY_Sinh_oo, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_oo2", (void *)_cffi_const_XKB_KEY_Sinh_oo2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_pa", (void *)_cffi_const_XKB_KEY_Sinh_pa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_pha", (void *)_cffi_const_XKB_KEY_Sinh_pha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ra", (void *)_cffi_const_XKB_KEY_Sinh_ra, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ri", (void *)_cffi_const_XKB_KEY_Sinh_ri, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_rii", (void *)_cffi_const_XKB_KEY_Sinh_rii, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ru2", (void *)_cffi_const_XKB_KEY_Sinh_ru2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ruu2", (void *)_cffi_const_XKB_KEY_Sinh_ruu2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_sa", (void *)_cffi_const_XKB_KEY_Sinh_sa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_sha", (void *)_cffi_const_XKB_KEY_Sinh_sha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ssha", (void *)_cffi_const_XKB_KEY_Sinh_ssha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_tha", (void *)_cffi_const_XKB_KEY_Sinh_tha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_thha", (void *)_cffi_const_XKB_KEY_Sinh_thha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_tta", (void *)_cffi_const_XKB_KEY_Sinh_tta, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ttha", (void *)_cffi_const_XKB_KEY_Sinh_ttha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_u", (void *)_cffi_const_XKB_KEY_Sinh_u, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_u2", (void *)_cffi_const_XKB_KEY_Sinh_u2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_uu", (void *)_cffi_const_XKB_KEY_Sinh_uu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_uu2", (void *)_cffi_const_XKB_KEY_Sinh_uu2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_va", (void *)_cffi_const_XKB_KEY_Sinh_va, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sinh_ya", (void *)_cffi_const_XKB_KEY_Sinh_ya, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SlowKeys_Enable", (void *)_cffi_const_XKB_KEY_SlowKeys_Enable, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_StickyKeys_Enable", (void *)_cffi_const_XKB_KEY_StickyKeys_Enable, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunAgain", (void *)_cffi_const_XKB_KEY_SunAgain, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunAltGraph", (void *)_cffi_const_XKB_KEY_SunAltGraph, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunAudioLowerVolume", (void *)_cffi_const_XKB_KEY_SunAudioLowerVolume, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunAudioMute", (void *)_cffi_const_XKB_KEY_SunAudioMute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunAudioRaiseVolume", (void *)_cffi_const_XKB_KEY_SunAudioRaiseVolume, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunCompose", (void *)_cffi_const_XKB_KEY_SunCompose, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunCopy", (void *)_cffi_const_XKB_KEY_SunCopy, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunCut", (void *)_cffi_const_XKB_KEY_SunCut, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunF36", (void *)_cffi_const_XKB_KEY_SunF36, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunF37", (void *)_cffi_const_XKB_KEY_SunF37, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunFA_Acute", (void *)_cffi_const_XKB_KEY_SunFA_Acute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunFA_Cedilla", (void *)_cffi_const_XKB_KEY_SunFA_Cedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunFA_Circum", (void *)_cffi_const_XKB_KEY_SunFA_Circum, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunFA_Diaeresis", (void *)_cffi_const_XKB_KEY_SunFA_Diaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunFA_Grave", (void *)_cffi_const_XKB_KEY_SunFA_Grave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunFA_Tilde", (void *)_cffi_const_XKB_KEY_SunFA_Tilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunFind", (void *)_cffi_const_XKB_KEY_SunFind, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunFront", (void *)_cffi_const_XKB_KEY_SunFront, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunOpen", (void *)_cffi_const_XKB_KEY_SunOpen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunPageDown", (void *)_cffi_const_XKB_KEY_SunPageDown, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunPageUp", (void *)_cffi_const_XKB_KEY_SunPageUp, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunPaste", (void *)_cffi_const_XKB_KEY_SunPaste, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunPowerSwitch", (void *)_cffi_const_XKB_KEY_SunPowerSwitch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunPowerSwitchShift", (void *)_cffi_const_XKB_KEY_SunPowerSwitchShift, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunPrint_Screen", (void *)_cffi_const_XKB_KEY_SunPrint_Screen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunProps", (void *)_cffi_const_XKB_KEY_SunProps, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunStop", (void *)_cffi_const_XKB_KEY_SunStop, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunSys_Req", (void *)_cffi_const_XKB_KEY_SunSys_Req, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunUndo", (void *)_cffi_const_XKB_KEY_SunUndo, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunVideoDegauss", (void *)_cffi_const_XKB_KEY_SunVideoDegauss, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunVideoLowerBrightness", (void *)_cffi_const_XKB_KEY_SunVideoLowerBrightness, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_SunVideoRaiseBrightness", (void *)_cffi_const_XKB_KEY_SunVideoRaiseBrightness, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Super_L", (void *)_cffi_const_XKB_KEY_Super_L, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Super_R", (void *)_cffi_const_XKB_KEY_Super_R, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Sys_Req", (void *)_cffi_const_XKB_KEY_Sys_Req, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_System", (void *)_cffi_const_XKB_KEY_System, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_T", (void *)_cffi_const_XKB_KEY_T, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_THORN", (void *)_cffi_const_XKB_KEY_THORN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Tab", (void *)_cffi_const_XKB_KEY_Tab, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Tabovedot", (void *)_cffi_const_XKB_KEY_Tabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Tcaron", (void *)_cffi_const_XKB_KEY_Tcaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Tcedilla", (void *)_cffi_const_XKB_KEY_Tcedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Terminate_Server", (void *)_cffi_const_XKB_KEY_Terminate_Server, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_baht", (void *)_cffi_const_XKB_KEY_Thai_baht, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_bobaimai", (void *)_cffi_const_XKB_KEY_Thai_bobaimai, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_chochan", (void *)_cffi_const_XKB_KEY_Thai_chochan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_chochang", (void *)_cffi_const_XKB_KEY_Thai_chochang, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_choching", (void *)_cffi_const_XKB_KEY_Thai_choching, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_chochoe", (void *)_cffi_const_XKB_KEY_Thai_chochoe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_dochada", (void *)_cffi_const_XKB_KEY_Thai_dochada, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_dodek", (void *)_cffi_const_XKB_KEY_Thai_dodek, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_fofa", (void *)_cffi_const_XKB_KEY_Thai_fofa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_fofan", (void *)_cffi_const_XKB_KEY_Thai_fofan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_hohip", (void *)_cffi_const_XKB_KEY_Thai_hohip, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_honokhuk", (void *)_cffi_const_XKB_KEY_Thai_honokhuk, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_khokhai", (void *)_cffi_const_XKB_KEY_Thai_khokhai, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_khokhon", (void *)_cffi_const_XKB_KEY_Thai_khokhon, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_khokhuat", (void *)_cffi_const_XKB_KEY_Thai_khokhuat, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_khokhwai", (void *)_cffi_const_XKB_KEY_Thai_khokhwai, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_khorakhang", (void *)_cffi_const_XKB_KEY_Thai_khorakhang, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_kokai", (void *)_cffi_const_XKB_KEY_Thai_kokai, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_lakkhangyao", (void *)_cffi_const_XKB_KEY_Thai_lakkhangyao, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_lekchet", (void *)_cffi_const_XKB_KEY_Thai_lekchet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_lekha", (void *)_cffi_const_XKB_KEY_Thai_lekha, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_lekhok", (void *)_cffi_const_XKB_KEY_Thai_lekhok, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_lekkao", (void *)_cffi_const_XKB_KEY_Thai_lekkao, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_leknung", (void *)_cffi_const_XKB_KEY_Thai_leknung, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_lekpaet", (void *)_cffi_const_XKB_KEY_Thai_lekpaet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_leksam", (void *)_cffi_const_XKB_KEY_Thai_leksam, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_leksi", (void *)_cffi_const_XKB_KEY_Thai_leksi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_leksong", (void *)_cffi_const_XKB_KEY_Thai_leksong, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_leksun", (void *)_cffi_const_XKB_KEY_Thai_leksun, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_lochula", (void *)_cffi_const_XKB_KEY_Thai_lochula, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_loling", (void *)_cffi_const_XKB_KEY_Thai_loling, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_lu", (void *)_cffi_const_XKB_KEY_Thai_lu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_maichattawa", (void *)_cffi_const_XKB_KEY_Thai_maichattawa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_maiek", (void *)_cffi_const_XKB_KEY_Thai_maiek, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_maihanakat", (void *)_cffi_const_XKB_KEY_Thai_maihanakat, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_maihanakat_maitho", (void *)_cffi_const_XKB_KEY_Thai_maihanakat_maitho, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_maitaikhu", (void *)_cffi_const_XKB_KEY_Thai_maitaikhu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_maitho", (void *)_cffi_const_XKB_KEY_Thai_maitho, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_maitri", (void *)_cffi_const_XKB_KEY_Thai_maitri, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_maiyamok", (void *)_cffi_const_XKB_KEY_Thai_maiyamok, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_moma", (void *)_cffi_const_XKB_KEY_Thai_moma, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_ngongu", (void *)_cffi_const_XKB_KEY_Thai_ngongu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_nikhahit", (void *)_cffi_const_XKB_KEY_Thai_nikhahit, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_nonen", (void *)_cffi_const_XKB_KEY_Thai_nonen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_nonu", (void *)_cffi_const_XKB_KEY_Thai_nonu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_oang", (void *)_cffi_const_XKB_KEY_Thai_oang, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_paiyannoi", (void *)_cffi_const_XKB_KEY_Thai_paiyannoi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_phinthu", (void *)_cffi_const_XKB_KEY_Thai_phinthu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_phophan", (void *)_cffi_const_XKB_KEY_Thai_phophan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_phophung", (void *)_cffi_const_XKB_KEY_Thai_phophung, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_phosamphao", (void *)_cffi_const_XKB_KEY_Thai_phosamphao, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_popla", (void *)_cffi_const_XKB_KEY_Thai_popla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_rorua", (void *)_cffi_const_XKB_KEY_Thai_rorua, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_ru", (void *)_cffi_const_XKB_KEY_Thai_ru, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_saraa", (void *)_cffi_const_XKB_KEY_Thai_saraa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_saraaa", (void *)_cffi_const_XKB_KEY_Thai_saraaa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_saraae", (void *)_cffi_const_XKB_KEY_Thai_saraae, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_saraaimaimalai", (void *)_cffi_const_XKB_KEY_Thai_saraaimaimalai, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_saraaimaimuan", (void *)_cffi_const_XKB_KEY_Thai_saraaimaimuan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_saraam", (void *)_cffi_const_XKB_KEY_Thai_saraam, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_sarae", (void *)_cffi_const_XKB_KEY_Thai_sarae, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_sarai", (void *)_cffi_const_XKB_KEY_Thai_sarai, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_saraii", (void *)_cffi_const_XKB_KEY_Thai_saraii, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_sarao", (void *)_cffi_const_XKB_KEY_Thai_sarao, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_sarau", (void *)_cffi_const_XKB_KEY_Thai_sarau, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_saraue", (void *)_cffi_const_XKB_KEY_Thai_saraue, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_sarauee", (void *)_cffi_const_XKB_KEY_Thai_sarauee, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_sarauu", (void *)_cffi_const_XKB_KEY_Thai_sarauu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_sorusi", (void *)_cffi_const_XKB_KEY_Thai_sorusi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_sosala", (void *)_cffi_const_XKB_KEY_Thai_sosala, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_soso", (void *)_cffi_const_XKB_KEY_Thai_soso, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_sosua", (void *)_cffi_const_XKB_KEY_Thai_sosua, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_thanthakhat", (void *)_cffi_const_XKB_KEY_Thai_thanthakhat, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_thonangmontho", (void *)_cffi_const_XKB_KEY_Thai_thonangmontho, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_thophuthao", (void *)_cffi_const_XKB_KEY_Thai_thophuthao, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_thothahan", (void *)_cffi_const_XKB_KEY_Thai_thothahan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_thothan", (void *)_cffi_const_XKB_KEY_Thai_thothan, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_thothong", (void *)_cffi_const_XKB_KEY_Thai_thothong, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_thothung", (void *)_cffi_const_XKB_KEY_Thai_thothung, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_topatak", (void *)_cffi_const_XKB_KEY_Thai_topatak, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_totao", (void *)_cffi_const_XKB_KEY_Thai_totao, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_wowaen", (void *)_cffi_const_XKB_KEY_Thai_wowaen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_yoyak", (void *)_cffi_const_XKB_KEY_Thai_yoyak, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thai_yoying", (void *)_cffi_const_XKB_KEY_Thai_yoying, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Thorn", (void *)_cffi_const_XKB_KEY_Thorn, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Touroku", (void *)_cffi_const_XKB_KEY_Touroku, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Tslash", (void *)_cffi_const_XKB_KEY_Tslash, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_U", (void *)_cffi_const_XKB_KEY_U, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Uacute", (void *)_cffi_const_XKB_KEY_Uacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ubelowdot", (void *)_cffi_const_XKB_KEY_Ubelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ubreve", (void *)_cffi_const_XKB_KEY_Ubreve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ucircumflex", (void *)_cffi_const_XKB_KEY_Ucircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Udiaeresis", (void *)_cffi_const_XKB_KEY_Udiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Udoubleacute", (void *)_cffi_const_XKB_KEY_Udoubleacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ugrave", (void *)_cffi_const_XKB_KEY_Ugrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Uhook", (void *)_cffi_const_XKB_KEY_Uhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Uhorn", (void *)_cffi_const_XKB_KEY_Uhorn, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Uhornacute", (void *)_cffi_const_XKB_KEY_Uhornacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Uhornbelowdot", (void *)_cffi_const_XKB_KEY_Uhornbelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Uhorngrave", (void *)_cffi_const_XKB_KEY_Uhorngrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Uhornhook", (void *)_cffi_const_XKB_KEY_Uhornhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Uhorntilde", (void *)_cffi_const_XKB_KEY_Uhorntilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukrainian_GHE_WITH_UPTURN", (void *)_cffi_const_XKB_KEY_Ukrainian_GHE_WITH_UPTURN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukrainian_I", (void *)_cffi_const_XKB_KEY_Ukrainian_I, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukrainian_IE", (void *)_cffi_const_XKB_KEY_Ukrainian_IE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukrainian_YI", (void *)_cffi_const_XKB_KEY_Ukrainian_YI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukrainian_ghe_with_upturn", (void *)_cffi_const_XKB_KEY_Ukrainian_ghe_with_upturn, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukrainian_i", (void *)_cffi_const_XKB_KEY_Ukrainian_i, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukrainian_ie", (void *)_cffi_const_XKB_KEY_Ukrainian_ie, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukrainian_yi", (void *)_cffi_const_XKB_KEY_Ukrainian_yi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukranian_I", (void *)_cffi_const_XKB_KEY_Ukranian_I, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukranian_JE", (void *)_cffi_const_XKB_KEY_Ukranian_JE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukranian_YI", (void *)_cffi_const_XKB_KEY_Ukranian_YI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukranian_i", (void *)_cffi_const_XKB_KEY_Ukranian_i, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukranian_je", (void *)_cffi_const_XKB_KEY_Ukranian_je, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ukranian_yi", (void *)_cffi_const_XKB_KEY_Ukranian_yi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Umacron", (void *)_cffi_const_XKB_KEY_Umacron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Undo", (void *)_cffi_const_XKB_KEY_Undo, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Uogonek", (void *)_cffi_const_XKB_KEY_Uogonek, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Up", (void *)_cffi_const_XKB_KEY_Up, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Uring", (void *)_cffi_const_XKB_KEY_Uring, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_User", (void *)_cffi_const_XKB_KEY_User, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Utilde", (void *)_cffi_const_XKB_KEY_Utilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_V", (void *)_cffi_const_XKB_KEY_V, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_VoidSymbol", (void *)_cffi_const_XKB_KEY_VoidSymbol, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_W", (void *)_cffi_const_XKB_KEY_W, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Wacute", (void *)_cffi_const_XKB_KEY_Wacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Wcircumflex", (void *)_cffi_const_XKB_KEY_Wcircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Wdiaeresis", (void *)_cffi_const_XKB_KEY_Wdiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Wgrave", (void *)_cffi_const_XKB_KEY_Wgrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_WonSign", (void *)_cffi_const_XKB_KEY_WonSign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_X", (void *)_cffi_const_XKB_KEY_X, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AddFavorite", (void *)_cffi_const_XKB_KEY_XF86AddFavorite, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86ApplicationLeft", (void *)_cffi_const_XKB_KEY_XF86ApplicationLeft, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86ApplicationRight", (void *)_cffi_const_XKB_KEY_XF86ApplicationRight, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioCycleTrack", (void *)_cffi_const_XKB_KEY_XF86AudioCycleTrack, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioForward", (void *)_cffi_const_XKB_KEY_XF86AudioForward, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioLowerVolume", (void *)_cffi_const_XKB_KEY_XF86AudioLowerVolume, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioMedia", (void *)_cffi_const_XKB_KEY_XF86AudioMedia, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioMicMute", (void *)_cffi_const_XKB_KEY_XF86AudioMicMute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioMute", (void *)_cffi_const_XKB_KEY_XF86AudioMute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioNext", (void *)_cffi_const_XKB_KEY_XF86AudioNext, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioPause", (void *)_cffi_const_XKB_KEY_XF86AudioPause, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioPlay", (void *)_cffi_const_XKB_KEY_XF86AudioPlay, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioPrev", (void *)_cffi_const_XKB_KEY_XF86AudioPrev, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioRaiseVolume", (void *)_cffi_const_XKB_KEY_XF86AudioRaiseVolume, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioRandomPlay", (void *)_cffi_const_XKB_KEY_XF86AudioRandomPlay, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioRecord", (void *)_cffi_const_XKB_KEY_XF86AudioRecord, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioRepeat", (void *)_cffi_const_XKB_KEY_XF86AudioRepeat, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioRewind", (void *)_cffi_const_XKB_KEY_XF86AudioRewind, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86AudioStop", (void *)_cffi_const_XKB_KEY_XF86AudioStop, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Away", (void *)_cffi_const_XKB_KEY_XF86Away, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Back", (void *)_cffi_const_XKB_KEY_XF86Back, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86BackForward", (void *)_cffi_const_XKB_KEY_XF86BackForward, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Battery", (void *)_cffi_const_XKB_KEY_XF86Battery, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Blue", (void *)_cffi_const_XKB_KEY_XF86Blue, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Bluetooth", (void *)_cffi_const_XKB_KEY_XF86Bluetooth, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Book", (void *)_cffi_const_XKB_KEY_XF86Book, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86BrightnessAdjust", (void *)_cffi_const_XKB_KEY_XF86BrightnessAdjust, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86CD", (void *)_cffi_const_XKB_KEY_XF86CD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Calculater", (void *)_cffi_const_XKB_KEY_XF86Calculater, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Calculator", (void *)_cffi_const_XKB_KEY_XF86Calculator, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Calendar", (void *)_cffi_const_XKB_KEY_XF86Calendar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Clear", (void *)_cffi_const_XKB_KEY_XF86Clear, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86ClearGrab", (void *)_cffi_const_XKB_KEY_XF86ClearGrab, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Close", (void *)_cffi_const_XKB_KEY_XF86Close, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Community", (void *)_cffi_const_XKB_KEY_XF86Community, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86ContrastAdjust", (void *)_cffi_const_XKB_KEY_XF86ContrastAdjust, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Copy", (void *)_cffi_const_XKB_KEY_XF86Copy, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Cut", (void *)_cffi_const_XKB_KEY_XF86Cut, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86CycleAngle", (void *)_cffi_const_XKB_KEY_XF86CycleAngle, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86DOS", (void *)_cffi_const_XKB_KEY_XF86DOS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Display", (void *)_cffi_const_XKB_KEY_XF86Display, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Documents", (void *)_cffi_const_XKB_KEY_XF86Documents, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Eject", (void *)_cffi_const_XKB_KEY_XF86Eject, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Excel", (void *)_cffi_const_XKB_KEY_XF86Excel, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Explorer", (void *)_cffi_const_XKB_KEY_XF86Explorer, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Favorites", (void *)_cffi_const_XKB_KEY_XF86Favorites, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Finance", (void *)_cffi_const_XKB_KEY_XF86Finance, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Forward", (void *)_cffi_const_XKB_KEY_XF86Forward, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86FrameBack", (void *)_cffi_const_XKB_KEY_XF86FrameBack, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86FrameForward", (void *)_cffi_const_XKB_KEY_XF86FrameForward, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Game", (void *)_cffi_const_XKB_KEY_XF86Game, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Go", (void *)_cffi_const_XKB_KEY_XF86Go, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Green", (void *)_cffi_const_XKB_KEY_XF86Green, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Hibernate", (void *)_cffi_const_XKB_KEY_XF86Hibernate, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86History", (void *)_cffi_const_XKB_KEY_XF86History, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86HomePage", (void *)_cffi_const_XKB_KEY_XF86HomePage, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86HotLinks", (void *)_cffi_const_XKB_KEY_XF86HotLinks, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86KbdBrightnessDown", (void *)_cffi_const_XKB_KEY_XF86KbdBrightnessDown, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86KbdBrightnessUp", (void *)_cffi_const_XKB_KEY_XF86KbdBrightnessUp, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86KbdLightOnOff", (void *)_cffi_const_XKB_KEY_XF86KbdLightOnOff, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Launch0", (void *)_cffi_const_XKB_KEY_XF86Launch0, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Launch1", (void *)_cffi_const_XKB_KEY_XF86Launch1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Launch2", (void *)_cffi_const_XKB_KEY_XF86Launch2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Launch3", (void *)_cffi_const_XKB_KEY_XF86Launch3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Launch4", (void *)_cffi_const_XKB_KEY_XF86Launch4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Launch5", (void *)_cffi_const_XKB_KEY_XF86Launch5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Launch6", (void *)_cffi_const_XKB_KEY_XF86Launch6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Launch7", (void *)_cffi_const_XKB_KEY_XF86Launch7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Launch8", (void *)_cffi_const_XKB_KEY_XF86Launch8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Launch9", (void *)_cffi_const_XKB_KEY_XF86Launch9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86LaunchA", (void *)_cffi_const_XKB_KEY_XF86LaunchA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86LaunchB", (void *)_cffi_const_XKB_KEY_XF86LaunchB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86LaunchC", (void *)_cffi_const_XKB_KEY_XF86LaunchC, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86LaunchD", (void *)_cffi_const_XKB_KEY_XF86LaunchD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86LaunchE", (void *)_cffi_const_XKB_KEY_XF86LaunchE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86LaunchF", (void *)_cffi_const_XKB_KEY_XF86LaunchF, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86LightBulb", (void *)_cffi_const_XKB_KEY_XF86LightBulb, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86LogGrabInfo", (void *)_cffi_const_XKB_KEY_XF86LogGrabInfo, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86LogOff", (void *)_cffi_const_XKB_KEY_XF86LogOff, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86LogWindowTree", (void *)_cffi_const_XKB_KEY_XF86LogWindowTree, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Mail", (void *)_cffi_const_XKB_KEY_XF86Mail, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86MailForward", (void *)_cffi_const_XKB_KEY_XF86MailForward, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Market", (void *)_cffi_const_XKB_KEY_XF86Market, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Meeting", (void *)_cffi_const_XKB_KEY_XF86Meeting, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Memo", (void *)_cffi_const_XKB_KEY_XF86Memo, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86MenuKB", (void *)_cffi_const_XKB_KEY_XF86MenuKB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86MenuPB", (void *)_cffi_const_XKB_KEY_XF86MenuPB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Messenger", (void *)_cffi_const_XKB_KEY_XF86Messenger, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86ModeLock", (void *)_cffi_const_XKB_KEY_XF86ModeLock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86MonBrightnessDown", (void *)_cffi_const_XKB_KEY_XF86MonBrightnessDown, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86MonBrightnessUp", (void *)_cffi_const_XKB_KEY_XF86MonBrightnessUp, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Music", (void *)_cffi_const_XKB_KEY_XF86Music, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86MyComputer", (void *)_cffi_const_XKB_KEY_XF86MyComputer, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86MySites", (void *)_cffi_const_XKB_KEY_XF86MySites, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86New", (void *)_cffi_const_XKB_KEY_XF86New, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86News", (void *)_cffi_const_XKB_KEY_XF86News, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Next_VMode", (void *)_cffi_const_XKB_KEY_XF86Next_VMode, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86OfficeHome", (void *)_cffi_const_XKB_KEY_XF86OfficeHome, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Open", (void *)_cffi_const_XKB_KEY_XF86Open, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86OpenURL", (void *)_cffi_const_XKB_KEY_XF86OpenURL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Option", (void *)_cffi_const_XKB_KEY_XF86Option, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Paste", (void *)_cffi_const_XKB_KEY_XF86Paste, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Phone", (void *)_cffi_const_XKB_KEY_XF86Phone, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Pictures", (void *)_cffi_const_XKB_KEY_XF86Pictures, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86PowerDown", (void *)_cffi_const_XKB_KEY_XF86PowerDown, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86PowerOff", (void *)_cffi_const_XKB_KEY_XF86PowerOff, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Prev_VMode", (void *)_cffi_const_XKB_KEY_XF86Prev_VMode, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Q", (void *)_cffi_const_XKB_KEY_XF86Q, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Red", (void *)_cffi_const_XKB_KEY_XF86Red, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Refresh", (void *)_cffi_const_XKB_KEY_XF86Refresh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Reload", (void *)_cffi_const_XKB_KEY_XF86Reload, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Reply", (void *)_cffi_const_XKB_KEY_XF86Reply, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86RockerDown", (void *)_cffi_const_XKB_KEY_XF86RockerDown, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86RockerEnter", (void *)_cffi_const_XKB_KEY_XF86RockerEnter, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86RockerUp", (void *)_cffi_const_XKB_KEY_XF86RockerUp, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86RotateWindows", (void *)_cffi_const_XKB_KEY_XF86RotateWindows, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86RotationKB", (void *)_cffi_const_XKB_KEY_XF86RotationKB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86RotationPB", (void *)_cffi_const_XKB_KEY_XF86RotationPB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Save", (void *)_cffi_const_XKB_KEY_XF86Save, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86ScreenSaver", (void *)_cffi_const_XKB_KEY_XF86ScreenSaver, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86ScrollClick", (void *)_cffi_const_XKB_KEY_XF86ScrollClick, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86ScrollDown", (void *)_cffi_const_XKB_KEY_XF86ScrollDown, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86ScrollUp", (void *)_cffi_const_XKB_KEY_XF86ScrollUp, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Search", (void *)_cffi_const_XKB_KEY_XF86Search, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Select", (void *)_cffi_const_XKB_KEY_XF86Select, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Send", (void *)_cffi_const_XKB_KEY_XF86Send, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Shop", (void *)_cffi_const_XKB_KEY_XF86Shop, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Sleep", (void *)_cffi_const_XKB_KEY_XF86Sleep, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Spell", (void *)_cffi_const_XKB_KEY_XF86Spell, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86SplitScreen", (void *)_cffi_const_XKB_KEY_XF86SplitScreen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Standby", (void *)_cffi_const_XKB_KEY_XF86Standby, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Start", (void *)_cffi_const_XKB_KEY_XF86Start, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Stop", (void *)_cffi_const_XKB_KEY_XF86Stop, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Subtitle", (void *)_cffi_const_XKB_KEY_XF86Subtitle, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Support", (void *)_cffi_const_XKB_KEY_XF86Support, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Suspend", (void *)_cffi_const_XKB_KEY_XF86Suspend, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Switch_VT_1", (void *)_cffi_const_XKB_KEY_XF86Switch_VT_1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Switch_VT_10", (void *)_cffi_const_XKB_KEY_XF86Switch_VT_10, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Switch_VT_11", (void *)_cffi_const_XKB_KEY_XF86Switch_VT_11, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Switch_VT_12", (void *)_cffi_const_XKB_KEY_XF86Switch_VT_12, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Switch_VT_2", (void *)_cffi_const_XKB_KEY_XF86Switch_VT_2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Switch_VT_3", (void *)_cffi_const_XKB_KEY_XF86Switch_VT_3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Switch_VT_4", (void *)_cffi_const_XKB_KEY_XF86Switch_VT_4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Switch_VT_5", (void *)_cffi_const_XKB_KEY_XF86Switch_VT_5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Switch_VT_6", (void *)_cffi_const_XKB_KEY_XF86Switch_VT_6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Switch_VT_7", (void *)_cffi_const_XKB_KEY_XF86Switch_VT_7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Switch_VT_8", (void *)_cffi_const_XKB_KEY_XF86Switch_VT_8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Switch_VT_9", (void *)_cffi_const_XKB_KEY_XF86Switch_VT_9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86TaskPane", (void *)_cffi_const_XKB_KEY_XF86TaskPane, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Terminal", (void *)_cffi_const_XKB_KEY_XF86Terminal, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Time", (void *)_cffi_const_XKB_KEY_XF86Time, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86ToDoList", (void *)_cffi_const_XKB_KEY_XF86ToDoList, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Tools", (void *)_cffi_const_XKB_KEY_XF86Tools, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86TopMenu", (void *)_cffi_const_XKB_KEY_XF86TopMenu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86TouchpadOff", (void *)_cffi_const_XKB_KEY_XF86TouchpadOff, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86TouchpadOn", (void *)_cffi_const_XKB_KEY_XF86TouchpadOn, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86TouchpadToggle", (void *)_cffi_const_XKB_KEY_XF86TouchpadToggle, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Travel", (void *)_cffi_const_XKB_KEY_XF86Travel, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86UWB", (void *)_cffi_const_XKB_KEY_XF86UWB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Ungrab", (void *)_cffi_const_XKB_KEY_XF86Ungrab, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86User1KB", (void *)_cffi_const_XKB_KEY_XF86User1KB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86User2KB", (void *)_cffi_const_XKB_KEY_XF86User2KB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86UserPB", (void *)_cffi_const_XKB_KEY_XF86UserPB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86VendorHome", (void *)_cffi_const_XKB_KEY_XF86VendorHome, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Video", (void *)_cffi_const_XKB_KEY_XF86Video, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86View", (void *)_cffi_const_XKB_KEY_XF86View, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86WLAN", (void *)_cffi_const_XKB_KEY_XF86WLAN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86WWW", (void *)_cffi_const_XKB_KEY_XF86WWW, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86WakeUp", (void *)_cffi_const_XKB_KEY_XF86WakeUp, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86WebCam", (void *)_cffi_const_XKB_KEY_XF86WebCam, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86WheelButton", (void *)_cffi_const_XKB_KEY_XF86WheelButton, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Word", (void *)_cffi_const_XKB_KEY_XF86Word, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Xfer", (void *)_cffi_const_XKB_KEY_XF86Xfer, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86Yellow", (void *)_cffi_const_XKB_KEY_XF86Yellow, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86ZoomIn", (void *)_cffi_const_XKB_KEY_XF86ZoomIn, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86ZoomOut", (void *)_cffi_const_XKB_KEY_XF86ZoomOut, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_XF86iTouch", (void *)_cffi_const_XKB_KEY_XF86iTouch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Xabovedot", (void *)_cffi_const_XKB_KEY_Xabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Y", (void *)_cffi_const_XKB_KEY_Y, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Yacute", (void *)_cffi_const_XKB_KEY_Yacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ybelowdot", (void *)_cffi_const_XKB_KEY_Ybelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ycircumflex", (void *)_cffi_const_XKB_KEY_Ycircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ydiaeresis", (void *)_cffi_const_XKB_KEY_Ydiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ygrave", (void *)_cffi_const_XKB_KEY_Ygrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Yhook", (void *)_cffi_const_XKB_KEY_Yhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Ytilde", (void *)_cffi_const_XKB_KEY_Ytilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Z", (void *)_cffi_const_XKB_KEY_Z, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Zabovedot", (void *)_cffi_const_XKB_KEY_Zabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Zacute", (void *)_cffi_const_XKB_KEY_Zacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Zcaron", (void *)_cffi_const_XKB_KEY_Zcaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Zen_Koho", (void *)_cffi_const_XKB_KEY_Zen_Koho, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Zenkaku", (void *)_cffi_const_XKB_KEY_Zenkaku, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Zenkaku_Hankaku", (void *)_cffi_const_XKB_KEY_Zenkaku_Hankaku, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_Zstroke", (void *)_cffi_const_XKB_KEY_Zstroke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_a", (void *)_cffi_const_XKB_KEY_a, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_aacute", (void *)_cffi_const_XKB_KEY_aacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_abelowdot", (void *)_cffi_const_XKB_KEY_abelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_abovedot", (void *)_cffi_const_XKB_KEY_abovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_abreve", (void *)_cffi_const_XKB_KEY_abreve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_abreveacute", (void *)_cffi_const_XKB_KEY_abreveacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_abrevebelowdot", (void *)_cffi_const_XKB_KEY_abrevebelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_abrevegrave", (void *)_cffi_const_XKB_KEY_abrevegrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_abrevehook", (void *)_cffi_const_XKB_KEY_abrevehook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_abrevetilde", (void *)_cffi_const_XKB_KEY_abrevetilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_acircumflex", (void *)_cffi_const_XKB_KEY_acircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_acircumflexacute", (void *)_cffi_const_XKB_KEY_acircumflexacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_acircumflexbelowdot", (void *)_cffi_const_XKB_KEY_acircumflexbelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_acircumflexgrave", (void *)_cffi_const_XKB_KEY_acircumflexgrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_acircumflexhook", (void *)_cffi_const_XKB_KEY_acircumflexhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_acircumflextilde", (void *)_cffi_const_XKB_KEY_acircumflextilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_acute", (void *)_cffi_const_XKB_KEY_acute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_adiaeresis", (void *)_cffi_const_XKB_KEY_adiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ae", (void *)_cffi_const_XKB_KEY_ae, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_agrave", (void *)_cffi_const_XKB_KEY_agrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ahook", (void *)_cffi_const_XKB_KEY_ahook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_amacron", (void *)_cffi_const_XKB_KEY_amacron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ampersand", (void *)_cffi_const_XKB_KEY_ampersand, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_aogonek", (void *)_cffi_const_XKB_KEY_aogonek, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_apostrophe", (void *)_cffi_const_XKB_KEY_apostrophe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_approxeq", (void *)_cffi_const_XKB_KEY_approxeq, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_approximate", (void *)_cffi_const_XKB_KEY_approximate, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_aring", (void *)_cffi_const_XKB_KEY_aring, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_asciicircum", (void *)_cffi_const_XKB_KEY_asciicircum, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_asciitilde", (void *)_cffi_const_XKB_KEY_asciitilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_asterisk", (void *)_cffi_const_XKB_KEY_asterisk, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_at", (void *)_cffi_const_XKB_KEY_at, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_atilde", (void *)_cffi_const_XKB_KEY_atilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_b", (void *)_cffi_const_XKB_KEY_b, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_babovedot", (void *)_cffi_const_XKB_KEY_babovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_backslash", (void *)_cffi_const_XKB_KEY_backslash, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ballotcross", (void *)_cffi_const_XKB_KEY_ballotcross, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_bar", (void *)_cffi_const_XKB_KEY_bar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_because", (void *)_cffi_const_XKB_KEY_because, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_blank", (void *)_cffi_const_XKB_KEY_blank, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_block", (void *)_cffi_const_XKB_KEY_block, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_botintegral", (void *)_cffi_const_XKB_KEY_botintegral, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_botleftparens", (void *)_cffi_const_XKB_KEY_botleftparens, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_botleftsqbracket", (void *)_cffi_const_XKB_KEY_botleftsqbracket, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_botleftsummation", (void *)_cffi_const_XKB_KEY_botleftsummation, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_botrightparens", (void *)_cffi_const_XKB_KEY_botrightparens, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_botrightsqbracket", (void *)_cffi_const_XKB_KEY_botrightsqbracket, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_botrightsummation", (void *)_cffi_const_XKB_KEY_botrightsummation, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_bott", (void *)_cffi_const_XKB_KEY_bott, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_botvertsummationconnector", (void *)_cffi_const_XKB_KEY_botvertsummationconnector, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braceleft", (void *)_cffi_const_XKB_KEY_braceleft, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braceright", (void *)_cffi_const_XKB_KEY_braceright, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_bracketleft", (void *)_cffi_const_XKB_KEY_bracketleft, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_bracketright", (void *)_cffi_const_XKB_KEY_bracketright, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_blank", (void *)_cffi_const_XKB_KEY_braille_blank, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dot_1", (void *)_cffi_const_XKB_KEY_braille_dot_1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dot_10", (void *)_cffi_const_XKB_KEY_braille_dot_10, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dot_2", (void *)_cffi_const_XKB_KEY_braille_dot_2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dot_3", (void *)_cffi_const_XKB_KEY_braille_dot_3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dot_4", (void *)_cffi_const_XKB_KEY_braille_dot_4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dot_5", (void *)_cffi_const_XKB_KEY_braille_dot_5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dot_6", (void *)_cffi_const_XKB_KEY_braille_dot_6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dot_7", (void *)_cffi_const_XKB_KEY_braille_dot_7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dot_8", (void *)_cffi_const_XKB_KEY_braille_dot_8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dot_9", (void *)_cffi_const_XKB_KEY_braille_dot_9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1", (void *)_cffi_const_XKB_KEY_braille_dots_1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12", (void *)_cffi_const_XKB_KEY_braille_dots_12, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_123", (void *)_cffi_const_XKB_KEY_braille_dots_123, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1234", (void *)_cffi_const_XKB_KEY_braille_dots_1234, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12345", (void *)_cffi_const_XKB_KEY_braille_dots_12345, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_123456", (void *)_cffi_const_XKB_KEY_braille_dots_123456, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1234567", (void *)_cffi_const_XKB_KEY_braille_dots_1234567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12345678", (void *)_cffi_const_XKB_KEY_braille_dots_12345678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1234568", (void *)_cffi_const_XKB_KEY_braille_dots_1234568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_123457", (void *)_cffi_const_XKB_KEY_braille_dots_123457, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1234578", (void *)_cffi_const_XKB_KEY_braille_dots_1234578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_123458", (void *)_cffi_const_XKB_KEY_braille_dots_123458, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12346", (void *)_cffi_const_XKB_KEY_braille_dots_12346, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_123467", (void *)_cffi_const_XKB_KEY_braille_dots_123467, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1234678", (void *)_cffi_const_XKB_KEY_braille_dots_1234678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_123468", (void *)_cffi_const_XKB_KEY_braille_dots_123468, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12347", (void *)_cffi_const_XKB_KEY_braille_dots_12347, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_123478", (void *)_cffi_const_XKB_KEY_braille_dots_123478, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12348", (void *)_cffi_const_XKB_KEY_braille_dots_12348, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1235", (void *)_cffi_const_XKB_KEY_braille_dots_1235, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12356", (void *)_cffi_const_XKB_KEY_braille_dots_12356, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_123567", (void *)_cffi_const_XKB_KEY_braille_dots_123567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1235678", (void *)_cffi_const_XKB_KEY_braille_dots_1235678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_123568", (void *)_cffi_const_XKB_KEY_braille_dots_123568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12357", (void *)_cffi_const_XKB_KEY_braille_dots_12357, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_123578", (void *)_cffi_const_XKB_KEY_braille_dots_123578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12358", (void *)_cffi_const_XKB_KEY_braille_dots_12358, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1236", (void *)_cffi_const_XKB_KEY_braille_dots_1236, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12367", (void *)_cffi_const_XKB_KEY_braille_dots_12367, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_123678", (void *)_cffi_const_XKB_KEY_braille_dots_123678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12368", (void *)_cffi_const_XKB_KEY_braille_dots_12368, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1237", (void *)_cffi_const_XKB_KEY_braille_dots_1237, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12378", (void *)_cffi_const_XKB_KEY_braille_dots_12378, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1238", (void *)_cffi_const_XKB_KEY_braille_dots_1238, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_124", (void *)_cffi_const_XKB_KEY_braille_dots_124, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1245", (void *)_cffi_const_XKB_KEY_braille_dots_1245, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12456", (void *)_cffi_const_XKB_KEY_braille_dots_12456, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_124567", (void *)_cffi_const_XKB_KEY_braille_dots_124567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1245678", (void *)_cffi_const_XKB_KEY_braille_dots_1245678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_124568", (void *)_cffi_const_XKB_KEY_braille_dots_124568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12457", (void *)_cffi_const_XKB_KEY_braille_dots_12457, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_124578", (void *)_cffi_const_XKB_KEY_braille_dots_124578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12458", (void *)_cffi_const_XKB_KEY_braille_dots_12458, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1246", (void *)_cffi_const_XKB_KEY_braille_dots_1246, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12467", (void *)_cffi_const_XKB_KEY_braille_dots_12467, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_124678", (void *)_cffi_const_XKB_KEY_braille_dots_124678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12468", (void *)_cffi_const_XKB_KEY_braille_dots_12468, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1247", (void *)_cffi_const_XKB_KEY_braille_dots_1247, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12478", (void *)_cffi_const_XKB_KEY_braille_dots_12478, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1248", (void *)_cffi_const_XKB_KEY_braille_dots_1248, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_125", (void *)_cffi_const_XKB_KEY_braille_dots_125, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1256", (void *)_cffi_const_XKB_KEY_braille_dots_1256, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12567", (void *)_cffi_const_XKB_KEY_braille_dots_12567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_125678", (void *)_cffi_const_XKB_KEY_braille_dots_125678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12568", (void *)_cffi_const_XKB_KEY_braille_dots_12568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1257", (void *)_cffi_const_XKB_KEY_braille_dots_1257, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12578", (void *)_cffi_const_XKB_KEY_braille_dots_12578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1258", (void *)_cffi_const_XKB_KEY_braille_dots_1258, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_126", (void *)_cffi_const_XKB_KEY_braille_dots_126, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1267", (void *)_cffi_const_XKB_KEY_braille_dots_1267, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_12678", (void *)_cffi_const_XKB_KEY_braille_dots_12678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1268", (void *)_cffi_const_XKB_KEY_braille_dots_1268, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_127", (void *)_cffi_const_XKB_KEY_braille_dots_127, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1278", (void *)_cffi_const_XKB_KEY_braille_dots_1278, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_128", (void *)_cffi_const_XKB_KEY_braille_dots_128, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_13", (void *)_cffi_const_XKB_KEY_braille_dots_13, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_134", (void *)_cffi_const_XKB_KEY_braille_dots_134, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1345", (void *)_cffi_const_XKB_KEY_braille_dots_1345, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_13456", (void *)_cffi_const_XKB_KEY_braille_dots_13456, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_134567", (void *)_cffi_const_XKB_KEY_braille_dots_134567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1345678", (void *)_cffi_const_XKB_KEY_braille_dots_1345678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_134568", (void *)_cffi_const_XKB_KEY_braille_dots_134568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_13457", (void *)_cffi_const_XKB_KEY_braille_dots_13457, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_134578", (void *)_cffi_const_XKB_KEY_braille_dots_134578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_13458", (void *)_cffi_const_XKB_KEY_braille_dots_13458, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1346", (void *)_cffi_const_XKB_KEY_braille_dots_1346, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_13467", (void *)_cffi_const_XKB_KEY_braille_dots_13467, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_134678", (void *)_cffi_const_XKB_KEY_braille_dots_134678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_13468", (void *)_cffi_const_XKB_KEY_braille_dots_13468, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1347", (void *)_cffi_const_XKB_KEY_braille_dots_1347, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_13478", (void *)_cffi_const_XKB_KEY_braille_dots_13478, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1348", (void *)_cffi_const_XKB_KEY_braille_dots_1348, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_135", (void *)_cffi_const_XKB_KEY_braille_dots_135, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1356", (void *)_cffi_const_XKB_KEY_braille_dots_1356, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_13567", (void *)_cffi_const_XKB_KEY_braille_dots_13567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_135678", (void *)_cffi_const_XKB_KEY_braille_dots_135678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_13568", (void *)_cffi_const_XKB_KEY_braille_dots_13568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1357", (void *)_cffi_const_XKB_KEY_braille_dots_1357, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_13578", (void *)_cffi_const_XKB_KEY_braille_dots_13578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1358", (void *)_cffi_const_XKB_KEY_braille_dots_1358, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_136", (void *)_cffi_const_XKB_KEY_braille_dots_136, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1367", (void *)_cffi_const_XKB_KEY_braille_dots_1367, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_13678", (void *)_cffi_const_XKB_KEY_braille_dots_13678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1368", (void *)_cffi_const_XKB_KEY_braille_dots_1368, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_137", (void *)_cffi_const_XKB_KEY_braille_dots_137, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1378", (void *)_cffi_const_XKB_KEY_braille_dots_1378, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_138", (void *)_cffi_const_XKB_KEY_braille_dots_138, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_14", (void *)_cffi_const_XKB_KEY_braille_dots_14, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_145", (void *)_cffi_const_XKB_KEY_braille_dots_145, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1456", (void *)_cffi_const_XKB_KEY_braille_dots_1456, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_14567", (void *)_cffi_const_XKB_KEY_braille_dots_14567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_145678", (void *)_cffi_const_XKB_KEY_braille_dots_145678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_14568", (void *)_cffi_const_XKB_KEY_braille_dots_14568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1457", (void *)_cffi_const_XKB_KEY_braille_dots_1457, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_14578", (void *)_cffi_const_XKB_KEY_braille_dots_14578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1458", (void *)_cffi_const_XKB_KEY_braille_dots_1458, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_146", (void *)_cffi_const_XKB_KEY_braille_dots_146, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1467", (void *)_cffi_const_XKB_KEY_braille_dots_1467, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_14678", (void *)_cffi_const_XKB_KEY_braille_dots_14678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1468", (void *)_cffi_const_XKB_KEY_braille_dots_1468, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_147", (void *)_cffi_const_XKB_KEY_braille_dots_147, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1478", (void *)_cffi_const_XKB_KEY_braille_dots_1478, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_148", (void *)_cffi_const_XKB_KEY_braille_dots_148, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_15", (void *)_cffi_const_XKB_KEY_braille_dots_15, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_156", (void *)_cffi_const_XKB_KEY_braille_dots_156, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1567", (void *)_cffi_const_XKB_KEY_braille_dots_1567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_15678", (void *)_cffi_const_XKB_KEY_braille_dots_15678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1568", (void *)_cffi_const_XKB_KEY_braille_dots_1568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_157", (void *)_cffi_const_XKB_KEY_braille_dots_157, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1578", (void *)_cffi_const_XKB_KEY_braille_dots_1578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_158", (void *)_cffi_const_XKB_KEY_braille_dots_158, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_16", (void *)_cffi_const_XKB_KEY_braille_dots_16, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_167", (void *)_cffi_const_XKB_KEY_braille_dots_167, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_1678", (void *)_cffi_const_XKB_KEY_braille_dots_1678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_168", (void *)_cffi_const_XKB_KEY_braille_dots_168, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_17", (void *)_cffi_const_XKB_KEY_braille_dots_17, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_178", (void *)_cffi_const_XKB_KEY_braille_dots_178, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_18", (void *)_cffi_const_XKB_KEY_braille_dots_18, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2", (void *)_cffi_const_XKB_KEY_braille_dots_2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_23", (void *)_cffi_const_XKB_KEY_braille_dots_23, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_234", (void *)_cffi_const_XKB_KEY_braille_dots_234, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2345", (void *)_cffi_const_XKB_KEY_braille_dots_2345, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_23456", (void *)_cffi_const_XKB_KEY_braille_dots_23456, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_234567", (void *)_cffi_const_XKB_KEY_braille_dots_234567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2345678", (void *)_cffi_const_XKB_KEY_braille_dots_2345678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_234568", (void *)_cffi_const_XKB_KEY_braille_dots_234568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_23457", (void *)_cffi_const_XKB_KEY_braille_dots_23457, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_234578", (void *)_cffi_const_XKB_KEY_braille_dots_234578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_23458", (void *)_cffi_const_XKB_KEY_braille_dots_23458, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2346", (void *)_cffi_const_XKB_KEY_braille_dots_2346, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_23467", (void *)_cffi_const_XKB_KEY_braille_dots_23467, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_234678", (void *)_cffi_const_XKB_KEY_braille_dots_234678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_23468", (void *)_cffi_const_XKB_KEY_braille_dots_23468, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2347", (void *)_cffi_const_XKB_KEY_braille_dots_2347, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_23478", (void *)_cffi_const_XKB_KEY_braille_dots_23478, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2348", (void *)_cffi_const_XKB_KEY_braille_dots_2348, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_235", (void *)_cffi_const_XKB_KEY_braille_dots_235, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2356", (void *)_cffi_const_XKB_KEY_braille_dots_2356, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_23567", (void *)_cffi_const_XKB_KEY_braille_dots_23567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_235678", (void *)_cffi_const_XKB_KEY_braille_dots_235678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_23568", (void *)_cffi_const_XKB_KEY_braille_dots_23568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2357", (void *)_cffi_const_XKB_KEY_braille_dots_2357, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_23578", (void *)_cffi_const_XKB_KEY_braille_dots_23578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2358", (void *)_cffi_const_XKB_KEY_braille_dots_2358, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_236", (void *)_cffi_const_XKB_KEY_braille_dots_236, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2367", (void *)_cffi_const_XKB_KEY_braille_dots_2367, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_23678", (void *)_cffi_const_XKB_KEY_braille_dots_23678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2368", (void *)_cffi_const_XKB_KEY_braille_dots_2368, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_237", (void *)_cffi_const_XKB_KEY_braille_dots_237, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2378", (void *)_cffi_const_XKB_KEY_braille_dots_2378, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_238", (void *)_cffi_const_XKB_KEY_braille_dots_238, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_24", (void *)_cffi_const_XKB_KEY_braille_dots_24, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_245", (void *)_cffi_const_XKB_KEY_braille_dots_245, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2456", (void *)_cffi_const_XKB_KEY_braille_dots_2456, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_24567", (void *)_cffi_const_XKB_KEY_braille_dots_24567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_245678", (void *)_cffi_const_XKB_KEY_braille_dots_245678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_24568", (void *)_cffi_const_XKB_KEY_braille_dots_24568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2457", (void *)_cffi_const_XKB_KEY_braille_dots_2457, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_24578", (void *)_cffi_const_XKB_KEY_braille_dots_24578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2458", (void *)_cffi_const_XKB_KEY_braille_dots_2458, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_246", (void *)_cffi_const_XKB_KEY_braille_dots_246, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2467", (void *)_cffi_const_XKB_KEY_braille_dots_2467, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_24678", (void *)_cffi_const_XKB_KEY_braille_dots_24678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2468", (void *)_cffi_const_XKB_KEY_braille_dots_2468, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_247", (void *)_cffi_const_XKB_KEY_braille_dots_247, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2478", (void *)_cffi_const_XKB_KEY_braille_dots_2478, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_248", (void *)_cffi_const_XKB_KEY_braille_dots_248, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_25", (void *)_cffi_const_XKB_KEY_braille_dots_25, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_256", (void *)_cffi_const_XKB_KEY_braille_dots_256, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2567", (void *)_cffi_const_XKB_KEY_braille_dots_2567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_25678", (void *)_cffi_const_XKB_KEY_braille_dots_25678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2568", (void *)_cffi_const_XKB_KEY_braille_dots_2568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_257", (void *)_cffi_const_XKB_KEY_braille_dots_257, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2578", (void *)_cffi_const_XKB_KEY_braille_dots_2578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_258", (void *)_cffi_const_XKB_KEY_braille_dots_258, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_26", (void *)_cffi_const_XKB_KEY_braille_dots_26, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_267", (void *)_cffi_const_XKB_KEY_braille_dots_267, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_2678", (void *)_cffi_const_XKB_KEY_braille_dots_2678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_268", (void *)_cffi_const_XKB_KEY_braille_dots_268, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_27", (void *)_cffi_const_XKB_KEY_braille_dots_27, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_278", (void *)_cffi_const_XKB_KEY_braille_dots_278, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_28", (void *)_cffi_const_XKB_KEY_braille_dots_28, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_3", (void *)_cffi_const_XKB_KEY_braille_dots_3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_34", (void *)_cffi_const_XKB_KEY_braille_dots_34, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_345", (void *)_cffi_const_XKB_KEY_braille_dots_345, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_3456", (void *)_cffi_const_XKB_KEY_braille_dots_3456, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_34567", (void *)_cffi_const_XKB_KEY_braille_dots_34567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_345678", (void *)_cffi_const_XKB_KEY_braille_dots_345678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_34568", (void *)_cffi_const_XKB_KEY_braille_dots_34568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_3457", (void *)_cffi_const_XKB_KEY_braille_dots_3457, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_34578", (void *)_cffi_const_XKB_KEY_braille_dots_34578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_3458", (void *)_cffi_const_XKB_KEY_braille_dots_3458, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_346", (void *)_cffi_const_XKB_KEY_braille_dots_346, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_3467", (void *)_cffi_const_XKB_KEY_braille_dots_3467, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_34678", (void *)_cffi_const_XKB_KEY_braille_dots_34678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_3468", (void *)_cffi_const_XKB_KEY_braille_dots_3468, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_347", (void *)_cffi_const_XKB_KEY_braille_dots_347, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_3478", (void *)_cffi_const_XKB_KEY_braille_dots_3478, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_348", (void *)_cffi_const_XKB_KEY_braille_dots_348, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_35", (void *)_cffi_const_XKB_KEY_braille_dots_35, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_356", (void *)_cffi_const_XKB_KEY_braille_dots_356, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_3567", (void *)_cffi_const_XKB_KEY_braille_dots_3567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_35678", (void *)_cffi_const_XKB_KEY_braille_dots_35678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_3568", (void *)_cffi_const_XKB_KEY_braille_dots_3568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_357", (void *)_cffi_const_XKB_KEY_braille_dots_357, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_3578", (void *)_cffi_const_XKB_KEY_braille_dots_3578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_358", (void *)_cffi_const_XKB_KEY_braille_dots_358, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_36", (void *)_cffi_const_XKB_KEY_braille_dots_36, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_367", (void *)_cffi_const_XKB_KEY_braille_dots_367, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_3678", (void *)_cffi_const_XKB_KEY_braille_dots_3678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_368", (void *)_cffi_const_XKB_KEY_braille_dots_368, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_37", (void *)_cffi_const_XKB_KEY_braille_dots_37, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_378", (void *)_cffi_const_XKB_KEY_braille_dots_378, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_38", (void *)_cffi_const_XKB_KEY_braille_dots_38, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_4", (void *)_cffi_const_XKB_KEY_braille_dots_4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_45", (void *)_cffi_const_XKB_KEY_braille_dots_45, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_456", (void *)_cffi_const_XKB_KEY_braille_dots_456, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_4567", (void *)_cffi_const_XKB_KEY_braille_dots_4567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_45678", (void *)_cffi_const_XKB_KEY_braille_dots_45678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_4568", (void *)_cffi_const_XKB_KEY_braille_dots_4568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_457", (void *)_cffi_const_XKB_KEY_braille_dots_457, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_4578", (void *)_cffi_const_XKB_KEY_braille_dots_4578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_458", (void *)_cffi_const_XKB_KEY_braille_dots_458, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_46", (void *)_cffi_const_XKB_KEY_braille_dots_46, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_467", (void *)_cffi_const_XKB_KEY_braille_dots_467, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_4678", (void *)_cffi_const_XKB_KEY_braille_dots_4678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_468", (void *)_cffi_const_XKB_KEY_braille_dots_468, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_47", (void *)_cffi_const_XKB_KEY_braille_dots_47, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_478", (void *)_cffi_const_XKB_KEY_braille_dots_478, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_48", (void *)_cffi_const_XKB_KEY_braille_dots_48, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_5", (void *)_cffi_const_XKB_KEY_braille_dots_5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_56", (void *)_cffi_const_XKB_KEY_braille_dots_56, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_567", (void *)_cffi_const_XKB_KEY_braille_dots_567, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_5678", (void *)_cffi_const_XKB_KEY_braille_dots_5678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_568", (void *)_cffi_const_XKB_KEY_braille_dots_568, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_57", (void *)_cffi_const_XKB_KEY_braille_dots_57, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_578", (void *)_cffi_const_XKB_KEY_braille_dots_578, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_58", (void *)_cffi_const_XKB_KEY_braille_dots_58, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_6", (void *)_cffi_const_XKB_KEY_braille_dots_6, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_67", (void *)_cffi_const_XKB_KEY_braille_dots_67, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_678", (void *)_cffi_const_XKB_KEY_braille_dots_678, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_68", (void *)_cffi_const_XKB_KEY_braille_dots_68, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_7", (void *)_cffi_const_XKB_KEY_braille_dots_7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_78", (void *)_cffi_const_XKB_KEY_braille_dots_78, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_braille_dots_8", (void *)_cffi_const_XKB_KEY_braille_dots_8, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_breve", (void *)_cffi_const_XKB_KEY_breve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_brokenbar", (void *)_cffi_const_XKB_KEY_brokenbar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_c", (void *)_cffi_const_XKB_KEY_c, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_c_h", (void *)_cffi_const_XKB_KEY_c_h, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_cabovedot", (void *)_cffi_const_XKB_KEY_cabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_cacute", (void *)_cffi_const_XKB_KEY_cacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_careof", (void *)_cffi_const_XKB_KEY_careof, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_caret", (void *)_cffi_const_XKB_KEY_caret, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_caron", (void *)_cffi_const_XKB_KEY_caron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ccaron", (void *)_cffi_const_XKB_KEY_ccaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ccedilla", (void *)_cffi_const_XKB_KEY_ccedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ccircumflex", (void *)_cffi_const_XKB_KEY_ccircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_cedilla", (void *)_cffi_const_XKB_KEY_cedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_cent", (void *)_cffi_const_XKB_KEY_cent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ch", (void *)_cffi_const_XKB_KEY_ch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_checkerboard", (void *)_cffi_const_XKB_KEY_checkerboard, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_checkmark", (void *)_cffi_const_XKB_KEY_checkmark, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_circle", (void *)_cffi_const_XKB_KEY_circle, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_club", (void *)_cffi_const_XKB_KEY_club, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_colon", (void *)_cffi_const_XKB_KEY_colon, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_comma", (void *)_cffi_const_XKB_KEY_comma, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_containsas", (void *)_cffi_const_XKB_KEY_containsas, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_copyright", (void *)_cffi_const_XKB_KEY_copyright, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_cr", (void *)_cffi_const_XKB_KEY_cr, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_crossinglines", (void *)_cffi_const_XKB_KEY_crossinglines, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_cuberoot", (void *)_cffi_const_XKB_KEY_cuberoot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_currency", (void *)_cffi_const_XKB_KEY_currency, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_cursor", (void *)_cffi_const_XKB_KEY_cursor, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_d", (void *)_cffi_const_XKB_KEY_d, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dabovedot", (void *)_cffi_const_XKB_KEY_dabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dagger", (void *)_cffi_const_XKB_KEY_dagger, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dcaron", (void *)_cffi_const_XKB_KEY_dcaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_A", (void *)_cffi_const_XKB_KEY_dead_A, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_E", (void *)_cffi_const_XKB_KEY_dead_E, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_I", (void *)_cffi_const_XKB_KEY_dead_I, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_O", (void *)_cffi_const_XKB_KEY_dead_O, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_U", (void *)_cffi_const_XKB_KEY_dead_U, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_a", (void *)_cffi_const_XKB_KEY_dead_a, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_abovecomma", (void *)_cffi_const_XKB_KEY_dead_abovecomma, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_abovedot", (void *)_cffi_const_XKB_KEY_dead_abovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_abovereversedcomma", (void *)_cffi_const_XKB_KEY_dead_abovereversedcomma, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_abovering", (void *)_cffi_const_XKB_KEY_dead_abovering, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_aboveverticalline", (void *)_cffi_const_XKB_KEY_dead_aboveverticalline, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_acute", (void *)_cffi_const_XKB_KEY_dead_acute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_belowbreve", (void *)_cffi_const_XKB_KEY_dead_belowbreve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_belowcircumflex", (void *)_cffi_const_XKB_KEY_dead_belowcircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_belowcomma", (void *)_cffi_const_XKB_KEY_dead_belowcomma, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_belowdiaeresis", (void *)_cffi_const_XKB_KEY_dead_belowdiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_belowdot", (void *)_cffi_const_XKB_KEY_dead_belowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_belowmacron", (void *)_cffi_const_XKB_KEY_dead_belowmacron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_belowring", (void *)_cffi_const_XKB_KEY_dead_belowring, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_belowtilde", (void *)_cffi_const_XKB_KEY_dead_belowtilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_belowverticalline", (void *)_cffi_const_XKB_KEY_dead_belowverticalline, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_breve", (void *)_cffi_const_XKB_KEY_dead_breve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_capital_schwa", (void *)_cffi_const_XKB_KEY_dead_capital_schwa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_caron", (void *)_cffi_const_XKB_KEY_dead_caron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_cedilla", (void *)_cffi_const_XKB_KEY_dead_cedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_circumflex", (void *)_cffi_const_XKB_KEY_dead_circumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_currency", (void *)_cffi_const_XKB_KEY_dead_currency, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_dasia", (void *)_cffi_const_XKB_KEY_dead_dasia, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_diaeresis", (void *)_cffi_const_XKB_KEY_dead_diaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_doubleacute", (void *)_cffi_const_XKB_KEY_dead_doubleacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_doublegrave", (void *)_cffi_const_XKB_KEY_dead_doublegrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_e", (void *)_cffi_const_XKB_KEY_dead_e, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_grave", (void *)_cffi_const_XKB_KEY_dead_grave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_greek", (void *)_cffi_const_XKB_KEY_dead_greek, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_hook", (void *)_cffi_const_XKB_KEY_dead_hook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_horn", (void *)_cffi_const_XKB_KEY_dead_horn, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_i", (void *)_cffi_const_XKB_KEY_dead_i, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_invertedbreve", (void *)_cffi_const_XKB_KEY_dead_invertedbreve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_iota", (void *)_cffi_const_XKB_KEY_dead_iota, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_longsolidusoverlay", (void *)_cffi_const_XKB_KEY_dead_longsolidusoverlay, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_lowline", (void *)_cffi_const_XKB_KEY_dead_lowline, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_macron", (void *)_cffi_const_XKB_KEY_dead_macron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_o", (void *)_cffi_const_XKB_KEY_dead_o, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_ogonek", (void *)_cffi_const_XKB_KEY_dead_ogonek, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_perispomeni", (void *)_cffi_const_XKB_KEY_dead_perispomeni, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_psili", (void *)_cffi_const_XKB_KEY_dead_psili, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_semivoiced_sound", (void *)_cffi_const_XKB_KEY_dead_semivoiced_sound, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_small_schwa", (void *)_cffi_const_XKB_KEY_dead_small_schwa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_stroke", (void *)_cffi_const_XKB_KEY_dead_stroke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_tilde", (void *)_cffi_const_XKB_KEY_dead_tilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_u", (void *)_cffi_const_XKB_KEY_dead_u, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dead_voiced_sound", (void *)_cffi_const_XKB_KEY_dead_voiced_sound, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_decimalpoint", (void *)_cffi_const_XKB_KEY_decimalpoint, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_degree", (void *)_cffi_const_XKB_KEY_degree, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_diaeresis", (void *)_cffi_const_XKB_KEY_diaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_diamond", (void *)_cffi_const_XKB_KEY_diamond, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_digitspace", (void *)_cffi_const_XKB_KEY_digitspace, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dintegral", (void *)_cffi_const_XKB_KEY_dintegral, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_division", (void *)_cffi_const_XKB_KEY_division, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dollar", (void *)_cffi_const_XKB_KEY_dollar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_doubbaselinedot", (void *)_cffi_const_XKB_KEY_doubbaselinedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_doubleacute", (void *)_cffi_const_XKB_KEY_doubleacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_doubledagger", (void *)_cffi_const_XKB_KEY_doubledagger, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_doublelowquotemark", (void *)_cffi_const_XKB_KEY_doublelowquotemark, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_downarrow", (void *)_cffi_const_XKB_KEY_downarrow, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_downcaret", (void *)_cffi_const_XKB_KEY_downcaret, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_downshoe", (void *)_cffi_const_XKB_KEY_downshoe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_downstile", (void *)_cffi_const_XKB_KEY_downstile, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_downtack", (void *)_cffi_const_XKB_KEY_downtack, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_dstroke", (void *)_cffi_const_XKB_KEY_dstroke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_e", (void *)_cffi_const_XKB_KEY_e, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_eabovedot", (void *)_cffi_const_XKB_KEY_eabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_eacute", (void *)_cffi_const_XKB_KEY_eacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ebelowdot", (void *)_cffi_const_XKB_KEY_ebelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ecaron", (void *)_cffi_const_XKB_KEY_ecaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ecircumflex", (void *)_cffi_const_XKB_KEY_ecircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ecircumflexacute", (void *)_cffi_const_XKB_KEY_ecircumflexacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ecircumflexbelowdot", (void *)_cffi_const_XKB_KEY_ecircumflexbelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ecircumflexgrave", (void *)_cffi_const_XKB_KEY_ecircumflexgrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ecircumflexhook", (void *)_cffi_const_XKB_KEY_ecircumflexhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ecircumflextilde", (void *)_cffi_const_XKB_KEY_ecircumflextilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ediaeresis", (void *)_cffi_const_XKB_KEY_ediaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_egrave", (void *)_cffi_const_XKB_KEY_egrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ehook", (void *)_cffi_const_XKB_KEY_ehook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_eightsubscript", (void *)_cffi_const_XKB_KEY_eightsubscript, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_eightsuperior", (void *)_cffi_const_XKB_KEY_eightsuperior, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_elementof", (void *)_cffi_const_XKB_KEY_elementof, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ellipsis", (void *)_cffi_const_XKB_KEY_ellipsis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_em3space", (void *)_cffi_const_XKB_KEY_em3space, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_em4space", (void *)_cffi_const_XKB_KEY_em4space, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_emacron", (void *)_cffi_const_XKB_KEY_emacron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_emdash", (void *)_cffi_const_XKB_KEY_emdash, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_emfilledcircle", (void *)_cffi_const_XKB_KEY_emfilledcircle, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_emfilledrect", (void *)_cffi_const_XKB_KEY_emfilledrect, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_emopencircle", (void *)_cffi_const_XKB_KEY_emopencircle, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_emopenrectangle", (void *)_cffi_const_XKB_KEY_emopenrectangle, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_emptyset", (void *)_cffi_const_XKB_KEY_emptyset, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_emspace", (void *)_cffi_const_XKB_KEY_emspace, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_endash", (void *)_cffi_const_XKB_KEY_endash, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_enfilledcircbullet", (void *)_cffi_const_XKB_KEY_enfilledcircbullet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_enfilledsqbullet", (void *)_cffi_const_XKB_KEY_enfilledsqbullet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_eng", (void *)_cffi_const_XKB_KEY_eng, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_enopencircbullet", (void *)_cffi_const_XKB_KEY_enopencircbullet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_enopensquarebullet", (void *)_cffi_const_XKB_KEY_enopensquarebullet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_enspace", (void *)_cffi_const_XKB_KEY_enspace, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_eogonek", (void *)_cffi_const_XKB_KEY_eogonek, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_equal", (void *)_cffi_const_XKB_KEY_equal, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_eth", (void *)_cffi_const_XKB_KEY_eth, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_etilde", (void *)_cffi_const_XKB_KEY_etilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_exclam", (void *)_cffi_const_XKB_KEY_exclam, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_exclamdown", (void *)_cffi_const_XKB_KEY_exclamdown, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ezh", (void *)_cffi_const_XKB_KEY_ezh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_f", (void *)_cffi_const_XKB_KEY_f, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_fabovedot", (void *)_cffi_const_XKB_KEY_fabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_femalesymbol", (void *)_cffi_const_XKB_KEY_femalesymbol, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ff", (void *)_cffi_const_XKB_KEY_ff, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_figdash", (void *)_cffi_const_XKB_KEY_figdash, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_filledlefttribullet", (void *)_cffi_const_XKB_KEY_filledlefttribullet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_filledrectbullet", (void *)_cffi_const_XKB_KEY_filledrectbullet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_filledrighttribullet", (void *)_cffi_const_XKB_KEY_filledrighttribullet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_filledtribulletdown", (void *)_cffi_const_XKB_KEY_filledtribulletdown, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_filledtribulletup", (void *)_cffi_const_XKB_KEY_filledtribulletup, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_fiveeighths", (void *)_cffi_const_XKB_KEY_fiveeighths, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_fivesixths", (void *)_cffi_const_XKB_KEY_fivesixths, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_fivesubscript", (void *)_cffi_const_XKB_KEY_fivesubscript, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_fivesuperior", (void *)_cffi_const_XKB_KEY_fivesuperior, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_fourfifths", (void *)_cffi_const_XKB_KEY_fourfifths, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_foursubscript", (void *)_cffi_const_XKB_KEY_foursubscript, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_foursuperior", (void *)_cffi_const_XKB_KEY_foursuperior, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_fourthroot", (void *)_cffi_const_XKB_KEY_fourthroot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_function", (void *)_cffi_const_XKB_KEY_function, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_g", (void *)_cffi_const_XKB_KEY_g, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_gabovedot", (void *)_cffi_const_XKB_KEY_gabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_gbreve", (void *)_cffi_const_XKB_KEY_gbreve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_gcaron", (void *)_cffi_const_XKB_KEY_gcaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_gcedilla", (void *)_cffi_const_XKB_KEY_gcedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_gcircumflex", (void *)_cffi_const_XKB_KEY_gcircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_grave", (void *)_cffi_const_XKB_KEY_grave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_greater", (void *)_cffi_const_XKB_KEY_greater, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_greaterthanequal", (void *)_cffi_const_XKB_KEY_greaterthanequal, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_guilder", (void *)_cffi_const_XKB_KEY_guilder, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_guillemotleft", (void *)_cffi_const_XKB_KEY_guillemotleft, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_guillemotright", (void *)_cffi_const_XKB_KEY_guillemotright, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_h", (void *)_cffi_const_XKB_KEY_h, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hairspace", (void *)_cffi_const_XKB_KEY_hairspace, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hcircumflex", (void *)_cffi_const_XKB_KEY_hcircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_heart", (void *)_cffi_const_XKB_KEY_heart, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_aleph", (void *)_cffi_const_XKB_KEY_hebrew_aleph, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_ayin", (void *)_cffi_const_XKB_KEY_hebrew_ayin, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_bet", (void *)_cffi_const_XKB_KEY_hebrew_bet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_beth", (void *)_cffi_const_XKB_KEY_hebrew_beth, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_chet", (void *)_cffi_const_XKB_KEY_hebrew_chet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_dalet", (void *)_cffi_const_XKB_KEY_hebrew_dalet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_daleth", (void *)_cffi_const_XKB_KEY_hebrew_daleth, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_doublelowline", (void *)_cffi_const_XKB_KEY_hebrew_doublelowline, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_finalkaph", (void *)_cffi_const_XKB_KEY_hebrew_finalkaph, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_finalmem", (void *)_cffi_const_XKB_KEY_hebrew_finalmem, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_finalnun", (void *)_cffi_const_XKB_KEY_hebrew_finalnun, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_finalpe", (void *)_cffi_const_XKB_KEY_hebrew_finalpe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_finalzade", (void *)_cffi_const_XKB_KEY_hebrew_finalzade, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_finalzadi", (void *)_cffi_const_XKB_KEY_hebrew_finalzadi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_gimel", (void *)_cffi_const_XKB_KEY_hebrew_gimel, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_gimmel", (void *)_cffi_const_XKB_KEY_hebrew_gimmel, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_he", (void *)_cffi_const_XKB_KEY_hebrew_he, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_het", (void *)_cffi_const_XKB_KEY_hebrew_het, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_kaph", (void *)_cffi_const_XKB_KEY_hebrew_kaph, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_kuf", (void *)_cffi_const_XKB_KEY_hebrew_kuf, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_lamed", (void *)_cffi_const_XKB_KEY_hebrew_lamed, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_mem", (void *)_cffi_const_XKB_KEY_hebrew_mem, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_nun", (void *)_cffi_const_XKB_KEY_hebrew_nun, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_pe", (void *)_cffi_const_XKB_KEY_hebrew_pe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_qoph", (void *)_cffi_const_XKB_KEY_hebrew_qoph, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_resh", (void *)_cffi_const_XKB_KEY_hebrew_resh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_samech", (void *)_cffi_const_XKB_KEY_hebrew_samech, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_samekh", (void *)_cffi_const_XKB_KEY_hebrew_samekh, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_shin", (void *)_cffi_const_XKB_KEY_hebrew_shin, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_taf", (void *)_cffi_const_XKB_KEY_hebrew_taf, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_taw", (void *)_cffi_const_XKB_KEY_hebrew_taw, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_tet", (void *)_cffi_const_XKB_KEY_hebrew_tet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_teth", (void *)_cffi_const_XKB_KEY_hebrew_teth, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_waw", (void *)_cffi_const_XKB_KEY_hebrew_waw, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_yod", (void *)_cffi_const_XKB_KEY_hebrew_yod, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_zade", (void *)_cffi_const_XKB_KEY_hebrew_zade, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_zadi", (void *)_cffi_const_XKB_KEY_hebrew_zadi, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_zain", (void *)_cffi_const_XKB_KEY_hebrew_zain, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hebrew_zayin", (void *)_cffi_const_XKB_KEY_hebrew_zayin, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hexagram", (void *)_cffi_const_XKB_KEY_hexagram, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_horizconnector", (void *)_cffi_const_XKB_KEY_horizconnector, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_horizlinescan1", (void *)_cffi_const_XKB_KEY_horizlinescan1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_horizlinescan3", (void *)_cffi_const_XKB_KEY_horizlinescan3, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_horizlinescan5", (void *)_cffi_const_XKB_KEY_horizlinescan5, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_horizlinescan7", (void *)_cffi_const_XKB_KEY_horizlinescan7, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_horizlinescan9", (void *)_cffi_const_XKB_KEY_horizlinescan9, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpBackTab", (void *)_cffi_const_XKB_KEY_hpBackTab, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpClearLine", (void *)_cffi_const_XKB_KEY_hpClearLine, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpDeleteChar", (void *)_cffi_const_XKB_KEY_hpDeleteChar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpDeleteLine", (void *)_cffi_const_XKB_KEY_hpDeleteLine, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpIO", (void *)_cffi_const_XKB_KEY_hpIO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpInsertChar", (void *)_cffi_const_XKB_KEY_hpInsertChar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpInsertLine", (void *)_cffi_const_XKB_KEY_hpInsertLine, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpKP_BackTab", (void *)_cffi_const_XKB_KEY_hpKP_BackTab, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpModelock1", (void *)_cffi_const_XKB_KEY_hpModelock1, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpModelock2", (void *)_cffi_const_XKB_KEY_hpModelock2, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpReset", (void *)_cffi_const_XKB_KEY_hpReset, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpSystem", (void *)_cffi_const_XKB_KEY_hpSystem, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpUser", (void *)_cffi_const_XKB_KEY_hpUser, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpYdiaeresis", (void *)_cffi_const_XKB_KEY_hpYdiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpblock", (void *)_cffi_const_XKB_KEY_hpblock, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpguilder", (void *)_cffi_const_XKB_KEY_hpguilder, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hplira", (void *)_cffi_const_XKB_KEY_hplira, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hplongminus", (void *)_cffi_const_XKB_KEY_hplongminus, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpmute_acute", (void *)_cffi_const_XKB_KEY_hpmute_acute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpmute_asciicircum", (void *)_cffi_const_XKB_KEY_hpmute_asciicircum, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpmute_asciitilde", (void *)_cffi_const_XKB_KEY_hpmute_asciitilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpmute_diaeresis", (void *)_cffi_const_XKB_KEY_hpmute_diaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hpmute_grave", (void *)_cffi_const_XKB_KEY_hpmute_grave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hstroke", (void *)_cffi_const_XKB_KEY_hstroke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ht", (void *)_cffi_const_XKB_KEY_ht, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_hyphen", (void *)_cffi_const_XKB_KEY_hyphen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_i", (void *)_cffi_const_XKB_KEY_i, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_iacute", (void *)_cffi_const_XKB_KEY_iacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ibelowdot", (void *)_cffi_const_XKB_KEY_ibelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ibreve", (void *)_cffi_const_XKB_KEY_ibreve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_icircumflex", (void *)_cffi_const_XKB_KEY_icircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_identical", (void *)_cffi_const_XKB_KEY_identical, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_idiaeresis", (void *)_cffi_const_XKB_KEY_idiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_idotless", (void *)_cffi_const_XKB_KEY_idotless, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ifonlyif", (void *)_cffi_const_XKB_KEY_ifonlyif, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_igrave", (void *)_cffi_const_XKB_KEY_igrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ihook", (void *)_cffi_const_XKB_KEY_ihook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_imacron", (void *)_cffi_const_XKB_KEY_imacron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_implies", (void *)_cffi_const_XKB_KEY_implies, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_includedin", (void *)_cffi_const_XKB_KEY_includedin, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_includes", (void *)_cffi_const_XKB_KEY_includes, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_infinity", (void *)_cffi_const_XKB_KEY_infinity, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_integral", (void *)_cffi_const_XKB_KEY_integral, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_intersection", (void *)_cffi_const_XKB_KEY_intersection, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_iogonek", (void *)_cffi_const_XKB_KEY_iogonek, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_itilde", (void *)_cffi_const_XKB_KEY_itilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_j", (void *)_cffi_const_XKB_KEY_j, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_jcircumflex", (void *)_cffi_const_XKB_KEY_jcircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_jot", (void *)_cffi_const_XKB_KEY_jot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_k", (void *)_cffi_const_XKB_KEY_k, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_A", (void *)_cffi_const_XKB_KEY_kana_A, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_CHI", (void *)_cffi_const_XKB_KEY_kana_CHI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_E", (void *)_cffi_const_XKB_KEY_kana_E, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_FU", (void *)_cffi_const_XKB_KEY_kana_FU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_HA", (void *)_cffi_const_XKB_KEY_kana_HA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_HE", (void *)_cffi_const_XKB_KEY_kana_HE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_HI", (void *)_cffi_const_XKB_KEY_kana_HI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_HO", (void *)_cffi_const_XKB_KEY_kana_HO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_HU", (void *)_cffi_const_XKB_KEY_kana_HU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_I", (void *)_cffi_const_XKB_KEY_kana_I, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_KA", (void *)_cffi_const_XKB_KEY_kana_KA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_KE", (void *)_cffi_const_XKB_KEY_kana_KE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_KI", (void *)_cffi_const_XKB_KEY_kana_KI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_KO", (void *)_cffi_const_XKB_KEY_kana_KO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_KU", (void *)_cffi_const_XKB_KEY_kana_KU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_MA", (void *)_cffi_const_XKB_KEY_kana_MA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_ME", (void *)_cffi_const_XKB_KEY_kana_ME, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_MI", (void *)_cffi_const_XKB_KEY_kana_MI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_MO", (void *)_cffi_const_XKB_KEY_kana_MO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_MU", (void *)_cffi_const_XKB_KEY_kana_MU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_N", (void *)_cffi_const_XKB_KEY_kana_N, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_NA", (void *)_cffi_const_XKB_KEY_kana_NA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_NE", (void *)_cffi_const_XKB_KEY_kana_NE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_NI", (void *)_cffi_const_XKB_KEY_kana_NI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_NO", (void *)_cffi_const_XKB_KEY_kana_NO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_NU", (void *)_cffi_const_XKB_KEY_kana_NU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_O", (void *)_cffi_const_XKB_KEY_kana_O, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_RA", (void *)_cffi_const_XKB_KEY_kana_RA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_RE", (void *)_cffi_const_XKB_KEY_kana_RE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_RI", (void *)_cffi_const_XKB_KEY_kana_RI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_RO", (void *)_cffi_const_XKB_KEY_kana_RO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_RU", (void *)_cffi_const_XKB_KEY_kana_RU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_SA", (void *)_cffi_const_XKB_KEY_kana_SA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_SE", (void *)_cffi_const_XKB_KEY_kana_SE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_SHI", (void *)_cffi_const_XKB_KEY_kana_SHI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_SO", (void *)_cffi_const_XKB_KEY_kana_SO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_SU", (void *)_cffi_const_XKB_KEY_kana_SU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_TA", (void *)_cffi_const_XKB_KEY_kana_TA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_TE", (void *)_cffi_const_XKB_KEY_kana_TE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_TI", (void *)_cffi_const_XKB_KEY_kana_TI, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_TO", (void *)_cffi_const_XKB_KEY_kana_TO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_TSU", (void *)_cffi_const_XKB_KEY_kana_TSU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_TU", (void *)_cffi_const_XKB_KEY_kana_TU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_U", (void *)_cffi_const_XKB_KEY_kana_U, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_WA", (void *)_cffi_const_XKB_KEY_kana_WA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_WO", (void *)_cffi_const_XKB_KEY_kana_WO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_YA", (void *)_cffi_const_XKB_KEY_kana_YA, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_YO", (void *)_cffi_const_XKB_KEY_kana_YO, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_YU", (void *)_cffi_const_XKB_KEY_kana_YU, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_a", (void *)_cffi_const_XKB_KEY_kana_a, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_closingbracket", (void *)_cffi_const_XKB_KEY_kana_closingbracket, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_comma", (void *)_cffi_const_XKB_KEY_kana_comma, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_conjunctive", (void *)_cffi_const_XKB_KEY_kana_conjunctive, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_e", (void *)_cffi_const_XKB_KEY_kana_e, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_fullstop", (void *)_cffi_const_XKB_KEY_kana_fullstop, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_i", (void *)_cffi_const_XKB_KEY_kana_i, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_middledot", (void *)_cffi_const_XKB_KEY_kana_middledot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_o", (void *)_cffi_const_XKB_KEY_kana_o, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_openingbracket", (void *)_cffi_const_XKB_KEY_kana_openingbracket, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_switch", (void *)_cffi_const_XKB_KEY_kana_switch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_tsu", (void *)_cffi_const_XKB_KEY_kana_tsu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_tu", (void *)_cffi_const_XKB_KEY_kana_tu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_u", (void *)_cffi_const_XKB_KEY_kana_u, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_ya", (void *)_cffi_const_XKB_KEY_kana_ya, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_yo", (void *)_cffi_const_XKB_KEY_kana_yo, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kana_yu", (void *)_cffi_const_XKB_KEY_kana_yu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kappa", (void *)_cffi_const_XKB_KEY_kappa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kcedilla", (void *)_cffi_const_XKB_KEY_kcedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_kra", (void *)_cffi_const_XKB_KEY_kra, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_l", (void *)_cffi_const_XKB_KEY_l, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_lacute", (void *)_cffi_const_XKB_KEY_lacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_latincross", (void *)_cffi_const_XKB_KEY_latincross, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_lbelowdot", (void *)_cffi_const_XKB_KEY_lbelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_lcaron", (void *)_cffi_const_XKB_KEY_lcaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_lcedilla", (void *)_cffi_const_XKB_KEY_lcedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_leftanglebracket", (void *)_cffi_const_XKB_KEY_leftanglebracket, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_leftarrow", (void *)_cffi_const_XKB_KEY_leftarrow, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_leftcaret", (void *)_cffi_const_XKB_KEY_leftcaret, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_leftdoublequotemark", (void *)_cffi_const_XKB_KEY_leftdoublequotemark, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_leftmiddlecurlybrace", (void *)_cffi_const_XKB_KEY_leftmiddlecurlybrace, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_leftopentriangle", (void *)_cffi_const_XKB_KEY_leftopentriangle, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_leftpointer", (void *)_cffi_const_XKB_KEY_leftpointer, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_leftradical", (void *)_cffi_const_XKB_KEY_leftradical, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_leftshoe", (void *)_cffi_const_XKB_KEY_leftshoe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_leftsinglequotemark", (void *)_cffi_const_XKB_KEY_leftsinglequotemark, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_leftt", (void *)_cffi_const_XKB_KEY_leftt, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_lefttack", (void *)_cffi_const_XKB_KEY_lefttack, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_less", (void *)_cffi_const_XKB_KEY_less, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_lessthanequal", (void *)_cffi_const_XKB_KEY_lessthanequal, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_lf", (void *)_cffi_const_XKB_KEY_lf, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_lira", (void *)_cffi_const_XKB_KEY_lira, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_logicaland", (void *)_cffi_const_XKB_KEY_logicaland, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_logicalor", (void *)_cffi_const_XKB_KEY_logicalor, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_longminus", (void *)_cffi_const_XKB_KEY_longminus, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_lowleftcorner", (void *)_cffi_const_XKB_KEY_lowleftcorner, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_lowrightcorner", (void *)_cffi_const_XKB_KEY_lowrightcorner, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_lstroke", (void *)_cffi_const_XKB_KEY_lstroke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_m", (void *)_cffi_const_XKB_KEY_m, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_mabovedot", (void *)_cffi_const_XKB_KEY_mabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_macron", (void *)_cffi_const_XKB_KEY_macron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_malesymbol", (void *)_cffi_const_XKB_KEY_malesymbol, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_maltesecross", (void *)_cffi_const_XKB_KEY_maltesecross, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_marker", (void *)_cffi_const_XKB_KEY_marker, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_masculine", (void *)_cffi_const_XKB_KEY_masculine, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_minus", (void *)_cffi_const_XKB_KEY_minus, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_minutes", (void *)_cffi_const_XKB_KEY_minutes, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_mu", (void *)_cffi_const_XKB_KEY_mu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_multiply", (void *)_cffi_const_XKB_KEY_multiply, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_musicalflat", (void *)_cffi_const_XKB_KEY_musicalflat, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_musicalsharp", (void *)_cffi_const_XKB_KEY_musicalsharp, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_mute_acute", (void *)_cffi_const_XKB_KEY_mute_acute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_mute_asciicircum", (void *)_cffi_const_XKB_KEY_mute_asciicircum, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_mute_asciitilde", (void *)_cffi_const_XKB_KEY_mute_asciitilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_mute_diaeresis", (void *)_cffi_const_XKB_KEY_mute_diaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_mute_grave", (void *)_cffi_const_XKB_KEY_mute_grave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_n", (void *)_cffi_const_XKB_KEY_n, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_nabla", (void *)_cffi_const_XKB_KEY_nabla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_nacute", (void *)_cffi_const_XKB_KEY_nacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ncaron", (void *)_cffi_const_XKB_KEY_ncaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ncedilla", (void *)_cffi_const_XKB_KEY_ncedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ninesubscript", (void *)_cffi_const_XKB_KEY_ninesubscript, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ninesuperior", (void *)_cffi_const_XKB_KEY_ninesuperior, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_nl", (void *)_cffi_const_XKB_KEY_nl, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_nobreakspace", (void *)_cffi_const_XKB_KEY_nobreakspace, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_notapproxeq", (void *)_cffi_const_XKB_KEY_notapproxeq, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_notelementof", (void *)_cffi_const_XKB_KEY_notelementof, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_notequal", (void *)_cffi_const_XKB_KEY_notequal, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_notidentical", (void *)_cffi_const_XKB_KEY_notidentical, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_notsign", (void *)_cffi_const_XKB_KEY_notsign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ntilde", (void *)_cffi_const_XKB_KEY_ntilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_numbersign", (void *)_cffi_const_XKB_KEY_numbersign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_numerosign", (void *)_cffi_const_XKB_KEY_numerosign, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_o", (void *)_cffi_const_XKB_KEY_o, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_oacute", (void *)_cffi_const_XKB_KEY_oacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_obarred", (void *)_cffi_const_XKB_KEY_obarred, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_obelowdot", (void *)_cffi_const_XKB_KEY_obelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ocaron", (void *)_cffi_const_XKB_KEY_ocaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ocircumflex", (void *)_cffi_const_XKB_KEY_ocircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ocircumflexacute", (void *)_cffi_const_XKB_KEY_ocircumflexacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ocircumflexbelowdot", (void *)_cffi_const_XKB_KEY_ocircumflexbelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ocircumflexgrave", (void *)_cffi_const_XKB_KEY_ocircumflexgrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ocircumflexhook", (void *)_cffi_const_XKB_KEY_ocircumflexhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ocircumflextilde", (void *)_cffi_const_XKB_KEY_ocircumflextilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_odiaeresis", (void *)_cffi_const_XKB_KEY_odiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_odoubleacute", (void *)_cffi_const_XKB_KEY_odoubleacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_oe", (void *)_cffi_const_XKB_KEY_oe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ogonek", (void *)_cffi_const_XKB_KEY_ogonek, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ograve", (void *)_cffi_const_XKB_KEY_ograve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ohook", (void *)_cffi_const_XKB_KEY_ohook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ohorn", (void *)_cffi_const_XKB_KEY_ohorn, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ohornacute", (void *)_cffi_const_XKB_KEY_ohornacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ohornbelowdot", (void *)_cffi_const_XKB_KEY_ohornbelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ohorngrave", (void *)_cffi_const_XKB_KEY_ohorngrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ohornhook", (void *)_cffi_const_XKB_KEY_ohornhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ohorntilde", (void *)_cffi_const_XKB_KEY_ohorntilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_omacron", (void *)_cffi_const_XKB_KEY_omacron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_oneeighth", (void *)_cffi_const_XKB_KEY_oneeighth, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_onefifth", (void *)_cffi_const_XKB_KEY_onefifth, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_onehalf", (void *)_cffi_const_XKB_KEY_onehalf, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_onequarter", (void *)_cffi_const_XKB_KEY_onequarter, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_onesixth", (void *)_cffi_const_XKB_KEY_onesixth, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_onesubscript", (void *)_cffi_const_XKB_KEY_onesubscript, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_onesuperior", (void *)_cffi_const_XKB_KEY_onesuperior, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_onethird", (void *)_cffi_const_XKB_KEY_onethird, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ooblique", (void *)_cffi_const_XKB_KEY_ooblique, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_openrectbullet", (void *)_cffi_const_XKB_KEY_openrectbullet, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_openstar", (void *)_cffi_const_XKB_KEY_openstar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_opentribulletdown", (void *)_cffi_const_XKB_KEY_opentribulletdown, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_opentribulletup", (void *)_cffi_const_XKB_KEY_opentribulletup, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ordfeminine", (void *)_cffi_const_XKB_KEY_ordfeminine, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfActivate", (void *)_cffi_const_XKB_KEY_osfActivate, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfAddMode", (void *)_cffi_const_XKB_KEY_osfAddMode, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfBackSpace", (void *)_cffi_const_XKB_KEY_osfBackSpace, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfBackTab", (void *)_cffi_const_XKB_KEY_osfBackTab, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfBeginData", (void *)_cffi_const_XKB_KEY_osfBeginData, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfBeginLine", (void *)_cffi_const_XKB_KEY_osfBeginLine, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfCancel", (void *)_cffi_const_XKB_KEY_osfCancel, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfClear", (void *)_cffi_const_XKB_KEY_osfClear, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfCopy", (void *)_cffi_const_XKB_KEY_osfCopy, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfCut", (void *)_cffi_const_XKB_KEY_osfCut, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfDelete", (void *)_cffi_const_XKB_KEY_osfDelete, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfDeselectAll", (void *)_cffi_const_XKB_KEY_osfDeselectAll, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfDown", (void *)_cffi_const_XKB_KEY_osfDown, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfEndData", (void *)_cffi_const_XKB_KEY_osfEndData, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfEndLine", (void *)_cffi_const_XKB_KEY_osfEndLine, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfEscape", (void *)_cffi_const_XKB_KEY_osfEscape, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfExtend", (void *)_cffi_const_XKB_KEY_osfExtend, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfHelp", (void *)_cffi_const_XKB_KEY_osfHelp, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfInsert", (void *)_cffi_const_XKB_KEY_osfInsert, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfLeft", (void *)_cffi_const_XKB_KEY_osfLeft, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfMenu", (void *)_cffi_const_XKB_KEY_osfMenu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfMenuBar", (void *)_cffi_const_XKB_KEY_osfMenuBar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfNextField", (void *)_cffi_const_XKB_KEY_osfNextField, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfNextMenu", (void *)_cffi_const_XKB_KEY_osfNextMenu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfPageDown", (void *)_cffi_const_XKB_KEY_osfPageDown, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfPageLeft", (void *)_cffi_const_XKB_KEY_osfPageLeft, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfPageRight", (void *)_cffi_const_XKB_KEY_osfPageRight, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfPageUp", (void *)_cffi_const_XKB_KEY_osfPageUp, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfPaste", (void *)_cffi_const_XKB_KEY_osfPaste, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfPrevField", (void *)_cffi_const_XKB_KEY_osfPrevField, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfPrevMenu", (void *)_cffi_const_XKB_KEY_osfPrevMenu, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfPrimaryPaste", (void *)_cffi_const_XKB_KEY_osfPrimaryPaste, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfQuickPaste", (void *)_cffi_const_XKB_KEY_osfQuickPaste, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfReselect", (void *)_cffi_const_XKB_KEY_osfReselect, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfRestore", (void *)_cffi_const_XKB_KEY_osfRestore, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfRight", (void *)_cffi_const_XKB_KEY_osfRight, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfSelect", (void *)_cffi_const_XKB_KEY_osfSelect, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfSelectAll", (void *)_cffi_const_XKB_KEY_osfSelectAll, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfUndo", (void *)_cffi_const_XKB_KEY_osfUndo, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_osfUp", (void *)_cffi_const_XKB_KEY_osfUp, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_oslash", (void *)_cffi_const_XKB_KEY_oslash, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_otilde", (void *)_cffi_const_XKB_KEY_otilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_overbar", (void *)_cffi_const_XKB_KEY_overbar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_overline", (void *)_cffi_const_XKB_KEY_overline, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_p", (void *)_cffi_const_XKB_KEY_p, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_pabovedot", (void *)_cffi_const_XKB_KEY_pabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_paragraph", (void *)_cffi_const_XKB_KEY_paragraph, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_parenleft", (void *)_cffi_const_XKB_KEY_parenleft, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_parenright", (void *)_cffi_const_XKB_KEY_parenright, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_partdifferential", (void *)_cffi_const_XKB_KEY_partdifferential, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_partialderivative", (void *)_cffi_const_XKB_KEY_partialderivative, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_percent", (void *)_cffi_const_XKB_KEY_percent, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_period", (void *)_cffi_const_XKB_KEY_period, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_periodcentered", (void *)_cffi_const_XKB_KEY_periodcentered, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_permille", (void *)_cffi_const_XKB_KEY_permille, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_phonographcopyright", (void *)_cffi_const_XKB_KEY_phonographcopyright, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_plus", (void *)_cffi_const_XKB_KEY_plus, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_plusminus", (void *)_cffi_const_XKB_KEY_plusminus, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_prescription", (void *)_cffi_const_XKB_KEY_prescription, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_prolongedsound", (void *)_cffi_const_XKB_KEY_prolongedsound, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_punctspace", (void *)_cffi_const_XKB_KEY_punctspace, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_q", (void *)_cffi_const_XKB_KEY_q, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_quad", (void *)_cffi_const_XKB_KEY_quad, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_question", (void *)_cffi_const_XKB_KEY_question, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_questiondown", (void *)_cffi_const_XKB_KEY_questiondown, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_quotedbl", (void *)_cffi_const_XKB_KEY_quotedbl, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_quoteleft", (void *)_cffi_const_XKB_KEY_quoteleft, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_quoteright", (void *)_cffi_const_XKB_KEY_quoteright, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_r", (void *)_cffi_const_XKB_KEY_r, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_racute", (void *)_cffi_const_XKB_KEY_racute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_radical", (void *)_cffi_const_XKB_KEY_radical, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_rcaron", (void *)_cffi_const_XKB_KEY_rcaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_rcedilla", (void *)_cffi_const_XKB_KEY_rcedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_registered", (void *)_cffi_const_XKB_KEY_registered, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_rightanglebracket", (void *)_cffi_const_XKB_KEY_rightanglebracket, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_rightarrow", (void *)_cffi_const_XKB_KEY_rightarrow, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_rightcaret", (void *)_cffi_const_XKB_KEY_rightcaret, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_rightdoublequotemark", (void *)_cffi_const_XKB_KEY_rightdoublequotemark, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_rightmiddlecurlybrace", (void *)_cffi_const_XKB_KEY_rightmiddlecurlybrace, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_rightmiddlesummation", (void *)_cffi_const_XKB_KEY_rightmiddlesummation, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_rightopentriangle", (void *)_cffi_const_XKB_KEY_rightopentriangle, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_rightpointer", (void *)_cffi_const_XKB_KEY_rightpointer, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_rightshoe", (void *)_cffi_const_XKB_KEY_rightshoe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_rightsinglequotemark", (void *)_cffi_const_XKB_KEY_rightsinglequotemark, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_rightt", (void *)_cffi_const_XKB_KEY_rightt, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_righttack", (void *)_cffi_const_XKB_KEY_righttack, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_s", (void *)_cffi_const_XKB_KEY_s, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_sabovedot", (void *)_cffi_const_XKB_KEY_sabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_sacute", (void *)_cffi_const_XKB_KEY_sacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_scaron", (void *)_cffi_const_XKB_KEY_scaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_scedilla", (void *)_cffi_const_XKB_KEY_scedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_schwa", (void *)_cffi_const_XKB_KEY_schwa, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_scircumflex", (void *)_cffi_const_XKB_KEY_scircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_script_switch", (void *)_cffi_const_XKB_KEY_script_switch, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_seconds", (void *)_cffi_const_XKB_KEY_seconds, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_section", (void *)_cffi_const_XKB_KEY_section, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_semicolon", (void *)_cffi_const_XKB_KEY_semicolon, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_semivoicedsound", (void *)_cffi_const_XKB_KEY_semivoicedsound, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_seveneighths", (void *)_cffi_const_XKB_KEY_seveneighths, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_sevensubscript", (void *)_cffi_const_XKB_KEY_sevensubscript, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_sevensuperior", (void *)_cffi_const_XKB_KEY_sevensuperior, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_signaturemark", (void *)_cffi_const_XKB_KEY_signaturemark, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_signifblank", (void *)_cffi_const_XKB_KEY_signifblank, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_similarequal", (void *)_cffi_const_XKB_KEY_similarequal, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_singlelowquotemark", (void *)_cffi_const_XKB_KEY_singlelowquotemark, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_sixsubscript", (void *)_cffi_const_XKB_KEY_sixsubscript, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_sixsuperior", (void *)_cffi_const_XKB_KEY_sixsuperior, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_slash", (void *)_cffi_const_XKB_KEY_slash, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_soliddiamond", (void *)_cffi_const_XKB_KEY_soliddiamond, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_space", (void *)_cffi_const_XKB_KEY_space, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_squareroot", (void *)_cffi_const_XKB_KEY_squareroot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ssharp", (void *)_cffi_const_XKB_KEY_ssharp, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_sterling", (void *)_cffi_const_XKB_KEY_sterling, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_stricteq", (void *)_cffi_const_XKB_KEY_stricteq, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_t", (void *)_cffi_const_XKB_KEY_t, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_tabovedot", (void *)_cffi_const_XKB_KEY_tabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_tcaron", (void *)_cffi_const_XKB_KEY_tcaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_tcedilla", (void *)_cffi_const_XKB_KEY_tcedilla, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_telephone", (void *)_cffi_const_XKB_KEY_telephone, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_telephonerecorder", (void *)_cffi_const_XKB_KEY_telephonerecorder, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_therefore", (void *)_cffi_const_XKB_KEY_therefore, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_thinspace", (void *)_cffi_const_XKB_KEY_thinspace, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_thorn", (void *)_cffi_const_XKB_KEY_thorn, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_threeeighths", (void *)_cffi_const_XKB_KEY_threeeighths, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_threefifths", (void *)_cffi_const_XKB_KEY_threefifths, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_threequarters", (void *)_cffi_const_XKB_KEY_threequarters, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_threesubscript", (void *)_cffi_const_XKB_KEY_threesubscript, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_threesuperior", (void *)_cffi_const_XKB_KEY_threesuperior, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_tintegral", (void *)_cffi_const_XKB_KEY_tintegral, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_topintegral", (void *)_cffi_const_XKB_KEY_topintegral, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_topleftparens", (void *)_cffi_const_XKB_KEY_topleftparens, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_topleftradical", (void *)_cffi_const_XKB_KEY_topleftradical, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_topleftsqbracket", (void *)_cffi_const_XKB_KEY_topleftsqbracket, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_topleftsummation", (void *)_cffi_const_XKB_KEY_topleftsummation, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_toprightparens", (void *)_cffi_const_XKB_KEY_toprightparens, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_toprightsqbracket", (void *)_cffi_const_XKB_KEY_toprightsqbracket, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_toprightsummation", (void *)_cffi_const_XKB_KEY_toprightsummation, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_topt", (void *)_cffi_const_XKB_KEY_topt, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_topvertsummationconnector", (void *)_cffi_const_XKB_KEY_topvertsummationconnector, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_trademark", (void *)_cffi_const_XKB_KEY_trademark, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_trademarkincircle", (void *)_cffi_const_XKB_KEY_trademarkincircle, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_tslash", (void *)_cffi_const_XKB_KEY_tslash, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_twofifths", (void *)_cffi_const_XKB_KEY_twofifths, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_twosubscript", (void *)_cffi_const_XKB_KEY_twosubscript, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_twosuperior", (void *)_cffi_const_XKB_KEY_twosuperior, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_twothirds", (void *)_cffi_const_XKB_KEY_twothirds, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_u", (void *)_cffi_const_XKB_KEY_u, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_uacute", (void *)_cffi_const_XKB_KEY_uacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ubelowdot", (void *)_cffi_const_XKB_KEY_ubelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ubreve", (void *)_cffi_const_XKB_KEY_ubreve, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ucircumflex", (void *)_cffi_const_XKB_KEY_ucircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_udiaeresis", (void *)_cffi_const_XKB_KEY_udiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_udoubleacute", (void *)_cffi_const_XKB_KEY_udoubleacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ugrave", (void *)_cffi_const_XKB_KEY_ugrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_uhook", (void *)_cffi_const_XKB_KEY_uhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_uhorn", (void *)_cffi_const_XKB_KEY_uhorn, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_uhornacute", (void *)_cffi_const_XKB_KEY_uhornacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_uhornbelowdot", (void *)_cffi_const_XKB_KEY_uhornbelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_uhorngrave", (void *)_cffi_const_XKB_KEY_uhorngrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_uhornhook", (void *)_cffi_const_XKB_KEY_uhornhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_uhorntilde", (void *)_cffi_const_XKB_KEY_uhorntilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_umacron", (void *)_cffi_const_XKB_KEY_umacron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_underbar", (void *)_cffi_const_XKB_KEY_underbar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_underscore", (void *)_cffi_const_XKB_KEY_underscore, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_union", (void *)_cffi_const_XKB_KEY_union, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_uogonek", (void *)_cffi_const_XKB_KEY_uogonek, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_uparrow", (void *)_cffi_const_XKB_KEY_uparrow, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_upcaret", (void *)_cffi_const_XKB_KEY_upcaret, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_upleftcorner", (void *)_cffi_const_XKB_KEY_upleftcorner, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_uprightcorner", (void *)_cffi_const_XKB_KEY_uprightcorner, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_upshoe", (void *)_cffi_const_XKB_KEY_upshoe, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_upstile", (void *)_cffi_const_XKB_KEY_upstile, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_uptack", (void *)_cffi_const_XKB_KEY_uptack, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_uring", (void *)_cffi_const_XKB_KEY_uring, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_utilde", (void *)_cffi_const_XKB_KEY_utilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_v", (void *)_cffi_const_XKB_KEY_v, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_variation", (void *)_cffi_const_XKB_KEY_variation, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_vertbar", (void *)_cffi_const_XKB_KEY_vertbar, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_vertconnector", (void *)_cffi_const_XKB_KEY_vertconnector, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_voicedsound", (void *)_cffi_const_XKB_KEY_voicedsound, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_vt", (void *)_cffi_const_XKB_KEY_vt, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_w", (void *)_cffi_const_XKB_KEY_w, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_wacute", (void *)_cffi_const_XKB_KEY_wacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_wcircumflex", (void *)_cffi_const_XKB_KEY_wcircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_wdiaeresis", (void *)_cffi_const_XKB_KEY_wdiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_wgrave", (void *)_cffi_const_XKB_KEY_wgrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_x", (void *)_cffi_const_XKB_KEY_x, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_xabovedot", (void *)_cffi_const_XKB_KEY_xabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_y", (void *)_cffi_const_XKB_KEY_y, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_yacute", (void *)_cffi_const_XKB_KEY_yacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ybelowdot", (void *)_cffi_const_XKB_KEY_ybelowdot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ycircumflex", (void *)_cffi_const_XKB_KEY_ycircumflex, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ydiaeresis", (void *)_cffi_const_XKB_KEY_ydiaeresis, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_yen", (void *)_cffi_const_XKB_KEY_yen, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ygrave", (void *)_cffi_const_XKB_KEY_ygrave, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_yhook", (void *)_cffi_const_XKB_KEY_yhook, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_ytilde", (void *)_cffi_const_XKB_KEY_ytilde, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_z", (void *)_cffi_const_XKB_KEY_z, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_zabovedot", (void *)_cffi_const_XKB_KEY_zabovedot, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_zacute", (void *)_cffi_const_XKB_KEY_zacute, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_zcaron", (void *)_cffi_const_XKB_KEY_zcaron, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_zerosubscript", (void *)_cffi_const_XKB_KEY_zerosubscript, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_zerosuperior", (void *)_cffi_const_XKB_KEY_zerosuperior, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "XKB_KEY_zstroke", (void *)_cffi_const_XKB_KEY_zstroke, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "keyboard_key", (void *)&_cffi_externpy__keyboard_key, _CFFI_OP(_CFFI_OP_EXTERN_PYTHON, 319), (void *)keyboard_key },
  { "output_resolution", (void *)&_cffi_externpy__output_resolution, _CFFI_OP(_CFFI_OP_EXTERN_PYTHON, 348), (void *)output_resolution },
  { "pointer_button", (void *)&_cffi_externpy__pointer_button, _CFFI_OP(_CFFI_OP_EXTERN_PYTHON, 318), (void *)pointer_button },
  { "pointer_motion", (void *)&_cffi_externpy__pointer_motion, _CFFI_OP(_CFFI_OP_EXTERN_PYTHON, 320), (void *)pointer_motion },
  { "view_created", (void *)&_cffi_externpy__view_created, _CFFI_OP(_CFFI_OP_EXTERN_PYTHON, 156), (void *)view_created },
  { "view_destroyed", (void *)&_cffi_externpy__view_destroyed, _CFFI_OP(_CFFI_OP_EXTERN_PYTHON, 287), (void *)view_destroyed },
  { "view_focus", (void *)&_cffi_externpy__view_focus, _CFFI_OP(_CFFI_OP_EXTERN_PYTHON, 290), (void *)view_focus },
  { "view_request_geometry", (void *)&_cffi_externpy__view_request_geometry, _CFFI_OP(_CFFI_OP_EXTERN_PYTHON, 346), (void *)view_request_geometry },
  { "view_request_move", (void *)&_cffi_externpy__view_request_move, _CFFI_OP(_CFFI_OP_EXTERN_PYTHON, 347), (void *)view_request_move },
  { "view_request_resize", (void *)&_cffi_externpy__view_request_resize, _CFFI_OP(_CFFI_OP_EXTERN_PYTHON, 349), (void *)view_request_resize },
  { "wlc_event_loop_add_fd", (void *)_cffi_f_wlc_event_loop_add_fd, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 101), (void *)_cffi_d_wlc_event_loop_add_fd },
  { "wlc_event_loop_add_timer", (void *)_cffi_f_wlc_event_loop_add_timer, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 97), (void *)_cffi_d_wlc_event_loop_add_timer },
  { "wlc_event_source_remove", (void *)_cffi_f_wlc_event_source_remove, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 184), (void *)_cffi_d_wlc_event_source_remove },
  { "wlc_event_source_timer_update", (void *)_cffi_f_wlc_event_source_timer_update, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3), (void *)_cffi_d_wlc_event_source_timer_update },
  { "wlc_exec", (void *)_cffi_f_wlc_exec, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 173), (void *)_cffi_d_wlc_exec },
  { "wlc_get_backend_type", (void *)_cffi_f_wlc_get_backend_type, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_N, 75), (void *)_cffi_d_wlc_get_backend_type },
  { "wlc_get_focused_output", (void *)_cffi_f_wlc_get_focused_output, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_N, 147), (void *)_cffi_d_wlc_get_focused_output },
  { "wlc_get_outputs", (void *)_cffi_f_wlc_get_outputs, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 137), (void *)_cffi_d_wlc_get_outputs },
  { "wlc_handle_get_user_data", (void *)_cffi_f_wlc_handle_get_user_data, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 149), (void *)_cffi_d_wlc_handle_get_user_data },
  { "wlc_handle_set_user_data", (void *)_cffi_f_wlc_handle_set_user_data, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 276), (void *)_cffi_d_wlc_handle_set_user_data },
  { "wlc_init", (void *)_cffi_f_wlc_init, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_N, 70), (void *)_cffi_d_wlc_init },
  { "wlc_keyboard_get_current_keys", (void *)_cffi_f_wlc_keyboard_get_current_keys, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 123), (void *)_cffi_d_wlc_keyboard_get_current_keys },
  { "wlc_keyboard_get_keysym_for_key", (void *)_cffi_f_wlc_keyboard_get_keysym_for_key, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 126), (void *)_cffi_d_wlc_keyboard_get_keysym_for_key },
  { "wlc_keyboard_get_utf32_for_key", (void *)_cffi_f_wlc_keyboard_get_utf32_for_key, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 126), (void *)_cffi_d_wlc_keyboard_get_utf32_for_key },
  { "wlc_keyboard_get_xkb_keymap", (void *)_cffi_f_wlc_keyboard_get_xkb_keymap, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_N, 116), (void *)_cffi_d_wlc_keyboard_get_xkb_keymap },
  { "wlc_keyboard_get_xkb_state", (void *)_cffi_f_wlc_keyboard_get_xkb_state, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_N, 118), (void *)_cffi_d_wlc_keyboard_get_xkb_state },
  { "wlc_log_set_handler", (void *)_cffi_f_wlc_log_set_handler, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 280), (void *)_cffi_d_wlc_log_set_handler },
  { "wlc_output_focus", (void *)_cffi_f_wlc_output_focus, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 193), (void *)_cffi_d_wlc_output_focus },
  { "wlc_output_get_gamma_size", (void *)_cffi_f_wlc_output_get_gamma_size, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 120), (void *)_cffi_d_wlc_output_get_gamma_size },
  { "wlc_output_get_mask", (void *)_cffi_f_wlc_output_get_mask, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 130), (void *)_cffi_d_wlc_output_get_mask },
  { "wlc_output_get_mutable_views", (void *)_cffi_f_wlc_output_get_mutable_views, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 133), (void *)_cffi_d_wlc_output_get_mutable_views },
  { "wlc_output_get_name", (void *)_cffi_f_wlc_output_get_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 72), (void *)_cffi_d_wlc_output_get_name },
  { "wlc_output_get_resolution", (void *)_cffi_f_wlc_output_get_resolution, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 113), (void *)_cffi_d_wlc_output_get_resolution },
  { "wlc_output_get_scale", (void *)_cffi_f_wlc_output_get_scale, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 130), (void *)_cffi_d_wlc_output_get_scale },
  { "wlc_output_get_sleep", (void *)_cffi_f_wlc_output_get_sleep, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 7), (void *)_cffi_d_wlc_output_get_sleep },
  { "wlc_output_get_views", (void *)_cffi_f_wlc_output_get_views, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 140), (void *)_cffi_d_wlc_output_get_views },
  { "wlc_output_get_virtual_resolution", (void *)_cffi_f_wlc_output_get_virtual_resolution, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 113), (void *)_cffi_d_wlc_output_get_virtual_resolution },
  { "wlc_output_set_gamma", (void *)_cffi_f_wlc_output_set_gamma, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 241), (void *)_cffi_d_wlc_output_set_gamma },
  { "wlc_output_set_mask", (void *)_cffi_f_wlc_output_set_mask, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 248), (void *)_cffi_d_wlc_output_set_mask },
  { "wlc_output_set_resolution", (void *)_cffi_f_wlc_output_set_resolution, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 236), (void *)_cffi_d_wlc_output_set_resolution },
  { "wlc_output_set_sleep", (void *)_cffi_f_wlc_output_set_sleep, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 196), (void *)_cffi_d_wlc_output_set_sleep },
  { "wlc_output_set_views", (void *)_cffi_f_wlc_output_set_views, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 65), (void *)_cffi_d_wlc_output_set_views },
  { "wlc_pointer_get_position", (void *)_cffi_f_wlc_pointer_get_position, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 187), (void *)_cffi_d_wlc_pointer_get_position },
  { "wlc_pointer_set_position", (void *)_cffi_f_wlc_pointer_set_position, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 190), (void *)_cffi_d_wlc_pointer_set_position },
  { "wlc_run", (void *)_cffi_f_wlc_run, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_N, 316), (void *)_cffi_d_wlc_run },
  { "wlc_set_compositor_ready_cb", (void *)_cffi_f_wlc_set_compositor_ready_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 313), (void *)_cffi_d_wlc_set_compositor_ready_cb },
  { "wlc_set_compositor_terminate_cb", (void *)_cffi_f_wlc_set_compositor_terminate_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 313), (void *)_cffi_d_wlc_set_compositor_terminate_cb },
  { "wlc_set_input_created_cb", (void *)_cffi_f_wlc_set_input_created_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 152), (void *)_cffi_d_wlc_set_input_created_cb },
  { "wlc_set_input_destroyed_cb", (void *)_cffi_f_wlc_set_input_destroyed_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 283), (void *)_cffi_d_wlc_set_input_destroyed_cb },
  { "wlc_set_keyboard_key_cb", (void *)_cffi_f_wlc_set_keyboard_key_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 164), (void *)_cffi_d_wlc_set_keyboard_key_cb },
  { "wlc_set_output_context_created_cb", (void *)_cffi_f_wlc_set_output_context_created_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 286), (void *)_cffi_d_wlc_set_output_context_created_cb },
  { "wlc_set_output_context_destroyed_cb", (void *)_cffi_f_wlc_set_output_context_destroyed_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 286), (void *)_cffi_d_wlc_set_output_context_destroyed_cb },
  { "wlc_set_output_created_cb", (void *)_cffi_f_wlc_set_output_created_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 155), (void *)_cffi_d_wlc_set_output_created_cb },
  { "wlc_set_output_destroyed_cb", (void *)_cffi_f_wlc_set_output_destroyed_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 286), (void *)_cffi_d_wlc_set_output_destroyed_cb },
  { "wlc_set_output_focus_cb", (void *)_cffi_f_wlc_set_output_focus_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 289), (void *)_cffi_d_wlc_set_output_focus_cb },
  { "wlc_set_output_render_post_cb", (void *)_cffi_f_wlc_set_output_render_post_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 286), (void *)_cffi_d_wlc_set_output_render_post_cb },
  { "wlc_set_output_render_pre_cb", (void *)_cffi_f_wlc_set_output_render_pre_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 286), (void *)_cffi_d_wlc_set_output_render_pre_cb },
  { "wlc_set_output_resolution_cb", (void *)_cffi_f_wlc_set_output_resolution_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 301), (void *)_cffi_d_wlc_set_output_resolution_cb },
  { "wlc_set_pointer_button_cb", (void *)_cffi_f_wlc_set_pointer_button_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 161), (void *)_cffi_d_wlc_set_pointer_button_cb },
  { "wlc_set_pointer_motion_cb", (void *)_cffi_f_wlc_set_pointer_motion_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 170), (void *)_cffi_d_wlc_set_pointer_motion_cb },
  { "wlc_set_pointer_scroll_cb", (void *)_cffi_f_wlc_set_pointer_scroll_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 167), (void *)_cffi_d_wlc_set_pointer_scroll_cb },
  { "wlc_set_touch_cb", (void *)_cffi_f_wlc_set_touch_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 158), (void *)_cffi_d_wlc_set_touch_cb },
  { "wlc_set_view_created_cb", (void *)_cffi_f_wlc_set_view_created_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 155), (void *)_cffi_d_wlc_set_view_created_cb },
  { "wlc_set_view_destroyed_cb", (void *)_cffi_f_wlc_set_view_destroyed_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 286), (void *)_cffi_d_wlc_set_view_destroyed_cb },
  { "wlc_set_view_focus_cb", (void *)_cffi_f_wlc_set_view_focus_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 289), (void *)_cffi_d_wlc_set_view_focus_cb },
  { "wlc_set_view_move_to_output_cb", (void *)_cffi_f_wlc_set_view_move_to_output_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 310), (void *)_cffi_d_wlc_set_view_move_to_output_cb },
  { "wlc_set_view_properties_updated_cb", (void *)_cffi_f_wlc_set_view_properties_updated_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 304), (void *)_cffi_d_wlc_set_view_properties_updated_cb },
  { "wlc_set_view_render_post_cb", (void *)_cffi_f_wlc_set_view_render_post_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 286), (void *)_cffi_d_wlc_set_view_render_post_cb },
  { "wlc_set_view_render_pre_cb", (void *)_cffi_f_wlc_set_view_render_pre_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 286), (void *)_cffi_d_wlc_set_view_render_pre_cb },
  { "wlc_set_view_request_geometry_cb", (void *)_cffi_f_wlc_set_view_request_geometry_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 295), (void *)_cffi_d_wlc_set_view_request_geometry_cb },
  { "wlc_set_view_request_move_cb", (void *)_cffi_f_wlc_set_view_request_move_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 298), (void *)_cffi_d_wlc_set_view_request_move_cb },
  { "wlc_set_view_request_resize_cb", (void *)_cffi_f_wlc_set_view_request_resize_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 307), (void *)_cffi_d_wlc_set_view_request_resize_cb },
  { "wlc_set_view_request_state_cb", (void *)_cffi_f_wlc_set_view_request_state_cb, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 292), (void *)_cffi_d_wlc_set_view_request_state_cb },
  { "wlc_terminate", (void *)_cffi_f_wlc_terminate, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_N, 316), (void *)_cffi_d_wlc_terminate },
  { "wlc_view_bring_above", (void *)_cffi_f_wlc_view_bring_above, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 267), (void *)_cffi_d_wlc_view_bring_above },
  { "wlc_view_bring_to_front", (void *)_cffi_f_wlc_view_bring_to_front, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 193), (void *)_cffi_d_wlc_view_bring_to_front },
  { "wlc_view_close", (void *)_cffi_f_wlc_view_close, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 193), (void *)_cffi_d_wlc_view_close },
  { "wlc_view_focus", (void *)_cffi_f_wlc_view_focus, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 193), (void *)_cffi_d_wlc_view_focus },
  { "wlc_view_get_app_id", (void *)_cffi_f_wlc_view_get_app_id, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 72), (void *)_cffi_d_wlc_view_get_app_id },
  { "wlc_view_get_class", (void *)_cffi_f_wlc_view_get_class, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 72), (void *)_cffi_d_wlc_view_get_class },
  { "wlc_view_get_geometry", (void *)_cffi_f_wlc_view_get_geometry, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 107), (void *)_cffi_d_wlc_view_get_geometry },
  { "wlc_view_get_instance", (void *)_cffi_f_wlc_view_get_instance, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 72), (void *)_cffi_d_wlc_view_get_instance },
  { "wlc_view_get_mask", (void *)_cffi_f_wlc_view_get_mask, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 130), (void *)_cffi_d_wlc_view_get_mask },
  { "wlc_view_get_output", (void *)_cffi_f_wlc_view_get_output, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 144), (void *)_cffi_d_wlc_view_get_output },
  { "wlc_view_get_parent", (void *)_cffi_f_wlc_view_get_parent, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 144), (void *)_cffi_d_wlc_view_get_parent },
  { "wlc_view_get_pid", (void *)_cffi_f_wlc_view_get_pid, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 91), (void *)_cffi_d_wlc_view_get_pid },
  { "wlc_view_get_state", (void *)_cffi_f_wlc_view_get_state, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 130), (void *)_cffi_d_wlc_view_get_state },
  { "wlc_view_get_title", (void *)_cffi_f_wlc_view_get_title, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 72), (void *)_cffi_d_wlc_view_get_title },
  { "wlc_view_get_type", (void *)_cffi_f_wlc_view_get_type, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 130), (void *)_cffi_d_wlc_view_get_type },
  { "wlc_view_get_visible_geometry", (void *)_cffi_f_wlc_view_get_visible_geometry, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 210), (void *)_cffi_d_wlc_view_get_visible_geometry },
  { "wlc_view_positioner_get_anchor", (void *)_cffi_f_wlc_view_positioner_get_anchor, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 77), (void *)_cffi_d_wlc_view_positioner_get_anchor },
  { "wlc_view_positioner_get_anchor_rect", (void *)_cffi_f_wlc_view_positioner_get_anchor_rect, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 107), (void *)_cffi_d_wlc_view_positioner_get_anchor_rect },
  { "wlc_view_positioner_get_constraint_adjustment", (void *)_cffi_f_wlc_view_positioner_get_constraint_adjustment, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 80), (void *)_cffi_d_wlc_view_positioner_get_constraint_adjustment },
  { "wlc_view_positioner_get_gravity", (void *)_cffi_f_wlc_view_positioner_get_gravity, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 83), (void *)_cffi_d_wlc_view_positioner_get_gravity },
  { "wlc_view_positioner_get_offset", (void *)_cffi_f_wlc_view_positioner_get_offset, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 110), (void *)_cffi_d_wlc_view_positioner_get_offset },
  { "wlc_view_positioner_get_size", (void *)_cffi_f_wlc_view_positioner_get_size, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 113), (void *)_cffi_d_wlc_view_positioner_get_size },
  { "wlc_view_send_below", (void *)_cffi_f_wlc_view_send_below, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 267), (void *)_cffi_d_wlc_view_send_below },
  { "wlc_view_send_to_back", (void *)_cffi_f_wlc_view_send_to_back, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 193), (void *)_cffi_d_wlc_view_send_to_back },
  { "wlc_view_set_geometry", (void *)_cffi_f_wlc_view_set_geometry, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 252), (void *)_cffi_d_wlc_view_set_geometry },
  { "wlc_view_set_mask", (void *)_cffi_f_wlc_view_set_mask, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 248), (void *)_cffi_d_wlc_view_set_mask },
  { "wlc_view_set_output", (void *)_cffi_f_wlc_view_set_output, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 267), (void *)_cffi_d_wlc_view_set_output },
  { "wlc_view_set_parent", (void *)_cffi_f_wlc_view_set_parent, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 267), (void *)_cffi_d_wlc_view_set_parent },
  { "wlc_view_set_state", (void *)_cffi_f_wlc_view_set_state, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 200), (void *)_cffi_d_wlc_view_set_state },
  { "wlc_view_set_type", (void *)_cffi_f_wlc_view_set_type, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 205), (void *)_cffi_d_wlc_view_set_type },
};

static const struct _cffi_field_s _cffi_fields[] = {
  { "origin", offsetof(struct wlc_geometry, origin),
              sizeof(((struct wlc_geometry *)0)->origin),
              _CFFI_OP(_CFFI_OP_NOOP, 338) },
  { "size", offsetof(struct wlc_geometry, size),
            sizeof(((struct wlc_geometry *)0)->size),
            _CFFI_OP(_CFFI_OP_NOOP, 339) },
  { "leds", offsetof(struct wlc_modifiers, leds),
            sizeof(((struct wlc_modifiers *)0)->leds),
            _CFFI_OP(_CFFI_OP_NOOP, 12) },
  { "mods", offsetof(struct wlc_modifiers, mods),
            sizeof(((struct wlc_modifiers *)0)->mods),
            _CFFI_OP(_CFFI_OP_NOOP, 12) },
  { "x", offsetof(struct wlc_point, x),
         sizeof(((struct wlc_point *)0)->x),
         _CFFI_OP(_CFFI_OP_NOOP, 5) },
  { "y", offsetof(struct wlc_point, y),
         sizeof(((struct wlc_point *)0)->y),
         _CFFI_OP(_CFFI_OP_NOOP, 5) },
  { "w", offsetof(struct wlc_size, w),
         sizeof(((struct wlc_size *)0)->w),
         _CFFI_OP(_CFFI_OP_NOOP, 12) },
  { "h", offsetof(struct wlc_size, h),
         sizeof(((struct wlc_size *)0)->h),
         _CFFI_OP(_CFFI_OP_NOOP, 12) },
};

static const struct _cffi_struct_union_s _cffi_struct_unions[] = {
  { "libinput_device", 334, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "wlc_event_source", 335, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "wlc_geometry", 336, _CFFI_F_CHECK_FIELDS,
    sizeof(struct wlc_geometry), offsetof(struct _cffi_align_struct_wlc_geometry, y), 0, 2 },
  { "wlc_modifiers", 337, _CFFI_F_CHECK_FIELDS,
    sizeof(struct wlc_modifiers), offsetof(struct _cffi_align_struct_wlc_modifiers, y), 2, 2 },
  { "wlc_point", 338, _CFFI_F_CHECK_FIELDS,
    sizeof(struct wlc_point), offsetof(struct _cffi_align_struct_wlc_point, y), 4, 2 },
  { "wlc_size", 339, _CFFI_F_CHECK_FIELDS,
    sizeof(struct wlc_size), offsetof(struct _cffi_align_struct_wlc_size, y), 6, 2 },
  { "xkb_keymap", 341, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "xkb_state", 343, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
};

static const struct _cffi_enum_s _cffi_enums[] = {
  { "wlc_backend_type", 324, _cffi_prim_int(sizeof(enum wlc_backend_type), ((enum wlc_backend_type)-1) <= 0),
    "WLC_BACKEND_NONE,WLC_BACKEND_DRM,WLC_BACKEND_X11" },
  { "wlc_button_state", 15, _cffi_prim_int(sizeof(enum wlc_button_state), ((enum wlc_button_state)-1) <= 0),
    "WLC_BUTTON_STATE_RELEASED,WLC_BUTTON_STATE_PRESSED" },
  { "wlc_event_bit", 325, _cffi_prim_int(sizeof(enum wlc_event_bit), ((enum wlc_event_bit)-1) <= 0),
    "WLC_EVENT_READABLE,WLC_EVENT_WRITABLE,WLC_EVENT_HANGUP,WLC_EVENT_ERROR" },
  { "wlc_key_state", 23, _cffi_prim_int(sizeof(enum wlc_key_state), ((enum wlc_key_state)-1) <= 0),
    "WLC_KEY_STATE_RELEASED,WLC_KEY_STATE_PRESSED" },
  { "wlc_led_bit", 326, _cffi_prim_int(sizeof(enum wlc_led_bit), ((enum wlc_led_bit)-1) <= 0),
    "WLC_BIT_LED_NUM,WLC_BIT_LED_CAPS,WLC_BIT_LED_SCROLL" },
  { "wlc_log_type", 178, _cffi_prim_int(sizeof(enum wlc_log_type), ((enum wlc_log_type)-1) <= 0),
    "WLC_LOG_INFO,WLC_LOG_WARN,WLC_LOG_ERROR,WLC_LOG_WAYLAND" },
  { "wlc_modifier_bit", 327, _cffi_prim_int(sizeof(enum wlc_modifier_bit), ((enum wlc_modifier_bit)-1) <= 0),
    "WLC_BIT_MOD_SHIFT,WLC_BIT_MOD_CAPS,WLC_BIT_MOD_CTRL,WLC_BIT_MOD_ALT,WLC_BIT_MOD_MOD2,WLC_BIT_MOD_MOD3,WLC_BIT_MOD_LOGO,WLC_BIT_MOD_MOD5" },
  { "wlc_positioner_anchor_bit", 328, _cffi_prim_int(sizeof(enum wlc_positioner_anchor_bit), ((enum wlc_positioner_anchor_bit)-1) <= 0),
    "WLC_BIT_ANCHOR_NONE,WLC_BIT_ANCHOR_TOP,WLC_BIT_ANCHOR_BOTTOM,WLC_BIT_ANCHOR_LEFT,WLC_BIT_ANCHOR_RIGHT" },
  { "wlc_positioner_constraint_adjustment_bit", 329, _cffi_prim_int(sizeof(enum wlc_positioner_constraint_adjustment_bit), ((enum wlc_positioner_constraint_adjustment_bit)-1) <= 0),
    "WLC_BIT_CONSTRAINT_ADJUSTMENT_NONE,WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_X,WLC_BIT_CONSTRAINT_ADJUSTMENT_SLIDE_Y,WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_X,WLC_BIT_CONSTRAINT_ADJUSTMENT_FLIP_Y,WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_X,WLC_BIT_CONSTRAINT_ADJUSTMENT_RESIZE_Y" },
  { "wlc_positioner_gravity_bit", 330, _cffi_prim_int(sizeof(enum wlc_positioner_gravity_bit), ((enum wlc_positioner_gravity_bit)-1) <= 0),
    "WLC_BIT_GRAVITY_NONE,WLC_BIT_GRAVITY_TOP,WLC_BIT_GRAVITY_BOTTOM,WLC_BIT_GRAVITY_LEFT,WLC_BIT_GRAVITY_RIGHT" },
  { "wlc_resize_edge", 331, _cffi_prim_int(sizeof(enum wlc_resize_edge), ((enum wlc_resize_edge)-1) <= 0),
    "WLC_RESIZE_EDGE_NONE,WLC_RESIZE_EDGE_TOP,WLC_RESIZE_EDGE_BOTTOM,WLC_RESIZE_EDGE_LEFT,WLC_RESIZE_EDGE_TOP_LEFT,WLC_RESIZE_EDGE_BOTTOM_LEFT,WLC_RESIZE_EDGE_RIGHT,WLC_RESIZE_EDGE_TOP_RIGHT,WLC_RESIZE_EDGE_BOTTOM_RIGHT" },
  { "wlc_scroll_axis_bit", 332, _cffi_prim_int(sizeof(enum wlc_scroll_axis_bit), ((enum wlc_scroll_axis_bit)-1) <= 0),
    "WLC_SCROLL_AXIS_VERTICAL,WLC_SCROLL_AXIS_HORIZONTAL" },
  { "wlc_touch_type", 29, _cffi_prim_int(sizeof(enum wlc_touch_type), ((enum wlc_touch_type)-1) <= 0),
    "WLC_TOUCH_DOWN,WLC_TOUCH_UP,WLC_TOUCH_MOTION,WLC_TOUCH_FRAME,WLC_TOUCH_CANCEL" },
  { "wlc_view_property_update_bit", 333, _cffi_prim_int(sizeof(enum wlc_view_property_update_bit), ((enum wlc_view_property_update_bit)-1) <= 0),
    "WLC_BIT_PROPERTY_TITLE,WLC_BIT_PROPERTY_CLASS,WLC_BIT_PROPERTY_APP_ID,WLC_BIT_PROPERTY_PID" },
  { "wlc_view_state_bit", 202, _cffi_prim_int(sizeof(enum wlc_view_state_bit), ((enum wlc_view_state_bit)-1) <= 0),
    "WLC_BIT_MAXIMIZED,WLC_BIT_FULLSCREEN,WLC_BIT_RESIZING,WLC_BIT_MOVING,WLC_BIT_ACTIVATED" },
  { "wlc_view_type_bit", 207, _cffi_prim_int(sizeof(enum wlc_view_type_bit), ((enum wlc_view_type_bit)-1) <= 0),
    "WLC_BIT_OVERRIDE_REDIRECT,WLC_BIT_UNMANAGED,WLC_BIT_SPLASH,WLC_BIT_MODAL,WLC_BIT_POPUP" },
};

static const struct _cffi_typename_s _cffi_typenames[] = {
  { "pid_t", 87 },
  { "wlc_handle", 8 },
};

static const struct _cffi_type_context_s _cffi_type_context = {
  _cffi_types,
  _cffi_globals,
  _cffi_fields,
  _cffi_struct_unions,
  _cffi_enums,
  _cffi_typenames,
  3250,  /* num_globals */
  8,  /* num_struct_unions */
  16,  /* num_enums */
  2,  /* num_typenames */
  NULL,  /* no includes */
  351,  /* num_types */
  1,  /* flags */
};

#ifdef __GNUC__
#  pragma GCC visibility push(default)  /* for -fvisibility= */
#endif

#ifdef PYPY_VERSION
PyMODINIT_FUNC
_cffi_pypyinit_wlc(const void *p[])
{
    if (((intptr_t)p[0]) >= 0x0A03) {
        _cffi_call_python_org = (void(*)(struct _cffi_externpy_s *, char *))p[1];
    }
    p[0] = (const void *)0x2601;
    p[1] = &_cffi_type_context;
}
#  ifdef _MSC_VER
     PyMODINIT_FUNC
#  if PY_MAJOR_VERSION >= 3
     PyInit_wlc(void) { return NULL; }
#  else
     initwlc(void) { }
#  endif
#  endif
#elif PY_MAJOR_VERSION >= 3
PyMODINIT_FUNC
PyInit_wlc(void)
{
  return _cffi_init("wlc", 0x2601, &_cffi_type_context);
}
#else
PyMODINIT_FUNC
initwlc(void)
{
  _cffi_init("wlc", 0x2601, &_cffi_type_context);
}
#endif

#ifdef __GNUC__
#  pragma GCC visibility pop
#endif
